
/*

  KLayout Layout Viewer
  Copyright (C) 2006-2007 Matthias Koefferlein

  This program is free software; you can redistribute it and/or modify
  it under the terms of the GNU General Public License as published by
  the Free Software Foundation; either version 2 of the License, or
  (at your option) any later version.

  This program is distributed in the hope that it will be useful,
  but WITHOUT ANY WARRANTY; without even the implied warranty of
  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
  GNU General Public License for more details.

  You should have received a copy of the GNU General Public License
  along with this program; if not, write to the Free Software
  Foundation, Inc., 51 Franklin St, Fifth Floor, Boston, MA  02110-1301  USA

*/


#include "dbUserObject.h"
#include "dbClipboard.h"
#include "tlString.h"
#include "tlAssert.h"
#include "layPlugin.h"
#include "layRenderer.h"
#include "laySnap.h"
#include "layLayoutView.h"
#include "layLayoutViewConfig.h"
#include "layLayoutCanvas.h"
#include "layProperties.h"
#include "layMainWindow.h"
#include "antService.h"
#include "antPropertiesPage.h"
#include "antConfig.h"
#include "ConfigurationDialogImpl.h"

namespace ant
{

// -------------------------------------------------------------
//  Functionality to draw a ruler object

static void
tick_spacings (double d, double min_d, int &minor_ticks, double &ticks)
{
  if (min_d > d) {

    minor_ticks = -1;
    ticks = -1;

  } else {

    const double log10 = log (10.0);

    //  as a safety measure, not too many ticks are created.
    min_d = std::max (min_d, 0.001 * d);

    double l1 = log (min_d) / log10;
    double l0 = floor (l1);
    l1 -= l0;

    if (l1 < 0.3) {
      minor_ticks = 5;
    } else if (l1 < 0.7) {
      minor_ticks = 2;
    } else {
      minor_ticks = 1;
    }

    ticks = exp (log10 * l0) * 10.0;

  }
}

/**
 *  @brief Draw a ruler with the given parameters
 *
 *  @param q1 The first point in pixel space
 *  @param q2 The second point in pixel space
 *  @param length_u The ruler length in micron
 *  @param min_spc_u The minimum tick spacing in micron
 *  @param label The label text to draw
 *  @param sel True to draw ruler in "selected" mode
 *  @param right True to draw the ruler with ticks to the right (as seem from p1 to p2 in transformed space)
 *  @param style The style with which to draw the ruler
 *  @param bitmap The bitmap to draw the ruler on
 */
void 
draw_ruler (const db::DPoint &q1, const db::DPoint &q2, double length_u, double min_spc_u, const std::string &label, bool sel, bool right, ant::Object::style_type style, lay::Bitmap *bitmap)
{
  double arrow_width = 8;
  double arrow_length = 12;
  double sel_width = 2;

  lay::Renderer renderer;
  renderer.reserve_edges (100);

  if (length_u < 1e-5 /*micron*/) {

    if (sel) {
      
      db::DBox b (q1 - db::DPoint (sel_width * 0.5, sel_width * 0.5),
                  q2 + db::DPoint (sel_width * 0.5, sel_width * 0.5));
      
      renderer.insert (db::DEdge (b.lower_left (), b.lower_right ()));
      renderer.insert (db::DEdge (b.lower_right (), b.upper_right ()));
      renderer.insert (db::DEdge (b.upper_right (), b.upper_left ()));
      renderer.insert (db::DEdge (b.upper_left (), b.lower_left ()));
      
      renderer.render_fill (*bitmap);
      renderer.render_contour (*bitmap);
      
    } else {
      renderer.insert (db::DEdge (q1, q1));
      renderer.render_contour (*bitmap);
    }

    renderer.clear ();

    renderer.insert (db::DBox (q1, q1), 
                     label,
                     db::DefaultFont, 
                     db::HAlignLeft, 
                     db::VAlignBottom,
                     db::fixpoint_trans::r0);

    renderer.render_texts (*bitmap);
                     
  } else {

    //  compute the tick distribution
    int tick_length = (style == ant::Object::STY_ruler ? 8 : 0);

    double ticks = -1.0;
    int minor_ticks = -1;
    
    if (tick_length > 0) {
      tick_spacings (length_u, min_spc_u, minor_ticks, ticks);
    }

    //  normal and unit vector
    
    double len = q1.double_distance (q2);
    if (style == ant::Object::STY_arrow_end && len < double (arrow_length) * 1.2) {
      arrow_length = len / 1.2;
      arrow_width = arrow_length * 2.0 / 3.0;
    } else if (style == ant::Object::STY_arrow_both && len < double (arrow_length) * 2.4) {
      arrow_length = len / 2.4;
      arrow_width = arrow_length * 2.0 / 3.0;
    }
    
    db::DPoint qq (q2.y () - q1.y (), q1.x () - q2.x ());
    qq *= 1.0 / len;
    if (!right) {
      qq = -qq;
    }

    db::DPoint qu = q2 - q1;
    qu *= 1.0 / len;
      
    //  produce polygon stuff
      
    if (sel) {
      
      db::DPoint qw = qq * (double (sel_width) * 0.5);
      
      db::DPoint dq1, dq2;
      if (style == ant::Object::STY_arrow_both || style == ant::Object::STY_arrow_start) {
        dq1 = qu * double (arrow_length - 1);
      }
      if (style == ant::Object::STY_arrow_both || style == ant::Object::STY_arrow_end) {
        dq2 = qu * double (-(arrow_length - 1));
      }
      
      renderer.insert (db::DEdge (q1 + dq1 + qw, q2 + dq2 + qw));
      renderer.insert (db::DEdge (q2 + dq2 + qw, q2 + dq2 - qw));
      renderer.insert (db::DEdge (q2 + dq2 - qw, q1 + dq1 - qw));
      renderer.insert (db::DEdge (q1 + dq1 - qw, q1 + dq1 + qw));
      
    }

    if (style == ant::Object::STY_arrow_end || style == ant::Object::STY_arrow_both) {
      renderer.insert (db::DEdge (q2, q2 + qq * double (arrow_width * 0.5) - qu * double (arrow_length)));
      renderer.insert (db::DEdge (q2 + qq * double (arrow_width * 0.5) - qu * double (arrow_length), q2 - qq * double (arrow_width * 0.5) - qu * double (arrow_length)));
      renderer.insert (db::DEdge (q2 - qq * double (arrow_width * 0.5) - qu * double (arrow_length), q2));
    } 
    
    if (style == ant::Object::STY_arrow_start || style == ant::Object::STY_arrow_both) {
      renderer.insert (db::DEdge (q1, q1 + qq * double (arrow_width * 0.5) + qu * double (arrow_length)));
      renderer.insert (db::DEdge (q1 + qq * double (arrow_width * 0.5) + qu * double (arrow_length), q1 - qq * double (arrow_width * 0.5) + qu * double (arrow_length)));
      renderer.insert (db::DEdge (q1 - qq * double (arrow_width * 0.5) + qu * double (arrow_length), q1));
    }

    renderer.render_fill (*bitmap);
    renderer.render_contour (*bitmap);
    renderer.clear ();
    
    //  produce edge and text stuff
    
    if (! sel) {
      renderer.insert (db::DEdge (q1, q2));
    }
    
    //  create three tick vectors in tv_text, tv_short and tv_long
    double tf = double (tick_length);
    db::DPoint tv_text;
    if (style == ant::Object::STY_arrow_start || style == ant::Object::STY_arrow_both || style == ant::Object::STY_arrow_end) {
      tv_text = qq * (arrow_width * 0.5 + 2.0);
    } else {
      tv_text = qq * (tf + 2.0);
    }
    db::DPoint tv_short = qq * tf * 0.5;
    db::DPoint tv_long  = qq * tf;

    bool text_left;
    if (fabs (qq.x ()) > 1e-6) {
      text_left = qq.x () > 0.0;
    } else {
      text_left = q2.x () < q1.x ();
    }
    bool text_bottom;
    if (fabs (qq.y ()) > 1e-6) {
      text_bottom = qq.y () > 0.0;
    } else {
      text_bottom = q1.y () > q2.y ();
    }

    if (tick_length > 0) {
      renderer.insert (db::DEdge (q1, q1 + tv_long));
      renderer.insert (db::DEdge (q2, q2 + tv_long));
    }
    
    renderer.insert (db::DBox (q2 + tv_text, q2 + tv_text), 
                     label,
                     db::DefaultFont, 
                     (text_left ? db::HAlignLeft : db::HAlignRight), 
                     (text_bottom ? db::VAlignBottom : db::VAlignTop),
                     db::fixpoint_trans::r0);
                     
    if (minor_ticks > 0 && ticks > 0.0) {

      db::DPoint q = q2 - q1;

      int nmax = int (floor ((length_u / ticks) * double (minor_ticks) - 1e-6));
      for (int n = 1; n <= nmax; ++n) {

        double r = ticks * double (n) / double (minor_ticks) / length_u;

        db::DPoint qq = q1 + q * r;
        qq = db::DPoint (floor (qq.x () + 0.5), floor (qq.y () + 0.5));

        if (n % minor_ticks == 0) {
          renderer.insert (db::DEdge (qq, qq + tv_long));
        } else {
          renderer.insert (db::DEdge (qq, qq + tv_short));
        }

      }

    }

    renderer.render_contour (*bitmap);
    renderer.render_texts (*bitmap);
  
  }
}

void 
draw_ruler (const ant::Object &ruler, const db::DCplxTrans &trans, bool sel, lay::Bitmap *bitmap)
{
  //  round the starting point, shift both, and round the end point 
  std::pair <db::DPoint, db::DPoint> v = lay::snap (trans * ruler.p1 (), trans * ruler.p2 ());
  db::DPoint q1 = v.first;
  db::DPoint q2 = v.second;
  
  double lu = ruler.p1 ().double_distance (ruler.p2 ());
  int min_tick_spc = 10;  //  min tick spacing in pixel
  double mu = double (min_tick_spc) / trans.ctrans (1.0);
  
  if (ruler.outline () == Object::OL_diag) {
    draw_ruler (q1, q2, lu, mu, ruler.text (), sel, q2.x () < q1.x (), ruler.style (), bitmap);
  }
  if (ruler.outline () == Object::OL_xy || ruler.outline () == Object::OL_diag_xy) {
    bool r = (q2.x () > q1.x ()) ^ (q2.y () < q1.y ());
    if (ruler.outline () == Object::OL_diag_xy) {
      draw_ruler (q1, q2, lu, mu, ruler.text (), sel, !r, ruler.style (), bitmap);
    }
    draw_ruler (q1, db::DPoint (q2.x (), q1.y ()), lu, mu, ruler.text_x (), sel, r, ruler.style (), bitmap);
    draw_ruler (db::DPoint (q2.x (), q1.y ()), q2, lu, mu, ruler.text_y (), sel, r, ruler.style (), bitmap);
  }
  if (ruler.outline () == Object::OL_yx || ruler.outline () == Object::OL_diag_yx) {
    bool r = (q2.x () > q1.x ()) ^ (q2.y () > q1.y ());
    if (ruler.outline () == Object::OL_diag_yx) {
      draw_ruler (q1, q2, lu, mu, ruler.text (), sel, !r, ruler.style (), bitmap);
    }
    draw_ruler (q1, db::DPoint (q1.x (), q2.y ()), lu, mu, ruler.text_x (), sel, r, ruler.style (), bitmap);
    draw_ruler (db::DPoint (q1.x (), q2.y ()), q2, lu, mu, ruler.text_y (), sel, r, ruler.style (), bitmap);
  }
  if (ruler.outline () == Object::OL_box) {
    bool r = (q2.x () > q1.x ()) ^ (q2.y () < q1.y ());
    draw_ruler (q1, db::DPoint (q2.x (), q1.y ()), lu, mu, ruler.text_x (), sel, r, ruler.style (), bitmap);
    draw_ruler (db::DPoint (q2.x (), q1.y ()), q2, lu, mu, ruler.text_y (), sel, r, ruler.style (), bitmap);
    draw_ruler (q1, db::DPoint (q1.x (), q2.y ()), lu, mu, "", sel, !r, ruler.style (), bitmap);
    draw_ruler (db::DPoint (q1.x (), q2.y ()), q2, lu, mu, ruler.text (), sel, !r, ruler.style (), bitmap);
  }
}


// -------------------------------------------------------------

//  search range for select in pixels
unsigned int search_range = 5;   // @@@ make variable?

// -------------------------------------------------------------

class ContourFinder
{
public:
  ContourFinder (const db::DPoint &original, const std::vector <db::DEdge> &cutlines)
    : m_any (false), m_any_exact (false), 
      m_original (original), m_tests (100 /* max. number of tests, TODO: make variable? */),
      mp_layout (0), m_cutlines (cutlines), mp_prop_sel (0)
  {
    //  nothing yet.
  }
  
  void
  find (lay::LayoutView *view, int cv_index, const std::vector<db::CplxTrans> &trans, const db::Layout &layout, const db::Box &region, const db::Cell &cell, unsigned int l,
        const std::set<db::PropertiesRepository::properties_id_type> *prop_sel, bool inv_prop_sel)
  {
    m_region = region; 
    mp_layout = &layout;
    mp_prop_sel = prop_sel;
    m_inv_prop_sel = inv_prop_sel;

    const lay::CellView &cv = view->cellview (cv_index);
    int ctx_levels = cv.specific_path ().size ();

    for (std::vector<db::CplxTrans>::const_iterator t = trans.begin (); t != trans.end () && m_tests > 0; ++t) {
      do_find (view, cv_index, cell, l, view->get_min_hier_levels () - ctx_levels, view->get_max_hier_levels () - ctx_levels, *t * cv.context_trans ());
    }
  }

  const db::DPoint &get_found () const
  {
    if (m_any_exact) {
      return m_closest_exact;
    } else if (m_any) {
      return m_closest;
    } else {
      return m_original;
    }
  }

  bool any () const
  {
    return (m_any || m_any_exact) && m_tests > 0;
  }

private:
  void 
  find_closest_exact (const db::DPoint &p)
  {
    db::DPoint dp (p * mp_layout->dbu ());
    if (! m_any_exact || m_original.distance (dp) < m_original.distance (m_closest_exact)) {
      m_closest_exact = dp;
      m_any_exact = true;
    }
  }

  void 
  find_closest (const db::DPoint &p)
  {
    db::DPoint dp (p * mp_layout->dbu ());
    if (! m_any || m_original.distance (dp) < m_original.distance (m_closest)) {
      m_closest = dp;
      m_any = true;
    }
  }

  void closest (const db::DPoint &p) 
  {
    if (m_cutlines.empty ()) {

      find_closest_exact (p);

    } else {

      //  Test, if there is a crossing between a cut line and the vertical
      //  or horizonal lines through the given point. Take this as the
      //  test point.
      for (std::vector <db::DEdge>::const_iterator cl = m_cutlines.begin (); cl != m_cutlines.end (); ++cl) {
        std::pair<bool, db::DPoint> ret;
        ret = db::DEdge (p, p + db::DPoint (1.0, 0.0)).cut_point (*cl * (1.0 / mp_layout->dbu ()));
        if (ret.first) {
          find_closest_exact (ret.second);
        }
        ret = db::DEdge (p, p + db::DPoint (0.0, 1.0)).cut_point (*cl * (1.0 / mp_layout->dbu ()));
        if (ret.first) {
          find_closest_exact (ret.second);
        }
      }

    }
  }

  void closest (const db::DEdge &e)
  {
    bool any_point = false;

    //  do the checks in dbu space rather than micron space because
    //  the tolerances are set up for this.
    for (std::vector <db::DEdge>::const_iterator cl = m_cutlines.begin (); cl != m_cutlines.end (); ++cl) {
      std::pair<bool, db::DPoint> ret = e.cut_point (*cl * (1.0 / mp_layout->dbu ()));
      if (ret.first) {
        //  if the projection exactly hits the edge and the point
        //  of crossing is inside the search region, take this as
        //  the exact hit. Exact hits have priority over projected
        //  hits.
        if (e.contains (ret.second) && m_region.contains (ret.second)) {
          find_closest_exact (ret.second);
        } else {
          find_closest (ret.second);
        }
        any_point = true;
      }
    }

    if (! any_point) {
    
      //  no certain direction to look into:
      //  compute the projection of the point and if within a 
      //  certain range, take the projected point.

      db::DCoord d = std::max (m_region.width (), m_region.height ());
      db::DPoint v (e.p2 () - e.p1 ());
      db::DPoint n (-v.y (), v.x ());
      double f = d / n.double_distance ();

      db::DPoint e1 (m_original.x () / mp_layout->dbu () - n.x () * f,
                     m_original.y () / mp_layout->dbu () - n.y () * f);
      db::DPoint e2 (m_original.x () / mp_layout->dbu () + n.x () * f,
                     m_original.y () / mp_layout->dbu () + n.y () * f);
      
      double dmin (std::numeric_limits <db::Coord>::min ());
      double dmax (std::numeric_limits <db::Coord>::max ());
      db::DBox dworld (dmin, dmin, dmax, dmax);

      if (dworld.contains (e1) && dworld.contains (e2)) {

        std::pair<bool, db::DPoint> ip = e.intersect_point (db::DEdge (e1, e2));
        if (ip.first) { 
          find_closest (ip.second);
        }

      }

    }

  }

  void
  do_find (lay::LayoutView *view, int cv_index, const db::Cell &cell, unsigned int l, int min_level, int max_level, const db::CplxTrans &t)
  {
    if (max_level <= 0 || ! (t * cell.bbox (l)).touches (m_region) || view->is_cell_hidden (cell.cell_index (), cv_index)) {
      return;
    }

    if (min_level <= 0) {

      const db::Shapes &shapes = cell.shapes (l);

      db::ShapeIterator shape = shapes.begin_touching (db::Box::from_double (t.inverted () * m_region), db::ShapeIterator::Polygons | db::ShapeIterator::Paths | db::ShapeIterator::Boxes, mp_prop_sel, m_inv_prop_sel);
      while (! shape.at_end () && m_tests > 0) {

        --m_tests;

        if (shape->is_polygon ()) {

          for (db::Shape::polygon_edge_iterator e = shape->begin_edge (); ! e.at_end (); ++e) {
            test_edge (t * *e);
          }

        } else if (shape->is_path ()) {

          //  test the "spine"
          db::Shape::point_iterator pt = shape->begin_point (); 
          if (pt != shape->end_point ()) {
            db::Point p (*pt);
            ++pt;
            for (; pt != shape->end_point (); ++pt) {
              test_edge (t * db::Edge (p, *pt));
              p = *pt;
            }
          }

          //  convert to polygon and test those edges
          db::Polygon poly;
          shape->polygon (poly);
          for (db::Polygon::polygon_edge_iterator e = poly.begin_edge (); ! e.at_end (); ++e) {
            test_edge (t * *e);
          }

        } else if (shape->is_box ()) {

          const db::Box &box = shape->box ();

          test_edge (t * db::Edge (box.p1 (), db::Point (box.left (), box.top ())));
          test_edge (t * db::Edge (db::Point (box.left (), box.top ()), box.p2 ()));
          test_edge (t * db::Edge (box.p2 (), db::Point (box.right (), box.bottom ())));
          test_edge (t * db::Edge (db::Point (box.right (), box.bottom ()), box.p1 ()));

        }

        ++shape;

      }

    }

    db::Box touch_box (db::Box::from_double (t.inverted () * m_region)); 
    db::box_convert <db::CellInst> bc (*mp_layout, l);

    db::Cell::touching_iterator inst = cell.begin_touching (touch_box); 
    while (! inst.at_end () && m_tests > 0) {

      const db::CellInstArray &cell_inst = inst->cell_inst ();
      for (db::CellInstArray::iterator p = cell_inst.begin_touching (touch_box, bc); ! p.at_end () && m_tests > 0; ++p) {

        --m_tests;

        do_find (view, cv_index, 
                 mp_layout->cell (cell_inst.object ().cell_index ()), l,
                 min_level - 1, max_level - 1,
                 t * cell_inst.complex_trans (*p));

      }

      ++inst;

    }

  }

  void
  test_edge (const db::DEdge &edg)
  {
//  experimentally disabled this:
//  vertex snap is just annoying when trying to measure the width of simulation contours ..
//  But: when measuring corner-to-corner distances it is very valuable ..
#if 1
    //  we hit the region with this edge: vertex snap.
    if (m_region.contains (edg.p1 ())) {
      closest (edg.p1 ());
    }
    if (m_region.contains (edg.p2 ())) {
      closest (edg.p2 ());
    }
#endif
    
    //  if the edge cuts through the active region: test the
    //  edge as a whole
    if (edg.clipped (m_region).first) {
      closest (edg);
    }
  }

  bool m_any, m_any_exact;
  db::DPoint m_closest, m_closest_exact;
  db::DPoint m_original;
  int m_tests;
  const db::Layout *mp_layout;
  db::DBox m_region;
  std::vector <db::DEdge> m_cutlines;
  const std::set<db::PropertiesRepository::properties_id_type> *mp_prop_sel;
  bool m_inv_prop_sel;
};


// -------------------------------------------------------------

View::View (ant::Service *rulers, const ant::Object *ruler, bool selected)
  : lay::ViewObject (rulers->widget ()), 
    mp_rulers (rulers), m_selected (selected), mp_ruler (ruler)
{
  //  .. nothing else ..
}

View::~View ()
{
  //  .. nothing else ..
}

void 
View::transform_by (const db::DCplxTrans &t)
{
  if (m_trans != t) {
    m_trans = t;
    redraw ();
  }
}

void 
View::ruler (const ant::Object *r)
{
  mp_ruler = r;
  redraw ();
}

void 
View::render (const lay::Viewport &vp, lay::ViewObjectCanvas &canvas) 
{ 
  if (! mp_ruler) {
    return;
  }

  QColor c (mp_rulers->color ());
  if (! c.isValid ()) {
    c = QColor (canvas.foreground_color ().rgb ());
  }

  //  obtain bitmap to render on
  lay::Bitmap *bitmap;
  if (mp_rulers->with_halo ()) {
    std::vector <lay::ViewOp> ops;
    ops.reserve (2);
    //  we use 2 and 3 for the bitmap index. Since selection markers are using 0 and 1, rulers
    //  that are dragged appear in front of them.
    ops.push_back (lay::ViewOp (canvas.background_color ().rgb (), lay::ViewOp::Copy, 0, lay::ViewOp::Rect, 3, 2));
    ops.push_back (lay::ViewOp (c.rgb (), lay::ViewOp::Copy, 0, lay::ViewOp::Rect, 1, 3));
    bitmap = canvas.bitmap (ops, vp.width (), vp.height ());
  } else {
    bitmap = canvas.bitmap (lay::ViewOp (c.rgb ()), vp.width (), vp.height ());
  }

  draw_ruler (*mp_ruler, vp.trans () * m_trans, m_selected, bitmap);
}

// -------------------------------------------------------------
//  ant::Service implementation

Service::Service (db::Manager *manager, lay::LayoutView *view)
  : lay::ViewService (view->view_object_widget ()), 
    lay::Editable (view),
    lay::Plugin (view),
    lay::Drawing (1/*number of bitmaps*/, view->drawings ()),
    db::Object (manager),
    m_halo (true),
    m_snap_mode (ant::Object::Any),
    m_grid (0.001),
    m_grid_snap (false), m_obj_snap (false), m_snap_range (1),
    m_max_number_of_rulers (-1 /*unlimited*/),
    mp_view (view),
    mp_active_ruler (0),
    m_drawing (false), m_current (),
    m_move_mode (MoveNone),
    m_current_template (0)
{ 
  //  .. nothing yet ..
}

Service::~Service ()
{
  for (std::vector<ant::View *>::iterator r = m_rulers.begin (); r != m_rulers.end (); ++r) {
    delete *r;
  }
  m_rulers.clear ();
}

bool 
Service::configure (const std::string &name, const std::string &value)
{
  bool taken = true;

  if (name == cfg_ruler_color) {

    QColor color;
    lay::ColorConverter ().from_string (value, color);

    //  make the color available for the dynamic view objects too.
    if (lay::test_and_set (m_color, color)) {
      widget ()->touch ();
    }

  } else if (name == cfg_ruler_halo) {

    bool halo;
    tl::from_string (value, halo);

    //  make the color available for the dynamic view objects too.
    if (lay::test_and_set (m_halo, halo)) {
      widget ()->touch ();
    }

  } else if (name == cfg_ruler_grid_micron) {

    double g = 0;
    tl::from_string (value, g);
    m_grid = g;
    taken = false; // to let others use the grid too.

  } else if (name == cfg_max_number_of_rulers) {

    int n = -1;
    tl::from_string (value, n);
    if (n != m_max_number_of_rulers) {
      m_max_number_of_rulers = n;
      reduce_rulers (n);
    }

  } else if (name == cfg_ruler_snap_range) {

    int n = 0;
    tl::from_string (value, n);
    m_snap_range = n;

  } else if (name == cfg_ruler_obj_snap) {
    tl::from_string (value, m_obj_snap);
  } else if (name == cfg_ruler_grid_snap) {
    tl::from_string (value, m_grid_snap);
  } else if (name == cfg_ruler_snap_mode) {

    ant::Object::angle_constraint_type sm = ant::Object::Any;
    ACConverter ().from_string (value, sm);
    m_snap_mode = sm;

  } else if (name == cfg_ruler_templates) {
    m_ruler_templates = ant::Template::from_string (value); 
  } else if (name == cfg_current_ruler_template) {
    
    int n = 0;
    tl::from_string (value, n);
    m_current_template = n;

  } else {
    taken = false;
  }

  return taken;

}

const ant::Template &
Service::current_template () const
{
  if (m_current_template >= m_ruler_templates.size ()) {
    static ant::Template s_def_template;
    return s_def_template;
  } else {
    return m_ruler_templates [m_current_template];
  }
}

void 
Service::config_finalize ()
{
  //  the changing of the view ops is done here since it may depend on the 
  //  background color which might be changed by another configure call later.
  std::vector <lay::ViewOp> view_ops;
  if (m_halo) {
    view_ops.push_back (lay::ViewOp (mp_view->background_color ().rgb (), lay::ViewOp::Copy, 0, lay::ViewOp::Rect, 3, 0));
  }
  if (m_color.isValid ()) {
    view_ops.push_back (lay::ViewOp (m_color.rgb (), lay::ViewOp::Copy, 0, lay::ViewOp::Rect, 1, 0));
  } else {
    view_ops.push_back (lay::ViewOp (mp_view->foreground_color ().rgb (), lay::ViewOp::Copy, 0, lay::ViewOp::Rect, 1, 0));
  }
  set_view_ops (view_ops);
}

void 
Service::clear_highlights ()
{
  for (std::vector<ant::View *>::iterator r = m_rulers.begin (); r != m_rulers.end (); ++r) {
    (*r)->visible (false);
  }
}

void 
Service::restore_highlights ()
{
  for (std::vector<ant::View *>::iterator r = m_rulers.begin (); r != m_rulers.end (); ++r) {
    (*r)->visible (true);
  }
}

void 
Service::highlight (unsigned int n)
{
  for (std::vector<ant::View *>::iterator r = m_rulers.begin (); r != m_rulers.end (); ++r) {
    (*r)->visible (n-- == 0);
  }
}

void 
Service::clear_rulers ()
{
  drag_cancel ();
  reduce_rulers (0);
}

void 
Service::drag_cancel () 
{
  if (m_drawing) {

    widget ()->ungrab_mouse (this);

    if (mp_active_ruler) {
      delete mp_active_ruler;
      mp_active_ruler = 0;
    }

    m_drawing = false;

  }
}

void 
Service::insert_ruler (const ant::Object &ruler, bool limit_number)
{
  //  determine the last id
  int idmax = -1;
  for (db::DUserObjectLayer::iterator r = mp_view->annotation_shapes ().begin (db::DUserObject::tag ()); r != mp_view->annotation_shapes ().end (db::DUserObject::tag ()); ++r) {
    const ant::Object *robj = dynamic_cast <const ant::Object *> (r->ptr ());
    if (robj) {
      if (robj->id () > idmax) {
        idmax = robj->id ();
      }
    }
  }

  //  create the ruler from the template
  ant::Object *new_ruler = new ant::Object (ruler);
  new_ruler->id (idmax + 1);
  mp_view->annotation_shapes ().basic_insert (db::DUserObject (new_ruler));

  //  delete surplus rulers
  if (limit_number) {
    reduce_rulers (m_max_number_of_rulers);
  }
}

/**
 *  @brief Helper function to determine which move mode to choose given a certain search box and ant::Object
 */
static bool
dragging_what (const ant::Object *robj, const db::DBox &search_dbox, ant::Service::MoveMode &mode, db::DPoint &p1)
{
  db::DPoint p12, p21;
  bool has_p12 = false, has_p21 = false;
  
  if (robj->outline () == ant::Object::OL_xy || robj->outline () == ant::Object::OL_diag_xy || robj->outline () == ant::Object::OL_box) {
    p12 = db::DPoint (robj->p2 ().x (), robj->p1 ().y ());
    has_p12 = true;
  }
  if (robj->outline () == ant::Object::OL_yx || robj->outline () == ant::Object::OL_diag_yx || robj->outline () == ant::Object::OL_box) {
    p21 = db::DPoint (robj->p1 ().x (), robj->p2 ().y ());
    has_p21 = true;
  }
 
  //  HINT: this was implemented returning a std::pair<MoveMode, db::DPoint>, but 
  //  I was not able to get it to work in gcc 4.1.2 in -O3 mode ...
  
  if (search_dbox.contains (robj->p1 ())) {
    p1 = robj->p2 ();
    mode = ant::Service::MoveP1;
    return true;
  }
  if (search_dbox.contains (robj->p2 ())) {
    p1 = robj->p1 ();
    mode = ant::Service::MoveP2;
    return true;
  }
  if (has_p12 && search_dbox.contains (p12)) {
    p1 = p12;
    mode = ant::Service::MoveP12;
    return true;
  }
  if (has_p21 && search_dbox.contains (p21)) {
    p1 = p21;
    mode = ant::Service::MoveP21;
    return true;
  }
  if (has_p12 && search_dbox.touches (db::DBox (p12, robj->p2 ()))) {
    p1 = db::DPoint (p12.x (), search_dbox.center ().y ());
    mode = ant::Service::MoveP2X;
    return true;
  }
  if (has_p21 && search_dbox.touches (db::DBox (p21, robj->p1 ()))) {
    p1 = db::DPoint (p21.x (), search_dbox.center ().y ());
    mode = ant::Service::MoveP1X;
    return true;
  }
  if (has_p12 && search_dbox.touches (db::DBox (p12, robj->p1 ()))) {
    p1 = db::DPoint (search_dbox.center ().x (), p12.y ());
    mode = ant::Service::MoveP1Y;
    return true;
  }
  if (has_p21 && search_dbox.touches (db::DBox (p21, robj->p2 ()))) {
    p1 = db::DPoint (search_dbox.center ().x (), p21.y ());
    mode = ant::Service::MoveP2Y;
    return true;
  }
  if ((robj->outline () == ant::Object::OL_diag || robj->outline () == ant::Object::OL_diag_xy || robj->outline () == ant::Object::OL_diag_yx)
    && db::DEdge (robj->p1 (), robj->p2 ()).distance_abs (search_dbox.center ()) <= search_dbox.width () * 0.5) {
    p1 = search_dbox.center ();
    mode = ant::Service::MoveRuler;
    return true;
  }
  
  return false;
}

bool 
Service::begin_move (bool sel, const db::DPoint &p)
{
  //  cancel any pending move or drag operations, reset mp_active_ruler
  widget ()->drag_cancel (); // KLUDGE: every service does this to the same service manager

  //  choose move mode
  if (sel) {

    m_move_mode = MoveSelected;
    m_p1 = p;
    m_trans = db::DTrans (-m_p1);

    for (std::vector <ant::View *>::iterator r = m_rulers.begin (); r != m_rulers.end (); ++r) {
      (*r)->thaw ();
    }
    return false;

  } else {
  
    m_move_mode = MoveNone;

    //  compute search box
    double l = double (search_range) / widget ()->mouse_event_trans ().mag ();
    db::DBox search_dbox = db::DBox (p, p).enlarged (db::DPoint (l, l));

    //  box-selection

    db::DUserObjectLayer::touching_iterator r = mp_view->annotation_shapes ().begin_touching (search_dbox, db::DUserObject::tag ());

    while (m_move_mode == MoveNone && ! r.at_end ()) {
      
      const ant::Object *robj = dynamic_cast <const ant::Object *> ((*r).ptr ());
      if (robj) {
        
        if (dragging_what (robj, search_dbox, m_move_mode, m_p1)) {
          
          //  found anything: make the moved ruler the selection
          clear_selection ();
          m_selected.insert (std::make_pair (mp_view->annotation_shapes ().begin (db::DUserObject::tag ()) + r.index (), 0));
          m_current = *robj;
          m_original = m_current; 
          m_rulers.push_back (new ant::View (this, &m_current, true));
          m_rulers.back ()->thaw ();
          return true;
          
        }

      }
      
      ++r;
      
    }

    //  nothing was found
    return false;

  }
}

void
Service::move_transform (const db::DPoint &p, db::fixpoint_trans tr)
{
  if (m_rulers.empty () || m_selected.empty ()) {
    return;
  }

  if (m_move_mode == MoveRuler) {

    db::DPoint dp = p;

    m_original.transform (db::DTrans (m_p1) * db::DTrans (tr) * db::DTrans (-m_p1));
    m_current.transform (db::DTrans (dp) * db::DTrans (tr) * db::DTrans (-dp));

    m_rulers [0]->redraw ();

  } else if (m_move_mode == MoveSelected) {

    m_trans *= db::DTrans (m_p1) * db::DTrans (tr) * db::DTrans (-m_p1);

    for (std::vector<ant::View *>::iterator r = m_rulers.begin (); r != m_rulers.end (); ++r) {
      (*r)->transform_by (m_trans);
    }

  }

}

void 
Service::move (const db::DPoint &p)
{
  if (m_rulers.empty () || m_selected.empty ()) {
    return;
  }

  if (m_move_mode == MoveP1) {
    
    m_current.p1 (snap2 (m_p1, p, &m_current).second);
    m_rulers [0]->redraw ();

  } else if (m_move_mode == MoveP2) {
    
    m_current.p2 (snap2 (m_p1, p, &m_current).second);
    m_rulers [0]->redraw ();

  } else if (m_move_mode == MoveP12) {
    
    db::DPoint p12 = snap2 (m_p1, p, &m_current).second;
    m_current.p1 (db::DPoint (m_current.p1 ().x(), p12.y ()));
    m_current.p2 (db::DPoint (p12.x (), m_current.p2 ().y ()));
    m_rulers [0]->redraw ();

  } else if (m_move_mode == MoveP21) {
    
    db::DPoint p21 = snap2 (m_p1, p, &m_current).second;
    m_current.p1 (db::DPoint (p21.x (), m_current.p1 ().y ()));
    m_current.p2 (db::DPoint (m_current.p2 ().x(), p21.y ()));
    m_rulers [0]->redraw ();

  } else if (m_move_mode == MoveP1X) {
    
    db::DPoint pc = snap2 (m_p1, p, &m_current).second;
    m_current.p1 (db::DPoint (pc.x (), m_current.p1 ().y ()));
    m_rulers [0]->redraw ();

  } else if (m_move_mode == MoveP2X) {
    
    db::DPoint pc = snap2 (m_p1, p, &m_current).second;
    m_current.p2 (db::DPoint (pc.x (), m_current.p2 ().y ()));
    m_rulers [0]->redraw ();

  } else if (m_move_mode == MoveP1Y) {
    
    db::DPoint pc = snap2 (m_p1, p, &m_current).second;
    m_current.p1 (db::DPoint (m_current.p1 ().x (), pc.y ()));
    m_rulers [0]->redraw ();

  } else if (m_move_mode == MoveP2Y) {
    
    db::DPoint pc = snap2 (m_p1, p, &m_current).second;
    m_current.p2 (db::DPoint (m_current.p2 ().x (), pc.y ()));
    m_rulers [0]->redraw ();

  } else if (m_move_mode == MoveRuler) {

    //  try two ways of snapping
    db::DPoint dp = p - m_p1;

    db::DPoint p1 = dp + m_original.p1 ();
    db::DPoint p2 = dp + m_original.p2 ();

    std::pair<bool, db::DPoint> r1 = snap1 (p1, m_obj_snap && m_original.snap ());
    db::DPoint q1 = r1.second;
    std::pair<bool, db::DPoint> r2 = snap1 (p2, m_obj_snap && m_original.snap ());
    db::DPoint q2 = r2.second;

    if ((!r2.first && r1.first) || ((r1.first || (!r1.first && !r2.first)) && q1.distance (p1) < q2.distance (p2))) {
      q2 = q1 + (m_original.p2 () - m_original.p1 ());
    } else {
      q1 = q2 + (m_original.p1 () - m_original.p2 ());
    }

    m_current.p1 (q1);
    m_current.p2 (q2);

    m_rulers [0]->redraw ();

  } else if (m_move_mode == MoveSelected) {

    db::DPoint dp = p - m_trans (m_p1);
    //  round the drag distance to grid if required: this is the least we can do in this case
    if (m_grid_snap) {
      dp = db::DPoint (lay::snap (dp.x (), m_grid), lay::snap (dp.y (), m_grid));
    } 

    m_trans = db::DTrans (dp) * m_trans;

    for (std::vector<ant::View *>::iterator r = m_rulers.begin (); r != m_rulers.end (); ++r) {
      (*r)->transform_by (m_trans);
    }

  }
}

void 
Service::end_move (const db::DPoint &p)
{
  if (! m_rulers.empty () && ! m_selected.empty ()) {

    if (m_move_mode == MoveSelected) {

      //  replace the rulers that were moved:
      for (std::map<obj_iterator, unsigned int>::const_iterator s = m_selected.begin (); s != m_selected.end (); ++s) {

        const ant::Object *robj = dynamic_cast<const ant::Object *> (s->first->ptr ());

        //  compute moved object and replace
        ant::Object *rnew = new ant::Object (*robj);
        rnew->transform (m_trans);
        mp_view->annotation_shapes ().replace (&*s->first, db::DUserObject (rnew));

      }

      //  and make selection "visible"
      selection_to_view ();

    } else if (m_move_mode != MoveNone) {

      //  replace the ruler that was moved
      mp_view->annotation_shapes ().replace (&*m_selected.begin ()->first, db::DUserObject (new ant::Object (m_current)));

      //  clear the selection (that was artifically created before)
      clear_selection ();

    }

  }

  //  termine the operation
  m_move_mode = MoveNone;

}

void
Service::selection_to_view ()
{
  //  the selection objects need to be recreated since we destroyed the old rulers
  for (std::vector<ant::View *>::iterator r = m_rulers.begin (); r != m_rulers.end (); ++r) {
    delete *r;
  }
  m_rulers.clear ();
  m_rulers.reserve (m_selected.size ());
  for (std::map<obj_iterator, unsigned int>::iterator r = m_selected.begin (); r != m_selected.end (); ++r) {
    r->second = m_rulers.size ();
    const ant::Object *robj = dynamic_cast<const ant::Object *> (r->first->ptr ());
    m_rulers.push_back (new ant::View (this, robj, true /*selected*/));
  }
}

db::DBox 
Service::selection_bbox ()
{
  db::DBox box;
  for (std::map<obj_iterator, unsigned int>::iterator r = m_selected.begin (); r != m_selected.end (); ++r) {
    const ant::Object *robj = dynamic_cast<const ant::Object *> (r->first->ptr ());
    box += robj->box ();
  }
  return box;
}

void 
Service::transform (const db::DCplxTrans &trans)
{
  //  replace the rulers that were transformed:
  for (std::map<obj_iterator, unsigned int>::const_iterator s = m_selected.begin (); s != m_selected.end (); ++s) {

    const ant::Object *robj = dynamic_cast<const ant::Object *> (s->first->ptr ());

    //  compute transformed object and replace
    ant::Object *rnew = new ant::Object (*robj);
    rnew->transform (trans);
    mp_view->annotation_shapes ().replace (&*s->first, db::DUserObject (rnew));

  }

  selection_to_view ();
}

void 
Service::edit_cancel () 
{
  //  Cancel any move operation
  if (m_move_mode != MoveNone) {

    clear_selection ();
    m_move_mode = MoveNone;

  }
}

bool 
Service::mouse_press_event (const db::DPoint &p, unsigned int buttons, bool prio) 
{
  return mouse_click_event (p, buttons, prio);
}

bool 
Service::mouse_click_event (const db::DPoint &p, unsigned int buttons, bool prio) 
{
  if (prio && (buttons & lay::LeftButton) != 0) {

    if (! m_drawing) {

      //  cancel any edit operations so far
      m_move_mode = MoveNone;

      //  reset selection
      clear_selection ();
         
      //  set the maximum number of rulers minus 1 to account for the new ruler
      //  and clear surplus rulers
      reduce_rulers (m_max_number_of_rulers - 1);

      m_p1 = snap1 (p, m_obj_snap && current_template ().snap ()).second;

      //  create and start dragging the ruler
      
      m_current = ant::Object (m_p1, m_p1, 0, current_template ());
      
      if (mp_active_ruler) {
        delete mp_active_ruler;
      }
      mp_active_ruler = new ant::View (this, &m_current, false /*not selected*/);
      mp_active_ruler->thaw ();
      m_drawing = true;

      widget ()->grab_mouse (this, false);

    } else {

      //  create the ruler object

      //  begin the transaction
      tl_assert (! manager ()->transacting ());
      manager ()->transaction (tl::translate ("Create ruler")); 

      insert_ruler (ant::Object (m_current.p1 (), m_current.p2 (), 0, current_template ()), true);

      //  stop dragging
      drag_cancel ();

      //  end the transaction
      manager ()->commit ();
      
    }

    return true;

  }

  return false;
}

bool 
Service::mouse_move_event (const db::DPoint &p, bool prio) 
{
  if (m_drawing && prio) {
    m_current.p2 (snap2 (m_p1, p, mp_active_ruler->ruler ()).second);
    mp_active_ruler->redraw ();
  }
  return false;
}

void 
Service::deactivated ()
{
  drag_cancel ();
}

std::pair <bool, db::DPoint>
Service::snap1 (const db::DPoint &pt, bool obj_snap, const std::vector <db::DEdge> &cutlines)
{
  db::DPoint dp (pt);

  ContourFinder finder (dp, cutlines);

  if (obj_snap && mp_view) {

    for (lay::LayerPropertiesConstIterator l = mp_view->begin_layers (); ! l.at_end (); ++l) {

      if (l->visible (true /*real*/)) {

        const lay::CellView &cv = mp_view->cellview (l->cellview_index ());
        if (cv.is_valid ()) {

          double px = dp.x () / cv->layout ().dbu ();
          double py = dp.y () / cv->layout ().dbu ();
          double dd = std::max (0.0, widget ()->mouse_event_trans ().inverted ().ctrans (m_snap_range) / cv->layout ().dbu ());
          
          double dmin (std::numeric_limits <db::Coord>::min ());
          double dmax (std::numeric_limits <db::Coord>::max ());
          db::DBox dworld (dmin, dmin, dmax, dmax);

          db::DBox dregion = dworld & db::DBox (px - dd, py - dd, px + dd, py + dd);

          if (! dregion.empty ()) {
            finder.find (mp_view, l->cellview_index (), l->trans (), cv->layout (), 
                         db::Box (db::Point::from_double (dregion.p1 ()), db::Point::from_double (dregion.p2 ())), 
                         *cv.cell (), 
                         l->layer_index (), 
                         &l->prop_sel (), l->inverse_prop_sel ()); 
          }

        }

      }

    }

  } 

  //  in grid snap mode, for the "object free" analysis snap to the grid now
  if (m_grid_snap) {
    dp = db::DPoint (lay::snap (dp.x (), m_grid), lay::snap (dp.y (), m_grid));
  }

  //  try to project the point onto the cutlines
  db::DPoint closest;
  bool anyp = false;

  for (std::vector<db::DEdge>::const_iterator c = cutlines.begin (); c != cutlines.end (); ++c) {
    std::pair<bool, db::DPoint> r = c->cut_point (db::DEdge (dp, dp + db::DPoint (-c->dy (), c->dx ())));
    if (r.first) {
      if (! anyp || dp.sq_distance (r.second) < dp.sq_distance (closest)) {
        anyp = true;
        closest = r.second;
      }
    }
  }

  if (finder.any () && anyp) {
    //  if both the projection and the finder are sucessful, decide by a heuristic criterion which to take
    //  (the projection gets a penalty (of the snap range) a counts less than the finder's choice)
    //  This avoids extreme distortions of the ruler due to projection on long edges.
    double penalty = widget ()->mouse_event_trans ().inverted ().ctrans (m_snap_range);
    if ((dp.distance (closest) + penalty) * 5.0 < dp.distance (finder.get_found ())) {
      return std::make_pair (false, closest);
    } else {
      return std::make_pair (true, finder.get_found ());
    }
  } else if (finder.any ()) {
    return std::make_pair (true, finder.get_found ());
  } else if (anyp) {
    return std::make_pair (false, closest);
  } else {
    return std::make_pair (false, dp);
  }
}

std::pair <bool, db::DPoint>
Service::snap2 (const db::DPoint &p1, const db::DPoint &p2, const ant::Object *obj)
{
  ant::Object::angle_constraint_type snap_mode = obj->angle_constraint () == ant::Object::Global ? m_snap_mode : obj->angle_constraint ();
  bool obj_snap = m_obj_snap && obj->snap ();

  std::vector <db::DEdge> cutlines;

  //  set up cutlines depending on the mode
  if (snap_mode == ant::Object::Ortho) {
    cutlines.reserve (2);
    cutlines.push_back (db::DEdge (p1, p1 + db::DPoint (0.0, 1.0)));
    cutlines.push_back (db::DEdge (p1, p1 + db::DPoint (1.0, 0.0)));
  } else if (snap_mode == ant::Object::Horizontal) {
    cutlines.push_back (db::DEdge (p1, p1 + db::DPoint (1.0, 0.0)));
  } else if (snap_mode == ant::Object::Vertical) {
    cutlines.push_back (db::DEdge (p1, p1 + db::DPoint (0.0, 1.0)));
  } else if (snap_mode == ant::Object::Diagonal) {
    cutlines.reserve (4);
    cutlines.push_back (db::DEdge (p1, p1 + db::DPoint (0.0, 1.0)));
    cutlines.push_back (db::DEdge (p1, p1 + db::DPoint (1.0, 0.0)));
    cutlines.push_back (db::DEdge (p1, p1 + db::DPoint (1.0, 1.0)));
    cutlines.push_back (db::DEdge (p1, p1 + db::DPoint (1.0, -1.0)));
  } 

  return snap1 (p2, obj_snap, cutlines); 
}

struct RulerIdComp 
{
  bool operator() (const db::DUserObjectLayer::iterator &a, const db::DUserObjectLayer::iterator &b) const
  {
    return dynamic_cast<const ant::Object *>(a->ptr ())->id () < dynamic_cast<const ant::Object *>(b->ptr ())->id ();
  }
};

void
Service::reduce_rulers (int num)
{
  db::DUserObjectLayer::iterator rfrom = mp_view->annotation_shapes ().begin (db::DUserObject::tag ()); 
  db::DUserObjectLayer::iterator rto = mp_view->annotation_shapes ().end (db::DUserObject::tag ());

  size_t n = std::distance (rfrom, rto);
  if (num >= 0 && int (n) > num) {

    //  clear selection
    clear_selection ();

    //  extract all rulers and other objects

    std::vector <db::DUserObjectLayer::iterator> positions;
    positions.reserve (n);

    for (db::DUserObjectLayer::iterator r = rfrom; r != rto; ++r) {
      const ant::Object *robj = dynamic_cast <const ant::Object *> (r->ptr ());
      if (robj) {
        positions.push_back (r);
      }
    }

    //  sort so we find the ones that are too old, remove them and sort the 
    //  remaining positions
    std::sort (positions.begin (), positions.end (), RulerIdComp ());
    positions.erase (positions.begin () + (positions.size () - num), positions.end ());
    std::sort (positions.begin (), positions.end ());

    //  now we can erase these positions
    mp_view->annotation_shapes ().erase_positions (db::DUserObject::tag (), positions.begin (), positions.end ());

  }

}

void 
Service::cut ()
{
  if (has_selection ()) {

    //  copy & delete the selected rulers
    copy_selected ();
    del_selected ();

  }
}

void 
Service::copy ()
{
  //  copy the selected rulers
  copy_selected ();
}

void
Service::copy_selected ()
{
  //  extract all selected rulers and paste in "micron" space
  for (std::map<obj_iterator, unsigned int>::iterator r = m_selected.begin (); r != m_selected.end (); ++r) {
    r->second = m_rulers.size ();
    const ant::Object *robj = dynamic_cast<const ant::Object *> (r->first->ptr ());
    db::Clipboard::instance () += new db::ClipboardValue<ant::Object> (*robj);
  }
}

void 
Service::paste ()
{
  if (db::Clipboard::instance ().begin () != db::Clipboard::instance ().end ()) {

    //  determine the last id
    int idmax = -1;
    for (db::DUserObjectLayer::iterator r = mp_view->annotation_shapes ().begin (db::DUserObject::tag ()); r != mp_view->annotation_shapes ().end (db::DUserObject::tag ()); ++r) {
      const ant::Object *robj = dynamic_cast <const ant::Object *> (r->ptr ());
      if (robj && robj->id () > idmax) {
        idmax = robj->id ();
      }
    }

    bool any_added = false;
    for (db::Clipboard::iterator c = db::Clipboard::instance ().begin (); c != db::Clipboard::instance ().end (); ++c) {
      const db::ClipboardValue<ant::Object> *value = dynamic_cast<const db::ClipboardValue<ant::Object> *> (*c);
      if (value) {
        ant::Object *ruler = new ant::Object (value->get ());
        ruler->id (++idmax);
        mp_view->annotation_shapes ().basic_insert (db::DUserObject (ruler));
        any_added = true;
      }
    }

  }

}

void 
Service::del ()
{
  if (has_selection ()) {

    //  delete the selected rulers
    del_selected ();

  }
}

void
Service::del_selected ()
{
  //  positions will hold a set of iterators that are to be erased
  std::vector <db::DUserObjectLayer::iterator> positions;
  positions.reserve (m_selected.size ());
  for (std::map<obj_iterator, unsigned int>::iterator r = m_selected.begin (); r != m_selected.end (); ++r) {
    positions.push_back (r->first);
  }

  //  clear selection
  clear_selection ();

  //  erase all and insert the ones that we want to keep 
  std::sort (positions.begin (), positions.end ());
  mp_view->annotation_shapes ().erase_positions (db::DUserObject::tag (), positions.begin (), positions.end ());
}

bool 
Service::has_selection ()
{
  return ! m_selected.empty ();
}

bool 
Service::select (obj_iterator obj, lay::Editable::SelectionMode mode)
{
  if (mode == lay::Editable::Replace || mode == lay::Editable::Add) {
    //  select
    if (m_selected.find (obj) == m_selected.end ()) {
      m_selected.insert (std::make_pair (obj, 0));
      return true;
    }
  } else if (mode == lay::Editable::Reset) {
    //  unselect
    if (m_selected.find (obj) != m_selected.end ()) {
      m_selected.erase (obj);
      return true;
    }
  } else {
    //  invert selection
    if (m_selected.find (obj) != m_selected.end ()) {
      m_selected.erase (obj);
    } else {
      m_selected.insert (std::make_pair (obj, 0));
    }
    return true;
  }
  return false;
}

void
Service::clear_selection ()
{
  select (db::DBox (), lay::Editable::Reset);
}

static bool 
is_selected (const ant::Object &ruler, const db::DPoint &pos, double enl, double &distance) 
{
  db::DBox b (ruler.p1 (), ruler.p2 ());

  //  enlarge this box by some pixels
  b.enlarge (db::DPoint (enl, enl));

  if (! b.contains (pos)) {
    return false;
  }
  
  db::DEdge edges[4];
  unsigned int nedges = 0;
  
  if (ruler.outline () == ant::Object::OL_diag || 
      ruler.outline () == ant::Object::OL_diag_xy ||
      ruler.outline () == ant::Object::OL_diag_yx) {
    edges [nedges++] = db::DEdge (ruler.p1 (), ruler.p2 ());
  }
  if (ruler.outline () == ant::Object::OL_xy ||
      ruler.outline () == ant::Object::OL_diag_xy ||     
      ruler.outline () == ant::Object::OL_box) {
    edges [nedges++] = db::DEdge (ruler.p1 (), db::DPoint (ruler.p2 ().x (), ruler.p1 ().y ()));
    edges [nedges++] = db::DEdge (db::DPoint (ruler.p2 ().x (), ruler.p1 ().y ()), ruler.p2 ());
  }
  if (ruler.outline () == ant::Object::OL_yx ||
      ruler.outline () == ant::Object::OL_diag_yx ||     
      ruler.outline () == ant::Object::OL_box) {
    edges [nedges++] = db::DEdge (ruler.p1 (), db::DPoint (ruler.p1 ().x (), ruler.p2 ().y ()));
    edges [nedges++] = db::DEdge (db::DPoint (ruler.p1 ().x (), ruler.p2 ().y ()), ruler.p2 ());
  }
    
  for (unsigned int i = 0; i < nedges; ++i) {
    double d = edges [i].distance_abs (pos);
    if (d <= enl) {
      distance = d;
      return true;
    } 
  }  

  return false;
}

static bool 
is_selected (const ant::Object &ruler, const db::DBox &box, double enl)
{
  return (box.contains (ruler.p1 ()) && box.contains (ruler.p2 ()));
}

double
Service::click_proximity (const db::DPoint &pos)
{
  //  compute search box
  double l = double (search_range) / widget ()->mouse_event_trans ().mag ();
  db::DBox search_dbox = db::DBox (pos, pos).enlarged (db::DPoint (l, l));

  //  point selection: look for the "closest" ruler
  double dmin = std::numeric_limits <double>::max ();
  bool any_found = false;

  db::DUserObjectLayer::touching_iterator r = mp_view->annotation_shapes ().begin_touching (search_dbox, db::DUserObject::tag ());
  while (! r.at_end ()) {
    const ant::Object *robj = dynamic_cast<const ant::Object *> ((*r).ptr ());
    if (robj) {
      double d;
      if (is_selected (*robj, pos, l, d)) {
        if (! any_found || d < dmin) {
          dmin = d;
        }
        any_found = true;
      }
    }
    ++r;
  }

  //  return the proximity value
  if (any_found) {
    return dmin;
  } else {
    return lay::Editable::click_proximity (pos); 
  } 
}

bool
Service::select (const db::DBox &box, lay::Editable::SelectionMode mode)
{
  bool needs_update = false;
  bool any_selected = false;

  //  clear before unless "add" is selected
  if (mode == lay::Editable::Replace) {
    if (! m_selected.empty ()) {
      m_selected.clear ();
      needs_update = true;
    }
  }

  if (box.empty ()) {

    //  unconditional selection
    if (mode == lay::Editable::Reset) {
      if (! m_selected.empty ()) {
        m_selected.clear ();
        needs_update = true;
      }
    } else {

      db::DUserObjectLayer::iterator rfrom = mp_view->annotation_shapes ().begin (db::DUserObject::tag ()); 
      db::DUserObjectLayer::iterator rto = mp_view->annotation_shapes ().end (db::DUserObject::tag ());

      //  extract all rulers
      for (db::DUserObjectLayer::iterator r = rfrom; r != rto; ++r) {
        const ant::Object *robj = dynamic_cast<const ant::Object *> ((*r).ptr ());
        if (robj) {
          any_selected = true;
          if (select (r, mode)) {
            needs_update = true;
          }
        }
      }
    }

  } else {

    //  compute search box
    double l = double (search_range) / widget ()->mouse_event_trans ().mag ();
    db::DBox search_dbox = box.enlarged (db::DPoint (l, l));

    if (! box.is_point ()) {

      //  box-selection
      db::DUserObjectLayer::touching_iterator r = mp_view->annotation_shapes ().begin_touching (search_dbox, db::DUserObject::tag ());
      while (! r.at_end ()) {
        const ant::Object *robj = dynamic_cast<const ant::Object *> ((*r).ptr ());
        if (robj) {
          if (is_selected (*robj, box, l)) {
            any_selected = true;
            if (select (mp_view->annotation_shapes ().begin (db::DUserObject::tag ()) + r.index (), mode)) {
              needs_update = true;
            }
          }
        }
        ++r;
      }

    } else {

      //  point selection: look for the "closest" ruler
      double dmin = std::numeric_limits <double>::max ();

      db::DUserObjectLayer::touching_iterator r = mp_view->annotation_shapes ().begin_touching (search_dbox, db::DUserObject::tag ());
      db::DUserObjectLayer::touching_iterator rmin (r);
      while (! r.at_end ()) {
        const ant::Object *robj = dynamic_cast<const ant::Object *> ((*r).ptr ());
        if (robj) {
          double d;
          if (is_selected (*robj, box.p1 (), l, d)) {
            if (! any_selected || d < dmin) {
              rmin = r;
              dmin = d;
            }
            any_selected = true;
          }
        }
        ++r;
      }

      //  select the one that was found
      if (any_selected) {
        select (mp_view->annotation_shapes ().begin (db::DUserObject::tag ()) + rmin.index (), mode);
        needs_update = true;
      }

    }

  }

  //  if required, update the list of ruler objects to display the selection
  if (needs_update) {
    selection_to_view ();
  }

  //  return true if at least one element was selected
  return any_selected;
}

lay::PropertiesPage *
Service::properties_page (QWidget *parent)
{
  return new PropertiesPage (this, parent);
}

void 
Service::get_selection (std::vector <obj_iterator> &sel) const
{
  sel.clear ();
  sel.reserve (m_selected.size ());

  //  positions will hold a set of iterators that are to be erased
  for (std::map<obj_iterator, unsigned int>::const_iterator r = m_selected.begin (); r != m_selected.end (); ++r) {
    sel.push_back (r->first);
  }
}

void
Service::change_ruler (obj_iterator pos, const ant::Object &to)
{
  //  replace the object
  mp_view->annotation_shapes ().replace (&*pos, db::DUserObject (new ant::Object (to)));

  //  and make selection "visible"
  selection_to_view ();
}

void 
Service::paint_on_bitmaps (const db::CplxTrans &trans,
                           const std::vector <lay::Bitmap *> &bitmaps)
{
  if (bitmaps.empty ()) {
    return;
  }

  db::Coord lim = std::numeric_limits<db::Coord>::max ();
  db::DBox world (trans * db::Box (db::Point (-lim, -lim), db::Point (lim, lim)));
  db::DBox vp = trans.inverted () * (world & db::DBox (db::DPoint (0.0, 0.0), db::DPoint (bitmaps.front ()->width (), bitmaps.front ()->height ())));

  lay::Renderer renderer;

  db::DUserObject::tag user_object_tag = db::DUserObject::tag ();
  db::DUserObjectLayer::touching_iterator user_object = mp_view->annotation_shapes ().begin_touching (vp, user_object_tag);
  while (! user_object.at_end ()) {
    const ant::Object *ruler = dynamic_cast <const ant::Object *> ((*user_object).ptr ());
    draw_ruler (*ruler, db::DCplxTrans (trans), false /*not selected*/, bitmaps.front ());
    ++user_object;
  }
}

AnnotationIterator
Service::begin_annotations () const
{
  return AnnotationIterator (mp_view->annotation_shapes ().begin (db::DUserObject::tag ()), mp_view->annotation_shapes ().end (db::DUserObject::tag ()));
}

void 
Service::menu_activated (const std::string &symbol)
{
  if (symbol == "ant::clear_all_rulers") {
    manager ()->transaction (tl::translate ("Clear all rulers")); 
    clear_rulers ();
    manager ()->commit ();
  } else if (symbol == "ant::configure") {
    ConfigurationDialogImpl config_dialog (&lay::MainWindow::instance (), &lay::MainWindow::instance (), "ant::Plugin");
    config_dialog.exec ();
  } else {
    lay::Plugin::menu_activated (symbol);
  }
}

// -------------------------------------------------------------

} // namespace ant


