
/*

  KLayout Layout Viewer
  Copyright (C) 2006-2007 Matthias Koefferlein

  This program is free software; you can redistribute it and/or modify
  it under the terms of the GNU General Public License as published by
  the Free Software Foundation; either version 2 of the License, or
  (at your option) any later version.

  This program is distributed in the hope that it will be useful,
  but WITHOUT ANY WARRANTY; without even the implied warranty of
  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
  GNU General Public License for more details.

  You should have received a copy of the GNU General Public License
  along with this program; if not, write to the Free Software
  Foundation, Inc., 51 Franklin St, Fifth Floor, Boston, MA  02110-1301  USA

*/



#ifndef HDR_dbCell
#define HDR_dbCell

#include <dbInstances.h>
#include <dbObject.h>
#include <dbMemStatistics.h>
#include <dbBox.h>
#include <dbBoxTree.h>
#include <dbBoxConvert.h>
#include <dbShape.h>
#include <dbShapes.h>
#include <dbCellInst.h>
#include <dbArray.h>
#include <tlVector.h>
#include <tlAlgorithm.h>

#include <map>
#include <set>

namespace db
{

template <class C> class cell_list;
template <class C> class cell_list_const_iterator;
template <class C> class cell_list_iterator;

template <class Coord> class generic_repository;

/**
 *  @brief The cell object
 * 
 *  A cell object consists of a set of shape containers (called layers),
 *  a set of child cell instances and auxiliary informations such as
 *  the parent instance list.
 *  A cell is identified through an index given to the cell upon instantiation.
 *  The cell index is valid in the context of a cell graph object which
 *  must issue the cell index.
 */

template <class Graph>
class cell
  : public db::Object
{
public:
  typedef Graph graph_type;
  typedef typename Graph::box_type box_type;
  typedef tl::vector<box_type> box_vector;
  typedef typename box_type::coord_type coord_type;
  typedef db::cell_inst<graph_type> cell_inst_type;
  typedef db::simple_trans<coord_type> array_trans;
  typedef db::array<cell_inst_type, array_trans> cell_inst_array_type;
  typedef typename Graph::cell_index_type cell_index_type;
  typedef db::shapes<coord_type> shapes_type;
  typedef typename db::shapes<coord_type>::shape_iterator shape_iterator;
  typedef tl::vector<db::shapes<coord_type> *> shapes_vector;
  typedef db::instances<graph_type> instances_type;
  typedef db::instance<graph_type> instance_type;
  typedef typename instances_type::touching_iterator touching_iterator;
  typedef typename instances_type::overlapping_iterator overlapping_iterator;
  typedef typename instances_type::const_iterator const_iterator;
  typedef typename instances_type::parent_inst_type parent_inst_type;
  typedef typename instances_type::parent_inst_iterator parent_inst_iterator;
  typedef typename instances_type::parent_cell_iterator parent_cell_iterator;
  typedef typename instances_type::child_cell_iterator child_cell_iterator;

  /**
   *  @brief Standard constructor: create an empty cell object
   * 
   *  Takes the manager object from the graph object.
   *
   *  @param ci The index of the cell
   *  @param g A reference to the graph object that owns the cell
   */
  cell (cell_index_type ci, graph_type &g) 
    : db::Object (g.manager ()), 
      m_cell_index (ci), mp_graph (&g), m_instances (this), m_hier_levels (0), m_bbox_needs_update (false),
      mp_last (0), mp_next (0)
  {
    //  .. nothing yet 
  }

  /**
   *  @brief The assignment operator 
   */
  cell &operator= (const cell &d)
  {
    if (this != &d) {

      m_cell_index = d.m_cell_index;

      //  because the shapes objects do not have a good 
      //  copy ctor which allows to attach them to a graph, 
      //  do the copy in this somewhat strange way ..
      clear_shapes_no_invalidate ();
      m_shapes_vector.reserve (d.m_shapes_vector.size ());
      for (unsigned int l = 0; l < d.m_shapes_vector.size (); ++l) {
        m_shapes_vector.push_back (new db::shapes<coord_type> (manager (), mp_graph));
        *(m_shapes_vector.back ()) = *(d.m_shapes_vector [l]);
      }

      m_instances = d.m_instances;
      m_bbox = d.m_bbox;
      m_bboxes = d.m_bboxes;
      m_hier_levels = d.m_hier_levels;
      m_bbox_needs_update = d.m_bbox_needs_update;

    }
    return *this;
  }

  /**
   *  @brief The copy ctor
   *
   *  In contrast to the assignment operator the copy ctor also
   *  copies the owning graph reference.
   */
  cell (const cell &d)
    : db::Object (d),
      mp_graph (d.mp_graph), m_instances (this), m_hier_levels (d.m_hier_levels),
      mp_last (0), mp_next (0)
  {
    operator= (d);
  }

  /**
   *  @brief The destructor
   */
  ~cell ()
  {
    clear_shapes ();
  }

  /**
   *  @brief The translation operator 
   */
  void translate (const cell &d, generic_repository<coord_type> &rep)
  {
    tl_assert (this != &d);

    m_cell_index = d.m_cell_index;

    //  because the shapes objects do not have a good 
    //  copy ctor which allows to attach them to a graph, 
    //  do the copy in this somewhat strange way ..
    clear_shapes_no_invalidate ();
    m_shapes_vector.reserve (d.m_shapes_vector.size ());
    for (unsigned int l = 0; l < d.m_shapes_vector.size (); ++l) {
      m_shapes_vector.push_back (new db::shapes<coord_type> (manager (), mp_graph));
      m_shapes_vector.back ()->translate (*(d.m_shapes_vector [l]), rep);
    }

    m_instances = d.m_instances;
    m_bbox = d.m_bbox;
    m_bboxes = d.m_bboxes;
    m_hier_levels = d.m_hier_levels;
    m_bbox_needs_update = d.m_bbox_needs_update;
  }

  /**
   *  @brief Return the const shapes list of the given layer
   *
   *  This method allows to access the shapes list on a certain layer.
   *  If the layer does not exist yet, a reference to an empty list is
   *  returned.
   *
   *  @param index The layer index of the shapes list to retrieve
   *
   *  @return A const reference to the shapes list
   */
  db::shapes<coord_type> &shapes (unsigned int index) 
  {
    if (m_shapes_vector.capacity () <= index) {
      m_shapes_vector.reserve (index + 1);
    }
    while (m_shapes_vector.size () <= index) {
      m_shapes_vector.push_back (new db::shapes<coord_type> (manager (), mp_graph));
    }
    return *(m_shapes_vector [index]);
  }

  /**
   *  @brief Return the shapes list of the given layer
   *
   *  This method allows to access the shapes list on a certain layer.
   *  If the layer does not exist yet, it is created.
   *
   *  @param index The layer index of the shapes list to retrieve
   *
   *  @return A reference to the shapes list
   */
  const db::shapes<coord_type> &shapes (unsigned int index) const
  {
    if (m_shapes_vector.size () > index) {
      return *(m_shapes_vector [index]);
    } else {
      //  Because of a gcc bug it seems to be not possible
      //  to instantiate a simple static object here:
      static const db::shapes<coord_type> *empty_shapes = 0;
      if (! empty_shapes) {
        empty_shapes = new db::shapes<coord_type> ();
      }
      return *empty_shapes;
    }
  }

  /**
   *  @brief Clear all shapes in the cell
   */
  void clear_shapes ()
  {
    mp_graph->invalidate_bboxes ();  //  HINT: must come before the change is done!
    clear_shapes_no_invalidate ();
  }

  /**
   *  @brief Clear the instance list
   */
  void clear_insts ()
  {
    if (! m_instances.empty ()) {
      m_instances.clear_insts ();
    }
  }

  /**
   *  @brief Swap the layers given
   */
  void swap (unsigned int i1, unsigned int i2)
  {
    if (i1 != i2) {
      shapes (i1).swap (shapes (i2));
      m_bbox_needs_update = true;
    }
  }

  /**
   *  @brief Clear the shapes on the given layer
   */
  void clear (unsigned int index)
  {
    shapes (index).clear ();
    //  db::Shapes::clear does not set the bbox dirty flag! (reason: after clear() there is nothing where this flag could be stored ..)
    m_bbox_needs_update = true;
  }

  /**
   *  @brief Return the number of layers in this cell
   *
   *  The number of layer is the number of layers actually 
   *  created. There may be more layers but those are definitly
   *  empty.
   */
  unsigned int layers () const
  {
    return m_shapes_vector.size ();
  }

  /**
   *  @brief Reserve space for the given number of layers
   */
  void reserve_layers (unsigned int s)
  {
    m_shapes_vector.reserve (s);
  }

  /** 
   *  @brief Erase a cell instance given by a instance proxy
   *
   *  Erasing a cell instance will destroy the sorting order and invalidate
   *  other instance proxies.
   *  sort() must be called before a region query can be done
   *  on the cell instances. 
   */
  void erase (const instance_type &ref)
  {
    m_instances.erase (ref);
  }

  /**
   *  @brief Erase a cell instance
   *  
   *  Erasing a cell instance will destroy the sorting order.
   *  sort() must be called before a region query can be done
   *  on the cell instances. update_bbox() must be called 
   *  explicitly before the bounding box can be updated.
   */
  void erase (const_iterator e)
  {
    m_instances.erase (e);
  }

  /**
   *  @brief Erasing of multiple instances
   *
   *  Erase a set of positions given by an iterator I: *(from,to).
   *  *I must render an "iterator" object.
   *  The iterators in the sequence from, to must be sorted in
   *  "operator<" order.
   *
   *  @param first The start of the sequence of iterators
   *  @param last The end of the sequence of iterators
   */
  void erase_insts (const std::vector<instance_type> &instances)
  {
    m_instances.erase_insts (instances);
  }

  /**
   *  @brief Insert a generic cell instance 
   *  
   *  Inserting a cell instance will destroy the sorting order.
   *  sort() must be called before a region query can be done
   *  on the cell instances. update_bbox() must be called 
   *  explicitly before the bounding box can be updated.
   */
  template <class Inst>
  void insert (const Inst &inst)
  {
    m_instances.insert (inst);
  }

  /**
   *  @brief Insert a sequence [from,to) of cell instances
   *  
   *  @param from The start of the sequence
   *  @param to The past-the-end pointer of the sequence
   */
  template <class I>
  void insert (I from, I to)
  {
    m_instances.insert (from, to);
  }

  /**
   *  @brief Insert an instance given by a instance reference 
   *
   *  This member may be used to copy an instance from one cell to another.
   *  Because of the inherent instability of the instance pointer it must not be 
   *  used to copy instances within one cell.
   *
   *  @param instance The instance reference of which to insert a copy
   */
  instance_type insert (const instance_type &instance)
  {
    return (m_instances.insert (instance));
  }

  /**
   *  @brief Transform the instance pointed to by the instance reference
   *
   *  @param t The transformation to apply
   */
  template <class Trans>
  void transform (const instance_type &ref, const Trans &t)
  {
    m_instances.transform (ref, t);
  }

  /**
   *  @brief Insert an instance given by a instance reference with a different cell index and property ID
   *
   *  This member may be used to map an instance to another layout object. 
   *
   *  @param ref The instance reference of which to insert a copy
   *  @param im The mapper to new cell index to use (for mapping to a different layout for example)
   *  @param pm The mapper to new cell property ID use (for mapping to a different layout for example)
   */
  template <class IndexMap, class PropIdMap>
  instance_type insert (const instance_type &ref, const IndexMap &im, const PropIdMap &pm)
  {
    return (m_instances.insert (ref, im, pm));
  }

  /** 
   *  @brief Replace the properties ID of an element (pointed to by the iterator) with the given one
   *
   *  The iterator must point to a instance already having a property ID. 
   *  If this is not the case, an exception is thrown
   */
  void replace_prop_id (const instance_type &ref, db::PropertiesRepository::properties_id_type prop_id)
  {
    m_instances.replace_prop_id (ref, prop_id);
  }

  /**
   *  @brief Replace the instance pointed to by the iterator with the given instance
   *
   *  The ability to replace an instance implies the ability to in-place convert the
   *  kind of instance, i.e. a instance without property cannot be replaced with one
   *  with properties.
   */
  template <class InstArray>
  void replace (const instance_type &ref, const InstArray &inst)
  {
    m_instances.replace (ref, inst);
  }

  /**
   *  @brief The cell index accessor method
   *
   *  @return The cell index of the cell
   */
  const cell_index_type &cell_index () const
  {
    return m_cell_index;
  }

  /**
   *  @brief Update the child-parent relationships
   * 
   *  This will update the child-parent relationships. Basically
   *  this means entering the cell as a parent into all it's child
   *  cells.
   */
  void update_relations ()
  {
    m_instances.update_relations (mp_graph, cell_index ());
  }

  /** 
   *  @brief Check if the bbox needs update because the shapes have changed
   *
   *  Returns true if the bounding box needs update because the
   *  shape bounding box (not one of the child bounding boxes) has changed.
   */
  bool is_shape_bbox_dirty () const
  {
    if (m_bbox_needs_update) {
      return true;
    }
    for (typename shapes_vector::const_iterator s = m_shapes_vector.begin (); s != m_shapes_vector.end (); ++s) {
      if ((*s)->is_bbox_dirty ()) {
        return true;
      }
    }
    return false;
  }

  /** 
   *  @brief Update the bbox 
   *
   *  This will update the bbox from the shapes and instances.
   *  This requires the bboxes of the child cells to be computed
   *  before. Practically this will be done by computing the
   *  bboxes bottom-up in the hierarchy.
   *  In addition, the number of hierarchy leves below is also
   *  updated.
   *
   *  @param layers The max. number of layers in the child cells
   *  @return true, if the bounding box has changed.
   */
  bool update_bbox (unsigned int layers)
  {
    unsigned int l;

    //  determine the bounding box
    box_type org_bbox = m_bbox;
    m_bbox = box_type ();

    //  prepare the box vector
    unsigned int n = std::max (layers, (unsigned int) m_shapes_vector.size ());
    if (n == 0) {
      m_bboxes.clear ();
      return false;
    }

    //  save the original boxes for simple compare
    box_vector org_bboxes;
    org_bboxes.resize (n);
    org_bboxes.swap (m_bboxes);
    
    //  compute the per-layer bboxes of the cell instances
    //  exploit the fact that these are sorted by instance,
    //  rotation and magnification.
    for (typename instances_type::sorted_inst_index_iterator o = m_instances.begin_sorted_inst_indices (); o != m_instances.end_sorted_inst_indices (); ) {

      typename instances_type::sorted_inst_index_iterator o1 = o;
      const cell_inst_array_type &o1_inst = m_instances.begin (*o)->cell_inst ();

      typename instances_type::sorted_inst_index_iterator oo = o;
      while (++oo != m_instances.end_sorted_inst_indices () && m_instances.begin (*oo)->cell_inst ().raw_equal (o1_inst)) 
        ;

      box_type raw_box;
      while (o != oo) {
        raw_box += m_instances.begin (*o)->cell_inst ().raw_bbox ();
        ++o;
      }

      for (l = 0; l < layers; ++l) {

        //  the per-layer bounding boxes
        db::box_convert <cell_inst_type> bc (*mp_graph, l);
        box_type lbox = o1_inst.bbox_from_raw_bbox (raw_box, bc);

        if (! lbox.empty ()) {
          m_bbox += lbox;
          m_bboxes[l] += lbox;
        }

      }
      
    }

    //  update the bboxes of the shapes lists
    for (l = 0; l < m_shapes_vector.size (); ++l) {

      m_shapes_vector [l]->update_bbox ();
      box_type sbox (m_shapes_vector [l]->bbox ());

      if (! sbox.empty ()) {
        m_bbox += sbox;
        m_bboxes[l] += sbox;
      }
     
    }

    //  reset "dirty child instances" flag
    m_bbox_needs_update = false;

    //  return true, if anything has changed with the box 
    bool changed = (org_bbox != m_bbox);
    for (l = 0; ! changed && l < n; ++l) {
      if (l >= org_bboxes.size ()) {
        changed = ! m_bboxes [l].empty ();
      } else {
        changed = (m_bboxes [l] != org_bboxes [l]);
      }
    }
    return changed;

  }

  /**
   *  @brief Sort the shapes lists 
   *
   *  This will sort the shapes lists for query of regions 
   *  on a per-shape basis. Since sorting of the shapes is
   *  guarded against redundant sorting (db::Layer::sort),
   *  we can safely call the sort method in any case.
   */
  void sort_shapes ()
  {
    for (typename shapes_vector::iterator s = m_shapes_vector.begin ();
         s != m_shapes_vector.end (); ++s) {
      (*s)->sort ();
    }
  }

  /**
   *  @brief Retrieve the bounding box of the cell
   *
   *  Before the bounding box can be retrieved, it must have
   *  been computed using update_bbox. This is performed by 
   *  requesting an update from the graph.
   *
   *  @return The bounding box that was computed by update_bbox
   */
  const box_type &bbox () const
  {
    mp_graph->update ();
    return m_bbox;
  }

  /**
   *  @brief Retrieve the per-layer bounding box of the cell
   *
   *  Before the bounding box can be retrieved, it must have
   *  been computed using update_bbox. This is performed by 
   *  requesting an update from the graph.
   *
   *  @return The bounding box that was computed by update_bbox
   */
  const box_type &bbox (unsigned int l) const
  {
    mp_graph->update ();
    if (l < m_bboxes.size ()) {
      return m_bboxes [l];
    } else {
      return ms_empty_box;
    }
  }

  /**
   *  @brief Region query for the instances in "overlapping" mode
   *
   *  This will return the region query iterator for the child cell
   *  instances overlapping with the given region b. 
   *
   *  @param b The region to query
   *
   *  @return The region iterator for the child cell instances
   */
  overlapping_iterator begin_overlapping (const box_type &b) const
  {
    mp_graph->update ();
    return m_instances.begin_overlapping (b, mp_graph);
  }

  /**
   *  @brief Region query for the instances in "touching" mode
   *
   *  This will return the region query iterator for the child cell
   *  instances touching the given region b. 
   *
   *  @param b The region to query
   *
   *  @return The region iterator for the child cell instances
   */
  touching_iterator begin_touching (const box_type &b) const
  {
    mp_graph->update ();
    return m_instances.begin_touching (b, mp_graph);
  }

  /**
   *  @brief The child cell iterator
   *
   *  This iterator will report the child cell indices, not every instance.
   *  Basically this is a filter.
   */
  child_cell_iterator begin_child_cells () const
  {
    return m_instances.begin_child_cells ();
  }

  /** 
   *  @brief Report the number of child cells
   *
   *  Just the number of child cells with distinct cell indices are
   *  reported.
   *  CAUTION: this method is SLOW!
   */
  size_t child_cells () const
  {
    return m_instances.child_cells ();
  }

  /**
   *  @brief The number of cell instances 
   */
  size_t cell_instances () const
  {
    return m_instances.cell_instances ();
  }

  /**
   *  @brief The cell instance access begin iterator
   *
   *  The iterator follows the "at_end" semantics.
   */
  const_iterator begin () const
  {
    return m_instances.begin ();
  }

  /**
   *  @brief The cell instance access begin iterator plus offset
   *
   *  The iterator returned will point to the i-th element.
   *  The iterator follows the "at_end" semantics.
   */
  const_iterator begin (size_t i) const
  {
    return m_instances.begin (i);
  }

  /**
   *  @brief The cell instance access by index in list of sorted instances
   *
   *  The iterator returned will point to the i-th element, where i is the index in the list
   *  of instances sorted by cell index.
   *  The iterator follows the "at_end" semantics.
   */
  const_iterator sorted_inst_ptr (size_t i) const
  {
    return m_instances.begin (m_instances.begin_sorted_inst_indices () [i]);
  }

  /**
   *  @brief The parent instance list begin iterator
   *
   *  The begin_parent_insts() allows to access to the parent instance list.
   */
  parent_inst_iterator begin_parent_insts () const
  {
    mp_graph->update ();
    return m_instances.begin_parent_insts (mp_graph);
  }

  /**
   *  @brief Report the number of parent cells 
   *
   *  Since the m_parent_insts vector just stores references to those cells
   *  that have distinct cell indices, we can simply report their length.
   */
  size_t parent_cells () const
  {
    return m_instances.parent_cells ();
  }

  /**
   *  @brief The parent cell iterator
   * 
   *  This iterator will iterate over the parent cells, just returning their
   *  cell index.
   */
  parent_cell_iterator begin_parent_cells () const
  {
    return m_instances.begin_parent_cells ();
  }

  /**
   *  @brief The parent cell end iterator
   */
  parent_cell_iterator end_parent_cells () const
  {
    return m_instances.end_parent_cells ();
  }

  /**
   *  @brief Tell if the cell is a top-level cell
   *
   *  A cell is a top-level cell if there are no parent instantiations.
   */
  bool is_top () const
  {
    mp_graph->update ();
    return m_instances.is_top ();
  }

  /**
   *  @brief Tell if the cell is a leaf cell
   *
   *  A cell is a leaf cell if there are no child instantiations.
   */
  bool is_leaf () const
  {
    return m_instances.empty ();
  }

  /**
   *  @brief Return the number of hierarchy levels below (expensive)
   */
  unsigned int hierarchy_levels () const
  {
    mp_graph->update ();
    return m_hier_levels;
  }

  /**
   *  @brief begin iterator of all shapes with a overlapping mode region query
   *
   *  This method ensures that the box tress are updated before the query is done.
   *  The db::Shape-specific iterator does not do that.
   *
   *  @param box The region to query the shapes from
   *  @param layer The layer from which to query the shapes
   *  @param flags The flags of shapes to query (see db::shape documentation)
   *  @param prop_sel A property selector (a set of property set ids)
   *  @param inv_prop_sel true, if prop_sel should select all ids that should not be iterated
   *  
   *  @return The iterator delivering all these shapes
   */
  shape_iterator begin_overlapping (unsigned int layer, const box_type &box, unsigned int flags, const typename shape_iterator::property_selector *prop_sel = 0, bool inv_prop_sel = false) const
  {
    mp_graph->update ();
    return shapes (layer).begin_overlapping (box, flags, prop_sel, inv_prop_sel);
  }

  /**
   *  @brief begin iterator of all shapes with a touching mode region query
   *
   *  This method ensures that the box tress are updated before the query is done.
   *  The db::Shape-specific iterator does not do that.
   *
   *  @param box The region to query the shapes from
   *  @param layer The layer from which to query the shapes
   *  @param flags The flags of shapes to query (see db::shape documentation)
   *  @param prop_sel A property selector (a set of property set ids)
   *  @param inv_prop_sel true, if prop_sel should select all ids that should not be iterated
   *  
   *  @return The iterator delivering all these shapes
   */
  shape_iterator begin_touching (unsigned int layer, const box_type &box, unsigned int flags, const typename shape_iterator::property_selector *prop_sel = 0, bool inv_prop_sel = false) const
  {
    mp_graph->update ();
    return shapes (layer).begin_touching (box, flags, prop_sel, inv_prop_sel);
  }

  /**
   *  @brief Collect all calling cells (either calling this cell directly or indirectly)
   *
   *  This method adds all cell indices of all cells that either directly or indirectly 
   *  call this cell to the set given. It is assumed that the if the set contains a cell, it
   *  will also contain all called cells, so it may act as a cache.
   */
  void collect_caller_cells (std::set<cell_index_type> &callers) const
  {
    for (parent_cell_iterator cc = begin_parent_cells (); cc != end_parent_cells (); ++cc) {
      if (callers.find (*cc) == callers.end ()) { 
        callers.insert (*cc);
        mp_graph->cell (*cc).collect_caller_cells (callers);
      }
    }
  }

  /**
   *  @brief Collect all cells called either directly or indirectly by this cell
   *
   *  This method adds all cell indices of all cells called either directly or indirectly 
   *  by this cell to the set given. It is assumed that the if the set contains a cell, it
   *  will also contain all called cells, so it may act as a cache.
   */
  void collect_called_cells (std::set<cell_index_type> &called) const
  {
    for (child_cell_iterator cc = begin_child_cells (); ! cc.at_end (); ++cc) {
      if (called.find (*cc) == called.end ()) { 
        called.insert (*cc);
        mp_graph->cell (*cc).collect_called_cells (called);
      }
    }
  }

  /**
   *  @brief Invalidate the instances bounding box (to be called by the instances container)
   */
  void invalidate_insts ()
  {
    mp_graph->invalidate_hier ();  //  HINT: must come before the change is done!
    mp_graph->invalidate_bboxes ();
    m_bbox_needs_update = true;
  }

  /**
   *  @brief Invalidate the hierarchical structure (to be called by the instances container)
   */
  void invalidate_hier ()
  {
    mp_graph->invalidate_hier ();  //  HINT: must come before the change is done!
  }

  /** 
   *  @brief Implementation of the redo method
   */
  void redo (db::Op *op) throw ()
  {
    //  actions are only queued by the instance list - this is should be 
    //  responsible for the handling of the latter.
    //  HACK: this is not really a nice concept, but it saves us a pointer to the manager.
    db::inst_op_base<instances_type> *instop = dynamic_cast< db::inst_op_base<instances_type> *> (op);
    if (instop) {
      instop->redo (&m_instances);
    } 
  }

  /** 
   *  @brief Implementation of the undo method
   */
  void undo (db::Op *op) throw ()
  {
    //  actions are only queued by the instance list - this is should be 
    //  responsible for the handling of the latter.
    //  HACK: this is not really a nice concept, but it saves us a pointer to the manager.
    db::inst_op_base<instances_type> *instop = dynamic_cast< db::inst_op_base<instances_type> *> (op);
    if (instop) {
      instop->undo (&m_instances);
    } 
  }

  /**
   *  @brief Collect memory usage statistics
   */
  void
  collect_mem_stat (db::MemStatistics &m) const
  {
    m.cell_info (m_cell_index);
    m.cell_info (mp_graph);
    m.cell_info (m_shapes_vector);                                         //  need more: see below
    m.cell_info (m_bbox);
    m.cell_info (m_bboxes); 
    m.cell_info (m_hier_levels);
    m.cell_info (m_bbox_needs_update);

    m_instances.collect_mem_stat (m);

    for (typename shapes_vector::const_iterator s = m_shapes_vector.begin (); s != m_shapes_vector.end (); ++s) {
      (*s)->collect_mem_stat (m);
    }
  }

private:
  friend class db::CellGraph;
  friend class db::cell_list<cell>;
  friend class db::cell_list_iterator<cell>;
  friend class db::cell_list_const_iterator<cell>;
  friend class db::instances<Graph>;

  cell_index_type m_cell_index;
  mutable graph_type *mp_graph;
  shapes_vector m_shapes_vector;
  instances_type m_instances;
  box_type m_bbox;
  box_vector m_bboxes;
  unsigned int m_hier_levels;
  bool m_bbox_needs_update;
  static box_type ms_empty_box;

  //  linked list, used by CellGraph
  cell *mp_last, *mp_next;

  //  clear the shapes without telling the graph
  void clear_shapes_no_invalidate ()
  {
    for (typename shapes_vector::iterator s = m_shapes_vector.begin ();
         s != m_shapes_vector.end (); ++s) {
      delete *s;
    }
    m_shapes_vector.clear ();
    m_bbox_needs_update = true;
  }

  //  helper function for computing the number of hierarchy levels
  //  must be called bottom-up
  unsigned int 
  count_hier_levels () const
  {
    unsigned int l = 0;

    for (const_iterator c = begin (); !c.at_end (); ++c) {
      l = std::max (l, mp_graph->cell (c->cell_index ()).m_hier_levels + 1);
    }

    return l;
  }

  /**
   *  @brief Return a reference to the instances object 
   */
  instances_type &instances ()
  {
    return m_instances;
  }

  /**
   *  @brief Return a reference to the instances object 
   */
  const instances_type &instances () const
  {
    return m_instances;
  }

  /**
   *  @brief Count the number of parent instances
   * 
   *  This will accumulate the number of parent instances in the vector
   *  provided. Later, this count is used to resize the parent instance
   *  vector.
   */
  void count_parent_insts (std::vector <size_t> &count) const
  {
    m_instances.count_parent_insts (count);
  }

  /**
   *  @brief Clear the parent instance list
   *
   *  Reserve the given number of entries.
   */
  void clear_parent_insts (size_t sz = 0)
  {
    m_instances.clear_parent_insts (sz);
  }

  /**
   *  @brief Sort the child instance list
   * 
   *  The child instances are first sorted by cell index, then by raw transformation
   *  (excluding displacement). This allows to simplify the bbox computation by 
   *  convolution of the displacements bboxes with the object bboxes.
   */
  void sort_child_insts ()
  {
    m_instances.sort_child_insts ();
  }

  /**
   *  @brief Sort the cell instance list
   *
   *  This will sort the cell instance list. As a prerequesite
   *  the cell's bounding boxes must have been computed.
   *
   *  @param layers The maximum number of layers in the child cells
   */
  void sort_inst_tree ()
  {
    m_instances.sort_inst_tree (mp_graph);

    //  update the number of hierarchy levels @@@ why here???
    m_hier_levels = count_hier_levels ();
  }

};

/** 
 *  @brief A typedef for a standard cell
 */
typedef db::cell<db::CellGraph> Cell;

} // namespace db

namespace gsi
{
  template <class C> struct ClassTag;
  struct CantCopySemanticsTag;
  struct CantDefaultCreateSemanticsTag;

  template <> struct ClassTag<db::Cell> {
    typedef CantCopySemanticsTag copy_semantics;
    typedef CantDefaultCreateSemanticsTag create_semantics;
  };
}

#endif

