
/*

  KLayout Layout Viewer
  Copyright (C) 2006-2007 Matthias Koefferlein

  This program is free software; you can redistribute it and/or modify
  it under the terms of the GNU General Public License as published by
  the Free Software Foundation; either version 2 of the License, or
  (at your option) any later version.

  This program is distributed in the hope that it will be useful,
  but WITHOUT ANY WARRANTY; without even the implied warranty of
  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
  GNU General Public License for more details.

  You should have received a copy of the GNU General Public License
  along with this program; if not, write to the Free Software
  Foundation, Inc., 51 Franklin St, Fifth Floor, Boston, MA  02110-1301  USA

*/


#include "dbCellGraph.h"
#include "dbMemStatistics.h"
#include "dbTrans.h"
#include "dbShapeRepository.h"
#include "tlTimer.h"
#include "tlLog.h"
#include "tlInternational.h"
#include "tlProgress.h"
#include "tlAssert.h"

#include <limits>

namespace db
{

// -------------------------------------------------------------------------

CellGraph::CellGraph (db::Manager *manager)
  : db::Object (manager),
    m_cells_size (0),
    m_invalid (0)
{
  //  nothing yet.
}

CellGraph::CellGraph (const CellGraph &d)
  : db::Object (d),
    db::LayoutStateModel (d),
    m_cells_size (0),
    m_invalid (0)
{
  tl_assert (! d.under_construction () && ! (d.manager () && d.manager ()->transacting ()));
  operator= (d);
}

CellGraph::~CellGraph () 
{
  //  since it the cell graph (or the derived layout) might produce some transactions that refer to 
  //  this object, we need to clear the manager's transaction list before the cell graph is deleted.
  if (manager ()) {
    manager ()->clear ();
  }
}

CellGraph &
CellGraph::operator= (const CellGraph &d)
{
  if (this != &d) {

    tl_assert (! d.under_construction () && ! (d.manager () && d.manager ()->transacting ()));

    db::LayoutStateModel::operator= (d);
   
    invalidate_hier ();

    //  There is no copy ctor for the cells. Therefore we need
    //  to copy the cells in this somewhat strange way ..
    m_free_cell_indices.clear ();
    m_cells.clear ();
    m_cells_size = 0;
    m_cell_ptrs.clear ();
    for (const_iterator c = d.begin (); c != d.end (); ++c) {
      new_cell () = *c;
    }

    m_top_down_list = d.m_top_down_list;

    m_free_indices = d.m_free_indices;
    m_layer_states = d.m_layer_states;
  
  }
  return *this;
}

void
CellGraph::translate (const CellGraph &d, GenericRepository &rep)
{
  tl_assert (this != &d);

  db::LayoutStateModel::operator= (d);

  invalidate_hier ();
   
  //  There is no copy ctor for the cells. Therefore we need
  //  to copy the cells in this somewhat strange way ..
  m_free_cell_indices.clear ();
  m_cells.clear ();
  m_cells_size = 0;
  m_cell_ptrs.clear ();
  for (const_iterator c = d.begin (); c != d.end (); ++c) {
    new_cell ().translate (*c, rep);
  }

  m_free_indices = d.m_free_indices;
  m_layer_states = d.m_layer_states;
}

void
CellGraph::remove_cell (cell_type &cell)
{
  cell_index_type ci = cell.cell_index ();
  tl_assert (is_valid_cell_index (ci));

  invalidate_hier ();

  m_cells.erase (iterator (m_cell_ptrs [ci]));

  m_cell_ptrs [ci] = 0;

  //  Using free cell indices does have one significant drawback:
  //  The cellview references cannot be uniquely classified as being invalid - because the
  //  ID might be reused. This causes problems, when a cell is being deleted and subsequently a
  //  cell is being created - a crash occures. Therefore the free index feature is disabled.
  //  If this causes memory consumption problems, it should be considered to use a map and
  //  an arbitrary ID.
  // m_free_cell_indices.push_back (ci);

  --m_cells_size;
}

bool 
CellGraph::is_valid_cell_index (cell_index_type ci) const
{
  return ci < m_cell_ptrs.size () && m_cell_ptrs [ci] != 0;
}

CellGraph::cell_type *
CellGraph::take_cell (cell_index_type ci)
{
  tl_assert (m_cell_ptrs [ci] != 0);

  invalidate_hier ();

  cell_type *cell = m_cells.take (iterator (m_cell_ptrs [ci]));
  --m_cells_size;

  m_cell_ptrs [ci] = 0;

  return cell;
}

void
CellGraph::insert_cell (cell_index_type ci, cell_type *new_cell)
{
  tl_assert (ci < m_cell_ptrs.size ());
  tl_assert (m_cell_ptrs [ci] == 0);

  invalidate_hier ();

  m_cells.push_back_ptr (new_cell);
  m_cell_ptrs [ci] = new_cell;

  ++m_cells_size;
}

CellGraph::cell_type &
CellGraph::new_cell ()
{
  invalidate_hier ();

  size_t new_index;
  if (m_free_cell_indices.empty ()) {
    new_index = m_cell_ptrs.size ();
  } else {
    new_index = m_free_cell_indices.back ();
    m_free_cell_indices.pop_back ();
  }

  cell_type *new_cell = new cell_type (new_index, *this);
  m_cells.push_back_ptr (new_cell);
  m_cell_ptrs.push_back (new_cell);

  ++m_cells_size;

  return *new_cell;
}

void 
CellGraph::update_relations ()
{
  for (iterator c = begin (); c != end (); ++c) {
    c->sort_child_insts ();
  }

  std::vector <size_t> parent_insts (cells (), 0);
  for (const_iterator c = begin (); c != end (); ++c) {
    c->count_parent_insts (parent_insts);
  }
  std::vector <size_t>::const_iterator n = parent_insts.begin ();
  for (iterator c = begin (); c != end (); ++c, ++n) {
    c->clear_parent_insts (*n);
  }
  for (iterator c = begin (); c != end (); ++c) {
    c->update_relations ();
  }
}

CellGraph::top_down_iterator 
CellGraph::end_top_cells () 
{
  update ();

  top_down_iterator e = m_top_down_list.begin ();
  while (e != end_top_down ()) {
    if (! cell (*e).is_top ()) {
      return e;
    }
    ++e;
  }
  return e;
}

CellGraph::top_down_const_iterator 
CellGraph::end_top_cells () const
{
  update ();

  top_down_const_iterator e = m_top_down_list.begin ();
  while (e != end_top_down ()) {
    if (! cell (*e).is_top ()) {
      return e;
    }
    ++e;
  }
  return e;
}

void 
CellGraph::do_update ()
{
  //  Since the cell's bbox method tries to call graph.update()
  //  we need to check for recursion
  if (under_construction ()) {
    return;
  }
  start_changes ();

  //  establish a progress report since this operation can take some time.
  //  HINT: because of some gcc bug, automatic destruction of the tl::Progress
  //  object does not work. We overcome this problem by creating the object with new 
  //  and catching exceptions.
  tl::Progress *pr = new tl::Progress (tl::translate ("Sorting layout"), m_cells_size, 1000);
  pr->set_desc ("");

  try {

    //  if the hierarchy has been changed so far, update
    //  the hierarchy management informations
    if (hier_dirty ()) {
      {
        tl::SelfTimer timer (tl::verbosity () >= 31, "Updating relations");
        pr->set_desc (tl::translate ("Updating relations"));
        update_relations ();
      } 
      {
        tl::SelfTimer timer (tl::verbosity () >= 31, "Topological sort");
        pr->set_desc (tl::translate ("Topological sorting"));
        tl_assert (topological_sort ());
      }
    }

    //  @@@ a boolean vector (with size as determined by number of cells)
    //  would probably be much faster!
    std::set<cell_index_type> dirty_parents;

    //  if something on the bboxes (either on shape level or on 
    //  cell bbox level - i.e. by child instances) has been changed,
    //  update the bbox informations. In addition sort the shapes
    //  lists of region queries, since they might have changed once
    //  the bboxes are dirty.
    if (bboxes_dirty ()) {
      {
        tl::SelfTimer timer (tl::verbosity () >= 31, "Updating bounding boxes");
        size_t layers = 0;
        pr->set (0);
        pr->set_desc (tl::translate ("Updating bounding boxes"));
        for (bottom_up_iterator c = begin_bottom_up (); c != end_bottom_up (); ++c) {
          ++*pr;
          cell_type &cp (cell (*c));
          if (cp.is_shape_bbox_dirty () || dirty_parents.find (*c) != dirty_parents.end ()) {
            if (cp.update_bbox (layers)) {
              //  the bounding box has changed - need to insert parents into "dirty parents" list
              for (cell_type::parent_inst_iterator p = cp.begin_parent_insts (); ! p.at_end (); ++p) {
                dirty_parents.insert (p->parent_cell_index ());
              }
            } 
          }
          if (cp.layers () > layers) {
            layers = cp.layers ();
          }
        }
      }
      {
        tl::SelfTimer timer (tl::verbosity () >= 31, "Sorting shapes");
        pr->set (0);
        pr->set_desc (tl::translate ("Sorting shapes"));
        for (bottom_up_iterator c = begin_bottom_up (); c != end_bottom_up (); ++c) {
          ++*pr;
          cell_type &cp (cell (*c));
          cp.sort_shapes ();
        }
      }
    }

    //  sort the instance trees now, since we have computed the bboxes
    if (hier_dirty () || ! dirty_parents.empty ()) {
      tl::SelfTimer timer (tl::verbosity () >= 31, "Sorting instances");
      size_t layers = 0;
      pr->set (0);
      pr->set_desc (tl::translate ("Sorting instances"));
      for (bottom_up_iterator c = begin_bottom_up (); c != end_bottom_up (); ++c) {
        ++*pr;
        cell_type &cp (cell (*c));
        if (hier_dirty () || dirty_parents.find (*c) != dirty_parents.end ()) {
          cp.sort_inst_tree ();
        }
        if (cp.layers () > layers) {
          layers = cp.layers ();
        }
      }
    }

  } catch (...) {
    //  HINT: see above.
    end_changes ();
    delete pr;
    throw;
  }

  end_changes ();
  delete pr;
}

void 
CellGraph::delete_layer (unsigned int n)
{
  tl_assert (n < layers () && m_layer_states [n] != Free);

  m_free_indices.push_back (n);
  m_layer_states [n] = Free;

  //  clear the shapes
  for (iterator c = begin (); c != end (); ++c) {
    c->clear (n);
  }
}

unsigned int 
CellGraph::insert_layer (bool special) 
{
  if (m_free_indices.size () > 0) {
    unsigned int i = m_free_indices.back ();
    m_free_indices.pop_back ();
    m_layer_states [i] = special ? Special : Normal;
    return i;
  } else {
    m_layer_states.push_back (special ? Special : Normal);
    unsigned int i = layers () - 1;
    return i;
  }
}

void 
CellGraph::insert_layer (unsigned int index, bool special) 
{
  if (index >= layers ()) {

    //  add layer to the end of the list.
    //  add as may freelist entries as required.
    while (index > layers ()) {
      m_free_indices.push_back (layers ());
      m_layer_states.push_back (Free);
    }
    m_layer_states.push_back (special ? Special : Normal);

  } else {

    tl_assert (m_layer_states [index] == Free);
    m_layer_states [index] = special ? Special : Normal;
  
  }

}

void 
CellGraph::reserve_layers (unsigned int n)
{
  for (iterator c = begin (); c != end (); ++c) {
    c->reserve_layers (n);
  }
  m_layer_states.reserve (n);
}

void
CellGraph::collect_mem_stat (db::MemStatistics &m) const
{
  m.layout_info (*((db::LayoutStateModel *) this));
  m.layout_info (m_cells_size);
  m.layout_info (m_cell_ptrs);
  m.layout_info (m_invalid);
  m.layout_info (m_top_down_list);
  m.layout_info (m_free_indices);
  m.layout_info (m_layer_states);

  //  this does not count the internal memory footprint of the cells
  m.layout_info (m_cells);

  for (cell_list::const_iterator i = m_cells.begin (); i != m_cells.end (); ++i) {
    i->collect_mem_stat (m);
  }
}

bool 
CellGraph::topological_sort ()
{
  m_top_down_list.clear ();
  m_top_down_list.reserve (m_cells_size);

  std::vector<size_t> num_parents (m_cell_ptrs.size (), 0);

  //  while there are cells to treat ..
  while (m_top_down_list.size () != m_cells_size) {

    cell_index_vector::const_iterator i1 = m_top_down_list.end ();

    //  Treat all cells that do not have all parents reported.
    //  For all such a cells, disable the parent counting, 
    //  add the cell's index to the top-down sorted list and
    //  increment the reported parent instance count in all the
    //  child cells.

    for (const_iterator c = begin (); c != end (); ++c) {
      if (c->parent_cells () == num_parents [c->cell_index ()]) {
        m_top_down_list.push_back (c->cell_index ());
        num_parents [c->cell_index ()] = std::numeric_limits<cell_index_type>::max ();
      }
    }

    //  For all these a cells, increment the reported parent instance 
    //  count in all the child cells.
    for (cell_index_vector::const_iterator ii = i1; ii != m_top_down_list.end (); ++ii) {
      for (cell_type::child_cell_iterator cc = cell (*ii).begin_child_cells (); ! cc.at_end (); ++cc) {
        tl_assert (num_parents [*cc] != std::numeric_limits<cell_index_type>::max ());
        num_parents [*cc] += 1;
      }
    }

    //  If no new cells have been reported this is basically a 
    //  sign of recursion in the graph.
    if (i1 == m_top_down_list.end ()) {
      return false;
    }
 
  }

  //  The cell graph is fine.
  return true;

}


}

