
/*

  KLayout Layout Viewer
  Copyright (C) 2006-2007 Matthias Koefferlein

  This program is free software; you can redistribute it and/or modify
  it under the terms of the GNU General Public License as published by
  the Free Software Foundation; either version 2 of the License, or
  (at your option) any later version.

  This program is distributed in the hope that it will be useful,
  but WITHOUT ANY WARRANTY; without even the implied warranty of
  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
  GNU General Public License for more details.

  You should have received a copy of the GNU General Public License
  along with this program; if not, write to the Free Software
  Foundation, Inc., 51 Franklin St, Fifth Floor, Boston, MA  02110-1301  USA

*/



#ifndef HDR_dbCellInst
#define HDR_dbCellInst

#include <dbBox.h>
#include <dbPropertiesRepository.h>
#include <dbTrans.h>

namespace db
{

class CellGraph;

/**
 *  @brief The cell instance class
 *
 *  The basic elements of a cell instance is a transformation
 *  and a cell reference (in this case through a cell index).
 */
template <class Graph>
class cell_inst
{
public:
  typedef Graph graph_type;
  typedef typename Graph::box_type box_type;
  typedef typename Graph::cell_index_type cell_index_type;
  typedef typename Box::coord_type coord_type;

  typedef db::object_tag< cell_inst<Graph> > tag;

  /**
   *  @brief Default ctor
   */
  cell_inst ()
    : m_cell_index (0)
  { }

  /**
   *  @brief Create a cell instance from the given index
   *  
   *  @param ci The cell index
   */
  cell_inst (cell_index_type ci)
    : m_cell_index (ci)
  { }

  /**
   *  @brief The cell index accessor
   */
  cell_index_type cell_index () const
  {
    return m_cell_index;
  }

  /**
   *  @brief Compute the bounding box
   *
   *  This method computes the bbox of the cell instance.
   *  As a requirement, the cell's bounding box must have been
   *  computed before.
   */
  box_type bbox (const graph_type &g) const
  {
    return g.cell (m_cell_index).bbox ();
  }

  /**
   *  @brief Compute the bounding box
   *
   *  This method computes the bbox of the cell instance
   *  given a certain layer.
   *  As a requirement, the cell's bounding boxes must have been
   *  computed before.
   */
  box_type bbox (const graph_type &g, unsigned int l) const
  {
    return g.cell (m_cell_index).bbox (l);
  }

  /**
   *  @brief Comparison: comparison for equality 
   */
  bool operator== (const cell_inst &d) const
  {
    return m_cell_index == d.m_cell_index;
  }

  /**
   *  @brief Comparison: compare by cell id 
   *
   *  This sorting order is used by the cell instances of the 
   *  cell.
   */
  bool operator< (const cell_inst &d) const
  {
    return m_cell_index < d.m_cell_index;
  }

  /**
   *  @brief Convert to a string
   */
  std::string to_string () const
  {
    return "[" + tl::to_string (m_cell_index) + "]";
  }

private:
  cell_index_type m_cell_index;
};

/** 
 *  @brief A typedef for a standard cell inst
 */
typedef db::cell_inst<db::CellGraph> CellInst;

} // namespace db

#endif

