
/*

  KLayout Layout Viewer
  Copyright (C) 2006-2007 Matthias Koefferlein

  This program is free software; you can redistribute it and/or modify
  it under the terms of the GNU General Public License as published by
  the Free Software Foundation; either version 2 of the License, or
  (at your option) any later version.

  This program is distributed in the hope that it will be useful,
  but WITHOUT ANY WARRANTY; without even the implied warranty of
  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
  GNU General Public License for more details.

  You should have received a copy of the GNU General Public License
  along with this program; if not, write to the Free Software
  Foundation, Inc., 51 Franklin St, Fifth Floor, Boston, MA  02110-1301  USA

*/



#include "dbGDS2Reader.h"
#include "dbGDS2.h"
#include "dbArray.h"

#include "tlException.h"
#include "tlString.h"
#include "tlClassRegistry.h"

namespace db
{

// ---------------------------------------------------------------
//  GDS2Reader


GDS2Reader::GDS2Reader (tl::InputStream &s)
  : m_stream (s), 
    m_progress (tl::translate ("Reading GDS2 file"), 0, 10000)
{
  m_progress.set_format (tl::translate ("%.0f MB"));
  m_progress.set_res (1.0);
  m_progress.set_unit (1024 * 1024);
}

GDS2Reader::~GDS2Reader ()
{
  //  .. nothing yet ..
}

const LayerMap &
GDS2Reader::read (db::Layout &layout, const LayerMap &map, bool create) throw (tl::Exception)
{
  m_layer_map = map;
  layout.start_changes ();
  do_read (layout, create);
  layout.end_changes ();
  return m_layer_map;
}

const LayerMap &
GDS2Reader::read (db::Layout &layout) throw (tl::Exception)
{
  m_layer_map = LayerMap ();
  layout.start_changes ();
  do_read (layout, true /*create layers*/);
  layout.end_changes ();
  return m_layer_map;
}

short 
GDS2Reader::get_record () throw (tl::Exception)
{  
  unsigned char *b = (unsigned char *) m_stream.get (4);
  if (! b) {
    error (tl::translate ("Unexpected end-of-file"));
  }

  //  @@@ this can be done more efficiently
  //  @@@ officially, the record length must be less then 0x8000.
  m_recnum++;
  m_reclen = (size_t (b[0]) << 8) | size_t (b[1]);
  short rec_id = (short (b[2]) << 8) | short (b[3]);

  if (m_reclen < 4) {
    error (tl::translate ("Invalid record length (less than 4)"));
  }
  if (m_reclen >= 0x8000) {
    warn (tl::translate ("record length larger than 0x8000 encountered: interpretating as unsigned"));
  }
  if (m_reclen % 2 == 1) {
    error (tl::translate ("Odd record length"));
  }

  m_reclen -= 4;

  if (m_reclen > 0) {
    mp_rec_buf = (unsigned char *) m_stream.get (m_reclen);
    if (! mp_rec_buf) {
      error (tl::translate ("Unexpected end-of-file"));
    }
  } else {
    mp_rec_buf = 0;
  }
   
  m_recptr = 0; 
  return rec_id;
}

inline int 
GDS2Reader::get_int () throw (tl::Exception)
{
  unsigned char *b = mp_rec_buf + m_recptr;
  m_recptr += 4;

  //  @@@ this can be done more efficiently ..
  return (int (b[0]) << 24) | (int (b[1]) << 16) | (int (b[2]) << 8) | int (b[3]);
}

inline short 
GDS2Reader::get_short () throw (tl::Exception)
{
  unsigned char *b = mp_rec_buf + m_recptr;
  m_recptr += 2;

  //  @@@ this can be done more efficiently ..
  return (short (b[0]) << 8) | short (b[1]);
}

inline unsigned short 
GDS2Reader::get_ushort () throw (tl::Exception)
{
  unsigned char *b = mp_rec_buf + m_recptr;
  m_recptr += 2;

  //  @@@ this can be done more efficiently ..
  return ((unsigned short) (b[0]) << 8) | (unsigned short) (b[1]);
}

inline double 
GDS2Reader::get_double () throw (tl::Exception)
{
  unsigned char *b = mp_rec_buf + m_recptr;
  m_recptr += 8;

  //  @@@ this can be done more efficiently ..
  double x = 0;
  for (int i = 1; i < 8; ++i) {
    x = x * 256.0 + double (b[i]);
  }

  if (b[0] & 0x80) {
    x = -x;
  }
  
  double e = double (b[0] & 0x7f) - (64.0 + 14);
  x *= pow (16.0, e);
  return x;
}

const char *
GDS2Reader::get_string () throw (tl::Exception)
{
  if (m_reclen == 0) {
    return "";
  }

  if (mp_rec_buf [m_reclen - 1] == 0) {
    //  we already have a terminating '\0': just return the string's location
    return (const char *) mp_rec_buf;
  } else {
    //  use the temporary buffer to create a zero-terminated string
    m_string_buf.assign ((const char *) mp_rec_buf, 0, m_reclen);
    return m_string_buf.c_str ();
  }
}

void
GDS2Reader::get_string (tl::string &s) const throw (tl::Exception)
{
  s.assign ((const char *) mp_rec_buf, 0, m_reclen);
}

void 
GDS2Reader::error (const std::string &msg) throw (tl::Exception)
{
  throw GDS2ReaderException (msg, m_stream.pos (), m_recnum, m_cellname.c_str ());
}

void 
GDS2Reader::warn (const std::string &msg) 
{
  // @@@ simple handling ..
  fprintf (stderr, tl::translate ("WARNING: %s (position=%ld, record number=%ld, cell=%s)\n").c_str (), msg.c_str (), m_stream.pos (), m_recnum, m_cellname.c_str ());
}

void
GDS2Reader::finish_element ()
{
  while (true) {

    short rec_id = get_record ();

    if (rec_id == sENDEL) {
      break;
    } else if (rec_id == sPROPATTR) {
      //  skip this record
    } else if (rec_id == sPROPVALUE) {
      //  skip this record
    } else {
      error (tl::translate ("ENDEL, PROPATTR or PROPVALUE record expected"));
    }

  } 
}


std::pair <bool, db::PropertiesRepository::properties_id_type> 
GDS2Reader::finish_element (db::PropertiesRepository &rep)
{
  bool any = false;
  long attr = 0;

  db::PropertiesRepository::properties_set properties;

  while (true) {

    short rec_id = get_record ();

    if (rec_id == sENDEL) {
      break;
    } else if (rec_id == sPROPATTR) {
      attr = long (get_ushort ());
    } else if (rec_id == sPROPVALUE) {

      const char *value = get_string ();
      properties.insert (std::make_pair (rep.prop_name_id (tl::Variant (attr)), 
                                         tl::Variant (value)));

      any = true;

    } else {
      error (tl::translate ("ENDEL, PROPATTR or PROPVALUE record expected"));
    }

  } 

  if (any) {
    return std::make_pair (true, rep.properties_id (properties));
  } else {
    return std::make_pair (false, 0);
  }
}


std::pair <bool, unsigned int> 
GDS2Reader::open_dl (db::Layout &layout, const LDPair &dl, bool create) throw (tl::Exception)
{
  std::pair<bool, unsigned int> ll = m_layer_map.logical (dl);
  if (ll.first || ! create) {
    return ll;
  } else {

    //  and create the layer
    db::LayerProperties lp;
    lp.layer = dl.layer;
    lp.datatype = dl.datatype;
    unsigned int ll = layout.insert_layer (lp);
    m_layer_map.map (dl, ll);

    return std::make_pair (true, ll);

  }
}

struct GDS2XY
{
  unsigned char x[4];
  unsigned char y[4];
};

struct GDS2XYToPoint
{
  db::Point operator() (const GDS2XY &p) 
  {
    //  @@@ this can be done more efficiently ..
    int x = (int (p.x[0]) << 24) | (int (p.x[1]) << 16) | (int (p.x[2]) << 8) | int (p.x[3]);
    int y = (int (p.y[0]) << 24) | (int (p.y[1]) << 16) | (int (p.y[2]) << 8) | int (p.y[3]);
    return db::Point (x, y);
  }
};

inline bool 
eq_x (const GDS2XY &a, const GDS2XY &b)
{
  return a.x[0] == b.x[0] && a.x[1] == b.x[1] && a.x[2] == b.x[2] && a.x[3] == b.x[3];
}

inline bool 
eq_y (const GDS2XY &a, const GDS2XY &b)
{
  return a.y[0] == b.y[0] && a.y[1] == b.y[1] && a.y[2] == b.y[2] && a.y[3] == b.y[3];
}

void 
GDS2Reader::do_read (db::Layout &layout, bool create_layers) throw (tl::Exception)
{
  m_recnum = 0;
  --m_recnum;
  m_reclen = 0;
  m_cellname = "";
  m_libname = "";
  m_create_layers = create_layers;

  //  read header
  if (get_record () != sHEADER) {
    error (tl::translate ("HEADER record expected"));
  }
  if (get_record () != sBGNLIB) {
    error (tl::translate ("BGNLIB record expected"));
  }

  //  read until 
  short rec_id = 0;
  do {
    rec_id = get_record ();
    if (rec_id == sLIBDIRSIZE ||
        rec_id == sSRFNAME ||
        rec_id == sREFLIBS ||
        rec_id == sFONTS ||
        rec_id == sATTRTABLE ||
        rec_id == sGENERATIONS ||
        rec_id == sFORMAT ||
        rec_id == sMASK ||
        rec_id == sENDMASKS) {
      //  OK and overread
    } else if (rec_id == sLIBNAME) {
      m_libname = get_string ();
    } else if (rec_id == sUNITS) {
      break;
    } else {
      error (tl::translate ("Invalid record or data type"));
    }
  } while (true);

  //  get units
  m_dbuu = get_double ();
  m_dbu = get_double ();
  layout.dbu (m_dbuu);

  //  this container has been found to grow quite a lot.
  //  using a list instead of a vector should make this more efficient.
  tl::vector<db::CellInstArray> instances;
  tl::vector<db::CellInstArrayWithProperties> instances_with_props;

  //  get cells
  while ((rec_id = get_record ()) == sBGNSTR) {

    m_progress.set (m_stream.pos ());

    //  erase current instance list 
    instances.erase (instances.begin (), instances.end ());
    instances_with_props.erase (instances_with_props.begin (), instances_with_props.end ());

    if (get_record () != sSTRNAME) {
      error (tl::translate ("STRNAME record expected"));
    }

    get_string (m_cellname);
    std::pair<bool, db::Layout::cell_index_type> c = layout.cell_by_name (m_cellname.c_str ()); 
    if (c.first) {
      //  cell already there: just add shapes (cell might have been created through forward reference)
      m_cellindex = c.second;
    } else {
      m_cellindex = layout.add_cell (m_cellname.c_str ());
    }
    db::Cell &cell = layout.cell (m_cellindex);

    //  read cell content
    while ((rec_id = get_record ()) != sENDSTR) { 

      m_progress.set (m_stream.pos ());

      if (rec_id == sBOUNDARY) {

        read_boundary (layout, cell);

      } else if (rec_id == sPATH) {

        read_path (layout, cell);

      } else if (rec_id == sSREF || rec_id == sAREF) {

        bool array = (rec_id == sAREF);
        read_ref (layout, cell, array, instances, instances_with_props);

      } else if (rec_id == sTEXT) {

        read_text (layout, cell);

      } else if (rec_id == sBOX) {

        read_box (layout, cell);

      } else if (rec_id == sNODE) {

        //  NODE records are ignored.
        warn (tl::translate ("NODE records are ignored"));
        while (get_record () != sENDEL) { }

      } else {
        error (tl::translate ("Invalid record or data type"));
      }
    
    }

    //  insert all instances collected
    if (! instances.empty ()) {
      cell.insert (instances.begin (), instances.end ());
    }
    if (! instances_with_props.empty ()) {
      cell.insert (instances_with_props.begin (), instances_with_props.end ());
    }

    m_cellname = "";

  }

  //  check, if the last record is a ENDLIB
  if (rec_id != sENDLIB) {
    error (tl::translate ("ENDLIB record expected"));
  }

}

void 
GDS2Reader::read_boundary (db::Layout &layout, db::Cell &cell)
{
  LDPair ld; 
  short rec_id = 0;

  do {
    rec_id = get_record ();
  } while (rec_id == sELFLAGS || rec_id == sPLEX);
  if (rec_id != sLAYER) {
    error (tl::translate ("LAYER record expected"));
  }
  ld.layer = get_ushort ();
  if (get_record () != sDATATYPE) {
    error (tl::translate ("DATATYPE record expected"));
  }
  ld.datatype = get_ushort ();

  if (get_record () != sXY) {
    error (tl::translate ("XY record expected"));
  }

  std::pair<bool, unsigned int> ll = open_dl (layout, ld, m_create_layers);
  if (ll.first) {

    GDS2XYToPoint pt_conv;

    //  create a box object if possible
    GDS2XY *xy = (GDS2XY *) mp_rec_buf;
    if ((m_reclen == 4 * sizeof (GDS2XY) || 
         (m_reclen == 5 * sizeof(GDS2XY) && pt_conv (xy[4]) == pt_conv (xy[0]))) && 
        ((eq_x (xy[0], xy[1]) && eq_x (xy[2], xy[3]) && 
          eq_y (xy[1], xy[2]) && eq_y (xy[0], xy[3])) ||
         (eq_x (xy[1], xy[2]) && eq_x (xy[0], xy[3]) && 
          eq_y (xy[0], xy[1]) && eq_y (xy[2], xy[3])))) {

      //  we can create a box object:
      db::Point p1 = pt_conv (*xy++);
      db::Point p2 = p1;
       
      while (xy < (GDS2XY *) mp_rec_buf + 4) {
        db::Point p (pt_conv (*xy++));
        if (p.x () < p1.x ()) {
          p1.set_x (p.x ());
        }
        if (p.y () < p1.y ()) {
          p1.set_y (p.y ());
        }
        if (p.x () > p2.x ()) {
          p2.set_x (p.x ());
        }
        if (p.y () > p2.y ()) {
          p2.set_y (p.y ());
        }
      }

      std::pair<bool, db::PropertiesRepository::properties_id_type> pp = finish_element (layout.properties_repository ());
      if (pp.first) {
        cell.shapes (ll.second).insert (db::BoxWithProperties (db::Box (p1, p2), pp.second));
      } else {
        cell.shapes (ll.second).insert (db::Box (p1, p2));
      }

    } else {

      //  convert the GDS2 record into the polygon.
      db::SimplePolygon poly;
      poly.assign_hull ((GDS2XY *) mp_rec_buf, (GDS2XY *) (mp_rec_buf + m_reclen), pt_conv);

      if (poly.hull ().size () < 3) {
        warn (tl::translate ("BOUNDARY with less than 3 points ignored"));
        finish_element ();
      } else {
        //  this will copy the polyon:
        std::pair<bool, db::PropertiesRepository::properties_id_type> pp = finish_element (layout.properties_repository ());
        if (pp.first) {
          cell.shapes (ll.second).insert (db::SimplePolygonRefWithProperties (db::SimplePolygonRef (poly, layout.shape_repository ()), pp.second));
        } else {
          cell.shapes (ll.second).insert (db::SimplePolygonRef (poly, layout.shape_repository ()));
        }
      }

    }

  } else {
    finish_element ();
  }
}

void 
GDS2Reader::read_path (db::Layout &layout, db::Cell &cell)
{
  LDPair ld; 
  short rec_id = 0;

  do {
    rec_id = get_record ();
  } while (rec_id == sELFLAGS || rec_id == sPLEX);
  if (rec_id != sLAYER) {
    error (tl::translate ("LAYER record expected"));
  }
  ld.layer = get_ushort ();
  if (get_record () != sDATATYPE) {
    error (tl::translate ("DATATYPE record expected"));
  }
  ld.datatype = get_ushort ();
    
  rec_id = get_record ();

  short type = 0; 
  if (rec_id == sPATHTYPE) {
    type = get_ushort (); 
    rec_id = get_record ();
  }

  if (type != 0 && type != 1 && type != 2 && type != 4) {
    warn (tl::translate ("Unsupported PATHTYPE"));
    type = 0;
  }

  db::Coord w = 0;
  if (rec_id == sWIDTH) {
    w = get_int ();
    rec_id = get_record ();
  }

  db::Coord bgn_ext = 0;
  db::Coord end_ext = 0;

  if (rec_id == sBGNEXTN) {
    bgn_ext = get_int ();
    rec_id = get_record ();
  } else {
    if (type == 2 || type == 1) {
      bgn_ext = w / 2;
    } 
  }

  if (rec_id == sENDEXTN) {
    end_ext = get_int ();
    rec_id = get_record ();
  } else {
    if (type == 2 || type == 1) {
      end_ext = w / 2;
    } 
  }

  if (rec_id != sXY) {
    error (tl::translate ("XY record expected"));
  }

  std::pair<bool, unsigned int> ll = open_dl (layout, ld, m_create_layers);
  if (ll.first) {

    //  convert the GDS2 record into the polygon.
    GDS2XYToPoint pt_conv;

    //  this will copy the path:
    db::Path path;
    path.width (w);
    path.extensions (bgn_ext, end_ext);
    path.assign ((GDS2XY *) mp_rec_buf, (GDS2XY *) (mp_rec_buf + m_reclen), pt_conv);
    path.round (type == 1);

    if (path.points () < 2) {
      warn (tl::translate ("PATH with less than 2 points ignored"));
      finish_element ();
    } else {
      std::pair<bool, db::PropertiesRepository::properties_id_type> pp = finish_element (layout.properties_repository ());
      if (pp.first) {
        cell.shapes (ll.second).insert (db::PathRefWithProperties (db::PathRef (path, layout.shape_repository ()), pp.second));
      } else {
        cell.shapes (ll.second).insert (db::PathRef (path, layout.shape_repository ()));
      }
    }

  } else {
    finish_element ();
  }
}

void 
GDS2Reader::read_text (db::Layout &layout, db::Cell &cell)
{
  LDPair ld; 
  short rec_id = 0;

  do {
    rec_id = get_record ();
  } while (rec_id == sELFLAGS || rec_id == sPLEX);
  if (rec_id != sLAYER) {
    error (tl::translate ("LAYER record expected"));
  }
  ld.layer = get_ushort ();
  if (get_record () != sTEXTTYPE) {
    error (tl::translate ("DATATYPE record expected"));
  }
  ld.datatype = get_ushort ();

  std::pair<bool, unsigned int> ll = open_dl (layout, ld, m_create_layers);
  if (ll.first) {

    rec_id = get_record ();
    if (rec_id == sPRESENTATION) {
      rec_id = get_record ();
    }
    if (rec_id == sPATHTYPE) {
      rec_id = get_record ();
    }
    if (rec_id == sWIDTH) {
      rec_id = get_record ();
    }

    bool mirror = false;
    int angle = 0;

    if (rec_id == sSTRANS) {

      short f = get_short ();
      if ((f & 0x8000) != 0) {
        mirror = true;
      }

      rec_id = get_record ();

    }

    if (rec_id == sMAG) {
      //  text size is ignored currently ...
      rec_id = get_record ();
    }

    if (rec_id == sANGLE) {

      double aorg = get_double ();
      double a = aorg / 90.0;
      if (a < -4 || a > 4) {
        warn (tl::sprintf (tl::translate ("Invalid text rotation angle (%g is less than -360 or larger than 360)").c_str (), aorg));
      }
      angle = int (a < 0 ? (a - 0.5) : (a + 0.5));
      if (fabs (double (angle) - a) > 1e-9) {
        warn (tl::sprintf (tl::translate ("Invalid text rotation angle (%g is not a multiple of 90)").c_str (), aorg));
      }
      while (angle < 0) {
        angle += 4;
      }
      while (angle >= 4) {
        angle -= 4;
      }

      rec_id = get_record ();
      
    }

    if (rec_id != sXY) {
      error (tl::translate ("XY record expected"));
    }
    db::Coord x = get_int ();
    db::Coord y = get_int ();

    db::Trans t (angle, mirror, db::Point (x, y));

    if (get_record () != sSTRING) {
      error (tl::translate ("STRING record expected"));
    }

    //  Create the text
    db::Text text (get_string (), t);

    std::pair<bool, db::PropertiesRepository::properties_id_type> pp = finish_element (layout.properties_repository ());
    if (pp.first) {
      cell.shapes (ll.second).insert (db::TextRefWithProperties (db::TextRef (text, layout.shape_repository ()), pp.second));
    } else {
      cell.shapes (ll.second).insert (db::TextRef (text, layout.shape_repository ()));
    }

  } else {
    finish_element ();
  }
}

void 
GDS2Reader::read_box (db::Layout &layout, db::Cell &cell)
{
  LDPair ld; 
  short rec_id = 0;

  do {
    rec_id = get_record ();
  } while (rec_id == sELFLAGS || rec_id == sPLEX);
  if (rec_id != sLAYER) {
    error (tl::translate ("LAYER record expected"));
  }
  ld.layer = get_ushort ();
  if (get_record () != sBOXTYPE) {
    error (tl::translate ("DATATYPE record expected"));
  }
  ld.datatype = get_ushort ();

  std::pair<bool, unsigned int> ll = open_dl (layout, ld, m_create_layers);
  if (ll.first) {

    if (get_record () != sXY) {
      error (tl::translate ("XY record expected"));
    }

    GDS2XY *xy = (GDS2XY *) mp_rec_buf;
    GDS2XYToPoint pt_conv;
    db::Box box;
    while (xy < (GDS2XY *) (mp_rec_buf + m_reclen)) {
      box += pt_conv (*xy++);
    }

    std::pair<bool, db::PropertiesRepository::properties_id_type> pp = finish_element (layout.properties_repository ());
    if (! box.empty ()) {
      if (pp.first) {
        cell.shapes (ll.second).insert (db::BoxWithProperties (box, pp.second));
      } else {
        cell.shapes (ll.second).insert (box);
      }
    }

  } else {
    finish_element ();
  }
}

void 
GDS2Reader::read_ref (db::Layout &layout, db::Cell &cell, bool array, tl::vector<db::CellInstArray> &instances, tl::vector<db::CellInstArrayWithProperties> &instances_with_props)
{
  short rec_id = 0;

  do {
    rec_id = get_record ();
  } while (rec_id == sELFLAGS || rec_id == sPLEX);
  if (rec_id != sSNAME) {
    error (tl::translate ("SNAME record expected"));
  }

  db::Layout::cell_index_type ci;

  {
    //  obtain cell or create new: "cn" is not valid beyond the 
    //  scope of the get_string() call
    const char *cn = get_string ();
    std::pair<bool, db::Layout::cell_index_type> c = layout.cell_by_name (cn); 
    if (c.first) {
      //  cell already there: just add shapes (cell might have been created through forward reference)
      ci = c.second;
    } else {
      ci = layout.add_cell (cn);
    }
  }

  bool mirror = false;
  int angle = 0;
  double angle_deg = 0.0;
  double mag = 1.0;
  bool is_mag = false;

  rec_id = get_record ();
  if (rec_id == sSTRANS) {
    short f = get_short ();
    if ((f & 0x8000) != 0) {
      mirror = true;
    }
    if ((f & (4 | 2)) != 0) {
      error (tl::translate ("Absolute transformations are supported"));
    }  
    rec_id = get_record ();
  }

  if (rec_id == sMAG) {
    mag = get_double ();
    if (fabs (mag - 1.0) > 1e-9) { 
      is_mag = true;
    }
    rec_id = get_record ();
  }

  if (rec_id == sANGLE) {
    angle_deg = get_double ();
    double a = angle_deg / 90.0;
    if (a < -4 || a > 4) {
      error (tl::sprintf (tl::translate ("Invalid rotation angle (%g is less than -360 or larger than 360)").c_str (), angle_deg));
    }
    angle = int (a < 0 ? (a - 0.5) : (a + 0.5));
    if (fabs (double (angle) - a) > 1e-9) {
      angle = -1; // indicates arbitrary orientation. Take angle_deg instead
    } else {
      if (angle < 0) {
        angle += ((4 - 1) - angle) & ~(4 - 1);
      }
      angle = angle % 4;
    }
    rec_id = get_record ();
  }

  db::CellInstArray inst;

  if (array) {

    //  Array reference
    if (rec_id != sCOLROW) {
      error (tl::translate ("COLROW record expected"));
    }

    int cols = get_ushort ();
    int rows = get_ushort ();

    //  Array reference
    if (get_record () != sXY) {
      error (tl::translate ("XY record expected"));
    }

    //  Create the instance
    db::Coord x = get_int ();
    db::Coord y = get_int ();
    db::Coord xc = get_int () - x;
    db::Coord yc = get_int () - y;
    db::Coord xr = get_int () - x;
    db::Coord yr = get_int () - y;

    if (cols > 1) {
      if (xc % cols != 0 || yc % cols != 0) {
        warn (tl::translate ("Off-grid AREF column vector"));
      }
      xc /= cols;
      yc /= cols;
    }

    if (rows > 1) {
      if (xr % rows != 0 || yr % rows != 0) {
        warn (tl::translate ("Off-grid AREF row vector"));
      }
      xr /= rows;
      yr /= rows;
    }

    //  insert the cell array
    if (is_mag || angle < 0) {
      inst = db::CellInstArray (db::CellInst (ci), 
                                db::CplxTrans (mag, angle_deg, mirror, db::Point (x, y)), db::Point (xc, yc), db::Point (xr, yr), cols, rows);
    } else {
      inst = db::CellInstArray (db::CellInst (ci), 
                                db::Trans (angle, mirror, db::Point (x, y)), db::Point (xc, yc), db::Point (xr, yr), cols, rows);
    }

  } else {

    //  Single reference
    if (rec_id != sXY) {
      error (tl::translate ("XY record expected"));
    }

    //  Create the instance
    db::Coord x = get_int ();
    db::Coord y = get_int ();

    if (is_mag || angle < 0) {
      inst = db::CellInstArray (db::CellInst (ci), db::CplxTrans (mag, angle_deg, mirror, db::Point (x, y)));
    } else {
      inst = db::CellInstArray (db::CellInst (ci), db::Trans (angle, mirror, db::Point (x, y)));
    }

  }  

  std::pair<bool, db::PropertiesRepository::properties_id_type> pp = finish_element (layout.properties_repository ());
  if (pp.first) {
    instances_with_props.push_back (db::CellInstArrayWithProperties (inst, pp.second));
  } else {
    instances.push_back (inst);
  }
}


}

