
/*

  KLayout Layout Viewer
  Copyright (C) 2006-2007 Matthias Koefferlein

  This program is free software; you can redistribute it and/or modify
  it under the terms of the GNU General Public License as published by
  the Free Software Foundation; either version 2 of the License, or
  (at your option) any later version.

  This program is distributed in the hope that it will be useful,
  but WITHOUT ANY WARRANTY; without even the implied warranty of
  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
  GNU General Public License for more details.

  You should have received a copy of the GNU General Public License
  along with this program; if not, write to the Free Software
  Foundation, Inc., 51 Franklin St, Fifth Floor, Boston, MA  02110-1301  USA

*/


#include "dbLayout.h"
#include "dbShape.h"
#include "tlStream.h"
#include "tlAssert.h"
#include "tlException.h"
#include "dbGDS2Writer.h"
#include "dbGDS2.h"
#include "dbSaveLayoutOptions.h"

#include <stdio.h>
#include <errno.h>

#include <limits>

namespace db
{

// ------------------------------------------------------------------
//  GDS2Writer implementation

GDS2Writer::GDS2Writer ()
{
  //  .. nothing yet ..
}

static bool 
is_valid_prop (const db::LayerProperties &prop, void *)
{
  return prop.layer >= 0 && prop.datatype >= 0;
}

static int safe_scale (double sf, int value)
{
  double i = floor (sf * value + 0.5);
  if (i < double (std::numeric_limits<int>::min ())) {
    throw tl::Exception ("Scaling failed: coordinate underflow");
  }
  if (i > double (std::numeric_limits<int>::max ())) {
    throw tl::Exception ("Scaling failed: coordinate overflow");
  }
  return int (i);
}

inline int scale (double sf, int value)
{
  if (sf == 1.0) {
    return value;
  } else {
    return safe_scale (sf, value);
  }
}

void
GDS2Writer::write (const db::Layout &layout, tl::OutputStream &stream, const db::SaveLayoutOptions &options) throw (tl::Exception)
{
  mp_stream = &stream;

  double dbu = (options.dbu () == 0.0) ? layout.dbu () : options.dbu ();
  double sf = options.scale_factor () * (layout.dbu () / dbu);
  if (fabs (sf - 1.0) < 1e-9) {
    //  to avoid rounding problems, set to 1.0 exactly if possible.
    sf = 1.0;
  }

  std::vector <std::pair <unsigned int, db::LayerProperties> > layers;
  options.get_valid_layers (layout, layers, &is_valid_prop);

  std::vector <db::CellGraph::cell_index_type> cells;
  options.get_cells (layout, cells);

  //  write header

  write_short (6);
  write_short (sHEADER);
  write_short (600);

  write_short (4 + 12 * 2);
  write_short (sBGNLIB);
  write_time ();
  write_time ();

  std::string libname ("LIB");
  write_string_record (sLIBNAME, libname);

  write_short (4 + 8 * 2);
  write_short (sUNITS);
  write_double (dbu);
  write_double (dbu * 1e-6);

  //  body

  for (std::vector<db::CellGraph::cell_index_type>::const_iterator cell = cells.begin (); cell != cells.end (); ++cell) {

    //  cell header 

    write_short (4 + 12 * 2);
    write_short (sBGNSTR);
    write_time ();
    write_time ();

    write_string_record (sSTRNAME, layout.cell_name (*cell));

    //  cell body 

    const db::Cell &cref (layout.cell (*cell));

    //  instances
    
    for (db::Cell::const_iterator inst = cref.begin (); ! inst.at_end (); ++inst) {

      db::Point a, b;
      unsigned long amax, bmax;

      bool is_reg = inst->is_regular_array (a, b, amax, bmax);

      write_short (4);
      write_short (is_reg ? sAREF : sSREF);

      write_string_record (sSNAME, layout.cell_name (inst->cell_index ()));

      db::Trans t = inst->front ();

      if (t.rot () != 0 || inst->is_complex ()) {

        write_short (6);
        write_short (sSTRANS);
        write_short (t.is_mirror () ? 0x8000 : 0);

        if (inst->is_complex ()) {
          write_short (4 + 8);
          write_short (sMAG);
          write_double (inst->complex_trans ().mag ());
          write_short (4 + 8);
          write_short (sANGLE);
          write_double (inst->complex_trans ().angle ());
        } else {
          if ((t.rot () % 4) != 0) {
            write_short (4 + 8);
            write_short (sANGLE);
            write_double ((t.rot () % 4) * 90.0);
          }
        }

      }

      if (is_reg) {
        write_short (4 + 2 * 2);
        write_short (sCOLROW);
        write_short (std::max ((unsigned long) 1, amax));
        write_short (std::max ((unsigned long) 1, bmax));
      }

      write_short (4 + (is_reg ? 3 : 1) * 2 * 4);
      write_short (sXY);
      write_int (scale (sf, t.disp ().x ()));
      write_int (scale (sf, t.disp ().y ()));

      if (is_reg) {
        write_int (scale (sf, t.disp ().x () + a.x () * amax));
        write_int (scale (sf, t.disp ().y () + a.y () * amax));
        write_int (scale (sf, t.disp ().x () + b.x () * bmax));
        write_int (scale (sf, t.disp ().y () + b.y () * bmax));
      }

      if (inst->has_prop_id () && inst->prop_id () != 0) {

        const db::PropertiesRepository::properties_set &props = layout.properties_repository ().properties (inst->prop_id ());
        for (db::PropertiesRepository::properties_set::const_iterator p = props.begin (); p != props.end (); ++p) {

          const tl::Variant &name = layout.properties_repository ().prop_name (p->first);

          long attr = -1;
          if (name.is_long ()) {
            attr = name.to_long ();
          }

          if (attr >= 0 && attr < 65535) {

            write_short (6);
            write_short (sPROPATTR);
            write_short (attr);

            write_string_record (sPROPVALUE, p->second.to_string ());

          }

        }

      }

      write_short (4);
      write_short (sENDEL);

    }

    //  shapes

    for (std::vector <std::pair <unsigned int, db::LayerProperties> >::const_iterator l = layers.begin (); l != layers.end (); ++l) {
 
      if (layout.is_valid_layer (l->first)) {

        int layer = l->second.layer;
        int datatype = l->second.datatype;

        db::ShapeIterator shape (cref.shapes (l->first).begin (db::ShapeIterator::Boxes | db::ShapeIterator::Polygons | db::ShapeIterator::Edges | db::ShapeIterator::Paths | db::ShapeIterator::Texts));
        while (! shape.at_end ()) {

          bool any = true;

          if (shape->is_text ()) {
            
            db::Trans trans = shape->text_trans ();
          
            write_short (4);
            write_short (sTEXT);

            write_short (4 + 2);
            write_short (sLAYER);
            write_short (layer);

            write_short (4 + 2);
            write_short (sTEXTTYPE);
            write_short (datatype);

            if (trans.rot () != 0) {

              write_short (6);
              write_short (sSTRANS);
              write_short (trans.is_mirror () ? 0x8000 : 0);

              if ((trans.rot () % 4) != 0) {
                write_short (4 + 8);
                write_short (sANGLE);
                write_double ((trans.rot () % 4) * 90.0);
              }

            }

            write_short (4 + 2 * 4);
            write_short (sXY);
            write_int (scale (sf, trans.disp ().x ()));
            write_int (scale (sf, trans.disp ().y ()));

            write_string_record (sSTRING, shape->text_string ().c_str ());

          } else if (shape->is_polygon ()) {

            //  @@@ how to deal with holes, max vertex count?
            if (shape->holes () > 0) {
              throw tl::Exception (tl::translate ("Cannot write polygons with holes to GDS2 streams"));
            }

            size_t n = 0;
            for (db::Shape::point_iterator e = shape->begin_hull (); e != shape->end_hull (); ++e) {
              ++n;
            }
            size_t nmax = 8000; // @@@
            if (n > nmax) {
              throw tl::Exception (tl::translate ("Cannot write polygons with more than %u points to GDS2 streams").c_str (), nmax);
            }

            if (n > 0) {

              write_short (4);
              write_short (sBOUNDARY);

              write_short (4 + 2);
              write_short (sLAYER);
              write_short (layer);

              write_short (4 + 2);
              write_short (sDATATYPE);
              write_short (datatype);

              write_short (4 + (n + 1) * 2 * 4);
              write_short (sXY);

              db::Shape::point_iterator e (shape->begin_hull ());
              for ( ; e != shape->end_hull (); ++e) {
                write_int (scale (sf, (*e).x ()));
                write_int (scale (sf, (*e).y ()));
              }

              //  and closing point
              e = shape->begin_hull ();
              if (e != shape->end_hull ()) {
                write_int (scale (sf, (*e).x ()));
                write_int (scale (sf, (*e).y ()));
              }

            } else {
              //  don't write polygons with no points
              any = false;
            }

          } else if (shape->is_edge ()) {

            db::Edge e (shape->edge ());

            write_short (4);
            write_short (sPATH);

            write_short (4 + 2);
            write_short (sLAYER);
            write_short (layer);

            write_short (4 + 2);
            write_short (sDATATYPE);
            write_short (datatype);

            write_short (4 + 2);
            write_short (sPATHTYPE);
            write_short (0);

            write_short (4 + 4);
            write_short (sWIDTH);
            write_int (0);

            write_short (4 + 2 * 2 * 4);
            write_short (sXY);
            write_int (scale (sf, e.p1 ().x ()));
            write_int (scale (sf, e.p1 ().y ()));
            write_int (scale (sf, e.p2 ().x ()));
            write_int (scale (sf, e.p2 ().y ()));

          } else if (shape->is_path ()) {

            //  instantiate the path and draw
            db::Path path;
            shape->path (path);

            write_short (4);
            write_short (sPATH);

            write_short (4 + 2);
            write_short (sLAYER);
            write_short (layer);

            write_short (4 + 2);
            write_short (sDATATYPE);
            write_short (datatype);

            short type = 0;
            db::Coord w = path.width ();
            db::Coord se = path.extensions ().first;
            db::Coord ee = path.extensions ().second;

            if (se == w / 2 && ee == w / 2) {
              type = path.round () ? 1 : 2;
            } else if (se == 0 && ee == 0) {
              type = 0;
            } else {
              type = 4;
            }

            write_short (4 + 2);
            write_short (sPATHTYPE);
            write_short (type);

            write_short (4 + 4);
            write_short (sWIDTH);
            write_int (scale (sf, w));

            if (type == 4) {

              write_short (4 + 4);
              write_short (sBGNEXTN);
              write_int (scale (sf, se));

              write_short (4 + 4);
              write_short (sENDEXTN);
              write_int (scale (sf, ee));

            }

            write_short (4 + path.points () * 2 * 4);
            write_short (sXY);
            for (db::Path::iterator p = path.begin (); p != path.end (); ++p) {
              write_int (scale (sf, (*p).x ()));
              write_int (scale (sf, (*p).y ()));
            }

          } else if (shape->is_box ()) {

            db::Box box (shape->box ());

            write_short (4);
            write_short (sBOUNDARY);

            write_short (4 + 2);
            write_short (sLAYER);
            write_short (layer);

            write_short (4 + 2);
            write_short (sDATATYPE);
            write_short (datatype);

            write_short (4 + 5 * 2 * 4);
            write_short (sXY);
            write_int (scale (sf, box.left ()));
            write_int (scale (sf, box.bottom ()));
            write_int (scale (sf, box.left ()));
            write_int (scale (sf, box.top ()));
            write_int (scale (sf, box.right ()));
            write_int (scale (sf, box.top ()));
            write_int (scale (sf, box.right ()));
            write_int (scale (sf, box.bottom ()));
            write_int (scale (sf, box.left ()));
            write_int (scale (sf, box.bottom ()));

          } else {
            any = false;
          }

          if (any) {

            if (shape->has_prop_id () && shape->prop_id () != 0) {

              const db::PropertiesRepository::properties_set &props = layout.properties_repository ().properties (shape->prop_id ());
              for (db::PropertiesRepository::properties_set::const_iterator p = props.begin (); p != props.end (); ++p) {

                const tl::Variant &name = layout.properties_repository ().prop_name (p->first);

                long attr = -1;
                if (name.is_long ()) {
                  attr = name.to_long ();
                }

                if (attr >= 0 && attr < 65535) {

                  write_short (6);
                  write_short (sPROPATTR);
                  write_short (attr);

                  write_string_record (sPROPVALUE, p->second.to_string ());

                }

              }

            }

            write_short (4);
            write_short (sENDEL);

          }

          ++shape;

        }

      }

    }

    //  end of cell

    write_short (4);
    write_short (sENDSTR);

  }

  write_short (4);
  write_short (sENDLIB);
}

void 
GDS2Writer::write_byte (unsigned char b)
{
  mp_stream->put ((const char *) &b, 1);
}

void 
GDS2Writer::write_short (int16_t i)
{
  char b[2];
  //  @@@ this can be done more efficiently?
  b[0] = ((i >> 8) & 0xff);
  b[1] = (i & 0xff);
  mp_stream->put (b, sizeof (b));
}

void 
GDS2Writer::write_int (int32_t l)
{
  char b[4];
  //  @@@ this can be done more efficiently?
  b[0] = ((l >> 24) & 0xff);
  b[1] = ((l >> 16) & 0xff);
  b[2] = ((l >> 8) & 0xff);
  b[3] = (l & 0xff);
  mp_stream->put (b, sizeof (b));
}

void 
GDS2Writer::write_double (double d)
{
  char b[8];
  
  b[0] = 0;
  if (d < 0) {
    b[0] = char (0x80);
    d = -d;
  }

  //  compute the next power of 16 that that value will fit in
  int e = 0;
  if (d < 1e-77 /*~16^-64*/) {
    d = 0;
  } else {
    double lg16 = log (d) / log (16.0);
    e = int (ceil (log (d) / log (16.0)));
    if (e == lg16) {
      ++e;
    }
  }

  d /= pow (16.0, e - 14);

  tl_assert (e >= -64 && e < 64);
  b[0] |= ((e + 64) & 0x7f);

  uint64_t m = uint64_t (d + 0.5);
  for (int i = 7; i > 0; --i) {
    b[i] = (m & 0xff);
    m >>= 8;
  }

  mp_stream->put (b, sizeof (b));
}

void 
GDS2Writer::write_time ()
{
  //  @@@ not implemented yet
  for (unsigned int i = 0; i < 6; ++i) {
    write_short (0);
  }
}

void 
GDS2Writer::write_string (const char *t)
{
  size_t l = strlen (t);
  mp_stream->put (t, l);
  if ((l & 1) != 0) {
    write_byte (0);
  }
}

void 
GDS2Writer::write_string (const std::string &t)
{
  size_t l = t.size ();
  mp_stream->put (t.c_str (), l);
  if ((l & 1) != 0) {
    write_byte (0);
  }
}

void 
GDS2Writer::write_string_record (short record, const std::string &t)
{
  write_short (4 + ((t.size () + 1) / 2) * 2);
  write_short (record);
  write_string (t);
}


} // namespace db

