
/*

  KLayout Layout Viewer
  Copyright (C) 2006-2007 Matthias Koefferlein

  This program is free software; you can redistribute it and/or modify
  it under the terms of the GNU General Public License as published by
  the Free Software Foundation; either version 2 of the License, or
  (at your option) any later version.

  This program is distributed in the hope that it will be useful,
  but WITHOUT ANY WARRANTY; without even the implied warranty of
  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
  GNU General Public License for more details.

  You should have received a copy of the GNU General Public License
  along with this program; if not, write to the Free Software
  Foundation, Inc., 51 Franklin St, Fifth Floor, Boston, MA  02110-1301  USA

*/


#ifndef HDR_dbInstances
#define HDR_dbInstances

#include "dbManager.h"
#include "dbStatic.h"
#include "dbBoxTree.h"
#include "dbBoxConvert.h"
#include "dbArray.h"
#include "dbPropertiesRepository.h"
#include "tlVector.h"
#include "tlUtils.h"

namespace db
{

class CellGraph;
template <class Graph> class cell;
template <class Graph> class instances;
template <class CellInst, class ArrayTrans> class cell_inst_index_box_convert;

/**
 *  @brief A base class for instance operations 
 *
 *  This class is used for the Op classes for the undo/redo queuing mechanism.
 */
template <class I>
class inst_op_base
  : public db::Op
{
public:
  inst_op_base () : db::Op () { }

  virtual void undo (I *instances) = 0;
  virtual void redo (I *instances) = 0;
};

/**
 *  @brief A undo/redo queue object for the instances
 *
 *  This class is used internally to queue an insert or erase operation
 *  into the db::Object manager's undo/redo queue.
 */
template <class Instances, class Inst>
class inst_op
  : public inst_op_base<Instances> 
{
public:
  typedef typename Instances::box_type box_type;
  typedef db::box_convert<Inst> box_convert_type;
  typedef db::box_tree<box_type, Inst, box_convert_type> tree_type;

  inst_op (bool insert, const Inst &sh)
    : m_insert (insert)
  {
    m_insts.push_back (sh);
  }
  
  template <class Iter>
  inst_op (bool insert, Iter from, Iter to)
    : m_insert (insert)
  {
    m_insts.insert (m_insts.end (), from, to);
  }

  template <class Iter>
  inst_op (bool insert, Iter from, Iter to, bool dummy)
    : m_insert (insert)
  {
    m_insts.reserve (std::distance (from, to));
    for (Iter i = from; i != to; ++i) {
      m_insts.push_back (**i);
    }
  }

  virtual void undo (Instances *insts)
  {
    if (m_insert) {
      erase (insts);
    } else {
      insert (insts);
    }
  }

  virtual void redo (Instances *insts)
  {
    if (m_insert) {
      insert (insts);
    } else {
      erase (insts);
    }
  }

private:
  bool m_insert;
  std::vector<Inst> m_insts;

  void insert (Instances *insts)
  {
    insts->insert (m_insts.begin (), m_insts.end ());
  }

  void erase (Instances *insts)
  {
    if (size_t (std::distance (((const Instances *) insts)->inst_tree (typename Inst::tag ()).begin (), ((const Instances *) insts)->inst_tree (typename Inst::tag ()).end ())) <= m_insts.size ()) {
      //  If all shapes are to be removed, just clear the instances
      insts->clear (typename Inst::tag ());
    } else {
      std::sort (m_insts.begin (), m_insts.end ());
      //  This is not quite effective but seems to be the simpliest way
      //  of implementing this: search for each element and erase these.
      std::vector<typename tree_type::const_iterator> to_erase;
      to_erase.reserve (m_insts.size ());
      for (typename tree_type::const_iterator linst = ((const Instances *) insts)->inst_tree (typename Inst::tag ()).begin (); linst != ((const Instances *) insts)->inst_tree (typename Inst::tag ()).end (); ++linst) {
        typename tree_type::const_iterator i = std::lower_bound (m_insts.begin (), m_insts.end (), *linst);
        if (i != m_insts.end () && *i == *linst) {
          to_erase.push_back (linst);
        }
      }
      insts->erase_positions (typename Inst::tag (), to_erase.begin (), to_erase.end ());
    }
  }
};

/**
 *  @brief The cell inst array per-layer box converter
 *
 *  This class is used as a function to convert a cell instance array
 *  to a box for a given layer. This requires that the per-layer 
 *  bounding boxes of the cell have been computed already.
 */

template <class CellInst, class ArrayTrans>
struct cell_inst_index_box_convert 
{
  typedef db::array <CellInst, ArrayTrans> cell_inst_array_type;
  typedef typename CellInst::graph_type graph_type;
  typedef typename CellInst::box_type box_type;
  typedef cell<graph_type> cell_type;

  typedef db::complex_bbox_tag complexity;

  cell_inst_index_box_convert (const graph_type &g, const cell_type &c, unsigned int l)
    : m_bc (g, l), mp_cell (&c)
  { }
  
  cell_inst_index_box_convert (const graph_type &g, const cell_type &c)
    : m_bc (g), mp_cell (&c)
  { }
  
  box_type operator() (size_t t) const
  {
    return box_type (mp_cell->begin (t)->bbox (m_bc));
  }

private:
  box_convert <CellInst> m_bc;
  const cell_type *mp_cell;
};

/**
 *  @brief A parent instance
 *
 *  A parent instance is basically an inverse instance: instead of pointing
 *  to the child cell, it is pointing to the parent cell and the transformation
 *  is representing the shift of the parent cell relative to the child cell.
 *  For memory performance, a parent instance is not stored as a instance but
 *  rather as a reference to a child instance and a reference to the cell which
 *  is the parent.
 *  The parent instance itself is computed on the fly. It is representative for
 *  a set of instances belonging to the same cell index. The special parent instance
 *  iterator takes care of producing the right sequence.
 */

template <class Graph>
class parent_inst
{
public: 
  typedef Graph graph_type;
  typedef typename Graph::cell_index_type cell_index_type;

  /**
   *  @brief The default ctor
   *
   *  This creates an invalid instance.
   */
  parent_inst ()
  {
    m_parent_cell_index = (cell_index_type) -1;
    m_index = 0;
  }

  /**
   *  @brief Create a parent instance using the parent cell and instance
   */
  parent_inst (cell_index_type parent_cell_index, size_t index)
    : m_parent_cell_index (parent_cell_index), m_index (index)
  {
    //  .. nothing yet ..
  }

  /**
   *  @brief Retrieve the reference to the parent cell 
   */
  cell_index_type parent_cell_index () const
  {
    return m_parent_cell_index;
  }

  /**
   *  @brief Access to the item index
   *
   *  This method is used mainly by the parent instance iterator
   */
  size_t index () const
  {
    return m_index;
  }

  /**
   *  @brief Equality
   */
  bool operator== (const parent_inst &d) const
  {
    return m_parent_cell_index == d.m_parent_cell_index && m_index == d.m_index;
  }

protected:
  cell_index_type m_parent_cell_index;
  size_t m_index;
};

/**
 *  @brief A parent instance representation
 *
 *  This object represents a parent instance that can be dereferenced standalone.
 *  For this, it requires a "Graph" object pointer for dereferencing.
 */

template <class Graph>
class parent_inst_rep
  : public parent_inst<Graph>
{
public: 
  typedef Graph graph_type;
  typedef typename parent_inst<Graph>::cell_index_type cell_index_type;
  typedef cell<Graph> cell_type;
  typedef typename Graph::box_type box_type;
  typedef typename box_type::coord_type coord_type;
  typedef db::simple_trans<coord_type> array_trans;
  typedef db::cell_inst<graph_type> cell_inst_type;
  typedef db::array<cell_inst_type, array_trans> cell_inst_array_type;

  /**
   *  @brief The default ctor
   */
  parent_inst_rep (const Graph *graph = 0)
    : parent_inst<Graph> (), mp_graph (graph)
  {
    //  .. nothing yet ..
  }

  /**
   *  @brief Create a representation from a parent_inst object
   */
  parent_inst_rep &operator= (const parent_inst<Graph> &pi)
  {
    parent_inst<Graph>::operator= (pi);
    return *this;
  }

  /**
   *  @brief Retrieve the child instance associated with this parent instance
   *
   *  Return an iterator pointing to the respective child instance.
   */
  typename cell_type::const_iterator child_inst () const
  {
    return mp_graph->cell (this->m_parent_cell_index).sorted_inst_ptr (this->m_index);
  }

  /**
   *  @brief Compute the inverse instance by which the parent is seen from the child
   */
  cell_inst_array_type inst () const
  {
    //  create a new parent instance by cloning and inverting the array
    cell_inst_array_type ci (child_inst ()->cell_inst ());
    ci.object () = cell_inst_type (this->m_parent_cell_index);
    ci.invert ();
    return ci;
  }

  /**
   *  @brief Increment the index
   */
  void inc ()
  {
    ++this->m_index;
  }

private:
  const Graph *mp_graph;
};

/**
 *  @brief A parent instance iterator
 *
 *  The iterator is first an iterator over a vector of parent instances but then
 *  takes care of decomposing a single instance into a set of such.
 */

template <class Graph>
class parent_inst_iterator
{
public: 
  typedef Graph graph_type;
  typedef typename Graph::cell_index_type cell_index_type;
  typedef parent_inst<Graph> parent_inst_type;
  typedef db::cell<graph_type> cell_type;
  typedef parent_inst_rep<Graph> value_type;
  typedef const value_type *pointer; 
  typedef value_type reference;   //  operator* returns a value
  typedef std::forward_iterator_tag iterator_category;
  typedef void difference_type;

  /**
   *  @brief Default ctor
   */
  parent_inst_iterator ()
    : mp_graph (0), m_iter (), m_end ()
  { }

  /**
   *  @brief Constructor
   *
   *  This creates an iterator pointing to the start of the sequence represented
   *  by the parent instance object at *iter or at the end of the parent instance
   *  vector.
   */
  parent_inst_iterator (const Graph *graph, typename tl::vector<parent_inst_type>::const_iterator iter, typename tl::vector<parent_inst_type>::const_iterator end)
    : mp_graph (graph), m_iter (iter), m_end (end), m_rep (graph)
  {
    if (m_iter != m_end) {
      m_rep = parent_inst_type (m_iter->parent_cell_index (), m_iter->index ());
    }
  }

  /** 
   *  @brief Equality operator
   */
  bool operator== (const parent_inst_iterator &b) const
  {
    return m_iter == b.m_iter && m_rep == b.m_rep;
  }

  /** 
   *  @brief Inequality operator
   */
  bool operator!= (const parent_inst_iterator &b) const
  {
    return !operator==(b);
  }

  /**
   *  @brief Increment operator
   */
  parent_inst_iterator &operator++() 
  {
    cell_index_type ci = m_rep.child_inst ()->cell_index ();
    m_rep.inc ();

    if (m_rep.index () == mp_graph->cell (m_rep.parent_cell_index ()).cell_instances () ||
        m_rep.child_inst ()->cell_index () != ci) {
      ++m_iter;
      if (m_iter != m_end) {
        m_rep = parent_inst_type (m_iter->parent_cell_index (), m_iter->index ());
      } else {
        m_rep = parent_inst_type ();
      }
    }

    return *this;
  }

  /**
   *  @brief Access to the parent instance
   *
   *  In order not to return a reference to the temporary instance, the reference is returned by value.
   *  This is important for the GSI binding since this way, a copy is created which is required when
   *  the iterator moves on.
   */
  value_type operator* () const
  {
    return m_rep;
  }

  /**
   *  @brief Access to the parent instance
   * 
   *  See above.
   */
  const value_type *operator-> () const
  {
    return &m_rep;
  }

  /**
   *  @brief Returns true, if the iterator is at the end
   */
  bool at_end () const
  {
    return m_iter == m_end;
  }
  
private:
  const Graph *mp_graph;
  typename tl::vector<parent_inst_type>::const_iterator m_iter, m_end;
  value_type m_rep;
};

/**
 *  @brief A parent cell iterator
 *
 *  The iterator will report just cell indices, not instances.
 */

template <class Graph>
class parent_cell_iterator
{
public: 
  typedef Graph graph_type;
  typedef typename Graph::cell_index_type cell_index_type;
  typedef parent_inst<Graph> parent_inst_type;
  typedef cell_index_type value_type;
  typedef void pointer;           //  no operator->
  typedef value_type reference;   //  operator* returns a value
  typedef std::forward_iterator_tag iterator_category;
  typedef void difference_type;

  /**
   *  @brief Default ctor
   */
  parent_cell_iterator ()
    : m_iter ()
  { }

  /**
   *  @brief Constructor
   */
  parent_cell_iterator (typename tl::vector<parent_inst_type>::const_iterator iter)
    : m_iter (iter)
  { }

  /** 
   *  @brief Equality operator
   */
  bool operator== (const parent_cell_iterator &b) const
  {
    return m_iter == b.m_iter;
  }

  /** 
   *  @brief Inequality operator
   */
  bool operator!= (const parent_cell_iterator &b) const
  {
    return m_iter != b.m_iter;
  }

  /**
   *  @brief Access to the actual instance
   *
   *  The actual instance is created on the fly by the update_rep method.
   *  This will create an artificial instance which takes into account the
   *  offset from the representative.
   */
  cell_index_type operator* () const
  {
    return m_iter->parent_cell_index ();
  }

  /**
   *  @brief Increment operator
   */
  parent_cell_iterator &operator++() 
  {
    ++m_iter;
    return *this;
  }

private:
  typename tl::vector<parent_inst_type>::const_iterator m_iter;
};

/**
 *  @brief A instance reference
 *
 *  In analogy to the "shape" reference (i.e. db::Shape), this instance
 *  reference points to a certain instance and provides an abstract interface.
 */

template <class Graph>
class instance
{
public:
  typedef Graph graph_type;
  typedef typename Graph::cell_index_type cell_index_type;
  typedef db::cell_inst<graph_type> cell_inst_type;
  typedef typename cell_inst_type::coord_type coord_type;
  typedef db::simple_trans<coord_type> array_trans;
  typedef db::array<cell_inst_type, array_trans> cell_inst_array_type;
  typedef db::object_with_properties<cell_inst_array_type> cell_inst_wp_array_type;

  /**
   *  @brief Initialize a reference with "nil"
   */
  instance ()
    : mp_inst (0), mp_inst_wp (0)
  { }
  
  /**
   *  @brief Initialize a reference with an instance pointer
   */
  instance (const cell_inst_array_type &inst)
    : mp_inst (&inst), mp_inst_wp (0)
  { }
  
  /**
   *  @brief Initialize a reference with an pointer to an instance with properties
   */
  instance (const cell_inst_wp_array_type &inst)
    : mp_inst (0), mp_inst_wp (&inst)
  { }

  /**
   *  @brief Get the basic pointer for instances with properties
   *
   *  The pointer will be 0, if this is not an instance with properties.
   */
  const cell_inst_wp_array_type *basic_ptr (typename cell_inst_wp_array_type::tag tag) const
  {
    return mp_inst_wp;
  }

  /**
   *  @brief Get the basic pointer for normal instances
   *
   *  The pointer will be 0, if this is not a normal instance.
   */
  const cell_inst_array_type *basic_ptr (typename cell_inst_array_type::tag tag) const
  {
    return mp_inst;
  }

  /**
   *  @brief Get the properties ID in an abstract way
   */
  db::PropertiesRepository::properties_id_type prop_id () const
  {
    if (mp_inst_wp) {
      return mp_inst_wp->properties_id ();
    } else {
      return 0;
    }
  }

  /**
   *  @brief Test, if the iterator is of a certain type
   */
  bool has_prop_id () const
  {
    return mp_inst_wp != 0;
  }

  /**
   *  @brief Get the basic instance (without properties)
   */
  const cell_inst_array_type &cell_inst () const
  {
    tl_assert (mp_inst_wp || mp_inst);
    if (mp_inst_wp) {
      return *mp_inst_wp;
    } else {
      return *mp_inst;
    }
  }

  /**
   *  @brief Return the cell index of the reference
   *
   *  This method is basically provided for convenience
   *  and is equivalent to ->cell_inst ().object ().cell_index ().
   */
  cell_index_type cell_index () const
  {
    return cell_inst ().object ().cell_index ();
  }

  /**
   *  @brief Check if the instance array is a regular one and return the parameters
   *
   *  This method is basically provided for convenience
   *  and is equivalent to ->cell_inst ().is_regular_array ().
   */
  bool is_regular_array (typename cell_inst_array_type::point_type &a, typename cell_inst_array_type::point_type &b, unsigned long &amax, unsigned long &bmax) const 
  {
    return cell_inst ().is_regular_array (a, b, amax, bmax);
  }

  /**
   *  @brief Return the complex transformation of this instance
   *
   *  This method is basically provided for convenience
   *  and is equivalent to ->cell_inst ().complex_trans ().
   */
  typename cell_inst_array_type::complex_trans_type complex_trans () const
  {
    return cell_inst ().complex_trans ();
  }

  /**
   *  @brief Return true, if the transformation of this instance is complex
   *
   *  This method is basically provided for convenience
   *  and is equivalent to ->cell_inst ().is_complex ().
   */
  bool is_complex () const
  {
    return cell_inst ().is_complex ();
  }

  /**
   *  @brief Return the first transformation of this array
   *
   *  This method is basically provided for convenience
   *  and is equivalent to ->cell_inst ().front ().
   */
  const typename cell_inst_array_type::trans_type &front () const
  {
    return cell_inst ().front ();
  }

  /**
   *  @brief Return the number of flat instances represented by this array
   *
   *  This method is basically provided for convenience
   *  and is equivalent to ->cell_inst ().size ().
   */
  size_t size () const
  {
    return cell_inst ().size ();
  }

  /**
   *  @brief Return the bounding box of this array 
   *
   *  This method is basically provided for convenience
   *  and is equivalent to ->cell_inst ().bbox (bc).
   */
  typename cell_inst_array_type::box_type bbox (const db::box_convert<cell_inst_array_type> &bc) const
  {
    return cell_inst ().bbox (bc);
  }

  /**
   *  @brief Return the iterator for the instances of the array
   *
   *  This method is basically provided for convenience
   *  and is equivalent to ->cell_inst ().begin ().
   */
  typename cell_inst_array_type::iterator begin () const 
  {
    return cell_inst ().begin ();
  }

  /**
   *  @brief Return the region query iterator for the instances of the array
   *
   *  This method is basically provided for convenience
   *  and is equivalent to ->cell_inst ().begin_touching ().
   */
  typename cell_inst_array_type::iterator begin_touching (const typename cell_inst_array_type::box_type &b, const Graph *g) const 
  {
    db::box_convert<cell_inst_array_type> bc (*g);
    return cell_inst ().begin_touching (b, bc);
  }

  /**
   *  @brief Equality
   *
   *  Equality of instances is not specified by the identity of the objects but by the
   *  identity of the pointers - both instances must reference the same object.
   */
  bool operator== (const instance &d) const
  {
    return (mp_inst == d.mp_inst && mp_inst_wp == d.mp_inst_wp);
  }

  /**
   *  @brief Value equality
   *
   *  In contrast to the operator==, this method compares values rather than pointers.
   */
  bool equals (const instance &d) const
  {
    return prop_id () == d.prop_id () && cell_inst () == d.cell_inst ();
  }

  /**
   *  @brief Inequality
   *
   *  This compares the value, not the pointers.
   */
  bool operator!= (const instance &d) const
  {
    return !operator== (d);
  }

  /**
   *  @brief Less operator
   *
   *  Ordering of instances is not specified by the identity of the objects but by the
   *  order of the pointers.
   *  Since pointers are volatile objects, the ordering is not strictly reproducable!
   *  The order is designed such that different types of instances are separated in
   *  an order sequence of instance proxies. This is an important fact for the 
   *  "erase" functionality of the instances container when erasing a set of objects.
   */
  bool operator< (const instance &d) const
  {
    if (has_prop_id () != d.has_prop_id ()) {
      return (has_prop_id () < d.has_prop_id ());
    } 
    if (has_prop_id ()) {
      return mp_inst_wp < d.mp_inst_wp;
    } else {
      return mp_inst < d.mp_inst;
    }
  }

  /**
   *  @brief Comparison of values
   *
   *  In contrast to operator<, this method compares values rather than pointers.
   */
  bool less (const instance &d) const
  {
    if (prop_id () != d.prop_id ()) {
      return prop_id () < d.prop_id ();
    }
    if (cell_inst () != d.cell_inst ()) {
      return cell_inst () < d.cell_inst ();
    }
    return false;
  }

private:
  const cell_inst_array_type *mp_inst;
  const cell_inst_wp_array_type *mp_inst_wp;
};

/**
 *  @brief A generic iterator for the instances 
 *
 *  This generic iterator is specialized with the Traits object, which
 *  determines how to initialize the iterators and what types to use.
 */

template <class Graph, class Traits>
class instance_iterator
{
public:
  typedef Graph graph_type;
  typedef typename Graph::cell_index_type cell_index_type;
  typedef db::cell_inst<graph_type> cell_inst_type;
  typedef typename cell_inst_type::coord_type coord_type;
  typedef db::simple_trans<coord_type> array_trans;
  typedef db::array<cell_inst_type, array_trans> cell_inst_array_type;
  typedef db::object_with_properties<cell_inst_array_type> cell_inst_wp_array_type;
  typedef db::instance<Graph> value_type;
  typedef db::instances<Graph> instances_type;
  typedef typename Traits::iter_type iter_type;
  typedef typename Traits::iter_wp_type iter_wp_type;
  typedef const value_type *pointer; 
  typedef value_type reference;   //  operator* returns a value
  typedef std::forward_iterator_tag iterator_category;
  typedef void difference_type;

  /**
   *  @brief Default ctor
   */
  instance_iterator ()
    : m_iter ()
  { }

  /**
   *  @brief Constructor
   */
  instance_iterator (const Traits &traits)
  { 
    traits.init (m_iter, m_iter_wp);
    update_ref ();
  }

  /**
   *  @brief Constructor with offset
   */
  instance_iterator (const Traits &traits, size_t offset)
  { 
    traits.init (m_iter, m_iter_wp, offset);
    update_ref ();
  }

  /**
   *  @brief Equality
   */
  bool operator== (const instance_iterator &d) const
  {
    return m_iter == d.m_iter && m_iter_wp == d.m_iter_wp;
  }

  /**
   *  @brief Inequality
   */
  bool operator!= (const instance_iterator &d) const
  {
    return !operator== (d);
  }

  /**
   *  @brief Access to the actual instance
   *  
   *  HINT: this method returns a value, not a "const value &" because this way, 
   *  we do not loose much (it is not a expensive object) but gain more stability in the
   *  GSI interface, in which we need to be careful not to return references to temporary objects.
   */
  value_type operator* () const
  {
    return m_ref;
  }

  /**
   *  @brief Access to the actual instance (pointer)
   */
  const value_type *operator-> () const
  {
    return &m_ref;
  }

  /**
   *  @brief Increment operator
   */
  instance_iterator &operator++() 
  {
    if (! m_iter.at_end ()) {
      ++m_iter;
    } else if (! m_iter_wp.at_end ()) {
      ++m_iter_wp;
    }
    update_ref ();
    return *this;
  }

  /**
   *  @brief Report the end of the iterator (in non-STL schemes)
   */
  bool at_end () const
  {
    if (! m_iter.at_end ()) {
      return false;
    }
    if (! m_iter_wp.at_end ()) {
      return false;
    }
    return true;
  }

  /**
   *  @brief Get the const iterator for a given type
   */
  const iter_type &get_iter (typename cell_inst_array_type::tag tag) const
  {
    return m_iter; 
  }

  /**
   *  @brief Get the const iterator for a given type
   */
  const iter_wp_type &get_iter (typename cell_inst_wp_array_type::tag tag) const
  {
    return m_iter_wp;
  }

private:
  iter_type m_iter;
  iter_wp_type m_iter_wp;
  value_type m_ref;

  void update_ref ()
  {
    if (! m_iter.at_end ()) {
      m_ref = *m_iter;
    } else if (! m_iter_wp.at_end ()) {
      m_ref = *m_iter_wp;
    } else {
      m_ref = value_type ();
    }
  }
};

/**
 *  @brief Traits for the normal iterator
 */

template <class Graph>
struct normal_instance_iterator_traits
{
  typedef Graph graph_type;
  typedef typename Graph::box_type box_type;
  typedef typename box_type::coord_type coord_type;
  typedef db::cell_inst<graph_type> cell_inst_type;
  typedef db::simple_trans<coord_type> array_trans;
  typedef db::array<cell_inst_type, array_trans> cell_inst_array_type;
  typedef db::object_with_properties<cell_inst_array_type> cell_inst_wp_array_type;
  typedef db::box_convert<cell_inst_array_type> cell_inst_array_box_converter;
  typedef db::box_convert<cell_inst_wp_array_type> cell_inst_wp_array_box_converter;
  typedef db::box_tree<box_type, cell_inst_array_type, cell_inst_array_box_converter> cell_inst_tree_type;
  typedef db::box_tree<box_type, cell_inst_wp_array_type, cell_inst_wp_array_box_converter> cell_inst_wp_tree_type;
  typedef db::instances<Graph> instances_type;

  typedef tl::iterator_pair<typename cell_inst_tree_type::const_iterator> iter_type;
  typedef tl::iterator_pair<typename cell_inst_wp_tree_type::const_iterator> iter_wp_type;

  normal_instance_iterator_traits (const instances_type *insts)
    : mp_insts (insts)
  { }

  void init (iter_type &iter, iter_wp_type &iter_wp) const
  {
    typename cell_inst_array_type::tag tag = typename cell_inst_array_type::tag ();
    typename cell_inst_wp_array_type::tag tag_wp = typename cell_inst_wp_array_type::tag ();
    iter = iter_type (mp_insts->inst_tree (tag).begin (), mp_insts->inst_tree (tag).end ());
    iter_wp = iter_wp_type (mp_insts->inst_tree (tag_wp).begin (), mp_insts->inst_tree (tag_wp).end ());
  }
 
  void init (iter_type &iter, iter_wp_type &iter_wp, size_t offset) const
  {
    //  Hint: this initialization is required for SunCC
    typename cell_inst_array_type::tag tag = typename cell_inst_array_type::tag ();
    typename cell_inst_wp_array_type::tag tag_wp = typename cell_inst_wp_array_type::tag ();

    if (mp_insts->inst_tree (tag).size () > offset) {
      iter = iter_type (mp_insts->inst_tree (tag).begin () + offset, mp_insts->inst_tree (tag).end ());
      return;
    } else {
      iter = iter_type (mp_insts->inst_tree (tag).end (), mp_insts->inst_tree (tag).end ());
      offset -= mp_insts->inst_tree (tag).size ();
    }

    if (mp_insts->inst_tree (tag_wp).size () > offset) {
      iter_wp = iter_wp_type (mp_insts->inst_tree (tag_wp).begin () + offset, mp_insts->inst_tree (tag_wp).end ());
      return;
    } else {
      iter_wp = iter_wp_type (mp_insts->inst_tree (tag_wp).end (), mp_insts->inst_tree (tag_wp).end ());
      offset -= mp_insts->inst_tree (tag_wp).size ();
    }
  }
 
public:
  const instances_type *mp_insts;
};

/**
 *  @brief Traits for the touching iterator
 */

template <class Graph>
struct touching_instance_iterator_traits
{
  typedef Graph graph_type;
  typedef typename Graph::box_type box_type;
  typedef typename box_type::coord_type coord_type;
  typedef db::cell_inst<graph_type> cell_inst_type;
  typedef db::simple_trans<coord_type> array_trans;
  typedef db::array<cell_inst_type, array_trans> cell_inst_array_type;
  typedef db::object_with_properties<cell_inst_array_type> cell_inst_wp_array_type;
  typedef db::box_convert<cell_inst_array_type> cell_inst_array_box_converter;
  typedef db::box_convert<cell_inst_wp_array_type> cell_inst_wp_array_box_converter;
  typedef db::box_tree<box_type, cell_inst_array_type, cell_inst_array_box_converter> cell_inst_tree_type;
  typedef db::box_tree<box_type, cell_inst_wp_array_type, cell_inst_wp_array_box_converter> cell_inst_wp_tree_type;
  typedef db::instances<Graph> instances_type;

  typedef typename cell_inst_tree_type::touching_iterator iter_type;
  typedef typename cell_inst_wp_tree_type::touching_iterator iter_wp_type;

  touching_instance_iterator_traits (const instances_type *insts, const box_type &box, const Graph *graph)
    : mp_insts (insts), m_box (box), mp_graph (graph)
  { }

  void init (iter_type &iter, iter_wp_type &iter_wp) const
  {
    //  Hint: this initialization is required for SunCC
    typename cell_inst_array_type::tag tag = typename cell_inst_array_type::tag ();
    typename cell_inst_wp_array_type::tag tag_wp = typename cell_inst_wp_array_type::tag ();

    iter = mp_insts->inst_tree (tag).begin_touching (m_box, cell_inst_array_box_converter (*mp_graph));
    iter_wp = mp_insts->inst_tree (tag_wp).begin_touching (m_box, cell_inst_wp_array_box_converter (*mp_graph));
  }
 
  void init (iter_type &iter, iter_wp_type &iter_wp, size_t offset) const
  {
    tl_assert (false); // not supported
  }

public:
  const instances_type *mp_insts;
  box_type m_box;
  const Graph *mp_graph;
};

/**
 *  @brief Traits for the overlapping iterator
 */

template <class Graph>
struct overlapping_instance_iterator_traits
{
  typedef Graph graph_type;
  typedef typename Graph::box_type box_type;
  typedef typename box_type::coord_type coord_type;
  typedef db::cell_inst<graph_type> cell_inst_type;
  typedef db::simple_trans<coord_type> array_trans;
  typedef db::array<cell_inst_type, array_trans> cell_inst_array_type;
  typedef db::object_with_properties<cell_inst_array_type> cell_inst_wp_array_type;
  typedef db::box_convert<cell_inst_array_type> cell_inst_array_box_converter;
  typedef db::box_convert<cell_inst_wp_array_type> cell_inst_wp_array_box_converter;
  typedef db::box_tree<box_type, cell_inst_array_type, cell_inst_array_box_converter> cell_inst_tree_type;
  typedef db::box_tree<box_type, cell_inst_wp_array_type, cell_inst_wp_array_box_converter> cell_inst_wp_tree_type;
  typedef db::instances<Graph> instances_type;

  typedef typename cell_inst_tree_type::overlapping_iterator iter_type;
  typedef typename cell_inst_wp_tree_type::overlapping_iterator iter_wp_type;

  overlapping_instance_iterator_traits (const instances_type *insts, const box_type &box, const Graph *graph)
    : mp_insts (insts), m_box (box), mp_graph (graph)
  { }

  void init (iter_type &iter, iter_wp_type &iter_wp) const
  {
    //  Hint: this initialization is required for SunCC
    typename cell_inst_array_type::tag tag = typename cell_inst_array_type::tag ();
    typename cell_inst_wp_array_type::tag tag_wp = typename cell_inst_wp_array_type::tag ();

    iter = mp_insts->inst_tree (tag).begin_overlapping (m_box, cell_inst_array_box_converter (*mp_graph));
    iter_wp = mp_insts->inst_tree (tag_wp).begin_overlapping (m_box, cell_inst_wp_array_box_converter (*mp_graph));
  }
 
  void init (iter_type &iter, iter_wp_type &iter_wp, size_t offset) const
  {
    tl_assert (false); // not supported
  }

public:
  const instances_type *mp_insts;
  box_type m_box;
  const Graph *mp_graph;
};

/**
 *  @brief A child cell iterator
 *
 *  The iterator will report just cell indices, not instances.
 */

template <class Graph>
class child_cell_iterator
{
public: 
  typedef Graph graph_type;
  typedef typename Graph::cell_index_type cell_index_type;
  typedef db::cell_inst<graph_type> cell_inst_type;
  typedef cell_index_type value_type;
  typedef db::instances<Graph> instances_type;
  typedef typename instances_type::sorted_inst_index_iterator inst_iterator_type;
  typedef void pointer;                //  no operator->
  typedef cell_index_type reference;   //  operator* returns a value
  typedef std::forward_iterator_tag iterator_category;
  typedef void difference_type;

  /**
   *  @brief Default ctor
   */
  child_cell_iterator ()
    : m_iter (), m_end (), mp_insts (0)
  { }

  /**
   *  @brief Constructor
   */
  child_cell_iterator (const instances_type *insts)
    : m_iter (insts->begin_sorted_inst_indices ()),
      m_end (insts->end_sorted_inst_indices ()),
      mp_insts (insts)
  { }

  /**
   *  @brief Access to the actual instance
   *
   *  The actual instance is created on the fly by the update_rep method.
   *  This will create an artificial instance which takes into account the
   *  offset from the representative.
   */
  cell_index_type operator* () const
  {
    return mp_insts->begin (*m_iter)->cell_inst ().object ().cell_index ();
  }

  /**
   *  @brief Increment operator
   */
  child_cell_iterator &operator++() 
  {
    cell_index_type ci = operator* ();
    do {
      ++m_iter;
    } while (m_iter != m_end && operator* () == ci);
    return *this;
  }

  /**
   *  @brief Report the end of the iterator (in non-STL schemes)
   */
  bool at_end () const
  {
    return (m_iter == m_end);
  }

private:
  inst_iterator_type m_iter, m_end;
  const instances_type *mp_insts;
};

/**
 *  @brief A helper function that compares child instances
 */
template <class I>
struct cell_inst_compare_f
{
  bool operator() (const std::pair <const I *, size_t> &a, const std::pair <const I *, size_t> &b) const
  {
    return a.first->raw_less (*b.first);
  }
};

/**
 *  @brief A general instance list
 *
 *  This list encapsulates a couple of instances of various types.
 *  The generic iterators are used to access these
 */

template <class Graph>
class instances 
{
public:
  typedef Graph graph_type;
  typedef db::cell<Graph> cell_type;
  typedef typename Graph::box_type box_type;
  typedef typename box_type::coord_type coord_type;
  typedef db::cell_inst<graph_type> cell_inst_type;
  typedef db::simple_trans<coord_type> array_trans;
  typedef db::array<cell_inst_type, array_trans> cell_inst_array_type;
  typedef db::object_with_properties<cell_inst_array_type> cell_inst_wp_array_type;
  typedef typename Graph::cell_index_type cell_index_type;
  typedef db::box_convert<cell_inst_array_type> cell_inst_array_box_converter;
  typedef db::box_convert<cell_inst_wp_array_type> cell_inst_wp_array_box_converter;
  typedef db::box_tree<box_type, cell_inst_array_type, cell_inst_array_box_converter> cell_inst_tree_type;
  typedef db::box_tree<box_type, cell_inst_wp_array_type, cell_inst_wp_array_box_converter> cell_inst_wp_tree_type;
  typedef instance_iterator<graph_type, normal_instance_iterator_traits<graph_type> > const_iterator;
  typedef instance_iterator<graph_type, overlapping_instance_iterator_traits<graph_type> > overlapping_iterator;
  typedef instance_iterator<graph_type, touching_instance_iterator_traits<graph_type> > touching_iterator;
  typedef db::parent_inst<graph_type> parent_inst_type;
  typedef tl::vector<parent_inst_type> parent_inst_vector;
  typedef db::parent_inst_iterator<graph_type> parent_inst_iterator;
  typedef db::parent_cell_iterator<graph_type> parent_cell_iterator;
  typedef db::child_cell_iterator<graph_type> child_cell_iterator;
  typedef tl::vector<size_t> sorted_inst_index_vector;
  typedef sorted_inst_index_vector::const_iterator sorted_inst_index_iterator;
  typedef db::instance<graph_type> instance_type;

  /**
   *  @brief Default constructor
   */
  instances (cell_type *cell)
    : mp_cell_inst_wp_tree (0), mp_cell_inst_tree (0), mp_cell (cell)
  {
    //  .. nothing yet ..
  }

  /**
   *  @brief The destructor
   */
  ~instances ()
  {
    do_clear_insts ();
  }

  /**
   *  @brief The assignment operator 
   */
  instances &operator= (const instances &d)
  {
    if (this != &d) {
      clear_insts ();
      if (d.mp_cell_inst_wp_tree) {
        inst_tree (typename cell_inst_wp_array_type::tag ()) = *d.mp_cell_inst_wp_tree;
      }
      if (d.mp_cell_inst_tree) {
        inst_tree (typename cell_inst_array_type::tag ()) = *d.mp_cell_inst_tree;
      }
      m_parent_insts = d.m_parent_insts;
    }
    return *this;
  }

  /**
   *  @brief Clear the instance list
   */
  void clear_insts ()
  {
    mp_cell->invalidate_insts ();
    if (mp_cell->manager () && mp_cell->manager ()->transacting ()) {
      if (mp_cell_inst_tree) {
        mp_cell->manager ()->queue (mp_cell, new db::inst_op<instances<Graph>, cell_inst_array_type> (false /*not insert*/, mp_cell_inst_tree->begin (), mp_cell_inst_tree->end ()));
      }
      if (mp_cell_inst_wp_tree) {
        mp_cell->manager ()->queue (mp_cell, new db::inst_op<instances<Graph>, cell_inst_wp_array_type> (false /*not insert*/, mp_cell_inst_wp_tree->begin (), mp_cell_inst_wp_tree->end ()));
      }
    }
    do_clear_insts ();
  }

  /**
   *  @brief Clear a instance list by tag
   *
   *  HINT: operations of this kind are not logged by the manager!
   *  They are supposed to be used in special cases only.
   */
  void clear (typename cell_inst_wp_array_type::tag) 
  {
    mp_cell->invalidate_insts ();
    if (mp_cell_inst_wp_tree) {
      delete mp_cell_inst_wp_tree;
      mp_cell_inst_wp_tree = 0;
    }
  }

  /**
   *  @brief Clear a instance list by tag
   *
   *  HINT: operations of this kind are not logged by the manager!
   *  They are supposed to be used in special cases only.
   */
  void clear (typename cell_inst_array_type::tag) 
  {
    mp_cell->invalidate_insts ();
    if (mp_cell_inst_tree) {
      delete mp_cell_inst_tree;
      mp_cell_inst_tree = 0;
    }
  }

  /**
   *  @brief Empty predicate - if no instances are given, this predicate is true
   */
  bool empty () const
  {
    if (mp_cell_inst_tree && ! mp_cell_inst_tree->size () == 0) {
      return false;
    }
    if (mp_cell_inst_wp_tree && ! mp_cell_inst_wp_tree->size () == 0) {
      return false;
    }
    return true;
  }

  /** 
   *  @brief Erase a cell instance given by a instance proxy
   *
   *  Erasing a cell instance will destroy the sorting order and invalidate
   *  other instance proxies.
   *  sort() must be called before a region query can be done
   *  on the cell instances. 
   */
  void erase (const instance_type &ref)
  {
    if (ref.has_prop_id ()) {
      erase_inst_by_tag (typename cell_inst_wp_array_type::tag (), ref);
    } else {
      erase_inst_by_tag (typename cell_inst_array_type::tag (), ref);
    } 
  }

  /**
   *  @brief Erase a cell instance
   *  
   *  Erasing a cell instance will destroy the sorting order.
   *  sort() must be called before a region query can be done
   *  on the cell instances. 
   */
  void erase (const_iterator e)
  {
    if (e->has_prop_id ()) {
      erase_inst_by_tag (typename cell_inst_wp_array_type::tag (), e);
    } else {
      erase_inst_by_tag (typename cell_inst_array_type::tag (), e);
    } 
  }

  /**
   *  @brief Erasing of multiple instances
   *
   *  Erase a set of positions given by an iterator I: *(from,to).
   *  *I must render an "iterator" object.
   *  The iterators in the sequence from, to must be sorted in
   *  "operator<" order.
   *
   *  @param first The start of the sequence of iterators
   *  @param last The end of the sequence of iterators
   */
  void erase_insts (const std::vector<instance_type> &instances)
  {
    for (typename std::vector<instance_type>::const_iterator i = instances.begin (); i != instances.end (); ) {

      typename std::vector<instance_type>::const_iterator inext = i;
      while (inext != instances.end () && inext->has_prop_id () == i->has_prop_id ()) {
        ++inext;
      }

      if (i->has_prop_id ()) {
        erase_insts_by_tag (typename cell_inst_wp_array_type::tag (), i, inext);
      } else {
        erase_insts_by_tag (typename cell_inst_array_type::tag (), i, inext);
      }

      i = inext;

    }
  }

  /**
   *  @brief Erasing of multiple elements
   *
   *  Erase a set of positions given by an iterator I: *(from,to).
   *  *I must render an "iterator" object.
   *  The iterators in the sequence from, to must be sorted in
   *  "operator<" order.
   *
   *  @param tag The specific instance type's tag
   *  @param first The start of the sequence of iterators
   *  @param last The end of the sequence of iterators
   */
  template <class Tag, class I>
  void erase_positions (Tag tag, I first, I last)
  {
    mp_cell->invalidate_insts ();  //  HINT: must come before the change is done!
    if (mp_cell->manager () && mp_cell->manager ()->transacting ()) {
      mp_cell->manager ()->queue (mp_cell, new db::inst_op<instances<Graph>, typename Tag::object_type> (false /*not insert*/, first, last, true /*dummy*/));
    }

    typedef db::box_tree<box_type, typename Tag::object_type, db::box_convert<typename Tag::object_type> > tree_type;
    tree_type &t = inst_tree (tag);
    t.erase_positions (first, last);
  }

  /**
   *  @brief Get the instance tree by shape type
   */
  const cell_inst_wp_tree_type &inst_tree (typename cell_inst_wp_array_type::tag) const
  {
    return *(mp_cell_inst_wp_tree ? mp_cell_inst_wp_tree : &ms_empty_wp_tree);
  }

  /**
   *  @brief Get the instance tree by shape type
   */
  const cell_inst_tree_type &inst_tree (typename cell_inst_array_type::tag) const
  {
    return *(mp_cell_inst_tree ? mp_cell_inst_tree : &ms_empty_tree);
  }

  /**
   *  @brief Insert a cell instance in a "editable safe" way
   *  
   *  Depending on the editable mode (db::editable_mode), not a cell instance array
   *  but a object_with_properties object will be created.
   *  Inserting a cell instance will destroy the sorting order.
   *  sort() must be called before a region query can be done
   *  on the cell instances.
   */
  template <class InstArray>
  void insert (const InstArray &inst)
  {
    if (db::editable_mode ()) {
      safe_insert (inst);
    } else {
      basic_insert (inst);
    }
  }

  /**
   *  @brief Insert a cell instance
   *  
   *  Inserting a cell instance will destroy the sorting order.
   *  sort() must be called before a region query can be done
   *  on the cell instances.
   */
  template <class InstArray>
  const InstArray &basic_insert (const InstArray &inst)
  {
    if (mp_cell->manager () && mp_cell->manager ()->transacting ()) {
      mp_cell->manager ()->queue (mp_cell, new db::inst_op<instances<Graph>, InstArray> (true /*insert*/, inst));
    }
    mp_cell->invalidate_insts ();
    return *inst_tree (typename InstArray::tag ()).insert (inst);
  }

  /**
   *  @brief Insert a sequence [from,to) of cell instances in a "editable safe" way
   *  
   *  @param from The start of the sequence
   *  @param to The past-the-end pointer of the sequence
   */
  template <class I>
  void insert (I from, I to)
  {
    if (db::editable_mode ()) {
      for (I i = from; i != to; ++i) {
        insert (*i);
      }
    } else {
      basic_insert (from, to);
    }
  }

  /**
   *  @brief Insert a sequence [from,to) of cell instances
   *  
   *  @param from The start of the sequence
   *  @param to The past-the-end pointer of the sequence
   */
  template <class I>
  void basic_insert (I from, I to)
  {
    typedef typename std::iterator_traits<I> it_traits;

    if (mp_cell->manager () && mp_cell->manager ()->transacting ()) {
      mp_cell->manager ()->queue (mp_cell, new db::inst_op<instances<Graph>, typename it_traits::value_type> (true /*insert*/, from, to));
    }
    mp_cell->invalidate_insts ();
    inst_tree (typename it_traits::value_type::tag ()).insert (from, to);
  }

  /**
   *  @brief Insert an instance given by a instance reference 
   *
   *  This member may be used to copy an instance from one cell to another.
   *  Because of the inherent instability of the instance pointer it must not be 
   *  used to copy instances within one cell.
   *
   *  @param instance The instance reference of which to insert a copy
   */
  instance_type insert (const instance_type &instance)
  {
    return (insert (instance, tl::ident_map<cell_index_type> (), tl::ident_map<db::PropertiesRepository::properties_id_type> ()));
  }

  /**
   *  @brief Insert an instance given by a instance reference with a different cell index and property ID
   *
   *  This member may be used to map an instance to another layout object. 
   *
   *  @param ref The instance reference of which to insert a copy
   *  @param im The mapper to new cell index to use (for mapping to a different layout for example)
   *  @param pm The mapper to new cell property ID use (for mapping to a different layout for example)
   */
  template <class IndexMap, class PropIdMap>
  instance_type insert (const instance_type &ref, const IndexMap &im, const PropIdMap &pm)
  {
    if (! ref.has_prop_id ()) {
      cell_inst_array_type inst = *ref.basic_ptr (typename cell_inst_array_type::tag ());
      inst.object () = cell_inst_type (im (ref.cell_index ()));
      return instance_type (basic_insert (inst));
    } else {
      cell_inst_wp_array_type inst_wp = *ref.basic_ptr (typename cell_inst_wp_array_type::tag ());
      inst_wp.object () = cell_inst_type (im (ref.cell_index ()));
      inst_wp.properties_id (pm (ref.prop_id ()));
      return instance_type (basic_insert (inst_wp));
    }
  }

  /** 
   *  @brief Replace the properties ID of an element (pointed to by the iterator) with the given one
   *
   *  The iterator must point to a instance already having a property ID. 
   *  If this is not the case, an exception is thrown
   */
  void replace_prop_id (const instance_type &ref, db::PropertiesRepository::properties_id_type prop_id)
  {
    tl_assert (ref.has_prop_id ());
    cell_inst_wp_array_type *pos = (cell_inst_wp_array_type *) ref.basic_ptr (typename cell_inst_wp_array_type::tag ());
    if (mp_cell->manager () && mp_cell->manager ()->transacting ()) {
      mp_cell->manager ()->queue (mp_cell, new db::inst_op<instances<Graph>, cell_inst_wp_array_type> (false /*not insert*/, *pos));
    }
    //  KLUDGE: even though the properties do not constitute the hierarchy, we call the invalidate method 
    //  to trigger a redraw for example.
    mp_cell->invalidate_hier ();
    pos->properties_id (prop_id);
    if (mp_cell->manager () && mp_cell->manager ()->transacting ()) {
      mp_cell->manager ()->queue (mp_cell, new db::inst_op<instances<Graph>, cell_inst_wp_array_type> (true /*insert*/, *pos));
    }
  }

  /**
   *  @brief Replace the instance pointed to by the iterator with the given instance
   *
   *  The ability to replace an instance implies the ability to in-place convert the
   *  kind of instance, i.e. a instance without property cannot be replaced with one
   *  with properties.
   */
  template <class InstArray>
  void replace (const instance_type &ref, const InstArray &inst)
  {
    if (! ref.has_prop_id ()) {
      replace (ref.basic_ptr (typename cell_inst_array_type::tag ()), (const cell_inst_array_type &) inst);
    } else {
      replace (ref.basic_ptr (typename cell_inst_wp_array_type::tag ()), cell_inst_wp_array_type (inst, ref.prop_id ()));
    }
  }

  /**
   *  @brief Transform the instance pointed to by the instance reference
   *
   *  @param t The transformation to apply
   */
  template <class Trans>
  void transform (const instance_type &ref, const Trans &t)
  {
    if (! ref.has_prop_id ()) {
      replace (ref, ref.basic_ptr (typename cell_inst_array_type::tag ())->transformed (t));
    } else {
      replace (ref, ref.basic_ptr (typename cell_inst_wp_array_type::tag ())->transformed (t));
    }
  }

  /**
   *  @brief Replace an instance given by an address
   */

  template <class InstArray>
  void replace (const InstArray *replace, const InstArray &with)
  {
    if (mp_cell->manager () && mp_cell->manager ()->transacting ()) {
      mp_cell->manager ()->queue (mp_cell, new db::inst_op<instances<Graph>, InstArray> (false /*not insert*/, *replace));
      mp_cell->manager ()->queue (mp_cell, new db::inst_op<instances<Graph>, InstArray> (true /*insert*/, with));
    }
    mp_cell->invalidate_insts ();
    *((InstArray *)replace) = with;
  }

  /**
   *  @brief Clear the parent instance list
   *
   *  Reserve the given number of entries.
   */
  void clear_parent_insts (size_t sz = 0)
  {
    m_parent_insts.clear ();
    m_parent_insts.reserve (sz);
  }

  /**
   *  @brief Count the number of parent instances
   * 
   *  This will accumulate the number of parent instances in the vector
   *  provided. Later, this count is used to resize the parent instance
   *  vector.
   */
  void count_parent_insts (std::vector <size_t> &count) const
  {
    cell_index_type last_ci = (cell_index_type) -1l;
    
    for (sorted_inst_index_iterator c = begin_sorted_inst_indices (); c != end_sorted_inst_indices (); ++c) {
      cell_index_type ci = begin (*c)->cell_inst ().object ().cell_index ();
      if (ci != last_ci) {
        last_ci = ci;
        ++count [ci];
      }
    }
  }

  /**
   *  @brief Establish the instance index list giving the instances by cell index
   */
  void sort_child_insts () 
  {
    std::vector< std::pair<const cell_inst_array_type *, size_t> > indices;
    indices.reserve (cell_instances ());

    size_t i = 0;
    for (const_iterator cc = begin (); ! cc.at_end (); ++cc, ++i) {
      indices.push_back (std::make_pair (& cc->cell_inst (), i));
    }
    std::sort (indices.begin (), indices.end (), cell_inst_compare_f<cell_inst_array_type> ());

    m_insts_by_cell_index = sorted_inst_index_vector ();
    m_insts_by_cell_index.reserve (indices.size ());
    for (typename std::vector< std::pair<const cell_inst_array_type *, size_t> >::const_iterator ii = indices.begin (); ii != indices.end (); ++ii) {
      m_insts_by_cell_index.push_back (ii->second);
    }
  }

  /**
   *  @brief Sort the cell instance list
   *
   *  This will sort the cell instance list. As a prerequesite
   *  the cell's bounding boxes must have been computed.
   */
  void sort_inst_tree (const Graph *g)
  {
    //  sort the overall box instance tree
    if (mp_cell_inst_wp_tree) {
      mp_cell_inst_wp_tree->sort (db::box_convert<cell_inst_wp_array_type> (*g));
    }
    if (mp_cell_inst_tree) {
      mp_cell_inst_tree->sort (db::box_convert<cell_inst_array_type> (*g));
    }
  }

  /**
   *  @brief Update the child-parent relationships
   * 
   *  This will update the child-parent relationships. Basically
   *  this means entering the cell as a parent into all it's child
   *  cells.
   */
  void update_relations (Graph *g, cell_index_type cell_index)
  {
    cell_index_type last_ci = (cell_index_type) -1l;

    size_t idx = 0;
    for (sorted_inst_index_iterator c = begin_sorted_inst_indices (); c != end_sorted_inst_indices (); ++c, ++idx) {
      cell_index_type ci = begin (*c)->cell_inst ().object ().cell_index ();
      if (ci != last_ci) {
        last_ci = ci;
        g->cell (ci).instances ().m_parent_insts.push_back (parent_inst_type (cell_index, idx));
      }
    }
  }

  /**
   *  @brief Region query for the instances in "overlapping" mode
   *
   *  This will return the region query iterator for the child cell
   *  instances overlapping with the given region b. 
   *
   *  @param b The region to query
   *
   *  @return The region iterator for the child cell instances
   */
  overlapping_iterator begin_overlapping (const box_type &b, const Graph *g) const
  {
    return overlapping_iterator (overlapping_instance_iterator_traits<Graph> (this, b, g));
  }

  /**
   *  @brief Region query for the instances in "touching" mode
   *
   *  This will return the region query iterator for the child cell
   *  instances touching the given region b. 
   *
   *  @param b The region to query
   *
   *  @return The region iterator for the child cell instances
   */
  touching_iterator begin_touching (const box_type &b, const Graph *g) const
  {
    return touching_iterator (touching_instance_iterator_traits<Graph> (this, b, g));
  }

  /**
   *  @brief The child cell iterator
   *
   *  This iterator will report the child cell indices, not every instance.
   *  Basically this is a filter.
   */
  child_cell_iterator begin_child_cells () const
  {
    return child_cell_iterator (this);
  }

  /** 
   *  @brief Report the number of child cells
   *
   *  Just the number of child cells with distinct cell indices are
   *  reported.
   *  CAUTION: this method is SLOW!
   */
  size_t child_cells () const
  {
    size_t n = 0;
    for (child_cell_iterator i = begin_child_cells (); ! i.at_end (); ++i) {
      ++n;
    }
    return n;
  }

  /**
   *  @brief The number of cell instances 
   */
  size_t cell_instances () const
  {
    size_t n = 0;
    if (mp_cell_inst_wp_tree) {
      n += mp_cell_inst_wp_tree->size ();
    } 
    if (mp_cell_inst_tree) {
      n += mp_cell_inst_tree->size ();
    }
    return n;
  }

  /**
   *  @brief The cell instance access begin iterator
   */
  const_iterator begin () const
  {
    return const_iterator (normal_instance_iterator_traits<Graph> (this));
  }

  /**
   *  @brief The iterator delivering the cell instance (indices) sorted by child cell ID
   */
  sorted_inst_index_iterator begin_sorted_inst_indices () const
  {
    return m_insts_by_cell_index.begin ();
  }

  /**
   *  @brief The iterator delivering the cell instance (indices) sorted by child cell ID (end)
   */
  sorted_inst_index_iterator end_sorted_inst_indices () const
  {
    return m_insts_by_cell_index.end ();
  }

  /**
   *  @brief The cell instance access begin iterator plus offset
   *
   *  The iterator returned will point to the i-th element.
   */
  const_iterator begin (size_t i) const
  {
    return const_iterator (normal_instance_iterator_traits<Graph> (this), i);
  }

  /**
   *  @brief The parent instance list begin iterator
   *
   *  The begin_parent_insts() allows to access to the parent instance list.
   */
  parent_inst_iterator begin_parent_insts (const Graph *g) const
  {
    return parent_inst_iterator (g, m_parent_insts.begin (), m_parent_insts.end ());
  }

  /**
   *  @brief Report the number of parent cells 
   *
   *  Since the m_parent_insts vector just stores references to those cells
   *  that have distinct cell indices, we can simply report their length.
   */
  size_t parent_cells () const
  {
    return m_parent_insts.size ();
  }

  /**
   *  @brief The parent cell iterator
   * 
   *  This iterator will iterate over the parent cells, just returning their
   *  cell index.
   */
  parent_cell_iterator begin_parent_cells () const
  {
    return parent_cell_iterator (m_parent_insts.begin ());
  }

  /**
   *  @brief The parent cell end iterator
   */
  parent_cell_iterator end_parent_cells () const
  {
    return parent_cell_iterator (m_parent_insts.end ());
  }

  /**
   *  @brief Tell if the cell is a top-level cell
   *
   *  A cell is a top-level cell if there are no parent instantiations.
   */
  bool is_top () const
  {
    return m_parent_insts.end () == m_parent_insts.begin ();
  }

  /**
   *  @brief Collect memory usage statistics
   */
  void
  collect_mem_stat (db::MemStatistics &m) const
  {
    m.cell_info (sizeof (mp_cell), sizeof (mp_cell));
    m.cell_info (sizeof (m_parent_insts), sizeof (m_parent_insts));        //  need more: see below
    m.cell_info (sizeof (mp_cell_inst_wp_tree), sizeof (mp_cell_inst_wp_tree));  //  need more: see below
    m.cell_info (sizeof (mp_cell_inst_tree), sizeof (mp_cell_inst_tree));        //  need more: see below
    m.cell_info (sizeof (m_insts_by_cell_index), sizeof (m_insts_by_cell_index));        //  need more: see below

    if (mp_cell_inst_wp_tree) {
      m.inst_trees (*mp_cell_inst_wp_tree);
    }
    if (mp_cell_inst_tree) {
      m.inst_trees (*mp_cell_inst_tree);
    }
    m.instances (m_parent_insts);
    m.instances (m_insts_by_cell_index);
    m.instances (-sizeof (m_parent_insts), -sizeof (m_parent_insts));        //  the actual object is counted for cell, not for instances (see above)
  }

private:
  //  in editable mode, only instance arrays with properties are permissable
  template <class InstArray>
  void safe_insert (const db::object_with_properties<InstArray> &inst)
  {
    basic_insert (inst);
  }

  //  in editable mode, only instance arrays with properties are permissable
  template <class InstArray>
  void safe_insert (const InstArray &inst)
  {
    basic_insert (db::object_with_properties<InstArray> (inst, 0));
  }

  cell_inst_wp_tree_type *mp_cell_inst_wp_tree;
  cell_inst_tree_type *mp_cell_inst_tree;
  parent_inst_vector m_parent_insts;
  sorted_inst_index_vector m_insts_by_cell_index;
  cell_type *mp_cell;
  static cell_inst_wp_tree_type ms_empty_wp_tree;
  static cell_inst_tree_type ms_empty_tree;

  //  no copy ctor
  instances (const instances &d);

  void do_clear_insts ()
  {
    if (mp_cell_inst_tree) {
      delete mp_cell_inst_tree;
      mp_cell_inst_tree = 0;
    }
    if (mp_cell_inst_wp_tree) {
      delete mp_cell_inst_wp_tree;
      mp_cell_inst_wp_tree = 0;
    }
  }

  cell_inst_wp_tree_type &inst_tree (typename cell_inst_wp_array_type::tag) 
  {
    if (! mp_cell_inst_wp_tree) {
      mp_cell_inst_wp_tree = new cell_inst_wp_tree_type ();
    }
    return *mp_cell_inst_wp_tree;
  }

  cell_inst_tree_type &inst_tree (typename cell_inst_array_type::tag) 
  {
    if (! mp_cell_inst_tree) {
      mp_cell_inst_tree = new cell_inst_tree_type ();
    }
    return *mp_cell_inst_tree;
  }

  template <class Tag>
  void erase_inst_by_tag (Tag tag, const_iterator e)
  {
    erase_inst_by_tag (tag, *(e.get_iter (tag).basic_iter ()));
  }

  template <class Tag>
  void erase_inst_by_tag (Tag tag, const instance_type &ref)
  {
    erase_inst_by_tag (tag, *(ref.basic_ptr (tag)));
  }

  template <class Tag>
  void erase_inst_by_tag (Tag tag, const typename Tag::object_type &obj)
  {
    mp_cell->invalidate_insts ();

    if (mp_cell->manager () && mp_cell->manager ()->transacting ()) {
      mp_cell->manager ()->queue (mp_cell, new db::inst_op<instances<Graph>, typename Tag::object_type> (false /*not insert*/, obj));
    }

    typedef db::box_tree<box_type, typename Tag::object_type, db::box_convert<typename Tag::object_type> > tree_type;
    tree_type &t = inst_tree (tag);
    t.erase (t.begin () + size_t (&obj - &*(((const tree_type *)&t)->begin ())));
  }

  template <class Tag>
  void erase_insts_by_tag (Tag tag, typename std::vector<instance_type>::const_iterator s1, typename std::vector<instance_type>::const_iterator s2)
  {
    typedef db::box_tree<box_type, typename Tag::object_type, db::box_convert<typename Tag::object_type> > tree_type;
    tree_type &t = inst_tree (tag);
    typename tree_type::iterator b = t.begin ();

    std::vector<typename tree_type::iterator> iters;
    iters.reserve (std::distance (s1, s2));

    for (typename std::vector<instance_type>::const_iterator s = s1; s != s2; ++s) {
      iters.push_back (b + size_t (s->basic_ptr (tag) - &*(((const tree_type *)&t)->begin ())));
    }

    erase_positions (Tag (), iters.begin (), iters.end ());
  }
};

//  the typedef for the standard instance reference
typedef db::instance<db::CellGraph> Instance;

//  the typedef for the standard instance container 
typedef db::instances<db::CellGraph> Instances;

}

#endif

