
/*

  KLayout Layout Viewer
  Copyright (C) 2006-2007 Matthias Koefferlein

  This program is free software; you can redistribute it and/or modify
  it under the terms of the GNU General Public License as published by
  the Free Software Foundation; either version 2 of the License, or
  (at your option) any later version.

  This program is distributed in the hope that it will be useful,
  but WITHOUT ANY WARRANTY; without even the implied warranty of
  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
  GNU General Public License for more details.

  You should have received a copy of the GNU General Public License
  along with this program; if not, write to the Free Software
  Foundation, Inc., 51 Franklin St, Fifth Floor, Boston, MA  02110-1301  USA

*/


#include "dbLayoutDiff.h"
#include "dbLayout.h"
#include "tlLog.h"

namespace db
{

static void 
collect_layers (const db::Layout &l, std::map< std::pair<int, int>, unsigned int> &layers, unsigned int flags /* not used currently */)
{
  for (unsigned int i = 0; i < l.layers (); ++i) {
    if (l.is_valid_layer (i)) {
      const LayerProperties &lp = l.get_properties (i);
      if (lp.layer >= 0 && lp.datatype >= 0) {
        layers.insert (std::make_pair (std::make_pair (lp.layer, lp.datatype), i));
      }
    }
  }
}

static void
collect_cells (const db::Layout &l, std::map <std::string, unsigned int> &cells)
{
  for (db::CellGraph::const_iterator c = l.begin (); c != l.end (); ++c) {
    cells.insert (std::make_pair (l.cell_name (c->cell_index ()), c->cell_index ()));
  }
}

struct PropNormalize 
{
  PropNormalize (const db::Layout *a, const db::Layout *b)
    : mp_a (a), mp_b (b)
  { 
    if (mp_b != mp_a) {
      m_pr = b->properties_repository ();
    }
  }

  unsigned long operator() (unsigned long pa) 
  {
    if (mp_b == mp_a) {
      return pa;
    } else {
      const db::PropertiesRepository::properties_set &ps (mp_a->properties_repository ().properties (pa));
      db::PropertiesRepository::properties_set ps_mapped;
      for (db::PropertiesRepository::properties_set::const_iterator p = ps.begin (); p != ps.end (); ++p) {
        const tl::Variant &name = mp_a->properties_repository ().prop_name (p->first);
        ps_mapped.insert (std::make_pair (m_pr.prop_name_id (name), p->second));
      }
      return m_pr.properties_id (ps_mapped);
    }
  }

  const db::PropertiesRepository &pr () { return m_pr; }

private:
  const db::Layout *mp_a, *mp_b;
  db::PropertiesRepository m_pr;
};

static void
collect_insts (const db::Layout &l, const db::Cell *cell, unsigned int flags, const std::map <db::CellGraph::cell_index_type, size_t> &cci, std::vector <db::CellInstArrayWithProperties> &insts, PropNormalize &pn)
{
  for (db::Cell::const_iterator i = cell->begin (); !i.at_end (); ++i) {

    db::CellInstArrayWithProperties inst;

    db::CellInst new_obj (cci.find (i->cell_index ())->second);

    unsigned long prop_id = 0;
    if (! (flags & layout_diff::f_no_properties)) {
      prop_id = pn (i->prop_id ());
    }

    db::Point a, b;
    unsigned long amax, bmax;
    if (i->is_regular_array (a, b, amax, bmax) && (amax > 1 || bmax > 1)) {

      //  normalize arrays
      if (amax == 1) {
        a = db::Point ();
      }
      if (bmax == 1) {
        b = db::Point ();
      }
      if (b < a) {
        std::swap (a, b);
        std::swap (amax, bmax);
      }

      inst = db::CellInstArrayWithProperties (db::CellInstArray (new_obj, i->complex_trans (), a, b, amax, bmax), prop_id);

    } else {
      inst = db::CellInstArrayWithProperties (db::CellInstArray (new_obj, i->complex_trans ()), prop_id);
    }

    insts.push_back (inst);

  }

  std::sort (insts.begin (), insts.end ());
}

static void
collect_polygons (const db::Layout &l, const db::Cell *c, unsigned int layer, unsigned int flags, std::vector< std::pair<db::Polygon, unsigned long> > &shapes, PropNormalize &pn)
{
  for (db::ShapeIterator s = c->shapes (layer).begin (db::ShapeIterator::Polygons | ((flags & db::layout_diff::f_boxes_as_polygons) ? db::ShapeIterator::Boxes : 0)); !s.at_end (); ++s) {
    unsigned long prop_id = 0;
    if (! (flags & layout_diff::f_no_properties)) {
      prop_id = pn (s->prop_id ());
    }
    shapes.push_back (std::make_pair (db::Polygon (), prop_id));
    s->polygon (shapes.back ().first);
  }

  std::sort (shapes.begin (), shapes.end ());
}

static void
collect_edges (const db::Layout &l, const db::Cell *c, unsigned int layer, unsigned int flags, std::vector< std::pair<db::Edge, unsigned long> > &shapes, PropNormalize &pn)
{
  for (db::ShapeIterator s = c->shapes (layer).begin (db::ShapeIterator::Edges); !s.at_end (); ++s) {
    unsigned long prop_id = 0;
    if (! (flags & layout_diff::f_no_properties)) {
      prop_id = pn (s->prop_id ());
    }
    shapes.push_back (std::make_pair (db::Edge (), prop_id));
    s->edge (shapes.back ().first);
  }

  std::sort (shapes.begin (), shapes.end ());
}

static void
collect_boxes (const db::Layout &l, const db::Cell *c, unsigned int layer, unsigned int flags, std::vector< std::pair<db::Box, unsigned long> > &shapes, PropNormalize &pn)
{
  for (db::ShapeIterator s = c->shapes (layer).begin (db::ShapeIterator::Boxes); !s.at_end (); ++s) {
    unsigned long prop_id = 0;
    if (! (flags & layout_diff::f_no_properties)) {
      prop_id = pn (s->prop_id ());
    }
    shapes.push_back (std::make_pair (db::Box (), prop_id));
    s->box (shapes.back ().first);
  }

  std::sort (shapes.begin (), shapes.end ());
}

static void
collect_texts (const db::Layout &l, const db::Cell *c, unsigned int layer, unsigned int flags, std::vector< std::pair<db::Text, unsigned long> > &shapes, PropNormalize &pn)
{
  for (db::ShapeIterator s = c->shapes (layer).begin (db::ShapeIterator::Texts); !s.at_end (); ++s) {

    unsigned long prop_id = 0;
    if (! (flags & layout_diff::f_no_properties)) {
      prop_id = pn (s->prop_id ());
    }
    shapes.push_back (std::make_pair (db::Text (), prop_id));
    s->text (shapes.back ().first);

    if (flags & layout_diff::f_no_text_orientation) {
      db::Text::trans_type tt (shapes.back ().first.trans ());
      tt = db::Text::trans_type (tt.disp ());
      shapes.back ().first.trans (tt);
    }

  }

  std::sort (shapes.begin (), shapes.end ());
}

static void
collect_paths (const db::Layout &l, const db::Cell *c, unsigned int layer, unsigned int flags, std::vector< std::pair<db::Path, unsigned long> > &shapes, PropNormalize &pn)
{
  for (db::ShapeIterator s = c->shapes (layer).begin (db::ShapeIterator::Paths); !s.at_end (); ++s) {
    unsigned long prop_id = 0;
    if (! (flags & layout_diff::f_no_properties)) {
      prop_id = pn (s->prop_id ());
    }
    shapes.push_back (std::make_pair (db::Path (), prop_id));
    s->path (shapes.back ().first);
  }

  std::sort (shapes.begin (), shapes.end ());
}

static void 
print_cell_inst (const db::CellInstArrayWithProperties &ci, const std::vector <std::string> &cell_names)
{
  tl::info << "  " << cell_names [ci.object ().cell_index ()] << " " << ci.complex_trans ().to_string () << tl::noendl;

  db::Point a, b;
  unsigned long amax, bmax;
  if (ci.is_regular_array (a, b, amax, bmax)) {
    tl::info << "[a=" << a.to_string () << ", b=" << b.to_string () << ", na=" << amax << ", nb=" << bmax << "]" << tl::noendl;
  } else {
    tl::info << "" << tl::noendl;
  }
  if (ci.properties_id () != 0) {
    tl::info << " [" << ci.properties_id () << "]";
    /*  to print the properties themselves:
    const db::PropertiesRepository::properties_set &props = pr.properties (ci.properties_id ());
    for (db::PropertiesRepository::properties_set::const_iterator p = props.begin (); p != props.end (); ++p) {
      const tl::Variant &name = pr.prop_name (p->first);
      if (name.is_long ()) {
        tl::info << "  {" << int (name.to_long ()) << " {" << p->second.to_string () << "}}";
      } else if (name.is_string ()) {
        tl::info << "  {{" << name.to_string () << "} {" << p->second.to_string () << "}}";
      }
    }
    tl::info << "}";
    */
  } else {
    tl::info << "";
  }
}

template <class SH>
void
print_diffs (const db::PropertiesRepository &pr, const std::vector <std::pair <SH, unsigned long> > &a, const std::vector <std::pair <SH, unsigned long> > &b)
{
  std::vector <std::pair <SH, unsigned long> > anotb;
  std::set_difference (a.begin (), a.end (), b.begin (), b.end (), std::back_inserter (anotb));
  for (typename std::vector <std::pair <SH, unsigned long> >::const_iterator s = anotb.begin (); s != anotb.end (); ++s) {
    tl::info << "  " << s->first.to_string () << tl::noendl;
    if (s->second != 0) {
      tl::info << " [" << s->second << "]";
      /*  to print the properties themselves:
      const db::PropertiesRepository::properties_set &props = pr.properties (s->second);
      for (db::PropertiesRepository::properties_set::const_iterator p = props.begin (); p != props.end (); ++p) {
        const tl::Variant &name = pr.prop_name (p->first);
        if (name.is_long ()) {
          tl::info << "  {" << int (name.to_long ()) << " {" << p->second.to_string () << "}}";
        } else if (name.is_string ()) {
          tl::info << "  {{" << name.to_string () << "} {" << p->second.to_string () << "}}";
        }
      }
      tl::info << "}";
      */
    } else {
      tl::info << "";
    }
  }
}

bool
compare_layouts (const db::Layout &a, const db::Layout &b, unsigned int flags)
{
  bool differs = false;

  if (fabs (a.dbu () - b.dbu ()) > 1e-9) {
    differs = true;
    if (flags & layout_diff::f_silent) {
      return false;
    }
    tl::error << "layout_diff: database units differ " << a.dbu () << " vs. " << b.dbu ();
  }

  bool verbose = (flags & layout_diff::f_verbose);

  PropNormalize prop_normalize_a (&a, &a);
  PropNormalize prop_normalize_b (&b, &a);

  //  compare layers

  std::map< std::pair<int, int>, unsigned int> layers_a;
  std::map< std::pair<int, int>, unsigned int> layers_b;

  collect_layers (a, layers_a, flags);
  collect_layers (b, layers_b, flags);

  std::vector< std::pair<int, int> > common_layers;

  for (std::map< std::pair<int, int>, unsigned int>::const_iterator la = layers_a.begin (); la != layers_a.end (); ++la) {
    std::map< std::pair<int, int>, unsigned int>::const_iterator lb = layers_b.find (la->first);
    if (lb == layers_b.end ()) {
      differs = true;
      if (flags & layout_diff::f_silent) {
        return false;
      }
      tl::error << "layout_diff: layer " << la->first.first << "," << la->first.second << " is not present in layout b, but in a";
    } else {
      common_layers.push_back (la->first);
      if (! (flags & layout_diff::f_no_layer_names) && a.get_properties (la->second).name != b.get_properties (lb->second).name) {
        differs = true;
        if (flags & layout_diff::f_silent) {
          return false;
        }
        tl::error << "layout_diff: layer names differ between layout a and b for layer " << la->first.first << "," << la->first.second << ": " 
                  << a.get_properties (la->second).name << " vs. " << b.get_properties (lb->second).name;
      }
    }
  }

  for (std::map< std::pair<int, int>, unsigned int>::const_iterator lb = layers_b.begin (); lb != layers_b.end (); ++lb) {
    std::map< std::pair<int, int>, unsigned int>::const_iterator la = layers_a.find (lb->first);
    if (la == layers_a.end ()) {
      differs = true;
      if (flags & layout_diff::f_silent) {
        return false;
      }
      tl::error << "layout_diff: layer " << lb->first.first << "," << lb->first.second << " is not present in layout a, but in b";
    } 
  }


  //  compare cells

  std::map <std::string, db::CellGraph::cell_index_type> cells_a; 
  std::map <std::string, db::CellGraph::cell_index_type> cells_b; 

  collect_cells (a, cells_a);
  collect_cells (b, cells_b);

  std::vector <std::string> common_cells;
  std::map <db::CellGraph::cell_index_type, size_t> common_cell_indices_a;
  std::vector <db::CellGraph::cell_index_type> common_cells_a;
  std::map <db::CellGraph::cell_index_type, size_t> common_cell_indices_b;
  std::vector <db::CellGraph::cell_index_type> common_cells_b;

  size_t cci = 0;
  for (std::map <std::string, db::CellGraph::cell_index_type>::const_iterator ca = cells_a.begin (); ca != cells_a.end (); ++ca) {
    std::map <std::string, db::CellGraph::cell_index_type>::const_iterator cb = cells_b.find (ca->first);
    if (cb == cells_b.end ()) {
      differs = true;
      if (flags & layout_diff::f_silent) {
        return false;
      }
      tl::error << "layout_diff: cell " << ca->first << " is not present in layout b, but in a";
    } else {
      common_cells.push_back (ca->first);
      common_cell_indices_a.insert (std::make_pair (ca->second, cci));
      common_cells_a.push_back (ca->second);
      common_cell_indices_b.insert (std::make_pair (cb->second, cci));
      common_cells_b.push_back (cb->second);
    }
  }

  for (std::map <std::string, db::CellGraph::cell_index_type>::const_iterator cb = cells_b.begin (); cb != cells_b.end (); ++cb) {
    std::map <std::string, db::CellGraph::cell_index_type>::const_iterator ca = cells_a.find (cb->first);
    if (ca == cells_a.end ()) {
      differs = true;
      if (flags & layout_diff::f_silent) {
        return false;
      }
      tl::error << "layout_diff: cell " << cb->first << " is not present in layout a, but in b";
    }
  }


  //  compare cell by cell
  
  for (unsigned int cci = 0; cci < common_cells.size (); ++cci) {

    const db::Cell *cell_a = &a.cell (common_cells_a [cci]);
    const db::Cell *cell_b = &b.cell (common_cells_b [cci]);

    if (cell_a->bbox () != cell_b->bbox ()) {
      differs = true;
      if (flags & layout_diff::f_silent) {
        return false;
      }
      tl::error << "layout_diff: bounding boxes differ for cell " << a.cell_name (common_cells_a [cci]) << ", " 
                << cell_a->bbox ().to_string () << " vs. " << cell_b->bbox ().to_string ();
    }

    std::vector <db::CellInstArrayWithProperties> insts_a;
    collect_insts (a, cell_a, flags, common_cell_indices_a, insts_a, prop_normalize_a);

    std::vector <db::CellInstArrayWithProperties> insts_b;
    collect_insts (b, cell_b, flags, common_cell_indices_b, insts_b, prop_normalize_b);

    if (insts_a != insts_b) {
      differs = true;
      if (flags & layout_diff::f_silent) {
        return false;
      }
      tl::error << "layout_diff: instances differ in cell " << common_cells [cci];
      if (verbose) {
        tl::info << "list for a:";
        for (std::vector <db::CellInstArrayWithProperties>::const_iterator s = insts_a.begin (); s != insts_a.end (); ++s) {
          print_cell_inst (*s, common_cells);
        }
        tl::info << "list for b:";
        for (std::vector <db::CellInstArrayWithProperties>::const_iterator s = insts_b.begin (); s != insts_b.end (); ++s) {
          print_cell_inst (*s, common_cells);
        }
        std::vector <db::CellInstArrayWithProperties> anotb;
        std::set_difference (insts_a.begin (), insts_a.end (), insts_b.begin (), insts_b.end (), std::back_inserter (anotb));
        tl::info << "Not in b but in a:";
        for (std::vector <db::CellInstArrayWithProperties>::const_iterator s = anotb.begin (); s != anotb.end (); ++s) {
          print_cell_inst (*s, common_cells);
        }
        std::vector <db::CellInstArrayWithProperties> bnota;
        std::set_difference (insts_b.begin (), insts_b.end (), insts_a.begin (), insts_a.end (), std::back_inserter (bnota));
        tl::info << "Not in a but in b:";
        for (std::vector <db::CellInstArrayWithProperties>::const_iterator s = bnota.begin (); s != bnota.end (); ++s) {
          print_cell_inst (*s, common_cells);
        }
      }
    }


    //  compare layer by layer
    
    for (std::vector< std::pair<int, int> >::const_iterator cl = common_layers.begin (); cl != common_layers.end (); ++cl) {

      unsigned int layer_a = layers_a.find (*cl)->second;
      unsigned int layer_b = layers_b.find (*cl)->second;

      if (cell_a->bbox (layer_a) != cell_b->bbox (layer_b)) {
        differs = true;
        if (flags & layout_diff::f_silent) {
          return false;
        }
        tl::error << "layout_diff: per-layer bounding boxes differ for cell " << a.cell_name (common_cells_a [cci]) << ", layer (" << cl->first << ", " << cl->second << "), " 
                  << cell_a->bbox (layer_a).to_string () << " vs. " << cell_b->bbox (layer_b).to_string ();
      }


      //  compare polygons

      std::vector <std::pair <db::Polygon, unsigned long> > polygons_a;
      collect_polygons (a, cell_a, layer_a, flags, polygons_a, prop_normalize_a);

      std::vector <std::pair <db::Polygon, unsigned long> > polygons_b;
      collect_polygons (b, cell_b, layer_b, flags, polygons_b, prop_normalize_b);

      if (polygons_a != polygons_b) {
        differs = true;
        if (flags & layout_diff::f_silent) {
          return false;
        }
        tl::error << "layout_diff: polygons differ";
        if (verbose) {
          tl::info << "Not in b but in a:";
          print_diffs (prop_normalize_b.pr (), polygons_a, polygons_b);
          tl::info << "Not in a but in b:";
          print_diffs (prop_normalize_b.pr (), polygons_b, polygons_a);
        }
      }


      //  compare paths

      std::vector <std::pair <db::Path, unsigned long> > paths_a;
      collect_paths (a, cell_a, layer_a, flags, paths_a, prop_normalize_a);

      std::vector <std::pair <db::Path, unsigned long> > paths_b;
      collect_paths (b, cell_b, layer_b, flags, paths_b, prop_normalize_b);

      if (paths_a != paths_b) {
        differs = true;
        if (flags & layout_diff::f_silent) {
          return false;
        }
        tl::error << "layout_diff: paths differ";
        if (verbose) {
          tl::info << "Not in b but in a:";
          print_diffs (prop_normalize_b.pr (), paths_a, paths_b);
          tl::info << "Not in a but in b:";
          print_diffs (prop_normalize_b.pr (), paths_b, paths_a);
        }
      }


      //  compare texts

      std::vector <std::pair <db::Text, unsigned long> > texts_a;
      collect_texts (a, cell_a, layer_a, flags, texts_a, prop_normalize_a);

      std::vector <std::pair <db::Text, unsigned long> > texts_b;
      collect_texts (b, cell_b, layer_b, flags, texts_b, prop_normalize_b);

      if (texts_a != texts_b) {
        differs = true;
        if (flags & layout_diff::f_silent) {
          return false;
        }
        tl::error << "layout_diff: texts differ";
        if (verbose) {
          tl::info << "Not in b but in a:";
          print_diffs (prop_normalize_b.pr (), texts_a, texts_b);
          tl::info << "Not in a but in b:";
          print_diffs (prop_normalize_b.pr (), texts_b, texts_a);
        }
      }


      //  compare boxes (unless this is done by the polygon compare code)
      if (! (flags & db::layout_diff::f_boxes_as_polygons)) {

        std::vector <std::pair <db::Box, unsigned long> > boxes_a;
        collect_boxes (a, cell_a, layer_a, flags, boxes_a, prop_normalize_a);

        std::vector <std::pair <db::Box, unsigned long> > boxes_b;
        collect_boxes (b, cell_b, layer_b, flags, boxes_b, prop_normalize_b);

        if (boxes_a != boxes_b) {
          differs = true;
          if (flags & layout_diff::f_silent) {
            return false;
          }
          tl::error << "layout_diff: boxes differ";
          if (verbose) {
            tl::info << "Not in b but in a:";
            print_diffs (prop_normalize_b.pr (), boxes_a, boxes_b);
            tl::info << "Not in a but in b:";
            print_diffs (prop_normalize_b.pr (), boxes_b, boxes_a);
          }
        }

      }

      //  compare edges

      std::vector <std::pair <db::Edge, unsigned long> > edges_a;
      collect_edges (a, cell_a, layer_a, flags, edges_a, prop_normalize_a);

      std::vector <std::pair <db::Edge, unsigned long> > edges_b;
      collect_edges (b, cell_b, layer_b, flags, edges_b, prop_normalize_b);

      if (edges_a != edges_b) {
        differs = true;
        if (flags & layout_diff::f_silent) {
          return false;
        }
        tl::error << "layout_diff: edges differ";
        if (verbose) {
          tl::info << "Not in b but in a:";
          print_diffs (prop_normalize_b.pr (), edges_a, edges_b);
          tl::info << "Not in a but in b:";
          print_diffs (prop_normalize_b.pr (), edges_b, edges_a);
        }
      }

    }

  }

  return ! differs;

}

}

