
/*

  KLayout Layout Viewer
  Copyright (C) 2006-2007 Matthias Koefferlein

  This program is free software; you can redistribute it and/or modify
  it under the terms of the GNU General Public License as published by
  the Free Software Foundation; either version 2 of the License, or
  (at your option) any later version.

  This program is distributed in the hope that it will be useful,
  but WITHOUT ANY WARRANTY; without even the implied warranty of
  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
  GNU General Public License for more details.

  You should have received a copy of the GNU General Public License
  along with this program; if not, write to the Free Software
  Foundation, Inc., 51 Franklin St, Fifth Floor, Boston, MA  02110-1301  USA

*/



#include "dbOASISReader.h"
#include "dbStream.h"
#include "dbObjectWithProperties.h"
#include "dbArray.h"

#include "tlException.h"
#include "tlString.h"
#include "tlClassRegistry.h"

namespace db
{

// ---------------------------------------------------------------
//  OASISReader


OASISReader::OASISReader (tl::InputStream &s)
  : m_stream (s), 
    m_progress (tl::translate ("Reading OASIS file"), 0, 10000),
    mm_repetition (this, "repetition"),
    mm_placement_cell (this, "placement-cell"),
    mm_placement_x (this, "playcement-x"),
    mm_placement_y (this, "playcement-y"),
    mm_layer (this, "layer"),
    mm_datatype (this, "datatype"),
    mm_textlayer (this, "textlayer"),
    mm_texttype (this, "texttype"),
    mm_text_x (this, "text-x"),
    mm_text_y (this, "text-y"),
    mm_text_string (this, "text-string"),
    mm_geometry_x (this, "geometry-x"),
    mm_geometry_y (this, "geometry-y"),
    mm_geometry_w (this, "geometry-w"),
    mm_geometry_h (this, "geometry-h"),
    mm_polygon_point_list (this, "polygon-point-list"),
    mm_path_halfwidth (this, "path-halfwidth"),
    mm_path_start_extension (this, "path-start-extension"),
    mm_path_end_extension (this, "path-end-extension"),
    mm_path_point_list (this, "path-point-list"),
    mm_ctrapezoid_type (this, "ctrapezoid-type"),
    mm_circle_radius (this, "circle-radius"),
    mm_last_property_name (this, "last-property-name"),
    mm_last_property_is_sprop (this, "last-property-is-stdprop"),
    mm_last_value_list(this, "last-value-list")
{
  m_progress.set_format (tl::translate ("%.0f MB"));
  m_progress.set_res (1.0);
  m_progress.set_unit (1024 * 1024);
}

OASISReader::~OASISReader ()
{
  //  .. nothing yet ..
}

const LayerMap &
OASISReader::read (db::Layout &layout, const LayerMap &map, bool create) throw (tl::Exception)
{
  m_layer_map = map;
  layout.start_changes ();
  do_read (layout, create);
  layout.end_changes ();
  return m_layer_map;
}

const LayerMap &
OASISReader::read (db::Layout &layout) throw (tl::Exception)
{
  m_layer_map = LayerMap ();
  layout.start_changes ();
  do_read (layout, true /*create layers*/);
  layout.end_changes ();
  return m_layer_map;
}

inline long 
OASISReader::get_long () throw (tl::Exception)
{
  unsigned long u = get_ulong ();
  if ((u & 1) != 0) {
    return -long (u >> 1);
  } else {
    return long (u >> 1);
  }
}

inline unsigned long 
OASISReader::get_ulong () throw (tl::Exception)
{
  unsigned long v = 0;
  unsigned long vm = 1;
  char c;
  
  do {
    unsigned char *b = (unsigned char *) m_stream.get (1);
    if (! b) {
      error (tl::translate ("Unexpected end-of-file"));
    }
    c = *b;
    if (vm > std::numeric_limits <unsigned long>::max () / 128 && 
        (unsigned long) (c & 0x7f) > (std::numeric_limits <unsigned long>::max () / vm)) {
      error (tl::translate ("Unsigned long value overflow"));
    }
    v += (unsigned long) (c & 0x7f) * vm;
    vm <<= 7;
  } while ((c & 0x80) != 0);

  return v;
}

inline int 
OASISReader::get_int () throw (tl::Exception)
{
  unsigned int u = get_uint ();
  if ((u & 1) != 0) {
    return -int (u >> 1);
  } else {
    return int (u >> 1);
  }
}

inline unsigned int 
OASISReader::get_uint () throw (tl::Exception)
{
  unsigned int v = 0;
  unsigned int vm = 1;
  char c;
  
  do {
    unsigned char *b = (unsigned char *) m_stream.get (1);
    if (! b) {
      error (tl::translate ("Unexpected end-of-file"));
    }
    c = *b;
    if (vm > std::numeric_limits <unsigned int>::max () / 128 && 
        (unsigned int) (c & 0x7f) > (std::numeric_limits <unsigned int>::max () / vm)) {
      error (tl::translate ("Unsigned integer value overflow"));
    }
    v += (unsigned int) (c & 0x7f) * vm;
    vm <<= 7;
  } while ((c & 0x80) != 0);

  return v;
}

std::string 
OASISReader::get_str () throw (tl::Exception)
{
  std::string s;
  get_str (s);
  return s;
}

void
OASISReader::get_str (std::string &s) throw (tl::Exception)
{
  size_t l = 0;
  get (l);

  char *b = (char *) m_stream.get (l);
  s.assign (b, l);
}

double
OASISReader::get_real () throw (tl::Exception)
{
  unsigned int t = get_uint ();

  if (t == 0) {

    return double (get_ulong ()); 

  } else if (t == 1) {

    return -double (get_ulong ()); 

  } else if (t == 2) {

    return 1.0 / double (get_ulong ()); 

  } else if (t == 3) {

    return -1.0 / double (get_ulong ()); 

  } else if (t == 4) {

    double d = double (get_ulong ());
    return d / double (get_ulong ()); 

  } else if (t == 5) {

    double d = double (get_ulong ());
    return -d / double (get_ulong ()); 

  } else if (t == 6) {

    union {
      float f;
      uint32_t i;
    } i2f;

    unsigned char *b = (unsigned char *) m_stream.get (sizeof (i2f.i));
    if (! b) {
      error (tl::translate ("Unexpected end-of-file"));
    }
    i2f.i = 0;
    b += sizeof (i2f.i);
    for (unsigned int i = 0; i < sizeof (i2f.i); ++i) {
      i2f.i = (i2f.i << 8) + uint32_t (*--b);
    }

    return double (i2f.f);

  } else if (t == 7) {

    union {
      double d;
      uint64_t i;
    } i2f;

    unsigned char *b = (unsigned char *) m_stream.get (sizeof (i2f.i));
    if (! b) {
      error (tl::translate ("Unexpected end-of-file"));
    }
    i2f.i = 0;
    b += sizeof (i2f.i);
    for (unsigned int i = 0; i < sizeof (i2f.i); ++i) {
      i2f.i = (i2f.i << 8) + uint64_t (*--b);
    }

    return double (i2f.d);

  } else {
    error (tl::sprintf (tl::translate ("Invalid real type %d").c_str (), t));
    return 0.0;
  }
}

db::Coord
OASISReader::get_ucoord (unsigned long grid) throw (tl::Exception)
{
  unsigned long lx = 0;
  get (lx);
  lx *= grid;
  if (lx > (unsigned long) (std::numeric_limits <db::Coord>::max ())) {
    error (tl::translate ("Coordinate value overflow"));
  }
  return db::Coord (lx);
}

db::Coord
OASISReader::get_coord (long grid) throw (tl::Exception)
{
  long lx = 0;
  get (lx);
  lx *= grid;
  if (lx < long (std::numeric_limits <db::Coord>::min ()) ||
      lx > long (std::numeric_limits <db::Coord>::max ())) {
    error (tl::translate ("Coordinate value overflow"));
  }
  return db::Coord (lx);
}

db::Point
OASISReader::get_2delta (long grid) throw (tl::Exception)
{
  unsigned long l1 = 0;
  get (l1);

  long lx = l1 >> 2;
  lx *= grid;
  if (lx > long (std::numeric_limits <db::Coord>::max ())) {
    error (tl::translate ("Coordinate value overflow"));
  }
  db::Coord x = lx;

  switch (l1 & 3) {
  case 0:
    return db::Point (x, 0);
  case 1:
    return db::Point (0, x);
  case 2:
    return db::Point (-x, 0);
  case 3:
  default:
    return db::Point (0, -x);
  }
}

db::Point
OASISReader::get_3delta (long grid) throw (tl::Exception)
{
  unsigned long l1 = 0;
  get (l1);

  long lx = l1 >> 3;
  lx *= grid;
  if (lx > long (std::numeric_limits <db::Coord>::max ())) {
    error (tl::translate ("Coordinate value overflow"));
  }
  db::Coord x = lx;

  switch (l1 & 7) {
  case 0:
    return db::Point (x, 0);
  case 1:
    return db::Point (0, x);
  case 2:
    return db::Point (-x, 0);
  case 3:
    return db::Point (0, -x);
  case 4:
    return db::Point (x, x);
  case 5:
    return db::Point (-x, x);
  case 6:
    return db::Point (-x, -x);
  case 7:
  default:
    return db::Point (x, -x);
  }
}

db::Point
OASISReader::get_gdelta (long grid) throw (tl::Exception)
{
  unsigned long l1 = 0;
  get (l1);

  if ((l1 & 1) != 0) {

    long lx = ((l1 & 2) == 0 ? long (l1 >> 2) : -long (l1 >> 2));
    lx *= grid;
    if (lx < long (std::numeric_limits <db::Coord>::min ()) ||
        lx > long (std::numeric_limits <db::Coord>::max ())) {
      error (tl::translate ("Coordinate value overflow"));
    }

    long ly;
    get (ly);
    ly *= grid;
    if (ly < long (std::numeric_limits <db::Coord>::min ()) ||
        ly > long (std::numeric_limits <db::Coord>::max ())) {
      error (tl::translate ("Coordinate value overflow"));
    }
    
    return db::Point (db::Coord (lx), db::Coord (ly));

  } else {

    long lx = l1 >> 4;
    lx *= grid;
    if (lx > long (std::numeric_limits <db::Coord>::max ())) {
      error (tl::translate ("Coordinate value overflow"));
    }
    db::Coord x = lx;

    switch ((l1 >> 1) & 7) {
    case 0:
      return db::Point (x, 0);
    case 1:
      return db::Point (0, x);
    case 2:
      return db::Point (-x, 0);
    case 3:
      return db::Point (0, -x);
    case 4:
      return db::Point (x, x);
    case 5:
      return db::Point (-x, x);
    case 6:
      return db::Point (-x, -x);
    case 7:
    default:
      return db::Point (x, -x);
    }

  }
}

void 
OASISReader::error (const std::string &msg) throw (tl::Exception)
{
  throw OASISReaderException (msg, m_stream.pos (), m_cellname.c_str ());
}

void 
OASISReader::warn (const std::string &msg) 
{
  // @@@ simple handling ..
  fprintf (stderr, tl::translate ("WARNING: %s (position=%ld, cell=%s)\n").c_str (), msg.c_str (), m_stream.pos (), m_cellname.c_str ());
}

std::pair <bool, unsigned int> 
OASISReader::open_dl (db::Layout &layout, const LDPair &dl, bool create) throw (tl::Exception)
{
  std::pair<bool, unsigned int> ll = m_layer_map.logical (dl);
  if (ll.first || ! create) {
    return ll;
  } else {

    //  and create the layer
    db::LayerProperties lp;
    lp.layer = dl.layer;
    lp.datatype = dl.datatype;

    //  resolve OASIS name if possible
    const tl::interval_map <db::ld_type, std::string> *names_dmap = m_layernames.mapped (dl.layer);
    if (names_dmap != 0) {
      const std::string *name = names_dmap->mapped (dl.datatype);
      if (name != 0) {
        lp.name = *name;
      }
    }
      
    unsigned int ll = layout.insert_layer (lp);
    m_layer_map.map (dl, ll);

    return std::make_pair (true, ll);

  }
}

/// A helper class to join two datatype layer name map members
struct LNameJoinOp1
{
  void operator() (std::string &a, const std::string &b)
  {
    if (a != b) {
      if (! a.empty ()) {
        a += ";";
      } 
      a += b;
    }
  }
};

/// A helper class to join two layer map members 
/// (this implementation basically merged the datatype maps)  
struct LNameJoinOp2
{
  void operator() (tl::interval_map<db::ld_type, std::string> &a, const tl::interval_map<db::ld_type, std::string> &b)
  {
    LNameJoinOp1 op1;
    a.add (b.begin (), b.end (), op1);
  }
};

static const char magic_bytes[] = { "%SEMI-OASIS\015\012" };

void 
OASISReader::do_read (db::Layout &layout, bool create_layers) throw (tl::Exception)
{
  unsigned char r;
  char *mb;

  //  read magic bytes
  mb = (char *) m_stream.get (sizeof (magic_bytes) - 1);
  if (! mb) {
    error (tl::translate ("File too short"));
  }
  if (strncmp (mb, magic_bytes, sizeof (magic_bytes) - 1) != 0) {
    error (tl::translate ("Format error (missing magic bytes)"));
  }

  //  read first record
  r = get_byte ();
  if (r != 1) {
    error (tl::translate ("Format error (START record expected)"));
  }

  std::string v = get_str ();
  if (v != "1.0") {
    error (tl::sprintf (tl::translate ("Format error (only version 1.0 is supported, file has version %s)").c_str (), v.c_str ()));
  }

  double res = get_real ();
  if (res < 1e-6) {
    error (tl::sprintf (tl::translate ("Invalid resolution of %g").c_str (), res));
  }

  //  compute database unit in pixel per meter
  m_dbu = 1.0e-6 / res;
  layout.dbu (m_dbu * 1e6);

  //  read over table offsets if required
  bool table_offsets_at_end = get_uint ();
  if (! table_offsets_at_end) {
    for (unsigned int i = 0; i < 12; ++i) {
      get_ulong ();
    }
  }

  //  define the name id counters
  unsigned long cellname_id = 0;
  unsigned long textstring_id = 0;
  unsigned long propstring_id = 0;
  unsigned long propname_id = 0;

  //  id mode (explicit or implicit)
  enum id_mode { any, expl, impl };
  id_mode cellname_id_mode = any;
  id_mode textstring_id_mode = any;
  id_mode propstring_id_mode = any;
  id_mode propname_id_mode = any;

  m_cellnames.clear ();
  m_textstrings.clear ();
  m_propstrings.clear ();
  m_propnames.clear ();
  m_layernames.clear ();
  m_cells_by_id.clear ();
  m_cells_by_name.clear ();
  m_defined_cells_by_id.clear ();
  m_defined_cells_by_name.clear ();

  tl::vector<db::CellInstArray> instances;
  tl::vector<db::CellInstArrayWithProperties> instances_with_props;

  //  read next record
  while (true) {

    r = get_byte ();

    if (r == 0 /*PAD*/) {

      //  simply skip.

    } else if (r == 2 /*END*/) {

      //  done
      break;

    } else if (r == 3 || r == 4 /*CELLNAME*/) {

      //  read a cell name
      std::string name = get_str ();

      //  and the associated id
      unsigned long id = cellname_id;
      if (r == 3) {
        if (cellname_id_mode == expl) {
          error (tl::translate ("Explicit and implicit CELLNAME modes cannot be mixed"));
        }
        cellname_id_mode = impl;
        ++cellname_id;
      } else {
        if (cellname_id_mode == impl) {
          error (tl::translate ("Explicit and implicit CELLNAME modes cannot be mixed"));
        }
        cellname_id_mode = expl;
        get (id);
      }

      if (! m_cellnames.insert (std::make_pair (id, name)).second) {
        error (tl::sprintf (tl::translate ("A CELLNAME with id %ld is present already").c_str (), id));
      }

    } else if (r == 5 || r == 6 /*TEXTSTRING*/) {

      //  read a text string
      std::string name = get_str ();

      //  and the associated id
      unsigned long id = textstring_id;
      if (r == 5) {
        if (textstring_id_mode == expl) {
          error (tl::translate ("Explicit and implicit TEXTSTRING modes cannot be mixed"));
        }
        textstring_id_mode = impl;
        ++textstring_id;
      } else {
        if (textstring_id_mode == impl) {
          error (tl::translate ("Explicit and implicit TEXTSTRING modes cannot be mixed"));
        }
        textstring_id_mode = expl;
        get (id);
      }

      if (! m_textstrings.insert (std::make_pair (id, name)).second) {
        error (tl::sprintf (tl::translate ("A TEXTSTRING with id %ld is present already").c_str (), id));
      }

    } else if (r == 7 || r == 8 /*PROPNAME*/) {

      //  read a property name
      std::string name = get_str ();

      //  and the associated id
      unsigned long id = propname_id;
      if (r == 7) {
        if (propname_id_mode == expl) {
          error (tl::translate ("Explicit and implicit PROPNAME modes cannot be mixed"));
        }
        propname_id_mode = impl;
        ++propname_id;
      } else {
        if (propname_id_mode == impl) {
          error (tl::translate ("Explicit and implicit PROPNAME modes cannot be mixed"));
        }
        propname_id_mode = expl;
        get (id);
      }

      if (! m_propnames.insert (std::make_pair (id, name)).second) {
        error (tl::sprintf (tl::translate ("A PROPNAME with id %ld is present already").c_str (), id));
      }

    } else if (r == 9 || r == 10 /*PROPSTRING*/) {

      //  read a property string
      std::string name = get_str ();

      //  and the associated id
      unsigned long id = propstring_id;
      if (r == 9) {
        if (propstring_id_mode == expl) {
          error (tl::translate ("Explicit and implicit PROPSTRING modes cannot be mixed"));
        }
        propstring_id_mode = impl;
        ++propstring_id;
      } else {
        if (propstring_id_mode == impl) {
          error (tl::translate ("Explicit and implicit PROPSTRING modes cannot be mixed"));
        }
        propstring_id_mode = expl;
        get (id);
      }

      if (! m_propstrings.insert (std::make_pair (id, name)).second) {
        error (tl::sprintf (tl::translate ("A PROPSTRING with id %ld is present already").c_str (), id));
      }

    } else if (r == 11 || r == 12 /*LAYERNAME*/) {

      //  read a layer name 
      std::string name = get_str ();

      db::ld_type dt1 = 0, dt2 = std::numeric_limits<db::ld_type>::max () - 1;
      db::ld_type l1 = 0, l2 = std::numeric_limits<db::ld_type>::max () - 1;
      unsigned int it;

      it = get_uint ();
      if (it == 0) {
        //  keep limits
      } else if (it == 1) {
        l2 = get_uint ();
      } else if (it == 2) {
        l1 = get_uint ();
      } else if (it == 3) {
        l1 = get_uint ();
        l2 = l1;
      } else if (it == 4) {
        l1 = get_uint ();
        l2 = get_uint ();
      } else {
        error (tl::translate ("Invalid LAYERNAME interval mode (layer)"));
      }

      it = get_uint ();
      if (it == 0) {
        //  keep limits
      } else if (it == 1) {
        dt2 = get_uint ();
      } else if (it == 2) {
        dt1 = get_uint ();
      } else if (it == 3) {
        dt1 = get_uint ();
        dt2 = dt1;
      } else if (it == 4) {
        dt1 = get_uint ();
        dt2 = get_uint ();
      } else {
        error (tl::translate ("Invalid LAYERNAME interval mode (datatype)"));
      }

      //  add to the layer name map
      tl::interval_map <db::ld_type, std::string> dt_map;
      LNameJoinOp1 op1;
      dt_map.add (dt1, dt2 + 1, name, op1);
      LNameJoinOp2 op2;
      m_layernames.add (l1, l2 + 1, dt_map, op2);

    } else if (r == 28 || r == 29 /*PROPERTY*/) {

      //  unrecognized property: parse over
      if (r == 28) {
        read_properties ();
      }

    } else if (r == 30 || r == 31 /*XNAME*/) {

      //  read a XNAME: it is simply ignored
      get_ulong ();
      get_str ();
      if (r == 31) {
        get_ulong ();
      }

    } else if (r == 13 || r == 14 /*CELL*/) {

      db::Layout::cell_index_type cell_index = 0;

      //  read a cell
      if (r == 13) {

        unsigned long id = 0;
        get (id);
        if (! m_defined_cells_by_id.insert (id).second) {
          error (tl::sprintf (tl::translate ("A cell with id %ld is defined already").c_str (), id));
        }

        std::map <unsigned long, db::Layout::cell_index_type>::const_iterator c = m_cells_by_id.find (id);
        if (c != m_cells_by_id.end ()) {
          cell_index = c->second;
        } else {
          std::map <unsigned long, std::string>::const_iterator name = m_cellnames.find (id);
          if (name == m_cellnames.end ()) {
            cell_index = layout.add_cell ();
            m_forward_references.insert (std::make_pair (id, cell_index));
          } else {
            cell_index = layout.add_cell (name->second.c_str ());
            m_cells_by_name.insert (std::make_pair (name->second, cell_index));
          }
          m_cells_by_id.insert (std::make_pair (id, cell_index));
        }

      } else {

        std::string name = get_str ();
        if (! m_defined_cells_by_name.insert (name).second) {
          error (tl::sprintf (tl::translate ("A cell with name %s is defined already").c_str (), name.c_str ()));
        }

        std::map <std::string, db::Layout::cell_index_type>::const_iterator c = m_cells_by_name.find (name);
        if (c != m_cells_by_name.end ()) {
          cell_index = c->second;
        } else {
          cell_index = layout.add_cell (name.c_str ());
          m_cells_by_name.insert (std::make_pair (name, cell_index));
        }

      }

      //  erase current instance list 
      instances.erase (instances.begin (), instances.end ());
      instances_with_props.erase (instances_with_props.begin (), instances_with_props.end ());

      do_read_cell (cell_index, layout, create_layers, instances, instances_with_props);

      //  insert all instances collected (inserting them once is 
      //  more effective than doing this every time)
      if (! instances.empty ()) {
        layout.cell (cell_index).insert (instances.begin (), instances.end ());
      }
      if (! instances_with_props.empty ()) {
        layout.cell (cell_index).insert (instances_with_props.begin (), instances_with_props.end ());
      }

    } else if (r == 34 /*CBLOCK*/) {

      unsigned int type = get_uint ();
      if (type != 0) {
        error (tl::sprintf (tl::translate ("Invalid CBLOCK compression type %d").c_str (), type));
      }

      get_uint ();  // uncomp-byte-count - not needed
      get_uint ();  // comp-byte-count - not needed

      //  put the stream into deflating mode
      m_stream.deflate ();

    } else {
      error (tl::sprintf (tl::translate ("Invalid record type on global level %d").c_str (), int (r)));
    }

  }

  //  read over tail and discard
  mb = (char *) m_stream.get (254);
  if (! mb) {
    error (tl::translate ("Format error (too few bytes after END record)"));
  }

  //  check if there are no more bytes
  mb = (char *) m_stream.get (254);
  if (mb) {
    error (tl::translate ("Format error (too many bytes after END record)"));
  }

  for (std::map <unsigned long, db::Layout::cell_index_type>::const_iterator fw = m_forward_references.begin (); fw != m_forward_references.end (); ++fw) {
    std::map <unsigned long, std::string>::const_iterator cn = m_cellnames.find (fw->first);
    if (cn == m_cellnames.end ()) {
      error (tl::sprintf (tl::translate ("No cellname declared for cell id %ld").c_str (), fw->first));
    } else {
      layout.rename_cell (fw->second, cn->second.c_str ());
    }
  }
}

std::pair <bool, db::PropertiesRepository::properties_id_type> 
OASISReader::read_element_properties (db::PropertiesRepository &rep)
{
  bool any = false;
  db::PropertiesRepository::properties_set properties;

  while (true) {

    unsigned char m = get_byte ();

    if (m == 28) {
      read_properties ();
    } else if (m != 29) {
      m_stream.unget (1);
      break;
    }

    //  @@@ :KLUDGE: instead of storing the name, the last property name id could be
    //               stored in mm_last_property_name @@@

    if (mm_last_property_name.get () == "S_GDS_PROPERTY" && mm_last_property_is_sprop.get ()) {

      if (mm_last_value_list.get ().size () != 2) {
        error (tl::translate ("S_GDS_PROPERTY must have a value list with exactly two elements"));
      }

      properties.insert (std::make_pair (rep.prop_name_id (mm_last_value_list.get () [0]), mm_last_value_list.get () [1]));

    } else if (mm_last_value_list.get ().size () == 0) {
      properties.insert (std::make_pair (rep.prop_name_id (tl::Variant (mm_last_property_name.get ())), 
                                         tl::Variant ()));
    } else if (mm_last_value_list.get ().size () == 1) {
      properties.insert (std::make_pair (rep.prop_name_id (tl::Variant (mm_last_property_name.get ())), 
                                         tl::Variant (mm_last_value_list.get () [0])));
    } else if (mm_last_value_list.get ().size () > 1) {
      properties.insert (std::make_pair (rep.prop_name_id (tl::Variant (mm_last_property_name.get ())), 
                                         tl::Variant (mm_last_value_list.get ().begin (), mm_last_value_list.get ().end ())));
    }

    any = true;

  } 

  if (any) {
    return std::make_pair (true, rep.properties_id (properties));
  } else {
    return std::make_pair (false, 0);
  }
}

void 
OASISReader::read_properties ()
{
  unsigned char m = get_byte ();

  if (m & 0x04) {
    if (m & 0x02) {

      unsigned long id;
      get (id);

      std::map <unsigned long, std::string>::const_iterator cid = m_propnames.find (id);
      if (cid == m_propnames.end ()) {
        error (tl::sprintf (tl::translate ("No property name string defined for id %ld (must be declared before name is used)").c_str (), id));
      }

      mm_last_property_name = cid->second;

    } else {
      mm_last_property_name = get_str ();
    }
  }

  mm_last_property_is_sprop = ((m & 0x01) != 0);

  if (! (m & 0x08)) {

    unsigned long n = ((unsigned long) (m >> 4)) & 0x0f;
    if (n == 15) {
      get (n);
    }

    mm_last_value_list.get_non_const ().clear ();
    mm_last_value_list.get_non_const ().reserve (n);

    while (n > 0) {

      unsigned char t = get_byte ();
      if (t < 8) {

        m_stream.unget (1);
        double v = get_real ();
        mm_last_value_list.get_non_const ().push_back (tl::Variant (v));

      } else if (t == 8) {

        unsigned long l;
        get (l);
        mm_last_value_list.get_non_const ().push_back (tl::Variant (long (l)));

      } else if (t == 9) {

        long l;
        get (l);
        mm_last_value_list.get_non_const ().push_back (tl::Variant (l));

      } else if (t == 10 || t == 11 || t == 12) {

        mm_last_value_list.get_non_const ().push_back (tl::Variant (get_str ()));

      } else if (t == 13 || t == 14 || t == 15) {

        unsigned long id;
        get (id);
        std::map <unsigned long, std::string>::const_iterator sid = m_propstrings.find (id);
        if (sid == m_propnames.end ()) {
          error (tl::sprintf (tl::translate ("No property value string defined for id %ld (must be declared before name is used)").c_str (), id));
        }
        mm_last_value_list.get_non_const ().push_back (tl::Variant (sid->second));

      } else {
        error (tl::sprintf (tl::translate ("Invalid property value type %d").c_str (), int (t)));
      }

      --n;

    }

    mm_last_value_list.set_initialized ();

  }
}


      
void 
OASISReader::read_pointlist (modal_variable <std::vector <db::Point> > &pointlist, bool for_polygon)
{
  unsigned int type = get_uint ();
  
  unsigned long n = 0;
  get (n);
  if (n == 0) {
    error (tl::translate ("Invalid point list: length is zero").c_str ());
  }

  pointlist.get_non_const ().clear ();
  if ((type == 0 || type == 1) && for_polygon) {
    //  because for polygons, the pointlist will be closed implicitly
    pointlist.get_non_const ().reserve (n + 2);
  } else {
    pointlist.get_non_const ().reserve (n + 1);
  }

  pointlist.get_non_const ().push_back (db::Point ());

  if (type == 0 || type == 1) {

    bool h = (type == 0);

    db::Point pos;
    for (unsigned long i = 0; i < n; ++i) {
      db::Coord d = get_coord ();
      if (h) {
        pos += db::Point (d, 0);
      } else {
        pos += db::Point (0, d);
      }
      h = ! h;
      pointlist.get_non_const ().push_back (pos);
    }

    //  synthesize the last point for polygons
    if (for_polygon) {
      if ((n % 2) != 0) {
        warn (tl::translate ("Type 0 or 1 point list with odd number of points is illegal"));
      }
      if (h) {
        pointlist.get_non_const ().push_back (db::Point (0, pos.y ()));
      } else {
        pointlist.get_non_const ().push_back (db::Point (pos.x (), 0));
      }
    }

  } else if (type == 2) {

    db::Point pos;
    for (unsigned long i = 0; i < n; ++i) {
      pos += get_2delta ();
      pointlist.get_non_const ().push_back (pos);
    }

  } else if (type == 3) {

    db::Point pos;
    for (unsigned long i = 0; i < n; ++i) {
      pos += get_3delta ();
      pointlist.get_non_const ().push_back (pos);
    }

  } else if (type == 4) {

    db::Point pos;
    for (unsigned long i = 0; i < n; ++i) {
      pos += get_gdelta ();
      pointlist.get_non_const ().push_back (pos);
    }

  } else if (type == 5) {

    db::Point pos;
    db::Point delta;
    for (unsigned long i = 0; i < n; ++i) {
      delta += get_gdelta ();
      pos += delta;
      pointlist.get_non_const ().push_back (pos);
    }

  } else {
    error (tl::sprintf (tl::translate ("Invalid point list type %d").c_str (), type));
  }

  pointlist.set_initialized ();
}

void
OASISReader::read_repetition ()
{
  unsigned char type = get_uint ();
  if (type == 0) {
    
    //  reuse modal variable

  } else if (type == 1) {

    unsigned long nx = 0, ny = 0;
    get (nx); 
    get (ny);

    db::Coord dx = get_ucoord ();
    db::Coord dy = get_ucoord ();

    mm_repetition = new RegularRepetition (db::Point (dx, 0), db::Point (0, dy), nx + 2, ny + 2);

  } else if (type == 2) {

    unsigned long nx = 0;
    get (nx); 

    db::Coord dx = get_ucoord ();

    mm_repetition = new RegularRepetition (db::Point (dx, 0), db::Point (0, 0), nx + 2, 1);

  } else if (type == 3) {

    unsigned long ny = 0;
    get (ny);

    db::Coord dy = get_ucoord ();

    mm_repetition = new RegularRepetition (db::Point (0, 0), db::Point (0, dy), 1, ny + 2);

  } else if (type == 4 || type == 5) {
    
    IrregularRepetition *rep = new IrregularRepetition ();
    mm_repetition = rep;

    unsigned long n = 0;
    get (n);

    unsigned long lgrid = 1;
    if (type == 5) {
      get (lgrid);
    }

    rep->reserve (n + 1);

    db::Coord x = 0;
    for (unsigned long i = 0; i <= n; ++i) {
      x += get_ucoord (lgrid);
      rep->push_back (db::Point (x, 0));
    }

  } else if (type == 6 || type == 7) {
    
    IrregularRepetition *rep = new IrregularRepetition ();
    mm_repetition = rep;

    unsigned long n = 0;
    get (n);

    unsigned long lgrid = 1;
    if (type == 7) {
      get (lgrid);
    }

    rep->reserve (n + 1);

    db::Coord x = 0;
    for (unsigned long i = 0; i <= n; ++i) {
      x += get_ucoord (lgrid);
      rep->push_back (db::Point (0, x));
    }

  } else if (type == 8) {

    unsigned long n = 0, m = 0;

    get (n); 
    get (m);
    db::Point dn = get_gdelta (); 
    db::Point dm = get_gdelta (); 

    mm_repetition = new RegularRepetition (dn, dm, n + 2, m + 2);

  } else if (type == 9) {

    unsigned long n = 0;
    get (n); 
    db::Point dn = get_gdelta (); 

    mm_repetition = new RegularRepetition (dn, db::Point (0, 0), n + 2, 1);

  } else if (type == 10) {

    IrregularRepetition *rep = new IrregularRepetition ();
    mm_repetition = rep;

    unsigned long n = 0;
    get (n);

    rep->reserve (n + 1);

    db::Point p;
    for (unsigned long i = 0; i <= n; ++i) {
      p += get_gdelta ();
      rep->push_back (p);
    }

  } else if (type == 11) {

    IrregularRepetition *rep = new IrregularRepetition ();
    mm_repetition = rep;

    unsigned long n = 0;
    get (n);

    unsigned long grid = 0;
    get (grid);

    rep->reserve (n + 1);

    db::Point p;
    for (unsigned long i = 0; i <= n; ++i) {
      p += get_gdelta (grid);
      rep->push_back (db::Point (p.x (), p.y ()));
    }

  } else {
    error (tl::sprintf (tl::translate ("Invalid repetition type %d").c_str (), type));
  }

}

void 
OASISReader::do_read_cell (db::Layout::cell_index_type cell_index, 
                           db::Layout &layout, bool create_layers, 
                           tl::vector<db::CellInstArray> &instances,
                           tl::vector<db::CellInstArrayWithProperties> &instances_with_props) throw (tl::Exception)
{
  m_progress.set (m_stream.pos ());
  m_cellname = layout.cell_name (cell_index);

  bool xy_absolute = true;

  //  reset modal variables
  mm_repetition.reset ();
  mm_placement_cell.reset ();
  mm_placement_x = 0;
  mm_placement_y = 0;
  mm_layer.reset ();
  mm_datatype.reset ();
  mm_textlayer.reset ();
  mm_texttype.reset ();
  mm_text_x = 0;
  mm_text_y = 0;
  mm_text_string.reset ();
  mm_geometry_x = 0;
  mm_geometry_y = 0;
  mm_geometry_w.reset ();
  mm_geometry_h.reset ();
  mm_polygon_point_list.reset ();
  mm_path_halfwidth.reset ();
  mm_path_start_extension.reset ();
  mm_path_end_extension.reset ();
  mm_path_point_list.reset ();
  mm_ctrapezoid_type.reset ();
  mm_circle_radius.reset ();
  mm_last_property_name.reset ();
  mm_last_property_is_sprop.reset ();
  mm_last_value_list.reset ();

  //  read next record
  while (true) {

    m_progress.set (m_stream.pos ());

    unsigned char r = get_byte ();

    if (r == 0 /*PAD*/) {

      //  simply skip.

    } else if (r == 15 /*XYABSOLUTE*/) {

      //  switch to absolute mode
      xy_absolute = true;

    } else if (r == 16 /*XYRELATIVE*/) {

      //  switch to relative mode
      xy_absolute = false;

    } else if (r == 17 || r == 18 /*PLACEMENT*/) {

      unsigned char m = get_byte ();

      //  locate cell
      if (m & 0x80) {

        if (m & 0x40) {

          //  cell by id
          unsigned long id;
          get (id);
          std::map <unsigned long, db::Layout::cell_index_type>::const_iterator cid = m_cells_by_id.find (id);
          if (cid == m_cells_by_id.end ()) {

            //  create the cell
            std::map <unsigned long, std::string>::const_iterator name = m_cellnames.find (id);
            if (name == m_cellnames.end ()) {
              mm_placement_cell = layout.add_cell ();
              m_forward_references.insert (std::make_pair (id, mm_placement_cell.get ()));
            } else {
              mm_placement_cell = layout.add_cell (name->second.c_str ());
              m_cells_by_name.insert (std::make_pair (name->second, mm_placement_cell.get ()));
            }
            m_cells_by_id.insert (std::make_pair (id, mm_placement_cell.get ()));

          } else {
            mm_placement_cell = cid->second;
          }

        } else {

          //  cell by name
          std::string name;
          get_str (name);

          std::map <std::string, db::Layout::cell_index_type>::const_iterator cid = m_cells_by_name.find (name);
          if (cid == m_cells_by_name.end ()) {

            //  create the cell
            mm_placement_cell = layout.add_cell (name.c_str ());
            m_cells_by_name.insert (std::make_pair (name, mm_placement_cell.get ()));

          } else {
            mm_placement_cell = cid->second;
          }

        }

      } 

      double mag = 1.0;
      bool mag_set = false;

      double angle_deg = 0.0; // only meaningful if angle < 0
      int angle = 0;
      bool mirror = false;

      if (r == 18) {

        if (m & 0x04) {
          mag = get_real ();
          if (fabs (mag - 1.0) > 1e-6) {
            mag_set = true;
          }
        }

        if (m & 0x02) {
          angle_deg = get_real ();
          double a = angle_deg / 90.0;
          if (a < -4 || a > 4) {
            error (tl::sprintf (tl::translate ("Invalid rotation angle (%g is less than -360 or larger than 360)").c_str (), angle_deg));
          }
          angle = int (a < 0 ? (a - 0.5) : (a + 0.5));
          if (fabs (double (angle) - a) > 1e-6) {
            angle = -1; // indicates arbitrary orientation. Take angle_deg instead
          } else {
            if (angle < 0) {
              angle += ((4 - 1) - angle) & ~(4 - 1);
            }
            angle = angle % 4;
          }
        }

      } else {
        angle = ((m & 0x06) >> 1);
      }
          
      mirror = (m & 0x01) != 0;

      if (m & 0x20) {
        db::Coord x;
        get (x);
        if (xy_absolute) {
          mm_placement_x = x;
        } else {
          mm_placement_x = x + mm_placement_x.get ();
        }
      }

      if (m & 0x10) {
        db::Coord y;
        get (y);
        if (xy_absolute) {
          mm_placement_y = y;
        } else {
          mm_placement_y = y + mm_placement_y.get ();
        }
      }

      db::Point pos (mm_placement_x.get (), mm_placement_y.get ());

      if (m & 0x8) {

        read_repetition ();

        std::pair<bool, db::PropertiesRepository::properties_id_type> pp = read_element_properties (layout.properties_repository ());

        db::Point a, b;
        size_t na, nb;
        if (mm_repetition.get ().is_regular (a, b, na, nb)) {

          db::CellInstArray inst;

          if (mag_set || angle < 0) {
            inst = db::CellInstArray (db::CellInst (mm_placement_cell.get ()), 
                                      db::CplxTrans (mag, angle_deg, mirror, pos), a, b, na, nb);
          } else {
            inst = db::CellInstArray (db::CellInst (mm_placement_cell.get ()), 
                                      db::Trans (angle, mirror, pos), a, b, na, nb);
          }

          if (pp.first) {
            instances_with_props.push_back (db::CellInstArrayWithProperties (inst, pp.second));
          } else {
            instances.push_back (inst);
          }

        } else {

          RepetitionIterator p = mm_repetition.get ().begin ();
          while (! p.at_end ()) {

            db::CellInstArray inst;

            if (mag_set || angle < 0) {
              inst = db::CellInstArray (db::CellInst (mm_placement_cell.get ()), db::CplxTrans (mag, angle_deg, mirror, pos + *p));
            } else {
              inst = db::CellInstArray (db::CellInst (mm_placement_cell.get ()), db::Trans (angle, mirror, pos + *p));
            }

            if (pp.first) {
              instances_with_props.push_back (db::CellInstArrayWithProperties (inst, pp.second));
            } else {
              instances.push_back (inst);
            }

            ++p;

          }

        }

      } else {

        std::pair<bool, db::PropertiesRepository::properties_id_type> pp = read_element_properties (layout.properties_repository ());

        db::CellInstArray inst;

        if (mag_set || angle < 0) {
          inst = db::CellInstArray (db::CellInst (mm_placement_cell.get ()), db::CplxTrans (mag, angle_deg, mirror, pos));
        } else {
          inst = db::CellInstArray (db::CellInst (mm_placement_cell.get ()), db::Trans (angle, mirror, pos));
        }

        if (pp.first) {
          instances_with_props.push_back (db::CellInstArrayWithProperties (inst, pp.second));
        } else {
          instances.push_back (inst);
        }

      }

    } else if (r == 19 /*TEXT*/) {

      unsigned char m = get_byte ();

      if (m & 0x40) {
        if (m & 0x20) {
          unsigned long id;
          get (id);
          std::map <unsigned long, std::string>::const_iterator tid = m_textstrings.find (id);
          if (tid == m_textstrings.end ()) {
            error (tl::sprintf (tl::translate ("No text string defined for id %ld (must be declared before text is used)").c_str (), id));
          }
          mm_text_string = tid->second;
        } else {
          mm_text_string = get_str ();
        }
      } 

      if (m & 0x1) {
        mm_textlayer = get_uint ();
      }

      if (m & 0x2) {
        mm_texttype = get_uint ();
      }

      if (m & 0x10) {
        db::Coord x;
        get (x);
        if (xy_absolute) {
          mm_text_x = x;
        } else {
          mm_text_x = x + mm_text_x.get ();
        }
      }

      if (m & 0x8) {
        db::Coord y;
        get (y);
        if (xy_absolute) {
          mm_text_y = y;
        } else {
          mm_text_y = y + mm_text_y.get ();
        }
      }

      db::Point pos (mm_text_x.get (), mm_text_y.get ());

      std::pair<bool, unsigned int> ll = open_dl (layout, LDPair (mm_textlayer.get (), mm_texttype.get ()), create_layers);

      if (m & 0x4) {

        read_repetition ();

        //  @@@ :KLUDGE: should not read properties if layer is not enabled!
        std::pair<bool, db::PropertiesRepository::properties_id_type> pp = read_element_properties (layout.properties_repository ());

        if (ll.first) {

          db::Text text (mm_text_string.get (), db::Trans ());
          db::Cell &cell = layout.cell (cell_index);

          //  If the repetion is a regular one, convert the repetition into
          //  a shape array (currently there is no shape array with properties ..)
          db::Point a, b;
          size_t na, nb;
          if (mm_repetition.get ().is_regular (a, b, na, nb)) {

            //  creating a TextPtr is most efficient with a normalized text because no displacement is provided
            db::Point d (text.box ().lower_left ());
            text.move (-d);
            db::TextPtr text_ptr (text, layout.shape_repository ());

            if (pp.first) {
              cell.shapes (ll.second).insert (db::object_with_properties<db::array<db::TextPtr, db::Disp> > (db::array<db::TextPtr, db::Disp> (text_ptr, db::Disp (d + pos), layout.array_repository (), a, b, na, nb), pp.second));
            } else {
              cell.shapes (ll.second).insert (db::array<db::TextPtr, db::Disp> (text_ptr, db::Disp (d + pos), layout.array_repository (), a, b, na, nb));
            }

          } else {

            RepetitionIterator p = mm_repetition.get ().begin ();
            db::TextRef text_ref (text, layout.shape_repository ());
            while (! p.at_end ()) {
              if (pp.first) {
                cell.shapes (ll.second).insert (db::TextRefWithProperties (text_ref.transformed (db::Disp (pos + *p)), pp.second));
              } else {
                cell.shapes (ll.second).insert (text_ref.transformed (db::Disp (pos + *p)));
              }
              ++p;
            }

          }

        }

      } else {

        std::pair<bool, db::PropertiesRepository::properties_id_type> pp = read_element_properties (layout.properties_repository ());

        if (ll.first) {
          db::Text text (mm_text_string.get (), db::Trans (pos));
          if (pp.first) {
            layout.cell (cell_index).shapes (ll.second).insert (db::TextRefWithProperties (db::TextRef (text, layout.shape_repository ()), pp.second));
          } else {
            layout.cell (cell_index).shapes (ll.second).insert (db::TextRef (text, layout.shape_repository ()));
          }
        }

      }

    } else if (r == 20 /*RECTANGLE*/) {

      unsigned char m = get_byte ();

      if (m & 0x1) {
        mm_layer = get_uint ();
      }

      if (m & 0x2) {
        mm_datatype = get_uint ();
      }

      if (m & 0x40) {
        mm_geometry_w = get_ucoord ();
      } 
      if (m & 0x80) {
        mm_geometry_h = mm_geometry_w; // @@@ really?
      } else {
        if (m & 0x20) {
          mm_geometry_h = get_ucoord ();
        } 
      }

      if (m & 0x10) {
        db::Coord x;
        get (x);
        if (xy_absolute) {
          mm_geometry_x = x;
        } else {
          mm_geometry_x = x + mm_geometry_x.get ();
        }
      }

      if (m & 0x8) {
        db::Coord y;
        get (y);
        if (xy_absolute) {
          mm_geometry_y = y;
        } else {
          mm_geometry_y = y + mm_geometry_y.get ();
        }
      }

      db::Box box (db::Point (mm_geometry_x.get (), mm_geometry_y.get ()),
                   db::Point (mm_geometry_x.get () + mm_geometry_w.get (), mm_geometry_y.get () + mm_geometry_h.get ()));

      std::pair<bool, unsigned int> ll = open_dl (layout, LDPair (mm_layer.get (), mm_datatype.get ()), create_layers);

      if (m & 0x4) {

        read_repetition ();

        std::pair<bool, db::PropertiesRepository::properties_id_type> pp = read_element_properties (layout.properties_repository ());

        if (ll.first) {

          db::Cell &cell = layout.cell (cell_index);

          //  If the repetion is a regular one, convert the repetition into
          //  a box array
          db::Point a, b;
          size_t na, nb;
          if (mm_repetition.get ().is_regular (a, b, na, nb)) {

            //  there is no box array currently, so we use a polygon array
            if (pp.first) {
              cell.shapes (ll.second).insert (db::object_with_properties<db::Shape::box_array_type> (db::Shape::box_array_type (box, db::unit_trans (), layout.array_repository (), a, b, na, nb), pp.second));
            } else {
              cell.shapes (ll.second).insert (db::Shape::box_array_type (box, db::unit_trans (), layout.array_repository (), a, b, na, nb));
            }

          } else {

            //  convert the OASIS record into the rectangle one by one.
            RepetitionIterator p = mm_repetition.get ().begin ();
            while (! p.at_end ()) {
              if (pp.first) {
                cell.shapes (ll.second).insert (db::BoxWithProperties (box.moved (*p), pp.second));
              } else {
                cell.shapes (ll.second).insert (box.moved (*p));
              }
              ++p;
            }

          }

        }

      } else {
        
        std::pair<bool, db::PropertiesRepository::properties_id_type> pp = read_element_properties (layout.properties_repository ());

        if (ll.first) {

          db::Cell &cell = layout.cell (cell_index);
          if (pp.first) {
            cell.shapes (ll.second).insert (db::BoxWithProperties (box, pp.second));
          } else {
            cell.shapes (ll.second).insert (box);
          }

        }

      }

    } else if (r == 21 /*POLYGON*/) {

      unsigned char m = get_byte ();

      if (m & 0x1) {
        mm_layer = get_uint ();
      }

      if (m & 0x2) {
        mm_datatype = get_uint ();
      }

      if (m & 0x20) {
        read_pointlist (mm_polygon_point_list, true);
      }

      if (m & 0x10) {
        db::Coord x;
        get (x);
        if (xy_absolute) {
          mm_geometry_x = x;
        } else {
          mm_geometry_x = x + mm_geometry_x.get ();
        }
      }

      if (m & 0x8) {
        db::Coord y;
        get (y);
        if (xy_absolute) {
          mm_geometry_y = y;
        } else {
          mm_geometry_y = y + mm_geometry_y.get ();
        }
      }

      db::Point pos (mm_geometry_x.get (), mm_geometry_y.get ());

      std::pair<bool, unsigned int> ll = open_dl (layout, LDPair (mm_layer.get (), mm_datatype.get ()), create_layers);

      if (m & 0x4) {

        read_repetition ();

        std::pair<bool, db::PropertiesRepository::properties_id_type> pp = read_element_properties (layout.properties_repository ());

        if (ll.first) {

          db::Cell &cell = layout.cell (cell_index);

          if (mm_polygon_point_list.get ().size () < 3) {
            warn (tl::translate ("POLYGON with less than 3 points ignored"));
          } else {

            //  If the repetion is a regular one, convert the repetition into
            //  a shape array (currently there is no shape array with properties ..)
            db::Point a, b;
            size_t na, nb;
            if (mm_repetition.get ().is_regular (a, b, na, nb)) {

              //  convert the OASIS record into the polygon.
              db::SimplePolygon poly;
              poly.assign_hull (mm_polygon_point_list.get ().begin (), mm_polygon_point_list.get ().end ());
              //  creating a SimplePolygonPtr is most efficient with a normalized polygon because no displacement is provided
              db::Point d (poly.box ().lower_left ());
              poly.move (-d);
              db::SimplePolygonPtr poly_ptr (poly, layout.shape_repository ());

              if (pp.first) {
                cell.shapes (ll.second).insert (db::object_with_properties<db::array<db::SimplePolygonPtr, db::Disp> > (db::array<db::SimplePolygonPtr, db::Disp> (poly_ptr, db::Disp (d + pos), layout.array_repository (), a, b, na, nb), pp.second));
              } else {
                cell.shapes (ll.second).insert (db::array<db::SimplePolygonPtr, db::Disp> (poly_ptr, db::Disp (d + pos), layout.array_repository (), a, b, na, nb));
              }

            } else {

              //  convert the OASIS record into the polygon.
              db::SimplePolygon poly;
              poly.assign_hull (mm_polygon_point_list.get ().begin (), mm_polygon_point_list.get ().end ());
              db::SimplePolygonRef poly_ref (poly, layout.shape_repository ());

              RepetitionIterator p = mm_repetition.get ().begin ();
              while (! p.at_end ()) {
                if (pp.first) {
                  cell.shapes (ll.second).insert (db::SimplePolygonRefWithProperties (poly_ref.transformed (db::Disp (pos + *p)), pp.second));
                } else {
                  cell.shapes (ll.second).insert (poly_ref.transformed (db::Disp (pos + *p)));
                }
                ++p;
              }
          
            }

          }

        }

      } else {

        std::pair<bool, db::PropertiesRepository::properties_id_type> pp = read_element_properties (layout.properties_repository ());

        if (ll.first) {

          if (mm_polygon_point_list.get ().size () < 3) {
            warn (tl::translate ("POLYGON with less than 3 points ignored"));
          } else {

            //  convert the OASIS record into the polygon.
            db::SimplePolygon poly;
            poly.assign_hull (mm_polygon_point_list.get ().begin (), mm_polygon_point_list.get ().end ());
            db::SimplePolygonRef poly_ref (poly, layout.shape_repository ());

            if (pp.first) {
              layout.cell (cell_index).shapes (ll.second).insert (db::SimplePolygonRefWithProperties (poly_ref.transformed (db::Disp (pos)), pp.second));
            } else {
              layout.cell (cell_index).shapes (ll.second).insert (poly_ref.transformed (db::Disp (pos)));
            }

          }

        }

      }

    } else if (r == 22 /*PATH*/) {

      unsigned char m = get_byte ();

      if (m & 0x1) {
        mm_layer = get_uint ();
      }

      if (m & 0x2) {
        mm_datatype = get_uint ();
      }

      if (m & 0x40) {
        mm_path_halfwidth = get_ucoord ();
      }

      if (m & 0x80) {

        unsigned int e = get_uint ();
        if ((e & 0x0c) == 0x0c) {
          mm_path_start_extension = get_coord ();
        } else if ((e & 0x04) == 0x04) {
          mm_path_start_extension = 0; //  @@@ is setting the start extension correct here?
        } else if ((e & 0x08) == 0x08) {
          mm_path_start_extension = mm_path_halfwidth.get (); //  @@@ is setting the start extension correct here?
        }
        if ((e & 0x03) == 0x03) {
          mm_path_end_extension = get_coord ();
        } else if ((e & 0x01) == 0x01) {
          mm_path_end_extension = 0; //  @@@ is setting the start extension correct here?
        } else if ((e & 0x02) == 0x02) {
          mm_path_end_extension = mm_path_halfwidth.get (); //  @@@ is setting the start extension correct here?
        }

      }

      if (m & 0x20) {
        read_pointlist (mm_path_point_list, false);
      }

      if (m & 0x10) {
        db::Coord x;
        get (x);
        if (xy_absolute) {
          mm_geometry_x = x;
        } else {
          mm_geometry_x = x + mm_geometry_x.get ();
        }
      }

      if (m & 0x8) {
        db::Coord y;
        get (y);
        if (xy_absolute) {
          mm_geometry_y = y;
        } else {
          mm_geometry_y = y + mm_geometry_y.get ();
        }
      }

      db::Point pos (mm_geometry_x.get (), mm_geometry_y.get ());

      std::pair<bool, unsigned int> ll = open_dl (layout, LDPair (mm_layer.get (), mm_datatype.get ()), create_layers);

      if (m & 0x4) {

        read_repetition ();

        std::pair<bool, db::PropertiesRepository::properties_id_type> pp = read_element_properties (layout.properties_repository ());

        if (ll.first) {

          if (mm_path_point_list.get ().size () < 2) {
            warn (tl::translate ("POLYGON with less than 2 points ignored"));
          } else {

            //  convert the OASIS record into the path.
            db::Path path;
            path.width (2 * mm_path_halfwidth.get ());
            path.extensions (mm_path_start_extension.get (), mm_path_end_extension.get ());
            path.assign (mm_path_point_list.get ().begin (), mm_path_point_list.get ().end ());

            db::Cell &cell = layout.cell (cell_index);

            //  If the repetion is a regular one, convert the repetition into
            //  a shape array (currently there is no shape array with properties ..)
            db::Point a, b;
            size_t na, nb;
            if (mm_repetition.get ().is_regular (a, b, na, nb)) {

              //  creating a PathPtr is most efficient with a normalized path because no displacement is provided
              db::Point d (path.box ().lower_left ());
              path.move (-d);
              db::PathPtr path_ptr (path, layout.shape_repository ());

              if (pp.first) {
                cell.shapes (ll.second).insert (db::object_with_properties<db::array<db::PathPtr, db::Disp> > (db::array<db::PathPtr, db::Disp> (path_ptr, db::Disp (d + pos), layout.array_repository (), a, b, na, nb), pp.second));
              } else {
                cell.shapes (ll.second).insert (db::array<db::PathPtr, db::Disp> (path_ptr, db::Disp (d + pos), layout.array_repository (), a, b, na, nb));
              }

            } else {

              db::PathRef path_ref (path, layout.shape_repository ());

              RepetitionIterator p = mm_repetition.get ().begin ();
              while (! p.at_end ()) {
                if (pp.first) {
                  cell.shapes (ll.second).insert (db::PathRefWithProperties (path_ref.transformed (db::Disp (pos + *p)), pp.second));
                } else {
                  cell.shapes (ll.second).insert (path_ref.transformed (db::Disp (pos + *p)));
                }
                ++p;
              }

            }
            
          }

        }

      } else {

        std::pair<bool, db::PropertiesRepository::properties_id_type> pp = read_element_properties (layout.properties_repository ());

        if (ll.first) {

          if (mm_path_point_list.get ().size () < 2) {
            warn (tl::translate ("PATH with less than 2 points ignored"));
          } else {

            //  convert the OASIS record into the path.
            db::Path path;
            path.width (2 * mm_path_halfwidth.get ());
            path.extensions (mm_path_start_extension.get (), mm_path_end_extension.get ());
            path.assign (mm_path_point_list.get ().begin (), mm_path_point_list.get ().end ());
            db::PathRef path_ref (path, layout.shape_repository ());

            if (pp.first) {
              layout.cell (cell_index).shapes (ll.second).insert (db::PathRefWithProperties (path_ref.transformed (db::Disp (pos)), pp.second));
            } else {
              layout.cell (cell_index).shapes (ll.second).insert (path_ref.transformed (db::Disp (pos)));
            }

          }

        }

      }

    } else if (r == 23 || r == 24 || r == 25 /*TRAPEZOID*/) {

      unsigned char m = get_byte ();

      if (m & 0x1) {
        mm_layer = get_uint ();
      }

      if (m & 0x2) {
        mm_datatype = get_uint ();
      }

      if (m & 0x40) {
        mm_geometry_w = get_ucoord ();
      }

      if (m & 0x20) {
        mm_geometry_h = get_ucoord ();
      }

      db::Coord delta_a = 0, delta_b = 0;
      if (r == 23 || r == 24) {
        delta_a = get_coord ();
      }
      if (r == 23 || r == 25) {
        delta_b = get_coord ();
      }

      if (m & 0x10) {
        db::Coord x;
        get (x);
        if (xy_absolute) {
          mm_geometry_x = x;
        } else {
          mm_geometry_x = x + mm_geometry_x.get ();
        }
      }

      if (m & 0x8) {
        db::Coord y;
        get (y);
        if (xy_absolute) {
          mm_geometry_y = y;
        } else {
          mm_geometry_y = y + mm_geometry_y.get ();
        }
      }

      db::Point pos (mm_geometry_x.get (), mm_geometry_y.get ());

      std::pair<bool, unsigned int> ll = open_dl (layout, LDPair (mm_layer.get (), mm_datatype.get ()), create_layers);

      db::Point pts [4];

      if (m & 0x80) {
        //  vertically
        pts [0] = db::Point (0, std::max (delta_a, 0));
        pts [1] = db::Point (0, mm_geometry_h.get () + std::min (delta_b, 0));
        pts [2] = db::Point (mm_geometry_w.get (), mm_geometry_h.get () - std::max (delta_b, 0));
        pts [3] = db::Point (mm_geometry_w.get (), -std::min (delta_a, 0));
      } else {
        //  horizontally
        pts [0] = db::Point (std::max (delta_a, 0), mm_geometry_h.get ());
        pts [1] = db::Point (mm_geometry_w.get () + std::min (delta_b, 0), mm_geometry_h.get ());
        pts [2] = db::Point (mm_geometry_w.get () - std::max (delta_b, 0), 0);
        pts [3] = db::Point (-std::min (delta_a, 0), 0);
      }

      if (m & 0x4) {

        read_repetition ();

        std::pair<bool, db::PropertiesRepository::properties_id_type> pp = read_element_properties (layout.properties_repository ());

        if (ll.first) {

          //  convert the OASIS record into the polygon.
          db::SimplePolygon poly;
          poly.assign_hull (pts, pts + 4);
          db::SimplePolygonRef poly_ref (poly, layout.shape_repository ());

          db::Cell &cell = layout.cell (cell_index);
          RepetitionIterator p = mm_repetition.get ().begin ();
          while (! p.at_end ()) {
            if (pp.first) {
              cell.shapes (ll.second).insert (db::SimplePolygonRefWithProperties (poly_ref.transformed (db::Disp (pos + *p)), pp.second));
            } else {
              cell.shapes (ll.second).insert (poly_ref.transformed (db::Disp (pos + *p)));
            }
            ++p;
          }

        }

      } else {

        std::pair<bool, db::PropertiesRepository::properties_id_type> pp = read_element_properties (layout.properties_repository ());

        if (ll.first) {

          //  convert the OASIS record into the polygon.
          db::SimplePolygon poly;
          poly.assign_hull (pts, pts + 4);
          db::SimplePolygonRef poly_ref (poly, layout.shape_repository ());

          if (pp.first) {
            layout.cell (cell_index).shapes (ll.second).insert (SimplePolygonRefWithProperties (poly_ref.transformed (db::Disp (pos)), pp.second));
          } else {
            layout.cell (cell_index).shapes (ll.second).insert (poly_ref.transformed (db::Disp (pos)));
          }

        }

      }

    } else if (r == 26 /*CTRAPEZOID*/) {

      unsigned char m = get_byte ();

      if (m & 0x1) {
        mm_layer = get_uint ();
      }

      if (m & 0x2) {
        mm_datatype = get_uint ();
      }

      if (m & 0x80) {
        mm_ctrapezoid_type = get_uint ();
      }

      if (m & 0x40) {
        mm_geometry_w = get_ucoord ();
      }

      if (m & 0x20) {
        mm_geometry_h = get_ucoord ();
      }

      if (m & 0x10) {
        db::Coord x;
        get (x);
        if (xy_absolute) {
          mm_geometry_x = x;
        } else {
          mm_geometry_x = x + mm_geometry_x.get ();
        }
      }

      if (m & 0x8) {
        db::Coord y;
        get (y);
        if (xy_absolute) {
          mm_geometry_y = y;
        } else {
          mm_geometry_y = y + mm_geometry_y.get ();
        }
      }

      db::Point pos (mm_geometry_x.get (), mm_geometry_y.get ());

      std::pair<bool, unsigned int> ll = open_dl (layout, LDPair (mm_layer.get (), mm_datatype.get ()), create_layers);

      db::Point pts [4];

      static db::Coord ctraps_table[][4][4] = {
        //  type 0
        {
          { 0, 0, 0, 0 },  // x=0*w+0*h, y=0*w+0*h ...
          { 0, 0, 0, 1 },
          { 1, -1, 0, 1 },
          { 1, 0, 0, 0 }
        },
        //  type 1
        {
          { 0, 0, 0, 0 },  
          { 0, 0, 0, 1 },
          { 1, 0, 0, 1 },
          { 1, -1, 0, 0 }
        },
        //  type 2
        {
          { 0, 0, 0, 0 },  
          { 0, 1, 0, 1 },
          { 1, 0, 0, 1 },
          { 1, 0, 0, 0 }
        },
        //  type 3
        {
          { 0, 1, 0, 0 },  
          { 0, 0, 0, 1 },
          { 1, 0, 0, 1 },
          { 1, 0, 0, 0 }
        },
        //  type 4
        {
          { 0, 0, 0, 0 },
          { 0, 1, 0, 1 },
          { 1, -1, 0, 1 },
          { 1, 0, 0, 0 }
        },
        //  type 5
        {
          { 0, 1, 0, 0 },  
          { 0, 0, 0, 1 },
          { 1, 0, 0, 1 },
          { 1, -1, 0, 0 }
        },
        //  type 6
        {
          { 0, 0, 0, 0 },  
          { 0, 1, 0, 1 },
          { 1, 0, 0, 1 },
          { 1, -1, 0, 0 }
        },
        //  type 7
        {
          { 0, 1, 0, 0 },  
          { 0, 0, 0, 1 },
          { 1, -1, 0, 1 },
          { 1, 0, 0, 0 }
        },
        //  type 8
        {
          { 0, 0, 0, 0 },  
          { 0, 0, 0, 1 },
          { 1, 0, -1, 1 },
          { 1, 0, 0, 0 }
        },
        //  type 9
        {
          { 0, 0, 0, 0 },  
          { 0, 0, -1, 1 },
          { 1, 0, 0, 1 },
          { 1, 0, 0, 0 }
        },
        //  type 10
        {
          { 0, 0, 0, 0 },  
          { 0, 0, 0, 1 },
          { 1, 0, 0, 1 },
          { 1, 0, 1, 0 }
        },
        //  type 11
        {
          { 0, 0, 1, 0 },  
          { 0, 0, 0, 1 },
          { 1, 0, 0, 1 },
          { 1, 0, 0, 0 }
        },
        //  type 12
        {
          { 0, 0, 0, 0 },  
          { 0, 0, 0, 1 },
          { 1, 0, -1, 1 },
          { 1, 0, 1, 0 }
        },
        //  type 13
        {
          { 0, 0, 1, 0 },  
          { 0, 0, -1, 1 },
          { 1, 0, 0, 1 },
          { 1, 0, 0, 0 }
        },
        //  type 14
        {
          { 0, 0, 0, 0 },  
          { 0, 0, -1, 1 },
          { 1, 0, 0, 1 },
          { 1, 0, 1, 0 }
        },
        //  type 15
        {
          { 0, 0, 1, 0 },  
          { 0, 0, 0, 1 },
          { 1, 0, -1, 1 },
          { 1, 0, 0, 0 }
        },
        //  type 16
        {
          { 0, 0, 0, 0 },  
          { 0, 0, 1, 0 },
          { 1, 0, 0, 0 },
          { 0, 0, 0, 0 }
        },
        //  type 17
        {
          { 0, 0, 0, 0 },  
          { 0, 0, 1, 0 },
          { 1, 0, 1, 0 },
          { 0, 0, 0, 0 }
        },
        //  type 18
        {
          { 0, 0, 0, 0 },  
          { 1, 0, 1, 0 },
          { 1, 0, 0, 0 },
          { 0, 0, 0, 0 }
        },
        //  type 19
        {
          { 0, 0, 1, 0 },  
          { 1, 0, 1, 0 },
          { 1, 0, 0, 0 },
          { 0, 0, 1, 0 }
        },
        //  type 20
        {
          { 0, 0, 0, 0 },  
          { 0, 1, 0, 1 },
          { 0, 2, 0, 0 },
          { 0, 0, 0, 0 }
        },
        //  type 21
        {
          { 0, 0, 0, 1 },  
          { 0, 2, 0, 1 },
          { 0, 1, 0, 0 },
          { 0, 0, 0, 1 }
        },
        //  type 22
        {
          { 0, 0, 0, 0 },  
          { 0, 0, 2, 0 },
          { 1, 0, 1, 0 },
          { 0, 0, 0, 0 }
        },
        //  type 23
        {
          { 1, 0, 0, 0 },  
          { 0, 0, 1, 0 },
          { 1, 0, 2, 0 },
          { 1, 0, 0, 0 }
        },
        //  type 24
        {
          { 0, 0, 0, 0 },  
          { 0, 0, 0, 1 },
          { 1, 0, 0, 1 },
          { 1, 0, 0, 0 }
        },
        //  type 25
        {
          { 0, 0, 0, 0 },  
          { 0, 0, 1, 0 },
          { 1, 0, 1, 0 },
          { 1, 0, 0, 0 }
        },
      };

      if (mm_ctrapezoid_type.get () >= sizeof (ctraps_table) / sizeof (ctraps_table [0])) {
        error (tl::sprintf (tl::translate ("Invalid CTRAPEZOID type %d").c_str (), int (mm_ctrapezoid_type.get ())));
      }

      db::Coord w = 0, h = 0;

      for (unsigned i = 0; i < 4; ++i) {

        db::Coord *m = ctraps_table [mm_ctrapezoid_type.get ()][i];

        db::Coord x = 0;
        if (m[0] != 0) x += m[0] * mm_geometry_w.get ();
        if (m[1] != 0) x += m[1] * mm_geometry_h.get ();

        db::Coord y = 0;
        if (m[2] != 0) y += m[2] * mm_geometry_w.get ();
        if (m[3] != 0) y += m[3] * mm_geometry_h.get ();

        pts [i] = db::Point (x, y);
        
        if (x > w) w = x;
        if (y > h) h = y;

      }

      //  set modal variables to the bbox of the shape
      mm_geometry_w = w;
      mm_geometry_h = h;

      unsigned int npts = 4;
      if (pts [npts - 1] == pts [0]) {
        --npts;
      }

      if (m & 0x4) {

        read_repetition ();

        std::pair<bool, db::PropertiesRepository::properties_id_type> pp = read_element_properties (layout.properties_repository ());

        if (ll.first) {

          //  convert the OASIS record into the polygon.
          db::SimplePolygon poly;
          poly.assign_hull (pts, pts + npts);
          db::SimplePolygonRef poly_ref (poly, layout.shape_repository ());

          db::Cell &cell = layout.cell (cell_index);
          RepetitionIterator p = mm_repetition.get ().begin ();
          while (! p.at_end ()) {
            if (pp.first) {
              cell.shapes (ll.second).insert (db::SimplePolygonRefWithProperties (poly_ref.transformed (db::Disp (pos + *p)), pp.second));
            } else {
              cell.shapes (ll.second).insert (poly_ref.transformed (db::Disp (pos + *p)));
            }
            ++p;
          }

        }

      } else {

        std::pair<bool, db::PropertiesRepository::properties_id_type> pp = read_element_properties (layout.properties_repository ());

        if (ll.first) {

          //  convert the OASIS record into the polygon.
          db::SimplePolygon poly;
          poly.assign_hull (pts, pts + npts);
          db::SimplePolygonRef poly_ref (poly, layout.shape_repository ());

          if (pp.first) {
            layout.cell (cell_index).shapes (ll.second).insert (db::SimplePolygonRefWithProperties (poly_ref.transformed (db::Disp (pos)), pp.second));
          } else {
            layout.cell (cell_index).shapes (ll.second).insert (poly_ref.transformed (db::Disp (pos)));
          }

        }

      }

    } else if (r == 27 /*CIRCLE*/) {

      unsigned char m = get_byte ();

      if (m & 0x1) {
        mm_layer = get_uint ();
      }

      if (m & 0x2) {
        mm_datatype = get_uint ();
      }

      if (m & 0x20) {
        mm_circle_radius = get_ucoord ();
      }

      if (m & 0x10) {
        db::Coord x;
        get (x);
        if (xy_absolute) {
          mm_geometry_x = x;
        } else {
          mm_geometry_x = x + mm_geometry_x.get ();
        }
      }

      if (m & 0x8) {
        db::Coord y;
        get (y);
        if (xy_absolute) {
          mm_geometry_y = y;
        } else {
          mm_geometry_y = y + mm_geometry_y.get ();
        }
      }

      db::Point pos (mm_geometry_x.get (), mm_geometry_y.get ());

      std::pair<bool, unsigned int> ll = open_dl (layout, LDPair (mm_layer.get (), mm_datatype.get ()), create_layers);

      //  ignore this circle if the radius is zero
      if (mm_circle_radius.get () <= 0) {
        ll.first = false;
      }

      db::Point pts [16];
      size_t npoints = sizeof (pts) / sizeof (pts[0]);
      size_t step = 1;

      static double ctab [][2] = {
        { 1.000000,0.000000 },
        { 0.923880,0.382683 },
        { 0.707107,0.707107 },
        { 0.382683,0.923880 },
        { 0.000000,1.000000 },
        { -0.382683,0.923880 },
        { -0.707107,0.707107 },
        { -0.923880,0.382683 },
        { -1.000000,0.000000 },
        { -0.923880,-0.382683 },
        { -0.707107,-0.707107 },
        { -0.382683,-0.923880 },
        { -0.000000,-1.000000 },
        { 0.382683,-0.923880 },
        { 0.707107,-0.707107 },
        { 0.923880,-0.382683 }
      };

      if (mm_circle_radius.get () <= 2) {
        step = npoints / 4;
        npoints = 4;
      } else if (mm_circle_radius.get () <= 30) {
        step = npoints / 8;
        npoints = 8;
      } 

      for (size_t i = 0; i < npoints && ll.first;  i++) {
        pts [i] = db::Point::from_double (db::DPoint (ctab [i * step][0] * mm_circle_radius.get (), 
                                                      ctab [i * step][1] * mm_circle_radius.get ()));
      }

      if (m & 0x4) {

        read_repetition ();

        std::pair<bool, db::PropertiesRepository::properties_id_type> pp = read_element_properties (layout.properties_repository ());

        if (ll.first) {

          //  convert the OASIS record into the polygon.
          db::SimplePolygon poly;
          poly.assign_hull (pts, pts + npoints);
          db::SimplePolygonRef poly_ref (poly, layout.shape_repository ());

          db::Cell &cell = layout.cell (cell_index);
          RepetitionIterator p = mm_repetition.get ().begin ();
          while (! p.at_end ()) {
            if (pp.first) {
              cell.shapes (ll.second).insert (db::SimplePolygonRefWithProperties (poly_ref.transformed (db::Disp (pos + *p)), pp.second));
            } else {
              cell.shapes (ll.second).insert (poly_ref.transformed (db::Disp (pos + *p)));
            }
            ++p;
          }

        }

      } else {

        std::pair<bool, db::PropertiesRepository::properties_id_type> pp = read_element_properties (layout.properties_repository ());

        if (ll.first) {

          //  convert the OASIS record into the polygon.
          db::SimplePolygon poly;
          poly.assign_hull (pts, pts + npoints);
          db::SimplePolygonRef poly_ref (poly, layout.shape_repository ());

          if (pp.first) {
            layout.cell (cell_index).shapes (ll.second).insert (SimplePolygonRefWithProperties (poly_ref.transformed (db::Disp (pos)), pp.second));
          } else {
            layout.cell (cell_index).shapes (ll.second).insert (poly_ref.transformed (db::Disp (pos)));
          }

        }

      }


    } else if (r == 28 || r == 29 /*PROPERTY*/) {

      //  unrecognized property: parse over
      if (r == 28) {
        read_properties ();
      }

    } else if (r == 32 /*XELEMENT*/) {

      //  read over
      get_ulong ();
      get_str ();

    } else if (r == 33 /*XGEOMETRY*/) {

      //  read over.

      unsigned char m = get_byte ();

      if (m & 0x1) {
        mm_layer = get_uint ();
      }

      if (m & 0x2) {
        mm_datatype = get_uint ();
      }

      //  data payload:
      get_str ();

      if (m & 0x10) {
        db::Coord x;
        get (x);
        if (xy_absolute) {
          mm_geometry_x = x;
        } else {
          mm_geometry_x = x + mm_geometry_x.get ();
        }
      }

      if (m & 0x8) {
        db::Coord y;
        get (y);
        if (xy_absolute) {
          mm_geometry_y = y;
        } else {
          mm_geometry_y = y + mm_geometry_y.get ();
        }
      }

      if (m & 0x4) {
        read_repetition ();
      }

    } else if (r == 34 /*CBLOCK*/) {

      unsigned int type = get_uint ();
      if (type != 0) {
        error (tl::sprintf (tl::translate ("Invalid CBLOCK compression type %d").c_str (), type));
      }

      get_uint ();  // uncomp-byte-count - not needed
      get_uint ();  // comp-byte-count - not needed

      //  put the stream into deflating mode
      m_stream.deflate ();

    } else {
      //  put the byte back into the stream
      m_stream.unget (1);
      return;
    }

  }

  m_cellname = "";

}

}

