
/*

  KLayout Layout Viewer
  Copyright (C) 2006-2007 Matthias Koefferlein

  This program is free software; you can redistribute it and/or modify
  it under the terms of the GNU General Public License as published by
  the Free Software Foundation; either version 2 of the License, or
  (at your option) any later version.

  This program is distributed in the hope that it will be useful,
  but WITHOUT ANY WARRANTY; without even the implied warranty of
  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
  GNU General Public License for more details.

  You should have received a copy of the GNU General Public License
  along with this program; if not, write to the Free Software
  Foundation, Inc., 51 Franklin St, Fifth Floor, Boston, MA  02110-1301  USA

*/


#include "dbOASISWriter.h"
#include "tlStream.h"

namespace db
{

OASISWriter::OASISWriter ()
  : mp_stream (0)
{
  // .. nothing yet ..
}

void 
OASISWriter::write_byte (unsigned char b)
{
  mp_stream->put ((const char *) &b, 1);
}

void 
OASISWriter::write (long n)
{
  if (n < 0) {
    write (((unsigned long) (-n) << 1) | 1);
  } else {
    write ((unsigned long) n << 1);
  }
}

void 
OASISWriter::write (unsigned long n)
{
  do {
    unsigned char b = n & 0x7f;
    n >>= 7;
    if (n > 0) {
      b |= 0x80;
    }
    mp_stream->put ((const char *) &b, 1);
  } while (n > 0);
}

void
OASISWriter::write (float d) 
{
  if (fabs (d) >= 0.5 && fabs (floor (d + 0.5) - d) < 1e-6 && fabs (d) < double (std::numeric_limits<long>::max ())) {

    //  whole number (negative or positive)
    if (d < 0.0) {
      write_byte (1);
      write ((unsigned long) floor (-d + 0.5));
    } else { 
      write_byte (0);
      write ((unsigned long) floor (d + 0.5));
    }

  } else {

    write_byte (6);

    //  4-Byte IEEE real
    union {
      float d;
      uint32_t i;
    } f2i;

    f2i.d = d;
    uint32_t i = f2i.i;
    char b[sizeof (f2i.i)];
    for (unsigned int n = sizeof (f2i.i); n > 0; --n) {
      b[n - 1] = i & 0xff; 
      i >>= 8;
    }
    mp_stream->put (b, sizeof (f2i.i));

  }
}

void
OASISWriter::write (double d) 
{
  if (fabs (d) >= 0.5 && fabs (floor (d + 0.5) - d) < 1e-10 && fabs (d) < double (std::numeric_limits<long>::max ())) {

    //  whole number (negative or positive)
    if (d < 0.0) {
      write_byte (1);
      write ((unsigned long) floor (-d + 0.5));
    } else { 
      write_byte (0);
      write ((unsigned long) floor (d + 0.5));
    }

  } else {

    write_byte (7);

    //  8-Byte IEEE real
    union {
      double d;
      uint64_t i;
    } f2i;

    f2i.d = d;
    uint64_t i = f2i.i;
    char b[sizeof (f2i.i)];
    for (unsigned int n = 0; n < sizeof (f2i.i); n++) {
      b[n] = i & 0xff; 
      i >>= 8;
    }
    mp_stream->put (b, sizeof (f2i.i));

  }
}

void 
OASISWriter::write_bstring (const char *s)
{
  unsigned long l = strlen (s);
  write (l);
  mp_stream->put (s, l);
}

void  
OASISWriter::write_astring (const char *s)
{
  unsigned long l = strlen (s);

  write (l);

  bool valid = true;
  for (const char *c = s; *c && valid; ++c) {
    if (*c < 0x20 || *c > 0x7e) {
      valid = false;
    }
  }

  if (valid) {
    mp_stream->put (s, l);
  } else {
    for (const char *c = s; *c; ++c) {
      if (*c < 0x20 || *c > 0x7e) {
        write_byte ('*');
      } else {
        mp_stream->put (c, 1);
      }
    }
  }
}

void  
OASISWriter::write_nstring (const char *s)
{
  unsigned long l = strlen (s);

  if (!l) {
    write_astring ("*");
  } else {

    write (l);

    bool valid = true;
    for (const char *c = s; *c && valid; ++c) {
      if (*c < 0x21 || *c > 0x7e) {
        valid = false;
      }
    }

    if (valid) {
      mp_stream->put (s, l);
    } else {
      for (const char *c = s; *c; ++c) {
        if (*c < 0x21 || *c > 0x7e) {
          write_byte ('*');
        } else {
          mp_stream->put (c, 1);
        }
      }
    }

  }
}

template <class R>
inline R safe_scale (double sf, R value)
{
  double i = floor (sf * value + 0.5);
  if (i < double (std::numeric_limits<R>::min ())) {
    throw tl::Exception ("Scaling failed: coordinate underflow");
  }
  if (i > double (std::numeric_limits<R>::max ())) {
    throw tl::Exception ("Scaling failed: coordinate overflow");
  }
  return R (i);
}

void
OASISWriter::write_gdelta (const db::Point &p)
{
  db::Coord x = p.x ();
  db::Coord y = p.y ();

  if (m_sf != 1.0) {
    x = safe_scale (m_sf, x);
    y = safe_scale (m_sf, y);
  }

  if (x == -y || x == y || x == 0 || y == 0) {

    unsigned long dir = 0;
    unsigned long l = 0;

    if (x > 0) {
      l = x;
      if (y == 0) {
        dir = 0;
      } else if (y < 0) {
        dir = 7;
      } else {
        dir = 4;
      }
    } else if (x == 0) {
      if (y < 0) {
        l = -y;
        dir = 3;
      } else {
        l = y;
        dir = 1;
      }
    } else if (x < 0) {
      l = -x;
      if (y == 0) {
        dir = 2;
      } else if (y < 0) {
        dir = 6;
      } else {
        dir = 5;
      }
    }

    write ((l << 4) | (dir << 1));

  } else {

    unsigned long d;
    if (x < 0) {
      d = ((unsigned long) -x << 2) | 3;
    } else {
      d = ((unsigned long) x << 2) | 1;
    }
    write (d);
    write (y);

  }
}

void
OASISWriter::write_coord (db::Coord c)
{
  if (m_sf == 1.0) {
    return write (c);
  } else {
    return write (safe_scale (m_sf, c));
  }
}

void
OASISWriter::write_ucoord (db::Coord c)
{
  if (m_sf == 1.0) {
    return write ((unsigned long) c);
  } else {
    return write (safe_scale (m_sf, (unsigned long) c));
  }
}

static bool 
is_valid_prop (const db::LayerProperties &prop, void *)
{
  return prop.layer >= 0 && prop.datatype >= 0;
}

void
OASISWriter::emit_property_def (const db::Layout &layout, db::PropertiesRepository::properties_id_type prop_id, unsigned int &propname_id, unsigned int &propstring_id)
{
  const db::PropertiesRepository::properties_set &props = layout.properties_repository ().properties (prop_id);
  for (db::PropertiesRepository::properties_set::const_iterator p = props.begin (); p != props.end (); ++p) {

    const tl::Variant &name = layout.properties_repository ().prop_name (p->first);
    const char *name_str = "S_GDS_PROPERTY";
    if (! name.is_long () || (name.to_long () >= 0x8000 && name.to_long () < 0)) {
      name_str = name.to_string ();
    }
    if (m_propnames.insert (std::make_pair (name_str, propname_id)).second) {
      write_byte (7);
      write_nstring (name_str);
      ++propname_id;
    }

    if (p->second.is_string ()) {
      if (m_propstrings.insert (std::make_pair (p->second.to_string (), propstring_id)).second) {
        write_byte (9);
        write_bstring (p->second.to_string ());
        ++propstring_id;
      }
    }

  }
}

void 
OASISWriter::write (const db::Layout &layout, tl::OutputStream &stream, const db::SaveLayoutOptions &options) throw (tl::Exception)
{
  mp_stream = &stream;

  double dbu = (options.dbu () == 0.0) ? layout.dbu () : options.dbu ();
  m_sf = options.scale_factor () * (layout.dbu () / dbu);
  if (fabs (m_sf - 1.0) < 1e-9) {
    //  to avoid rounding problems, set to 1.0 exactly if possible.
    m_sf = 1.0;
  }

  std::vector <std::pair <unsigned int, db::LayerProperties> > layers;
  options.get_valid_layers (layout, layers, &is_valid_prop);

  std::vector <db::CellGraph::cell_index_type> cells;
  options.get_cells (layout, cells);

  //  write header

  char magic[] = "%SEMI-OASIS\015\012";
  mp_stream->put (magic, sizeof (magic) - 1);

  //  START record
  write_byte (1); 
  write_bstring ("1.0");
  write (1.0 / dbu);
  write_byte (0);  //  offset-flag
  //  offset table:
  for (unsigned int i = 0; i < 12; ++i) {
    write_byte (0);
  }

  //  Prepare name tables
  unsigned int id;
 
  m_textstrings.clear ();
  m_propnames.clear ();
  m_propstrings.clear ();

  for (std::vector<db::CellGraph::cell_index_type>::const_iterator cell = cells.begin (); cell != cells.end (); ++cell) {
    //  CELLNAME (explicit)
    write_byte (4);
    write_nstring (layout.cell_name (*cell));
    write ((unsigned long) *cell);
  }

  //  build text string table

  id = 0;
  for (std::vector<db::CellGraph::cell_index_type>::const_iterator cell = cells.begin (); cell != cells.end (); ++cell) {

    const db::Cell &cref (layout.cell (*cell));
    for (std::vector <std::pair <unsigned int, db::LayerProperties> >::const_iterator l = layers.begin (); l != layers.end (); ++l) {
      db::ShapeIterator shape (cref.shapes (l->first).begin (db::ShapeIterator::Texts));
      while (! shape.at_end ()) {
        if (m_textstrings.insert (std::make_pair (shape->text_string (), id)).second) {
          write_byte (5);
          write_astring (shape->text_string ().c_str ());
          ++id;
        }
        ++shape;
      }
    }

  }

  //  build property name and value string tables

  unsigned int propname_id = 0;
  unsigned int propstring_id = 0;
  std::set <unsigned long> prop_ids_done;

  for (std::vector<db::CellGraph::cell_index_type>::const_iterator cell = cells.begin (); cell != cells.end (); ++cell) {

    const db::Cell &cref (layout.cell (*cell));

    for (db::Cell::const_iterator inst = cref.begin (); ! inst.at_end (); ++inst) {
      if (inst->has_prop_id () && inst->prop_id () != 0 && prop_ids_done.find (inst->prop_id ()) == prop_ids_done.end ()) {
        prop_ids_done.insert (inst->prop_id ());
        emit_property_def (layout, inst->prop_id (), propname_id, propstring_id);
      }
    }

    for (std::vector <std::pair <unsigned int, db::LayerProperties> >::const_iterator l = layers.begin (); l != layers.end (); ++l) {
      db::ShapeIterator shape (cref.shapes (l->first).begin (db::ShapeIterator::Boxes | db::ShapeIterator::Polygons | db::ShapeIterator::Edges | db::ShapeIterator::Paths | db::ShapeIterator::Texts));
      while (! shape.at_end ()) {
        if (shape->has_prop_id () && shape->prop_id () != 0 && prop_ids_done.find (shape->prop_id ()) == prop_ids_done.end ()) {
          prop_ids_done.insert (shape->prop_id ());
          emit_property_def (layout, shape->prop_id (), propname_id, propstring_id);
        }
        ++shape;
      }
    }

  }

  //  write layernames table
  
  for (std::vector <std::pair <unsigned int, db::LayerProperties> >::const_iterator l = layers.begin (); l != layers.end (); ++l) {

    if (! l->second.name.empty ()) {

      //  write mappings to text layer and shape layers
      write_byte (11);
      write_nstring (l->second.name.c_str ());
      write_byte (3);
      write ((unsigned long) l->second.layer);
      write_byte (3);
      write ((unsigned long) l->second.datatype);

      write_byte (12);
      write_nstring (l->second.name.c_str ());
      write_byte (3);
      write ((unsigned long) l->second.layer);
      write_byte (3);
      write ((unsigned long) l->second.datatype);

    }

  }

  //  body

  for (std::vector<db::CellGraph::cell_index_type>::const_iterator cell = cells.begin (); cell != cells.end (); ++cell) {

    //  reset modal variables
    mm_repetition.reset ();
    mm_placement_cell.reset ();
    mm_placement_x = 0;
    mm_placement_y = 0;
    mm_layer.reset ();
    mm_datatype.reset ();
    mm_textlayer.reset ();
    mm_texttype.reset ();
    mm_text_x = 0;
    mm_text_y = 0;
    mm_text_string.reset ();
    mm_geometry_x = 0;
    mm_geometry_y = 0;
    mm_geometry_w.reset ();
    mm_geometry_h.reset ();
    mm_polygon_point_list.reset ();
    mm_path_halfwidth.reset ();
    mm_path_start_extension.reset ();
    mm_path_end_extension.reset ();
    mm_path_point_list.reset ();
    mm_ctrapezoid_type.reset ();
    mm_circle_radius.reset ();
    mm_last_property_name.reset ();
    mm_last_property_is_sprop.reset ();
    mm_last_value_list.reset ();

    //  cell header 

    write_byte (13);  // CELL
    write ((unsigned long) *cell);

    //  cell body 
    const db::Cell &cref (layout.cell (*cell));

    //  instances
    write_insts (layout, cref);

    //  shapes
    for (std::vector <std::pair <unsigned int, db::LayerProperties> >::const_iterator l = layers.begin (); l != layers.end (); ++l) {
      write_texts (layout, cref, l->second, l->first);
      write_polygons (layout, cref, l->second, l->first);
      write_edges (layout, cref, l->second, l->first);
      write_paths (layout, cref, l->second, l->first);
      write_boxes (layout, cref, l->second, l->first);
    }

    //  end of cell

  }

  //  END record

  write_byte (2);
  unsigned int pad = 256 - 5;
  write (pad);
  for (unsigned int i = 0; i < pad; ++i) {
    write_byte (0);
  }
  write_byte (0);

}

void 
OASISWriter::write (const Repetition &rep)
{
  db::Point a, b;
  size_t amax, bmax;

  bool is_reg = rep.is_regular (a, b, amax, bmax);

  //  currently there are only regular repetitions
  //  TODO: optimize for orthogonal cases 
  tl_assert (is_reg);
  tl_assert (amax >= 2 || bmax >= 2);

  if (amax == 1 || bmax == 1) {

    if (bmax == 1) {
      b = a;
      bmax = amax;
    }

    if (b.x () == 0 && b.y () >= 0) {
      write_byte (3);
      write (bmax - 2);
      write_ucoord (b.y ());
    } else if (b.y () == 0 && b.x () >= 0) {
      write_byte (2);
      write (bmax - 2);
      write_ucoord (b.x ());
    } else {
      write_byte (9);
      write (bmax - 2);
      write_gdelta (b);
    }

  } else if (b.x () == 0 && b.y () >= 0 && a.y () == 0 && a.x () >= 0) {

    write_byte (1);
    write (amax - 2);
    write (bmax - 2);
    write_ucoord (a.x ());
    write_ucoord (b.y ());

  } else if (b.y () == 0 && b.x () >= 0 && a.x () == 0 && a.y () >= 0) {

    write_byte (1);
    write (bmax - 2);
    write (amax - 2);
    write_ucoord (b.x ());
    write_ucoord (a.y ());

  } else {

    write_byte (8);
    write (amax - 2);
    write (bmax - 2);
    write_gdelta (a);
    write_gdelta (b);

  }
}

void 
OASISWriter::write_insts (const db::Layout &layout, const db::Cell &cell)
{
  for (db::Cell::const_iterator inst = cell.begin (); ! inst.at_end (); ++inst) {

    Repetition rep;
    db::Point a, b;
    unsigned long amax, bmax;

    db::Trans t = inst->front ();

    bool is_reg = inst->is_regular_array (a, b, amax, bmax);
    if (is_reg && (amax > 1 || bmax > 1)) {
      rep = Repetition (new RegularRepetition (a, b, amax, bmax));
    }

    unsigned char info = 0x40;  // by reference number
    if (mm_placement_cell != inst->cell_index ()) {
      info |= 0x80;
    }
    if (mm_placement_x != t.disp ().x ()) {
      info |= 0x20;
    }
    if (mm_placement_y != t.disp ().y ()) {
      info |= 0x10;
    }
    if (rep != Repetition ()) {
      info |= 0x08;
    }

    if (t.is_mirror ()) {
      info |= 0x01;
    }

    if (inst->is_complex ()) {
      write_byte (18);
      write_byte (info | 0x06);
    } else {
      write_byte (17);
      write_byte (info | ((t.rot () & 0x03) << 1));
    }

    if (info & 0x80) {
      mm_placement_cell = inst->cell_index ();
      write ((unsigned long) mm_placement_cell.get ());
    }

    if (inst->is_complex ()) {
      write (inst->complex_trans ().mag ());
      write (inst->complex_trans ().angle ());
    }
    
    if (info & 0x20) {
      mm_placement_x = t.disp ().x ();
      write_coord (mm_placement_x.get ());
    }
    if (info & 0x10) {
      mm_placement_y = t.disp ().y ();
      write_coord (mm_placement_y.get ());
    }

    if (info & 0x08) {
      if (mm_repetition == rep) {
        write_byte (0);  // reuse
      } else {
        mm_repetition = rep;
        write (mm_repetition.get ());
      }
    }

    if (inst->has_prop_id () && inst->prop_id () != 0) {
      write_props (layout, layout.properties_repository ().properties (inst->prop_id ()));
    }

  }
}

void
OASISWriter::write_props (const db::Layout &layout, const db::PropertiesRepository::properties_set &props)
{
  std::vector<tl::Variant> pv_list; 

  for (db::PropertiesRepository::properties_set::const_iterator p = props.begin (); p != props.end (); ++p) {

    const tl::Variant &name = layout.properties_repository ().prop_name (p->first);

    const char *name_str = "S_GDS_PROPERTY";
    bool sflag = true;

    pv_list.clear ();
    const std::vector<tl::Variant> *pvl = &pv_list;

    if (! name.is_long () || (name.to_long () >= 0x8000 && name.to_long () < 0)) {

      name_str = name.to_string ();
      sflag = false;

      if (p->second.is_list ()) {
        pvl = &p->second.get_list ();
      } else if (!p->second.is_nil ()) {
        pv_list.reserve (1);
        pv_list.push_back (p->second);
      }

    } else {

      pv_list.reserve (2);
      pv_list.push_back (name.to_long ());
      pv_list.push_back (p->second.to_string ());

    }

    bool same_name = (mm_last_property_name == name_str);
    bool same_value = (mm_last_value_list == *pvl);
    bool same_sflag = (mm_last_property_is_sprop == sflag);

    if (same_name && same_value && same_sflag) {
      write_byte (29); // repeat property 
    } else {

      write_byte (28);

      unsigned char info = sflag ? 1 : 0;
      if (same_value) {
        info |= 0x08;
      } else {
        if (pvl->size () >= 15) {
          info |= 0xf0;
        } else {
          info |= ((unsigned char)pvl->size ()) << 4;
        }
      }

      if (! same_name) {
        info |= 0x06;
      }

      write_byte (info);

      if (! same_name) {
        std::map <std::string, unsigned long>::const_iterator pni = m_propnames.find (name_str);
        tl_assert (pni != m_propnames.end ());
        write (pni->second);
        mm_last_property_name = name_str;
      }

      if (! same_value) {

        if (pvl->size () >= 15) {
          write ((unsigned long) pvl->size ());
        }

        //  write property values
        for (unsigned long i = 0; i < pvl->size (); ++i) {
          const tl::Variant &v = (*pvl)[i];
          if (v.is_double ()) {
            write (v.to_double ());
          } else if (v.is_long ()) {
            write_byte (9);
            write (v.to_long ());
          } else {
            const char *pvs = v.to_string ();
            std::map <std::string, unsigned long>::const_iterator pvi = m_propstrings.find (pvs);
            if (pvi != m_propstrings.end ()) {
              write_byte (14);
              write (pvi->second);
            } else {
              write_byte (11);
              write_bstring (pvs);
            }
          }
        }

        mm_last_value_list = *pvl;

      }

      mm_last_property_is_sprop = sflag;

    }

  }
}

void
OASISWriter::write_pointlist (const std::vector<db::Point> &pointlist, bool for_polygons)
{
  tl_assert ((for_polygons && pointlist.size () > 1) || (! for_polygons && pointlist.size () > 0));

  //  determine type: 0 (manhattan, horizontal first), 1 (manhattan, vert. first), -1 other
  db::Point plast (0, 0);
  int type = -1;
  int hvlast = -1;
  for (std::vector<db::Point>::const_iterator p = pointlist.begin (); p != pointlist.end (); ++p) {
    int hv = -1;
    if (p->x () == plast.x ()) {
      hv = 1;
    } else if (p->y () == plast.y ()) {
      hv = 0;
    } else {
      type = -1;
      break;
    }
    if (type == -1) {
      type = hv;
    } else if (hv != !hvlast) {
      type = -1;
      break;
    }
    hvlast = hv;
    plast = *p;
  }

  //  test last displacement for polygons
  if (for_polygons && type >= 0) {
    if (plast.x () == 0) {
      if (hvlast != 0) {
        type = -1;
      }
    } else if (plast.y () == 0) {
      if (hvlast != 1) {
        type = -1;
      }
    } else {
      type = -1;
    }
  }

  if (type == 0 || type == 1) {

    //  manhattan pointlist
    write_byte (type);
    size_t implicit = for_polygons ? 1 : 0;
    write ((unsigned long) (pointlist.size () - implicit));

    db::Point plast (0, 0);
    for (std::vector<db::Point>::const_iterator p = pointlist.begin (); p != pointlist.end () - implicit; ++p) {
      db::Coord d = p->x () - plast.x ();
      if (d == 0) {
        d = p->y () - plast.y ();
      }
      write_coord (d);
      plast = *p;
    }

  } else {

    //  generic pointlist
    write_byte (4);
    write ((unsigned long) pointlist.size ());
    db::Point plast (0, 0);
    for (std::vector<db::Point>::const_iterator p = pointlist.begin (); p != pointlist.end (); ++p) {
      write_gdelta (*p - plast);
      plast = *p;
    }

  }
}

void 
OASISWriter::write_texts (const db::Layout &layout, const db::Cell &cell, const db::LayerProperties &lprops, unsigned int lindex)
{
  int layer = lprops.layer;
  int datatype = lprops.datatype;

  db::ShapeIterator shape;
  shape = cell.shapes (lindex).begin (db::ShapeIterator::Texts);

  while (! shape.at_end ()) {

    db::Trans trans = shape->text_trans ();
    std::map <std::string, unsigned long>::const_iterator ts = m_textstrings.find (shape->text_string ());
    tl_assert (ts != m_textstrings.end ());
    unsigned long text_id = ts->second;

    unsigned char info = 0x20;

    if (mm_text_string != shape->text_string ()) {
      info |= 0x40;
    }
    if (mm_textlayer != layer) {
      info |= 0x01;
    }
    if (mm_texttype != datatype) {
      info |= 0x02;
    }
    if (mm_text_x != trans.disp ().x ()) {
      info |= 0x10;
    }
    if (mm_text_y != trans.disp ().y ()) {
      info |= 0x08;
    }

    write_byte (19);
    write_byte (info);
    if (info & 0x40) {
      mm_text_string = shape->text_string ();
      write ((unsigned long) text_id);
    }
    if (info & 0x01) {
      mm_textlayer = layer;
      write ((unsigned long) layer);
    }
    if (info & 0x02) {
      mm_texttype = datatype;
      write ((unsigned long) datatype);
    }
    if (info & 0x10) {
      mm_text_x = trans.disp ().x ();
      write_coord (mm_text_x.get ());
    }
    if (info & 0x08) {
      mm_text_y = trans.disp ().y ();
      write_coord (mm_text_y.get ());
    }

    if (shape->has_prop_id () && shape->prop_id () != 0) {
      write_props (layout, layout.properties_repository ().properties (shape->prop_id ()));
    }

    ++shape;

  }
}

void 
OASISWriter::write_polygons (const db::Layout &layout, const db::Cell &cell, const db::LayerProperties &lprops, unsigned int lindex)
{
  int layer = lprops.layer;
  int datatype = lprops.datatype;

  db::ShapeIterator shape;
  shape = cell.shapes (lindex).begin (db::ShapeIterator::Polygons);

  std::vector<db::Point> pointlist;

  while (! shape.at_end ()) {

    //  @@@ how to deal with holes, max vertex count?
    if (shape->holes () > 0) {
      throw tl::Exception (tl::translate ("Cannot write polygons with holes to OASIS streams"));
    }
   
    size_t n = 0;
    for (db::Shape::point_iterator e = shape->begin_hull (); e != shape->end_hull (); ++e) {
      ++n;
    }
    pointlist.erase (pointlist.begin (), pointlist.end ());
    pointlist.reserve (n - 1);

    db::Shape::point_iterator e = shape->begin_hull ();
    db::Point start = *e;
    ++e;
    for ( ; e != shape->end_hull (); ++e) {
      pointlist.push_back (*e - start);
    }

    if (pointlist.size () < 2) {
      throw tl::Exception (tl::translate ("Polygons with less than three points cannot be written to OASIS files"));
    }

    unsigned char info = 0x00;

    if (mm_layer != layer) {
      info |= 0x01;
    }
    if (mm_datatype != datatype) {
      info |= 0x02;
    }
    if (mm_geometry_x != start.x ()) {
      info |= 0x10;
    }
    if (mm_geometry_y != start.y ()) {
      info |= 0x08;
    }
    if (mm_polygon_point_list != pointlist) {
      info |= 0x20;
    }

    write_byte (21);
    write_byte (info);

    if (info & 0x01) {
      mm_layer = layer;
      write ((unsigned long) layer);
    }
    if (info & 0x02) {
      mm_datatype = datatype;
      write ((unsigned long) datatype);
    }
    if (info & 0x20) {
      mm_polygon_point_list = pointlist;
      write_pointlist (pointlist, true /*for polygons*/);
    }
    if (info & 0x10) {
      mm_geometry_x = start.x ();
      write_coord (start.x ());
    }
    if (info & 0x08) {
      mm_geometry_y = start.y ();
      write_coord (start.y ());
    }

    if (shape->has_prop_id () && shape->prop_id () != 0) {
      write_props (layout, layout.properties_repository ().properties (shape->prop_id ()));
    }

    ++shape;

  }
}

void 
OASISWriter::write_boxes (const db::Layout &layout, const db::Cell &cell, const db::LayerProperties &lprops, unsigned int lindex)
{
  int layer = lprops.layer;
  int datatype = lprops.datatype;

  db::ShapeIterator shape;
  shape = cell.shapes (lindex).begin (db::ShapeIterator::Boxes);

  while (! shape.at_end ()) {

    unsigned char info = 0x00;
    db::Box box (shape->box ());
    
    if (mm_layer != layer) {
      info |= 0x01;
    }
    if (mm_datatype != datatype) {
      info |= 0x02;
    }

    if (box.width () == box.height ()) {
      info |= 0x80;  // square
    } else {
      if (mm_geometry_h != box.height ()) {
        info |= 0x20;
      }
    }
    if (mm_geometry_w != box.width ()) {
      info |= 0x40;
    }

    if (mm_geometry_x != box.left ()) {
      info |= 0x10;
    }
    if (mm_geometry_y != box.bottom ()) {
      info |= 0x08;
    }
    
    write_byte (20);
    write_byte (info);

    if (info & 0x01) {
      mm_layer = layer;
      write ((unsigned long) layer);
    }
    if (info & 0x02) {
      mm_datatype = datatype;
      write ((unsigned long) datatype);
    }

    mm_geometry_w = box.width ();
    mm_geometry_h = box.height ();

    if (info & 0x40) {
      write_ucoord (mm_geometry_w.get ());
    }
    if (info & 0x20) {
      write_ucoord (mm_geometry_h.get ());
    }

    if (info & 0x10) {
      mm_geometry_x = box.left ();
      write_coord (mm_geometry_x.get ());
    }
    if (info & 0x08) {
      mm_geometry_y = box.bottom ();
      write_coord (mm_geometry_y.get ());
    }

    if (shape->has_prop_id () && shape->prop_id () != 0) {
      write_props (layout, layout.properties_repository ().properties (shape->prop_id ()));
    }

    ++shape;

  }
}

void 
OASISWriter::write_paths (const db::Layout &layout, const db::Cell &cell, const db::LayerProperties &lprops, unsigned int lindex)
{
  int layer = lprops.layer;
  int datatype = lprops.datatype;

  db::ShapeIterator shape;
  shape = cell.shapes (lindex).begin (db::ShapeIterator::Paths);

  std::vector<db::Point> pointlist;

  while (! shape.at_end ()) {

    std::pair<db::Path::coord_type, db::Path::coord_type> ext (0, 0);
    //  for round paths, circles are places to mimic the extensions
    if (! shape->round_path ()) {
      ext = shape->path_extensions ();
    }

    size_t n = 0;
    for (db::Shape::point_iterator e = shape->begin_point (); e != shape->end_point (); ++e) {
      ++n;
    }
    pointlist.erase (pointlist.begin (), pointlist.end ());
    pointlist.reserve (n - 1);

    db::Shape::point_iterator e = shape->begin_point ();
    db::Point start = *e;
    ++e;
    for ( ; e != shape->end_point (); ++e) {
      pointlist.push_back (*e - start);
    }

    if (pointlist.size () < 1) {
      throw tl::Exception (tl::translate ("Paths with less than two points cannot be written to OASIS files"));
    }

    db::Coord hw = shape->path_width () / 2;
    if (hw * 2 != shape->path_width ()) {
      throw tl::Exception (tl::translate ("Paths with odd width cannot be written to OASIS files"));
    }

    unsigned char info = 0x00;

    if (mm_layer != layer) {
      info |= 0x01;
    }
    if (mm_datatype != datatype) {
      info |= 0x02;
    }
    if (mm_geometry_x != start.x ()) {
      info |= 0x10;
    }
    if (mm_geometry_y != start.y ()) {
      info |= 0x08;
    }
    if (mm_path_point_list != pointlist) {
      info |= 0x20;
    }
    if (mm_path_start_extension != ext.first || mm_path_end_extension != ext.second) {
      info |= 0x80;
    }
    if (mm_path_halfwidth != hw) {
      info |= 0x40;
    }

    write_byte (22);
    write_byte (info);

    if (info & 0x01) {
      mm_layer = layer;
      write ((unsigned long) layer);
    }
    if (info & 0x02) {
      mm_datatype = datatype;
      write ((unsigned long) datatype);
    }
    if (info & 0x40) {
      mm_path_halfwidth = hw;
      write_ucoord (hw);
    }

    if (info & 0x80) {

      unsigned char ext_scheme = 0;
      if (mm_path_start_extension == ext.first) {
        //  00
      } else if (ext.first == 0) {
        ext_scheme |= 0x04;
      } else if (ext.first == hw) {
        ext_scheme |= 0x08;
      } else {
        ext_scheme |= 0x0c;
      }
      if (mm_path_end_extension == ext.second) {
        //  00
      } else if (ext.second == 0) {
        ext_scheme |= 0x01;
      } else if (ext.second == hw) {
        ext_scheme |= 0x02;
      } else {
        ext_scheme |= 0x03;
      }

      write_byte (ext_scheme);

      if ((ext_scheme & 0x0c) == 0x0c) {
        write_coord (ext.first);
      }
      if ((ext_scheme & 0x03) == 0x03) {
        write_coord (ext.second);
      }

      mm_path_start_extension = ext.first;
      mm_path_end_extension = ext.second;

    }

    if (info & 0x20) {
      mm_path_point_list = pointlist;
      write_pointlist (pointlist, false /*=for paths*/);
    }
    if (info & 0x10) {
      mm_geometry_x = start.x ();
      write_coord (start.x ());
    }
    if (info & 0x08) {
      mm_geometry_y = start.y ();
      write_coord (start.y ());
    }

    if (shape->has_prop_id () && shape->prop_id () != 0) {
      write_props (layout, layout.properties_repository ().properties (shape->prop_id ()));
    }

    if (shape->round_path ()) {

      //  write two circles at the path ends to mimic the round path ends.

      unsigned char info = 0;
      if (mm_circle_radius != hw) {
        info |= 0x20;
      }
      if (mm_geometry_x != start.x ()) {
        info |= 0x10;
      }
      if (mm_geometry_y != start.y ()) {
        info |= 0x08;
      }

      write_byte (27);
      write_byte (info);

      if (info & 0x20) {
        mm_circle_radius = hw;
        write_ucoord (hw);
      }
      if (info & 0x10) {
        mm_geometry_x = start.x ();
        write_coord (start.x ());
      }
      if (info & 0x08) {
        mm_geometry_y = start.y ();
        write_coord (start.y ());
      }

      if (shape->has_prop_id () && shape->prop_id () != 0) {
        write_props (layout, layout.properties_repository ().properties (shape->prop_id ()));
      }

      db::Point end (start);
      if (! pointlist.empty ()) {

        end += pointlist.back ();

        info = 0;
        if (mm_geometry_x != end.x ()) {
          info |= 0x10;
        }
        if (mm_geometry_y != end.y ()) {
          info |= 0x08;
        }

        write_byte (27);
        write_byte (info);

        if (info & 0x10) {
          mm_geometry_x = end.x ();
          write_coord (end.x ());
        }
        if (info & 0x08) {
          mm_geometry_y = end.y ();
          write_coord (end.y ());
        }

        if (shape->has_prop_id () && shape->prop_id () != 0) {
          write_props (layout, layout.properties_repository ().properties (shape->prop_id ()));
        }

      }

    }

    ++shape;

  }
}

void 
OASISWriter::write_edges (const db::Layout &layout, const db::Cell &cell, const db::LayerProperties &lprops, unsigned int lindex)
{
  int layer = lprops.layer;
  int datatype = lprops.datatype;

  db::ShapeIterator shape;
  shape = cell.shapes (lindex).begin (db::ShapeIterator::Edge);

  std::vector<db::Point> pointlist;

  while (! shape.at_end ()) {

    db::Edge e (shape->edge ());
    pointlist.reserve (1);
    pointlist.erase (pointlist.begin (), pointlist.end ());
    pointlist.push_back (e.p2 () - e.p1 ());

    unsigned char info = 0x00;

    if (mm_layer != layer) {
      info |= 0x01;
    }
    if (mm_datatype != datatype) {
      info |= 0x02;
    }
    if (mm_geometry_x != e.p1 ().x ()) {
      info |= 0x10;
    }
    if (mm_geometry_y != e.p1 ().y ()) {
      info |= 0x08;
    }
    if (mm_path_point_list != pointlist) {
      info |= 0x20;
    }
    if (mm_path_start_extension != 0 || mm_path_end_extension != 0) {
      info |= 0x80;
    }
    if (mm_path_halfwidth != 0) {
      info |= 0x40;
    }

    write_byte (21);
    write_byte (info);

    if (info & 0x01) {
      mm_layer = layer;
      write ((unsigned long) layer);
    }
    if (info & 0x02) {
      mm_datatype = datatype;
      write ((unsigned long) datatype);
    }
    if (info & 0x40) {
      mm_path_halfwidth = 0;
      write (0);
    }

    if (info & 0x80) {
      write_byte (0x05);  // flush
      mm_path_start_extension = 0;
      mm_path_end_extension = 0;
    }

    if (info & 0x20) {
      mm_path_point_list = pointlist;
      write_pointlist (pointlist, false /*=for paths*/);
    }
    if (info & 0x10) {
      mm_geometry_x = e.p1 ().x ();
      write_coord (e.p1 ().x ());
    }
    if (info & 0x08) {
      mm_geometry_y = e.p1 ().y ();
      write_coord (e.p1 ().y ());
    }

    if (shape->has_prop_id () && shape->prop_id () != 0) {
      write_props (layout, layout.properties_repository ().properties (shape->prop_id ()));
    }

    ++shape;

  }
}

}

