
/*

  KLayout Layout Viewer
  Copyright (C) 2006-2007 Matthias Koefferlein

  This program is free software; you can redistribute it and/or modify
  it under the terms of the GNU General Public License as published by
  the Free Software Foundation; either version 2 of the License, or
  (at your option) any later version.

  This program is distributed in the hope that it will be useful,
  but WITHOUT ANY WARRANTY; without even the implied warranty of
  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
  GNU General Public License for more details.

  You should have received a copy of the GNU General Public License
  along with this program; if not, write to the Free Software
  Foundation, Inc., 51 Franklin St, Fifth Floor, Boston, MA  02110-1301  USA

*/


#ifndef HDR_dbShape
#define HDR_dbShape

#include "tlException.h"
#include "dbManager.h"
#include "dbPolygon.h"
#include "dbPath.h"
#include "dbEdge.h"
#include "dbText.h"
#include "dbBox.h"
#include "dbShapes.h"
#include "dbArray.h"

#include <set>

namespace db 
{

/**
 *  @brief A generic edge iterator 
 *
 *  This edge iterator is used to provide edge iterators for simple and
 *  complex polygons as well.
 *  The implementation is based on a union scheme. It assumes that the
 *  iterators represented by the generic instance do not require a special
 *  copy semantics but can be copied by bytewise copy simply.
 */
template <class C>
class generic_polygon_edge_iterator
{
public:
  typedef C coord_type;
  typedef db::edge<C> edge_type;
  typedef db::disp_trans<C> disp_type;
  typedef db::polygon<C> polygon_type;
  typedef typename polygon_type::polygon_edge_iterator polygon_edge_iterator_type;
  typedef db::simple_polygon<C> simple_polygon_type;
  typedef typename simple_polygon_type::polygon_edge_iterator simple_polygon_edge_iterator_type;
  typedef db::polygon_ref<polygon_type, disp_type> polygon_ref_type;
  typedef typename polygon_ref_type::polygon_edge_iterator polygon_ref_edge_iterator_type;
  typedef db::polygon_ref<simple_polygon_type, disp_type> simple_polygon_ref_type;
  typedef typename simple_polygon_ref_type::polygon_edge_iterator simple_polygon_ref_edge_iterator_type;
  typedef edge_type value_type;
  typedef void pointer;
  typedef value_type reference; 
  typedef std::bidirectional_iterator_tag iterator_category;
  typedef void difference_type;

  enum iterator_type
  {
    None, SimplePolygon, SimplePolygonRef, Polygon, PolygonRef
  };

  /**
   *  @brief Default ctor
   */
  generic_polygon_edge_iterator ()
  {
    m_type = None;
  }
    
  /**
   *  @brief This constructor creates an edge iterator at "begin" for polygons
   * 
   *  The iterator stores begin and end and provides a "at_end" method to
   *  test if the sequence is at the end.
   *
   *  @param begin The start of the sequence
   */
  generic_polygon_edge_iterator (const polygon_edge_iterator_type &begin)
  {
    m_type = Polygon;
    new (m_d.iter) polygon_edge_iterator_type (begin);
  }
    
  /**
   *  @brief This constructor creates an edge iterator at "begin" for polygon references
   * 
   *  See former ctor.
   */
  generic_polygon_edge_iterator (const polygon_ref_edge_iterator_type &begin)
  {
    m_type = PolygonRef;
    new (m_d.iter) polygon_ref_edge_iterator_type (begin);
  }
    
  /**
   *  @brief This constructor creates an edge iterator at "begin" for simple polygons
   * 
   *  See former ctor.
   */
  generic_polygon_edge_iterator (const simple_polygon_edge_iterator_type &begin)
  {
    m_type = SimplePolygon;
    new (m_d.iter) simple_polygon_edge_iterator_type (begin);
  }
    
  /**
   *  @brief This constructor creates an edge iterator at "begin" for simple polygon references
   * 
   *  See former ctor.
   */
  generic_polygon_edge_iterator (const simple_polygon_ref_edge_iterator_type &begin)
  {
    m_type = SimplePolygonRef;
    new (m_d.iter) simple_polygon_ref_edge_iterator_type (begin);
  }
    
  /**
   *  @brief at_end predicate
   */
  bool at_end () const
  {
    return generic_const_f<bool, at_end_f> ();
  }

  /**
   *  @brief at_begin predicate
   */
  bool at_begin () const
  {
    return generic_const_f<bool, at_end_f> ();
  }

  /**
   *  @brief Increment operator
   */
  generic_polygon_edge_iterator &operator++ () 
  {
    generic_f<bool, inc_f> ();
    return *this;
  }

  /**
   *  @brief Decrement operator
   */
  generic_polygon_edge_iterator &operator-- () 
  {
    generic_f<bool, dec_f> ();
    return *this;
  }

  /**
   *  @brief access operator
   */
  edge_type operator* () const
  {
    return generic_const_f<edge_type, deref_f> ();
  }
  
private:
  //  This union is there to compute the maximum size required to hold all
  //  kind of iterators 
  union iter_size {
    char sz1 [sizeof (polygon_edge_iterator_type)];
    char sz2 [sizeof (polygon_ref_edge_iterator_type)];
    char sz3 [sizeof (simple_polygon_edge_iterator_type)];
    char sz4 [sizeof (simple_polygon_ref_edge_iterator_type)];
  };

  //  This member must be first to guarantee alignment on 64bit systems:
  //  The strange construction and the local dummy class helps to guarantee alignment of the "iter" space
  union {
    class _align_helper { long l; } _ah;
    char iter [sizeof (iter_size)];
  } m_d;

  iterator_type m_type;

  struct at_end_f {
    template <class Iter>
    bool operator() (const Iter &iter) const
    {
      return iter.at_end ();
    } 
  };

  struct inc_f {
    template <class Iter>
    bool operator() (Iter &iter) const
    {
      iter.operator++ ();
      return false;
    } 
  };

  struct dec_f {
    template <class Iter>
    bool operator() (Iter &iter) const
    {
      iter.operator-- ();
      return false;
    } 
  };

  struct deref_f {
    template <class Iter>
    edge_type operator() (const Iter &iter) const
    {
      return iter.operator* ();
    } 
  };

  struct at_begin_f {
    template <class Iter>
    bool operator() (const Iter &iter) const
    {
      return iter.at_begin ();
    } 
  };

  template <class Ret, class Func>
  Ret generic_const_f () const
  {
    Func f;
    if (m_type == Polygon) {
      return f (*((const polygon_edge_iterator_type *) m_d.iter));
    } else if (m_type == PolygonRef) {
      return f (*((const polygon_ref_edge_iterator_type *) m_d.iter));
    } else if (m_type == SimplePolygon) {
      return f (*((const simple_polygon_edge_iterator_type *) m_d.iter));
    } else {
      return f (*((const simple_polygon_ref_edge_iterator_type *) m_d.iter));
    }
  }

  template <class Ret, class Func>
  Ret generic_f () 
  {
    Func f;
    if (m_type == Polygon) {
      return f (*((polygon_edge_iterator_type *) m_d.iter));
    } else if (m_type == PolygonRef) {
      return f (*((polygon_ref_edge_iterator_type *) m_d.iter));
    } else if (m_type == SimplePolygon) {
      return f (*((simple_polygon_edge_iterator_type *) m_d.iter));
    } else {
      return f (*((simple_polygon_ref_edge_iterator_type *) m_d.iter));
    }
  }
};

/**
 *  @brief A generic point iterator 
 *
 *  This point iterator is used to provide point iterators for simple and
 *  complex polygons as well as paths.
 *  The implementation is based on a union scheme. It assumes that the
 *  iterators represented by the generic instance do not require a special
 *  copy semantics but can be copied by bytewise copy simply.
 */
template <class C>
class generic_point_iterator
{
public:
  typedef C coord_type;
  typedef db::point<C> point_type;
  typedef db::polygon<C> polygon_type;
  typedef db::disp_trans<C> disp_type;
  typedef typename polygon_type::polygon_contour_iterator polygon_point_iterator_type;
  typedef db::polygon_ref<polygon_type, disp_type> polygon_ref_type;
  typedef typename polygon_ref_type::polygon_contour_iterator polygon_ref_point_iterator_type;
  typedef db::path<C> path_type;
  typedef typename path_type::iterator path_point_iterator_type;
  typedef db::path_ref<path_type, disp_type> path_ref_type;
  typedef typename path_ref_type::iterator path_ref_point_iterator_type;
  typedef point_type value_type;
  typedef void pointer;
  typedef value_type reference; 
  typedef std::bidirectional_iterator_tag iterator_category;
  typedef void difference_type;

  enum iterator_type
  {
    Polygon, PolygonRef, Path, PathRef
  };

  /**
   *  @brief This constructor creates an point iterator at "begin" for polygons
   * 
   *  The iterator stores begin and end and provides a "at_end" method to
   *  test if the sequence is at the end.
   *  This method also applies to simple polygon contours.
   *
   *  @param begin The start of the sequence
   */
  generic_point_iterator (const polygon_point_iterator_type &begin)
  {
    m_type = Polygon;
    new (m_d.iter) polygon_point_iterator_type (begin);
  }
    
  /**
   *  @brief This constructor creates an point iterator at "begin" for polygon references
   * 
   *  This method also applies to simple polygon references.
   *  See former ctor.
   */
  generic_point_iterator (const polygon_ref_point_iterator_type &begin)
  {
    m_type = PolygonRef;
    new (m_d.iter) polygon_ref_point_iterator_type (begin);
  }
    
  /**
   *  @brief This constructor creates an point iterator at "begin" for paths
   * 
   *  See former ctor.
   */
  generic_point_iterator (const path_point_iterator_type &begin)
  {
    m_type = Path;
    new (m_d.iter) path_point_iterator_type (begin);
  }
    
  /**
   *  @brief This constructor creates an point iterator at "begin" for path references
   * 
   *  See former ctor.
   */
  generic_point_iterator (const path_ref_point_iterator_type &begin)
  {
    m_type = PathRef;
    new (m_d.iter) path_ref_point_iterator_type (begin);
  }
    
  /**
   *  @brief at_end predicate
   */
  bool at_end () const
  {
    return generic_const_f<bool, at_end_f> ();
  }

  /**
   *  @brief at_begin predicate
   */
  bool at_begin () const
  {
    return generic_const_f<bool, at_end_f> ();
  }

  /**
   *  @brief Increment operator
   */
  generic_point_iterator &operator++ () 
  {
    generic_f<bool, inc_f> ();
    return *this;
  }

  /**
   *  @brief Decrement operator
   */
  generic_point_iterator &operator-- () 
  {
    generic_f<bool, dec_f> ();
    return *this;
  }

  /**
   *  @brief access operator
   */
  point_type operator* () const
  {
    return generic_const_f<point_type, deref_f> ();
  }

  /**
   *  @brief Sorting order 
   */
  bool operator< (const generic_point_iterator &d) const
  {
    if (m_type != d.m_type) {
      return m_type < d.m_type;
    }

    if (m_type == Polygon) {
      return *((const polygon_point_iterator_type *) m_d.iter) < *((const polygon_point_iterator_type *) d.m_d.iter);
    } else if (m_type == PolygonRef) {
      return *((const polygon_ref_point_iterator_type *) m_d.iter) < *((const polygon_ref_point_iterator_type *) d.m_d.iter);
    } else if (m_type == Path) {
      return *((const path_point_iterator_type *) m_d.iter) < *((const path_point_iterator_type *) d.m_d.iter);
    } else {
      return *((const path_ref_point_iterator_type *) m_d.iter) < *((const path_ref_point_iterator_type *) d.m_d.iter);
    }
  }

  /**
   *  @brief Equality 
   */
  bool operator== (const generic_point_iterator &d) const
  {
    tl_assert (m_type == d.m_type);

    if (m_type == Polygon) {
      return *((const polygon_point_iterator_type *) m_d.iter) == *((const polygon_point_iterator_type *) d.m_d.iter);
    } else if (m_type == PolygonRef) {
      return *((const polygon_ref_point_iterator_type *) m_d.iter) == *((const polygon_ref_point_iterator_type *) d.m_d.iter);
    } else if (m_type == Path) {
      return *((const path_point_iterator_type *) m_d.iter) == *((const path_point_iterator_type *) d.m_d.iter);
    } else {
      return *((const path_ref_point_iterator_type *) m_d.iter) == *((const path_ref_point_iterator_type *) d.m_d.iter);
    }
  }

  /**
   *  @brief Inequality
   */
  bool operator!= (const generic_point_iterator &d) const
  {
    return !operator== (d);
  }
  
private:
  //  This union is there to compute the maximum size required to hold all
  //  kind of iterators 
  union iter_size {
    char sz1 [sizeof (polygon_point_iterator_type)];
    char sz2 [sizeof (polygon_ref_point_iterator_type)];
    char sz3 [sizeof (path_point_iterator_type)];
    char sz4 [sizeof (path_ref_point_iterator_type)];
  };

  //  This member must be first to guarantee alignment on 64bit systems:
  //  The strange construction and the local dummy class helps to guarantee alignment of the "iter" space
  union {
    class _align_helper { long l; } _ah;
    char iter [sizeof (iter_size)];
  } m_d;

  iterator_type m_type;

  struct at_end_f {
    template <class Iter>
    bool operator() (const Iter &iter) const
    {
      return iter.at_end ();
    } 
  };

  struct inc_f {
    template <class Iter>
    bool operator() (Iter &iter) const
    {
      iter.operator++ ();
      return false;
    } 
  };

  struct dec_f {
    template <class Iter>
    bool operator() (Iter &iter) const
    {
      iter.operator-- ();
      return false;
    } 
  };

  struct deref_f {
    template <class Iter>
    point_type operator() (const Iter &iter) const
    {
      return iter.operator* ();
    } 
  };

  struct at_begin_f {
    template <class Iter>
    bool operator() (const Iter &iter) const
    {
      return iter.at_begin ();
    } 
  };

  template <class Ret, class Func>
  Ret generic_const_f () const
  {
    Func f;
    if (m_type == Polygon) {
      return f (*((const polygon_point_iterator_type *) m_d.iter));
    } else if (m_type == PolygonRef) {
      return f (*((const polygon_ref_point_iterator_type *) m_d.iter));
    } else if (m_type == Path) {
      return f (*((const path_point_iterator_type *) m_d.iter));
    } else {
      return f (*((const path_ref_point_iterator_type *) m_d.iter));
    }
  }

  template <class Ret, class Func>
  Ret generic_f () 
  {
    Func f;
    if (m_type == Polygon) {
      return f (*((polygon_point_iterator_type *) m_d.iter));
    } else if (m_type == PolygonRef) {
      return f (*((polygon_ref_point_iterator_type *) m_d.iter));
    } else if (m_type == Path) {
      return f (*((path_point_iterator_type *) m_d.iter));
    } else {
      return f (*((path_ref_point_iterator_type *) m_d.iter));
    }
  }
};

/**
 *  @brief A shape proxy 
 *
 *  The shape proxy is basically a pointer to a shape of different kinds 
 *  No copy of the shape is created: if the shape proxy is copied the copy still
 *  points to the original shape. If the original shape is modified or deleted,
 *  the shape proxy will also point to a modified or invalid shape.
 *  The proxy can be "null" which means a invalid reference.
 */ 
template <class C>
class shape
{
public:
  typedef C coord_type;
  typedef db::simple_trans<C> trans_type;
  typedef db::disp_trans<C> disp_type;
  typedef db::polygon<C> polygon_type;
  typedef db::simple_polygon<C> simple_polygon_type;
  typedef db::polygon_ref<polygon_type, disp_type> polygon_ref_type;
  typedef db::polygon_ref<polygon_type, unit_trans> polygon_ptr_type;
  typedef db::array<polygon_ptr_type, disp_type> polygon_ptr_array_type;
  typedef db::polygon_ref<simple_polygon_type, disp_type> simple_polygon_ref_type;
  typedef db::polygon_ref<simple_polygon_type, unit_trans> simple_polygon_ptr_type;
  typedef db::array<simple_polygon_ptr_type, disp_type> simple_polygon_ptr_array_type;
  typedef db::path<C> path_type;
  typedef db::path_ref<path_type, disp_type> path_ref_type;
  typedef db::path_ref<path_type, unit_trans> path_ptr_type;
  typedef db::array<path_ptr_type, disp_type> path_ptr_array_type;
  typedef db::edge<C> edge_type;
  typedef db::text<C> text_type;
  typedef db::text_ref<text_type, disp_type> text_ref_type;
  typedef db::text_ref<text_type, unit_trans> text_ptr_type;
  typedef db::array<text_ptr_type, disp_type> text_ptr_array_type;
  typedef db::user_object<C> user_object_type;
  typedef db::box<C> box_type;
  typedef db::array<box_type, unit_trans> box_array_type;
  typedef db::box<C, short> short_box_type;
  typedef db::array<short_box_type, unit_trans> short_box_array_type;
  typedef db::point<C> point_type;
  typedef db::generic_polygon_edge_iterator<C> polygon_edge_iterator;
  typedef db::generic_point_iterator<C> point_iterator;

  //  Hint: in the following enum it is important that the ArrayMember types come *after*
  //  the non-member types. See db::shapes::erase with multi-shape erase capabilities.
  enum object_type {

    Null,

    Polygon,
    PolygonRef,
    PolygonPtrArray,
    PolygonPtrArrayMember,
    SimplePolygon,
    SimplePolygonRef,
    SimplePolygonPtrArray,
    SimplePolygonPtrArrayMember,
    Edge,
    Path,
    PathRef,
    PathPtrArray,
    PathPtrArrayMember,
    Box,
    BoxArray,
    BoxArrayMember,
    ShortBox,
    ShortBoxArray,
    ShortBoxArrayMember,
    Text,
    TextRef,
    TextPtrArray,
    TextPtrArrayMember,
    UserObject
  };
    
  /** 
   *  @brief Construct a shape proxy as a null object
   */
  shape ()
    : m_with_props (false), m_type (Null)
  {
    m_generic.any = 0;
  }

  /**
   *  @brief Construct as a proxy to a certain object
   */
  template <class Obj>
  shape (const Obj &obj)
  {
    m_with_props = false;
    init (obj);
    m_generic.any = &obj;
  }

  /** 
   *  @brief Construct a shape proxy as a shape with properties
   */
  template <class Sh>
  void init (const db::object_with_properties<Sh> &swp)
  {
    init ((const Sh &) swp);
    m_with_props = true;
  }

  /**
   *  @brief Construct as a proxy to a certain object
   *
   *  This form will copy the object itself. A temporary object may be passed to "obj".
   *  An implementation is provided only for the array types
   */
  template <class Obj, class Trans>
  shape (const Obj &obj, const Trans &trans)
  {
    m_with_props = false;
    init_member (obj, trans);
    m_generic.any = &obj;
  }

  /** 
   *  @brief Construct a shape proxy as a shape with properties
   *
   *  This form will copy the object itself. A temporary object may be passed to "obj".
   *  An implementation is provided only for a few types.
   */
  template <class Sh, class Trans>
  void init_member (const db::object_with_properties<Sh> &swp, const Trans &trans)
  {
    init_member ((const Sh &) swp, trans);
    m_with_props = true;
  }

  /** 
   *  @brief Construct a shape proxy as a reference to a polygon
   */
  void init (const polygon_type &polygon)
  {
    m_type = Polygon;
  }

  /** 
   *  @brief Construct a shape proxy as a reference to a polygon reference
   */
  void init (const polygon_ref_type &polygon_ref)
  {
    m_type = PolygonRef;
  }

  /** 
   *  @brief Construct a shape proxy as a reference to a polygon array member
   */
  void init_member (const polygon_ptr_array_type &polygon_aref, const trans_type &trans)
  {
    m_trans = trans;
    m_type = PolygonPtrArrayMember;
  }

  /** 
   *  @brief Construct a shape proxy as a reference to a polygon reference array
   */
  void init (const polygon_ptr_array_type &polygon_aref)
  {
    m_type = PolygonPtrArray;
  }

  /** 
   *  @brief Construct a shape proxy as a reference to a simple olygon
   */
  void init (const simple_polygon_type &polygon)
  {
    m_type = SimplePolygon;
  }

  /** 
   *  @brief Construct a shape proxy as a reference to a polygon reference
   */
  void init (const simple_polygon_ref_type &polygon_ref)
  {
    m_type = SimplePolygonRef;
  }

  /** 
   *  @brief Construct a shape proxy as a reference to a simple polygon array member
   */
  void init_member (const simple_polygon_ptr_array_type &simple_polygon_aref, const trans_type &trans)
  {
    m_trans = trans;
    m_type = SimplePolygonPtrArrayMember;
  }

  /** 
   *  @brief Construct a shape proxy as a reference to a simple polygon reference array
   */
  void init (const simple_polygon_ptr_array_type &simple_polygon_aref)
  {
    m_type = SimplePolygonPtrArray;
  }

  /** 
   *  @brief Construct a shape proxy as a reference to a path
   */
  void init (const path_type &path)
  {
    m_type = Path;
  }

  /** 
   *  @brief Construct a shape proxy as a reference to a path reference
   */
  void init (const path_ref_type &path)
  {
    m_type = PathRef;
  }

  /** 
   *  @brief Construct a shape proxy as a reference to a path array member
   */
  void init_member (const path_ptr_array_type &path_aref, const trans_type &trans)
  {
    m_trans = trans;
    m_type = PathPtrArrayMember;
  }

  /** 
   *  @brief Construct a shape proxy as a reference to a path reference array
   */
  void init (const path_ptr_array_type &path_aref)
  {
    m_type = PathPtrArray;
  }

  /** 
   *  @brief Construct a shape proxy as a reference to a text
   */
  void init (const text_type &text)
  {
    m_type = Text;
  }

  /** 
   *  @brief Construct a shape proxy as a reference to a text reference
   */
  void init (const text_ref_type &text_ref)
  {
    m_type = TextRef;
  }

  /** 
   *  @brief Construct a shape proxy as a reference to a path array member
   */
  void init_member (const text_ptr_array_type &text_aref, const trans_type &trans)
  {
    m_trans = trans;
    m_type = TextPtrArrayMember;
  }

  /** 
   *  @brief Construct a shape proxy as a reference to a text reference array
   */
  void init (const text_ptr_array_type &text_aref)
  {
    m_type = PathPtrArray;
  }

  /** 
   *  @brief Construct a shape proxy as a reference to a edge
   */
  void init (const edge_type &edge)
  {
    m_type = Edge;
  }

  /** 
   *  @brief Construct a shape proxy as a reference to a box
   */
  void init (const box_type &box)
  {
    m_type = Box;
  }

  /** 
   *  @brief Construct a shape proxy as a reference to a box array member
   */
  void init_member (const box_array_type &box_array, const trans_type &trans)
  {
    m_trans = trans;
    m_type = BoxArrayMember;
  }

  /** 
   *  @brief Construct a shape proxy as a reference to a box array
   */
  void init (const box_array_type &box_array)
  {
    m_type = BoxArray;
  }

  /** 
   *  @brief Construct a shape proxy as a reference to a small box
   */
  void init (const short_box_type &box)
  {
    m_type = ShortBox;
  }

  /** 
   *  @brief Construct a shape proxy as a reference to a box array member
   */
  void init_member (const short_box_array_type &box_array, const trans_type &trans)
  {
    m_trans = trans;
    m_type = ShortBoxArrayMember;
  }

  /** 
   *  @brief Construct a shape proxy as a reference to a box array
   */
  void init (const short_box_array_type &box_array)
  {
    m_type = ShortBoxArray;
  }

  /** 
   *  @brief Construct a shape proxy as a reference to a user object
   */
  void init (const user_object_type &obj)
  {
    m_type = UserObject;
  }

  /**
   *  @brief Determine, if the shape is of "object_with_properties" type.
   *
   *  Usually, the properties id (see prop_id()) should be used.
   *  In order to determine however, which object "basic_ptr" points to,
   *  this method is required.
   *
   *  @return true, if the shape is of "object_with_properties<Sh> type".
   */
  bool with_props () const
  {
    return m_with_props;
  }

  /**
   *  @brief Get the properties Id associated with the shape
   */
  db::PropertiesRepository::properties_id_type prop_id () const
  {
    if (m_with_props) {
      switch (m_type) {
      case Polygon:
        return m_generic.ppolygon->properties_id ();
      case PolygonRef:
        return m_generic.ppolygon_ref->properties_id ();
      case PolygonPtrArray:
      case PolygonPtrArrayMember:
        return m_generic.ppolygon_aref->properties_id ();
      case SimplePolygon:
        return m_generic.psimple_polygon->properties_id ();
      case SimplePolygonRef:
        return m_generic.psimple_polygon_ref->properties_id ();
      case SimplePolygonPtrArray:
      case SimplePolygonPtrArrayMember:
        return m_generic.psimple_polygon_aref->properties_id ();
      case Edge:
        return m_generic.pedge->properties_id ();
      case Path:
        return m_generic.ppath->properties_id ();
      case PathRef:
        return m_generic.ppath_ref->properties_id ();
      case PathPtrArray:
      case PathPtrArrayMember:
        return m_generic.ppath_aref->properties_id ();
      case Box:
        return m_generic.pbox->properties_id ();
      case BoxArray:
      case BoxArrayMember:
        return m_generic.pbox_array->properties_id ();
      case ShortBox:
        return m_generic.pshort_box->properties_id ();
      case ShortBoxArray:
      case ShortBoxArrayMember:
        return m_generic.pshort_box_array->properties_id ();
      case Text:
        return m_generic.ptext->properties_id ();
      case TextRef:
        return m_generic.ptext_ref->properties_id ();
      case TextPtrArray:
      case TextPtrArrayMember:
        return m_generic.ptext_aref->properties_id ();
      case UserObject:
        return m_generic.puser_object->properties_id ();
      default:
        return 0;
      }
    } else {
      return 0;
    }
  }

  /**
   *  @brief Check, if the shape is associated with a properties Id
   */
  bool has_prop_id () const
  {
    return m_with_props;
  }

  /**
   *  @brief Create the "begin" point iterator
   *
   *  This method applies to paths.
   */
  point_iterator begin_point () const
  {
    if (m_type == Path) {
      return point_iterator (path ().begin ());
    } else if (m_type == PathRef || m_type == PathPtrArrayMember) {
      return point_iterator (path_ref ().begin ());
    } else {
      tl_assert (false);
    }
  }

  /**
   *  @brief Create the "end" point iterator
   *
   *  This method applies to paths.
   */
  point_iterator end_point () const
  {
    if (m_type == Path) {
      return point_iterator (path ().end ());
    } else if (m_type == PathRef || m_type == PathPtrArrayMember) {
      return point_iterator (path_ref ().end ());
    } else {
      tl_assert (false);
    }
  }

  /**
   *  @brief Create the "begin" hull point iterator
   *
   *  This method applies to polygons.
   */
  point_iterator begin_hull () const
  {
    if (m_type == SimplePolygon) {
      return point_iterator (simple_polygon ().begin_hull ());
    } else if (m_type == SimplePolygonRef || m_type == SimplePolygonPtrArrayMember) {
      return point_iterator (simple_polygon_ref ().begin_hull ());
    } else if (m_type == Polygon) {
      return point_iterator (polygon ().begin_hull ());
    } else if (m_type == PolygonRef || m_type == PolygonPtrArrayMember) {
      return point_iterator (polygon_ref ().begin_hull ());
    } else {
      tl_assert (false);
    }
  }

  /**
   *  @brief Create the "end" hull point iterator
   *
   *  This method applies to polygons.
   */
  point_iterator end_hull () const
  {
    if (m_type == SimplePolygon) {
      return point_iterator (simple_polygon ().end_hull ());
    } else if (m_type == SimplePolygonRef || m_type == SimplePolygonPtrArrayMember) {
      return point_iterator (simple_polygon_ref ().end_hull ());
    } else if (m_type == Polygon) {
      return point_iterator (polygon ().end_hull ());
    } else if (m_type == PolygonRef || m_type == PolygonPtrArrayMember) {
      return point_iterator (polygon_ref ().end_hull ());
    } else {
      tl_assert (false);
    }
  }

  /**
   *  @brief Create the "begin" hole point iterator
   *
   *  This method applies to polygons.
   *  Simple polygons deliver an empty sequence.
   *
   *  @param hole The hole index (see holes () method)
   */
  point_iterator begin_hole (unsigned int hole) const
  {
    if (m_type == SimplePolygon) {
      return point_iterator (simple_polygon ().begin_hole (hole));
    } else if (m_type == SimplePolygonRef || m_type == SimplePolygonPtrArrayMember) {
      return point_iterator (simple_polygon_ref ().begin_hole (hole));
    } else if (m_type == Polygon) {
      return point_iterator (polygon ().begin_hole (hole));
    } else if (m_type == PolygonRef || m_type == PolygonPtrArrayMember) {
      return point_iterator (polygon_ref ().begin_hole (hole));
    } else {
      tl_assert (false);
    }
  }

  /**
   *  @brief Create the "end" hole point iterator
   *
   *  This method applies to polygons.
   *  Simple polygons deliver an empty sequence.
   *
   *  @param hole The hole index (see holes () method)
   */
  point_iterator end_hole (unsigned int hole) const
  {
    if (m_type == SimplePolygon) {
      return point_iterator (simple_polygon ().end_hole (hole));
    } else if (m_type == SimplePolygonRef || m_type == SimplePolygonPtrArrayMember) {
      return point_iterator (simple_polygon_ref ().end_hole (hole));
    } else if (m_type == Polygon) {
      return point_iterator (polygon ().end_hole (hole));
    } else if (m_type == PolygonRef || m_type == PolygonPtrArrayMember) {
      return point_iterator (polygon_ref ().end_hole (hole));
    } else {
      tl_assert (false);
    }
  }

  /**
   *  @brief Return the number of holes
   *
   *  This method applies to polygons.
   *  Simple polygons deliver a zero value.
   */
  unsigned int holes () const
  {
    if (m_type == SimplePolygon) {
      return simple_polygon ().holes ();
    } else if (m_type == SimplePolygonRef || m_type == SimplePolygonPtrArrayMember) {
      return simple_polygon_ref ().obj ().holes ();
    } else if (m_type == Polygon) {
      return polygon ().holes ();
    } else if (m_type == PolygonRef || m_type == PolygonPtrArrayMember) {
      return polygon_ref ().obj ().holes ();
    } else {
      tl_assert (false);
    }
  }

  /**
   *  @brief Create and return an edge iterator
   */
  polygon_edge_iterator begin_edge () const
  {
    if (m_type == SimplePolygon) {
      return polygon_edge_iterator (simple_polygon ().begin_edge ());
    } else if (m_type == SimplePolygonRef || m_type == SimplePolygonPtrArrayMember) {
      return polygon_edge_iterator (simple_polygon_ref ().begin_edge ());
    } else if (m_type == Polygon) {
      return polygon_edge_iterator (polygon ().begin_edge ());
    } else if (m_type == PolygonRef || m_type == PolygonPtrArrayMember) {
      return polygon_edge_iterator (polygon_ref ().begin_edge ());
    } else {
      tl_assert (false);
    }
  }

  /** 
   *  @brief Return the type of the shape reference
   */
  object_type type () const
  {
    return m_type;
  }

  /** 
   *  @brief Test if the shape proxy is a null object
   */
  bool is_null () const
  {
    return m_type == Null;
  }

  /** 
   *  @brief Return the internal object reference as a void pointer
   *
   *  If the object that the shape is referencing is known, the void pointer may be
   *  cast to the actual pointer.
   */
  const void *basic_ptr () const
  {
    return m_generic.any;
  }

  /** 
   *  @brief Return a reference to the polygon if one is referenced
   */
  const polygon_type &polygon () const
  {
    tl_assert (m_type == Polygon);
    return m_with_props ? *m_generic.ppolygon : *m_generic.polygon;
  }

  /** 
   *  @brief Return a polygon reference if one is referenced
   */
  polygon_ref_type polygon_ref () const
  {
    tl_assert (m_type == PolygonRef || m_type == PolygonPtrArrayMember);
    if (m_type == PolygonRef) {
      return m_with_props ? *m_generic.ppolygon_ref : *m_generic.polygon_ref;
    } else {
      tl_assert (m_trans.rot () == 0);
      return polygon_ref_type (m_with_props ? &m_generic.ppolygon_aref->object ().obj () : &m_generic.polygon_aref->object ().obj (), m_trans.disp ());
    }
  }

  /** 
   *  @brief Return a reference to the simple polygon if one is referenced
   */
  const simple_polygon_type &simple_polygon () const
  {
    tl_assert (m_type == SimplePolygon);
    return m_with_props ? *m_generic.psimple_polygon : *m_generic.simple_polygon;
  }

  /** 
   *  @brief Return a simple polygon reference if one is referenced
   */
  simple_polygon_ref_type simple_polygon_ref () const
  {
    tl_assert (m_type == SimplePolygonRef || m_type == SimplePolygonPtrArrayMember);
    if (m_type == SimplePolygonRef) {
      return m_with_props ? *m_generic.psimple_polygon_ref : *m_generic.simple_polygon_ref;
    } else {
      tl_assert (m_trans.rot () == 0);
      return simple_polygon_ref_type (m_with_props ? &m_generic.psimple_polygon_aref->object ().obj () : &m_generic.simple_polygon_aref->object ().obj (), m_trans.disp ());
    }
  }

  /** 
   *  @brief Test if the shape proxy points to a polygon
   */
  bool is_polygon () const
  {
    return (m_type == Polygon || m_type == PolygonRef || m_type == PolygonPtrArrayMember || 
            m_type == SimplePolygon || m_type == SimplePolygonRef || m_type == SimplePolygonPtrArrayMember);
  }

  /** 
   *  @brief Instantiate the polygon object
   * 
   *  If a polygon is referenced, this object is instantiated
   *  by this method. Paths are converted to polygons. Boxes are also converted.
   */
  void polygon (polygon_type &p) const
  {
    if (m_type == Polygon) {
      p = m_with_props ? *m_generic.ppolygon : *m_generic.polygon;
    } else if (m_type == PolygonRef || m_type == PolygonPtrArrayMember) {
      polygon_ref ().instantiate (p);
    } else if (m_type == SimplePolygon) {
      p.clear ();
      if (m_with_props) {
        p.assign_hull (m_generic.psimple_polygon->begin_hull (), m_generic.psimple_polygon->end_hull (), false /*no additional compression*/);
      } else {
        p.assign_hull (m_generic.simple_polygon->begin_hull (), m_generic.simple_polygon->end_hull (), false /*no additional compression*/);
      }
    } else if (m_type == SimplePolygonRef || m_type == SimplePolygonPtrArrayMember) {
      p.clear ();
      p.assign_hull (simple_polygon_ref ().obj ().begin_hull (), simple_polygon_ref ().obj ().end_hull (), simple_polygon_ref ().trans (), false /*no additional compression*/);
    } else if (m_type == Path) {
      p = m_generic.path->polygon ();
    } else if (m_type == PathRef || m_type == PathPtrArrayMember) {
      p = path_ref ().obj ().polygon ();
      p.transform (path_ref ().trans ());
    } else if (is_box ()) {
      p = box ();
    } else {
      tl_assert (false);
    }
  }

  /** 
   *  @brief Test if the shape proxy points to a simple polygon
   */
  bool is_simple_polygon () const
  {
    return (m_type == SimplePolygon || m_type == SimplePolygonRef || m_type == SimplePolygonPtrArrayMember);
  }

  /** 
   *  @brief Instantiate the simple polygon object
   * 
   *  If a simple polygon is referenced, this object is instantiated
   *  by this method. Paths are converted to polygons. Boxes are also converted.
   */
  void simple_polygon (simple_polygon_type &p) const
  {
    if (m_type == SimplePolygon) {
      p.clear ();
      if (m_with_props) {
        p.assign_hull (m_generic.psimple_polygon->begin_hull (), m_generic.psimple_polygon->end_hull (), false /*no additional compression*/);
      } else {
        p.assign_hull (m_generic.simple_polygon->begin_hull (), m_generic.simple_polygon->end_hull (), false /*no additional compression*/);
      }
    } else if (m_type == SimplePolygonRef || m_type == SimplePolygonPtrArrayMember) {
      p.clear ();
      p.assign_hull (simple_polygon_ref ().obj ().begin_hull (), simple_polygon_ref ().obj ().end_hull (), simple_polygon_ref ().trans (), false /*no additional compression*/);
    } else if (m_type == Path) {
      p = m_generic.path->simple_polygon ();
    } else if (m_type == PathRef || m_type == PathPtrArrayMember) {
      p = path_ref ().obj ().simple_polygon ();
      p.transform (path_ref ().trans ());
    } else if (is_box ()) {
      p = box ();
    } else {
      tl_assert (false);
    }
  }

  /** 
   *  @brief Return a reference to the path if one is referenced
   */
  const path_type &path () const
  {
    tl_assert (m_type == Path);
    return m_with_props ? *m_generic.ppath : *m_generic.path;
  }

  /** 
   *  @brief Return a path reference if one is referenced
   */
  path_ref_type path_ref () const
  {
    tl_assert (m_type == PathRef || m_type == PathPtrArrayMember);
    if (m_type == PathRef) {
      return m_with_props ? *m_generic.ppath_ref : *m_generic.path_ref;
    } else {
      tl_assert (m_trans.rot () == 0);
      return path_ref_type (m_with_props ? &m_generic.ppath_aref->object ().obj () : &m_generic.path_aref->object ().obj (), m_trans.disp ());
    }
  }

  /** 
   *  @brief Test if the shape proxy points to a path
   */
  bool is_path () const
  {
    return (m_type == Path || m_type == PathRef || m_type == PathPtrArrayMember);
  }

  /**
   *  @brief Obtain the path width
   *
   *  Applies to paths only. Will assert if not a path.
   */
  db::Path::coord_type path_width () const
  {
    if (m_type == Path) {
      return m_with_props ? m_generic.ppath->width () : m_generic.path->width ();
    } else {
      return path_ref ().obj ().width ();
    }
  }
  
  /**
   *  @brief Obtain the path extensions
   *
   *  Applies to paths only. Will assert if not a path.
   *
   *  @return A pair consisting of the begin and end extensions.
   */
  std::pair<db::Path::coord_type, db::Path::coord_type> path_extensions () const
  {
    if (m_type == Path) {
      return m_with_props ? m_generic.ppath->extensions () : m_generic.path->extensions ();
    } else {
      return path_ref ().obj ().extensions ();
    }
  }
  
  /**
   *  @brief Obtain the round path flag
   *
   *  Applies to paths only. Will assert if not a path.
   *
   *  @return True, if the path has round ends
   */
  bool round_path () const
  {
    if (m_type == Path) {
      return m_with_props ? m_generic.ppath->round () : m_generic.path->round ();
    } else {
      return path_ref ().obj ().round ();
    }
  }
  
  /** 
   *  @brief Instantiate the path object
   * 
   *  If a path is referenced, this object is instantiated
   *  by this method.
   */
  void path (path_type &p) const
  {
    if (m_type == Path) {
      p = m_with_props ? *m_generic.ppath : *m_generic.path;
    } else if (m_type == PathRef || m_type == PathPtrArrayMember) {
      path_ref ().instantiate (p);
    } else {
      tl_assert (false);
    }
  }

  /** 
   *  @brief Return a reference to the edge if one is referenced
   */
  const edge_type &edge () const
  {
    tl_assert (m_type == Edge);
    return m_with_props ? *m_generic.pedge : *m_generic.edge;
  }

  /** 
   *  @brief Test if the shape proxy points to a edge
   */
  bool is_edge () const
  {
    return (m_type == Edge);
  }

  /** 
   *  @brief Instantiate the edge object
   * 
   *  If a edge is referenced, this object is instantiated
   *  by this method.
   */
  void edge (edge_type &e) const
  {
    e = edge ();
  }

  /** 
   *  @brief Return a reference to the text if one is referenced
   */
  const text_type &text () const
  {
    tl_assert (m_type == Text);
    return m_with_props ? *m_generic.ptext : *m_generic.text;
  }

  /** 
   *  @brief Return a text reference if one is referenced
   */
  text_ref_type text_ref () const
  {
    tl_assert (m_type == TextRef || m_type == TextPtrArrayMember);
    if (m_type == TextRef) {
      return m_with_props ? *m_generic.ptext_ref : *m_generic.text_ref;
    } else {
      tl_assert (m_trans.rot () == 0);
      return text_ref_type (m_with_props ? &m_generic.ptext_aref->object ().obj () : &m_generic.text_aref->object ().obj (), m_trans.disp ());
    }
  }

  /** 
   *  @brief Test if the shape proxy points to a text
   */
  bool is_text () const
  {
    return (m_type == Text || m_type == TextRef || m_type == TextPtrArrayMember);
  }

  /** 
   *  @brief Instantiate the text object
   * 
   *  If a text is referenced, this object is instantiated
   *  by this method.
   */
  void text (text_type &t) const
  {
    if (m_type == Text) {
      t = m_with_props ? *m_generic.ptext : *m_generic.text;
    } else {
      text_ref ().instantiate (t);
    }
  }

  /**
   *  @brief Obtain the text string
   *
   *  Applies to texts only. Will assert if not a text.
   */
  const std::string &text_string () const
  {
    if (m_type == Text) {
      return m_with_props ? m_generic.ptext->string () : m_generic.text->string ();
    } else {
      return text_ref ().obj ().string ();
    }
  }
  
  /**
   *  @brief Obtain the text transformation
   *
   *  Applies to texts only. Will assert if not a text.
   */
  db::Text::trans_type text_trans () const
  {
    if (m_type == Text) {
      return m_with_props ? m_generic.ptext->trans () : m_generic.text->trans ();
    } else {
      return trans_type (text_ref ().trans ()) * text_ref ().obj ().trans ();
    }
  }

  /**
   *  @brief Obtain the text size
   *
   *  Applies to texts only. Will assert if not a text.
   */
  db::Text::coord_type text_size () const
  {
    if (m_type == Text) {
      return m_with_props ? m_generic.ptext->size () : m_generic.text->size ();
    } else {
      return text_ref ().obj ().size ();
    }
  }

  /**
   *  @brief Obtain the text's font
   *
   *  Applies to texts only. Will assert if not a text.
   */
  db::Font text_font () const
  {
    if (m_type == Text) {
      return m_with_props ? m_generic.ptext->font () : m_generic.text->font ();
    } else {
      return text_ref ().obj ().font ();
    }
  }

  /** 
   *  @brief Return a box if one is referenced
   */
  box_type box () const
  {
    tl_assert (m_type == Box || m_type == ShortBox || m_type == BoxArrayMember || m_type == ShortBoxArrayMember);
    if (m_type == Box) {
      return m_with_props ? *m_generic.pbox : *m_generic.box;
    } else if (m_type == ShortBox) {
      return m_with_props ? *m_generic.pshort_box : *m_generic.short_box;
    } else if (m_type == BoxArrayMember) {
      return m_trans * (m_with_props ? m_generic.pbox_array->object () : m_generic.box_array->object ());
    } else {
      return m_trans * box_type (m_with_props ? m_generic.pshort_box_array->object () : m_generic.short_box_array->object ());
    }
  }

  /** 
   *  @brief Test if the shape proxy points to a box
   */
  bool is_box () const
  {
    return (m_type == Box || m_type == ShortBox || m_type == BoxArrayMember || m_type == ShortBoxArrayMember);
  }

  /** 
   *  @brief Instantiate the box object
   * 
   *  If a box is referenced, this object is instantiated
   *  by this method.
   */
  void box (box_type &b) const
  {
    b = box ();
  }

  /** 
   *  @brief Return a reference to the user object if one is referenced
   */
  const user_object_type &user_object () const
  {
    tl_assert (m_type == UserObject);
    return m_with_props ? *m_generic.puser_object : *m_generic.user_object;
  }

  /** 
   *  @brief Test if the shape proxy points to a user object
   */
  bool is_user_object () const
  {
    return (m_type == UserObject);
  }

  /** 
   *  @brief Instantiate the user object
   * 
   *  If a user object is referenced, this object is instantiated
   *  by this method.
   */
  void user_object (user_object_type &u) const
  {
    tl_assert (m_type == UserObject);
    u = m_with_props ? *m_generic.puser_object : *m_generic.user_object;
  }

  /** 
   *  @brief Compute the bounding box of the shape the is referenced
   */
  box_type bbox () const
  {
    switch (m_type) {
    case Null:
      return box_type ();
    case Polygon:
      return m_with_props ? m_generic.ppolygon->box () : m_generic.polygon->box ();
    case PolygonRef:
    case PolygonPtrArrayMember:
      return polygon_ref ().box ();
    case PolygonPtrArray:
      if (m_with_props) {
        return m_generic.ppolygon_aref->bbox (db::box_convert<polygon_ptr_type> ());
      } else {
        return m_generic.polygon_aref->bbox (db::box_convert<polygon_ptr_type> ());
      }
    case SimplePolygon:
      return m_generic.simple_polygon->box ();
    case SimplePolygonRef:
    case SimplePolygonPtrArrayMember:
      return simple_polygon_ref ().box ();
    case SimplePolygonPtrArray:
      if (m_with_props) {
        return m_generic.psimple_polygon_aref->bbox (db::box_convert<simple_polygon_ptr_type> ());
      } else {
        return m_generic.simple_polygon_aref->bbox (db::box_convert<simple_polygon_ptr_type> ());
      }
    case Text:
      return m_with_props ? m_generic.ptext->box () : m_generic.text->box ();
    case TextRef:
    case TextPtrArrayMember:
      return text_ref ().box ();
    case TextPtrArray:
      if (m_with_props) {
        return m_generic.ptext_aref->bbox (db::box_convert<text_ptr_type> ());
      } else {
        return m_generic.text_aref->bbox (db::box_convert<text_ptr_type> ());
      }
    case Edge:
      if (m_with_props) {
        return box_type (m_generic.pedge->p1 (), m_generic.edge->p2 ());
      } else {
        return box_type (m_generic.edge->p1 (), m_generic.edge->p2 ());
      }
    case Path:
      return m_with_props ? m_generic.ppath->box () : m_generic.path->box ();
    case PathRef:
    case PathPtrArrayMember:
      return path_ref ().box ();
    case PathPtrArray:
      if (m_with_props) {
        return m_generic.ppath_aref->bbox (db::box_convert<path_ptr_type> ());
      } else {
        return m_generic.path_aref->bbox (db::box_convert<path_ptr_type> ());
      }
    case BoxArray:
      if (m_with_props) {
        return m_generic.pbox_array->bbox (db::box_convert<box_type> ());
      } else {
        return m_generic.box_array->bbox (db::box_convert<box_type> ());
      }
    case ShortBoxArray:
      if (m_with_props) {
        return m_generic.pshort_box_array->bbox (db::box_convert<box_type> ());
      } else {
        return m_generic.short_box_array->bbox (db::box_convert<box_type> ());
      }
    case Box:
    case ShortBox:
    case BoxArrayMember:
    case ShortBoxArrayMember:
      return box ();
    case UserObject:
      return m_with_props ? m_generic.puser_object->box () : m_generic.user_object->box ();
    default:
      return box_type ();
    }
  }

  /**
   *  @brief Attribute that tells if the shape is member of an array
   *
   *  In this case the shape is not a freestanding shape but one instance of an
   *  array shape.
   */
  bool is_array_member () const
  {
    return m_type == PolygonPtrArrayMember ||
           m_type == SimplePolygonPtrArrayMember ||
           m_type == PathPtrArrayMember ||
           m_type == BoxArrayMember ||
           m_type == ShortBoxArrayMember ||
           m_type == TextPtrArrayMember;
  }

  /**
   *  @brief Array instance member transformation
   *
   *  This attribute is valid only if is_array_member is true.
   *  The transformation returned describes the relative transformation of the 
   *  array member addressed.
   */
  const trans_type &array_trans () const
  {
    return m_trans;
  }

  /**
   *  @brief inequality 
   */
  bool operator!= (const shape &d) const
  {
    return ! operator== (d);
  }

  /**
   *  @brief equality 
   *
   *  Equality of shapes is not specified by the identity of the objects but by the
   *  identity of the pointers - both shapes must reference the same object.
   */
  bool operator== (const shape &d) const
  {
    if (m_type != d.m_type) {
      return false;
    }
    if (m_generic.any != d.m_generic.any) {
      return false;
    }
    return m_trans == d.m_trans;
  }

  /**
   *  @brief less-than operator 
   *
   *  Ordering of shapes is not specified by the identity of the objects but by the
   *  order of the pointers.
   *  Since pointers are volatile objects, the ordering is not strictly reproducable!
   */
  bool operator< (const shape &d) const
  {
    if (m_type != d.m_type) {
      return m_type < d.m_type;
    }
    if (m_generic.any != d.m_generic.any) {
      return m_generic.any < d.m_generic.any;
    }
    return m_trans < d.m_trans;
  }

  size_t mem_used () const
  {
    return sizeof (*this);
  }

  size_t mem_reqd () const
  {
    return sizeof (*this);
  }

public:
  template <class Coord> friend class db::shapes;
 
  union generic {

    const polygon_type *polygon;
    const polygon_ref_type *polygon_ref;
    const polygon_ptr_array_type *polygon_aref;
    const simple_polygon_type *simple_polygon;
    const simple_polygon_ref_type *simple_polygon_ref;
    const simple_polygon_ptr_array_type *simple_polygon_aref;
    const text_type *text;
    const text_ref_type *text_ref;
    const text_ptr_array_type *text_aref;
    const edge_type *edge;
    const path_type *path;
    const path_ref_type *path_ref;
    const path_ptr_array_type *path_aref;
    const box_type *box;
    const box_array_type *box_array;
    const short_box_type *short_box;
    const short_box_array_type *short_box_array;
    const user_object_type *user_object;

    const db::object_with_properties<polygon_type> *ppolygon;
    const db::object_with_properties<polygon_ref_type> *ppolygon_ref;
    const db::object_with_properties<polygon_ptr_array_type> *ppolygon_aref;
    const db::object_with_properties<simple_polygon_type> *psimple_polygon;
    const db::object_with_properties<simple_polygon_ref_type> *psimple_polygon_ref;
    const db::object_with_properties<simple_polygon_ptr_array_type> *psimple_polygon_aref;
    const db::object_with_properties<text_type> *ptext;
    const db::object_with_properties<text_ref_type> *ptext_ref;
    const db::object_with_properties<text_ptr_array_type> *ptext_aref;
    const db::object_with_properties<edge_type> *pedge;
    const db::object_with_properties<path_type> *ppath;
    const db::object_with_properties<path_ref_type> *ppath_ref;
    const db::object_with_properties<path_ptr_array_type> *ppath_aref;
    const db::object_with_properties<box_type> *pbox;
    const db::object_with_properties<box_array_type> *pbox_array;
    const db::object_with_properties<short_box_type> *pshort_box;
    const db::object_with_properties<short_box_array_type> *pshort_box_array;
    const db::object_with_properties<user_object_type> *puser_object;

    const void *any;

  };

  generic m_generic;
  trans_type m_trans;
  bool m_with_props : 8;
  object_type m_type : 24;
};

template <class C>
size_t mem_used (const shape<C> &s)
{
  return s.mem_used (); 
}

template <class C>
size_t mem_reqd (const shape<C> &s)
{
  return s.mem_reqd (); 
}


/** 
 *  @brief A typedef for a standard shape
 */
typedef db::shape<db::Coord> Shape;


/**
 *  @brief A generic shape iterator
 *
 *  This iterator can iterate any kind of shape from a "shapes" container.
 *  It allows to select certain kind of shapes. The dereferecing operator
 *  returns a shape proxy object that can be used to access the actual shape.
 *  It can iterator over all shapes and over a region selected.
 *  If the iterator is constructed, it always points to the beginning of the 
 *  list of shapes in the container or the first shape selected by the region.
 *  The iterator is incremented as usual with the ++ operator. The end of the
 *  sequence is tested with the at_end method.
 */
template <class C>
class shape_iterator
{
public:
  typedef C coord_type;
  typedef db::polygon<C> polygon_type;
  typedef db::simple_polygon<C> simple_polygon_type;
  typedef db::disp_trans<C> disp_type;
  typedef db::polygon_ref<polygon_type, disp_type> polygon_ref_type;
  typedef db::polygon_ref<simple_polygon_type, disp_type> simple_polygon_ref_type;
  typedef db::polygon_ref<polygon_type, unit_trans> polygon_ptr_type;
  typedef db::array<polygon_ptr_type, disp_type> polygon_ptr_array_type;
  typedef typename polygon_ptr_array_type::iterator polygon_ptr_array_iterator_type;
  typedef db::polygon_ref<simple_polygon_type, unit_trans> simple_polygon_ptr_type;
  typedef db::array<simple_polygon_ptr_type, disp_type> simple_polygon_ptr_array_type;
  typedef typename simple_polygon_ptr_array_type::iterator simple_polygon_ptr_array_iterator_type;
  typedef db::path<C> path_type;
  typedef db::path_ref<path_type, disp_type> path_ref_type;
  typedef db::path_ref<path_type, unit_trans> path_ptr_type;
  typedef db::array<path_ptr_type, disp_type> path_ptr_array_type;
  typedef typename path_ptr_array_type::iterator path_ptr_array_iterator_type;
  typedef db::edge<C> edge_type;
  typedef db::text<C> text_type;
  typedef db::text_ref<text_type, disp_type> text_ref_type;
  typedef db::text_ref<text_type, unit_trans> text_ptr_type;
  typedef db::array<text_ptr_type, disp_type> text_ptr_array_type;
  typedef typename text_ptr_array_type::iterator text_ptr_array_iterator_type;
  typedef db::box<C> box_type;
  typedef db::array<box_type, unit_trans> box_array_type;
  typedef typename box_array_type::iterator box_array_iterator_type;
  typedef db::box<C, short> short_box_type;
  typedef db::array<short_box_type, unit_trans> short_box_array_type;
  typedef typename short_box_array_type::iterator short_box_array_iterator_type;
  typedef db::point<C> point_type;
  typedef db::user_object<C> user_object_type;
  typedef db::shape<C> shape_type;
  typedef db::shapes<C> shapes_type;
  typedef db::PropertiesRepository::properties_id_type properties_id_type;
  typedef std::set<properties_id_type> property_selector;
  typedef shape_type value_type;
  typedef const shape_type *pointer; 
  typedef shape_type reference;   //  operator* returns a value
  typedef std::forward_iterator_tag iterator_category;
  typedef void difference_type;

  enum region_mode
  {
    None,
    Overlapping,
    Touching
  };

  enum object_type 
  {
    Polygon                = 0,
    PolygonRef             = 1,
    PolygonPtrArray        = 2,
    SimplePolygon          = 3,
    SimplePolygonRef       = 4,
    SimplePolygonPtrArray  = 5,
    Edge                   = 6,
    Path                   = 7,
    PathRef                = 8,
    PathPtrArray           = 9,
    Box                    = 10,
    BoxArray               = 11,
    ShortBox               = 12,
    ShortBoxArray          = 13,
    Text                   = 14,
    TextRef                = 15,
    TextPtrArray           = 16,
    UserObject             = 17,
    Null                   = 18 //  must be last!
  };

  enum flags_type 
  {
    //  These flags are a combination of all bits representing the
    //  object type (Bit 0 for Polygon and so on)
    Polygons          = (1 << Polygon) 
                      | (1 << PolygonRef) 
                      | (1 << PolygonPtrArray) 
                      | (1 << SimplePolygon) 
                      | (1 << SimplePolygonRef) 
                      | (1 << SimplePolygonPtrArray),
    Edges             = (1 << Edge),
    Paths             = (1 << Path) 
                      | (1 << PathRef) 
                      | (1 << PathPtrArray),
    Boxes             = (1 << Box)
                      | (1 << BoxArray)
                      | (1 << ShortBox)
                      | (1 << ShortBoxArray),
    Texts             = (1 << Text) 
                      | (1 << TextRef) 
                      | (1 << TextPtrArray),
    UserObjects       = (1 << UserObject),
    Properties        = (1 << Null),
    All               = (1 << Null) - 1,
    AllWithProperties = (1 << (Null + 1)) - 1,
    Nothing           = 0
  };
    
  /**
   *  @brief The default constructor of the iterator
   */
  shape_iterator ()
    : m_region_mode (None), m_type (Null), m_box (), m_flags (0), 
      mp_shapes (0), mp_prop_sel (0), m_inv_prop_sel (false), m_array_iterator_valid (false)
  {
    m_valid = false;
    m_with_props = false;
  }

  /**
   *  @brief The copy constructor of the iterator
   */
  shape_iterator (const shape_iterator<C> &d)
    : m_region_mode (None), m_type (Null), m_box (), m_flags (0), 
      mp_shapes (0), mp_prop_sel (0), m_inv_prop_sel (false), m_array_iterator_valid (false)
  {
    m_valid = false;
    m_with_props = false;
    operator= (d);
  }

  /**
   *  @brief The normal iterator
   *
   *  This constructor creates an iterator that iterates over all shapes of
   *  the selected kind. The kind of shape is selected by the "flags" argument
   *  which is an or-ed sum of flags_type constants.
   *
   *  @param shapes The shapes container to iterate over
   *  @param flags The kind of shapes to iterate over (or-ed constants of flags_type)
   */
  shape_iterator (const shapes_type &shapes, unsigned int flags = All, const property_selector *prop_sel = 0, bool inv_prop_sel = false)
    : m_region_mode (None), m_type (object_type (0)), m_box (), m_flags (flags), 
      mp_shapes (&shapes), mp_prop_sel (prop_sel), m_inv_prop_sel (inv_prop_sel), m_array_iterator_valid (false)
  {
    //  optimize: empty property selection plus inverse = no property selection at all
    //            any property selection and not inverse = only shapes with properties
    if (mp_prop_sel) {
      if (mp_prop_sel->empty () && m_inv_prop_sel) {
        mp_prop_sel = 0;
        m_inv_prop_sel = false;
      } else if (! m_inv_prop_sel) {
        m_flags |= Properties;
      }
    }
    m_valid = false;
    m_with_props = false;
    inc ();
  }

  /**
   *  @brief The region iterator
   *
   *  This constructor creates an iterator that iterates over all shapes of
   *  the selected kind that interact with the given region with their bounding box. 
   *  The kind of shape is selected by the "flags" argument which is an or-ed sum 
   *  of flags_type constants. The region_mode argument can be either Overlapping
   *  or Touching. In the former case all shapes whose bounding box overlaps with
   *  the given region are selected. In the latter case all shapes whose bounding
   *  box touches the given region are selected.
   *
   *  @param shapes The shapes container to iterate over
   *  @param box The region to select
   *  @param mode See above
   *  @param flags The kind of shapes to iterate over (or-ed constants of flags_type)
   */
  shape_iterator (const shapes_type &shapes, const box_type &box, region_mode mode, unsigned int flags = All, const property_selector *prop_sel = 0, bool inv_prop_sel = false)
    : m_region_mode (mode), m_type (object_type (0)), m_box (box), m_flags (flags), 
      mp_shapes (&shapes), mp_prop_sel (prop_sel), m_inv_prop_sel (inv_prop_sel), m_array_iterator_valid (false)
  {
    //  optimize: empty property selection plus inverse = no property selection at all
    //            any property selection and not inverse = only shapes with properties
    if (mp_prop_sel) {
      if (mp_prop_sel->empty () && m_inv_prop_sel) {
        mp_prop_sel = 0;
        m_inv_prop_sel = false;
      } else if (! m_inv_prop_sel) {
        m_flags |= Properties;
      }
    }
    m_valid = false;
    m_with_props = false;
    inc ();
  }

  /**
   *  @brief The destructor
   */
  ~shape_iterator ()
  {
    cleanup ();
  }

  /**
   *  @brief Assignment
   */
  shape_iterator &operator= (const shape_iterator<C> &d)
  {
    if (&d != this) {

      cleanup ();

      m_d = d.m_d;
      m_valid = d.m_valid;
      m_with_props = d.m_with_props;
      m_region_mode = d.m_region_mode;
      m_type = d.m_type;
      m_box = d.m_box;
      m_shape = d.m_shape;
      m_array = d.m_array;
      m_flags = d.m_flags;
      mp_shapes = d.mp_shapes;
      mp_prop_sel = d.mp_prop_sel;
      m_inv_prop_sel = d.m_inv_prop_sel;
      m_array_iterator_valid = d.m_array_iterator_valid;

      if (m_type != Null) {

        if (m_array_iterator_valid) {

          if (m_type == PolygonPtrArray) {

            polygon_ptr_array_iterator_type *d_arr_iter = (polygon_ptr_array_iterator_type *) d.m_ad.iter;
            polygon_ptr_array_iterator_type *arr_iter = (polygon_ptr_array_iterator_type *) m_ad.iter;
            new (arr_iter) polygon_ptr_array_iterator_type (*d_arr_iter);

          } else if (m_type == SimplePolygonPtrArray) {

            simple_polygon_ptr_array_iterator_type *d_arr_iter = (simple_polygon_ptr_array_iterator_type *) d.m_ad.iter;
            simple_polygon_ptr_array_iterator_type *arr_iter = (simple_polygon_ptr_array_iterator_type *) m_ad.iter;
            new (arr_iter) simple_polygon_ptr_array_iterator_type (*d_arr_iter);

          } else if (m_type == PathPtrArray) {

            path_ptr_array_iterator_type *d_arr_iter = (path_ptr_array_iterator_type *) d.m_ad.iter;
            path_ptr_array_iterator_type *arr_iter = (path_ptr_array_iterator_type *) m_ad.iter;
            new (arr_iter) path_ptr_array_iterator_type (*d_arr_iter);

          } else if (m_type == TextPtrArray) {

            text_ptr_array_iterator_type *d_arr_iter = (text_ptr_array_iterator_type *) d.m_ad.iter;
            text_ptr_array_iterator_type *arr_iter = (text_ptr_array_iterator_type *) m_ad.iter;
            new (arr_iter) text_ptr_array_iterator_type (*d_arr_iter);

          } else if (m_type == BoxArray) {

            box_array_iterator_type *d_arr_iter = (box_array_iterator_type *) d.m_ad.iter;
            box_array_iterator_type *arr_iter = (box_array_iterator_type *) m_ad.iter;
            new (arr_iter) box_array_iterator_type (*d_arr_iter);

          } else if (m_type == ShortBoxArray) {

            short_box_array_iterator_type *d_arr_iter = (short_box_array_iterator_type *) d.m_ad.iter;
            short_box_array_iterator_type *arr_iter = (short_box_array_iterator_type *) m_ad.iter;
            new (arr_iter) short_box_array_iterator_type (*d_arr_iter);

          }

        }

      }

    }
    return *this;
  }

  /** 
   *  @brief Access the object 
   * 
   *  This method delivers a proxy object that can be copied but still points
   *  to the original objects. This must be taken under consideration if the
   *  container that is iterated over is destroyed: the shape object will then
   *  become invalid instantly.
   *  Hint: for automation purposes, this function must return a copy, not a reference!
   *  For a reference we have operator->
   */
  shape_type operator* () const
  {
    return m_shape;
  }

  /** 
   *  @brief Access the object (see operator*)
   */
  const shape_type *operator-> () const
  {
    return &m_shape;
  }

  /**
   *  @brief Increment the iterator
   */
  shape_iterator &operator++ () 
  {
    inc ();
    return *this;
  }

  /**
   *  @brief Test if the iterator is at the end
   *
   *  @return true if there are no more elements
   */
  bool at_end () const
  {
    return m_type == Null;
  }

private:
  //  a helper union for the iter_size union 
  //  (basically computing the size required for all iterators for a certain shape/array type)
  template <class Shape>
  class per_shape_iter_size {
    union {
      char sz_n  [sizeof (typename db::layer<Shape>::iterator)];
      char sz_np [sizeof (typename db::layer< db::object_with_properties<Shape> >::iterator)];
      char sz_t  [sizeof (typename db::layer<Shape>::touching_iterator)];
      char sz_tp [sizeof (typename db::layer< db::object_with_properties<Shape> >::touching_iterator)];
      char sz_o  [sizeof (typename db::layer<Shape>::overlapping_iterator)];
      char sz_op [sizeof (typename db::layer< db::object_with_properties<Shape> >::overlapping_iterator)];
    } all_iterators;
  };
  
  //  this union is simply there to determine the maximum size required for all
  //  possible iterators.
  union iter_size {
    char sz1  [sizeof (per_shape_iter_size <polygon_type>)];
    char sz2  [sizeof (per_shape_iter_size <polygon_ref_type>)];
    char sz3  [sizeof (per_shape_iter_size <polygon_ptr_array_type>)];
    char sz4  [sizeof (per_shape_iter_size <simple_polygon_type>)];
    char sz5  [sizeof (per_shape_iter_size <simple_polygon_ref_type>)];
    char sz6  [sizeof (per_shape_iter_size <simple_polygon_ptr_array_type>)];
    char sz7  [sizeof (per_shape_iter_size <path_type>)];
    char sz8  [sizeof (per_shape_iter_size <path_ref_type>)];
    char sz9  [sizeof (per_shape_iter_size <path_ptr_array_type>)];
    char sz10 [sizeof (per_shape_iter_size <edge_type>)];
    char sz11 [sizeof (per_shape_iter_size <box_type>)];
    char sz12 [sizeof (per_shape_iter_size <box_array_type>)];
    char sz13 [sizeof (per_shape_iter_size <short_box_type>)];
    char sz14 [sizeof (per_shape_iter_size <short_box_array_type>)];
    char sz15 [sizeof (per_shape_iter_size <text_type>)];
    char sz16 [sizeof (per_shape_iter_size <text_ref_type>)];
    char sz17 [sizeof (per_shape_iter_size <text_ptr_array_type>)];
    char sz18 [sizeof (per_shape_iter_size <user_object_type>)];
  };

  //  this union is simply there to determine the maximum size required for all
  //  array iterators.
  union array_iter_size {
    char ai1  [sizeof (polygon_ptr_array_iterator_type)];
    char ai2  [sizeof (simple_polygon_ptr_array_iterator_type)];
    char ai3  [sizeof (path_ptr_array_iterator_type)];
    char ai4  [sizeof (text_ptr_array_iterator_type)];
    char ai5  [sizeof (box_array_iterator_type)];
    char ai6  [sizeof (short_box_array_iterator_type)];
  };

  //  This member must be first to guarantee alignment on 64bit systems:
  //  The strange construction and the local dummy class helps to guarantee alignment of the "iter" space
  union {
    class _align_helper { long l; } _ah;
    char iter [sizeof (iter_size)];
  } m_d;

  //  This member must be first to guarantee alignment on 64bit systems:
  //  The strange construction and the local dummy class helps to guarantee alignment of the "iter" space
  union {
    class _align_helper { long l; } _ah;
    char iter [sizeof (array_iter_size)];
  } m_ad;
  
  bool m_valid, m_with_props;
  region_mode m_region_mode;
  object_type m_type;
  box_type m_box;
  shape_type m_shape;
  shape_type m_array;
  unsigned int m_flags;
  const shapes_type *mp_shapes;
  const property_selector *mp_prop_sel;
  bool m_inv_prop_sel;
  bool m_array_iterator_valid;

  template <class Iter> 
  void skip_array_iter ()
  {
    Iter *arr_iter = (Iter *) m_ad.iter;
    arr_iter->~Iter ();
  }

  void skip_array ()
  {
    if (m_array_iterator_valid) {
      if (m_type == PolygonPtrArray) {
        skip_array_iter<polygon_ptr_array_iterator_type> ();
      } else if (m_type == SimplePolygonPtrArray) {
        skip_array_iter<simple_polygon_ptr_array_iterator_type> ();
      } else if (m_type == PathPtrArray) {
        skip_array_iter<path_ptr_array_iterator_type> ();
      } else if (m_type == TextPtrArray) {
        skip_array_iter<text_ptr_array_iterator_type> ();
      } else if (m_type == BoxArray) {
        skip_array_iter<box_array_iterator_type> ();
      } else if (m_type == ShortBoxArray) {
        skip_array_iter<short_box_array_iterator_type> ();
      }
      m_array_iterator_valid = false;
    }
  }

  void cleanup ()
  {
    //  this trick destroys all iterators that have been allocated in the generic union
    if (m_type != Null) {

      skip_array ();
      m_flags = 0;
      inc ();
      tl_assert (m_type == Null);

    }
  }

  struct NoRegionTag { };
  struct TouchingRegionTag { };
  struct OverlappingRegionTag { };

  template <class Sh>
  bool advance (NoRegionTag)
  {
    bool sel = (m_flags & (1 << (unsigned int) m_type)) != 0;
    bool props_only = (m_flags & Properties) != 0;

    if (! m_with_props && ! props_only) {
    
      typename db::layer<Sh>::iterator *iter = (typename db::layer<Sh>::iterator *) m_d.iter;

      if (!m_valid && sel) {
        typename db::layer<Sh>::iterator i (mp_shapes->begin (typename Sh::tag ()));
        if (i != mp_shapes->end (typename Sh::tag ())) {
          new (iter) typename db::layer<Sh>::iterator (i);
          m_valid = true;
        }
      } 

      if (m_valid) {
        if (!sel || *iter == mp_shapes->end (typename Sh::tag ())) {
          m_valid = false;
        } else {
          m_shape = **iter;
          ++*iter;
          return true;
        }
      }

      m_with_props = true;

    }

    if (m_with_props || props_only) {
    
      typename db::layer< db::object_with_properties<Sh> >::iterator *iter = (typename db::layer< db::object_with_properties<Sh> >::iterator *) m_d.iter;
      typename db::layer< db::object_with_properties<Sh> >::iterator iend (mp_shapes->end (typename db::object_with_properties <Sh>::tag ()));

      if (!m_valid && sel) {
        typename db::layer< db::object_with_properties<Sh> >::iterator i (mp_shapes->begin (typename db::object_with_properties <Sh>::tag ()));
        if (mp_prop_sel) {
          while (i != iend && (mp_prop_sel->find (i->properties_id ()) == mp_prop_sel->end ()) != m_inv_prop_sel) {
            ++i;
          }
        }
        if (i != iend) {
          new (iter) typename db::layer< db::object_with_properties<Sh> >::iterator (i);
          m_valid = true;
        }
      } 

      if (m_valid) {
        if (!sel || *iter == iend) {
          m_valid = false;
        } else {
          m_shape = **iter;
          do {
            ++*iter;
          } while (mp_prop_sel && *iter != iend && (mp_prop_sel->find ((*iter)->properties_id ()) == mp_prop_sel->end ()) != m_inv_prop_sel);
          return true;
        }
      }

    }

    m_with_props = false;
    return false;
  }

  template <class Sh>
  bool advance (TouchingRegionTag)
  {
    bool sel = (m_flags & (1 << (unsigned int) m_type)) != 0;
    bool props_only = (m_flags & Properties) != 0;

    if (! m_with_props && ! props_only) {

      typename db::layer<Sh>::touching_iterator *iter = (typename db::layer<Sh>::touching_iterator *) m_d.iter;

      if (!m_valid && sel) {
        typename db::layer<Sh>::touching_iterator i (mp_shapes->begin_touching (m_box, typename Sh::tag ()));
        if (!i.at_end ()) {
          new (iter) typename db::layer<Sh>::touching_iterator (i);
          m_valid = true;
        }
      } 
      if (m_valid) {
        if (!sel || iter->at_end ()) {
          m_valid = false;
        } else {
          m_shape = **iter;
          ++*iter;
          return true;
        }
      }

      m_with_props = true;

    }

    if (m_with_props || props_only) {

      typename db::layer< db::object_with_properties<Sh> >::touching_iterator *iter = (typename db::layer< db::object_with_properties<Sh> >::touching_iterator *) m_d.iter;

      if (!m_valid && sel) {
        typename db::layer< db::object_with_properties<Sh> >::touching_iterator i (mp_shapes->begin_touching (m_box, typename  db::object_with_properties<Sh> ::tag ()));
        if (mp_prop_sel) {
          while (!i.at_end () && (mp_prop_sel->find (i->properties_id ()) == mp_prop_sel->end ()) != m_inv_prop_sel) {
            ++i;
          }
        }
        if (!i.at_end ()) {
          new (iter) typename db::layer< db::object_with_properties<Sh> >::touching_iterator (i);
          m_valid = true;
        }
      } 
      if (m_valid) {
        if (!sel || iter->at_end ()) {
          m_valid = false;
          m_with_props = false;
        } else {
          m_shape = **iter;
          do {
            ++*iter;
          } while (mp_prop_sel && !iter->at_end () && (mp_prop_sel->find ((*iter)->properties_id ()) == mp_prop_sel->end ()) != m_inv_prop_sel);
          return true;
        }
      }

    }

    m_with_props = false;
    return false;
  }

  template <class Sh>
  bool advance (OverlappingRegionTag)
  {
    bool sel = (m_flags & (1 << (unsigned int) m_type)) != 0;
    bool props_only = (m_flags & Properties) != 0;

    if (! m_with_props && ! props_only) {

      typename db::layer<Sh>::overlapping_iterator *iter = (typename db::layer<Sh>::overlapping_iterator *) m_d.iter;

      if (!m_valid && sel) {
        typename db::layer<Sh>::overlapping_iterator i (mp_shapes->begin_overlapping (m_box, typename Sh::tag ()));
        if (!i.at_end ()) {
          new (iter) typename db::layer<Sh>::overlapping_iterator (i);
          m_valid = true;
        }
      } 
      if (m_valid) {
        if (!sel || iter->at_end ()) {
          m_valid = false;
        } else {
          m_shape = **iter;
          ++*iter;
          return true;
        }
      }

      m_with_props = true;

    }

    if (m_with_props || props_only) {

      typename db::layer< db::object_with_properties<Sh> >::overlapping_iterator *iter = (typename db::layer< db::object_with_properties<Sh> >::overlapping_iterator *) m_d.iter;

      if (!m_valid && sel) {
        typename db::layer< db::object_with_properties<Sh> >::overlapping_iterator i (mp_shapes->begin_overlapping (m_box, typename  db::object_with_properties<Sh> ::tag ()));
        if (mp_prop_sel) {
          while (!i.at_end () && (mp_prop_sel->find (i->properties_id ()) == mp_prop_sel->end ()) != m_inv_prop_sel) {
            ++i;
          }
        }
        if (!i.at_end ()) {
          new (iter) typename db::layer< db::object_with_properties<Sh> >::overlapping_iterator (i);
          m_valid = true;
        }
      } 
      if (m_valid) {
        if (!sel || iter->at_end ()) {
          m_valid = false;
        } else {
          m_shape = **iter;
          do {
            ++*iter;
          } while (mp_prop_sel && !iter->at_end () && (mp_prop_sel->find ((*iter)->properties_id ()) == mp_prop_sel->end ()) != m_inv_prop_sel);
          return true;
        }
      }

    }

    m_with_props = false;
    return false;
  }

  template <class Array>
  void init_array_iter (NoRegionTag) 
  {
    typedef typename Array::iterator array_iterator;

    array_iterator *arr_iter = (array_iterator *) m_ad.iter;
    if (m_with_props) { 
      new (arr_iter) array_iterator (((db::object_with_properties<Array> *) m_array.basic_ptr ())->begin ());
    } else {
      new (arr_iter) array_iterator (((Array *) m_array.basic_ptr ())->begin ());
    }
  }

  template <class Array>
  void init_array_iter (TouchingRegionTag) 
  {
    typedef typename Array::iterator array_iterator;
    typedef typename Array::object_type shape_ptr_type;

    array_iterator *arr_iter = (array_iterator *) m_ad.iter;
    db::box_convert<shape_ptr_type> bc;
    if (m_with_props) { 
      new (arr_iter) array_iterator (((db::object_with_properties<Array> *) m_array.basic_ptr ())->begin_touching (m_box, bc));
    } else {
      new (arr_iter) array_iterator (((Array *) m_array.basic_ptr ())->begin_touching (m_box, bc));
    }
  }

  template <class Array>
  void init_array_iter (OverlappingRegionTag) 
  {
    typedef typename Array::iterator array_iterator;
    typedef typename Array::object_type shape_ptr_type;

    array_iterator *arr_iter = (array_iterator *) m_ad.iter;
    db::box_convert<shape_ptr_type> bc;
    box_type box (m_box);
    box.enlarge (point_type (-1, -1));
    if (m_with_props) { 
      new (arr_iter) array_iterator (((db::object_with_properties<Array> *) m_array.basic_ptr ())->begin_touching (box, bc));
    } else {
      new (arr_iter) array_iterator (((Array *) m_array.basic_ptr ())->begin_touching (box, bc));
    }
  }

  template <class Array, class RegionTag>
  bool advance_aref ()
  {
    typedef typename Array::iterator array_iterator;
    typedef Array array_type;
    typedef typename db::layer<Array>::iterator layer_iterator;

    while (true) {

      if (m_array_iterator_valid) {
        array_iterator *arr_iter = (array_iterator *) m_ad.iter;
        if (! arr_iter->at_end ()) {
          break;
        } else {
          arr_iter->~array_iterator ();
          m_array_iterator_valid = false;
        }
      }

      RegionTag region_tag = RegionTag ();
      if (! advance<Array> (region_tag)) {
        return false;
      }

      m_array = m_shape;
      init_array_iter <Array> (region_tag);
      m_array_iterator_valid = true;

    }

    array_iterator *arr_iter = (array_iterator *) m_ad.iter;
    typename array_iterator::result_type t = **arr_iter;
    ++*arr_iter;

    //  HINT: since the array references store "pointers" without an intrinsic 
    //        transformation, we can drop this:
    //  t = t * (*iter)->obj ().trans (); 
 
    //  This creates a local reference object to reference an array member
    if (m_with_props) { 
      m_shape = shape_type (*((db::object_with_properties<Array> *) m_array.basic_ptr ()), t);
    } else {
      m_shape = shape_type (*((Array *) m_array.basic_ptr ()), t);
    }

    return true;
  }

  template <class RegionTag>
  void advance_generic ()
  {
    RegionTag region_tag;

    switch (m_type) {
    //  The order of the switch statements must be identical to the order of
    //  enum constants
    case Polygon:
      if (advance<polygon_type> (region_tag)) return;
      m_type = object_type ((unsigned int) m_type + 1);
    case PolygonRef:
      if (advance<polygon_ref_type> (region_tag)) return;
      m_type = object_type ((unsigned int) m_type + 1);
    case PolygonPtrArray:
      if (advance_aref<polygon_ptr_array_type, RegionTag> ()) return;
      m_type = object_type ((unsigned int) m_type + 1);
    case SimplePolygon:
      if (advance<simple_polygon_type> (region_tag)) return;
      m_type = object_type ((unsigned int) m_type + 1);
    case SimplePolygonRef:
      if (advance<simple_polygon_ref_type> (region_tag)) return;
      m_type = object_type ((unsigned int) m_type + 1);
    case SimplePolygonPtrArray:
      if (advance_aref<simple_polygon_ptr_array_type, RegionTag> ()) return;
      m_type = object_type ((unsigned int) m_type + 1);
    case Edge:
      if (advance<edge_type> (region_tag)) return;
      m_type = object_type ((unsigned int) m_type + 1);
    case Path:
      if (advance<path_type> (region_tag)) return;
      m_type = object_type ((unsigned int) m_type + 1);
    case PathRef:
      if (advance<path_ref_type> (region_tag)) return;
      m_type = object_type ((unsigned int) m_type + 1);
    case PathPtrArray:
      if (advance_aref<path_ptr_array_type, RegionTag> ()) return;
      m_type = object_type ((unsigned int) m_type + 1);
    case Box:
      if (advance<box_type> (region_tag)) return;
      m_type = object_type ((unsigned int) m_type + 1);
    case BoxArray:
      if (advance_aref<box_array_type, RegionTag> ()) return;
      m_type = object_type ((unsigned int) m_type + 1);
    case ShortBox:
      if (advance<short_box_type> (region_tag)) return;
      m_type = object_type ((unsigned int) m_type + 1);
    case ShortBoxArray:
      if (advance_aref<short_box_array_type, RegionTag> ()) return;
      m_type = object_type ((unsigned int) m_type + 1);
    case Text:
      if (advance<text_type> (region_tag)) return;
      m_type = object_type ((unsigned int) m_type + 1);
    case TextRef:
      if (advance<text_ref_type> (region_tag)) return;
      m_type = object_type ((unsigned int) m_type + 1);
    case TextPtrArray:
      if (advance_aref<text_ptr_array_type, RegionTag> ()) return;
      m_type = object_type ((unsigned int) m_type + 1);
    case UserObject:
      if (advance<user_object_type> (region_tag)) return;
      m_type = object_type ((unsigned int) m_type + 1);
    case Null:
      break;
    }
  }

  void inc ()
  {
    if (m_region_mode == None) {
      advance_generic<NoRegionTag> ();
    } else if (m_region_mode == Touching) {
      advance_generic<TouchingRegionTag> ();
    } else if (m_region_mode == Overlapping) {
      advance_generic<OverlappingRegionTag> ();
    }
  }

};

/** 
 *  @brief A typedef for a standard shape iterator
 */
typedef db::shape_iterator<db::Coord> ShapeIterator;

}  // namespace db
  
#endif

