
/*

  KLayout Layout Viewer
  Copyright (C) 2006-2007 Matthias Koefferlein

  This program is free software; you can redistribute it and/or modify
  it under the terms of the GNU General Public License as published by
  the Free Software Foundation; either version 2 of the License, or
  (at your option) any later version.

  This program is distributed in the hope that it will be useful,
  but WITHOUT ANY WARRANTY; without even the implied warranty of
  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
  GNU General Public License for more details.

  You should have received a copy of the GNU General Public License
  along with this program; if not, write to the Free Software
  Foundation, Inc., 51 Franklin St, Fifth Floor, Boston, MA  02110-1301  USA

*/


#include "dbStreamLayers.h"

#include "tlException.h"
#include "tlString.h"

#include <cctype>

namespace db
{

// ---------------------------------------------------------------
//  LayerMap

/// A helper class to join two datatype map members
struct LmapJoinOp1
{
  void operator() (unsigned int &a, unsigned int b)
  {
    //  TODO: could also do?: throw an exception ..
    a = b;
  }
};

/// A helper class to join two layer map members 
/// (this implementation basically merged the datatype maps)  
struct LmapJoinOp2
{
  void operator() (LayerMap::datatype_map &a, const LayerMap::datatype_map &b)
  {
    LmapJoinOp1 op1;
    a.add (b.begin (), b.end (), op1);
  }
};

/// Utility typedefs for the expression parser  
typedef std::pair<ld_type, ld_type> ld_interval;

/// Utility typedefs for the expression parser  
typedef std::vector<ld_interval> ld_interval_vector;


LayerMap::LayerMap ()
  : m_ld_map ()
{
  //  .. nothing yet ..
}

std::pair<bool, unsigned int> 
LayerMap::logical (const LDPair &p) const
{
  const datatype_map *dm = m_ld_map.mapped (p.layer);
  if (dm) {
    const unsigned int *l = dm->mapped (p.datatype);
    if (l) {
      return std::make_pair (true, *l);
    }
  }
  return std::make_pair (false, 0);
}

std::string
LayerMap::mapping_str (unsigned int ll) const
{
  std::string s;
  bool f1 = true;

  for (ld_map::const_iterator l = m_ld_map.begin (); l != m_ld_map.end (); ++l) {

    std::string t = tl::to_string (l->first.first);
    if (l->first.first < l->first.second - 1) {
      t += "-";
      t += tl::to_string (l->first.second - 1);
    }
    t += "/";

    bool f2 = true;
    for (datatype_map::const_iterator d = l->second.begin (); d != l->second.end (); ++d) {
      if (d->second == ll) {

        //  create a string representation
        if (!f2) {
          s += ",";
        } else {
          if (!f1) {
            s += ";";
          }
          f1 = false;
          s += t;
        }
        f2 = false;

        s += tl::to_string (d->first.first);
        if (d->first.first < d->first.second - 1) {
          s += "-";
          s += tl::to_string (d->first.second - 1);
        }

      }

    }
    
  }

  return s;
}

LDPair 
LayerMap::mapping (unsigned int ll) const
{
  for (ld_map::const_iterator l = m_ld_map.begin (); l != m_ld_map.end (); ++l) {
    for (datatype_map::const_iterator d = l->second.begin (); d != l->second.end (); ++d) {
      if (d->second == ll) {
        return LDPair (l->first.first, d->first.first);
      }
    }
  }
  return LDPair::invalid ();
}

void 
LayerMap::map (const LDPair &p, unsigned int l)
{
  insert (p, p, l);
}

void 
LayerMap::map (const LDPair &p1, const LDPair &p2, unsigned int l)
{
  insert (p1, p2, l);
}

/// Utility function for the expression parser:  
/// skip blanks in a expression
const char * 
skip_blanks (const char *&cp)
{
  while (*cp && std::isspace (*cp)) {
    ++cp;
  }
  return cp;
}

/// Utility function for the expression parser:  
/// Parse an interval
const char *
parse_interval (const char *&cp, ld_interval &p) throw (LayerSpecFormatException)
{
  ld_type n1 = 0, n2 = 0;

  skip_blanks (cp); 
  while (*cp && std::isdigit (*cp)) {
    n1 = n1 * 10 + ld_type (*cp - '0');
  }

  if (*skip_blanks (cp) == '-') {

    ++cp;

    skip_blanks (cp);
    while (*cp && std::isdigit (*cp)) {
      n2 = n2 * 10 + ld_type (*cp - '0');
    }

  } else {
    n2 = n1;
  }

  p.first = n1;
  p.second = n2;
  return cp;
}

/// Utility function for the expression parser:  
/// Parse an interval list
const char *
parse_intervals (const char *&cp, ld_interval_vector &v) throw (LayerSpecFormatException)
{
  --cp;
  do {
    ++cp;

    v.push_back (ld_interval (0, 0));
    parse_interval (cp, v.back ());

  } while (*skip_blanks (cp) == ',');

  return cp;
}

void 
LayerMap::map (const std::string &expr, unsigned int l) throw (LayerSpecFormatException)
{
  const char *cp = expr.c_str ();

  --cp;
  do {
    ++cp;

    ld_interval_vector vl, vd;

    parse_intervals (cp, vl);

    if (*skip_blanks (cp) == '/') {
      parse_intervals (cp, vd);
    } else {
      vd.push_back (ld_interval (0, 0));
    }

    datatype_map dm;
    for (ld_interval_vector::const_iterator di = vd.begin (); di != vd.end (); ++di) {
      LmapJoinOp1 op1;
      dm.add (di->first, di->second + 1, l, op1);
    }
    for (ld_interval_vector::const_iterator li = vl.begin (); li != vl.end (); ++li) {
      LmapJoinOp2 op2;
      m_ld_map.add (li->first, li->second + 1, dm, op2);
    }

  } while (*skip_blanks (cp) == ';');

  if (*cp != '\0') {
    throw LayerSpecFormatException (cp);
  }

}

void
LayerMap::insert (const LDPair &p1, const LDPair &p2, unsigned int l) 
{
  //  create a single-interval list for the datatype range
  LayerMap::datatype_map dt;
  LmapJoinOp1 op1;
  dt.add (p1.datatype, p2.datatype + 1, l, op1);

  //  add this to the layers using the special join operator that
  //  combines the datatype intervals
  LmapJoinOp2 op2;
  m_ld_map.add (p1.layer, p2.layer + 1, dt, op2);
}

}

