
/*

  KLayout Layout Viewer
  Copyright (C) 2006-2007 Matthias Koefferlein

  This program is free software; you can redistribute it and/or modify
  it under the terms of the GNU General Public License as published by
  the Free Software Foundation; either version 2 of the License, or
  (at your option) any later version.

  This program is distributed in the hope that it will be useful,
  but WITHOUT ANY WARRANTY; without even the implied warranty of
  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
  GNU General Public License for more details.

  You should have received a copy of the GNU General Public License
  along with this program; if not, write to the Free Software
  Foundation, Inc., 51 Franklin St, Fifth Floor, Boston, MA  02110-1301  USA

*/


#include "layLayoutView.h"
#include "edtMainService.h"
#include "edtService.h"
#include "edtConfig.h"
#include "edtDialogs.h"
#include "edtEditorOptionsPages.h"

namespace edt
{

// -----------------------------------------------------------------------------
//  Main Service implementation

MainService::MainService (db::Manager *manager, lay::LayoutView *view, lay::MainWindow *mw)
  : lay::Plugin (view),
    lay::Editable (view),
    db::Object (manager),
    mp_view (view),
    mp_main_window (mw),
    mp_generic_options_page (new edt::EditorOptionsGeneric ()),
    m_needs_update (true)
{ 
  //  collect the options pages and build the options dialog
  std::vector<edt::Service *> edt_services = mp_view->get_plugins <edt::Service> ();

  std::vector<edt::EditorOptionsPage *> pages;
  pages.push_back (mp_generic_options_page);

  for (std::vector<edt::Service *>::const_iterator es = edt_services.begin (); es != edt_services.end (); ++es) {
    std::vector<edt::EditorOptionsPage *> pp = (*es)->get_editor_options_pages ();
    for (std::vector<edt::EditorOptionsPage *>::iterator pi = pp.begin (); pi != pp.end (); ++pi) {
      (*pi)->activate (false);
    }
    pages.insert (pages.end (), pp.begin (), pp.end ());
  }

  mp_obj_prop_dialog = new edt::EditorOptionsPages (pages, mw);
}

MainService::~MainService ()
{
  if (mp_obj_prop_dialog) {
    delete mp_obj_prop_dialog;
    mp_obj_prop_dialog = 0;
  }
  if (mp_generic_options_page) {
    delete mp_generic_options_page;
    mp_generic_options_page = 0;
  }
}

void 
MainService::menu_activated (const std::string &symbol)
{
  if (symbol == "edt::descend") {
    cm_descend ();
  } else if (symbol == "edt::ascend") {
    cm_ascend ();
  } else if (symbol == "edt::edit_options") {
    cm_edit_options ();
  } else if (symbol == "edt::sel_change_layer") {
    cm_change_layer ();
  }
}

/**
 *  @brief A helper class to determine the common part of a set of instance elements
 */
class CommonInsts
{
public:
  CommonInsts () 
    : m_valid (true), m_first (true), m_ambigous (false), m_cv_index (0) 
  { 
    //  .. nothing yet ..  
  }

  void add (const ObjectInstPath &path, unsigned int n)
  {
    if (! m_valid) {

      //  don't do anything

    } else if (m_first) {

      m_common_inst.reserve (n);
      unsigned int nn = 0;
      for (ObjectInstPath::iterator p = path.begin (); p != path.end () && nn < n; ++nn) {
        m_common_inst.push_back (*p);
      }
      m_cv_index = path.cv_index ();
      m_first = false;

    } else if (m_cv_index != path.cv_index ()) {

      m_valid = false;

    } else {

      std::vector<db::InstElement>::iterator i1 = m_common_inst.begin ();
      ObjectInstPath::iterator i2 = path.begin ();
      while (i1 != m_common_inst.end () && i2 != path.end () && *i1 == *i2) {
        ++i1; ++i2;
      }
      if (i1 != m_common_inst.end ()) {
        m_ambigous = true;
        m_common_inst.erase (i1, m_common_inst.end ());
      }

    }
  }

  const std::vector<db::InstElement> &common_path () const
  {
    return m_common_inst;
  }

  bool valid () const
  {
    return m_valid;
  }

  bool anything () const
  {
    return ! m_first;
  }

  bool ambigous () const
  {
    return m_ambigous && m_common_inst.empty ();
  }

  bool empty () const
  {
    return m_common_inst.empty ();
  }

  unsigned int size () const
  {
    return m_common_inst.size ();
  }

  unsigned int cv_index () const
  {
    return m_cv_index;
  }

private:
  std::vector<db::InstElement> m_common_inst;
  bool m_valid;
  bool m_first;
  bool m_ambigous;
  unsigned int m_cv_index;
};


void  
MainService::cm_descend ()
{
  CommonInsts common_inst;

  std::vector<edt::Service *> edt_services = view ()->get_plugins <edt::Service> ();
  for (std::vector<edt::Service *>::const_iterator es = edt_services.begin (); es != edt_services.end () && common_inst.valid (); ++es) {
    for (edt::Service::objects::const_iterator sel = (*es)->selection ().begin (); sel != (*es)->selection ().end () && common_inst.valid (); ++sel) {
      common_inst.add (*sel, 1);
    }
  }

  //  cannot descend - we are at the lowest level already
  if (common_inst.empty ()) {
    return;
  }

  if (! common_inst.anything ()) {
    throw tl::Exception (tl::translate ("Select an object to determine into which instance to descend").c_str ());
  }
  if (! common_inst.valid () || common_inst.ambigous ()) {
    throw tl::Exception (tl::translate ("Selection is ambigous - cannot determine into which instance to descend").c_str ());
  }

  //  remove the common path and create a new set of selections

  std::vector< std::vector<ObjectInstPath> > new_selections;
  new_selections.reserve (edt_services.size ());

  for (std::vector<edt::Service *>::const_iterator es = edt_services.begin (); es != edt_services.end (); ++es) {

    new_selections.push_back (std::vector<ObjectInstPath> ());
    new_selections.back ().reserve ((*es)->selection ().size ());

    for (edt::Service::objects::const_iterator sel = (*es)->selection ().begin (); sel != (*es)->selection ().end (); ++sel) {

      new_selections.back ().push_back (*sel);
      edt::ObjectInstPath &new_sel = new_selections.back ().back ();
      new_sel.remove_front (common_inst.size ());
      //  it may happen that this way we dive into the instance selected. The resulting selection
      //  is not pointing to any instance any more and must be discarded therefore:
      if (new_sel.is_cell_inst () && new_sel.begin () == new_sel.end ()) {
        new_selections.back ().pop_back ();
      }

    }

  }

  //  this will clear the selection:
  view ()->descend (common_inst.common_path (), common_inst.cv_index ());
  view ()->set_current_cell_path (common_inst.cv_index (), view ()->cellview (common_inst.cv_index ()).combined_unspecific_path ());

  //  set the new selections
  unsigned int index = 0;
  for (std::vector<edt::Service *>::const_iterator es = edt_services.begin (); es != edt_services.end (); ++es, ++index) {
    (*es)->set_selection (new_selections [index].begin (), new_selections [index].end ());
  }

}

void  
MainService::cm_ascend ()
{
  //  add one path component and create a new set of selections

  std::vector<edt::Service *> edt_services = view ()->get_plugins <edt::Service> ();

  std::vector< std::vector<ObjectInstPath> > new_selections;
  new_selections.reserve (edt_services.size ());
  for (std::vector<edt::Service *>::const_iterator es = edt_services.begin (); es != edt_services.end (); ++es) {
    new_selections.push_back (std::vector<ObjectInstPath> ());
    new_selections.back ().insert (new_selections.back ().end (), (*es)->selection ().begin (), (*es)->selection ().end ());
  }

  //  this will clear the selection:
  for (int cv_index = 0; cv_index < int (view ()->cellviews ()); ++cv_index) {

    db::InstElement removed = view ()->ascend (cv_index);
    if (removed != db::InstElement ()) {

      db::Layout::cell_index_type new_top = view ()->cellview (cv_index).cell_index ();
      view ()->set_current_cell_path (cv_index, view ()->cellview (cv_index).combined_unspecific_path ());

      //  create and the new selections
      unsigned int index = 0;
      for (std::vector<edt::Service *>::const_iterator es = edt_services.begin (); es != edt_services.end (); ++es, ++index) {
        for (std::vector<ObjectInstPath>::iterator sel = new_selections [index].begin (); sel != new_selections [index].end (); ++sel) {
          if (int (sel->cv_index ()) == cv_index) {
            sel->insert_front (new_top, removed);
          }
        }
      }

    }

  }

  unsigned int index = 0;
  for (std::vector<edt::Service *>::const_iterator es = edt_services.begin (); es != edt_services.end (); ++es, ++index) {
    (*es)->set_selection (new_selections [index].begin (), new_selections [index].end ());
  }

}

void 
MainService::cm_change_layer ()
{
  tl_assert (view ()->is_editable ());

  int cv_index = -1;

  std::vector<edt::Service *> edt_services = view ()->get_plugins <edt::Service> ();

  unsigned int new_layer = 0;

  for (std::vector<edt::Service *>::const_iterator es = edt_services.begin (); es != edt_services.end (); ++es) {
    for (edt::Service::obj_iterator s = (*es)->selection ().begin (); s != (*es)->selection ().end (); ++s) {
      if (! s->is_cell_inst ()) {
        if (cv_index >= 0 && cv_index != int (s->cv_index ())) {
          throw tl::Exception (tl::translate ("Selections originate from different layouts - cannot switch layer in this case."));
        }
        if (cv_index < 0) {
          new_layer = s->layer ();
        }
        cv_index = int (s->cv_index ());
      }
    }
  }

  if (cv_index >= 0) {

    //  HINT: the following code is copied from ShapeEditService::get_edit_layer - it should
    //  be at some common place

    lay::LayerPropertiesConstIterator cl = view ()->current_layer ();
    if (cl.is_null ()) {
      throw tl::Exception (tl::translate ("Please select a layer first").c_str ());
    }
    
    if (cv_index != cl->cellview_index ()) {
      throw tl::Exception (tl::translate ("Shapes cannot be moved to a different layout").c_str ());
    }

    const lay::CellView &cv = view ()->cellview (cv_index);
    int layer = cl->layer_index ();

    if (cv_index < 0 || ! cv.is_valid ()) {
      throw tl::Exception (tl::translate ("Please select a cell first").c_str ());
    }

    if (layer < 0 || ! cv->layout ().is_valid_layer ((unsigned int) layer)) {

      if (cl->has_children ()) {
        throw tl::Exception (tl::translate ("Please select a valid drawing layer first").c_str ());
      } else {

        //  create this layer now
        const lay::ParsedLayerSource &source = cl->source (true /*real*/);

        db::LayerProperties db_lp;
        if (source.has_name ()) {
          db_lp.name = source.name ();
        }
        db_lp.layer = source.layer ();
        db_lp.datatype = source.datatype ();

        cv->layout ().insert_layer (db_lp);

        //  update the layer index inside the layer view
        cl->realize_source ();
          
        //  Hint: we could have taken the new index from insert_layer, but this 
        //  is a nice test:
        layer = cl->layer_index ();
        tl_assert (layer >= 0);

      }

    }

    view ()->cancel_edits ();

    manager ()->transaction (tl::translate ("Change layer"));

    db::Layout &layout = view ()->cellview (cv_index)->layout ();

    //  Collect the shapes per cell and layer first. 
    //  This is important since the selection potentially contains the same shape multiple times.
    std::map <std::pair <db::Layout::cell_index_type, unsigned int>, std::vector <db::Shape> > shapes_per_cell;
    for (std::vector<edt::Service *>::const_iterator es = edt_services.begin (); es != edt_services.end (); ++es) {
      for (edt::Service::obj_iterator s = (*es)->selection ().begin (); s != (*es)->selection ().end (); ++s) {
        if (!s->is_cell_inst () && int (s->layer ()) != layer) {
          shapes_per_cell.insert (std::make_pair (std::make_pair (s->cell_index (), s->layer ()), std::vector <db::Shape> ())).first->second.push_back (s->shape ());
        }
      }
    }

    //  Then insert and delete
    for (std::map <std::pair <db::Layout::cell_index_type, unsigned int>, std::vector <db::Shape> >::iterator spc = shapes_per_cell.begin (); spc != shapes_per_cell.end (); ++spc) {

      //  remove duplicates from the shape reference list
      std::sort (spc->second.begin (), spc->second.end ());
      std::vector <db::Shape>::iterator unique_shapes_end = std::unique (spc->second.begin (), spc->second.end ());
      spc->second.erase (unique_shapes_end, spc->second.end ());

      db::Shapes &source = layout.cell (spc->first.first).shapes (spc->first.second);
      db::Shapes &target = layout.cell (spc->first.first).shapes (layer);

      for (std::vector <db::Shape>::iterator sh = spc->second.begin (); sh != spc->second.end (); ++sh) {
        target.insert (*sh, layout.shape_repository (), tl::ident_map<db::Layout::properties_id_type> ());
      }

      //  HINT: we can erase the whole array ("true") since we ensured is_editable (thus supressing 
      //  shape arrays) in the assertion above.
      source.erase_shapes (spc->second, true);

    }
    
    //  The selection is no longer valid
    view ()->clear_selection ();

    manager ()->commit ();

  } else {
    throw tl::Exception (tl::translate ("Nothing selected to switch layers for"));
  }

}

void  
MainService::cm_edit_options ()
{
  if (! mp_obj_prop_dialog->isVisible ()) {
    mp_obj_prop_dialog->setup ();
    mp_obj_prop_dialog->show ();
  }
}

bool 
MainService::configure (const std::string &name, const std::string &value)
{
  if (name == cfg_edit_grid ||
      name == cfg_edit_move_angle_mode ||
      name == cfg_edit_connect_angle_mode ||
      name == cfg_edit_top_level_selection) {
    m_needs_update = true;
  }

  return false;  //  not taken - others need to handle this as well

}

void 
MainService::config_finalize ()
{
  if (m_needs_update) {
    m_needs_update = false;
    if (mp_obj_prop_dialog->isVisible ()) {
      mp_obj_prop_dialog->setup ();
    }
  }
}

void 
MainService::paste ()
{
  if (view ()->is_editable ()) {

    int cv_index = view ()->active_cellview_index ();

    std::vector<unsigned int> new_layers;

    //  paste the content into the active cellview.
    for (db::Clipboard::iterator c = db::Clipboard::instance ().begin (); c != db::Clipboard::instance ().end (); ++c) {
      const db::ClipboardValue<edt::ClipboardData> *value = dynamic_cast<const db::ClipboardValue<edt::ClipboardData> *> (*c);
      if (value) {

        const lay::CellView &cv = view ()->cellview (cv_index);
        if (! cv.is_valid ()) {
          throw tl::Exception (tl::translate ("No cell selected to paste something into"));
        }

        std::vector<unsigned int> nl = value->get ().insert (cv->layout (), cv.context_trans ().inverted (), &cv->layout ().cell (cv.cell_index ()));
        new_layers.insert (new_layers.end (), nl.begin (), nl.end ());

      }
    }

    //  Add new layers to the view if required.
    view ()->add_new_layers (new_layers, cv_index);
    view ()->update_content ();

  }
}

} // namespace edt


