
/*

  KLayout Layout Viewer
  Copyright (C) 2006-2007 Matthias Koefferlein

  This program is free software; you can redistribute it and/or modify
  it under the terms of the GNU General Public License as published by
  the Free Software Foundation; either version 2 of the License, or
  (at your option) any later version.

  This program is distributed in the hope that it will be useful,
  but WITHOUT ANY WARRANTY; without even the implied warranty of
  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
  GNU General Public License for more details.

  You should have received a copy of the GNU General Public License
  along with this program; if not, write to the Free Software
  Foundation, Inc., 51 Franklin St, Fifth Floor, Boston, MA  02110-1301  USA

*/


#include <QtGui/QMessageBox>

#include "dbVector.h"
#include "layLayoutView.h"
#include "laySnap.h"
#include "edtPartialService.h"
#include "edtService.h"
#include "edtConfig.h"
#include "edtDialogs.h"
#include "edtFinder.h"
#include "edtSnap.h"
#include "edtEditorOptionsPages.h"

namespace edt
{

//  max. number of tries in single-click selection before giving up
static int point_sel_tests = 10000; // @@@

//  search range for select in pixels
static unsigned int search_range = 5;   // @@@ make variable?

// -------------------------------------------------------------

/**
 *  @brief A move constraint 
 *
 *  A move constraint describes the degrees of freedom for a single point.
 *  Such a constraint can be: fixed (no freedom), unconstrained (point can move
 *  both in x and y direction) and freedom along an axis.
 */
class Constraint
{
public:
  typedef enum { Free, Fixed, OneDim } constraint_mode;

  /**
   *  @brief Construct an unconstrained constraint
   */
  Constraint ()
    : m_mode (Free)
  { }

  /**
   *  @brief Construct an "fixed" constraint
   */
  Constraint (int)
    : m_mode (Fixed)
  { }

  /**
   *  @brief Construct an "1-dimensional" constraint
   *
   *  @param axis Gives the direction in which the point can move, if 0 same as fixed constraint
   */
  Constraint (db::Vector axis)
    : m_axis (axis)
  { 
    if (axis == db::Vector ()) {
      m_mode = Fixed;
    } else {
      m_mode = OneDim;
    } 
  }

  /**
   *  @brief Type accessor
   */
  constraint_mode mode () const
  {
    return m_mode;
  }

  /**
   *  @brief Add constraints
   *
   *  Merging of constraints means to allow movement additionally in the same directions
   *  than given by the second constraint.
   */
  Constraint &operator+= (const Constraint &b)
  {
    if (m_mode == Fixed || b.mode () == Free) {
      *this = b;
    } else if (m_mode == Free || b.mode () == Fixed) {
      //  nothing to do.
    } else {
      //  must both be OneDim here.
      if (db::vprod_sign (m_axis, b.m_axis) != 0) {
        m_mode = Free;
      }
    }
    return *this;
  }

  /**
   *  @brief Unite constraints
   *
   *  Additionally impose a constraint on this movement.
   */
  Constraint &operator*= (const Constraint &b)
  {
    if (m_mode == Free || b.mode () == Fixed) {
      *this = b;
    } else if (m_mode == Fixed || b.mode () == Free) {
      //  nothing to do.
    } else {
      //  must both be OneDim here.
      if (db::vprod_sign (m_axis, b.m_axis) != 0) {
        m_mode = Fixed;
      }
    }
    return *this;
  }

  /**
   *  @brief Move a point by the given vector, given the imposed constraints
   *
   *  The movement is performed "as far as possible", i.e. projecting the axis to the
   *  move vector, not vice versa.
   */
  template <class C>
  db::point<C> move (const db::point<C> &p, const db::vector<C> &v)
  {
    if (v == db::vector<C> ()) {
      return p;
    } else if (m_mode == Free) {
      return p + v;
    } else if (m_mode == Fixed) {
      return p;
    } else {
      double proj = db::sprod (db::vector<C> (m_axis), v);
      //  do not allow movement for small angles to avoid creation of nasty spikes
      if (fabs (proj) < m_axis.double_length () * v.double_length () * 1e-2) {
        return p;
      } else {
        return p + db::vector<C>::from_double (m_axis * (v.sq_double_length () / proj));
      }
    }
  }

  /**
   *  @brief Transform by a given transformation
   */
  template <class T>
  Constraint &transform (const T &t)
  {
    m_axis.transform (t);
    return *this;
  }

  /**
   *  @brief Return the transformed version
   */
  template <class T>
  Constraint transformed (const T &t) const
  {
    Constraint c (*this);
    return c.transform (t);
  }

private:
  constraint_mode m_mode;
  db::Vector m_axis;
};

// -------------------------------------------------------------
//  A cache for the transformation variants for a certain layer and cell view index
//  for a lay::LayoutView

class TransformationVariants
{
public:
  TransformationVariants (const lay::LayoutView *view, bool per_cv_and_layer = true, bool per_cv = true);

  const std::vector<db::CplxTrans> *per_cv_and_layer (unsigned int cv, unsigned int layer) const;
  const std::vector<db::CplxTrans> *per_cv (unsigned int cv) const;

private:
  std::map <unsigned int, std::vector<db::CplxTrans> > m_per_cv_tv;
  std::map < std::pair<unsigned int, unsigned int>, std::vector<db::CplxTrans> > m_per_cv_and_layer_tv;
};

TransformationVariants::TransformationVariants (const lay::LayoutView *view, bool per_cv_and_layer, bool per_cv)
{
  //  build the transformation variants cache

  for (lay::LayerPropertiesConstIterator l = view->begin_layers (); !l.at_end (); ++l) {

    if (! l->has_children ()) {

      unsigned int cvi = (l->cellview_index () >= 0) ? (unsigned int) l->cellview_index () : 0;

      if (per_cv) {
        std::vector<db::CplxTrans> &tv = m_per_cv_tv.insert (std::make_pair (cvi, std::vector<db::CplxTrans> ())).first->second;
        tv.insert (tv.end (), l->trans ().begin (), l->trans ().end ());
      }

      if (l->layer_index () >= 0 && per_cv_and_layer) {
        std::vector<db::CplxTrans> &tv = m_per_cv_and_layer_tv.insert (std::make_pair (std::make_pair (cvi, (unsigned int) l->layer_index ()), std::vector<db::CplxTrans> ())).first->second;
        tv.insert (tv.end (), l->trans ().begin (), l->trans ().end ());
      }

    }

  }

  //  remove duplicates of the list of transformations (not related to layers)
  for (std::map <unsigned int, std::vector<db::CplxTrans> >::iterator t = m_per_cv_tv.begin (); t != m_per_cv_tv.end (); ++t) {
    std::sort (t->second.begin (), t->second.end ());
    std::vector<db::CplxTrans>::iterator new_last = std::unique (t->second.begin (), t->second.end ());
    t->second.erase (new_last, t->second.end ());
  }

  //  remove duplicates of the list of transformations (related to layers)
  for (std::map < std::pair<unsigned int, unsigned int>, std::vector<db::CplxTrans> >::iterator t = m_per_cv_and_layer_tv.begin (); t != m_per_cv_and_layer_tv.end (); ++t) {
    std::sort (t->second.begin (), t->second.end ());
    std::vector<db::CplxTrans>::iterator new_last = std::unique (t->second.begin (), t->second.end ());
    t->second.erase (new_last, t->second.end ());
  }
}

const std::vector<db::CplxTrans> *
TransformationVariants::per_cv_and_layer (unsigned int cv, unsigned int layer) const
{
  std::map <std::pair<unsigned int, unsigned int>, std::vector<db::CplxTrans> >::const_iterator t = m_per_cv_and_layer_tv.find (std::make_pair (cv, layer));
  if (t != m_per_cv_and_layer_tv.end ()) {
    return &t->second;
  } else {
    return 0;
  }
}

const std::vector<db::CplxTrans> *
TransformationVariants::per_cv (unsigned int cv) const
{
  std::map <unsigned int, std::vector<db::CplxTrans> >::const_iterator t = m_per_cv_tv.find (cv);
  if (t != m_per_cv_tv.end ()) {
    return &t->second;
  } else {
    return 0;
  }
}


// -------------------------------------------------------------
//  Some utilities

static bool
insert_point_path (const db::Path &p, const std::set<EdgeWithIndex> &sel, db::Point &ins, db::Path &new_path)
{
  new_path.width (p.width ());
  new_path.round (p.round ());
  new_path.extensions (p.bgn_ext (), p.end_ext ());

  std::vector<db::Point> ctr;
  ctr.reserve (p.points () + 1);

  bool found = false;

  unsigned int n = 0;
  for (db::Path::iterator pt = p.begin (); pt != p.end (); ++n) {
    db::Point p1 = *pt;
    ++pt;
    if (pt != p.end ()) {
      db::Point p2 = *pt;
      ctr.push_back (p1);
      if (! found && sel.find (EdgeWithIndex (db::Edge (p1, p2), n, n + 1)) != sel.end ()) {
        //  project the point onto the edge
        std::pair <bool, db::Point> projected = db::Edge (p1, p2).projected (ins);
        if (projected.first) {
          ins = projected.second;
          ctr.push_back (ins);
          found = true;
        }
      }
    } else {
      ctr.push_back (p1);
    }
  }

  if (found) {
    new_path.assign (ctr.begin (), ctr.end ());
  }
  return true;
}

static void 
assign_path_compressed (db::Path &p, std::vector <db::Point> &ctr)
{
  //  compress contour (remove redundant points)
  //  and assign to path

  std::vector<db::Point>::iterator wp = ctr.begin ();
  std::vector<db::Point>::iterator wp0 = wp;
  std::vector<db::Point>::const_iterator rp = ctr.begin ();
  db::Point pm1 = *rp;
  if (wp != ctr.end ()) {
    ++wp;
    ++rp;
    while (rp != ctr.end ()) {
      db::Point p0 = *rp;
      ++rp;
      if (rp != ctr.end ()) {
        db::Point pp1 = *rp;
        if (! (db::vprod_sign (pp1 - p0, p0 - pm1) == 0 && db::sprod_sign (pp1 - p0, p0 - pm1) >= 0)) {
          *wp++ = p0;
          pm1 = p0;
        }
      } else {
        *wp++ = p0;
      }
    }
  }

  p.assign (wp0, wp);
}

static db::Path
del_points_path (const db::Path &p, const std::set<EdgeWithIndex> &sel)
{
  db::Path new_path;
  new_path.width (p.width ());
  new_path.round (p.round ());
  new_path.extensions (p.bgn_ext (), p.end_ext ());

  std::vector<db::Point> ctr;
  ctr.reserve (p.points ());

  unsigned int n = 0;
  for (db::Path::iterator pt = p.begin (); pt != p.end (); ++pt, ++n) {
    db::Point p1 = *pt;
    if (sel.find (EdgeWithIndex (db::Edge (p1, p1), n, n)) == sel.end ()) {
      ctr.push_back (p1);
    }
  }

  assign_path_compressed (new_path, ctr);

  return new_path;
}

static void 
modify_path (db::Path &p, const std::map <PointWithIndex, db::Point> &new_points, const std::map <EdgeWithIndex, db::Edge> &new_edges, bool compress = false)
{
  std::vector<db::Point> ctr;
  ctr.reserve (p.points ());

  std::map <PointWithIndex, db::Point>::const_iterator np;
  std::map <EdgeWithIndex, db::Edge>::const_iterator ne;

  unsigned int n = 0;
  for (db::Path::iterator pt = p.begin (); pt != p.end (); ++n) {

    db::Point p1 = *pt;
    db::Point p1org = p1;
    np = new_points.find (PointWithIndex (p1, n));
    if (np != new_points.end ()) {
      p1 = np->second;
    }

    ++pt;
    if (pt != p.end ()) {

      db::Point p2 = *pt;
      db::Point p2org = p2;
      np = new_points.find (PointWithIndex (p2, n + 1));
      if (np != new_points.end ()) {
        p2 = np->second;
      }

      ne = new_edges.find (EdgeWithIndex (db::Edge (p1org, p2org), n, n + 1));

      ctr.push_back (p1);

      if (ne != new_edges.end () && ne->second.p1 () != p1) {
        ctr.push_back (ne->second.p1 ());
      }
      if (ne != new_edges.end () && ne->second.p2 () != p2) {
        ctr.push_back (ne->second.p2 ());
      }

    } else {
      ctr.push_back (p1);
    }

  }

  if (compress) {
    assign_path_compressed (p, ctr);
  } else {
    p.assign (ctr.begin (), ctr.end ());
  }
}

bool
insert_point_poly (const db::Polygon &p, const std::set<EdgeWithIndex> &sel, db::Point &ins, db::Polygon &new_poly)
{
  bool found = false;

  std::vector<db::Point> ctr;
  size_t points = p.hull ().size ();
  ctr.reserve (points + 1);

  unsigned int n = 0;
  db::Shape::polygon_edge_iterator ee;
  for (db::Shape::polygon_edge_iterator e = p.begin_edge (); ! e.at_end (); e = ee, ++n) {

    ee = e;
    ++ee;
    unsigned int nn = ee.at_end () ? 0 : n + 1;

    ctr.push_back ((*e).p1 ());
    if (! found && sel.find (EdgeWithIndex (*e, n, nn)) != sel.end ()) {
      //  project the point onto the edge
      std::pair <bool, db::Point> projected = (*e).projected (ins);
      if (projected.first) {
        ins = projected.second;
        ctr.push_back (ins);
        found = true;
      }
    }

  }

  if (found) {

    tl_assert (p.holes () == 0);
    new_poly.assign_hull (ctr.begin (), ctr.end (), false /*don't compress*/);

  }

  return found;
}

static db::Polygon
del_points_poly (const db::Polygon &p, const std::set<EdgeWithIndex> &sel)
{
  db::Polygon new_poly;

  std::vector<db::Point> ctr;
  size_t points = p.hull ().size ();
  ctr.reserve (points);

  unsigned int n = 0;
  for (db::Shape::polygon_edge_iterator e = p.begin_edge (); ! e.at_end (); ++e, ++n) {
    db::Point p1 = (*e).p1 ();
    if (sel.find (EdgeWithIndex (db::Edge (p1, p1), n, n)) == sel.end ()) {
      ctr.push_back (p1);
    }
  }

  //  @@@ TODO: consider holes.
  tl_assert (p.holes () == 0);
  new_poly.assign_hull (ctr.begin (), ctr.end (), true /*compress*/);

  return new_poly;
}

static void
modify_polygon (db::Polygon &p, 
                const std::map <PointWithIndex, db::Point> &new_points, 
                const std::map <EdgeWithIndex, db::Edge> &new_edges,
                bool compress = false)
{
  std::vector<db::Point> ctr;
  size_t points = p.hull ().size ();
  ctr.reserve (points);

  std::map <PointWithIndex, db::Point>::const_iterator np;
  std::map <EdgeWithIndex, db::Edge>::const_iterator ne;

  unsigned int n = 0;
  db::Shape::polygon_edge_iterator ee;
  for (db::Shape::polygon_edge_iterator e = p.begin_edge (); ! e.at_end (); e = ee, ++n) {

    ee = e;
    ++ee;
    unsigned int nn = ee.at_end () ? 0 : n + 1;

    db::Point p1 = (*e).p1 ();
    np = new_points.find (PointWithIndex (p1, n));
    if (np != new_points.end ()) {
      p1 = np->second;
    }

    db::Point p2 = (*e).p2 ();
    np = new_points.find (PointWithIndex (p2, nn));
    if (np != new_points.end ()) {
      p2 = np->second;
    }

    ne = new_edges.find (EdgeWithIndex (*e, n, nn));

    ctr.push_back (p1);

    if (ne != new_edges.end () && ne->second.p1 () != p1) {
      ctr.push_back (ne->second.p1 ());
    }
    if (ne != new_edges.end () && ne->second.p2 () != p2) {
      ctr.push_back (ne->second.p2 ());
    }

  }

  //  @@@ TODO: consider holes in the contour list ...
  tl_assert (p.holes () == 0); // @@@ 
  p.assign_hull (ctr.begin (), ctr.end (), compress);

}

static void 
constrain (std::map <PointWithIndex, Constraint> &constr, const EdgeWithIndex &edge)
{
  constr.insert (std::make_pair (edge.pi1 (), Constraint ())).first->second *= Constraint (edge.d ());
  constr.insert (std::make_pair (edge.pi2 (), Constraint ())).first->second *= Constraint (edge.d ());
}

static void
create_shift_sets (edt::PartialService::partial_objects::const_iterator r, std::map <PointWithIndex, db::Point> &new_points, std::map <EdgeWithIndex, db::Edge> &new_edges, db::Vector mv)
{
  //  Set up a map of new edges and new points
  for (std::set <EdgeWithIndex>::const_iterator e = r->second.begin (); e != r->second.end (); ++e) {
    if (e->p1 () != e->p2 ()) {
      new_edges.insert (std::make_pair (*e, db::Edge (*e)));
    } else {
      new_points.insert (std::make_pair (PointWithIndex (e->p1 (), e->n), e->p1 ()));
    }
  }

  //  new_points should only contain the selected points, not the start and end points of selected edges
  for (std::set <EdgeWithIndex>::const_iterator e = r->second.begin (); e != r->second.end (); ++e) {
    if (e->p1 () != e->p2 ()) {
      new_points.erase (e->pi1 ());
      new_points.erase (e->pi2 ());
    }
  }

  std::map <PointWithIndex, Constraint> point_constr;

  if (r->first.shape ().is_polygon ()) {

    unsigned int n = 0;
    db::Shape::polygon_edge_iterator ee;
    for (db::Shape::polygon_edge_iterator e = r->first.shape ().begin_edge (); ! e.at_end (); e = ee, ++n) {
      ee = e;
      ++ee;
      unsigned int nn = ee.at_end () ? 0 : n + 1;
      if ((*e).p1 () != (*e).p2 () && r->second.find (EdgeWithIndex (*e, n, nn)) == r->second.end ()) {
        constrain (point_constr, EdgeWithIndex (*e, n, nn));
      }
    }

  } else if (r->first.shape ().is_path ()) {

    if (r->first.shape ().begin_point () != r->first.shape ().end_point ()) {

      db::Shape::point_iterator pt = r->first.shape ().begin_point ();
      db::Point p1 = *pt;

      unsigned int n = 0;
      while (true) {

        ++pt;
        if (pt == r->first.shape ().end_point ()) {
          break;
        }

        EdgeWithIndex e (db::Edge (p1, *pt), n, n + 1);
        if (e.p1 () != e.p2 () && r->second.find (e) == r->second.end ()) {
          constrain (point_constr, e);
        }

        p1 = *pt;
        ++n;

      }

    }

  } else if (r->first.shape ().is_box ()) {

    //  convert to polygon and test those edges
    db::Polygon poly (r->first.shape ().box ());
    unsigned int n = 0;
    db::Shape::polygon_edge_iterator ee;
    for (db::Shape::polygon_edge_iterator e = poly.begin_edge (); ! e.at_end (); e = ee, ++n) {
      ee = e;
      ++ee;
      unsigned int nn = ee.at_end () ? 0 : n + 1;
      EdgeWithIndex ewi (*e, n, nn);
      if ((*e).p1 () != (*e).p2 () && r->second.find (ewi) == r->second.end ()) {
        //  add some moveable edges to impose manhattan constraints
        if (new_points.find (ewi.pi1 ()) != new_points.end () || 
            new_points.find (ewi.pi2 ()) != new_points.end ()) {
          new_edges.insert (std::make_pair (ewi, db::Edge ((*e))));
        } else {
          constrain (point_constr, ewi);
        }
      }
    }

  } 

  //  Simply move the points
  for (std::map <PointWithIndex, db::Point>::iterator np = new_points.begin (); np != new_points.end (); ++np) {
    np->second += mv;
  }

  //  The edges are treated somewhat more elaborate:
  for (std::map <EdgeWithIndex, db::Edge>::iterator ne = new_edges.begin (); ne != new_edges.end (); ++ne) {

    std::map <PointWithIndex, Constraint>::iterator c;

    //  compute normal of move vector
    db::Vector nmv;
    if (ne->first.d () != db::Vector ()) {
      nmv = mv - db::Vector::from_double (ne->first.d () * (double (db::sprod (mv, ne->first.d ())) / ne->first.d ().sq_double_length ()));
    }

    db::Point p1 = ne->second.p1 (), p2 = ne->second.p2 ();
    db::Point p1e = p1, p2e = p2;

    c = point_constr.find (ne->first.pi1 ());
    if (c != point_constr.end ()) {
      p1e = p1 = c->second.move (p1, nmv);
      //  if the movement was inhibited, create a new "detached" edge 
      if (c->second.mode () == Constraint::OneDim && p1 == ne->second.p1 ()) {
        p1e = p1 + nmv;
      }
    } else {
      p1 += mv;
      p1e = p1;
    }

    c = point_constr.find (ne->first.pi2 ());
    if (c != point_constr.end ()) {
      p2e = p2 = c->second.move (p2, nmv);
      //  if the movement was inhibited, create a new "detached" edge 
      if (c->second.mode () == Constraint::OneDim && p2 == ne->second.p2 ()) {
        p2e = p2 + nmv;
      }
    } else {
      p2 += mv;
      p2e = p2;
    }

    ne->second = db::Edge (p1e, p2e);

    //  insert the end points into the point list in order to find them by looking up a point alone
    new_points.insert (std::make_pair (ne->first.pi1 (), db::Point ())).first->second = p1;
    new_points.insert (std::make_pair (ne->first.pi2 (), db::Point ())).first->second = p2;

  }

}

// -------------------------------------------------------------
//  PartialShapeFinder declaration

/**
 *  @brief Partial shape finder utility class
 *
 *  This class specializes the finder to finding vertices or edges of shapes.
 */
class PartialShapeFinder
  : public Finder
{
public:
  struct StopException { };

  typedef std::vector<std::pair <edt::ObjectInstPath, std::vector <EdgeWithIndex> > > founds_vector_type;
  typedef founds_vector_type::const_iterator iterator;

  PartialShapeFinder (bool point_mode, bool top_level_sel, db::ShapeIterator::flags_type flags);

  bool find (lay::LayoutView *view, const lay::LayerProperties &lprops, const db::DBox &region_mu);
  bool find (lay::LayoutView *view, const db::DBox &region_mu);

  iterator begin () const
  {
    return m_founds.begin ();
  }

  iterator end () const
  {
    return m_founds.end ();
  }

private:
  virtual void 
  visit_cell (const db::Cell &cell, const db::Box &search_box, const db::CplxTrans &t, int /*level*/);

  lay::LayoutView *mp_view;
  founds_vector_type m_founds;
  db::ShapeIterator::flags_type m_flags;
  unsigned int m_cv_index;
  db::Layout::cell_index_type m_topcell;
  const std::set<db::PropertiesRepository::properties_id_type> *mp_prop_sel;
  bool m_inv_prop_sel;
  int m_tries;
  tl::Progress m_progress;
};

// -------------------------------------------------------------
//  PartialShapeFinder implementation

PartialShapeFinder::PartialShapeFinder (bool point_mode, bool top_level_sel, db::ShapeIterator::flags_type flags)
  : Finder (point_mode, top_level_sel), 
    mp_view (0), 
    m_flags (flags), 
    mp_prop_sel (0), 
    m_inv_prop_sel (false), 
    m_progress (tl::translate ("Selecting ..."))
{
  m_progress.set_unit (1000);
  m_progress.set_format ("");
}

bool
PartialShapeFinder::find (lay::LayoutView *view, const db::DBox &region_mu)
{
  for (lay::LayerPropertiesConstIterator lprop = view->begin_layers (); ! lprop.at_end (); ++lprop) {
    if (lprop->visible (true /*real*/)) {
      find (view, *lprop, region_mu);
    }
  }

  return ! m_founds.empty ();
}

bool 
PartialShapeFinder::find (lay::LayoutView *view, const lay::LayerProperties &lprops, const db::DBox &region_mu)
{
  m_cv_index = lprops.cellview_index ();
  mp_view = view;

  const lay::CellView &cv = view->cellview (m_cv_index);
  if (! cv.is_valid ()) {
    return false;
  }

  m_topcell = cv.cell_index ();

  double dbu = cv->layout ().dbu ();
  db::Box region = db::Box::from_double (region_mu * (1.0 / dbu));

  m_tries = point_sel_tests;
  mp_prop_sel = &lprops.prop_sel ();
  m_inv_prop_sel = lprops.inverse_prop_sel ();

  int min_level = view->get_min_hier_levels ();
  int max_level = view->get_max_hier_levels ();
  if (lprops.hier_levels ().has_from_level ()) {
    min_level = lprops.hier_levels ().from_level ();
  }
  if (lprops.hier_levels ().has_to_level ()) {
    max_level = lprops.hier_levels ().to_level ();
  }

  //  actually find
  try {
    start (view, cv, m_cv_index, lprops.trans (), region, min_level, max_level, int (lprops.layer_index ()));
  } catch (StopException) {
    return false;
  }

  //  return true if anything was found
  return ! m_founds.empty ();
}

void 
PartialShapeFinder::visit_cell (const db::Cell &cell, const db::Box &search_box, const db::CplxTrans &t, int level)
{
  if (! point_mode ()) {

    ++m_progress;

    const db::Shapes &shapes = cell.shapes (layer ());

    db::ShapeIterator shape = shapes.begin_touching (search_box, m_flags, mp_prop_sel, m_inv_prop_sel);
    while (! shape.at_end ()) {

      ++m_progress;

      //  in point mode just store that found that has the least "distance"
      m_founds.push_back (founds_vector_type::value_type ());

      ObjectInstPath &inst_path = m_founds.back ().first;
      std::vector<EdgeWithIndex> &edges = m_founds.back ().second;

      inst_path.set_cv_index (m_cv_index);
      inst_path.set_topcell (m_topcell);
      inst_path.assign_path (path ().begin (), path ().end ());
      inst_path.set_layer (layer ());
      inst_path.set_shape (*shape);

      //  in point mode, test the edges and use a "closest" criterion
      if (shape->is_polygon ()) {

        unsigned int n = 0; 
        db::Shape::polygon_edge_iterator ee;
        for (db::Shape::polygon_edge_iterator e = shape->begin_edge (); ! e.at_end (); e = ee, ++n) {

          ee = e;
          ++ee;
          unsigned int nn = ee.at_end () ? 0 : n + 1;

          if (search_box.contains ((*e).p1 ())) {
            edges.push_back (EdgeWithIndex (db::Edge ((*e).p1 (), (*e).p1 ()), n, n));
            if (search_box.contains ((*e).p2 ())) {
              edges.push_back (EdgeWithIndex (*e, n, nn));
            } 
          } 

        }

      } else if (shape->is_path ()) {

        bool pl_set = false;
        db::Point pl;
        unsigned int n = 0;
        for (db::Shape::point_iterator pt = shape->begin_point (); pt != shape->end_point (); ++pt, ++n) {
          if (search_box.contains (*pt)) {
            edges.push_back (EdgeWithIndex (db::Edge (*pt, *pt), n, n));
            if (pl_set && search_box.contains (pl)) {
              edges.push_back (EdgeWithIndex (db::Edge (pl, *pt), n - 1, n));
            }
          }
          pl = *pt;
          pl_set = true;
        }

      } else if (shape->is_box ()) {

        const db::Box &box = shape->box ();

        //  convert to polygon and test those edges
        db::Polygon poly (box);
        unsigned int n = 0; 
        db::Shape::polygon_edge_iterator ee;
        for (db::Shape::polygon_edge_iterator e = poly.begin_edge (); ! e.at_end (); e = ee, ++n) {

          ee = e;
          ++ee;
          unsigned int nn = ee.at_end () ? 0 : n + 1;

          if (search_box.contains ((*e).p1 ())) {
            edges.push_back (EdgeWithIndex (db::Edge ((*e).p1 (), (*e).p1 ()), n, n));
            if (search_box.contains ((*e).p2 ())) {
              edges.push_back (EdgeWithIndex (*e, n, nn));
            } 
          } 

        }

      } else if (shape->is_text ()) {

        db::Point tp (shape->text_trans () * db::Point ());
        if (search_box.contains (tp)) {
          edges.push_back (EdgeWithIndex (db::Edge (tp, tp), 0, 0));
        }

      }

      //  do not select shapes that do not have at least one edge selected
      if (edges.empty ()) {
        m_founds.pop_back ();
      }

      ++shape;

    } 

  } else {

    if (--m_tries < 0) {
      throw StopException ();
    }

    const db::Shapes &shapes = cell.shapes (layer ());
    std::vector <EdgeWithIndex> edge_sel;

    db::ShapeIterator shape = shapes.begin_touching (search_box, m_flags, mp_prop_sel, m_inv_prop_sel);
    while (! shape.at_end ()) {

      bool match = false;
      double d = std::numeric_limits<double>::max ();

      edge_sel.clear ();

      if (--m_tries < 0) {
        throw StopException ();
      }

      db::Point point (search_box.center ());

      //  in point mode, test the edges and use a "closest" criterion
      if (shape->is_polygon ()) {

        unsigned int n = 0;
        db::Shape::polygon_edge_iterator ee;
        for (db::Shape::polygon_edge_iterator e = shape->begin_edge (); ! e.at_end (); e = ee, ++n) {

          ee = e;
          ++ee;
          unsigned int nn = ee.at_end () ? 0 : n + 1;

          unsigned int r = test_edge (t * *e, d, match);
          if (r) {
            edge_sel.clear ();
            if ((r & 1) == 1) {
              edge_sel.push_back (EdgeWithIndex (db::Edge ((*e).p1 (), (*e).p1 ()), n, n));
            }
            if ((r & 2) == 2) {
              edge_sel.push_back (EdgeWithIndex (db::Edge ((*e).p2 (), (*e).p2 ()), nn, nn));
            }
            if (r == 3) {
              edge_sel.push_back (EdgeWithIndex (*e, n, nn));
            }
          }

        }

      } else if (shape->is_path ()) {

        //  test the "spine"
        db::Shape::point_iterator pt = shape->begin_point (); 
        if (pt != shape->end_point ()) {
          db::Point p (*pt);
          ++pt;
          unsigned int n = 0;
          for (; pt != shape->end_point (); ++pt, ++n) {
            unsigned int r = test_edge (t * db::Edge (p, *pt), d, match);
            if (r) {
              edge_sel.clear ();
              if ((r & 1) == 1) {
                edge_sel.push_back (EdgeWithIndex (db::Edge (p, p), n, n));
              }
              if ((r & 2) == 2) {
                edge_sel.push_back (EdgeWithIndex (db::Edge (*pt, *pt), n + 1, n + 1));
              }
              if (r == 3) {
                edge_sel.push_back (EdgeWithIndex (db::Edge (p, *pt), n, n + 1));
              }
            }
            p = *pt;
          }
        }

      } else if (shape->is_box ()) {

        const db::Box &box = shape->box ();

        //  convert to polygon and test those edges
        db::Polygon poly (box);

        unsigned int n = 0;
        db::Shape::polygon_edge_iterator ee;
        for (db::Shape::polygon_edge_iterator e = poly.begin_edge (); ! e.at_end (); e = ee, ++n) {

          ee = e;
          ++ee;
          unsigned int nn = ee.at_end () ? 0 : n + 1;

          unsigned int r = test_edge (t * *e, d, match);
          if (r) {
            edge_sel.clear ();
            if ((r & 1) == 1) {
              edge_sel.push_back (EdgeWithIndex (db::Edge ((*e).p1 (), (*e).p1 ()), n, n));
            }
            if ((r & 2) == 2) {
              edge_sel.push_back (EdgeWithIndex (db::Edge ((*e).p2 (), (*e).p2 ()), nn, nn));
            }
            if (r == 3) {
              edge_sel.push_back (EdgeWithIndex (*e, n, nn));
            }
          }

        }

      } else if (shape->is_text ()) {

        db::Point tp (shape->text_trans () * db::Point ());
        if (search_box.contains (tp)) {
          d = tp.distance (search_box.center ());
          edge_sel.clear ();
          edge_sel.push_back (EdgeWithIndex (db::Edge (tp, tp), 0, 0));
          match = true;
        }

      }

      if (match && closer (d)) {

        //  in point mode just store that found that has the least "distance"
        if (m_founds.empty ()) {
          m_founds.push_back (founds_vector_type::value_type ());
        }

        ObjectInstPath &inst_path = m_founds.back ().first;

        inst_path.set_cv_index (m_cv_index);
        inst_path.set_topcell (m_topcell);
        inst_path.assign_path (path ().begin (), path ().end ());
        inst_path.set_layer (layer ());
        inst_path.set_shape (*shape);

        m_founds.back ().second = edge_sel;

      }

      ++shape;

    }

  }

}

// -----------------------------------------------------------------------------
//  Main Service implementation

PartialService::PartialService (db::Manager *manager, lay::LayoutView *view, lay::MainWindow *mw)
  : lay::ViewService (view->view_object_widget ()), 
    lay::Editable (view),
    lay::Plugin (view),
    db::Object (manager),
    mp_view (view),
    mp_main_window (mw),
    m_dragging (false),
    m_keep_selection (true),
    mp_box (0),
    m_color (0),
    m_buttons (0),
    m_connect_ac (edt::Any), m_move_ac (edt::Any),
    m_top_level_sel (false)
{ 
  // ...
}

PartialService::~PartialService ()
{
  resize_markers (0);
  resize_inst_markers (0);

  if (mp_box) {
    delete mp_box;
    mp_box = 0;
  }
}

void  
PartialService::deactivated ()
{
  //  clear selection when this mode is left
  partial_select (db::DBox (), lay::Editable::Reset);
}

void  
PartialService::activated ()
{
  //  clear selection when this mode gets active
  view ()->clear_selection ();
}

void 
PartialService::set_colors (QColor /*background*/, QColor color)
{
  m_color = color.rgb ();
  if (mp_box) {
    mp_box->set_color (m_color);
  }
}

void 
PartialService::menu_activated (const std::string &symbol)
{
  // ...
}

bool 
PartialService::configure (const std::string &name, const std::string &value)
{
  edt::EditGridConverter egc;
  edt::ACConverter acc;

  if (name == cfg_edit_global_grid) {
    egc.from_string (value, m_global_grid);
  } else if (name == cfg_edit_grid) {
    egc.from_string (value, m_edit_grid);
    return true;  //  taken
  } else if (name == cfg_edit_move_angle_mode) {
    acc.from_string (value, m_move_ac);
    return true;  //  taken
  } else if (name == cfg_edit_connect_angle_mode) {
    acc.from_string (value, m_connect_ac);
    return true;  //  taken
  } else if (name == cfg_edit_top_level_selection) {
    tl::from_string (value, m_top_level_sel);
  }

  return false;  //  not taken
}

void 
PartialService::config_finalize ()
{
  // ...
}

//  TODO: this is duplicate code from Service::snap ..
db::DPoint 
PartialService::snap (const db::DPoint &p) const
{
  //  snap according to the grid
  if (m_edit_grid == db::DPoint ()) {
    return lay::snap_xy (p, m_global_grid);
  } else if (m_edit_grid.x () >= 1e-6) {
    return lay::snap_xy (p, m_edit_grid);
  } else {
    return p;
  }
}

void 
PartialService::edit_cancel ()
{
  //  stop dragging, clear selection
  m_dragging = false;

  if (mp_box) {
    delete mp_box;
    mp_box = 0;
  }

  widget ()->ungrab_mouse (this);

  partial_select (db::DBox (), lay::Editable::Reset);
}

bool  
PartialService::mouse_move_event (const db::DPoint &p, bool prio)
{
  if (m_dragging) {

    // drag the vertex or edge/segment
    m_current = snap (p - m_start) + m_start;
    selection_to_view ();

  } else if (prio && mp_box) {

    m_p2 = p;
    mp_box->set_points (m_p1, m_p2);

  } 

  //  pass on this event to other handlers
  return false;
}

bool  
PartialService::mouse_press_event (const db::DPoint &p, unsigned int buttons, bool prio)
{
  if (! view ()->is_editable ()) {
    return false;
  }

  //  only respond to left button clicks
  if ((buttons & lay::LeftButton) == 0) {
    return false;
  }

  //  only respond to first order events
  if (prio) {
    return false;
  }

  if (m_dragging) {

    //  eat events if already dragging
    return true;

  } else if (! mp_box) {

    if (m_selection.empty ()) {

      //  nothing is selected yet: 
      //  try to select something here.
      //  (select is allowed to throw an exception)
      
      try {
        partial_select (db::DBox (p, p), lay::Editable::Replace);
      } catch (tl::Exception &ex) {
        tl::error << ex.msg ();
        QMessageBox::critical (0, tl::translate ("Error").c_str (), ex.msg ().c_str ());
        //  clear selection
        partial_select (db::DBox (), lay::Editable::Reset);
      }

    }

    if (m_selection.empty () || ((buttons & lay::ShiftButton) != 0) || ((buttons & lay::ControlButton) != 0)) {

      //  if nothing was selected by this point or Ctrl or Shift was pressed, start dragging a box

      view ()->stop_redraw (); // @@@ how to restart if selection is aborted?
      m_buttons = buttons;

      if (mp_box) {
        delete mp_box;
      }

      m_p1 = p;
      m_p2 = p;
      mp_box = new lay::RubberBox (widget (), m_color, p, p);
      mp_box->set_stipple (6); // coarse hatched

      widget ()->grab_mouse (this, true);

    } else {

      //  something was selected: start dragging this ..
      m_dragging = true;
      m_keep_selection = true;
      m_current = m_start = p;

      widget ()->grab_mouse (this, true);

    }

    return true;

  }

  return false;
}

bool  
PartialService::mouse_click_event (const db::DPoint &p, unsigned int buttons, bool prio)
{
  if (! view ()->is_editable ()) {
    return false;
  }

  //  only respond to left button clicks
  if ((buttons & lay::LeftButton) == 0) {
    return false;
  }

  //  only respond to first order events
  if (! prio) {
    return false;
  }

  if (m_dragging) {

    if (m_current != m_start) {

      //  stop dragging
      widget ()->ungrab_mouse (this);
      
      manager ()->transaction (tl::translate ("Partial move")); 

      //  heuristically, if there is just one edge selected: do not confine to the movement
      //  angle constraint - the edge usually is confined enough
      db::DTrans move_trans;
      if (m_selection.size () == 1 && m_selection.begin ()->second.size () == 3 /*p1,p2,edge*/) {
        move_trans = db::DTrans (m_current - m_start);
      } else {
        //  TODO: DTrans should have a ctor that takes a vector
        move_trans = db::DTrans (db::DPoint () + edt::snap_angle (m_current - m_start, m_move_ac));
      }

      //  build the transformation variants cache
      TransformationVariants tv (view ());

      for (partial_objects::iterator r = m_selection.begin (); r != m_selection.end (); ++r) {

        const lay::CellView &cv = view ()->cellview (r->first.cv_index ());

        if (! r->first.is_cell_inst ()) {

          const std::vector<db::CplxTrans> *tv_list = tv.per_cv_and_layer (r->first.cv_index (), r->first.layer ());
          if (tv_list && ! tv_list->empty ()) {

            //  use only the first one of the explicit transformations 
            //  TODO: clarify how this can be implemented in a more generic form or leave it thus.

            db::CplxTrans gt (cv.context_trans () * r->first.trans ());
            db::DCplxTrans tt = (*tv_list) [0] * db::CplxTrans (cv->layout ().dbu ()) * gt;
            db::Vector move_vector (db::Point::from_double (tt.inverted () * (move_trans * (tt * db::DPoint ()))));

            std::map <EdgeWithIndex, db::Edge> new_edges;
            std::map <PointWithIndex, db::Point> new_points;
            create_shift_sets (r, new_points, new_edges, move_vector);

            //  modify the shapes and insert

            db::Shapes &shapes = cv->layout ().cell (r->first.cell_index ()).shapes (r->first.layer ());
            db::GenericRepository &rep = cv->layout ().shape_repository ();

            if (r->first.shape ().is_polygon ()) {

              db::Polygon poly;
              r->first.shape ().polygon (poly);

              //  warning: poly is modified:
              modify_polygon (poly, new_points, new_edges, true /*compress*/);

              shapes.replace (r->first.shape (), rep, poly);

            } else if (r->first.shape ().is_path ()) {

              db::Path path;
              r->first.shape ().path (path);

              //  warning: path is modified:
              modify_path (path, new_points, new_edges, true /*compress*/);

              shapes.replace (r->first.shape (), rep, path);

            } else if (r->first.shape ().is_box ()) {

              db::Polygon poly;
              r->first.shape ().polygon (poly);

              //  warning: poly is modified:
              modify_polygon (poly, new_points, new_edges, true /*compress*/);

              shapes.replace (r->first.shape (), rep, poly.box ());

            } else if (r->first.shape ().is_text ()) {

              db::Text t;
              r->first.shape ().text (t);

              db::Point tp (r->first.shape ().text_trans () * db::Point ());
              std::map <PointWithIndex, db::Point>::const_iterator np = new_points.find (PointWithIndex (tp, 0));

              if (np != new_points.end ()) {
                t.transform (db::Trans (np->second - tp));
                shapes.replace (r->first.shape (), rep, t);
              }

            }

            //  transform the selection
            
            std::set <EdgeWithIndex> new_sel;
            new_sel.swap (r->second);

            for (std::set <EdgeWithIndex>::const_iterator s = new_sel.begin (); s != new_sel.end () && m_keep_selection; ++s) {
              if (s->p1 () == s->p2 ()) {
                std::map <PointWithIndex, db::Point>::const_iterator np = new_points.find (s->pi1 ());
                if (np != new_points.end ()) {
                  r->second.insert (EdgeWithIndex (db::Edge (np->second, np->second), s->n, s->n));
                } else {
                  r->second.insert (*s);
                }
              } else {
                std::map <EdgeWithIndex, db::Edge>::const_iterator ne = new_edges.find (*s);
                if (ne != new_edges.end ()) {
                  r->second.insert (EdgeWithIndex (ne->second, s->n, s->nn));
                } else {
                  r->second.insert (*s);
                }
              }
            }

          }

        }

      }

      //  then move all instances.
      
      //  TODO: DTrans should have a ctor that takes a vector
      db::DTrans move_trans_inst = db::DTrans (db::DPoint () + edt::snap_angle (m_current - m_start, m_move_ac));

      //  sort the selected objects (the instances) by the cell they are in
      //  The key is a pair: cell_index, cv_index
      std::map <std::pair <db::Layout::cell_index_type, unsigned int>, std::vector <partial_objects::const_iterator> > insts_by_cell;
      for (partial_objects::const_iterator r = m_selection.begin (); r != m_selection.end (); ++r) {
        if (r->first.is_cell_inst ()) {
          insts_by_cell.insert (std::make_pair (std::make_pair (r->first.cell_index (), r->first.cv_index ()), std::vector <partial_objects::const_iterator> ())).first->second.push_back (r);
        }
      }

      std::vector <std::pair <db::Instance, db::CplxTrans> > insts_to_transform;
      for (std::map <std::pair <db::Layout::cell_index_type, unsigned int>, std::vector <partial_objects::const_iterator> >::const_iterator ibc = insts_by_cell.begin (); ibc != insts_by_cell.end (); ++ibc) {

        insts_to_transform.clear ();
        insts_to_transform.reserve (ibc->second.size ());
        for (std::vector <partial_objects::const_iterator>::const_iterator i = ibc->second.begin (); i != ibc->second.end (); ++i) {
          insts_to_transform.push_back (std::make_pair ((*i)->first.back ().inst_ptr, (*i)->first.trans ()));
        }

        const lay::CellView &cv = view ()->cellview (ibc->first.second);
        if (cv.is_valid ()) {

          const std::vector<db::CplxTrans> *tv_list = tv.per_cv (ibc->first.second);
          if (tv_list && ! tv_list->empty ()) {

            db::DCplxTrans tt = (*tv_list) [0] * cv.context_trans () * db::CplxTrans (cv->layout ().dbu ());
            db::CplxTrans move_trans_dbu (tt.inverted () * db::DCplxTrans (move_trans_inst) * tt);

            std::sort (insts_to_transform.begin (), insts_to_transform.end ());
            std::vector <std::pair <db::Instance, db::CplxTrans> >::const_iterator unique_end = std::unique (insts_to_transform.begin (), insts_to_transform.end ());
            db::Cell &cell = cv->layout ().cell (ibc->first.first);
            for (std::vector <std::pair <db::Instance, db::CplxTrans> >::const_iterator inst = insts_to_transform.begin (); inst != unique_end; ++inst) {

              db::ICplxTrans mt (inst->second.inverted () * move_trans_dbu * inst->second);
              cell.transform (inst->first, mt);

            }

          }

        }

      }

      manager ()->commit ();

    }

    if (! m_keep_selection) {
      m_selection.clear ();
    }

    m_dragging = false;
    selection_to_view ();

    return true;

  } else if (widget ()->mouse_event_viewport ().contains (p)) { 

    lay::Editable::SelectionMode mode = lay::Editable::Replace;
    bool shift = ((buttons & lay::ShiftButton) != 0);
    bool ctrl = ((buttons & lay::ControlButton) != 0);
    if (shift && ctrl) {
      mode = lay::Editable::Invert;
    } else if (shift) {
      mode = lay::Editable::Add;
    } else if (ctrl) {
      mode = lay::Editable::Reset;
    } 

    //  select is allowed to throw an exception 
    try {

      //  compute search box
      double l = double (search_range) / widget ()->mouse_event_trans ().mag ();
      db::DBox search_box = db::DBox (p, p).enlarged (db::DPoint (l, l));

      //  check, if there is a selected shape under the mouse - in this case, we do not do a new selection
      PartialShapeFinder finder (true /*point mode*/, m_top_level_sel, db::ShapeIterator::All);
      finder.find (view (), search_box);

      //  check, if there is a selected shape under the mouse - in this case, we do not do a new selection
      InstFinder inst_finder (true /*point mode*/, m_top_level_sel, true /*full arrays*/);
      inst_finder.find (view (), search_box);

      //  collect the founds from the finder
      //  consider a new selection if new objects are selected or the current selection is shape-only 
      //  (this may happen if points have been inserted)
      bool new_selection = ((finder.begin () == finder.end () && inst_finder.begin () == inst_finder.end ()) 
                             || mode != lay::Editable::Replace);

      for (PartialShapeFinder::iterator f = finder.begin (); f != finder.end () && ! new_selection; ++f) {
        partial_objects::const_iterator sel = m_selection.find (f->first);
        if (sel == m_selection.end () || sel->second.empty ()) {
          new_selection = true;
        }
      }

      if (finder.begin () == finder.end ()) {

        for (InstFinder::iterator f = inst_finder.begin (); f != inst_finder.end () && ! new_selection; ++f) {
          partial_objects::const_iterator sel = m_selection.find (*f);
          if (sel == m_selection.end ()) {
            new_selection = true;
          }
        }

      }

      if (new_selection) {

        if (mode == lay::Editable::Replace) {
          m_selection.clear ();
        }

        //  collect the founds from the finder
        for (PartialShapeFinder::iterator f = finder.begin (); f != finder.end (); ++f) {

          if (mode == lay::Editable::Replace || mode == lay::Editable::Add) {
            //  select
            partial_objects::iterator sel = m_selection.find (f->first);
            if (sel == m_selection.end ()) {
              sel = m_selection.insert (std::make_pair (f->first, std::set <EdgeWithIndex> ())).first;
            } 
            sel->second.insert (f->second.begin (), f->second.end ());
          } else if (mode == lay::Editable::Reset) {
            //  unselect
            if (m_selection.find (f->first) != m_selection.end ()) {
              m_selection.erase (f->first);
            }
          } else {
            //  invert selection
            if (m_selection.find (f->first) != m_selection.end ()) {
              m_selection.erase (f->first);
            } else {
              m_selection.insert (std::make_pair (f->first, std::set<EdgeWithIndex> ())).first->second.insert (f->second.begin (), f->second.end ());
            }
          }

        }

      }
      
      //  start dragging with that single selection
      if (mode == lay::Editable::Replace && ! m_selection.empty ()) {
        m_dragging = true;
        m_keep_selection = ! new_selection;
        m_current = m_start = p;
      }

      selection_to_view ();

    } catch (tl::Exception &ex) {
      tl::error << ex.msg ();
      QMessageBox::critical (0, tl::translate ("Error").c_str (), ex.msg ().c_str ());
      //  clear selection
      partial_select (db::DBox (), lay::Editable::Reset);
    }

    return true;

  }

  return false;
}

bool  
PartialService::mouse_double_click_event (const db::DPoint &p, unsigned int buttons, bool prio)
{
  if (! view ()->is_editable ()) {
    return false;
  }

  if ((buttons & lay::LeftButton) != 0 && prio) {

    //  stop dragging
    widget ()->ungrab_mouse (this);
    m_dragging = false;
      
    partial_select (db::DBox (p, p), lay::Editable::Replace);

    if (! m_selection.empty ()) {

      partial_objects::iterator r = m_selection.begin (); // we assert above that we have at least one selected element
      if (! r->first.is_cell_inst ()) {

        manager ()->transaction (tl::translate ("Insert point")); 

        //  snap the point
        db::DPoint new_point_d = snap (p);

        //  build the transformation variants cache
        TransformationVariants tv (view (), true /*per cv and layer*/, false /*per cv*/); 

        const std::vector<db::CplxTrans> *tv_list = tv.per_cv_and_layer (r->first.cv_index (), r->first.layer ());
        if (tv_list && ! tv_list->empty ()) {

          const lay::CellView &cv = view ()->cellview (r->first.cv_index ());

          db::DCplxTrans tt = (*tv_list) [0] * db::CplxTrans (cv->layout ().dbu ()) * (cv.context_trans () * r->first.trans ());
          db::Point new_point = db::Point::from_double (tt.inverted () * new_point_d);

          //  modify the shapes and replace

          db::Shapes &shapes = cv->layout ().cell (r->first.cell_index ()).shapes (r->first.layer ());
          db::GenericRepository &rep = cv->layout ().shape_repository ();

          if (r->first.shape ().is_polygon ()) {

            db::Polygon poly;
            r->first.shape ().polygon (poly);

            db::Polygon new_poly;
            if (insert_point_poly (poly, r->second, new_point, new_poly)) {
              shapes.replace (r->first.shape (), rep, new_poly);
            }

            r->second.clear ();

          } else if (r->first.shape ().is_path ()) {

            db::Path path;
            r->first.shape ().path (path);

            db::Path new_path;
            if (insert_point_path (path, r->second, new_point, new_path)) {
              shapes.replace (r->first.shape (), rep, new_path);
            }

            r->second.clear ();

          } else if (r->first.shape ().is_box ()) {

            //  convert the box into a polygon

            db::Polygon poly (r->first.shape ().box ());

            db::Polygon new_poly;
            if (insert_point_poly (poly, r->second, new_point, new_poly)) {

              ObjectInstPath obj = r->first;

              shapes.erase_shape (r->first.shape (), false /*just this instance*/);

              obj.set_shape (db::Shape (shapes.basic_insert (db::PolygonWithProperties (new_poly, r->first.shape ().prop_id ()))));

              m_selection.clear ();
              m_selection.insert (std::make_pair (obj, std::set<EdgeWithIndex> ()));

            }

          }

        }

        //  in case there was more than one object selected: restrict the selection to the first one.
        r = m_selection.begin (); 
        ++r;
        m_selection.erase (r, m_selection.end ());

        manager ()->commit ();

        selection_to_view ();

      }

    }

    return true;

  } else {
    return false;
  }
}

bool  
PartialService::mouse_release_event (const db::DPoint &p, bool prio)
{
  if (prio && mp_box) {

    widget ()->ungrab_mouse (this);

    delete mp_box;
    mp_box = 0;

    if (widget ()->mouse_event_viewport ().contains (p)) { 

      lay::Editable::SelectionMode mode = lay::Editable::Replace;
      bool shift = ((m_buttons & lay::ShiftButton) != 0);
      bool ctrl = ((m_buttons & lay::ControlButton) != 0);
      if (shift && ctrl) {
        mode = lay::Editable::Invert;
      } else if (shift) {
        mode = lay::Editable::Add;
      } else if (ctrl) {
        mode = lay::Editable::Reset;
      } 

      //  select is allowed to throw an exception 
      try {
        partial_select (db::DBox (m_p1, m_p2), mode);
      } catch (tl::Exception &ex) {
        tl::error << ex.msg ();
        QMessageBox::critical (0, tl::translate ("Error").c_str (), ex.msg ().c_str ());
        //  clear selection
        partial_select (db::DBox (), lay::Editable::Reset);
      }

    }

    return true;

  }

  return false;
}

bool 
PartialService::has_selection ()
{
  return ! m_selection.empty ();
}

void 
PartialService::del ()
{
  //  stop dragging
  widget ()->ungrab_mouse (this);
  
  std::map <std::pair <db::Layout::cell_index_type, std::pair <unsigned int, unsigned int> >, std::vector <partial_objects::const_iterator> > shapes_to_delete_by_cell;

  for (partial_objects::iterator r = m_selection.begin (); r != m_selection.end (); ++r) {

    if (! r->first.is_cell_inst ()) {

      const lay::CellView &cv = view ()->cellview (r->first.cv_index ());

      //  modify the shapes and replace

      db::Shapes &shapes = cv->layout ().cell (r->first.cell_index ()).shapes (r->first.layer ());
      db::GenericRepository &rep = cv->layout ().shape_repository ();

      if (r->first.shape ().is_polygon ()) {

        db::Polygon poly;
        r->first.shape ().polygon (poly);

        db::Polygon new_poly = del_points_poly (poly, r->second);
        if (new_poly.hull ().size () < 3) {
          shapes_to_delete_by_cell.insert (std::make_pair (std::make_pair (r->first.cell_index (), std::make_pair (r->first.cv_index (), r->first.layer ())), std::vector <partial_objects::const_iterator> ())).first->second.push_back (r);
        } else {
          shapes.replace (r->first.shape (), rep, new_poly);
        }

      } else if (r->first.shape ().is_path ()) {

        db::Path path;
        r->first.shape ().path (path);

        db::Path new_path = del_points_path (path, r->second);
        if (new_path.points () < 2) {
          shapes_to_delete_by_cell.insert (std::make_pair (std::make_pair (r->first.cell_index (), std::make_pair (r->first.cv_index (), r->first.layer ())), std::vector <partial_objects::const_iterator> ())).first->second.push_back (r);
        } else {
          shapes.replace (r->first.shape (), rep, new_path);
        }

      } else if (r->first.shape ().is_box ()) {

        //  if more than one point is deleted, the box basically collapses, if one point is deleted
        //  nothing changes on the box.
        if (r->second.size () > 1) {
          shapes_to_delete_by_cell.insert (std::make_pair (std::make_pair (r->first.cell_index (), std::make_pair (r->first.cv_index (), r->first.layer ())), std::vector <partial_objects::const_iterator> ())).first->second.push_back (r);
        }

      } else if (r->first.shape ().is_text ()) {

        shapes_to_delete_by_cell.insert (std::make_pair (std::make_pair (r->first.cell_index (), std::make_pair (r->first.cv_index (), r->first.layer ())), std::vector <partial_objects::const_iterator> ())).first->second.push_back (r);

      }

    }

  }

  //  delete all shapes that are really lost
  std::vector <db::Shape> shapes_to_delete;
  for (std::map <std::pair <db::Layout::cell_index_type, std::pair <unsigned int, unsigned int> >, std::vector <partial_objects::const_iterator> >::const_iterator sbc = shapes_to_delete_by_cell.begin (); sbc != shapes_to_delete_by_cell.end (); ++sbc) {

    shapes_to_delete.clear ();
    shapes_to_delete.reserve (sbc->second.size ());
    for (std::vector <partial_objects::const_iterator>::const_iterator s = sbc->second.begin (); s != sbc->second.end (); ++s) {
      shapes_to_delete.push_back ((*s)->first.shape ());
    }

    const lay::CellView &cv = view ()->cellview (sbc->first.second.first);
    if (cv.is_valid ()) {
      std::sort (shapes_to_delete.begin (), shapes_to_delete.end ());
      cv->layout ().cell (sbc->first.first).shapes (sbc->first.second.second).erase_shapes (shapes_to_delete, true /*whole array*/);
    }

  }
  
  //  then delete all instances.
  
  //  sort the selected objects (the instances) by the cell they are in
  //  The key is a pair: cell_index, cv_index
  std::map <std::pair <db::Layout::cell_index_type, unsigned int>, std::vector <partial_objects::const_iterator> > insts_by_cell;
  for (partial_objects::const_iterator r = m_selection.begin (); r != m_selection.end (); ++r) {
    if (r->first.is_cell_inst ()) {
      insts_by_cell.insert (std::make_pair (std::make_pair (r->first.cell_index (), r->first.cv_index ()), std::vector <partial_objects::const_iterator> ())).first->second.push_back (r);
    }
  }

  std::vector <db::Instance> insts_to_delete;
  for (std::map <std::pair <db::Layout::cell_index_type, unsigned int>, std::vector <partial_objects::const_iterator> >::const_iterator ibc = insts_by_cell.begin (); ibc != insts_by_cell.end (); ++ibc) {

    insts_to_delete.clear ();
    insts_to_delete.reserve (ibc->second.size ());
    for (std::vector <partial_objects::const_iterator>::const_iterator i = ibc->second.begin (); i != ibc->second.end (); ++i) {
      insts_to_delete.push_back ((*i)->first.back ().inst_ptr);
    }

    const lay::CellView &cv = view ()->cellview (ibc->first.second);
    if (cv.is_valid ()) {
      std::sort (insts_to_delete.begin (), insts_to_delete.end ());
      cv->layout ().cell (ibc->first.first).erase_insts (insts_to_delete);
    }

  }

  m_selection.clear ();
  m_dragging = false;
  selection_to_view ();
}

lay::InstanceMarker *
PartialService::new_inst_marker (size_t &nmarker, unsigned int cv_index)
{
  lay::InstanceMarker *marker;

  if (nmarker >= m_inst_markers.size ()) {
    marker = new lay::InstanceMarker (view (), cv_index);
    m_inst_markers.push_back (marker);
  } else {
    marker = m_inst_markers [nmarker];
  }

  ++nmarker;

  return marker;
}

lay::Marker *
PartialService::new_marker (size_t &nmarker, unsigned int cv_index)
{
  lay::Marker *marker;

  if (nmarker >= m_markers.size ()) {
    marker = new lay::Marker (view (), cv_index);
    m_markers.push_back (marker);
  } else {
    marker = m_markers [nmarker];
  }

  ++nmarker;

  marker->set_vertex_size (-1 /*default*/);
  marker->set_dither_pattern (1 /*hollow*/);
  marker->set_frame_pattern (0 /*solid*/);
  marker->set_line_width (-1 /*default*/);
  marker->set_halo (-1 /*default*/);

  return marker;
}

void 
PartialService::enter_path (db::Path &p, size_t &nmarker, partial_objects::const_iterator sel, const std::map <PointWithIndex, db::Point> &new_points, const std::map <EdgeWithIndex, db::Edge> &new_edges, const db::CplxTrans &gt, const std::vector<db::CplxTrans> &tv)
{
  lay::Marker *marker = new_marker (nmarker, sel->first.cv_index ());

  marker->set_dither_pattern (3 /*dotted*/);
  marker->set_frame_pattern (2 /*dotted*/);
  marker->set_line_width (1);
  marker->set_halo (0);
  modify_path (p, new_points, new_edges);
  marker->set (p, gt, tv);
}

void 
PartialService::enter_polygon (db::Polygon &p, size_t &nmarker, partial_objects::const_iterator sel, const std::map <PointWithIndex, db::Point> &new_points, const std::map <EdgeWithIndex, db::Edge> &new_edges, const db::CplxTrans &gt, const std::vector<db::CplxTrans> &tv)
{
  lay::Marker *marker = new_marker (nmarker, sel->first.cv_index ());

  marker->set_dither_pattern (3 /*dotted*/);
  marker->set_frame_pattern (2 /*dotted*/);
  marker->set_line_width (1);
  marker->set_halo (0);
  modify_polygon (p, new_points, new_edges);
  marker->set (p, gt, tv);
}

void 
PartialService::enter_vertices (size_t &nmarker, partial_objects::const_iterator sel, const std::map <PointWithIndex, db::Point> &new_points, const std::map <EdgeWithIndex, db::Edge> &new_edges, const db::CplxTrans &gt, const std::vector<db::CplxTrans> &tv)
{
  //  TODO: create vertex markers only for vertices that are not for an edge 
  //  and use "fat" vertices on the edge markers.

  for (std::set <EdgeWithIndex>::const_iterator e = sel->second.begin (); e != sel->second.end (); ++e) {

    if (e->p1 () == e->p2 ()) {
  
      lay::Marker *marker = new_marker (nmarker, sel->first.cv_index ());

      db::Point pnew = e->p1 ();
      std::map <PointWithIndex, db::Point>::const_iterator np = new_points.find (PointWithIndex (pnew, e->n));
      if (np != new_points.end ()) {
        pnew = np->second;
      }

      marker->set (db::Edge (pnew, pnew), gt, tv);

    }

  }

}

void 
PartialService::enter_edge (const EdgeWithIndex &e, size_t &nmarker, partial_objects::const_iterator sel, const std::map <PointWithIndex, db::Point> &new_points, const std::map <EdgeWithIndex, db::Edge> &new_edges, const db::CplxTrans &gt, const std::vector<db::CplxTrans> &tv)
{
  db::Point ep1 (e.p1 ());
  db::Point ep2 (e.p2 ());

  bool p1_sel = sel->second.find (EdgeWithIndex (db::Edge (ep1, ep1), e.n, e.n)) != sel->second.end ();
  bool p2_sel = sel->second.find (EdgeWithIndex (db::Edge (ep2, ep2), e.nn, e.nn)) != sel->second.end ();
  bool p12_sel = sel->second.find (e) != sel->second.end ();
  
  if (p1_sel || p2_sel || p12_sel) {

    //  map points to moved ones
    std::map <PointWithIndex, db::Point>::const_iterator np;
    np = new_points.find (e.pi1 ());
    if (np != new_points.end ()) {
      ep1 = np->second;
    }
    np = new_points.find (e.pi2 ());
    if (np != new_points.end ()) {
      ep2 = np->second;
    }

    db::Edge enew (ep1, ep2);

    std::map <EdgeWithIndex, db::Edge>::const_iterator ne;
    ne = new_edges.find (e);
    if (ne != new_edges.end ()) {

      enew = ne->second;

      if (enew.p1 () != ep1) {
        lay::Marker *marker = new_marker (nmarker, sel->first.cv_index ());
        marker->set_vertex_size (0);
        marker->set (db::Edge (ep1, enew.p1 ()), gt, tv);
      }

      if (enew.p2 () != ep2) {
        lay::Marker *marker = new_marker (nmarker, sel->first.cv_index ());
        marker->set_vertex_size (0);
        marker->set (db::Edge (enew.p2 (), ep2), gt, tv);
      }

    }

    if (p2_sel && !p12_sel) {

      lay::Marker *marker = new_marker (nmarker, sel->first.cv_index ());
      marker->set_vertex_size (0);

      db::Edge ee (db::Point::from_double ((ep1 - ep2) * 0.25 + ep2), ep2);
      marker->set (ee, gt, tv);

    } 

    if (p1_sel && !p12_sel) {

      lay::Marker *marker = new_marker (nmarker, sel->first.cv_index ());
      marker->set_vertex_size (0);

      db::Edge ee (ep1, db::Point::from_double ((ep2 - ep1) * 0.25 + ep1));
      marker->set (ee, gt, tv);

    } 

    if (p12_sel) {
      lay::Marker *marker = new_marker (nmarker, sel->first.cv_index ());
      marker->set_vertex_size (0);
      marker->set (enew, gt, tv);
    }

  }

}

void 
PartialService::selection_to_view ()
{
  //  if dragging, establish the current displacement
  db::DTrans move_trans;
  if (m_dragging) {
    //  heuristically, if there is just one edge selected: do not confine to the movement
    //  angle constraint - the edge usually is confined enough
    if (m_selection.size () == 1 && m_selection.begin ()->second.size () == 3 /*p1,p2,edge*/) {
      move_trans = db::DTrans (m_current - m_start);
    } else {
      //  TODO: DTrans should have a ctor that takes a vector
      move_trans = db::DTrans (db::DPoint () + edt::snap_angle (m_current - m_start, m_move_ac));
    }
  }

  //  build the transformation variants cache
  TransformationVariants tv (view ());

  size_t n_marker = 0;
  size_t n_inst_marker = 0;

  for (partial_objects::const_iterator r = m_selection.begin (); r != m_selection.end (); ++r) {

    const lay::CellView &cv = view ()->cellview (r->first.cv_index ());

    if (! r->first.is_cell_inst ()) {

      const std::vector<db::CplxTrans> *tv_list = tv.per_cv_and_layer (r->first.cv_index (), r->first.layer ());
      if (tv_list && !tv_list->empty ()) {

        //  use only the first one of the explicit transformations 
        //  TODO: clarify how this can be implemented in a more generic form or leave it thus.
        db::CplxTrans gt (cv.context_trans () * r->first.trans ());
        db::DCplxTrans tt = (*tv_list) [0] * db::CplxTrans (cv->layout ().dbu ()) * gt;
        db::Vector move_vector (db::Point::from_double (tt.inverted () * (move_trans * (tt * db::DPoint ()))));

        //  create the shift sets describing how points and edges are being moved
        
        std::map <EdgeWithIndex, db::Edge> new_edges;
        std::map <PointWithIndex, db::Point> new_points;

        if (m_dragging) {
          create_shift_sets (r, new_points, new_edges, move_vector);
        }

        //  create the markers to represent vertices and edges

        enter_vertices (n_marker, r, new_points, new_edges, gt, *tv_list);

        if (r->first.shape ().is_polygon ()) {

          unsigned int n = 0;
          db::Shape::polygon_edge_iterator ee;
          for (db::Shape::polygon_edge_iterator e = r->first.shape ().begin_edge (); ! e.at_end (); e = ee, ++n) {
            ee = e;
            ++ee;
            unsigned int nn = ee.at_end () ? 0 : n + 1;
            enter_edge (EdgeWithIndex (*e, n, nn), n_marker, r, new_points, new_edges, gt, *tv_list);
          }

          db::Polygon poly;
          r->first.shape ().polygon (poly);

          //  warning: poly is modified:
          enter_polygon (poly, n_marker, r, new_points, new_edges, gt, *tv_list);

        } else if (r->first.shape ().is_path ()) {

          if (r->first.shape ().begin_point () != r->first.shape ().end_point ()) {

            db::Shape::point_iterator pt = r->first.shape ().begin_point ();
            db::Point p1 = *pt;

            unsigned int n = 0;
            while (true) {

              ++pt;
              if (pt == r->first.shape ().end_point ()) {
                break;
              }
              
              enter_edge (EdgeWithIndex (db::Edge (p1, *pt), n, n + 1), n_marker, r, new_points, new_edges, gt, *tv_list);

              p1 = *pt;
              ++n;

            }

            //  ... put this somewhere else: @@@
            db::Path path;
            r->first.shape ().path (path);

            //  warning: path is modified:
            enter_path (path, n_marker, r, new_points, new_edges, gt, *tv_list);

          }

        } else if (r->first.shape ().is_box ()) {

          //  convert to polygon and test those edges
          db::Polygon poly (r->first.shape ().box ());
          unsigned int n = 0;
          db::Shape::polygon_edge_iterator ee;
          for (db::Shape::polygon_edge_iterator e = poly.begin_edge (); ! e.at_end (); e = ee, ++n) {
            ee = e;
            ++ee;
            unsigned int nn = ee.at_end () ? 0 : n + 1;
            enter_edge (EdgeWithIndex (*e, n, nn), n_marker, r, new_points, new_edges, gt, *tv_list);
          }

          //  warning: poly is modified:
          enter_polygon (poly, n_marker, r, new_points, new_edges, gt, *tv_list);

        } else if (r->first.shape ().is_text ()) {

          db::Point tp (r->first.shape ().text_trans () * db::Point ());
          enter_edge (EdgeWithIndex (db::Edge (tp, tp), 0, 0), n_marker, r, new_points, new_edges, gt, *tv_list);

        }

      }

    } else {

      //  compute the global transformation including movement, context and explicit transformation
      db::CplxTrans gt = db::CplxTrans (db::Trans (db::Point::from_double (move_trans.disp () * (1.0 / cv->layout ().dbu ()))) * db::Trans (move_trans.fp_trans ()));
      gt *= (cv.context_trans () * r->first.trans ());

      const std::vector<db::CplxTrans> *tv_list = tv.per_cv (r->first.cv_index ());
      if (tv_list  && ! tv_list->empty ()) {
        lay::InstanceMarker *marker = new_inst_marker (n_inst_marker, r->first.cv_index ());
        marker->set (r->first.back ().inst_ptr, gt, *tv_list);
      }

    }

  }

  //  delete superfluous markers
  resize_markers (n_marker);
  resize_inst_markers (n_inst_marker);
}

void 
PartialService::resize_markers (size_t n)
{
  for (std::vector<lay::Marker *>::iterator r = m_markers.begin () + n; r != m_markers.end (); ++r) {
    delete *r;
  }
  m_markers.erase (m_markers.begin () + n, m_markers.end ());
}

void 
PartialService::resize_inst_markers (size_t n)
{
  for (std::vector<lay::InstanceMarker *>::iterator r = m_inst_markers.begin () + n; r != m_inst_markers.end (); ++r) {
    delete *r;
  }
  m_inst_markers.erase (m_inst_markers.begin () + n, m_inst_markers.end ());
}

bool 
PartialService::partial_select (const db::DBox &box, lay::Editable::SelectionMode mode)
{
  //  compute search box
  double l = double (search_range) / widget ()->mouse_event_trans ().mag ();
  db::DBox search_box = box.enlarged (db::DPoint (l, l));

  bool needs_update = false;
  bool any_selected = false;

  //  clear before unless "add" is selected
  if (mode == lay::Editable::Replace) {
    if (! m_selection.empty ()) {
      m_selection.clear ();
      needs_update = true;
    }
  }

  if (box.empty ()) {

    //  unconditional selection
    if (mode == lay::Editable::Reset) {
      if (! m_selection.empty ()) {
        m_selection.clear ();
        needs_update = true;
      }
    } else {

      //  extract all shapes

      //  @@@ not implemented yet @@@

    }

  } else {

    PartialShapeFinder finder (box.is_point (), m_top_level_sel, db::ShapeIterator::All);
    finder.find (view (), search_box);

    //  collect the founds from the finder
    for (PartialShapeFinder::iterator f = finder.begin (); f != finder.end (); ++f) {

      if (mode == lay::Editable::Replace || mode == lay::Editable::Add) {
        //  select
        partial_objects::iterator sel = m_selection.find (f->first);
        if (sel == m_selection.end ()) {
          sel = m_selection.insert (std::make_pair (f->first, std::set <EdgeWithIndex> ())).first;
        } 
        sel->second.insert (f->second.begin (), f->second.end ());
      } else if (mode == lay::Editable::Reset) {
        //  unselect
        if (m_selection.find (f->first) != m_selection.end ()) {
          m_selection.erase (f->first);
        }
      } else {
        //  invert selection
        if (m_selection.find (f->first) != m_selection.end ()) {
          m_selection.erase (f->first);
        } else {
          m_selection.insert (std::make_pair (f->first, std::set <EdgeWithIndex> ())).first->second.insert (f->second.begin (), f->second.end ());
        }
      }

      needs_update = true;
      any_selected = true;

    }

    //  check, if there is a selected instance inside the box - in this case, we do not do a new selection
    if (! box.is_point ()) {

      InstFinder inst_finder (box.is_point (), m_top_level_sel, true /*full arrays*/);
      inst_finder.find (view (), search_box);

      //  collect the founds from the finder
      for (InstFinder::iterator f = inst_finder.begin (); f != inst_finder.end (); ++f) {

        if (mode == lay::Editable::Replace || mode == lay::Editable::Add) {
          //  select
          m_selection.insert (std::make_pair (*f, std::set <EdgeWithIndex> ()));
        } else if (mode == lay::Editable::Reset) {
          //  unselect
          if (m_selection.find (*f) != m_selection.end ()) {
            m_selection.erase (*f);
          }
        } else {
          //  invert selection
          if (m_selection.find (*f) != m_selection.end ()) {
            m_selection.erase (*f);
          } else {
            m_selection.insert (std::make_pair (*f, std::set <EdgeWithIndex> ()));
          }
        }

        needs_update = true;
        any_selected = true;

      }

    }
      
  }

  //  if required, update the list of ruler objects to display the selection
  if (needs_update) {
    selection_to_view ();
  }

  return any_selected;
}

} // namespace edt


