
/*

  KLayout Layout Viewer
  Copyright (C) 2006-2007 Matthias Koefferlein

  This program is free software; you can redistribute it and/or modify
  it under the terms of the GNU General Public License as published by
  the Free Software Foundation; either version 2 of the License, or
  (at your option) any later version.

  This program is distributed in the hope that it will be useful,
  but WITHOUT ANY WARRANTY; without even the implied warranty of
  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
  GNU General Public License for more details.

  You should have received a copy of the GNU General Public License
  along with this program; if not, write to the Free Software
  Foundation, Inc., 51 Franklin St, Fifth Floor, Boston, MA  02110-1301  USA

*/


#include "dbClipboard.h"
#include "edtService.h"
#include "edtEditorOptionsPages.h"
#include "edtDialogs.h"
#include "edtFinder.h"
#include "edtSnap.h"
#include "layLayoutView.h"
#include "laySnap.h"
#include "tlProgress.h"

namespace edt
{

// -------------------------------------------------------------

//  search range for select in pixels
static unsigned int search_range = 5;   // @@@ make variable?

// -------------------------------------------------------------

Service::Service (db::Manager *manager, lay::LayoutView *view, db::ShapeIterator::flags_type flags)
  : lay::ViewService (view->view_object_widget ()), 
    lay::Editable (view),
    lay::Plugin (view),
    db::Object (manager),
    mp_view (view),
    mp_edit_marker (0), 
    m_editing (false), m_immediate (false), 
    m_cell_inst_service (false),
    m_flags (flags),
    m_line_width (1), m_vertex_size (3),
    m_move_sel (false), m_moving (false),
    m_connect_ac (edt::Any), m_move_ac (edt::Any)
{ 
  //  .. nothing yet ..
}

Service::Service (db::Manager *manager, lay::LayoutView *view)
  : lay::ViewService (view->view_object_widget ()), 
    lay::Editable (view),
    lay::Plugin (view),
    db::Object (manager),
    mp_view (view),
    mp_edit_marker (0), 
    m_editing (false), m_immediate (false), 
    m_cell_inst_service (true),
    m_flags (db::ShapeIterator::Nothing),
    m_line_width (1), m_vertex_size (3),
    m_move_sel (false), m_moving (false),
    m_connect_ac (edt::Any), m_move_ac (edt::Any),
    m_top_level_sel (false)
{ 
  //  .. nothing yet ..
}

Service::~Service ()
{
  for (std::vector<lay::ViewObject *>::iterator r = m_markers.begin (); r != m_markers.end (); ++r) {
    delete *r;
  }
  m_markers.clear ();
  set_edit_marker (0);
}

db::DPoint 
Service::snap (db::DPoint p, const db::DPoint &plast, bool connect) const
{
  p = plast + edt::snap_angle (db::DVector (p - plast), connect ? m_connect_ac : m_move_ac);

  //  snap according to the grid
  bool snap = true;
  if (m_edit_grid == db::DPoint ()) {
    p = lay::snap_xy (p, m_global_grid);
  } else if (m_edit_grid.x () < 1e-6) {
    snap = false;
  } else {
    p = lay::snap_xy (p, m_edit_grid);
  }

  return p;
}

bool
Service::configure (const std::string &name, const std::string &value)
{
  edt::EditGridConverter egc;
  edt::ACConverter acc;

  if (name == cfg_edit_global_grid) {
    egc.from_string (value, m_global_grid);
  } else if (name == cfg_edit_grid) {
    egc.from_string (value, m_edit_grid);
    return true;  //  taken
  } else if (name == cfg_edit_move_angle_mode) {
    acc.from_string (value, m_move_ac);
    return true;  //  taken
  } else if (name == cfg_edit_connect_angle_mode) {
    acc.from_string (value, m_connect_ac);
    return true;  //  taken
  } else if (name == cfg_edit_top_level_selection) {
    tl::from_string (value, m_top_level_sel);
  }

  return false;  //  not taken
}

void 
Service::clear_highlights ()
{
  for (std::vector<lay::ViewObject *>::iterator r = m_markers.begin (); r != m_markers.end (); ++r) {
    (*r)->visible (false);
  }
}

void 
Service::restore_highlights ()
{
  for (std::vector<lay::ViewObject *>::iterator r = m_markers.begin (); r != m_markers.end (); ++r) {
    (*r)->visible (true);
  }
}

void 
Service::highlight (unsigned int n)
{
  for (std::vector<lay::ViewObject *>::iterator r = m_markers.begin (); r != m_markers.end (); ++r) {
    (*r)->visible (n-- == 0);
  }
}

void 
Service::cut ()
{
  if (has_selection () && view ()->is_editable ()) {
    //  copy & delete the selected objects
    copy_selected ();
    del_selected ();
  }
}

void 
Service::copy ()
{
  if (view ()->is_editable ()) {
    //  copy the selected objects
    copy_selected ();
  }
}

void
Service::copy_selected ()
{
  edt::CopyModeDialog mode_dialog (view ());

  bool any_inst = false;
  for (objects::const_iterator r = m_selection.begin (); r != m_selection.end () && ! any_inst; ++r) {
    any_inst = r->is_cell_inst ();
  }

  unsigned int inst_mode = 0; 

  if (! any_inst || mode_dialog.exec (inst_mode)) {

    //  create one ClipboardData object per cv_index because, this one assumes that there is 
    //  only one source layout object.
    std::set <unsigned int> cv_indices;
    for (objects::const_iterator r = m_selection.begin (); r != m_selection.end (); ++r) {
      cv_indices.insert (r->cv_index ());
    }

    for (std::set <unsigned int>::const_iterator cvi = cv_indices.begin (); cvi != cv_indices.end (); ++cvi) {

      db::ClipboardValue<edt::ClipboardData> *cd = new db::ClipboardValue<edt::ClipboardData> ();

      //  add the selected objects to the clipboard data objects.
      const lay::CellView &cv = view ()->cellview (*cvi);
      for (objects::const_iterator r = m_selection.begin (); r != m_selection.end (); ++r) {
        if (r->cv_index () == *cvi) {
          if (! r->is_cell_inst ()) {
            cd->get ().add (cv->layout (), r->layer (), r->shape (), cv.context_trans () * r->trans ());
          } else {
            cd->get ().add (cv->layout (), r->back ().inst_ptr, inst_mode, cv.context_trans () * r->trans ());
          }
        }
      }

      db::Clipboard::instance () += cd;

    }

  }
}

bool  
Service::begin_move (bool sel, const db::DPoint &p)
{
  if (view ()->is_editable () && sel) {

    m_move_start = p;
    m_move_trans = db::DTrans ();
    m_move_sel = sel;
    m_moving = true;

    for (std::vector<lay::ViewObject *>::iterator r = m_markers.begin (); r != m_markers.end (); ++r) {
      (*r)->thaw ();
    }

  }

  return false;
}

void  
Service::move (const db::DPoint &pu)
{
  db::DPoint p = snap (pu - m_move_start, db::DPoint (), false /*move*/) + snap (m_move_start);
  if (view ()->is_editable () && m_moving) {
    move_markers (db::DTrans (p) * db::DTrans (m_move_trans.fp_trans ()) * db::DTrans (-snap (m_move_start)));
  }
}

void  
Service::move_transform (const db::DPoint &pu, db::fixpoint_trans tr)
{
  db::DPoint p = snap (pu - m_move_start, db::DPoint (), false /*move*/) + snap (m_move_start);
  if (view ()->is_editable () && m_moving) {
    move_markers (db::DTrans (p) * db::DTrans (m_move_trans.fp_trans () * tr) * db::DTrans (-snap (m_move_start)));
  }
}

void  
Service::end_move (const db::DPoint & /*p*/)
{
  if (view ()->is_editable () && m_moving) {
    transform (db::DCplxTrans (m_move_trans));
    move_cancel (); // formally this functionality fits here
  }
}

db::DBox  
Service::selection_bbox ()
{
  db::DBox box;
  for (objects::const_iterator r = m_selection.begin (); r != m_selection.end (); ++r) {
    const db::Layout &layout = view ()->cellview (r->cv_index ())->layout ();
    db::box_convert<db::CellInst> bc (layout);
    if (! r->is_cell_inst ()) {
      box += (r->trans () * r->shape ().bbox ()) * layout.dbu ();
    } else {
      box += (r->trans () * r->back ().bbox (bc)) * layout.dbu ();
    }
  }
  return box;
}

void 
Service::transform (const db::DCplxTrans &trans)
{
  //  HINT: sorting the selected shapes/instances ensures that a shape/instance is not moved twice.
  //  This may happen due to per-instance selection of lower-level shapes.

  //  first transform all shapes.
  
  //  sort the selected objects (the shapes) by the cell they are in
  //  The key is a triple: cell_index, cv_index, layer
  std::map <std::pair <db::Layout::cell_index_type, std::pair <unsigned int, unsigned int> >, std::vector <objects::const_iterator> > shapes_by_cell;
  for (objects::const_iterator r = m_selection.begin (); r != m_selection.end (); ++r) {
    if (! r->is_cell_inst ()) {
      shapes_by_cell.insert (std::make_pair (std::make_pair (r->cell_index (), std::make_pair (r->cv_index (), r->layer ())), std::vector <objects::const_iterator> ())).first->second.push_back (r);
    }
  }

  //  build the transformation variants cache
  //  TODO: move this into LayoutView, where the other transform variant things are?
  //  TODO: this is duplicate code!

  std::map <unsigned int, std::vector<db::CplxTrans> > per_cv_tv;
  std::map < std::pair<unsigned int, unsigned int>, std::vector<db::CplxTrans> > per_cv_and_layer_tv;

  for (lay::LayerPropertiesConstIterator l = view ()->begin_layers (); !l.at_end (); ++l) {

    if (! l->has_children ()) {

      unsigned int cvi = (l->cellview_index () >= 0) ? (unsigned int) l->cellview_index () : 0;
      std::vector<db::CplxTrans> &tv = per_cv_tv.insert (std::make_pair (cvi, std::vector<db::CplxTrans> ())).first->second;
      tv.insert (tv.end (), l->trans ().begin (), l->trans ().end ());

      if (l->layer_index () >= 0) {
        std::vector<db::CplxTrans> &tv = per_cv_and_layer_tv.insert (std::make_pair (std::make_pair (cvi, (unsigned int) l->layer_index ()), std::vector<db::CplxTrans> ())).first->second;
        tv.insert (tv.end (), l->trans ().begin (), l->trans ().end ());
      }

    }

  }

  //  remove duplicates of the list of transformations (not related to layers)
  for (std::map <unsigned int, std::vector<db::CplxTrans> >::iterator t = per_cv_tv.begin (); t != per_cv_tv.end (); ++t) {
    std::sort (t->second.begin (), t->second.end ());
    std::vector<db::CplxTrans>::iterator new_last = std::unique (t->second.begin (), t->second.end ());
    t->second.erase (new_last, t->second.end ());
  }

  //  remove duplicates of the list of transformations (related to layers)
  for (std::map < std::pair<unsigned int, unsigned int>, std::vector<db::CplxTrans> >::iterator t = per_cv_and_layer_tv.begin (); t != per_cv_and_layer_tv.end (); ++t) {
    std::sort (t->second.begin (), t->second.end ());
    std::vector<db::CplxTrans>::iterator new_last = std::unique (t->second.begin (), t->second.end ());
    t->second.erase (new_last, t->second.end ());
  }

  std::vector <std::pair <db::Shape, db::CplxTrans> > shapes_to_transform;
  for (std::map <std::pair <db::Layout::cell_index_type, std::pair <unsigned int, unsigned int> >, std::vector <objects::const_iterator> >::const_iterator sbc = shapes_by_cell.begin (); sbc != shapes_by_cell.end (); ++sbc) {

    const lay::CellView &cv = view ()->cellview (sbc->first.second.first);
    if (cv.is_valid ()) {

      shapes_to_transform.clear ();
      shapes_to_transform.reserve (sbc->second.size ());
      for (std::vector <objects::const_iterator>::const_iterator s = sbc->second.begin (); s != sbc->second.end (); ++s) {
        shapes_to_transform.push_back (std::make_pair ((*s)->shape (), cv.context_trans () * (*s)->trans ()));
      }

      std::map < std::pair<unsigned int, unsigned int>, std::vector<db::CplxTrans> >::const_iterator tv_list = per_cv_and_layer_tv.find (sbc->first.second);
      if (tv_list != per_cv_and_layer_tv.end ()) {

        db::DCplxTrans tt = tv_list->second [0] * db::CplxTrans (cv->layout ().dbu ());
        db::CplxTrans move_trans_dbu (tt.inverted () * trans * tt);

        db::GenericRepository &rep = cv->layout ().shape_repository ();

        std::sort (shapes_to_transform.begin (), shapes_to_transform.end ());

        std::vector <std::pair <db::Shape, db::CplxTrans> >::const_iterator unique_end = std::unique (shapes_to_transform.begin (), shapes_to_transform.end ());

        db::Shapes &shapes = cv->layout ().cell (sbc->first.first).shapes (sbc->first.second.second);
        for (std::vector <std::pair <db::Shape, db::CplxTrans> >::const_iterator shape = shapes_to_transform.begin (); shape != unique_end; ++shape) {
          //  mt = move transformation in DBU units
          db::ICplxTrans mt (shape->second.inverted () * move_trans_dbu * shape->second);
          shapes.transform (shape->first, rep, mt);
        }

      }

    }

  }
  
  //  then move all instances.
  
  //  sort the selected objects (the instances) by the cell they are in
  //  The key is a pair: cell_index, cv_index
  std::map <std::pair <db::Layout::cell_index_type, unsigned int>, std::vector <objects::const_iterator> > insts_by_cell;
  for (objects::const_iterator r = m_selection.begin (); r != m_selection.end (); ++r) {
    if (r->is_cell_inst ()) {
      insts_by_cell.insert (std::make_pair (std::make_pair (r->cell_index (), r->cv_index ()), std::vector <objects::const_iterator> ())).first->second.push_back (r);
    }
  }

  std::vector <std::pair <db::Instance, db::CplxTrans> > insts_to_transform;
  for (std::map <std::pair <db::Layout::cell_index_type, unsigned int>, std::vector <objects::const_iterator> >::const_iterator ibc = insts_by_cell.begin (); ibc != insts_by_cell.end (); ++ibc) {

    insts_to_transform.clear ();
    insts_to_transform.reserve (ibc->second.size ());
    for (std::vector <objects::const_iterator>::const_iterator i = ibc->second.begin (); i != ibc->second.end (); ++i) {
      insts_to_transform.push_back (std::make_pair ((*i)->back ().inst_ptr, (*i)->trans ()));
    }

    const lay::CellView &cv = view ()->cellview (ibc->first.second);
    if (cv.is_valid ()) {

      std::map <unsigned int, std::vector<db::CplxTrans> >::const_iterator tv_list = per_cv_tv.find (ibc->first.second);
      if (tv_list != per_cv_tv.end ()) {

        db::DCplxTrans tt = tv_list->second [0] * cv.context_trans () * db::CplxTrans (cv->layout ().dbu ());
        db::CplxTrans move_trans_dbu (tt.inverted () * trans * tt);

        std::sort (insts_to_transform.begin (), insts_to_transform.end ());
        std::vector <std::pair <db::Instance, db::CplxTrans> >::const_iterator unique_end = std::unique (insts_to_transform.begin (), insts_to_transform.end ());
        db::Cell &cell = cv->layout ().cell (ibc->first.first);
        for (std::vector <std::pair <db::Instance, db::CplxTrans> >::const_iterator inst = insts_to_transform.begin (); inst != unique_end; ++inst) {

          db::ICplxTrans mt (inst->second.inverted () * move_trans_dbu * inst->second);
          cell.transform (inst->first, mt);

        }

      }

    }

  }
}

void   
Service::move_cancel ()
{
  if (m_move_trans != db::DTrans () && m_moving) {

    for (std::vector<lay::ViewObject *>::iterator r = m_markers.begin (); r != m_markers.end (); ++r) {
      (*r)->freeze ();
    }

    m_move_trans = db::DTrans ();
    m_move_start = db::DPoint ();

    //  reset to unmoved or clear selection and do not do anything 
    if (m_move_sel) {
      selection_to_view ();
    } else {
      clear_selection ();
    }

    m_moving = false;

  }
}

bool   
Service::mouse_move_event (const db::DPoint &p, bool prio)
{
  if (view ()->is_editable () && (m_editing || m_immediate)) {
    if (! m_editing) {
      //  in this mode, ignore exceptions here since it is rather annoying to have messages popping
      //  up then.
      try {
        do_begin_edit (p);
        m_editing = true;
      } catch (...) {
        set_edit_marker (0);
      }
    }
    if (m_editing) {
      do_mouse_move (p);
    }
  }
  return false;  // not taken to allow the mouse tracker to receive events as well
}

bool   
Service::mouse_press_event (const db::DPoint &p, unsigned int buttons, bool prio)
{
  if (view ()->is_editable () && prio) {
    
    if ((buttons & lay::LeftButton) != 0) {

      if (! m_editing) {

        view ()->cancel ();  //  cancel any pending edit operations and clear the selection
        set_edit_marker (0);
        do_begin_edit (p);
        m_editing = true;

      } else {
        if (do_mouse_click (p)) {
          m_editing = false;
          set_edit_marker (0);
          do_finish_edit ();
        }
      }

      return true;

    }

  } 

  return false;
}

bool   
Service::mouse_double_click_event (const db::DPoint &p, unsigned int buttons, bool prio)
{
  if (m_editing && prio && (buttons & lay::LeftButton) != 0) {
    do_finish_edit ();
    m_editing = false;
    set_edit_marker (0);
    return true;
  } else {
    return false;
  }
}

bool   
Service::mouse_click_event (const db::DPoint &p, unsigned int buttons, bool prio)
{
  if (view ()->is_editable () && prio && (buttons & lay::RightButton) != 0 && m_editing) {
    do_mouse_transform (p, db::fixpoint_trans::r90);
    return true;
  } else {
    return mouse_press_event (p, buttons, prio);
  }
}

void
Service::activated ()
{
  //  make all editor option pages visible
  std::vector<edt::EditorOptionsPage *> ppages = get_editor_options_pages ();
  for (std::vector<edt::EditorOptionsPage *>::const_iterator pp = ppages.begin (); pp != ppages.end (); ++pp) {
    (*pp)->activate (true);
  }

  if (view ()->is_editable ()) {
    view ()->cancel ();  //  cancel any pending edit operations and clear the selection
    set_edit_marker (0);
    m_immediate = do_activated ();
    m_editing = false;
  }
}

void   
Service::deactivated ()
{
  //  make all editor option pages invisible
  std::vector<edt::EditorOptionsPage *> ppages = get_editor_options_pages ();
  for (std::vector<edt::EditorOptionsPage *>::const_iterator pp = ppages.begin (); pp != ppages.end (); ++pp) {
    (*pp)->activate (false);
  }

  edit_cancel ();

  m_immediate = false;
}

void   
Service::edit_cancel ()
{
  move_cancel ();
  if (m_editing) {
    do_cancel_edit ();
    m_editing = false;
    set_edit_marker (0);
  }
}

void 
Service::del ()
{
  if (has_selection () && view ()->is_editable ()) {
    //  delete the selected objects
    del_selected ();
  }
}

void
Service::del_selected ()
{
  //  first delete all shapes.
  
  //  sort the selected objects (the shapes) by the cell they are in
  //  The key is a triple: cell_index, cv_index, layer
  std::map <std::pair <db::Layout::cell_index_type, std::pair <unsigned int, unsigned int> >, std::vector <objects::const_iterator> > shapes_by_cell;
  for (objects::const_iterator r = m_selection.begin (); r != m_selection.end (); ++r) {
    if (! r->is_cell_inst ()) {
      shapes_by_cell.insert (std::make_pair (std::make_pair (r->cell_index (), std::make_pair (r->cv_index (), r->layer ())), std::vector <objects::const_iterator> ())).first->second.push_back (r);
    }
  }

  std::vector <db::Shape> shapes_to_delete;
  for (std::map <std::pair <db::Layout::cell_index_type, std::pair <unsigned int, unsigned int> >, std::vector <objects::const_iterator> >::const_iterator sbc = shapes_by_cell.begin (); sbc != shapes_by_cell.end (); ++sbc) {

    shapes_to_delete.clear ();
    shapes_to_delete.reserve (sbc->second.size ());
    for (std::vector <objects::const_iterator>::const_iterator s = sbc->second.begin (); s != sbc->second.end (); ++s) {
      shapes_to_delete.push_back ((*s)->shape ());
    }

    const lay::CellView &cv = view ()->cellview (sbc->first.second.first);
    if (cv.is_valid ()) {
      std::sort (shapes_to_delete.begin (), shapes_to_delete.end ());
      cv->layout ().cell (sbc->first.first).shapes (sbc->first.second.second).erase_shapes (shapes_to_delete, true /*whole array*/);
    }

  }
  
  //  then delete all instances.
  
  //  sort the selected objects (the instances) by the cell they are in
  //  The key is a pair: cell_index, cv_index
  std::map <std::pair <db::Layout::cell_index_type, unsigned int>, std::vector <objects::const_iterator> > insts_by_cell;
  for (objects::const_iterator r = m_selection.begin (); r != m_selection.end (); ++r) {
    if (r->is_cell_inst ()) {
      insts_by_cell.insert (std::make_pair (std::make_pair (r->cell_index (), r->cv_index ()), std::vector <objects::const_iterator> ())).first->second.push_back (r);
    }
  }

  std::vector <db::Instance> insts_to_delete;
  for (std::map <std::pair <db::Layout::cell_index_type, unsigned int>, std::vector <objects::const_iterator> >::const_iterator ibc = insts_by_cell.begin (); ibc != insts_by_cell.end (); ++ibc) {

    insts_to_delete.clear ();
    insts_to_delete.reserve (ibc->second.size ());
    for (std::vector <objects::const_iterator>::const_iterator i = ibc->second.begin (); i != ibc->second.end (); ++i) {
      insts_to_delete.push_back ((*i)->back ().inst_ptr);
    }

    const lay::CellView &cv = view ()->cellview (ibc->first.second);
    if (cv.is_valid ()) {
      std::sort (insts_to_delete.begin (), insts_to_delete.end ());
      cv->layout ().cell (ibc->first.first).erase_insts (insts_to_delete);
    }

  }
}

bool 
Service::has_selection ()
{
  return ! m_selection.empty ();
}

double 
Service::click_proximity (const db::DPoint &pos)
{
  //  compute search box
  double l = double (search_range) / widget ()->mouse_event_trans ().mag ();
  db::DBox search_box = db::DBox (pos, pos).enlarged (db::DPoint (l, l));

  if (m_cell_inst_service) {

    InstFinder finder (true, view ()->is_editable () && m_top_level_sel, view ()->is_editable () /*full arrays in editable mode*/);

    //  go through all cell views
    std::set< std::pair<db::CplxTrans, int> > variants = view ()->cv_transform_variants ();
    for (std::set< std::pair<db::CplxTrans, int> >::const_iterator v = variants.begin (); v != variants.end (); ++v) {
      finder.find (view (), v->second, v->first, search_box);
    }

    //  Return the finder's proximity value
    if (finder.begin () != finder.end ()) {
      return finder.proximity ();
    } else {
      return lay::Editable::click_proximity (pos); 
    }

  } else {

    ShapeFinder finder (true, view ()->is_editable () && m_top_level_sel, m_flags);

    //  go through all visible layers of all cellviews
    for (lay::LayerPropertiesConstIterator lprop = view ()->begin_layers (); ! lprop.at_end () ; ++lprop) {
      if (lprop->visible (true /*real*/)) {
        finder.find (view (), *lprop, search_box);
      }
    }

    //  Return the finder's proximity value
    if (finder.begin () != finder.end ()) {
      return finder.proximity ();
    } else {
      return lay::Editable::click_proximity (pos); 
    }

  }
}

bool 
Service::select (const db::DBox &box, lay::Editable::SelectionMode mode)
{
  //  compute search box
  double l = double (search_range) / widget ()->mouse_event_trans ().mag ();
  db::DBox search_box = box.enlarged (db::DPoint (l, l));

  bool needs_update = false;
  bool any_selected = false;

  //  clear before unless "add" is selected
  if (mode == lay::Editable::Replace) {
    if (! m_selection.empty ()) {
      m_selection.clear ();
      needs_update = true;
    }
  }

  if (box.empty ()) {

    //  unconditional selection
    if (mode == lay::Editable::Reset) {
      if (! m_selection.empty ()) {
        m_selection.clear ();
        needs_update = true;
      }
    } else {

      //  extract all shapes

      //  @@@ not implemented yet @@@

    }

  } else if (m_cell_inst_service) {

    InstFinder finder (box.is_point (), view ()->is_editable () && m_top_level_sel, view ()->is_editable () /*full arrays in editable mode*/);

    //  go through all cell views
    std::set< std::pair<db::CplxTrans, int> > variants = view ()->cv_transform_variants ();
    for (std::set< std::pair<db::CplxTrans, int> >::const_iterator v = variants.begin (); v != variants.end (); ++v) {
      finder.find (view (), v->second, v->first, search_box);
    }

    //  collect the founds from the finder
    for (InstFinder::iterator f = finder.begin (); f != finder.end (); ++f) {
      select (*f, mode);
      needs_update = true;
      any_selected = true;
    }

  } else {

    ShapeFinder finder (box.is_point (), view ()->is_editable () && m_top_level_sel, m_flags);

    //  go through all visible layers of all cellviews
    for (lay::LayerPropertiesConstIterator lprop = view ()->begin_layers (); ! lprop.at_end (); ++lprop) {
      if (lprop->visible (true /*real*/)) {
        finder.find (view (), *lprop, search_box);
      }
    }

    //  collect the founds from the finder
    for (ShapeFinder::iterator f = finder.begin (); f != finder.end (); ++f) {
      select (*f, mode);
      needs_update = true;
      any_selected = true;
    }
      
  }

  //  if required, update the list of ruler objects to display the selection
  if (needs_update) {
    selection_to_view ();
  }

  return any_selected;
}

void 
Service::get_selection (std::vector <ObjectInstPath> &sel) const
{
  sel.clear ();
  sel.reserve (m_selection.size ());

  //  positions will hold a set of iterators that are to be erased
  for (std::set<ObjectInstPath>::const_iterator r = m_selection.begin (); r != m_selection.end (); ++r) {
    sel.push_back (*r);
  }
}

bool 
Service::select (const ObjectInstPath &obj, lay::Editable::SelectionMode mode)
{
  if (mode == lay::Editable::Replace || mode == lay::Editable::Add) {
    //  select
    if (m_selection.find (obj) == m_selection.end ()) {
      m_selection.insert (obj);
      return true;
    }
  } else if (mode == lay::Editable::Reset) {
    //  unselect
    if (m_selection.find (obj) != m_selection.end ()) {
      m_selection.erase (obj);
      return true;
    }
  } else {
    //  invert selection
    if (m_selection.find (obj) != m_selection.end ()) {
      m_selection.erase (obj);
    } else {
      m_selection.insert (obj);
    }
    return true;
  }
  return false;
}

void 
Service::clear_selection ()
{
  select (db::DBox (), lay::Editable::Reset);
}

void  
Service::move_markers (const db::DTrans &t)
{
  if (m_move_trans != t) {

    for (std::vector<lay::ViewObject *>::iterator r = m_markers.begin (); r != m_markers.end (); ++r) {

      lay::GenericMarkerBase *marker = dynamic_cast<lay::GenericMarkerBase *> (*r);
      if (marker) {

        const lay::CellView &cv = view ()->cellview (marker->cv_index ());
        double f = 1.0 / cv->layout ().dbu ();
        db::CplxTrans dt = (db::CplxTrans (f) * db::CplxTrans (t)) * (db::CplxTrans (f) * db::CplxTrans (m_move_trans)).inverted ();
        marker->set (dt * marker->trans ());

      }

    }

    m_move_trans = t;

  }
}

void 
Service::selection_to_view ()
{
  //  the selection objects need to be recreated since we destroyed the old rulers
  for (std::vector<lay::ViewObject *>::iterator r = m_markers.begin (); r != m_markers.end (); ++r) {
    delete *r;
  }
  m_markers.clear ();

  //  Hint: this is a lower bound:
  m_markers.reserve (m_selection.size ());

  //  build the transformation variants cache
  //  TODO: move this into LayoutView, where the other transform variant things are?

  std::map <unsigned int, std::vector<db::CplxTrans> > per_cv_tv;
  std::map < std::pair<unsigned int, unsigned int>, std::vector<db::CplxTrans> > per_cv_and_layer_tv;

  for (lay::LayerPropertiesConstIterator l = view ()->begin_layers (); !l.at_end (); ++l) {

    if (! l->has_children ()) {

      unsigned int cvi = (l->cellview_index () >= 0) ? (unsigned int) l->cellview_index () : 0;
      std::vector<db::CplxTrans> &tv = per_cv_tv.insert (std::make_pair (cvi, std::vector<db::CplxTrans> ())).first->second;
      tv.insert (tv.end (), l->trans ().begin (), l->trans ().end ());

      if (l->layer_index () >= 0) {
        std::vector<db::CplxTrans> &tv = per_cv_and_layer_tv.insert (std::make_pair (std::make_pair (cvi, (unsigned int) l->layer_index ()), std::vector<db::CplxTrans> ())).first->second;
        tv.insert (tv.end (), l->trans ().begin (), l->trans ().end ());
      }

    }

  }

  //  remove duplicates of the list of transformations (not related to layers)
  for (std::map <unsigned int, std::vector<db::CplxTrans> >::iterator t = per_cv_tv.begin (); t != per_cv_tv.end (); ++t) {
    std::sort (t->second.begin (), t->second.end ());
    std::vector<db::CplxTrans>::iterator new_last = std::unique (t->second.begin (), t->second.end ());
    t->second.erase (new_last, t->second.end ());
  }

  //  remove duplicates of the list of transformations (related to layers)
  for (std::map < std::pair<unsigned int, unsigned int>, std::vector<db::CplxTrans> >::iterator t = per_cv_and_layer_tv.begin (); t != per_cv_and_layer_tv.end (); ++t) {
    std::sort (t->second.begin (), t->second.end ());
    std::vector<db::CplxTrans>::iterator new_last = std::unique (t->second.begin (), t->second.end ());
    t->second.erase (new_last, t->second.end ());
  }

  for (std::set<ObjectInstPath>::iterator r = m_selection.begin (); r != m_selection.end (); ++r) {

    const lay::CellView &cv = view ()->cellview (r->cv_index ());

    //  compute the global transformation including movement, context and explicit transformation
    db::CplxTrans gt = db::CplxTrans (db::Trans (db::Point::from_double (m_move_trans.disp () * (1.0 / cv->layout ().dbu ()))) * db::Trans (m_move_trans.fp_trans ()));
    gt *= (cv.context_trans () * r->trans ());

    tl_assert (r->is_cell_inst () == m_cell_inst_service);

    if (m_cell_inst_service) {

      std::map <unsigned int, std::vector<db::CplxTrans> >::const_iterator tv_list = per_cv_tv.find (r->cv_index ());
      if (tv_list != per_cv_tv.end ()) {
      
        if (view ()->is_editable ()) {
          lay::InstanceMarker *marker = new lay::InstanceMarker (view (), r->cv_index ());
          marker->set (r->back ().inst_ptr, gt, tv_list->second);
          m_markers.push_back (marker);
        } else {
          lay::Marker *marker = new lay::Marker (view (), r->cv_index ());
          db::box_convert<db::CellInst> bc (cv->layout ());
          marker->set (bc (r->back ().inst_ptr.cell_inst ().object ()), gt * r->back ().inst_ptr.cell_inst ().complex_trans (*r->back ().array_inst), tv_list->second);
          m_markers.push_back (marker);
        }

      }

    } else {

      std::map < std::pair<unsigned int, unsigned int>, std::vector<db::CplxTrans> >::const_iterator tv_list = per_cv_and_layer_tv.find (std::make_pair (r->cv_index (), r->layer ()));
      if (tv_list != per_cv_and_layer_tv.end ()) {

        lay::ShapeMarker *marker = new lay::ShapeMarker (view (), r->cv_index ());
        marker->set (r->shape (), gt, tv_list->second);
        m_markers.push_back (marker);

      }

    }

  }
}

void 
Service::set_selection (std::vector <ObjectInstPath>::const_iterator s1, std::vector <ObjectInstPath>::const_iterator s2)
{
  m_selection.clear ();
  m_selection.insert (s1, s2);
  selection_to_view ();
}

// -------------------------------------------------------------

} // namespace edt


