
/*

  KLayout Layout Viewer
  Copyright (C) 2006-2007 Matthias Koefferlein

  This program is free software; you can redistribute it and/or modify
  it under the terms of the GNU General Public License as published by
  the Free Software Foundation; either version 2 of the License, or
  (at your option) any later version.

  This program is distributed in the hope that it will be useful,
  but WITHOUT ANY WARRANTY; without even the implied warranty of
  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
  GNU General Public License for more details.

  You should have received a copy of the GNU General Public License
  along with this program; if not, write to the Free Software
  Foundation, Inc., 51 Franklin St, Fifth Floor, Boston, MA  02110-1301  USA

*/



#include "gsi.h"
#include "dbLayout.h"
#include "dbBoxConvert.h"

namespace gsi
{

// ---------------------------------------------------------------
//  db::Cell binding

static void erase_inst (db::Cell *c, size_t i)
{
  c->erase (c->begin (i));
}

static db::Shapes::shape_iterator begin_shapes (const db::Cell *s, unsigned int layer_index, unsigned int flags)
{
  return s->shapes (layer_index).begin (flags);
}

static db::Shapes::shape_iterator begin_touching_shapes (const db::Cell *s, unsigned int layer_index, const db::Box &box, unsigned int flags)
{
  return s->begin_touching (layer_index, box, flags);
}

static db::Shapes::shape_iterator begin_overlapping_shapes (const db::Cell *s, unsigned int layer_index, const db::Box &box, unsigned int flags)
{
  return s->begin_overlapping (layer_index, box, flags);
}

static db::Instance child_inst_c (const db::Cell *c, size_t i)
{
  return *(c->begin (i));
}

static void insert_inst_with_props (db::Cell *c, const db::Cell::cell_inst_array_type &inst, db::PropertiesRepository::properties_id_type id)
{
  c->insert (db::CellInstArrayWithProperties (inst, id));
}

template <>
Class<db::Cell> Class<db::Cell>::decl ("Cell", 
  gsi::method ("shapes", (db::Cell::shapes_type &(db::Cell::*) (unsigned int)) &db::Cell::shapes,
    "@brief Return the shapes list of the given layer\n"
    "\n"
    "This method allows to access the shapes list on a certain layer.\n"
    "If the layer does not exist yet, it is created.\n"
    "\n"
    "@param index The layer index of the shapes list to retrieve\n"
    "\n"
    "@return A reference to the shapes list\n"
  ) +
  gsi::method ("clear_shapes", &db::Cell::clear_shapes,
    "@brief Clear all shapes in the cell\n"
  ) +
  gsi::method ("clear_insts", &db::Cell::clear_insts,
    "@brief Clear the instance list\n"
  ) +
  gsi::method ("swap", &db::Cell::swap,
    "@brief Swap the layers given\n"
    "@args layer_index1, layer_index2\n"
  ) +
  gsi::method ("clear", &db::Cell::clear,
    "@brief Clear the shapes on the given layer\n"
    "@args layer_index\n"
  ) +
  gsi::method_ext ("erase_inst", &erase_inst,
    "@brief Erase a cell instance (array) with the given index\n"
    "@args index\n"
    "The index is the position at which the instance appears in the \"each_inst\" iteration."
  ) +
  gsi::method ("insert", (void (db::Cell::*)(const db::Cell::cell_inst_array_type &)) &db::Cell::insert,
    "@brief Insert a cell instance (array)\n"
    "@args cell_inst_array\n"
    "In version 0.15, this method no longer returns a reference to the object created."
  ) +
  gsi::method_ext ("insert", &insert_inst_with_props,
    "@brief Insert a cell instance (array) with properties\n"
    "@args cell_inst_array, property_id\n"
    "The property Id must be obtained from the \\Layout object's property_id method which "
    "associates a property set with a property Id."
  ) +
  gsi::method ("cell_index", &db::Cell::cell_index,
    "@brief The cell index accessor method\n"
    "\n"
    "@return The cell index of the cell\n"
  ) +
  gsi::method_ext ("child_inst", &child_inst_c,
    "@brief Child instance by index\n"
    "@args index\n"
    "\n"
    "@param index The index of the cell instance to return. Must be between 0 and the value returned by \\child_instances.\n"
    "@return Returns the \\Instance object pointing to the specified instance\n"
    "\n"
    "Starting with version 0.15, this method returns an \\Instance object rather than a \\CellInstArray reference."
  ) +
  gsi::method ("child_instances", &db::Cell::cell_instances,
    "@brief Number of child instances\n"
    "\n"
    "@return Returns the number of cell instances\n"
  ) +
  gsi::method ("bbox", (const db::Cell::box_type &(db::Cell::*) () const) &db::Cell::bbox,
    "@brief Retrieve the bounding box of the cell\n"
    "\n"
    "@return The bounding box of the cell\n"
  ) +
  gsi::method ("bbox_per_layer", (const db::Cell::box_type &(db::Cell::*) (unsigned int) const) &db::Cell::bbox,
    "@brief Retrieve the per-layer bounding box of the cell\n"
    "@args layer_index\n"
    "\n"
    "@return The bounding box of the cell considering only the given layer\n"
  ) +
/* Not enabled since we do not have per-layer instance trees.
  gsi::iterator ("each_overlapping_inst_per_layer", (db::Cell::overlapping_layer_iterator (db::Cell::*) (const db::Cell::box_type &, unsigned int) const) &db::Cell::begin_overlapping,
    "@brief Per-layer region query for the instances in \"overlapping\" mode\n"
    "@args b, l\n"
    "\n"
    "This will iterate over all child cell\n"
    "instances overlapping with the given region b for the given layer.\n"
    "The end iterator is given by end_overlapping ().\n"
    "\n"
    "@param b The region to query\n"
    "@param l The layer to do the query for\n"
  ) +
  gsi::iterator ("each_touching_inst_per_layer", (db::Cell::touching_layer_iterator (db::Cell::*) (const db::Cell::box_type &, unsigned int) const) &db::Cell::begin_touching,
    "@brief Per-layer region query for the instances in \"touching\" mode\n"
    "@args b, l\n"
    "\n"
    "This will iterate over all child cell\n"
    "instances touching the given region b for the given layer.\n"
    "The end iterator is given by end_touching ().\n"
    "\n"
    "@param b The region to query\n"
    "@param l The layer to do the query for\n"
  ) +
*/
  gsi::iterator ("each_overlapping_inst", (db::Cell::overlapping_iterator (db::Cell::*) (const db::Cell::box_type &b) const) &db::Cell::begin_overlapping,
    "@brief Region query for the instances in \"overlapping\" mode\n"
    "@args b\n"
    "\n"
    "This will iterate over all child cell\n"
    "instances overlapping with the given region b. \n"
    "\n"
    "@param b The region to query\n"
    "\n"
    "Starting with version 0.15, this iterator delivers \\Instance objects rather than \\CellInstArray objects."
  ) +
  gsi::iterator ("each_touching_inst", (db::Cell::touching_iterator (db::Cell::*) (const db::Cell::box_type &b) const) &db::Cell::begin_touching,
    "@brief Region query for the instances in \"touching\" mode\n"
    "@args b\n"
    "\n"
    "This will iterate over all child cell\n"
    "instances touching the given region b. \n"
    "\n"
    "@param b The region to query\n"
    "\n"
    "Starting with version 0.15, this iterator delivers \\Instance objects rather than \\CellInstArray objects."
  ) +
  gsi::iterator ("each_child_cell", &db::Cell::begin_child_cells, 
    "@brief Iterate over all child cells\n"
    "\n"
    "This iterator will report the child cell indices, not every instance.\n"
  ) +
  gsi::method ("child_cells", &db::Cell::child_cells,
    "@brief Report the number of child cells\n"
    "\n"
    "The number of child cells (not child instances!) is reported.\n"
    "CAUTION: this method is SLOW!\n"
  ) +
  gsi::iterator ("each_inst", (db::Cell::const_iterator (db::Cell::*) () const) &db::Cell::begin, 
    "@brief Iterate over all child instances (which may actually be instance arrays)\n"
    "\n"
    "Starting with version 0.15, this iterator delivers \\Instance objects rather than \\CellInstArray objects."
  ) +
  gsi::iterator ("each_parent_inst", &db::Cell::begin_parent_insts,
    "@brief Iterate over the parent instance list (which may actually be instance arrays)\n"
  ) +
  gsi::method ("parent_cells", &db::Cell::parent_cells,
    "@brief Report the number of parent cells \n"
    "\n"
    "The number of parent cells (cells which reference our cell) is reported."
  ) +
  gsi::iterator ("each_parent_cell", &db::Cell::begin_parent_cells, &db::Cell::end_parent_cells, 
    "@brief Iterate over all parent cells\n"
    "\n"
    "This iterator will iterate over the parent cells, just returning their\n"
    "cell index.\n"
  ) +
  gsi::method ("is_top?", &db::Cell::is_top,
    "@brief Tell if the cell is a top-level cell\n"
    "\n"
    "A cell is a top-level cell if there are no parent instantiations.\n"
  ) +
  gsi::method ("is_leaf?", &db::Cell::is_leaf,
    "@brief Tell if the cell is a leaf cell\n"
    "\n"
    "A cell is a leaf cell if there are no child instantiations.\n"
  ) +
  gsi::iterator_ext ("each_shape", &begin_shapes, 
    "@brief Iterate all shapes of a given layer\n"
    "@args layer_index, flags\n"
    "\n"
    "@param flags An \"or\"-ed combination of the s_.. constants of the \\Shapes class\n"
    "@param layer_index The layer on which to run the query\n" 
    "\n"
    "This iterator is equivalent to 'shapes(layer).each'."
  ) +
  //  Hint: don't use db::Shapes::begin_touching. It does not update the box trees automatically
  gsi::iterator_ext ("each_touching_shape", &begin_touching_shapes, 
    "@brief Iterate all shapes of a given layer that touch the given box\n"
    "@args layer_index, box, flags\n"
    "\n"
    "@param flags An \"or\"-ed combination of the s_.. constants of the \\Shapes class\n"
    "@param box The box by which to query the shapes\n"
    "@param layer_index The layer on which to run the query\n" 
  ) +
  //  Hint: don't use db::Shapes::begin_overlapping. It does not update the box trees automatically
  gsi::iterator_ext ("each_overlapping_shape", &begin_overlapping_shapes, 
    "@brief Iterate all shapes of a given layer that overlap the given box\n"
    "@args layer_index, box, flags\n"
    "\n"
    "@param flags An \"or\"-ed combination of the s_.. constants of the \\Shapes class\n"
    "@param box The box by which to query the shapes\n"
    "@param layer_index The layer on which to run the query\n" 
  ) +
  gsi::method ("hierarchy_levels", &db::Cell::hierarchy_levels,
    "@brief Return the number of hierarchy levels below (expensive)\n"
  ),
  "@brief The cell object\n"
  "\n"
  "A cell object consists of a set of shape containers (called layers),\n"
  "a set of child cell instances and auxiliary informations such as\n"
  "the parent instance list.\n"
  "A cell is identified through an index given to the cell upon instantiation.\n"
  "Cell instances refer to single instances or array instances. Both are encapsulated in the\n"
  "same object, the \\CellInstArray object. In the simple case, this object refers to a single instance.\n"
  "In the general case, this object may refer to a regular array of cell instances as well.\n"
);

// ---------------------------------------------------------------
//  db::instance<db::CellGraph> binding (to "Instance")

static bool is_regular_array_i (const db::Instance *inst) 
{
  db::CellInstArray::point_type a, b;
  unsigned long na = 0, nb = 0;
  return inst->is_regular_array (a, b, na, nb);
}

static db::CellInstArray::point_type array_a_i (const db::Instance *inst) 
{
  db::CellInstArray::point_type a, b;
  unsigned long na = 0, nb = 0;
  inst->is_regular_array (a, b, na, nb);
  return a;
}

static db::CellInstArray::point_type array_b_i (const db::Instance *inst) 
{
  db::CellInstArray::point_type a, b;
  unsigned long na = 0, nb = 0;
  inst->is_regular_array (a, b, na, nb);
  return b;
}

static unsigned long array_na_i (const db::Instance *inst) 
{
  db::CellInstArray::point_type a, b;
  unsigned long na = 0, nb = 0;
  inst->is_regular_array (a, b, na, nb);
  return na;
}

static unsigned long array_nb_i (const db::Instance *inst) 
{
  db::CellInstArray::point_type a, b;
  unsigned long na = 0, nb = 0;
  inst->is_regular_array (a, b, na, nb);
  return nb;
}

template <>
Class<db::Instance> Class<db::Instance>::decl ("Instance", 
  method ("prop_id", &db::Instance::prop_id,
    "@brief Get the properties Id associated with the instance\n"
  ) +
  gsi::method ("has_prop_id?", &db::Instance::has_prop_id,
    "@brief Check, if the instance is associated with a properties Id\n"
  ) +
  method ("parent_cell_index", &db::parent_inst_rep<db::CellGraph>::parent_cell_index,
    "@brief Retrieve the reference to the parent cell \n"
  ) +
  method ("cell_index", &db::Instance::cell_index,
    "@brief Get the index of the cell this instance refers to\n"
  ) +
  gsi::method_ext ("is_regular_array?", &is_regular_array_i,
    "@brief Test, if this instance is a regular array\n"
  ) +
  gsi::method_ext ("a", &array_a_i,
    "@brief Return the displacement vector for the 'a' axis\n"
  ) +
  gsi::method_ext ("b", &array_b_i,
    "@brief Return the displacement vector for the 'b' axis\n"
  ) +
  gsi::method_ext ("na", &array_na_i,
    "@brief Return the number of instances in the 'a' axis\n"
  ) +
  gsi::method_ext ("nb", &array_nb_i,
    "@brief Return the number of instances in the 'b' axis\n"
  ) +
  gsi::method ("cplx_trans", (db::CellInstArray::complex_trans_type (db::Instance::*) () const) &db::Instance::complex_trans,
    "@brief Get the complex transformation of the instance or the first instance in the array\n"
    "This method is always valid compared to \\trans, since simple transformations can be expressed as complex transformations as well."
  ) +
  gsi::method ("trans", &db::Instance::front,
    "@brief Get the transformation of the first instance in the array\n"
    "The transformation returned is only valid if the array does not represent a complex transformation array"
  ) +
  gsi::method ("size", &db::Instance::size,
    "@brief The number of single instances in the instance array\n"
    "If the instance represents a single instance, the count is 1. Otherwise it is na*nb."
  ) +
  gsi::method ("is_complex?", &db::Instance::is_complex,
    "@brief Test, if the array is a complex array\n"
    "\n"
    "Returns true if the array represents complex instances (that is, with magnification and \n"
    "arbitrary rotation angles).\n"
  ) +
  gsi::method ("cell_inst", &db::Instance::cell_inst,
    "@brief Get the basic CellInstArray object associated with this instance reference."
  ) +
  gsi::method ("<", &db::Instance::operator<,
    "@brief Provide an order criterion for two Instance objects\n"
    "@args b\n"
    "Warning: this operator is just provided to establish any order, not a particular one."
  ) +
  gsi::method ("!=", &db::Instance::operator!=,
    "@brief Inequality of two Instance objects\n"
    "@args b\n"
    "Warning: this operator returns true if both objects refer to the same instance, not just identical ones."
  ) +
  gsi::method ("==", &db::Instance::operator==,
    "@brief Equality of two Instance objects\n"
    "@args b\n"
    "See the hint on the < operator."
  ),
  "@brief An instance proxy\n"
  "\n"
  "An instance proxy is basically a pointer to an instance of different kinds, \n"
  "similar to \\Shape, the shape proxy. \\Instance objects can be duplicated without\n"
  "creating copies of the instances itself: the copy will still point to the same instance\n"
  "than the original.\n"
);

// ---------------------------------------------------------------
//  db::parent_inst_rep<db::CellGraph> binding (to "ParentInstArray")

db::Instance child_inst (const db::parent_inst_rep<db::CellGraph> *pinst)
{
  return *pinst->child_inst ();
}

template <>
Class< db::parent_inst_rep<db::CellGraph> > Class< db::parent_inst_rep<db::CellGraph> >::decl ("ParentInstArray", 
  method ("parent_cell_index", &db::parent_inst_rep<db::CellGraph>::parent_cell_index,
    "@brief Retrieve the reference to the parent cell \n"
  ) +
  method_ext ("child_inst", &child_inst,
    "@brief Retrieve the child instance associated with this parent instance\n"
    "\n"
    "Starting with version 0.15, this method returns an \\Instance object rather than a \\CellInstArray reference."
  ) +
  method ("inst", &db::parent_inst_rep<db::CellGraph>::inst,
    "@brief Compute the inverse instance by which the parent is seen from the child\n"
  ),
  "@brief A parent instance\n"
  "\n"
  "A parent instance is basically an inverse instance: instead of pointing\n"
  "to the child cell, it is pointing to the parent cell and the transformation\n"
  "is representing the shift of the parent cell relative to the child cell.\n"
  "For memory performance, a parent instance is not stored as a instance but\n"
  "rather as a reference to a child instance and a reference to the cell which\n"
  "is the parent.\n"
  "The parent instance itself is computed on the fly. It is representative for\n"
  "a set of instances belonging to the same cell index. The special parent instance\n"
  "iterator takes care of producing the right sequence.\n"
);

// ---------------------------------------------------------------
//  db::CellInstArray binding

static db::CellInstArray 
new_v ()
{
  return db::CellInstArray ();
}

static db::CellInstArray 
new_cell_inst (db::Cell::cell_index_type ci, const db::CellInstArray::trans_type &t)
{
  return db::CellInstArray (db::CellInst (ci), t);
}

static db::CellInstArray 
new_cell_inst_cplx (db::Cell::cell_index_type ci, const db::CellInstArray::complex_trans_type &t)
{
  return db::CellInstArray (db::CellInst (ci), t);
}

static db::CellInstArray 
new_cell_inst_array (db::Cell::cell_index_type ci, const db::CellInstArray::trans_type &t, 
                     const db::CellInstArray::point_type &a, const db::CellInstArray::point_type &b, unsigned int na, unsigned int nb)
{
  return db::CellInstArray (db::CellInst (ci), t, a, b, na, nb);
}

static db::CellInstArray 
new_cell_inst_array_cplx (db::Cell::cell_index_type ci, const db::CellInstArray::complex_trans_type &t,
                          const db::CellInstArray::point_type &a, const db::CellInstArray::point_type &b, unsigned int na, unsigned int nb)
{
  return db::CellInstArray (db::CellInst (ci), t, a, b, na, nb);
}

static db::CellInstArray::box_type bbox (const db::CellInstArray *a, const db::Layout &layout) 
{
  db::box_convert <db::CellInst> bc (layout);
  return a->bbox (bc);
}

static db::CellInstArray::box_type bbox_per_layer (const db::CellInstArray *a, const db::Layout &layout, unsigned int layer_index) 
{
  db::box_convert <db::CellInst> bc (layout, layer_index);
  return a->bbox (bc);
}

static db::CellInst::cell_index_type cell_index (const db::CellInstArray *a) 
{
  return a->object ().cell_index ();
}

static bool is_regular_array (const db::CellInstArray *arr) 
{
  db::CellInstArray::point_type a, b;
  unsigned long na = 0, nb = 0;
  return arr->is_regular_array (a, b, na, nb);
}

static db::CellInstArray::point_type array_a (const db::CellInstArray *arr) 
{
  db::CellInstArray::point_type a, b;
  unsigned long na = 0, nb = 0;
  arr->is_regular_array (a, b, na, nb);
  return a;
}

static db::CellInstArray::point_type array_b (const db::CellInstArray *arr) 
{
  db::CellInstArray::point_type a, b;
  unsigned long na = 0, nb = 0;
  arr->is_regular_array (a, b, na, nb);
  return b;
}

static unsigned long array_na (const db::CellInstArray *arr) 
{
  db::CellInstArray::point_type a, b;
  unsigned long na = 0, nb = 0;
  arr->is_regular_array (a, b, na, nb);
  return na;
}

static unsigned long array_nb (const db::CellInstArray *arr) 
{
  db::CellInstArray::point_type a, b;
  unsigned long na = 0, nb = 0;
  arr->is_regular_array (a, b, na, nb);
  return nb;
}

template <>
Class<db::CellInstArray> Class<db::CellInstArray>::decl ("CellInstArray", 
  gsi::method ("new", &new_v, 
    "@brief Default constructor"
  ) +
  gsi::method ("new|#new_inst", &new_cell_inst, 
    "@brief Create a single cell instance\n"
    "@args cell_index, trans\n"
    "@param cell_index The cell to instantiate\n"
    "@param trans The transformation by which to instantiate the cell\n"
  ) +
  gsi::method ("new|#new_inst_cplx", &new_cell_inst_cplx, 
    "@brief Create a single cell instance with a complex transformation\n"
    "@args cell_index, trans\n"
    "@param cell_index The cell to instantiate\n"
    "@param trans The complex transformation by which to instantiate the cell\n"
  ) +
  gsi::method ("new|#new_inst_array", &new_cell_inst_array, 
    "@brief Create a single cell instance\n"
    "@args cell_index, trans, a, b, na, nb\n"
    "@param cell_index The cell to instantiate\n"
    "@param trans The transformation by which to instantiate the cell\n"
    "@param a The displacement vector of the array in the 'a' axis\n"
    "@param b The displacement vector of the array in the 'b' axis\n"
    "@param na The number of placements in the 'a' axis\n"
    "@param nb The number of placements in the 'b' axis\n"
  ) +
  gsi::method ("new|#new_inst_array_cplx", &new_cell_inst_array_cplx, 
    "@brief Create a single cell instance with a complex transformation\n"
    "@args cell_index, trans\n"
    "@param cell_index The cell to instantiate\n"
    "@param trans The complex transformation by which to instantiate the cell\n"
    "@param a The displacement vector of the array in the 'a' axis\n"
    "@param b The displacement vector of the array in the 'b' axis\n"
    "@param na The number of placements in the 'a' axis\n"
    "@param nb The number of placements in the 'b' axis\n"
  ) +
  gsi::method_ext ("bbox", &bbox,
    "@brief The bounding box of the array\n"
    "@args layout\n"
    "The bounding box incorporates all instances that the array represents. It needs the layout object to access the "
    "actual cell from the cell index."
  ) +
  gsi::method_ext ("bbox_per_layer", &bbox_per_layer,
    "@brief The bounding box of the array with respect to one layer\n"
    "@args layout, layer_index\n"
    "The bounding box incorporates all instances that the array represents. It needs the layout object to access the "
    "actual cell from the cell index."
  ) +
  gsi::method ("size", &db::CellInstArray::size,
    "@brief The number of single instances in the array\n"
    "If the instance represents a single instance, the count is 1. Otherwise it is na*nb."
  ) +
  gsi::method_ext ("cell_index", &cell_index,
    "@brief Get the cell index of the cell instantiated \n"
  ) +
  gsi::method ("cplx_trans", (db::CellInstArray::complex_trans_type (db::CellInstArray::*) () const) &db::CellInstArray::complex_trans,
    "@brief Get the complex transformation of the first instance in the array\n"
    "This method is always valid compared to \\trans, since simple transformations can be expressed as complex transformations as well."
  ) +
  gsi::method ("trans", &db::CellInstArray::front,
    "@brief Get the transformation of the first instance in the array\n"
    "The transformation returned is only valid if the array does not represent a complex transformation array"
  ) +
  gsi::method ("invert", &db::CellInstArray::invert,
    "@brief Invert an array reference\n"
    "\n"
    "The inverted array reference describes in which transformations the parent cell is\n"
    "seen from the current cell."
  ) +
  gsi::method ("<", &db::CellInstArray::operator<,
    "@brief Less operator\n"
  ) +
  gsi::method ("==", &db::CellInstArray::operator==,
    "@brief Compare operator for equality\n"
  ) +
  gsi::method ("!=", &db::CellInstArray::operator!=,
    "@brief Compare operator for inequality\n"
  ) +
  gsi::method ("is_complex?", &db::CellInstArray::is_complex,
    "@brief Test, if the array is a complex array\n"
    "\n"
    "Returns true if the array represents complex instances (that is, with magnification and \n"
    "arbitrary rotation angles).\n"
  ) +
  gsi::method_ext ("is_regular_array?", &is_regular_array,
    "@brief Test, if this instance is a regular array\n"
  ) +
  gsi::method_ext ("a", &array_a,
    "@brief Return the displacement vector for the 'a' axis\n"
  ) +
  gsi::method_ext ("b", &array_b,
    "@brief Return the displacement vector for the 'b' axis\n"
  ) +
  gsi::method_ext ("na", &array_na,
    "@brief Return the number of instances in the 'a' axis\n"
  ) +
  gsi::method_ext ("nb", &array_nb,
    "@brief Return the number of instances in the 'b' axis\n"
  ),
  "@brief A single or array cell instance\n"
  "This object represents either single or array cell instances. A cell instance array is a "
  "regular array, described by two displacement vectors (a, b) and the instance count along that axes (na, nb). "
  "\n\n"
  "In addition, this object represents either instances with simple transformations or "
  "instances with complex transformations. The latter includes magnified instances and instances "
  "rotated by an arbitrary angle."
);

}


