
/*

  KLayout Layout Viewer
  Copyright (C) 2006-2007 Matthias Koefferlein

  This program is free software; you can redistribute it and/or modify
  it under the terms of the GNU General Public License as published by
  the Free Software Foundation; either version 2 of the License, or
  (at your option) any later version.

  This program is distributed in the hope that it will be useful,
  but WITHOUT ANY WARRANTY; without even the implied warranty of
  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
  GNU General Public License for more details.

  You should have received a copy of the GNU General Public License
  along with this program; if not, write to the Free Software
  Foundation, Inc., 51 Franklin St, Fifth Floor, Boston, MA  02110-1301  USA

*/


#include "gsi.h"
#include "dbPoint.h"
#include "dbEdge.h"

namespace gsi
{

// ---------------------------------------------------------------
//  edge binding

template <class C>
struct edge_defs
{
  typedef typename C::coord_type coord_type;
  typedef typename C::box_type box_type;
  typedef typename C::point_type point_type;
  typedef typename C::distance_type distance_type;
  typedef typename C::area_type area_type;
  typedef db::simple_trans<coord_type> simple_trans_type;
  typedef db::complex_trans<coord_type, double> complex_trans_type;

  static C new_v ()
  {
    return C ();
  }

  static C new_pp (const point_type &p1, const point_type &p2)
  {
    return C (p1, p2);
  }

  static C new_xyxy (coord_type x1, coord_type y1, coord_type x2, coord_type y2)
  {
    return C (point_type (x1, y1), point_type (x2, y2));
  }

  static box_type bbox (const C *e)
  {
    return box_type (e->p1 (), e->p2 ());
  }

  static gsi::Methods methods ()
  {
    return
    method ("new", &new_v, 
      "@brief Default constructor: creates a degenerated edge 0,0 to 0,0"
    ) +
    method ("new|#new_xyxy", &new_xyxy, 
      "@brief Constructor with two coordinates given as single values\n"
      "\n"
      "@args x1, y1, x2, y2\n"
      "\n"
      "Two points are given to create a new edge."
    ) +
    method ("new|#new_pp", &new_pp, 
      "@brief Constructor with two points\n"
      "\n"
      "@args p1, p2\n"
      "\n"
      "Two points are given to create a new edge."
    ) +
    method ("<", &C::operator<,
      "@brief Less operator\n"
      "@args e\n"
      "@param e The object to compare against\n"
      "@return True, if the edge is 'less' as the other edge with respect to first and second point"
    ) +
    method ("==", &C::operator==,
      "@brief Equality test\n"
      "@args e\n"
      "@param e The object to compare against"
    ) +
    method ("!=", &C::operator!=,
      "@brief Inequality test\n"
      "@args e\n"
      "@param e The object to compare against"
    ) +
    method ("moved", &C::moved,
      "@brief Returns the moved edge\n"
      "@args p\n"
      "\n"
      "Moves the edge by the given offset and returns the \n"
      "moved edge. The edge is not modified.\n"
      "\n"
      "@param p The distance to move the edge.\n"
      "\n"
      "@return The moved edge.\n"
    ) +
    method ("enlarged", &C::enlarged,
      "@brief Returns the enlarged edge\n"
      "@args p\n"
      "\n"
      "Enlarges the edge by the given offset and returns the \n"
      "moved edge. The edge is not modified. Enlargement means\n"
      "that the first point is shifted by -p, the second by p.\n"
      "\n"
      "@param p The distance to move the edge.\n"
      "\n"
      "@return The moved edge.\n"
    ) +
    method ("transformed", &C::template transformed<simple_trans_type>,
      "@brief Transform the edge.\n"
      "@args t\n"
      "\n"
      "Transforms the edge with the given transformation.\n"
      "Does not modify the edge but returns the transformed edge.\n"
      "\n"
      "@param t The transformation to apply.\n"
      "\n"
      "@return The transformed edge.\n"
    ) +
    method ("transformed_cplx", &C::template transformed<complex_trans_type>,
      "@brief Transform the edge.\n"
      "@args t\n"
      "\n"
      "Transforms the edge with the given complex transformation.\n"
      "Does not modify the edge but returns the transformed edge.\n"
      "\n"
      "@param t The transformation to apply.\n"
      "\n"
      "@return The transformed edge.\n"
    ) +
    method ("move", &C::move,
      "@brief Moves the edge.\n"
      "@args p\n"
      "\n"
      "Moves the edge by the given offset and returns the \n"
      "moved edge. The edge is overwritten.\n"
      "\n"
      "@param p The distance to move the edge.\n"
      "\n"
      "@return The moved edge.\n"
    ) +
    method ("enlarge", &C::enlarge,
      "@brief Enlarges the edge.\n"
      "@args p\n"
      "\n"
      "Enlarges the edge by the given distance and returns the \n"
      "enlarged edge. The edge is overwritten.\n"
      "\n"
      "@param p The distance to move the edge points.\n"
      "\n"
      "@return The enlarged edge.\n"
    ) +
    method ("p1", &C::p1,
      "@brief The first point.\n"
    ) +
    method ("p2", &C::p2,
      "@brief The second point.\n"
    ) +
    method ("dx", &C::dx,
      "@brief The horizontal extend of the edge.\n"
    ) +
    method ("dy", &C::dy,
      "@brief The vertical extend of the edge.\n"
    ) +
    method ("x1", &C::x1,
      "@brief Shortcut for p1.x\n"
    ) +
    method ("y1", &C::y1,
      "@brief Shortcut for p1.y\n"
    ) +
    method ("x2", &C::x2,
      "@brief Shortcut for p2.x\n"
    ) +
    method ("y2", &C::y2,
      "@brief Shortcut for p2.y\n"
    ) +
    method ("dx_abs", &C::dx_abs,
      "@brief The absolute value of the horizontal extend of the edge.\n"
    ) +
    method ("dy_abs", &C::dy_abs,
      "@brief The vertical extend of the edge.\n"
    ) +
    method_ext ("bbox", &bbox,
      "@brief Return the bounding box of the edge.\n"
    ) +
    method ("is_degenerate?", &C::is_degenerate,
      "@brief Test for degenerated edge\n"
      "\n"
      "An edge is degenerate, if both end and start point are identical."
    ) +
    method ("length", &C::length,
      "@brief The length of the edge\n"
    ) +
    method ("sq_length", &C::sq_length,
      "@brief The square of the length of the edge\n"
    ) +
    method ("ortho_length", &C::ortho_length,
      "@brief The orthogonal length of the edge (\"manhattan-length\")\n"
      "\n"
      "@return The orthogonal length (abs(dx)+abs(dy))\n"
    ) +
    method ("to_s", (std::string (C::*) () const) &C::to_string,
      "@brief Convert to a string\n"
    ) +
    method ("is_parallel?", &C::parallel,
      "@brief Test for being parallel\n"
      "@args e\n"
      "\n"
      "@param e The edge to test against\n"
      "\n"
      "@return True if both edges are parallel\n"
    ) +
    method ("contains?", &C::contains,
      "@brief Test whether a point is on an edge.\n"
      "@args p\n"
      "\n"
      "A point is on a edge if it is on (or at least closer \n"
      "than a grid point to) the edge.\n"
      "\n"
      "@param p The point to test with the edge.\n"
      "\n"
      "@return True if the point is on the edge.\n"
    ) +
    method ("contains_excl?", &C::contains_excl,
      "@brief Test whether a point is on an edge excluding the endpoints.\n"
      "@args p\n"
      "\n"
      "A point is on a edge if it is on (or at least closer \n"
      "than a grid point to) the edge.\n"
      "\n"
      "@param p The point to test with the edge.\n"
      "\n"
      "@return True if the point is on the edge but not equal p1 or p2.\n"
    ) +
    method ("coincident?", &C::coincident,
      "@brief Coincidence check.\n"
      "@args e\n"
      "\n"
      "Checks whether a edge is coincident with another edge. \n"
      "Coincidence is defined by being parallel and that \n"
      "at least one point of one edge is on the other edge.\n"
      "\n"
      "@param e the edge to test with\n"
      "\n"
      "@return True if the edges are coincident.\n"
    ) +
    method ("intersect?", &C::intersect,
      "@brief Intersection test. \n"
      "@args e\n"
      "\n"
      "Returns true if the edges intersect. \n"
      "If the edges coincide, they also intersect.\n"
      "For degenerated edges, the intersection is mapped to\n"
      "point containment tests.\n"
      "\n"
      "@param e The edge to test.\n"
    ) +
    method ("distance", &C::distance,
      "@brief Distance between the edge and a point.\n"
      "@args p\n"
      "\n"
      "Returns the distance between the edge and the point. The \n"
      "distance is signed which is negative if the point is to the\n"
      "\"left\" of the edge and positive if the point is to the \"right\".\n"
      "The distance is measured by projecting the point onto the\n"
      "line through the edge. If the edge is degenerated, the distance\n"
      "is not defined.\n"
      "\n"
      "@param p The point to test.\n"
      "\n"
      "@return The distance\n"
    ) +
    method ("side_of", &C::side_of,
      "@brief Side of the pointdistanc\n"
      "@args p\n"
      "\n"
      "Returns -1 if the point is \"left\" of the edge, 0 if on\n"
      "and 1 if the point is \"right\" of the edge.\n"
      "\n"
      "@param p The point to test.\n"
      "\n"
      "@return The side value\n"
    ) +
    method ("distance_abs", &C::distance_abs,
      "@brief Absolute distance between the edge and a point.\n"
      "@args p\n"
      "\n"
      "Returns the distance between the edge and the point. \n"
      "\n"
      "@param p The point to test.\n"
      "\n"
      "@return The distance\n"
    ) +
    method ("swap_points", &C::swap_points,
      "@brief Swap the points of the edge\n"
    ) +
    method ("crossed_by?", &C::crossed_by,
      "@brief Check, if an edge is cut by a line (given by an edge)\n"
      "@args e\n"
      "\n"
      "This method returns true if p1 is in one semispace \n"
      "while p2 is in the other or one of them is on the line\n"
      "through the edge \"e\"\n"
    );
  }
};

static db::Edge edge_from_dedge (const db::DEdge &b)
{
  return db::Edge::from_double (b);
}

template <>
Class<db::Edge> Class<db::Edge>::decl ("Edge", 
  method ("from_dedge", &edge_from_dedge, 
    "@brief Construct an integer coordinate edge from a floating-point coordinate edge\n"
    "\n"
    "@args double_edge\n"
    "\n"
    "Create a integer coordinate edge from a floating-point coordinate edge"
  ) +
  edge_defs<db::Edge>::methods (),
  "@brief An edge class\n"
  "\n"
  "An edge is a connection between points, usually participating in a larger context "
  "such as a polygon. An edge has a defined direction (from p1 to p2). "
);

static db::DEdge dedge_from_iedge (const db::Edge &b)
{
  return db::DEdge (b);
}

template <>
Class<db::DEdge> Class<db::DEdge>::decl ("DEdge", 
  method ("from_iedge", &dedge_from_iedge, 
    "@brief Construct a floating-point coordinate edge from an integer coordinate edge\n"
    "\n"
    "@args int_edge\n"
    "\n"
    "Create a floating-point coordinate edge from an integer edge"
  ) +
  edge_defs<db::DEdge>::methods (),
  "@brief An edge class\n"
  "\n"
  "An edge is a connection between points, usually participating in a larger context "
  "such as a polygon. An edge has a defined direction (from p1 to p2). "
);

}

