
/*

  KLayout Layout Viewer
  Copyright (C) 2006-2007 Matthias Koefferlein

  This program is free software; you can redistribute it and/or modify
  it under the terms of the GNU General Public License as published by
  the Free Software Foundation; either version 2 of the License, or
  (at your option) any later version.

  This program is distributed in the hope that it will be useful,
  but WITHOUT ANY WARRANTY; without even the implied warranty of
  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
  GNU General Public License for more details.

  You should have received a copy of the GNU General Public License
  along with this program; if not, write to the Free Software
  Foundation, Inc., 51 Franklin St, Fifth Floor, Boston, MA  02110-1301  USA

*/


#include "gsi.h"
#include "dbLayout.h"
#include "dbWriter.h"
#include "dbSaveLayoutOptions.h"
#include "tlStream.h"

namespace gsi
{

// ---------------------------------------------------------------
//  db::LayerProperties binding

static const std::string &lp_get_name (const db::LayerProperties *lp)
{
  return lp->name;
}

static void lp_set_name (db::LayerProperties *lp, const std::string &s)
{
  lp->name = s;
}

static int lp_get_layer (const db::LayerProperties *lp)
{
  return lp->layer;
}

static void lp_set_layer (db::LayerProperties *lp, int n)
{
  lp->layer = n;
}

static int lp_get_datatype (const db::LayerProperties *lp)
{
  return lp->datatype;
}

static void lp_set_datatype (db::LayerProperties *lp, int n)
{
  lp->datatype = n;
}

//  since there already exists a "LayerProperties" object, we call this one "LayerInfo"
template <>
Class<db::LayerProperties> Class<db::LayerProperties>::decl ("LayerInfo", 
  gsi::method_ext ("name=", &lp_set_name, 
    "@brief Set the layer name\n"
    "The name is set on OASIS input for example, if the layer has a name.\n"
    "@args name"
  ) +
  gsi::method_ext ("name", &lp_get_name, 
    "@brief Read accessor for the layer name\n"
  ) +
  gsi::method_ext ("layer=", &lp_set_layer, 
    "@brief Set the layer number\n"
    "@args layer"
  ) +
  gsi::method_ext ("layer", &lp_get_layer, 
    "@brief Read accessor for the layer number\n"
  ) +
  gsi::method_ext ("datatype=", &lp_set_datatype, 
    "@brief Set the datatype\n"
    "@args datatype"
  ) +
  gsi::method_ext ("datatype", &lp_get_datatype, 
    "@brief Read accessor for the datatype\n"
  ),
  "@brief A structure encapsulating the layer properties\n"
  "\n"
  "The layer properties describe how a layer is stored in a GDSII or OASIS file for example."
);

// ---------------------------------------------------------------
//  db::Layout binding

static db::Layout::cell_index_type cell_by_name (db::Layout *l, const char *name)
{
  std::pair<bool, db::Layout::cell_index_type> c = l->cell_by_name (name);
  if (! c.first) {
    throw tl::Exception (tl::sprintf (tl::translate ("No such cell: '%s'").c_str (), name));
  }
  return c.second;
}

static db::PropertiesRepository::properties_id_type properties_id (db::Layout *layout, const std::vector<tl::Variant> &properties)
{
  db::PropertiesRepository::properties_set props;

  for (std::vector<tl::Variant>::const_iterator v = properties.begin (); v != properties.end (); ++v) {
    if (! v->is_list () || v->get_list ().size () != 2) {
      throw tl::Exception (tl::translate ("Expected a list of pairs of variants (found at least one that is not a pair)"));
    }
    db::PropertiesRepository::property_names_id_type name_id = layout->properties_repository ().prop_name_id (v->get_list ()[0]);
    props.insert (std::make_pair (name_id, v->get_list () [1]));
  }

  return layout->properties_repository ().properties_id (props);
}

static std::vector<tl::Variant> properties (const db::Layout *layout, db::PropertiesRepository::properties_id_type id)
{
  std::vector<tl::Variant> ret;

  if (layout->properties_repository ().is_valid_properties_id (id)) {

    const db::PropertiesRepository::properties_set &props = layout->properties_repository ().properties (id);
    ret.reserve (props.size ());

    for (db::PropertiesRepository::properties_set::const_iterator p = props.begin (); p != props.end (); ++p) {
      ret.push_back (tl::Variant::empty_list ());
      ret.back ().get_list ().reserve (2);
      ret.back ().get_list ().push_back (layout->properties_repository ().prop_name (p->first));
      ret.back ().get_list ().push_back (p->second);
    }

  }

  return ret;
}

static void 
write_options (const db::Layout *layout, const std::string &filename, bool gzip, const db::SaveLayoutOptions &options)
{
  db::Writer writer (options);
  if (gzip) {
    tl::OutputZLibFile file (filename);
    writer.write (*layout, file);
  } else {
    tl::OutputFile file (filename);
    writer.write (*layout, file);
  }
}

template <>
Class<db::Layout> Class<db::Layout>::decl ("Layout", 
  gsi::method_ext ("properties_id", &properties_id, 
    "@brief Get the properties ID for a given properties set\n"
    "@args properties\n"
    "\n"
    "Before a set of properties can be attached to a shape, it must be converted into an ID that "
    "is unique for that set. The properties set must be given as a list of pairs of variants, "
    "each pair describing a name and a value. The name acts as the key for the property and does not need to be a string (it can be an integer or double value as well).\n"
    "The backward conversion can be performed with the 'properties' method.\n"
    "\n"
    "@param properties The array of pairs of variants (both elements can be integer, double or string)\n"
    "@return The unique properties ID for that set"
  ) +
  gsi::method_ext ("properties", &properties, 
    "@brief Get the properties set for a given properties ID\n"
    "@args properties_id\n"
    "\n"
    "Basically performs the backward conversion of the 'properties_id' method. "
    "Given a properties ID, returns the properties set as an array of pairs of "
    "variants. In this array, each key and the value are stored as pairs (arrays with two elements).\n"
    "If the properties ID is not valid, an empty array is returned.\n"
    "\n"
    "@param properties_id The properties ID to get the properties for\n"
    "@return The array of variants (see \\properties_id)\n"
  ) +
  gsi::method ("has_cell?", &db::Layout::has_cell, 
    "@brief Tell, if the cell with a given name exists\n"
    "@args name\n"
    "Returns true, if the layout has a cell with the given name"
  ) +
  gsi::method_ext ("cell_by_name", &cell_by_name, 
    "@brief Get the cell index for a given name\n"
    "@args name\n"
    "Returns the cell index for the cell with the given name. If no cell with this "
    "name exists, an exception is thrown."
  ) +
  gsi::method ("cell_name", &db::Layout::cell_name, 
    "@brief Get the name for a cell with the given index\n"
    "@args index\n"
  ) +
  gsi::method ("add_cell", &db::Layout::add_cell, 
    "@brief Add a cell with the given name\n"
    "@args name\n"
    "@return The index of the newly created cell."
  ) +
  gsi::method ("rename_cell", &db::Layout::rename_cell, 
    "@brief Rename the cell with the given index to the given name\n"
    "@args index, name\n"
  ) +
  gsi::method ("start_changes", &db::Layout::start_changes,
    "@brief Signal the start of an operation bringing the layout into invalid state\n"
    "\n"
    "This method should be called whenever the layout is\n"
    "about to be brought into an invalid state. After calling\n"
    "this method, \\under_construction returns false which \n"
    "tells foreign code (such as \\update which might be called\n"
    "asynchroneously for example because of a repaint event)\n"
    "not to use this layout object.\n\n"
    "This state is cancelled by the \\end_changes method.\n"
    "The start_changes method can be called multiple times\n"
    "and must be cancelled the same number of times.\n\n"
    "Using this method is only required currently if a repaint event may happen while\n"
    "the layout object is in an invalid state."
  ) +
  gsi::method ("end_changes", &db::Layout::end_changes,
    "@brief Cancel the \"in changes\" state (see \"start_changes\")\n"
  ) +
  gsi::method ("under_construction", &db::Layout::under_construction,
    "@brief Tell if the layout object is under construction\n"
    "\n"
    "A layout object is either under construction if a transaction\n"
    "is ongoing or the layout is brought into invalid state by\n"
    "\"start_changes\".\n"
  ) +
  gsi::method ("update", (void (db::Layout::*) ()) &db::Layout::update,
    "@brief Update the internals of the layout\n"
    "This method updates the internal state of the layout. Usually this is done automatically\n"
    "This method is provided to ensure this state explicitly."
  ) +
  gsi::method ("dbu=", (void (db::Layout::*) (double)) &db::Layout::dbu,
    "@brief Database unit write accessor\n"
    "@args dbu\n"
  ) +
  gsi::method ("dbu", (double (db::Layout::*) () const) &db::Layout::dbu,
    "@brief Database unit read accessor\n"
  ) +
  gsi::method ("insert_layer", (unsigned int (db::Layout::*) (const db::LayerProperties &)) &db::Layout::insert_layer,
    "@brief Insert a new layer with the given properties\n"
    "@args props\n"
    "@return The index of the newly created layer\n"
  ) +
  gsi::method ("insert_layer_at", (void (db::Layout::*) (unsigned int, const db::LayerProperties &)) &db::Layout::insert_layer,
    "@brief Insert a new layer with the given properties at the given index\n"
    "@args index, props\n"
  ) +
  gsi::method ("insert_special_layer", (unsigned int (db::Layout::*) (const db::LayerProperties &)) &db::Layout::insert_special_layer,
    "@brief Insert a new special layer with the given properties\n"
    "@args props\n"
    "\n"
    "Special layers can be used to represent objects that should not participate in normal viewing or other "
    "related operations. Special layers are not reported as valid layers.\n"
    "\n"
    "@return The index of the newly created layer\n"
  ) +
  gsi::method ("insert_special_layer_at", (void (db::Layout::*) (unsigned int, const db::LayerProperties &)) &db::Layout::insert_special_layer,
    "@brief Insert a new special layer with the given properties at the given index\n"
    "@args index, props\n"
    "\n"
    "See \\insert_special_layer for a description of special layers."
  ) +
  gsi::method ("set_info", &db::Layout::set_properties,
    "@brief Set the info structure for a specified layer\n"
    "@args index, props\n"
  ) +
  gsi::method ("get_info", &db::Layout::get_properties,
    "@brief Get the info structure for a specified layer\n"
    "@args index\n"
  ) +
  gsi::method ("cells", &db::Layout::cells,
    "@brief Return the number of cells\n"
    "\n"
    "@return The number of cells (the maximum cell index)\n"
  ) +
  gsi::method ("cell", (db::Layout::cell_type &(db::Layout::*) (db::Layout::cell_index_type)) &db::Layout::cell,
    "@brief Address a cell by index\n"
    "@args i\n"
    "\n"
    "@param i The cell index\n"
    "@return A reference to the cell\n"
  ) +
  gsi::iterator ("each_cell", (db::Layout::iterator (db::Layout::*) ()) &db::Layout::begin, (db::Layout::iterator (db::Layout::*) ()) &db::Layout::end,
    "@brief Iterate the unsorted cell list\n"
  ) +
  gsi::iterator ("each_cell_bottom_up", (db::Layout::bottom_up_iterator (db::Layout::*) ()) &db::Layout::begin_bottom_up, (db::Layout::bottom_up_iterator (db::Layout::*) ()) &db::Layout::end_bottom_up,
    "@brief Iterate the bottom-up sorted cell list\n"
    "\n"
    "In bottom-up traversal a cell is not delivered before\n"
    "the last child cell of this cell has been delivered.\n"
    "The bottom-up iterator does not deliver cells but cell\n"
    "indices actually.\n"
  ) +
  gsi::iterator ("each_cell_top_down", (db::Layout::top_down_iterator (db::Layout::*) ()) &db::Layout::begin_top_down, (db::Layout::top_down_iterator (db::Layout::*) ()) &db::Layout::end_top_down,
    "@brief Iterate of the top-down sorted cell list\n"
    "\n"
    "The top-down cell list has the property of delivering all\n"
    "cells before they are instantiated. In addition the first\n"
    "cells are all top cells. There is at least one top cell.\n"
    "The top-down iterator does not deliver cells but cell\n"
    "indices actually.\n"
    "@brief begin iterator of the top-down sorted cell list\n"
  ) +
  gsi::iterator ("each_top_cell", (db::Layout::top_down_iterator (db::Layout::*) ()) &db::Layout::begin_top_down, (db::Layout::top_down_iterator (db::Layout::*) ()) &db::Layout::end_top_cells,
    "@brief Iterate the top cells\n"
    "A layout may have an arbitrary number of top cells. The usual case however is that there is one "
    "top cell."
  ) +
  gsi::method ("delete_layer", &db::Layout::delete_layer,
    "@brief Delete a layer\n"
    "@args layer_index\n"
    "\n"
    "This does free the shapes of the cells and remembers the\n"
    "layer's index for recycling.\n"
  ) +
  gsi::method ("layers", &db::Layout::layers,
    "@brief Return the number of layers\n"
    "The number of layers reports the maximum (plus 1) layer index used so far. Not all of the layers with "
    "an index in the range of 0 to layers-1 needs to be a valid layer. These layers can be either valid, "
    "special or unused. Use \\is_valid_layer? and \\is_special_layer? to test for the first two states.\n"
  ) +
  gsi::method ("is_valid_layer?", &db::Layout::is_valid_layer,
    "@brief Tell, if a layer index is a valid index\n"
    "@args layer_index\n"
    "\n"
    "@return true, if this is so\n"
  ) +
  gsi::method ("is_special_layer?", &db::Layout::is_special_layer,
    "@brief Tell, if a layer index is a special layer index\n"
    "@args layer_index\n"
    "\n"
    "@return true, if this is so\n"
  ) +
  gsi::method_ext ("write", &write_options,
    "@brief Write the layout to a stream file\n"
    "@args filename, gzip, options\n"
    "@param filename The file to which to write the layout\n"
    "@param gzip True, if the file should be gzipped\n"
    "@param options The option set to use for writing. See \\SaveLayoutOptions for details\n"
  ),
  "@brief The layout object\n"
  "\n"
  "The layout object basically wraps the cell graphs and\n"
  "adds functionality for managing cell names and layer names.\n"
  "The cell graph is a container for the cells and their hierarchical arrangement.\n"
  "The cell graph is constructed by creating cells and adding child instances to it.\n"
);

db::SaveLayoutOptions new_v ()
{
  return db::SaveLayoutOptions ();
}

template <>
Class<db::SaveLayoutOptions> Class<db::SaveLayoutOptions>::decl ("SaveLayoutOptions", 
  gsi::method ("new", &new_v,
    "@brief Default constructor\n"
    "\n"
    "By default, the scale factor will be 1.0, the database unit is set to \n"
    "\"same as original\" and all layers are selected as well as all cells.\n"
    "The default format is GDS2."
  ) +
  gsi::method ("format=", &db::SaveLayoutOptions::set_format,
    "@brief Select a format\n"
    "@args format\n"
    "The format string can be either \"GDS2\" or \"OASIS\". Other formats may be available if\n"
    "a suitable plugin is installed."
  ) +
  gsi::method ("format", &db::SaveLayoutOptions::format,
    "@brief Get the format name \n"
  ) + 
  gsi::method ("add_layer", &db::SaveLayoutOptions::add_layer,
    "@brief Add a layer to be saved \n"
    "\n"
    "@args layer_index\n"
    "\n"
    "Add the layer with the given index to the layers to be saved.\n"
    "If all layers have been selected previously, this state will \n"
    "be cleared. By default, the layer properties given in the\n"
    "layout to be saved will be used. If certain properties are \n"
    "specified, these will be used.\n"
  ) + 
  gsi::method ("select_all_layers", &db::SaveLayoutOptions::select_all_layers,
    "@brief Select all layers to be saved\n"
    "\n"
    "This is the default.\n"
  ) + 
  gsi::method ("add_cell", &db::SaveLayoutOptions::add_cell,
    "@brief Add a cell to save\n"
    "\n"
    "@args cell_index\n"
    "\n"
    "Give the cell index of the cell. The actual cell is determined when the layout is given. \n"
    "Clears the \"all_cells\" flag.\n"
  ) + 
  gsi::method ("select_all_cells", &db::SaveLayoutOptions::select_all_cells,
    "@brief Select all cells to save\n"
  ) + 
  gsi::method ("dbu=", &db::SaveLayoutOptions::set_dbu,
    "@brief Set the database unit to be saved\n"
    "@args dbu\n"
  ) + 
  gsi::method ("dbu", &db::SaveLayoutOptions::dbu,
    "@brief Get the database unit \n"
  ) + 
  gsi::method ("scale_factor=", &db::SaveLayoutOptions::set_scale_factor,
    "@brief Set the scaling factor for the saving \n"
    "@args scale_factor\n"
  ) + 
  gsi::method ("scale_factor", &db::SaveLayoutOptions::scale_factor,
    "@brief Get the scaling factor \n"
  ),
  "@brief Options for saving layouts\n"
  "This object is used by Layout::write to specify the options. Usually the default constructor\n"
  "provides a suitable object."
);

}

