
/*

  KLayout Layout Viewer
  Copyright (C) 2006-2007 Matthias Koefferlein

  This program is free software; you can redistribute it and/or modify
  it under the terms of the GNU General Public License as published by
  the Free Software Foundation; either version 2 of the License, or
  (at your option) any later version.

  This program is distributed in the hope that it will be useful,
  but WITHOUT ANY WARRANTY; without even the implied warranty of
  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
  GNU General Public License for more details.

  You should have received a copy of the GNU General Public License
  along with this program; if not, write to the Free Software
  Foundation, Inc., 51 Franklin St, Fifth Floor, Boston, MA  02110-1301  USA

*/


#include "gsi.h"
#include "dbPoint.h"
#include "dbBox.h"

namespace gsi
{

// ---------------------------------------------------------------
//  point binding

template <class C>
struct point_defs 
{
  typedef typename C::coord_type coord_type;

  static C new_v ()
  {
    return C ();
  }

  static C new_xy (coord_type x, coord_type y)
  {
    return C (x, y);
  }

  static gsi::Methods methods ()
  {
    return
    method ("new", &new_v,
      "@brief Default constructor: creates a point at 0,0"
    ) +
    method ("new|#new_xy", &new_xy,
      "@brief Constructor for a point from two coordinate values\n"
      "\n"
      "@args x, y\n"
    ) +
    method ("+", (C (C::*) (const C &) const) &C::add,
      "@brief Add one point to another\n"
      "\n"
      "@args p\n"
      "\n"
      "Add point p to self by adding the coordinates"
    ) +
    method ("-", (C (C::*) (const C &) const) &C::subtract,
      "@brief Subtract one point from another\n"
      "\n"
      "@args p\n"
      "\n"
      "Subtract point p from self by subtracting the coordinates"
    ) +
    method ("<", &C::operator<,
      "@brief \"less\" comparison operator\n"
      "\n"
      "@args p\n"
      "\n"
      "This operator is provided to establish a sorting\n"
      "order\n"
    ) +
    method ("==", &C::operator==,
      "@brief Equality test operator\n"
      "\n"
      "@args p\n"
    ) +
    method ("!=", &C::operator!=,
      "@brief Inequality test operator\n"
      "\n"
      "@args p\n"
    ) +
    method ("x", &C::x,
      "@brief Accessor to the x coordinate\n"
    ) +
    method ("y", &C::y,
      "@brief Accessor to the y coordinate\n"
    ) +
    method ("x=", &C::set_x,
      "@brief Write accessor to the x coordinate\n"
    ) +
    method ("y=", &C::set_y,
      "@brief Write accessor to the y coordinate\n"
    ) +
    method ("*", &C::operator*,
      "@brief Scaling by some factor\n"
      "\n"
      "@args f\n"
      "\n"
      "Scaling may involve rounding for integer coordinate points."
    ) +
    method ("distance", (double (C::*) (const C &) const) &C::double_distance,
      "@brief The euclidian distance to another point\n"
      "\n"
      "@args d\n"
      "\n"
      "@param d The other point to compute the distance to.\n"
    ) +
    method ("sq_distance", (double (C::*) (const C &) const) &C::sq_double_distance,
      "@brief The square euclidian distance to another point\n"
      "\n"
      "@args d\n"
      "\n"
      "@param d The other point to compute the distance to.\n"
    ) +
    method ("to_s", (std::string (C::*) () const) &C::to_string,
      "@brief String conversion\n"
    );
  }

};

static db::DPoint dpoint_from_ipoint (const db::Point &p)
{
  return db::DPoint (p);
}

template <>
Class<db::DPoint> Class<db::DPoint>::decl ("DPoint", 
  method ("from_ipoint", &dpoint_from_ipoint,
    "@brief Create a floating-point coordinate point from an integer coordinate point\n"
    "\n"
    "@args p\n"
  ) +
  point_defs<db::DPoint>::methods (),
  "@brief A point class with double (floating-point) coordinates\n"
);

static db::Point point_from_dpoint (const db::DPoint &p)
{
  return db::Point::from_double (p);
}

template <>
Class<db::Point> Class<db::Point>::decl ("Point", 
  method ("from_dpoint", &point_from_dpoint,
    "@brief Create a double point from an integer point\n"
    "\n"
    "@args p\n"
    "\n"
    "Creates an double point from an integer point p" 
  ) +
  point_defs<db::Point>::methods (),
  "@brief An integer point class\n"
);

}

