
/*

  KLayout Layout Viewer
  Copyright (C) 2006-2007 Matthias Koefferlein

  This program is free software; you can redistribute it and/or modify
  it under the terms of the GNU General Public License as published by
  the Free Software Foundation; either version 2 of the License, or
  (at your option) any later version.

  This program is distributed in the hope that it will be useful,
  but WITHOUT ANY WARRANTY; without even the implied warranty of
  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
  GNU General Public License for more details.

  You should have received a copy of the GNU General Public License
  along with this program; if not, write to the Free Software
  Foundation, Inc., 51 Franklin St, Fifth Floor, Boston, MA  02110-1301  USA

*/


#include "gsi.h"
#include "dbPoint.h"
#include "dbPolygon.h"

namespace gsi
{

// ---------------------------------------------------------------
//  simple polygon binding

template <class C>
struct simple_polygon_defs
{
  typedef typename C::coord_type coord_type;
  typedef typename C::box_type box_type;
  typedef typename C::point_type point_type;
  typedef typename C::distance_type distance_type;
  typedef typename C::area_type area_type;
  typedef db::simple_trans<coord_type> simple_trans_type;
  typedef db::complex_trans<coord_type, double> complex_trans_type;

  static void set_points (C *c, const std::vector<point_type> &pts)
  {
    c->assign_hull (pts.begin (), pts.end ());
  }

  static size_t num_points (C *c)
  {
    return c->hull ().size ();
  }

  static C *new_v ()
  {
    return new C;
  }

  static C *new_p (const std::vector<point_type> &pts)
  {
    C *c = new C;
    c->assign_hull (pts.begin (), pts.end ());
    return c;
  }

  static C *new_b (const box_type &box)
  {
    return new C (box);
  }

  static bool inside (const C *poly, point_type pt)
  {
    return db::inside_poly (poly->begin_edge (), pt);
  }

  static C transformed (const C *poly, const simple_trans_type &t)
  {
    return poly->transformed (t);
  }

  static db::simple_polygon<double> transformed_cplx (const C *poly, const complex_trans_type &t)
  {
    return poly->transformed (t);
  }

  static gsi::Methods methods ()
  {
    return
    constructor ("new", &new_v, 
      "@brief Default constructor: creates an empty (invalid) polygon"
    ) +
    constructor ("new|#new_p", &new_p, 
      "@brief Constructor given the points of the simple polygon\n"
      "\n"
      "@args pts\n"
      "\n"
      "@param pts The points forming the simple polygon\n"
    ) +
    constructor ("new|#new_b", &new_b, 
      "@brief Constructor converting a box to a polygon\n"
      "\n"
      "@args box\n"
      "\n"
      "@param box The box to convert to a polygon\n"
    ) +
    method ("==", &C::operator==,
        "@brief Equality test\n"
    ) +
    method ("!=", &C::operator!=,
        "@brief Inequality test\n"
    ) +
    method_ext ("points=", &set_points,
      "@brief Set the points of the simple polygon\n"
      "@args p\n"
      "@param p An array of points to assign to the simple polygon"
    ) +
    method_ext ("points", &num_points,
      "@brief Get the number of points"
    ) +
    iterator ("each_point", &C::begin_hull, &C::end_hull, 
      "@brief Iterate over the points that make up the simple polygon"
    ) +
    iterator ("each_edge", &C::begin_edge, 
      "@brief Iterate over the edges that make up the simple polygon"
    ) +
    method_ext ("inside", &inside,
      "@brief Test, if the given point is inside the polygon\n"
      "@args p\n"
      "If the given point is inside the polygon, true is returned. "
      "This tests works well only if the polygon is not self-overlapping and oriented clockwise. "
    ) +
    method ("move", &C::move,
      "@brief Moves the simple polygon.\n"
      "@args p\n"
      "\n"
      "Moves the simple polygon by the given offset and returns the \n"
      "moved simple polygon. The polygon is overwritten.\n"
      "\n"
      "@param p The distance to move the simple polygon.\n"
      "\n"
      "@return The moved simple polygon.\n"
    ) +
    method ("moved", &C::moved,
      "@brief Returns the moved simple polygon\n"
      "@args p\n"
      "\n"
      "Moves the simple polygon by the given offset and returns the \n"
      "moved simple polygon. The polygon is not modified.\n"
      "\n"
      "@param p The distance to move the simple polygon.\n"
      "\n"
      "@return The moved simple polygon.\n"
    ) +
    method_ext ("transformed", &transformed,
      "@brief Transform the simple polygon.\n"
      "@args t\n"
      "\n"
      "Transforms the simple polygon with the given transformation.\n"
      "Does not modify the simple polygon but returns the transformed polygon.\n"
      "\n"
      "@param t The transformation to apply.\n"
      "\n"
      "@return The transformed simple polygon.\n"
    ) +
    method_ext ("transformed_cplx", &transformed_cplx,
      "@brief Transform the simple polygon.\n"
      "@args t\n"
      "\n"
      "Transforms the simple polygon with the given complex transformation.\n"
      "Does not modify the simple polygon but returns the transformed polygon.\n"
      "\n"
      "@param t The transformation to apply.\n"
      "\n"
      "@return The transformed simple polygon.\n"
    ) +
    method ("to_s", (std::string (C::*) () const) &C::to_string,
      "@brief Convert to a string\n"
    ) +
    method ("area", &C::area,
      "@brief The area of the polygon\n"
      "The area is correct only if the polygon is not self-overlapping and the polygon is oriented clockwise."
    ) +
    method ("bbox", &C::box,
      "@brief Return the bounding box of the simple polygon"
    );
  }
};

db::SimplePolygon sp_from_dsp (const db::DSimplePolygon &d)
{
  return db::SimplePolygon (db::SimplePolygon::from_double (d));
}

db::DSimplePolygon dsp_from_sp (const db::SimplePolygon &d)
{
  return db::DSimplePolygon (d, db::CplxTrans ());
}

template <>
Class<db::SimplePolygon> Class<db::SimplePolygon>::decl ("SimplePolygon", 
  simple_polygon_defs<db::SimplePolygon>::methods () +
  method ("from_dpoly", &sp_from_dsp, 
    "@brief Construct an integer-coordinate polygon from a floating-point coordinate one"
    "\n"
    "@args double_poly\n"
    "\n"
    "This method has been added in version 0.15."
  ),
  "@brief A simple polygon class\n"
  "\n"
  "A simple polygon consists of an outer hull only.\n"
  "The hull contour consists of several points. The point\n"
  "list is normalized such that the leftmost, lowest point is \n"
  "the first one. The orientation is normalized such that\n"
  "the orientation of the hull contour is clockwise.\n"
  "\n"
  "It is in no way checked that the contours are not overlapping\n"
  "This must be ensured by the user of the object\n"
  "when filling the contours.\n"
);

template <>
Class<db::DSimplePolygon> Class<db::DSimplePolygon>::decl ("DSimplePolygon", 
  simple_polygon_defs<db::DSimplePolygon>::methods () +
  method ("from_ipoly", &dsp_from_sp, 
    "@brief Construct a floating-point coordinate polygon from an integer coordinate one"
    "\n"
    "@args int_poly\n"
    "\n"
    "This method has been added in version 0.15."
  ),
  "@brief A simple polygon class\n"
  "\n"
  "A simple polygon consists of an outer hull only.\n"
  "The contour consists of several points. The point\n"
  "list is normalized such that the leftmost, lowest point is \n"
  "the first one. The orientation is normalized such that\n"
  "the orientation of the hull contour is clockwise.\n"
  "\n"
  "It is in no way checked that the contours are not over-\n"
  "lapping. This must be ensured by the user of the object\n"
  "when filling the contours.\n"
);

// ---------------------------------------------------------------
//  polygon binding

template <class C>
struct polygon_defs
{
  typedef typename C::coord_type coord_type;
  typedef typename C::box_type box_type;
  typedef typename C::point_type point_type;
  typedef typename C::distance_type distance_type;
  typedef typename C::area_type area_type;
  typedef db::simple_trans<coord_type> simple_trans_type;
  typedef db::complex_trans<coord_type, double> complex_trans_type;

  static void set_hull (C *c, const std::vector<point_type> &pts)
  {
    return c->assign_hull (pts.begin (), pts.end ());
  }

  static size_t num_points_hull (C *c)
  {
    return c->hull ().size ();
  }

  static size_t num_points_hole (C *c, unsigned int n)
  {
    return c->contour (n + 1).size ();
  }

  static void insert_hole (C *c, const std::vector<point_type> &pts)
  {
    return c->insert_hole (pts.begin (), pts.end ());
  }

  static C *new_v ()
  {
    return new C;
  }

  static C *new_p (const std::vector<point_type> &pts)
  {
    C *c = new C;
    c->assign_hull (pts.begin (), pts.end ());
    return c;
  }

  static C *new_b (const box_type &box)
  {
    return new C (box);
  }

  static bool inside (const C *poly, point_type pt)
  {
    return db::inside_poly (poly->begin_edge (), pt);
  }

  static C transformed (const C *poly, const simple_trans_type &t)
  {
    return poly->transformed (t);
  }

  static db::polygon<double> transformed_cplx (const C *poly, const complex_trans_type &t)
  {
    return poly->transformed (t);
  }

  static gsi::Methods methods ()
  {
    return
    constructor ("new", &new_v, 
      "@brief Default constructor: creates an empty (invalid) polygon"
    ) +
    constructor ("new|#new_p", &new_p, 
      "@brief Constructor given the points of the polygon hull\n"
      "\n"
      "@args pts\n"
      "\n"
      "@param pts The points forming the polygon hull\n"
    ) +
    constructor ("new|#new_b", &new_b, 
      "@brief Constructor converting a box to a polygon\n"
      "\n"
      "@args box\n"
      "\n"
      "@param box The box to convert to a polygon\n"
    ) +
    method ("<", &C::operator<,
      "@brief Less operator\n"
      "@args p\n"
      "@param p The object to compare against\n"
      "This operator is provided to establish some, not necessarily a certain sorting order"
    ) +
    method ("==", &C::operator==,
      "@brief Equality test\n"
      "@args p\n"
      "@param p The object to compare against"
    ) +
    method ("!=", &C::operator!=,
      "@brief Inequality test\n"
      "@args p\n"
      "@param p The object to compare against"
    ) +
    method_ext ("hull=", &set_hull,
      "@brief Set the points of the hull of polygon\n"
      "@args p\n"
      "@param p An array of points to assign to the polygon's hull"
    ) +
    method_ext ("points_hull", &num_points_hull,
      "@brief Get the number of points of the hull"
    ) +
    method_ext ("points_hole", &num_points_hole,
      "@brief Get the number of points of the given hole\n"
      "@args n\n"
      "The argument gives the index of the hole of which the number of points "
      "are requested. The index must be less than the number of holes (see \\holes). "
    ) +
    method_ext ("insert_hole", &insert_hole,
      "@brief Insert a hole with the given points\n"
      "@args p\n"
      "@param p An array of points to insert as a new hole"
    ) +
    iterator ("each_point_hull", &C::begin_hull, &C::end_hull, 
      "@brief Iterate over the points that make up the hull"
    ) +
    iterator ("each_point_hole", &C::begin_hole, &C::end_hole, 
      "@brief Iterate over the points that make up the nth hole\n"
      "@args n\n"
      "The hole number must be less than the number of holes (see \\holes)"
    ) +
    method ("holes", &C::holes,
      "@brief Get the number of holes"
    ) +
    iterator ("each_edge", &C::begin_edge, 
      "@brief Iterate over the edges that make up the polygon"
    ) +
    method_ext ("inside", &inside,
      "@brief Test, if the given point is inside the polygon\n"
      "@args p\n"
      "If the given point is inside the polygon, true is returned. "
      "This tests works well only if the polygon is not self-overlapping and oriented clockwise. "
    ) +
    method ("move", &C::move,
      "@brief Moves the polygon.\n"
      "@args p\n"
      "\n"
      "Moves the polygon by the given offset and returns the \n"
      "moved polygon. The polygon is overwritten.\n"
      "\n"
      "@param p The distance to move the polygon.\n"
      "\n"
      "@return The moved polygon.\n"
    ) +
    method ("moved", &C::moved,
      "@brief Returns the moved polygon\n"
      "@args p\n"
      "\n"
      "Moves the polygon by the given offset and returns the \n"
      "moved polygon. The polygon is not modified.\n"
      "\n"
      "@param p The distance to move the polygon.\n"
      "\n"
      "@return The moved polygon.\n"
    ) +
    method_ext ("transformed", &transformed,
      "@brief Transform the polygon\n"
      "@args t\n"
      "\n"
      "Transforms the polygon with the given transformation.\n"
      "Does not modify the polygon but returns the transformed polygon.\n"
      "\n"
      "@param t The transformation to apply.\n"
      "\n"
      "@return The transformed polygon.\n"
    ) +
    method_ext ("transformed_cplx", &transformed_cplx,
      "@brief Transform the polygon with a complex transformation\n"
      "@args t\n"
      "\n"
      "Transforms the polygon with the given complex transformation.\n"
      "Does not modify the polygon but returns the transformed polygon.\n"
      "\n"
      "@param t The transformation to apply.\n"
      "\n"
      "@return The transformed polygon.\n"
    ) +
    method ("to_s", (std::string (C::*) () const) &C::to_string,
      "@brief Convert to a string\n"
    ) +
    method ("area", &C::area,
      "@brief The area of the polygon\n"
      "The area is correct only if the polygon is not self-overlapping and the polygon is oriented clockwise."
    ) +
    method ("bbox", &C::box,
      "@brief Return the bounding box of the polygon"
    );
  }
};

db::Polygon p_from_dp (const db::DPolygon &d)
{
  return db::Polygon (db::Polygon::from_double (d));
}

db::DPolygon dp_from_p (const db::Polygon &d)
{
  return db::DPolygon (d, db::cast_op<db::DPoint, db::Point> ());
}

template <>
Class<db::Polygon> Class<db::Polygon>::decl ("Polygon", 
  polygon_defs<db::Polygon>::methods () +
  method ("from_dpoly", &p_from_dp, 
    "@brief Construct an integer-coordinate polygon from a floating-point coordinate one"
    "\n"
    "@args double_poly\n"
    "\n"
    "This method has been added in version 0.15."
  ),
  "@brief A polygon class\n"
  "\n"
  "A polygon consists of an outer hull and zero to many\n"
  "holes. Each contour consists of several points. The point\n"
  "list is normalized such that the leftmost, lowest point is \n"
  "the first one. The orientation is normalized such that\n"
  "the orientation of the hull contour is clockwise, while\n"
  "the orientation of the holes is counterclockwise.\n"
  "\n"
  "It is in no way checked that the contours are not over-\n"
  "lapping. This must be ensured by the user of the object\n"
  "when filling the contours.\n"
);

template <>
Class<db::DPolygon> Class<db::DPolygon>::decl ("DPolygon", 
  polygon_defs<db::DPolygon>::methods () +
  method ("from_ipoly", &dp_from_p, 
    "@brief Construct a floating-point coordinate polygon from an integer coordinate one"
    "\n"
    "@args int_poly\n"
    "\n"
    "This method has been added in version 0.15."
  ),
  "@brief A polygon class\n"
  "\n"
  "A polygon consists of an outer hull and zero to many\n"
  "holes. Each contour consists of several points. The point\n"
  "list is normalized such that the leftmost, lowest point is \n"
  "the first one. The orientation is normalized such that\n"
  "the orientation of the hull contour is clockwise, while\n"
  "the orientation of the holes is counterclockwise.\n"
  "\n"
  "It is in no way checked that the contours are not over-\n"
  "lapping. This must be ensured by the user of the object\n"
  "when filling the contours.\n"
);

}

