
/*

  KLayout Layout Viewer
  Copyright (C) 2006-2007 Matthias Koefferlein

  This program is free software; you can redistribute it and/or modify
  it under the terms of the GNU General Public License as published by
  the Free Software Foundation; either version 2 of the License, or
  (at your option) any later version.

  This program is distributed in the hope that it will be useful,
  but WITHOUT ANY WARRANTY; without even the implied warranty of
  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
  GNU General Public License for more details.

  You should have received a copy of the GNU General Public License
  along with this program; if not, write to the Free Software
  Foundation, Inc., 51 Franklin St, Fifth Floor, Boston, MA  02110-1301  USA

*/


#include "gsi.h"
#include "dbPoint.h"
#include "dbBox.h"

namespace gsi
{

// ---------------------------------------------------------------
//  simple_trans binding

template <class C>
struct trans_defs 
{
  typedef typename C::coord_type coord_type;
  typedef typename C::displacement_type displacement_type;

  static int trans_r0 ()   { return db::fixpoint_trans::r0;   }
  static int trans_r90 ()  { return db::fixpoint_trans::r90;  }
  static int trans_r180 () { return db::fixpoint_trans::r180; }
  static int trans_r270 () { return db::fixpoint_trans::r270; }
  static int trans_m0 ()   { return db::fixpoint_trans::m0;   }
  static int trans_m45 ()  { return db::fixpoint_trans::m45;  }
  static int trans_m90 ()  { return db::fixpoint_trans::m90;  }
  static int trans_m135 () { return db::fixpoint_trans::m135; }

  static C new_v ()
  {
    return C ();
  }

  static C new_f (int f)
  {
    return C (db::fixpoint_trans (f));
  }

  static C new_xy (coord_type x, coord_type y)
  {
    return C (displacement_type (x, y));
  }

  static C new_fxy (int f, coord_type x, coord_type y)
  {
    return C (f, displacement_type (x, y));
  }

  static C new_rmxy (int r, bool m, coord_type x, coord_type y)
  {
    return C (r, m, displacement_type (x, y));
  }

  static C new_u (const displacement_type &u)
  {
    return C (u);
  }

  static C new_fu (int f, const displacement_type &u)
  {
    return C (f, u);
  }

  static C new_rmu (int r, bool m, displacement_type &u)
  {
    return C (r, m, u);
  }

  static gsi::Methods methods ()
  {
    return
    method ("new", &new_v, 
      "@brief Creates a unit transformation\n"
    ) +
    method ("new|#new_f", &new_f, 
      "@brief Conversion constructor from a fixpoint transformation\n"
      "\n"
      "@args f\n"
      "\n"
      "This constructor will create a transformation with a fixpoint transformation\n"
      "but no displacement.\n"
      "\n"
      "@param f The rotation/mirror code (r0 .. m135 constants)\n"
    ) +
    method ("new|#new_rmu", &new_rmu,
      "@brief The standard constructor using angle and mirror flag\n"
      "\n"
      "@args rot, mirr, u\n"
      "\n"
      "The sequence of operations is: mirroring at x axis,\n"
      "rotation, application of displacement.\n"
      "\n"
      "@param rot The rotation in units of 90 degree\n"
      "@param mirrx True, if mirrored at x axis\n"
      "@param u The displacement\n"
    ) +
    method ("new|#new_rmxy", &new_rmxy,
      "@brief The standard constructor using angle and mirror flag and two coordinate values for displacement\n"
      "\n"
      "@args rot, mirr, u\n"
      "\n"
      "The sequence of operations is: mirroring at x axis,\n"
      "rotation, application of displacement.\n"
      "\n"
      "@param rot The rotation in units of 90 degree\n"
      "@param mirrx True, if mirrored at x axis\n"
      "@param x The horizontal displacement\n"
      "@param y The vertical displacement\n"
    ) +
    method ("new|#new_fu", &new_fu, 
      "@brief The standard constructor using a code rather than angle and mirror\n"
      "\n"
      "@args f, u\n"
      "\n"
      "@param f The rotation/mirror code (r0 .. m135 constants)\n"
      "@param u The displacement\n"
    ) +
    method ("new|#new_fxy", &new_fxy, 
      "@brief The standard constructor using a code rather than angle and mirror and two coordinate values for displacement\n"
      "\n"
      "@args f, x, y\n"
      "\n"
      "@param f The rotation/mirror code (r0 .. m135 constants)\n"
      "@param x The horizontal displacement\n"
      "@param y The vertical displacement\n"
    ) +
    method ("new|#new_u", &new_u,
      "@brief The standard constructor using a displacement only\n"
      "\n"
      "@args u\n"
      "\n"
      "@param u The displacement\n"
    ) +
    method ("new|#new_xy", &new_xy,
      "@brief The standard constructor using a displacement given as two coordinates\n"
      "\n"
      "@args x, y\n"
      "\n"
      "@param x The horizontal displacement\n"
      "@param y The vertical displacement\n"
    ) +
    method ("inverted", &C::inverted, 
      "@brief Inversion\n"
      "\n"
      "Returns the inverted transformation\n"
      "\n"
      "@return The inverted transformation\n"
    ) +
    method ("invert", &C::invert,
      "@brief In-place inversion\n"
      "\n"
      "Inverts the transformation and replaces this object by the\n"
      "inverted one.\n"
      "\n"
      "@return The inverted transformation\n"
    ) +
    method ("ctrans", &C::template ctrans<coord_type>,
      "@brief The transformation of a distance\n"
      "\n"
      "@args d\n"
      "\n"
      "The \"ctrans\" method transforms the given distance.\n"
      "e = t(d). For the simple transformations, there\n"
      "is no magnification and no modification of the distance\n"
      "therefore.\n"
      "\n"
      "@param d The distance to transform\n"
      "@return The transformed distance\n"
    ) +
    method ("trans", &C::template trans<coord_type>, 
      "@brief The transformation of a point \n"
      "\n"
      "@args p\n"
      "\n"
      "The \"trans\" method transforms the given point.\n"
      "q = t(p)\n"
      "\n"
      "@param p The point to transform\n"
      "@return The transformed point\n"
    ) +
    method ("*", &C::concat,
      "@brief Multiplication (concatenation) of transformations\n"
      "\n"
      "@args *\n"
      "\n"
      "The * operator returns self*t (\"t is applied before this transformation\").\n"
      "\n"
      "@param t The transformation to apply before\n"
      "@return The modified transformation\n"
    ) +
    method ("<", &C::operator<,
      "@brief A sorting criterion\n"
    ) +
    method ("==", &C::operator==,
      "@brief Equality test\n"
    ) +
    method ("!=", &C::operator!=,
      "@brief Inequality test\n"
    ) +
    method ("to_s", (std::string (C::*) () const) &C::to_string,
      "@brief String conversion\n"
    ) +
    method ("disp", &C::disp,
      "@brief Accessor to the point\n"
    ) +
    method ("rot", &C::rot,
      "@brief Accessor to the angle/mirror code\n"
    ) +
    method ("angle", &C::angle,
      "@brief Accessor to the angle in units of 90 degree\n"
    ) +
    method ("is_mirror?", &C::is_mirror,
      "@brief Accessor to the mirror flag\n"
    ) +
    method ("r0", &trans_r0,
      "@brief Angle/mirror code for the \"unrotated\" transformation\n"
    ) +
    method ("r90", &trans_r90,
      "@brief Angle/mirror code for the \"rotated by 90 degree counterclockwise\" transformation\n"
    ) +
    method ("r180", &trans_r180,
      "@brief Angle/mirror code for the \"rotated by 180 degree counterclockwise\" transformation\n"
    ) +
    method ("r270", &trans_r270,
      "@brief Angle/mirror code for the \"rotated by 270 degree counterclockwise\" transformation\n"
    ) +
    method ("m0", &trans_m0,
      "@brief Angle/mirror code for the \"mirrored at the x-axis\" transformation\n"
    ) +
    method ("m45", &trans_m45,
      "@brief Angle/mirror code for the \"mirrored at the 45 degree axis\" transformation\n"
    ) +
    method ("m90", &trans_m90,
      "@brief Angle/mirror code for the \"mirrored at the y (90 degree) axis\" transformation\n"
    ) +
    method ("m135", &trans_m135,
      "@brief Angle/mirror code for the \"mirrored at the 135 degree axis\" transformation\n"
    );
  }
};

db::Trans trans_from_dtrans (const db::DTrans &dtrans)
{
  return db::Trans (dtrans.rot (), db::Point::from_double (dtrans.disp ()));
}

template <>
Class<db::Trans> Class<db::Trans>::decl ("Trans", 
  method ("from_dtrans", &trans_from_dtrans, 
    "@brief Conversion constructor from a floating-point transformation\n"
    "\n"
    "@args double_trans\n"
  ) +
  trans_defs<db::Trans>::methods (),
  "@brief A simple transformation\n"
  "\n"
  "The simple transformation applies a displacement vector\n"
  "and a simple fixpoint transformation. This version acts on integer coordinates.\n"
);

db::DTrans dtrans_from_itrans (const db::Trans &itrans)
{
  return db::DTrans (itrans);
}

template <>
Class<db::DTrans> Class<db::DTrans>::decl ("DTrans", 
  method ("from_itrans", &dtrans_from_itrans, 
    "@brief Conversion constructor from an integer coordinate transformation\n"
    "\n"
    "@args int_trans\n"
  ) +
  trans_defs<db::DTrans>::methods (),
  "@brief A simple transformation\n"
  "\n"
  "The simple transformation applies a displacement vector\n"
  "and a simple fixpoint transformation. This version acts on double coordinates.\n"
);

// ---------------------------------------------------------------
//  complex_trans binding

template <class C>
struct cplx_trans_defs 
{
  typedef typename C::coord_type coord_type;
  typedef typename C::displacement_type displacement_type;
  typedef db::simple_trans<coord_type> simple_trans_type;

  static int trans_r0 ()   { return db::fixpoint_trans::r0;   }
  static int trans_r90 ()  { return db::fixpoint_trans::r90;  }
  static int trans_r180 () { return db::fixpoint_trans::r180; }
  static int trans_r270 () { return db::fixpoint_trans::r270; }
  static int trans_m0 ()   { return db::fixpoint_trans::m0;   }
  static int trans_m45 ()  { return db::fixpoint_trans::m45;  }
  static int trans_m90 ()  { return db::fixpoint_trans::m90;  }
  static int trans_m135 () { return db::fixpoint_trans::m135; }

  static C new_v ()
  {
    return C ();
  }

  static C new_f (int f)
  {
    return C (f);
  }

  static C new_t (const simple_trans_type &t)
  {
    return C (t, 1.0, 1.0);
  }

  static C new_tm (const simple_trans_type &t, double m)
  {
    return C (t, 1.0, m);
  }

  static C new_m (double m)
  {
    return C (m);
  }

  static C new_mrmu (double mag, double r, bool m, const displacement_type &u)
  {
    return C (mag, r, m, u);
  }

  static gsi::Methods methods ()
  {
    return
    method ("new", &new_v, 
      "@brief Creates a unit transformation\n"
    ) +
    method ("new|#new_f", &new_f, 
      "@brief Conversion constructor from a fixpoint transformation\n"
      "\n"
      "@args f\n"
      "\n"
      "This constructor will create a transformation with a fixpoint transformation\n"
      "but no displacement.\n"
      "\n"
      "@param f The rotation/mirror code (r0 .. m135 constants)\n"
    ) +
    method ("new|#new_m", &new_m,
      "@brief Constructor from a magnification\n"
      "\n"
      "@args m\n"
      "\n"
      "Creates a magnifying transformation without displacement and rotation given the magnification m."
    ) +
    method ("new|#new_tm", &new_tm,
      "@brief Constructor from a simple transformation and a magnification\n"
      "\n"
      "@args t, m\n"
      "\n"
      "Creates a magnifying transformation from a simple transformation and a magnification."
    ) +
    method ("new|#new_t", &new_t,
      "@brief Constructor from a simple transformation alone\n"
      "\n"
      "@args t\n"
      "\n"
      "Creates a magnifying transformation from a simple transformation and a magnification of 1.0."
    ) +
    method ("new|#new_mrmu", &new_mrmu,
      "@brief The standard constructor using magnification, angle, mirror flag and displacement\n"
      "@args mag, rot, mirrx, u\n"
      "\n"
      "The sequence of operations is: magnification, mirroring at x axis,\n"
      "rotation, application of displacement.\n"
      "\n"
      "@param mag The magnification\n"
      "@param rot The rotation angle in units of degree\n"
      "@param mirrx True, if mirrored at x axis\n"
      "@param u The displacement\n"
    ) +
    method ("inverted", &C::inverted, 
      "@brief Inversion\n"
      "\n"
      "Returns the inverted transformation\n"
      "\n"
      "@return The inverted transformation\n"
    ) +
    method ("invert", &C::invert,
      "@brief In-place inversion\n"
      "\n"
      "Inverts the transformation and replaces this transformation by the\n"
      "inverted one.\n"
      "\n"
      "@return The inverted transformation\n"
    ) +
    method ("ctrans", &C::template ctrans<coord_type>,
      "@brief The transformation of a distance\n"
      "@args d\n"
      "\n"
      "The \"ctrans\" method transforms the given distance.\n"
      "e = t(d). For the simple transformations, there\n"
      "is no magnification and no modification of the distance\n"
      "therefore.\n"
      "\n"
      "@param d The distance to transform\n"
      "@return The transformed distance\n"
    ) +
    method ("trans", &C::template trans<coord_type>, 
      "@brief The transformation of a point \n"
      "@args p\n"
      "\n"
      "The \"trans\" method transforms the given point.\n"
      "q = t(p)\n"
      "\n"
      "@param p The point to transform\n"
      "@return The transformed point\n"
    ) +
    method ("*", &C::concat,
      "@brief Multiplication (concatenation) of transformations\n"
      "@args t\n"
      "\n"
      "The * operator returns self*t (\"t is applied before this transformation\").\n"
      "\n"
      "@param t The transformation to apply before\n"
      "@return The modified transformation\n"
    ) +
    method ("<", &C::operator<,
      "@brief A sorting criterion\n"
    ) +
    method ("==", &C::operator==,
      "@brief Equality test\n"
    ) +
    method ("!=", &C::operator!=,
      "@brief Inequality test\n"
    ) +
    method ("to_s", (std::string (C::*) () const) &C::to_string,
      "@brief String conversion\n"
    ) +
    method ("disp", (const displacement_type &(C::*)() const) &C::disp,
      "@brief Accessor to the point\n"
    ) +
    method ("disp=", (void (C::*) (const displacement_type &)) &C::disp,
      "@brief Write accessor to the point\n"
      "@args u\n"
      "@param u The new displacement"
    ) +
    method ("rot", &C::rot,
      "@brief Return the respective rotation code if possible\n"
      "\n"
      "If this transformation is orthogonal (is_ortho () == true), then this method\n"
      "will return the corresponding fixpoint transformation, not taking into account\n"
      "magnification and displacement. If the transformation is not orthogonal, the result\n"
      "reflects the quadrant the rotation goes into with the guarantee to reproduce the \n"
      "correct quadrant in the exact case.\n"
    ) +
    method ("is_mirror?", &C::is_mirror,
      "@brief Accessor to the mirror flag\n"
    ) +
    method ("mirror=", &C::mirror,
      "@brief Write accessor to the mirror flag\n"
      "@args m\n"
      "\"mirroring\" describes a reflection at the x-axis which is included in the transformation prior to rotation."
      "@param m The new mirror flag"
    ) +
    method ("is_unity?", &C::is_unity,
      "@brief Test, whether this is a unit transformation\n"
    ) +
    method ("is_ortho?", &C::is_ortho,
      "@brief Test, if the transformation is an orthogonal transformation\n"
      "\n"
      "If the rotation is by a multiple of 90 degree, this method will return true.\n"
    ) +
    method ("s_trans", &C::template s_trans<simple_trans_type>,
      "@brief Extract the simple transformation part\n"
      "\n"
      "The simple transformation part does not reflect magnification not arbitrary angles.\n"
      "On the angle contribution up to a multiple of 90 degree is reflected.\n"
    ) +
    method ("angle", (double (C::*) () const) &C::angle,
      "@brief Read accessor for the angle\n"
      "\n"
      "To check, if the transformation represents a rotation by a angle that\n"
      "is a multiple of 90 degree, use this predicate.\n"
      "\n"
      "@return The rotation angle this transformation provides in degree units (0..360 deg).\n"
    ) +
    method ("angle=", (void (C::*) (double)) &C::angle,
      "@brief Write accessor for the angle\n"
      "@args a\n"
      "@param a The new angle"
    ) +
    method ("mag", (double (C::*) () const) &C::mag,
      "@brief Read accessor to the magnification\n"
    ) +
    method ("is_mag?", &C::is_mag,
      "@brief Test, if the transformation is a magnifying one\n"
      "\n"
      "This is the recommended test for checking if the transformation represents\n"
      "a magnification.\n"
    ) +
    method ("mag=", (void (C::*) (double)) &C::mag,
      "@brief Write accessor to the magnification\n"
      "@args m\n"
      "@param m The new magnification"
    ) +
    method ("r0", &trans_r0,
      "@brief Angle/mirror code for the \"unrotated\" transformation\n"
    ) +
    method ("r90", &trans_r90,
      "@brief Angle/mirror code for the \"rotated by 90 degree counterclockwise\" transformation\n"
    ) +
    method ("r180", &trans_r180,
      "@brief Angle/mirror code for the \"rotated by 180 degree counterclockwise\" transformation\n"
    ) +
    method ("r270", &trans_r270,
      "@brief Angle/mirror code for the \"rotated by 270 degree counterclockwise\" transformation\n"
    ) +
    method ("m0", &trans_m0,
      "@brief Angle/mirror code for the \"mirrored at the x-axis\" transformation\n"
    ) +
    method ("m45", &trans_m45,
      "@brief Angle/mirror code for the \"mirrored at the 45 degree axis\" transformation\n"
    ) +
    method ("m90", &trans_m90,
      "@brief Angle/mirror code for the \"mirrored at the y (90 degree) axis\" transformation\n"
    ) +
    method ("m135", &trans_m135,
      "@brief Angle/mirror code for the \"mirrored at the 135 degree axis\" transformation\n"
    );
  }
};

db::DCplxTrans dcplx_trans_from_itrans (const db::CplxTrans &itrans)
{
  return db::DCplxTrans (itrans);
}

template <>
Class<db::DCplxTrans> Class<db::DCplxTrans>::decl ("DCplxTrans", 
  method ("from_itrans", &dcplx_trans_from_itrans, 
    "@brief Conversion constructor from an integer coordinate transformation\n"
    "\n"
    "@args int_trans\n"
  ) +
  cplx_trans_defs<db::DCplxTrans>::methods (),
  "@brief A complex transformation\n"
  "\n"
  "A complex transformation provides magnification, mirroring at the x-axis, rotation by an arbitary\n"
  "angle and a displacement. A complex transformation provides a superset of the simple transformation.\n"
  "In many applications, a complex transformation computes floating-point coordinates to minimize rounding effects.\n"
);

db::CplxTrans cplx_trans_from_dtrans (const db::DCplxTrans &dtrans)
{
  return db::CplxTrans (dtrans);
}

template <>
Class<db::CplxTrans> Class<db::CplxTrans>::decl ("CplxTrans", 
  method ("from_dtrans", &cplx_trans_from_dtrans, 
    "@brief Conversion constructor from an floating-point transformation\n"
    "\n"
    "@args dbl_trans\n"
  ) +
  cplx_trans_defs<db::CplxTrans>::methods (),
  "@brief A complex transformation\n"
  "\n"
  "A complex transformation provides magnification, mirroring at the x-axis, rotation by an arbitary\n"
  "angle and a displacement.\n"
);

}

