
/*

  KLayout Layout Viewer
  Copyright (C) 2006-2007 Matthias Koefferlein

  This program is free software; you can redistribute it and/or modify
  it under the terms of the GNU General Public License as published by
  the Free Software Foundation; either version 2 of the License, or
  (at your option) any later version.

  This program is distributed in the hope that it will be useful,
  but WITHOUT ANY WARRANTY; without even the implied warranty of
  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
  GNU General Public License for more details.

  You should have received a copy of the GNU General Public License
  along with this program; if not, write to the Free Software
  Foundation, Inc., 51 Franklin St, Fifth Floor, Boston, MA  02110-1301  USA

*/


#include "gsi.h"
#include "layApplication.h"
#include "layMainWindow.h"
#include "layLayoutView.h"
#include "dbSaveLayoutOptions.h"

namespace gsi
{

static void remove_stipple (lay::LayoutView *view, unsigned int index)
{
  lay::DitherPattern pattern (view->dither_pattern ());

  if (index >= (unsigned int) std::distance (pattern.begin (), pattern.begin_custom ()) && 
      index < (unsigned int) std::distance (pattern.begin (), pattern.end ())) {
    lay::DitherPatternInfo p;
    pattern.replace_pattern (index, p); 
    pattern.renumber ();
    view->set_dither_pattern (pattern);
  }
}

static void clear_stipples (lay::LayoutView *view)
{
  lay::DitherPattern no_stipples;
  view->set_dither_pattern (no_stipples);
}

static unsigned int add_stipple (lay::LayoutView *view, const std::string &name, const std::vector<unsigned int> &data, unsigned int bits)
{
  lay::DitherPattern pattern (view->dither_pattern ());

  lay::DitherPatternInfo p;
  p.set_name (name);
  if (bits > 0 && ! data.empty ()) {
    p.set_pattern (&(*data.begin ()), std::min (32u, bits), std::min (size_t (32), data.size ()));
  }
  unsigned int index = pattern.add_pattern (p);

  view->set_dither_pattern (pattern);

  return index;
}

template <>
Class<lay::LayoutView> Class<lay::LayoutView>::decl ("LayoutView", 
  method ("stop_redraw", &lay::LayoutView::stop_redraw,
    "@brief Stop the redraw thread\n"
    "\n"
    "It is very important to stop the redraw thread before applying changes to the "
    "layout or the cell views and the LayoutView configuration. This is usually done automatically. "
    "For rare cases, where this is not the case, this method is provided.\n"
  ) +
  method ("set_title", &lay::LayoutView::set_title,
    "@brief Set the title of the view\n"
    "\n"
    "@args title\n"
    "@param title The title string to use\n"
    "\n"
    "Override the standard title of the view indicating the file names loaded by "
    "the specified title string. The title string can be reset with \\reset_title to "
    "the standard title again."
  ) +
  method ("reset_title", &lay::LayoutView::reset_title,
    "@brief Reset the title to the standard title\n"
    "\n"
    "See \\set_title and \\title for a description about how titles are handled."
  ) +
  method ("title", &lay::LayoutView::title,
    "@brief Return the view's title string\n"
    "\n"
    "@return The title string\n"
    "\n"
    "The title string is either a string composed of the file names loaded (in some "
    "\"readable\" manner) or a customized title string set by \\set_title."
  ) +
  method ("save_layer_props", &lay::LayoutView::save_layer_props, 
    "@brief Save the layer properties\n"
    "\n"
    "@args fn\n"
    "\n"
    "Save the layer properties to the file given in \"fn\""
  ) +
  method ("load_layer_props", &lay::LayoutView::load_layer_props, 
    "@brief Load the layer properties\n"
    "\n"
    "@args fn\n"
    "\n"
    "Load the layer properties from the file given in \"fn\""
  ) +
  method ("min_hier_levels=", &lay::LayoutView::set_min_hier_levels,
    "@brief Set the minimum hierarchy level at which to display geometries\n"
    "\n"
    "@args level\n"
    "@param level The minimum level above which to display something\n"
    "\n"
    "This methods allows to set the minimum hierarchy level above which to display geometries."
    "This method may cause a redraw if required."
  ) +
  method ("min_hier_levels?", &lay::LayoutView::get_min_hier_levels,
    "@brief Query the minimum hierarchy level at which to display geometries\n"
    "\n"
    "@return The minimum level at which to display geometries"
  ) +
  method ("max_hier_levels=", &lay::LayoutView::set_max_hier_levels,
    "@brief Set the maximum hierarchy level up to which to display geometries\n"
    "\n"
    "@args level\n"
    "@param level The maximum level below which to display something\n"
    "\n"
    "This methods allows to set the maximum hierarchy below which to display geometries."
    "This method may cause a redraw if required."
  ) +
  method ("max_hier_levels?", &lay::LayoutView::get_max_hier_levels,
    "@brief Query the maximum hierarchy level up to which to display geometries\n"
    "\n"
    "@return The maximum level up to which to display geometries"
  ) +
  method ("reload_layout", &lay::LayoutView::reload_layout,
    "@brief Reload the given cellview\n"
    "\n"
    "@args cv\n"
    "\n"
    "@param cv The index of the cellview to reload"
  ) + 
  method ("create_layout", &lay::LayoutView::create_layout,
    "@brief Create a new, empty layout\n"
    "\n"
    "@args add_cellview\n"
    "\n"
    "The add_cellview parameter controls whether to create a new cellview (true)\n"
    "or clear all cellviews before.\n"
    "\n"
    "@return The index of the cellview created.\n"
  ) +
  method ("erase_cellview", &lay::LayoutView::erase_cellview,
    "@brief Erase the cellview with the given index\n"
    "\n"
    "@args index\n"
    "\n"
    "This closes the given cellview and unloads the layout associated with it, unless referred to by another cellview."
  ) +
  method ("rename_cellview", &lay::LayoutView::rename_cellview,
    "@brief Rename the cellview with the given index\n"
    "\n"
    "@args name, index\n"
    "\n"
    "If the name is not unique, a unique name will be constructed from the name given.\n"
    "The name may be different from the filename but is associated with the layout object.\n"
    "If a layout is shared between multiple cellviews (which may happen due to a clone of the layout view\n"
    "for example), both cellviews are renamed.\n"
  ) +
  method ("load_layout", &lay::LayoutView::load_layout,
    "@brief Load a (new) file into the layout\n"
    "\n"
    "@args filename, add_cellview\n"
    "\n"
    "Loads the file given by the \"filename\" parameter.\n"
    "The add_cellview param controls whether to create a new cellview\n"
    "or clear all cellviews before.\n"
    "\n"
    "@return The index of the cellview loaded.\n"
  ) +
  method ("active_cellview_index", &lay::LayoutView::active_cellview_index,
    "@brief Get the index of the active cellview (shown in hierarchy browser)\n"
  ) +
  method ("set_active_cellview_index", &lay::LayoutView::set_active_cellview_index,
    "@brief Make the cellview with the given index the active one (shown in hierarchy browser)\n"
    "@args index\n"
    "See \\active_cellview_index."
  ) +
  method ("get_current_cell_path", &lay::LayoutView::get_current_cell_path,
    "@brief Cell path of the current cell\n"
    "\n"
    "@args cv_index\n"
    "\n"
    "The current cell is the one highlighted in the browser with the focus rectangle. The \n"
    "current path is returned for the cellview given by cv_index.\n"
    "The cell path is a list of cell indices from the top cell to the current cell.\n"
    "\n"
    "@param cv_index The cellview index for which to get the current path from (usally this will be the active cellview index)"
  ) +
  method ("set_current_cell_path", (void (lay::LayoutView::*) (int, const lay::LayoutView::cell_path_type &)) &lay::LayoutView::set_current_cell_path,
    "@brief Set the path to the current cell\n"
    "\n"
    "@args cv_index\n"
    "\n"
    "The current cell is the one highlighted in the browser with the focus rectangle. The\n"
    "cell given by the path is hightlighted and scrolled into view.\n"
    "To select the cell to be drawn, use the \\select_cell or \\select_cell_path method.\n"
    "\n"
    "@param cv_index The cellview index for which to set the current path for (usally this will be the active cellview index)\n"
    "@param path The path to the current cell\n"
  ) +
  method ("cellviews", &lay::LayoutView::cellviews,
    "@brief Get the number of cellviews\n"
  ) + 
  method ("cellview", &lay::LayoutView::cellview,
    "@brief Get the cellview object for a given index\n"
    "\n"
    "@args cv_index\n"
    "\n"
    "@param cv_index The cellview index for which to get the object for"
  ) + 
  method ("zoom_fit", &lay::LayoutView::zoom_fit,
    "@brief Fit the contents of the current view into the window"
  ) +
  method ("zoom_box", &lay::LayoutView::zoom_box,
    "@brief Set the viewport to the given box\n"
    "\n"
    "@args box\n"
    "@param box The box to which to set the view in micron coordinates\n"
  ) +
  method ("zoom_in", &lay::LayoutView::zoom_in,
    "@brief Zoom in somewhat"
  ) +
  method ("zoom_out", &lay::LayoutView::zoom_out,
    "@brief Zoom out somewhat"
  ) +
  method ("pan_up", &lay::LayoutView::pan_up,
    "@brief Pan upward" 
  ) +
  method ("pan_down", &lay::LayoutView::pan_down,
    "@brief Pan down" 
  ) +
  method ("pan_left", &lay::LayoutView::pan_left,
    "@brief Pan to the left" 
  ) +
  method ("pan_right", &lay::LayoutView::pan_right,
    "@brief Pan to the right" 
  ) +
  method ("pan_center", &lay::LayoutView::pan_center,
    "@brief Pan to the given point\n" 
    "\n"
    "@args p\n"
    "\n"
    "The window is positioned such that \"p\" becomes the new center"
  ) +
  method ("box", &lay::LayoutView::box,
    "@brief Return the displayed box in micron space" 
  ) +
  method ("bookmark_view", &lay::LayoutView::bookmark_view,
    "@brief Bookmark the current view under the given name\n"
    "\n"
    "@args name\n"
    "\n"
    "@param name The name under which to bookmark the current state"
  ) +
  method ("cancel", &lay::LayoutView::cancel,
    "@brief Cancel all edit operations"
  ) +
  method ("stop", &lay::LayoutView::stop,
    "@brief Stop redraw thread and close any browsers\n"
    "This method usually does not need to be called explicitly. The redraw thread is stopped automatically."
  ) +
  method ("enable_edits", &lay::LayoutView::enable_edits,
    "@brief Enable or disable editing\n"
    "\n"
    "@args enable\n"
    "\n"
    "@param enable True to enable edits"
  ) +
  method ("select_cell_path", (void (lay::LayoutView::*) (const lay::LayoutView::cell_path_type &, int)) &lay::LayoutView::select_cell,
    "@brief Select a cell by cell index for a certain cell view\n"
    "\n"
    "@args cell_index, cv_index\n"
    "\n"
    "Select the current (top) cell by specifying a cell indexand the cellview index for which this cell should become the currently shown one. The path to the cell is constructed by "
    "selecting one that leads to a top cell.\n"
    "This method selects the cell to be drawn. In constrast, the \\set_current_cell_path method selects "
    "the cell that is highlighted in the cell tree (but not necessarily drawn)."
  ) +
  method ("select_cell", (void (lay::LayoutView::*) (lay::LayoutView::cell_index_type, int)) &lay::LayoutView::select_cell,
    "@brief Select a cell by index for a certain cell view\n"
    "\n"
    "@args cell_index, cv_index\n"
    "\n"
    "Select the current (top) cell by specifying a path (a list of cell indices from top to "
    "the actual cell) and the cellview index for which this cell should become the currently "
    "shown one.\n"
    "This method selects the cell to be drawn. In constrast, the \\set_current_cell_path method selects "
    "the cell that is highlighted in the cell tree (but not necessarily drawn)."
  ) +
  method ("descend", &lay::LayoutView::descend, 
    "@brief Descend further into the hierarchy.\n"
    "@args path, index\n"
    "Adds the given path (given as an array of InstElement objects) to the specific path of the "
    "cellview with the given index. In effect, the cell addressed by the terminal of the new path "
    "components can be shown in the context of the upper cells, if the minimum hierarchy level is "
    "set to a negative value.\n"
    "The path is assumed to originate from the current cell and contain specific instances sorted from "
    "top to bottom."
  ) +
  method ("ascend", &lay::LayoutView::ascend, 
    "@brief Ascend upwards in the hierarchy.\n"
    "@args index\n"
    "Removes one element from the specific path of the cellview with the given index. Returns the element "
    "removed."
  ) +
  method ("is_cell_hidden", &lay::LayoutView::is_cell_hidden,
    "@brief Tell, if the cell is hidden\n"
    "\n"
    "@args cell_index, cv_index\n"
    "\n"
    "@return True, if the cell with \"cell_index\" is hidden for the cellview \"cv_index\""
  ) +
  method ("hide_cell", &lay::LayoutView::hide_cell,
    "@brief Hide the given cell for the given cellview\n"
    "\n"
    "@args cell_index, cv_index\n"
  ) +
  method ("show_cell", &lay::LayoutView::show_cell,
    "@brief Show the given cell for the given cellview (cancel effect of \\hide_cell)\n"
    "\n"
    "@args cell_index, cv_index\n"
  ) +
  method ("show_all_cells", &lay::LayoutView::show_all_cells,
    "@brief Make all cells shown (cancel effects of \\hide_cell)"
  ) +
  method ("update_content", &lay::LayoutView::force_update_content,
    "@brief Update the layout view to the current state\n"
    "\n"
    "This method triggers an update of the hierarchy tree and layer view tree. Usually, this "
    "method does not need to be called. The widgets are updated automatically in most cases.\n"
    "\n"
    "Currently, this method MUST be called however, after the layer view tree has been changed by "
    "the \\insert_layer, \\replace_layer_node or \\delete_layer methods.\n" 
  ) +
  method ("max_hier", &lay::LayoutView::max_hier,
    "@brief Select all hierarchy levels available\n"
    "\n"
    "Show the layout in full depth down to the deepest level of hierarchy. "
    "This method may cause a redraw."
  ) +
  method ("save_screenshot", &lay::LayoutView::save_screenshot,
    "@brief Save a screenshot to the given file\n"
    "\n"
    "@args filename\n"
    "@param filename The file to which to write the screenshot to.\n"
    "\n"
    "The screenshot is written as a PNG file to the given file. "
    "This requires the drawing to be complete. Ideally, synchroneous mode is switched on "
    "for the application to guarantee this condition. The image will have the size of the viewport "
    "showing the current layout."
  ) +
  method ("save_image", &lay::LayoutView::save_image,
    "@brief Save the layout as an image to the given file\n"
    "\n"
    "@args filename, width, height\n"
    "@param filename The file to which to write the screenshot to.\n"
    "@param width The width of the image to render in pixel.\n"
    "@param height The height of the image to render in pixel.\n"
    "\n"
    "The image is written as a PNG file to the given file. "
    "The image is drawn synchroneously with the given width and height. Drawing may take some time. "
  ) +
  method ("save_as", &lay::LayoutView::save_as,
    "@brief Save a layout to the given stream file\n"
    "\n"
    "@args index, filename, gzip, options\n"
    "@param index The cellview index of the layout to save.\n"
    "@param filename The file to write.\n"
    "@param gzip True, if the file should be gzipped.\n"
    "@param options Writer options.\n"
    "\n"
    "The given layout (with the given index) is written to the stream file with the given options. "
    "'options' is a \\SaveLayoutOptions object that specifies which format to write and further options such "
    "as scaling factor etc.\n"
    "Calling this method is equivalent to calling 'write' on the respective layout object.\n"
  ) +
  method ("set_layer_properties", (void (lay::LayoutView::*) (const lay::LayerPropertiesConstIterator &, const lay::LayerProperties &)) &lay::LayoutView::set_properties, 
    "@brief Set the layer properties of the layer pointed to by the iterator\n"
    "@args iter, props\n"
    "\n"
    "This method replaces the layer properties of the element pointed to by \"iter\" by the properties "
    "given by \"props\". It will not change the hierarchy but just the properties of the given node."
  ) +
  method ("replace_layer_node", &lay::LayoutView::replace_layer_node, 
    "@brief Replace the layer node at the position given by \"iter\" with a new one\n"
    "@args iter, node\n"
  ) +
  method ("insert_layer", &lay::LayoutView::insert_layer, 
    "@brief Insert the given layer properties node into the list before the given position\n"
    "@args iter, node\n"
    "\n"
    "This method inserts the new properties node before the position given by \"iter\" and returns "
    "a const reference to the element created.\n"
  ) +
  method ("delete_layer", &lay::LayoutView::delete_layer, 
    "@brief Delete the layer properties node \n"
    "@args iter\n"
    "\n"
    "This method deletes the object that the iterator points to and invalidates\n"
    "the iterator since the object that the iterator points to is no longer valid.\n"
  ) +
  method ("begin_layers", &lay::LayoutView::begin_layers, 
    "@brief Begin iterator for the layers\n"
    "\n"
    "This iterator delivers the layers of this view, either in a recursive or non-recursive\n"
    "fashion, depending which iterator increment methods are used.\n"
    "The iterator delivered by \\end_layers is the past-the-end iterator. It can be compared\n"
    "agains a current iterator to check, if there are no further elements.\n"
  ) +
  method ("end_layers", &lay::LayoutView::end_layers, 
    "@brief End iterator for the layers\n"
    "See \\begin_layers for a description about this iterator\n"
  ) +
  method ("clear_layers", &lay::LayoutView::clear_layers, 
    "@brief Clear all layers\n"
  ) +
  method_ext ("remove_stipple", &remove_stipple,
    "@brief Remove the stipple pattern with the given index\n"
    "@args index\n"
    "The pattern with an index less than 16 cannot be removed. "
    "If a stipple pattern is removed that is still used, the results are not predictable. "
  ) +
  method_ext ("clear_stipples", &clear_stipples,
    "@brief Remove all stipple pattern\n"
    "All stipple pattern except the fixed ones are removed. If any of the custom stipple pattern is "
    "still used by the layers displayed, the results will not be predictable."
  ) +
  method_ext ("add_stipple", &add_stipple,
    "@brief Add a stipple pattern\n"
    "@args name, data, bits\n"
    "\n"
    "'data' is an array of unsigned integers describing the bits that make up the stipple "
    "pattern. If the array has less than 32 entries, the pattern will be repeated vertically. "
    "The number of bits used can be less than 32 bit which can be specified by the 'bits' parameter. "
    "Logically, the pattern will be put at the end of the list.\n"
    "\n"
    "@param name The name under which this pattern will appear in the stipple editor\n"
    "@param data See above\n"
    "@param bits See above\n"
    "@return The index of the newly created stipple pattern, which can be used as the dither pattern index of \\LayerProperties."
  ) +
  method ("current_layer", &lay::LayoutView::current_layer, 
    "@brief Get the current layer view\n"
    "\n"
    "Returns the \\LayerPropertiesIterator pointing to the current layer view (the one that has the focus). "
    "If no layer view is active currently, a null iterator is returned.\n"
  ) +
  method ("selected_layers", &lay::LayoutView::selected_layers, 
    "@brief Get the selected layers\n"
    "\n"
    "Returns an array of \\LayerPropertiesIterator objects pointing to the currently selected layers. "
    "If no layer view is selected currently, an empty array is returned.\n"
  ) +
  method ("add_cellview_list_observer", &lay::LayoutView::add_cellview_list_observer,
    "@brief Add a cellview list observer\n"
    "@args observer\n"
    "\n"
    "If a cellview is added or removed, this observer is triggered.\n"
  ) +
  method ("remove_cellview_list_observer", &lay::LayoutView::remove_cellview_list_observer,
    "@brief Remove a cellview list observer\n"
    "@args observer\n"
  ) +
  method ("add_cellview_observer", &lay::LayoutView::add_cellview_observer,
    "@brief Add a cellview observer\n"
    "@args observer\n"
    "\n"
    "If a cellview is changed (i.e. the cell is changed) this event is sent.\n"
    "The integer argument slot (\"signal_int\") of the observer will be triggered as well "
    "with the index of the cellview that has changed.\n"
  ) +
  method ("remove_cellview_observer", &lay::LayoutView::remove_cellview_observer,
    "@brief Remove a cellview observer\n"
    "@args observer\n"
  ) +
  method ("add_file_open_observer", &lay::LayoutView::add_file_open_observer,
    "@brief Add a file open observer\n"
    "@args observer\n"
    "\n"
    "If a new file is loaded, this observer is triggered.\n"
  ) +
  method ("remove_file_open_observer", &lay::LayoutView::remove_file_open_observer,
    "@brief Remove a file open observer\n"
    "@args observer\n"
  ) +
  method ("add_viewport_changed_observer", &lay::LayoutView::add_viewport_changed_observer,
    "@brief Add a viewport changed observer\n"
    "@args observer\n"
    "\n"
    "If the viewport (the rectangle that is shown) changes, this observer\n"
    "is triggered.\n"
  ) +
  method ("remove_viewport_changed_observer", &lay::LayoutView::remove_viewport_changed_observer,
    "@brief Remove a viewport changed observer\n"
    "@args observer\n"
  ) +
  method ("add_layer_list_observer", &lay::LayoutView::add_layer_list_observer,
    "@brief Add a layer list observer\n"
    "@args observer\n"
    "\n"
    "If the layer list changes, the observer's \"signal_int\" and \"signal\" slot is triggered. \n"
    "The int's bit 0 is set, if the properties have changed. If the arguments bit 1 is\n"
    "set, the hierarchy has changed.\n"
  ) +
  method ("remove_layer_list_observer", &lay::LayoutView::remove_layer_list_observer,
    "@brief Remove a layer list observer\n"
    "@args observer\n"
  ) +
  method ("add_cell_visibility_observer", &lay::LayoutView::add_cell_visibility_observer,
    "@brief Add a cell visibility observer\n"
    "@args observer\n"
    "\n"
    "If a cell is hidden or shown, this observer is triggered.\n"
  ) +
  method ("remove_cell_visibility_observer", &lay::LayoutView::remove_cell_visibility_observer,
    "@brief Remove a cell visibility observer\n"
    "@args observer\n"
  ) +
  //  HINT: the cast is important to direct GSI to the LayoutView method rather than the
  //  Plugin method (in which case we get a segmentation violation ..)
  method ("clear_config", (void (lay::LayoutView::*)()) &lay::LayoutView::clear_config,
    "@brief Clear the local configuration parameters\n"
    "\n"
    "See \\set_config for a description of the local configuration parameters."
  ) +
  method ("get_config", (std::string (lay::LayoutView::*)(const std::string &name) const) &lay::LayoutView::config_get,
    "@brief Query a local configuration parameter\n"
    "\n"
    "@args name\n"
    "@param name The name of the configuration parameter whose value shall be obtained (a string)\n"
    "\n"
    "@return The value of the parameter\n"
    "\n"
    "See \\set_config for a description of the local configuration parameters."
  ) +
  method ("set_config", (void (lay::LayoutView::*)(const std::string &name, const std::string &value)) &lay::LayoutView::config_set,
    "@brief Set a local configuration parameter with the given name to the given value\n"
    "\n"
    "@args name, value\n"
    "@param name The name of the configuration parameter to set\n"
    "@param value The value to which to set the configuration parameter\n"
    "\n"
    "This method sets a local configuration parameter with the given name to the given value. "
    "Values can only be strings. Numerical values have to be converted into strings first. "
    "Local configuration parameters override global configurations for this specific view. "
    "This allows for example to override global settings of background colors. "
    "Any local settings are not written to the configuration file. "
  ),
  "@brief The view object presenting one or more layout objects\n"
  "\n"
  "The visual part of the view is the tab panel in the main window. The non-visual part "
  "are the redraw thread, the layout handles, cell lists, layer view lists etc. "
  "This object controls these aspects of the view and controls the appearance of the data."
);

static db::Layout &get_layout (const lay::CellView *cv)
{
  return (*cv)->layout ();
}

static const std::string &name (const lay::CellView *cv)
{
  return (*cv)->name ();
}

static const std::string &filename (const lay::CellView *cv)
{
  return (*cv)->filename ();
}

template <>
Class<lay::CellView> Class<lay::CellView>::decl ("CellView", 
  method ("==", &lay::CellView::operator==,
    "@brief Equality: compares the cell the view points to, not the path\n"
    "@args other\n"
  ) +
  method ("is_valid?", &lay::CellView::is_valid,
    "@brief Test if the view points to a valid cell\n"
  ) +
  method ("set_path", &lay::CellView::set_unspecific_path,
    "@brief Set the unspecific part of the path explicitly\n"
    "@args path\n"
    "\n"
    "Setting the unspecific part of the path will clear the context path component and\n"
    "update the context and target cell.\n"
  ) +
  method ("set_context_path", &lay::CellView::set_specific_path,
    "@brief Set the context path explicitly\n"
    "@args path\n"
    "\n"
    "This method assumes that the unspecific part of the path \n"
    "is established already and that the context path starts\n"
    "from the context cell.\n"
  ) +
  method ("set_cell", (void (lay::CellView::*) (lay::CellView::cell_index_type)) &lay::CellView::set_cell,
    "@brief Set the path to the given cell\n"
    "\n"
    "This method will construct any path to this cell, not a \n"
    "particular one. It will clear the context path\n"
    "and update the context and target cell.\n"
  ) +
  method ("set_cell_name", (void (lay::CellView::*) (const std::string &)) &lay::CellView::set_cell,
    "@brief Set the cell by name\n"
    "\n"
    "If the name is not a valid one, the cellview will become\n"
    "invalid.\n"
    "This method will construct any path to this cell, not a \n"
    "particular one. It will clear the context path\n"
    "and update the context and target cell.\n"
  ) +
  method ("reset_cell", &lay::CellView::reset_cell,
    "@brief Reset the cell \n"
    "\n"
    "The cellview will become invalid. The layout object will\n"
    "still be attached to the cellview.\n"
  ) +
  method ("ctx_cell_index", &lay::CellView::ctx_cell_index,
    "@brief Get the context cell's index\n"
  ) +
  method ("ctx_cell", &lay::CellView::ctx_cell,
    "@brief Get the reference to the context cell currently addressed\n"
  ) +
  method ("cell_index", &lay::CellView::cell_index,
    "@brief Get the target cell's index\n"
  ) +
  method ("cell", &lay::CellView::cell,
    "@brief Get the reference to the target cell currently addressed\n"
  ) +
  method_ext ("filename", &gsi::filename,
    "@brief Get filename associated with the layout behind the cellview\n"
  ) +
  method_ext ("name", &gsi::name,
    "@brief Get the unique name associated with the layout behind the cellview\n"
  ) +
  method ("path", &lay::CellView::unspecific_path,
    "@brief Get the cell's unspecific part of the path leading to the context cell\n"
  ) +
  method ("context_path", &lay::CellView::specific_path,
    "@brief Get the cell's context path\n"
    "The context path leads from the context cell to the target cell in a specific "
    "fashion, i.e. describing each instance in detail, not just be cell indices. If "
    "the context and target cell are identical, the context path is empty."
  ) +
  method_ext ("layout", &get_layout,
    "@brief Get the reference to the layout object addressed by this view\n"
  ),
  "@brief A \"cell view\" reference\n"
  "\n"
  "A cell view reference points to a certain cell within a certain layout.\n"
  "The layout pointer can be nil, indicating that it is invalid.\n"
  "Also, the cell view describes a cell within that layout. The cell\n"
  "is addressed by an cell_index or a cell object reference.\n"
  "\n"
  "The cell is not only identified by it's index or object but as well \n"
  "by the path leading to that cell. This path describes how to find the\n"
  "cell in the context of it's parent cells. \n"
  "\n"
  "The path is in fact composed in two ways: once in an unspecific fashion,\n"
  "just describing which parent cells are used. The target of this path\n"
  "is called the context cell. It is accessible by the \\ctx_cell_index\n"
  "or \\ctx_cell methods.\n"
  "\n"
  "Additionally the path may further identify a certain instance of a certain\n"
  "subcell in the context cell. This is done through a set of \\InstElement\n"
  "objects. The target of this context path is the actual cell addressed by the\n"
  "cellview. This target cell is accessible by the \\cell_index or \\cell methods.\n"
  "In the viewer, the target cell is shown in the context of the context cell.\n"
  "The hierarchy levels are counted from the context cell, which is on level 0.\n"
  "If the context path is empty, the context cell is identical with the target cell.\n"
); 

}

