
/*

  KLayout Layout Viewer
  Copyright (C) 2006-2007 Matthias Koefferlein

  This program is free software; you can redistribute it and/or modify
  it under the terms of the GNU General Public License as published by
  the Free Software Foundation; either version 2 of the License, or
  (at your option) any later version.

  This program is distributed in the hope that it will be useful,
  but WITHOUT ANY WARRANTY; without even the implied warranty of
  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
  GNU General Public License for more details.

  You should have received a copy of the GNU General Public License
  along with this program; if not, write to the Free Software
  Foundation, Inc., 51 Franklin St, Fifth Floor, Boston, MA  02110-1301  USA

*/


#include "layAbstractMenu.h"
#include "layMainWindow.h"
#include "layExceptions.h"
#include "tlAssert.h"
#include "gtf.h"
#include "gsi.h"

#include <QtGui/QAction>
#include <QtGui/QMenu>
#include <QtGui/QMenuBar>
#include <QtGui/QToolBar>
#include <QtGui/QToolButton>

namespace lay
{

// ---------------------------------------------------------------
//  AbstractMenuItem implementation

AbstractMenuItem::AbstractMenuItem () 
  : mp_menu (0), m_has_submenu (false)
{
  //  ... nothing yet ..
}

AbstractMenuItem::AbstractMenuItem (const AbstractMenuItem &) 
  : mp_menu (0), m_has_submenu (false)
{ 
  //  ... nothing yet ..
}

void 
AbstractMenuItem::setup_item (const std::string &pn, const std::string &s, const Action &a)
{
  m_basename.clear ();

  tl::Extractor ex (s.c_str ());

  m_name = pn;
  if (! m_name.empty ()) {
    m_name += ".";
  }

  if (! ex.at_end ()) {

    ex.read (m_basename, ":");
    m_name += m_basename;

    while (ex.test (":")) {
      std::string g;
      ex.read (g, ":");
      m_groups.insert (g);
    }

  }

  set_action (a);
}

void 
AbstractMenuItem::set_action (const Action &a)
{
  bool enabled = m_action.is_enabled ();
  bool visible = m_action.is_visible ();
  m_action = a;
  m_action.set_enabled (enabled);
  m_action.set_visible (visible);

  m_action.qaction ()->setObjectName (m_basename.c_str ());
}

void 
AbstractMenuItem::set_action_title (const std::string &s)
{
  m_action.set_title (s);
}

void  
AbstractMenuItem::set_has_submenu ()
{
  m_has_submenu = true;
}

void  
AbstractMenuItem::set_menu (QMenu *menu)
{
  if (mp_menu) {
    delete mp_menu;
  }
  mp_menu = menu;
  if (mp_menu) {
    mp_menu->setObjectName (m_basename.c_str ());
  }
}

// ---------------------------------------------------------------
//  Action implementation

ActionHandle::ActionHandle (QAction *action)
  : mp_action (action), 
    m_ref_count (0)
{ 
  //  catch the destroyed signal to tell if the QAction object is deleted.
  connect (action, SIGNAL (destroyed (QObject *)), this, SLOT (destroyed (QObject *)));
}

ActionHandle::~ActionHandle ()
{
  if (mp_action) {
    delete mp_action;
  }
  mp_action = 0;
}

void 
ActionHandle::add_ref () 
{
  ++m_ref_count;
}

void 
ActionHandle::remove_ref ()
{
  if (--m_ref_count == 0) {
    delete this;
  }
}

QAction *
ActionHandle::ptr () const
{
  return mp_action;
}

void 
ActionHandle::destroyed (QObject *obj)
{
  mp_action = 0;
}


// ---------------------------------------------------------------
//  Action implementation

Action::Action ()
{
  mp_handle = new ActionHandle (new QAction (& lay::MainWindow::instance ()));
  gtf::action_connect (mp_handle->ptr (), SIGNAL (triggered ()), this, SLOT (triggered_slot ()));
  mp_handle->add_ref ();
}

Action::Action (const std::string &title)
{
  mp_handle = new ActionHandle (AbstractMenu::create_action (title));
  gtf::action_connect (mp_handle->ptr (), SIGNAL (triggered ()), this, SLOT (triggered_slot ()));
  mp_handle->add_ref ();
}

Action::Action (QAction *other)
{
  mp_handle = new ActionHandle (other);
  gtf::action_connect (mp_handle->ptr (), SIGNAL (triggered ()), this, SLOT (triggered_slot ()));
  mp_handle->add_ref ();
}

Action::Action (const Action &action)
{
  mp_handle = action.mp_handle;
  if (mp_handle) {
    gtf::action_connect (mp_handle->ptr (), SIGNAL (triggered ()), this, SLOT (triggered_slot ()));
    mp_handle->add_ref ();
  }
}

Action &
Action::operator= (const Action &action)
{
  if (this != &action) {
    if (mp_handle) {
      if (mp_handle->ptr ()) {
        gtf::action_disconnect (mp_handle->ptr (), SIGNAL (triggered ()), this, SLOT (triggered_slot ()));
      }
      mp_handle->remove_ref ();
      mp_handle = 0;
    }
    mp_handle = action.mp_handle;
    if (mp_handle) {
      gtf::action_connect (mp_handle->ptr (), SIGNAL (triggered ()), this, SLOT (triggered_slot ()));
      mp_handle->add_ref ();
    }
  }
  return *this;
}

Action::~Action ()
{
  if (mp_handle) {
    if (mp_handle->ptr ()) {
      gtf::action_disconnect (mp_handle->ptr (), SIGNAL (triggered ()), this, SLOT (triggered_slot ()));
    }
    mp_handle->remove_ref ();
    mp_handle = 0;
  }
}

void
Action::trigger ()
{
  qaction ()->trigger ();
}

void
Action::triggered ()
{
  //  .. no action yet, the reimplementation must provide some ..
}

void
Action::triggered_slot ()
{
  BEGIN_PROTECTED
  triggered ();
  END_PROTECTED
}

void 
Action::set_title (const std::string &t)
{
  qaction ()->setText (QString (t.c_str ()));
}

std::string 
Action::get_title () const
{
  return std::string (qaction ()->text ().ascii ());
}

void 
Action::set_shortcut (const std::string &s)
{
  qaction ()->setShortcut (QKeySequence (QString (s.c_str ())));
}

std::string 
Action::get_shortcut () const
{
  return std::string (qaction ()->shortcut ().toString ().ascii ());
}

QAction *
Action::qaction () const
{
  tl_assert (mp_handle);
  tl_assert (mp_handle->ptr ());
  return mp_handle->ptr ();
}

bool
Action::is_checkable () const
{
  return qaction ()->isCheckable ();
}

bool
Action::is_checked () const
{
  return qaction ()->isChecked ();
}

bool
Action::is_enabled () const
{
  return qaction ()->isEnabled ();
}

bool
Action::is_visible () const
{
  return qaction ()->isVisible ();
}

void
Action::set_enabled (bool b)
{
  return qaction ()->setEnabled (b);
}

void
Action::set_visible (bool v)
{
  return qaction ()->setVisible (v);
}

void
Action::set_checked (bool c)
{
  return qaction ()->setChecked (c);
}

void
Action::set_checkable (bool c)
{
  return qaction ()->setCheckable (c);
}

void 
Action::set_icon (const std::string &filename)
{
  if (filename.empty ()) {
    qaction ()->setIcon (QIcon ());
  } else {
    qaction ()->setIcon (QIcon (QString (filename.c_str ())));
  }
}

std::string 
Action::get_icon_text () const
{
  return qaction ()->iconText ().ascii ();
}

void 
Action::set_icon_text (const std::string &icon_text)
{
  if (icon_text.empty ()) {
    qaction ()->setIconText (QString ());
  } else {
    qaction ()->setIconText (QString (icon_text.c_str ()));
  }
}

// ---------------------------------------------------------------
//  ConfigureAction implementation

ConfigureAction::ConfigureAction ()
  : Action ()
{
  //  .. nothing yet ..
}

ConfigureAction::ConfigureAction (const ConfigureAction &action)
  : Action (action), m_cname (action.m_cname), m_cvalue (action.m_cvalue)
{
  //  .. nothing yet ..
} 

ConfigureAction::ConfigureAction (const std::string &cname, const std::string &cvalue)
  : Action (), m_cname (cname), m_cvalue (cvalue)
{
  //  .. nothing yet ..
} 

ConfigureAction::ConfigureAction (const std::string &title, const std::string &cname, const std::string &cvalue)
  : Action (title), m_cname (cname), m_cvalue (cvalue)
{
  //  .. nothing yet ..
} 

ConfigureAction &
ConfigureAction::operator= (const ConfigureAction &action)
{
  if (&action != this) {
    Action::operator= (action);
    m_cname = action.m_cname;
    m_cvalue = action.m_cvalue;
  }
  return *this;
}

void 
ConfigureAction::triggered ()
{
  MainWindow::instance ().config_set (m_cname, m_cvalue);
  MainWindow::instance ().config_end ();
}

// ---------------------------------------------------------------
//  AbstractMenu implementation

QAction *
AbstractMenu::create_action (const std::string &s)
{
  QString title;
  QString shortcut;
  QString res;

  bool has_shortcut = false;
  bool has_icon = false;

  const char *p = s.c_str ();
  while (*p && *p != '(' && *p != '<') {
    title += *p++;
  }

  if (*p == '(') {
    has_shortcut = true;
    ++p;
    while (*p && *p != ')') {
      shortcut += *p++;
    }
    if (*p == ')') {
      ++p;
    }
    while (*p && isspace (*p)) {
      ++p;
    }
  }

  if (*p == '<') {
    has_icon = true;
    ++p;
    while (*p && *p != '>') {
      res += *p++;
    }
  }
  
  QAction *a;
  if (has_icon) {
    a = new QAction (QIcon (res), title, &lay::MainWindow::instance ());
  } else {
    a = new QAction (title, &lay::MainWindow::instance ());
  }

  if (has_shortcut) {
    a->setShortcut (QKeySequence (shortcut));
  }

  return a;
}

AbstractMenu::AbstractMenu (MainWindow *mw)
  : mp_mw (mw)
{
  //  .. nothing else ..
}

AbstractMenu::~AbstractMenu ()
{
  reset_menu_objects (m_root);
}

void 
AbstractMenu::init (const MenuLayoutEntry *layout)
{
  m_root.set_has_submenu ();
  transfer (layout, m_root);
}

void 
AbstractMenu::reset_menu_objects (AbstractMenuItem &item)
{
  for (std::list<AbstractMenuItem>::iterator c = item.children.begin (); c != item.children.end (); ++c) {
    reset_menu_objects (*c);
  }
  item.set_menu (0);
}

void 
AbstractMenu::build (QMenuBar *mbar, QToolBar *tbar)
{
  m_helper_menu_items.clear ();
  mbar->clear ();
  tbar->clear ();

  for (std::list<AbstractMenuItem>::iterator c = m_root.children.begin (); c != m_root.children.end (); ++c) {

    if (c->has_submenu ()) {

      if (c->name () == "@toolbar") {
        build (tbar, c->children);
      } else if (c->name () [0] == '@') {

        if (c->menu () == 0) {
          c->set_menu (new QMenu (c->action ().get_title ().c_str (), mp_mw));
          c->set_action (Action (c->menu ()->menuAction ()));
        }

        build (c->menu (), c->children);

      } else {

        if (c->menu () == 0) {
          c->set_menu (mbar->addMenu (c->action ().get_title ().c_str ()));
          c->set_action (Action (c->menu ()->menuAction ()));
        } else {
          mbar->addMenu (c->menu ());
        }

        build (c->menu (), c->children);

      }

    } else {
      mbar->addAction (c->action ().qaction ());
    }

  }
}

void 
AbstractMenu::build (QMenu *m, std::list<AbstractMenuItem> &items)
{
  m->clear (); 

  for (std::list<AbstractMenuItem>::iterator c = items.begin (); c != items.end (); ++c) {
    if (c->has_submenu ()) {
      //  HINT: the action acts as a container for the title. Unfortunately, we cannot create a 
      //  menu with a given action. The action is provided by addMenu instead.
      QMenu *menu = m->addMenu (c->action ().get_title ().c_str ());
      c->set_action (Action (menu->menuAction ()));
      c->set_menu (menu);
      build (menu, c->children);
    } else {
      m->addAction (c->action ().qaction ());
    }
  }
}

void 
AbstractMenu::build (QToolBar *t, std::list<AbstractMenuItem> &items)
{
  for (std::list<AbstractMenuItem>::iterator c = items.begin (); c != items.end (); ++c) {

    if (! c->children.empty ()) {

      //  To support tool buttons with menu we have to attach a helper menu 
      //  item to the QAction object. 
      //  TODO: this hurts if we use this QAction otherwise. In this case, this
      //  QAction would get a menu too. However, hopefully this usage is constrained
      //  to special toobar buttons only.
      //  In order to be able to manage the QMenu ourselves, we must not give it a parent.
      QMenu *menu = new QMenu (0);
      m_helper_menu_items.push_back (menu); // will be owned by the stable vector 
      c->action ().qaction ()->setMenu (menu);
      if (! c->has_submenu ()) {
        t->addAction (c->action ().qaction ());
      }
      build (menu, c->children);

    } else {
      t->addAction (c->action ().qaction ());
    }

  }

}

QMenu *
AbstractMenu::detached_menu (const std::string &name)
{
  AbstractMenuItem *item = find_item_exact ("@" + name);
  tl_assert (item != 0);
  return item->menu ();
}

QMenu *
AbstractMenu::menu (const std::string &path)
{
  AbstractMenuItem *item = find_item_exact (path);
  if (item) {
    return item->menu ();
  } else {
    return 0;
  }
}

bool
AbstractMenu::is_valid (const std::string &path)
{
  AbstractMenuItem *item = find_item_exact (path);
  return item != 0;
}

bool
AbstractMenu::is_menu (const std::string &path)
{
  AbstractMenuItem *item = find_item_exact (path);
  return item != 0 && item->has_submenu ();
}

Action 
AbstractMenu::action (const std::string &path)
{
  AbstractMenuItem *item = find_item_exact (path);
  if (! item) {
    throw tl::Exception (tl::translate ("Not a valid menu item path: ") + path);
  }
  return item->action ();
}

std::vector<std::string> 
AbstractMenu::items (const std::string &path)
{
  std::vector<std::string> res;

  AbstractMenuItem *item = find_item_exact (path);
  if (item) {
    res.reserve (item->children.size ());
    for (std::list <AbstractMenuItem>::iterator c = item->children.begin (); c != item->children.end (); ++c) {
      res.push_back (c->name ());
    }
  }

  return res;
}

void 
AbstractMenu::insert_item (const std::string &path, const std::string &name, const Action &action)
{
  std::pair<AbstractMenuItem *, std::list<AbstractMenuItem>::iterator > ins = find_item (path);
  if (ins.first) {
    ins.first->children.insert (ins.second, AbstractMenuItem ());
    --ins.second;
    ins.second->setup_item (ins.first->name (), name, action);
  }
  emit changed ();
}

void 
AbstractMenu::insert_separator (const std::string &path, const std::string &name)
{
  std::pair<AbstractMenuItem *, std::list<AbstractMenuItem>::iterator > ins = find_item (path);
  if (ins.first) {
    ins.first->children.insert (ins.second, AbstractMenuItem ());
    --ins.second;
    QAction *action = new QAction (mp_mw);
    action->setSeparator (true);
    ins.second->setup_item (ins.first->name (), name, Action (action));
  }
  emit changed ();
}

void 
AbstractMenu::insert_menu (const std::string &path, const std::string &name, const Action &action)
{
  std::pair<AbstractMenuItem *, std::list<AbstractMenuItem>::iterator > ins = find_item (path);
  if (ins.first) {
    ins.first->children.insert (ins.second, AbstractMenuItem ());
    --ins.second;
    ins.second->setup_item (ins.first->name (), name, action);
    ins.second->set_has_submenu ();
  }
  emit changed ();
}

void 
AbstractMenu::insert_menu (const std::string &path, const std::string &name, const std::string &title)
{
  insert_menu (path, name, Action (create_action (title)));
}

void 
AbstractMenu::delete_item (const std::string &path)
{
  std::pair<AbstractMenuItem *, std::list<AbstractMenuItem>::iterator > ins = find_item (path);
  if (ins.first) {
    if (ins.second == ins.first->children.end ()) {
      throw tl::Exception (tl::translate ("delete_item cannot delete past-the-end item"));
    }
    reset_menu_objects (*ins.second);
    ins.first->children.erase (ins.second);
  }
  emit changed ();
}

AbstractMenuItem *
AbstractMenu::find_item_exact (const std::string &path)
{
  tl::Extractor extr (path.c_str ());
  AbstractMenuItem *item = &m_root;

  while (! extr.at_end ()) {

    if (extr.test ("#")) {

      unsigned int n = 0;
      extr.try_read (n);

      std::list<AbstractMenuItem>::iterator c = item->children.begin ();
      ++n;
      while (--n > 0 && c != item->children.end ()) {
        ++c;
      }
      if (n > 0) {
        return 0;
      } 

      item = &*c;

    } else {

      std::string n;
      extr.read (n, ".");
      std::string name (item->name ());
      if (! name.empty ()) {
        name += ".";
      }
      name += n;

      AbstractMenuItem *p = item;

      item = 0;
      for (std::list<AbstractMenuItem>::iterator c = p->children.begin (); c != p->children.end () && item == 0; ++c) {
        if (c->name () == name) {
          item = &*c;
        }
      }

      if (! item) {
        return 0;
      }

    }

    extr.test (".");

  }

  return item;
}

std::pair<AbstractMenuItem *, std::list<AbstractMenuItem>::iterator> 
AbstractMenu::find_item (const std::string &path)
{
  tl::Extractor extr (path.c_str ());
  AbstractMenuItem *parent = &m_root;
  std::list<AbstractMenuItem>::iterator iter = m_root.children.end ();

  while (true) {

    if (extr.test ("#")) {

      unsigned int n = 0;
      extr.try_read (n);
      iter = parent->children.begin ();
      ++n;
      while (--n > 0 && iter != parent->children.end ()) {
        ++iter;
      }
      if (n > 0) {
        return std::make_pair ((AbstractMenuItem *) 0, m_root.children.end ());
      }

    } else if (extr.test ("begin")) {
      return std::make_pair (parent, parent->children.begin ());
    } else if (extr.test ("end")) {
      return std::make_pair (parent, parent->children.end ());
    } else {

      std::string n;
      extr.read (n, ".");
      std::string name (parent->name ());
      if (! name.empty ()) {
        name += ".";
      }
      name += n;

      AbstractMenuItem *p = parent;
      parent = 0;
      for (std::list<AbstractMenuItem>::iterator c = p->children.begin (); c != p->children.end (); ++c) {
        if (c->name () == name) {
          parent = p;
          iter = c;
          break;
        }
      }

      if (! parent) {
        return std::make_pair ((AbstractMenuItem *) 0, m_root.children.end ());
      }

    }

    if (extr.at_end ()) {
      return std::make_pair (parent, iter);
    }

    extr.test (".");

    parent = &*iter;

  }

}

void 
AbstractMenu::transfer (const MenuLayoutEntry *layout, AbstractMenuItem &item)
{
  while (layout->name) {

    item.children.push_back (AbstractMenuItem ());

    QAction *a = new QAction (mp_mw);
    item.children.back ().setup_item (item.name (), layout->name, Action (a));

    if (! layout->title.empty ()) {

      item.children.back ().action ().qaction ()->setSeparator (false);
      item.children.back ().set_action_title (layout->title);

      if (! layout->key.isEmpty ()) {
        item.children.back ().action ().qaction ()->setShortcut (layout->key);
      }
      if (layout->slot) {
        gtf::action_connect (item.children.back ().action ().qaction (), SIGNAL (triggered ()), mp_mw, layout->slot);
      }

    } else {
      item.children.back ().action ().qaction ()->setSeparator (true);
    }

    if (layout->submenu) {
      item.children.back ().set_has_submenu ();
      transfer (layout->submenu, item.children.back ());
    }

    ++layout;

  }
}

std::vector<std::string> 
AbstractMenu::group (const std::string &name) const
{
  std::vector<std::string> grp;
  collect_group (grp, name, m_root);
  return grp;
}

void 
AbstractMenu::collect_group (std::vector<std::string> &grp, const std::string &name, const AbstractMenuItem &item) const
{
  for (std::list<AbstractMenuItem>::const_iterator c = item.children.begin (); c != item.children.end (); ++c) {
    if (c->groups ().find (name) != c->groups ().end ()) {
      grp.push_back (c->name ());
    }
    collect_group (grp, name, *c);
  }
}

}

