
/*

  KLayout Layout Viewer
  Copyright (C) 2006-2007 Matthias Koefferlein

  This program is free software; you can redistribute it and/or modify
  it under the terms of the GNU General Public License as published by
  the Free Software Foundation; either version 2 of the License, or
  (at your option) any later version.

  This program is distributed in the hope that it will be useful,
  but WITHOUT ANY WARRANTY; without even the implied warranty of
  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
  GNU General Public License for more details.

  You should have received a copy of the GNU General Public License
  along with this program; if not, write to the Free Software
  Foundation, Inc., 51 Franklin St, Fifth Floor, Boston, MA  02110-1301  USA

*/


#include "layApplication.h"
#include "layLayoutViewConfig.h"
#include "layMainWindow.h"
#include "layVersion.h"
#include "gtf.h"
#include "dbStatic.h"
#include "tlException.h"
#include "tlAssert.h"
#include "tlLog.h"
#include "tlString.h"
#include "tlException.h"
#include "tlInternational.h"

#include <QtGui/QIcon>
#include <QtCore/QDir>
#include <QtGui/QAction>

#ifdef _WIN32
#include <windows.h>
#else
#include <dlfcn.h>
#endif

#include <iostream>

namespace lay
{

static Application *ms_instance = 0;

Application::Application (int &argc, char *argv [])
  : m_write_config_file (true),
    m_gtf_replay_rate (0),
    m_gtf_replay_stop (-1),
    m_same_view (false),
    m_sync_mode (false),
    m_no_gui (false),
    m_editable (false),
    m_editable_set (false),
    m_qapp (argc, argv),
    mp_mw (0),
    mp_recorder (0)
{
  tl_assert (ms_instance == 0);
  ms_instance = this;

  std::string gtf_record;
  bool gtf_save_incremental = false;

  //  locate the layviewrc file in the home directory
  m_config_file = QDir::home ().absoluteFilePath (".layviewrc").ascii ();

  for (int i = 1; i < argc; ++i) {

    std::string a = argv [i];

    if (a == "-d" && (i + 1) < argc) {

      int v = std::max (0, atoi (argv [++i]));
      tl::verbosity (v);

    } else if (a == "-l" && (i + 1) < argc) {

      m_layer_props_file = argv [++i];

    } else if (m_ruby_interpreter.ruby_available () && a == "-rd" && (i + 1) < argc) {

      std::string v;
      const char *p = argv [++i];
      const char *n0 = p;
      while (*p && *p != '=') ++p;
      std::string n (n0, 0, p - n0);
      m_variables.push_back (std::make_pair (n, v));
      if (*p == '=') {
        m_variables.back ().second = ++p;
      }

    } else if (m_ruby_interpreter.ruby_available () && a == "-rm" && (i + 1) < argc) {

      m_ruby_modules.push_back (argv [++i]);

    } else if (m_ruby_interpreter.ruby_available () && a == "-r" && (i + 1) < argc) {

      m_ruby_script = argv [++i];

    } else if (a == "-gr" && (i + 1) < argc) {

      gtf_record = argv [++i];

    } else if (a == "-gi") {

      gtf_save_incremental = true;

    } else if (a == "-gp" && (i + 1) < argc) {

      m_gtf_replay = argv [++i];
      //  test mode replay forces sync mode to true
      m_sync_mode = true;

    } else if (a == "-gx" && (i + 1) < argc) {

      m_gtf_replay_rate = std::max (0, atoi (argv [++i]));

    } else if (a == "-gb" && (i + 1) < argc) {

      m_gtf_replay_stop = std::max (0, atoi (argv [++i]));

    } else if (a == "-c" && (i + 1) < argc) {

      m_config_file = argv [++i];

    } else if (a == "-p" && (i + 1) < argc) {

#ifdef _WIN32
      //  there is no "dlopen" on mingw, so we need to emulate it.
      HINSTANCE handle = LoadLibrary (argv[++i]);
      if (! handle) {
        throw tl::Exception (tl::translate ("Unable to load plugin: %s with error message: %s ").c_str (), argv[i], GetLastError ());
      }
#else
      void *handle;
      handle = dlopen (argv[++i], RTLD_LAZY);
      if (! handle) {
        throw tl::Exception (tl::translate ("Unable to load plugin: %s").c_str (), argv[i]);
      }
#endif

      if (tl::verbosity () >= 10) {
        tl::info << "Loaded plugin '" << argv[i] << "'";
      }

    } else if (a == "-s") {

      m_same_view = true;
      
    } else if (a == "-e") {

      m_editable = true;
      m_editable_set = true;
      
    } else if (a == "-ne") {

      m_editable = false;
      m_editable_set = true;
      
    } else if (a == "-t") {

      m_write_config_file = false;
      
    } else if (a == "-z") {

      m_no_gui = true;
      
    } else if (a == "-x") {

      m_sync_mode = true;
      
    } else if (a == "-v") {

      tl::info << lay::Version::name () << " " << lay::Version::version ();
      return;
      
    } else if (a == "-h") {

      tl::info << usage () << tl::noendl;
      return;

    } else if (a[0] == '-') {

      throw tl::Exception (tl::translate ("Unknown option: ") + a);

    } else {
      m_files.push_back (a);
    }

  }

  //  If the editable flag was not set, use it from the 
  //  configuration. Since it is too early now, we cannot use the
  //  configuration once it is read
  if (! m_editable_set) {
    lay::MainWindow cfg; // TODO: should be PluginRoot or something similar, not MainWindow!!!
    cfg.read_config (m_config_file);
    try {
      cfg.config_get (cfg_edit_mode, m_editable);
    } catch (...) { }
  }

  db::set_editable_mode (m_editable);

  m_qapp.setWindowIcon (QIcon (":logo.png"));

  if (! gtf_record.empty ()) {
    //  since the recorder tracks QAction connections etc., it must be instantiated before every other 
    //  object performing a gtf::action_connect for example
    mp_recorder = new gtf::Recorder (&m_qapp, gtf_record);
    mp_recorder->save_incremental (gtf_save_incremental);
  }

  mp_mw = new lay::MainWindow (&m_qapp, "main_window");
  m_qapp.setMainWidget (mp_mw);
}

Application::~Application () 
{
  if (mp_mw) {
    delete mp_mw;
    mp_mw = 0;
  }
  if (mp_recorder) {
    delete mp_recorder;
    mp_recorder = 0;
  }
  ms_instance = 0;
}

Application *
Application::instance () 
{
  return ms_instance;
}

std::string
Application::version () const
{
  return std::string (lay::Version::name ()) + " " + lay::Version::version ();
}
      
std::string 
Application::usage () const
{
  std::string r;
  r = std::string (lay::Version::exe_name ()) + " [<options>] [<file>] ..\n";
  r += tl::translate ("options") + "\n";
  r += tl::translate ("  -c <config file>   Use this configuration file") + "\n";
  r += tl::translate ("  -d <debug level>   Set debug level") + "\n";
  r += tl::translate ("  -e                 Editable mode (allow editing of files)") + "\n";
  r += tl::translate ("  -ne                Readonly mode (editing of files is disabled)") + "\n";
  r += tl::translate ("  -gr <file name>    Record GUI test file") + "\n";
  r += tl::translate ("  -gp <file name>    Replay GUI test file") + "\n";
  r += tl::translate ("  -gb <line number>  Replay GUI test file up to (including) line") + "\n";
  r += tl::translate ("  -gx <millisec>     Replay rate for GUI test file") + "\n";
  r += tl::translate ("  -gi                Incremental logs for GUI test file") + "\n";
  r += tl::translate ("  -l <layer file>    Use layer definition file") + "\n";
  r += tl::translate ("  -p <plugin>        Load the plugin (can be used multiple times)") + "\n";
  if (m_ruby_interpreter.ruby_available ()) {
    r += tl::translate ("  -r <script>        Execute ruby main script on startup (after having loaded files etc.)") + "\n";
    r += tl::translate ("  -rm <script>       Execute ruby module on startup (can be used multiple times)") + "\n";
    r += tl::translate ("  -rd <name>=<value> Specify ruby variable") + "\n";
  }
  r += tl::translate ("  -s                 Load files into same view") + "\n";
  r += tl::translate ("  -t                 Don't update the configuration file on exit") + "\n";
  r += tl::translate ("  -v                 Print program version and exit") + "\n";
  r += tl::translate ("  -x                 Synchroneous redraw mode") + "\n";
  r += tl::translate ("  -z                 Non-GUI mode") + "\n";
  return r;
}

int
Application::run ()
{
  if (! mp_mw) {
    return 0;
  }

  mp_mw->set_synchroneous (m_sync_mode);

  if (! m_no_gui) {
    mp_mw->setCaption (version ().c_str ());
    mp_mw->resize (800, 600);
    mp_mw->show ();
  }

  int result = 0;

  gtf::Player player (&m_qapp);
  if (! m_gtf_replay.empty ()) {
    player.load (m_gtf_replay);
  }

  if (mp_recorder) {
    mp_recorder->start ();
  }

  mp_mw->read_config (m_config_file);

  if (m_ruby_interpreter.ruby_available ()) {
    for (std::vector< std::pair<std::string, std::string> >::const_iterator v = m_variables.begin (); v != m_variables.end (); ++v) {
      m_ruby_interpreter.define_variable (v->first, v->second);
    }
    for (std::vector<std::string>::const_iterator m = m_ruby_modules.begin (); m != m_ruby_modules.end (); ++m) {
      m_ruby_interpreter.load_file (m->c_str ());
    }
  }

  for (std::vector <std::string>::const_iterator f = m_files.begin (); f != m_files.end (); ++f) {
    mp_mw->load_layout (*f, m_same_view ? 2 /*same view*/ : 1 /*new view*/);
  }

  if (! m_layer_props_file.empty ()) {
    mp_mw->load_layer_properties (m_layer_props_file, true /*all views*/);
    //  because the layer may carry transformations, we need to refit the cellviews.
    for (unsigned int v = 0; v != mp_mw->views (); ++v) {
      mp_mw->view (v)->zoom_fit ();
    }
  }

  if (! m_gtf_replay.empty ()) {
    player.replay (m_gtf_replay_rate, m_gtf_replay_stop);
  }

  if (! m_ruby_script.empty () && m_ruby_interpreter.ruby_available ()) {
    m_ruby_interpreter.load_file (m_ruby_script.c_str ());
    //  TODO: allow the ruby interpreter to return some value 
    result = 0;
  } else {
    result = exec ();
  }

  //  save the recorded test events
  if (mp_recorder && mp_recorder->recording ()) {
    mp_recorder->stop ();
    mp_recorder->save ();
  }

  if (m_write_config_file) {
    mp_mw->write_config (m_config_file); 
  }

  return result;
}

static void 
dump_children (QObject *obj, int level = 0)
{
  QObjectList children = obj->children ();
  if (! children.isEmpty () || ! obj->objectName ().isEmpty ()) {
    std::string info;
    for (int i = 0; i < level; ++i) {
      info += "  ";
    }
    if (obj->objectName ().isEmpty ()) {
      info += "<unnamed>";
    } else {
      info += obj->objectName ().ascii ();
    }
    tl::info << info;
    for (QObjectList::const_iterator child = children.begin (); child != children.end (); ++child) {
      dump_children (*child, level + 1);
    }
  }
}

int 
Application::exec ()
{
  if (m_no_gui) {
    return 0;
  } else {

    //  if requested, dump the 
    if (tl::verbosity () >= 30) {

      QWidgetList tl_widgets = QApplication::topLevelWidgets (); 

      tl::info << tl::translate ("Widget tree:");
      for (QWidgetList::const_iterator tl = tl_widgets.begin (); tl != tl_widgets.end (); ++tl) { 
        if (! (*tl)->objectName ().isEmpty ()) {
          dump_children (*tl); 
        }
      }
      tl::info << "";

      tl::info << tl::translate ("Actions list:");
      for (QWidgetList::const_iterator tl = tl_widgets.begin (); tl != tl_widgets.end (); ++tl) { 
        if (! (*tl)->objectName ().isEmpty ()) {
          QList<QAction *> actions = (*tl)->findChildren<QAction *> ();
          if (! actions.isEmpty ()) {
            tl::info << (*tl)->objectName ().ascii () << ":";
            for (QList<QAction *>::const_iterator a = actions.begin (); a != actions.end (); ++a) {
              if (! (*a)->objectName ().isEmpty ()) {
                tl::info << "  " << (*a)->objectName ().ascii ();
              }
            }
          }
        }
      }
      tl::info << "";

    }

    return m_qapp.exec ();

  }
}

void
Application::process_events ()
{
  if (! m_no_gui) {
    QApplication::processEvents ();
  }
}

bool 
Application::write_config (const std::string &config_file)
{
  tl_assert (mp_mw != 0);
  return mp_mw->write_config (config_file);
}

void 
Application::clear_config ()
{
  tl_assert (mp_mw != 0);
  return mp_mw->clear_config ();
}

bool 
Application::read_config (const std::string &config_file)
{
  tl_assert (mp_mw != 0);
  return mp_mw->read_config (config_file);
}

void 
Application::set_config (const std::string &name, const std::string &value)
{
  tl_assert (mp_mw != 0);
  mp_mw->config_set (name, value);
}

std::string 
Application::get_config (const std::string &name) const
{
  tl_assert (mp_mw != 0);
  return mp_mw->config_get (name);
}

std::vector<std::string> 
Application::get_config_names () const
{
  tl_assert (mp_mw != 0);
  std::vector<std::string> names;
  mp_mw->get_config_names (names);
  return names;
}

}

