
/*

  KLayout Layout Viewer
  Copyright (C) 2006-2007 Matthias Koefferlein

  This program is free software; you can redistribute it and/or modify
  it under the terms of the GNU General Public License as published by
  the Free Software Foundation; either version 2 of the License, or
  (at your option) any later version.

  This program is distributed in the hope that it will be useful,
  but WITHOUT ANY WARRANTY; without even the implied warranty of
  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
  GNU General Public License for more details.

  You should have received a copy of the GNU General Public License
  along with this program; if not, write to the Free Software
  Foundation, Inc., 51 Franklin St, Fifth Floor, Boston, MA  02110-1301  USA

*/


#include "BrowserDialog.h"
#include "layBrowserPanel.h"
#include "layBrowserDialog.h"
#include "layMainWindow.h"
#include "layExceptions.h"

namespace lay 
{

BrowserDialog::BrowserDialog ()
  //  like all the other dialogs, we do not specify a parent, so the non-modal window will no stay in 
  //  front of the main window (but: then any QInputDialog opened in the loadUrl method will push
  //  the browser dialog to the back since it refers to the main window ..)
  : QDialog (0, "html_browser"),
    m_back_dm (this, &BrowserDialog::back)
{
  init ();
}

void 
BrowserDialog::init ()
{
  m_enable_load = true;

  mp_source = 0;

  Ui::BrowserDialog::setupUi (this);

  browser->setReadOnly (true);
  browser->set_handler (this);

  connect (back_pb, SIGNAL (clicked ()), this, SLOT (back ()));
  connect (forward_pb, SIGNAL (clicked ()), this, SLOT (forward ()));
  connect (home_pb, SIGNAL (clicked ()), this, SLOT (home ()));

  show ();
}

BrowserDialog::~BrowserDialog ()
{
  browser->set_handler (0);
  set_source (0);
}

void 
BrowserDialog::load (const std::string &s)
{
  browser->setSource (QUrl (s.c_str ()));
}

void 
BrowserDialog::set_source (BrowserSource *source)
{
  if (mp_source) {
    BrowserSource *s = mp_source;
    mp_source = 0;
    s->attach (0);
  }
  mp_source = source;
  if (mp_source) {
    mp_source->attach (this);
  }
}

void 
BrowserDialog::set_home (const std::string &url)
{
  m_home = url;
  home ();
}

void 
BrowserDialog::reload ()
{
  //  clear caches to force a reload
  m_cached_url = "";
  m_cached_text = "";

  //  disable reload while we are in a loadResource call - clearing the cache will be sufficient
  if (m_enable_load) {
    browser->reload ();
  }
}

void 
BrowserDialog::back ()
{
  browser->backward ();
}

void 
BrowserDialog::forward ()
{
  browser->forward ();
}

void 
BrowserDialog::home ()
{
  browser->setSource (QUrl (m_home.c_str ()));
}

QVariant 
BrowserDialog::loadResource (int type, const QUrl &url)
{
  QVariant ret;

  //  recursion sentinel: avoid recursion by any action within mp_source->get that causes a "loadResource"
  if (! m_enable_load) {
    //  return any dummy in this case - otherwise the QTestBrowser complains about not having anything.
    return QVariant (" ");
  }
  m_enable_load = false;

  //  Qt sets the override cursor in response to link clicks - this is not appropriate for some
  //  GSI callback implementations that show InputDialogs for example. Therefore we install out own
  //  (normal) override cursor.
  QApplication::setOverrideCursor (QCursor (Qt::ArrowCursor));

  BEGIN_PROTECTED
  if (mp_source) {
    std::string u = url.toString ().ascii ();
    std::string s;
    if (u == m_cached_url) {
      s = m_cached_text;
    } else {
      s = mp_source->get (u);
    }
    if (s.empty ()) {
      s = " "; // QTextBrowser needs at least something
      //  The only way (as far as I know in Qt <4.2) to suppress navigation to 
      //  the Url is to schedule a delayed "back" signal. In Qt >= 4.2 we could register
      //  an external handler for "int" schemes that would do nothing ..
      m_back_dm ();
    } else {
      //  to avoid regeneration of text on artificial "back" events, the last page is cached
      m_cached_text = s;
      m_cached_url = u;
    }
    ret = QVariant (s.c_str ());
  } 
  END_PROTECTED

  QApplication::restoreOverrideCursor ();

  m_enable_load = true;
  return ret;
}

void
BrowserDialog::accept ()
{
  BEGIN_PROTECTED
  closed ();
  END_PROTECTED
  QDialog::accept ();
}

// -------------------------------------------------------------

QVariant 
BrowserPanel::loadResource (int type, const QUrl &url)
{
  if (mp_handler && type == QTextDocument::HtmlResource && url.scheme () == "int") {
    return mp_handler->loadResource (type, url);
  } else {
    return QTextBrowser::loadResource (type, url);
  }
}

// -------------------------------------------------------------

BrowserSource::BrowserSource ()
  : mp_dialog (0)
{
  //  .. nothing yet ..
}

BrowserSource::BrowserSource (const std::string &html)
  : mp_dialog (0), m_default_html (html)
{
  //  .. nothing yet ..
}

BrowserSource::~BrowserSource ()
{
  attach (0);
}

std::string 
BrowserSource::get (const std::string &url) 
{
  return m_default_html;
}

void 
BrowserSource::attach (lay::BrowserDialog *d)
{
  if (mp_dialog != d && mp_dialog) {
    mp_dialog->set_source (0);
  }
  mp_dialog = d;
}

}

