
/*

  KLayout Layout Viewer
  Copyright (C) 2006-2007 Matthias Koefferlein

  This program is free software; you can redistribute it and/or modify
  it under the terms of the GNU General Public License as published by
  the Free Software Foundation; either version 2 of the License, or
  (at your option) any later version.

  This program is distributed in the hope that it will be useful,
  but WITHOUT ANY WARRANTY; without even the implied warranty of
  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
  GNU General Public License for more details.

  You should have received a copy of the GNU General Public License
  along with this program; if not, write to the Free Software
  Foundation, Inc., 51 Franklin St, Fifth Floor, Boston, MA  02110-1301  USA

*/


#include "layDialogs.h"
#include "tlString.h"
#include "tlException.h"
#include "tlXMLParser.h"
#include "layExceptions.h"
#include "layLayerProperties.h"
#include "layFileDialog.h"


namespace lay
{

// --------------------------------------------------------------------------------
//  NewLayoutPropertiesDialog implementation

NewLayoutPropertiesDialog::NewLayoutPropertiesDialog (QWidget *parent)
  : QDialog (parent, "new_layout_properties_dialog")
{
  Ui::NewLayoutPropertiesDialog::setupUi (this);
}

NewLayoutPropertiesDialog::~NewLayoutPropertiesDialog ()
{
  //  .. nothing yet ..
}

bool 
NewLayoutPropertiesDialog::exec (std::string &cell_name, double &dbu, double &size)
{
  window_le->setText (tl::to_string (size).c_str ());
  dbu_le->setText (tl::to_string (dbu).c_str ());
  topcell_le->setText (cell_name.c_str ());

  if (QDialog::exec ()) {

    tl::from_string (window_le->text ().ascii (), size);
    tl::from_string (dbu_le->text ().ascii (), dbu);
    cell_name = topcell_le->text ().ascii ();
    return true;

  } else {
    return false;
  }
}

void 
NewLayoutPropertiesDialog::accept ()
{
BEGIN_PROTECTED;

  double x = 0.0;
  tl::from_string (window_le->text ().ascii (), x);
  tl::from_string (dbu_le->text ().ascii (), x);

  if (topcell_le->text ().isEmpty ()) {
    throw tl::Exception (tl::translate ("The topcell must be specified"));
  }
  
  QDialog::accept ();

END_PROTECTED;
}

// --------------------------------------------------------------------------------
//  NewCellPropertiesDialog implementation

NewCellPropertiesDialog::NewCellPropertiesDialog (QWidget *parent)
  : QDialog (parent, "new_cell_properties_dialog"),
    mp_layout (0)
{
  Ui::NewCellPropertiesDialog::setupUi (this);
}

NewCellPropertiesDialog::~NewCellPropertiesDialog ()
{
  //  .. nothing yet ..
}

bool 
NewCellPropertiesDialog::exec (const db::Layout *layout, std::string &cell_name, double &size)
{
  mp_layout = layout;

  name_le->setText (cell_name.c_str ());
  window_le->setText (tl::to_string (size).c_str ());

  if (QDialog::exec ()) {

    tl::from_string (window_le->text ().ascii (), size);
    cell_name = name_le->text ().ascii ();
    return true;

  } else {
    return false;
  }
}

void 
NewCellPropertiesDialog::accept ()
{
BEGIN_PROTECTED;

  double x = 0.0;
  tl::from_string (window_le->text ().ascii (), x);

  if (mp_layout->cell_by_name (name_le->text ().ascii ()).first) {
    throw tl::Exception (tl::translate ("A cell with that name already exists: %s").c_str (), name_le->text ().ascii ());
  }
  
  QDialog::accept ();

END_PROTECTED;
}

// --------------------------------------------------------------------------------
//  NewLayerPropertiesDialog implementation

NewLayerPropertiesDialog::NewLayerPropertiesDialog (QWidget *parent)
  : QDialog (parent, "new_layer_properties_dialog")
{
  Ui::NewLayerPropertiesDialog::setupUi (this);
}

NewLayerPropertiesDialog::~NewLayerPropertiesDialog ()
{
  //  .. nothing yet ..
}

bool 
NewLayerPropertiesDialog::exec (db::LayerProperties &src)
{
  if (src.layer >= 0) {
    layer_le->setText (tl::to_string (src.layer).c_str ());
  } else {
    layer_le->setText ("");
  }
  if (src.datatype >= 0) {
    datatype_le->setText (tl::to_string (src.datatype).c_str ());
  } else {
    datatype_le->setText ("");
  }
  name_le->setText (src.name.c_str ());

  if (QDialog::exec ()) {

    if (layer_le->text () != "") {
      int l = 0;
      tl::from_string (layer_le->text ().ascii (), l);
      src.layer = l;
    } else {
      src.layer = -1;
    }
    if (datatype_le->text () != "") {
      int d = 0;
      tl::from_string (datatype_le->text ().ascii (), d);
      src.datatype = d;
    } else {
      src.datatype = -1;
    }
    src.name = name_le->text ().ascii ();
    return true;

  } else {
    return false;
  }
}

// --------------------------------------------------------------------------------
//  MoveOptionsDialog implementation

MoveOptionsDialog::MoveOptionsDialog (QWidget *parent)
  : QDialog (parent, "move_options_dialog")
{
  Ui::MoveOptionsDialog::setupUi (this);
}

MoveOptionsDialog::~MoveOptionsDialog ()
{
  //  .. nothing yet ..
}

bool 
MoveOptionsDialog::exec (db::DPoint &disp)
{
  disp_x_le->setText (tl::to_string (disp.x ()).c_str ());
  disp_y_le->setText (tl::to_string (disp.y ()).c_str ());

  if (QDialog::exec ()) {

    double x = 0.0, y = 0.0;
    tl::from_string (disp_x_le->text ().ascii (), x);
    tl::from_string (disp_y_le->text ().ascii (), y);

    disp = db::DPoint (x, y);

    return true;

  } else {
    return false;
  }
}

void 
MoveOptionsDialog::accept ()
{
BEGIN_PROTECTED;
  double x = 0.0;
  tl::from_string (disp_x_le->text ().ascii (), x);
  tl::from_string (disp_y_le->text ().ascii (), x);
  QDialog::accept ();
END_PROTECTED;
}

// --------------------------------------------------------------------------------
//  RenameCellDialog implementation

RenameCellDialog::RenameCellDialog (QWidget *parent)
  : QDialog (parent, "rename_cell_dialog"), mp_layout (0)
{
  Ui::RenameCellDialog::setupUi (this);
}

RenameCellDialog::~RenameCellDialog ()
{
  //  .. nothing yet ..
}

void 
RenameCellDialog::accept ()
{
BEGIN_PROTECTED;
  if (name_le->text () == "") {
    throw tl::Exception (tl::translate ("A name must be given"));
  }
  if (mp_layout->cell_by_name (name_le->text ().ascii ()).first) {
    throw tl::Exception (tl::translate ("A cell with that name already exists"));
  }
  QDialog::accept ();
END_PROTECTED;
}

bool 
RenameCellDialog::exec (const db::Layout &layout, std::string &name)
{
  mp_layout = &layout;
  name_le->setText (name.c_str ());
  if (QDialog::exec ()) {
    name = name_le->text ().ascii ();
    return true;
  } else {
    return false;
  }
}

// --------------------------------------------------------------------------------
//  DeleteCellModeDialog implementation

DeleteCellModeDialog::DeleteCellModeDialog (QWidget *parent)
  : QDialog (parent, "delete_cell_mode_dialog")
{
  Ui::DeleteCellModeDialog::setupUi (this);
}

DeleteCellModeDialog::~DeleteCellModeDialog ()
{
  //  .. nothing yet ..
}

bool 
DeleteCellModeDialog::exec (int &delete_mode)
{
  QRadioButton *buttons [3] = { shallow_rb, deep_rb, full_rb };

  for (int i = 0; i < int (sizeof (buttons) / sizeof (buttons [0])); ++i) {
    buttons [i]->setChecked (delete_mode == i);
  }

  if (QDialog::exec ()) {
    for (int i = 0; i < int (sizeof (buttons) / sizeof (buttons [0])); ++i) {
      if (buttons [i]->isChecked ()) {
        delete_mode = i;
      }
    }
    return true;
  } else {
    return false;
  }
}

}

