
/*

  KLayout Layout Viewer
  Copyright (C) 2006-2007 Matthias Koefferlein

  This program is free software; you can redistribute it and/or modify
  it under the terms of the GNU General Public License as published by
  the Free Software Foundation; either version 2 of the License, or
  (at your option) any later version.

  This program is distributed in the hope that it will be useful,
  but WITHOUT ANY WARRANTY; without even the implied warranty of
  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
  GNU General Public License for more details.

  You should have received a copy of the GNU General Public License
  along with this program; if not, write to the Free Software
  Foundation, Inc., 51 Franklin St, Fifth Floor, Boston, MA  02110-1301  USA

*/


#include <string>

#include <QtGui/QTreeView>
#include <QtGui/QHeaderView>
#include <QtGui/QComboBox>
#include <QtGui/QResizeEvent>
#include <QtGui/QMenu>
#include <QtGui/QApplication>

#include "dbClipboard.h"
#include "dbClipboardData.h"
#include "layHierarchyControlPanel.h"
#include "layCellTreeModel.h"
#include "layLayoutView.h"
#include "layMainWindow.h"
#include "layAbstractMenu.h"
#include "layDialogs.h"
#include "tlInternational.h"
#include "tlString.h"
#include "gtf.h"

namespace lay
{

// --------------------------------------------------------------------
//  A helper class the identifies clipboard data 

class CellClipboardData
  : public db::ClipboardData
{
public:
  CellClipboardData () { }
};

// --------------------------------------------------------------------
//  HCPCellTreeWidget implementation

HCPCellTreeWidget::HCPCellTreeWidget (QWidget *parent, const char *name)
  : QTreeView (parent)
{
  setObjectName (name);
}

void 
HCPCellTreeWidget::mouseDoubleClickEvent (QMouseEvent *event)
{
  QModelIndex index (indexAt (event->pos ()));
  if (index.isValid ()) {
    emit cell_double_clicked (index);
  }
}

void 
HCPCellTreeWidget::mousePressEvent (QMouseEvent *event)
{
  if (event->button () == Qt::MidButton) {
    //  eat this event.
  } else {
    QTreeView::mousePressEvent (event);
  }
}

void 
HCPCellTreeWidget::mouseReleaseEvent (QMouseEvent *event)
{
  if (event->button () == Qt::MidButton) {
    QModelIndex index (indexAt (event->pos ()));
    if (index.isValid ()) {
      emit cell_middle_clicked (index);
    }
  } else {
    QTreeView::mouseReleaseEvent (event);
  }
}


// --------------------------------------------------------------------
//  HierarchyControlPanel implementation

void 
HierarchyControlPanel::init_menu (lay::AbstractMenu &menu)
{
  MenuLayoutEntry context_menu [] = {
    MenuLayoutEntry ("new_cell:edit:edit_mode",     tl::translate ("New Cell"),    SLOT (cm_new_cell ())),
    MenuLayoutEntry ("delete_cell:edit:edit_mode",  tl::translate ("Delete Cell"), SLOT (cm_cell_delete ())),
    MenuLayoutEntry ("rename_cell:edit:edit_mode",  tl::translate ("Rename Cell"), SLOT (cm_cell_rename ())),
    MenuLayoutEntry ("visibility_group:edit_mode"),
    MenuLayoutEntry ("show_as_top",       tl::translate ("Show As New Top"),       SLOT (cm_cell_select ())),
    MenuLayoutEntry ("visibility_group"),
    MenuLayoutEntry ("hide_cell",         tl::translate ("Hide"),                  SLOT (cm_cell_hide ())),
    MenuLayoutEntry ("show_cell",         tl::translate ("Show"),                  SLOT (cm_cell_show ())),
    MenuLayoutEntry ("show_all",          tl::translate ("Show All"),              SLOT (cm_cell_show_all ())),
    MenuLayoutEntry ("utils_group"),
    MenuLayoutEntry ("open_current",      tl::translate ("Open Current Cell"),     SLOT (cm_open_current_cell ())),
    MenuLayoutEntry ("file_group"),
    MenuLayoutEntry ("save_cell_as",      tl::translate ("Save Current Cell As"),  SLOT (cm_save_current_cell_as ())),
    MenuLayoutEntry::last ()
  };

  MenuLayoutEntry main_menu [] = {
    MenuLayoutEntry ("@hcp_context_menu", "",                                      context_menu),
    MenuLayoutEntry::last ()
  };

  menu.init (main_menu);
}

HierarchyControlPanel::HierarchyControlPanel (lay::LayoutView *view, QWidget *parent, const char *name)
  : QFrame (parent, name), 
    m_enable_cb (true), 
    mp_view (view),
    m_visibility_needs_update (true),
    m_active_index (0),
    m_do_update_content_dm (this, &HierarchyControlPanel::do_update_content)
{
  mp_selector = new QComboBox (this, "cellview_selection");

  connect (mp_selector, SIGNAL (activated (int)), this, SLOT (selection_changed (int)));

  setSizePolicy (QSizePolicy (QSizePolicy::Minimum, QSizePolicy::Preferred, 0, 0));

  mp_view->add_cellview_list_observer (*this);
  mp_view->add_hier_changed_observer (*this);

  do_update_content ();
}

HierarchyControlPanel::~HierarchyControlPanel ()
{
  //  .. nothing yet ..
}

QSize
HierarchyControlPanel::sizeHint () const
{
  int w = 120; // TODO: better(?): mp_cell_list->sizeHint ().width ();
  return QSize (w, 0);
}

void 
HierarchyControlPanel::resize (int w, int h)
{
  QFrame::resize (w, h);
  rearrange (w, h);
}

void 
HierarchyControlPanel::resizeEvent (QResizeEvent *re)
{
  rearrange (re->size ().width (), re->size ().height ());
}

void 
HierarchyControlPanel::setGeometry (int x, int y, int w, int h)
{
  QFrame::setGeometry (x, y, w, h);
  rearrange (w, h);
}

void 
HierarchyControlPanel::rearrange (int w, int h)
{
  if (h < 0) {

    mp_selector->hide ();
    for (std::vector <QTreeView *>::const_iterator f = mp_cell_lists.begin (); f != mp_cell_lists.end (); ++f) {
      (*f)->hide ();
    }

  } else {

    int hs = mp_selector->sizeHint ().height ();
    if (mp_cell_lists.size () <= 1) {
      hs = 0;
      mp_selector->hide ();
    } else {
      mp_selector->setGeometry (0, 0, w, hs);
      mp_selector->show ();
    }
    int i = 0;
    for (std::vector <QTreeView *>::const_iterator f = mp_cell_lists.begin (); f != mp_cell_lists.end (); ++f, ++i) {
      if (i == m_active_index) {
        (*f)->show ();
      } else {
        (*f)->hide ();
      }
      (*f)->setGeometry (0, hs, w, h - hs);
    }

  }
}

bool 
HierarchyControlPanel::event (QEvent *e)
{
  if (e->type () == QEvent::MaxUser) {
    //  GTF probe event
    e->accept ();
    return true;
  } else {
    return QFrame::event (e);
  }
}

void 
HierarchyControlPanel::context_menu (const QPoint &p)
{
  if (! mp_cell_lists.empty ()) {
    QMenu *ctx_menu = lay::MainWindow::instance ().menu ().detached_menu ("hcp_context_menu");
    ctx_menu->exec (mp_cell_lists [0]->mapToGlobal (p));
  }
}

void
HierarchyControlPanel::cm_cell_select ()
{
  cell_path_type path;
  current_cell (active (), path);
  emit cell_selected (path, active ());
}

void 
HierarchyControlPanel::middle_clicked (const QModelIndex &index)
{
  if (index.isValid ()) {

    //  build the path to the cell we clicked at
    cell_path_type path;
    CellTreeItem *item = (CellTreeItem *) index.internalPointer ();
    while (item) {
      path.push_back (item->cell_index ());
      item = item->parent ();
    }
    if (! path.empty ()) {
      std::reverse (path.begin (), path.end ());
    }

    emit cell_selected (path, active ());

  }
}

void 
HierarchyControlPanel::double_clicked (const QModelIndex &index)
{
  if (index.isValid ()) {
    mp_view->manager ()->transaction (tl::translate ("Show or hide cell")); 
    CellTreeItem *item = (CellTreeItem *) index.internalPointer ();
    if (mp_view->is_cell_hidden (item->cell_index (), m_active_index)) {
      mp_view->show_cell (item->cell_index (), m_active_index);
    } else {
      mp_view->hide_cell (item->cell_index (), m_active_index);
    }
    mp_view->manager ()->commit ();
  }
}

void
HierarchyControlPanel::cm_cell_hide ()
{
  mp_view->manager ()->transaction (tl::translate ("Hide cell")); 
  CellTreeItem *item = current_item ();
  if (item) {
    mp_view->hide_cell (item->cell_index (), m_active_index);
  }
  mp_view->manager ()->commit ();
}

void
HierarchyControlPanel::cm_cell_show ()
{
  mp_view->manager ()->transaction (tl::translate ("Show cell")); 
  CellTreeItem *item = current_item ();
  if (item) {
    mp_view->show_cell (item->cell_index (), m_active_index);
  }
  mp_view->manager ()->commit ();
}

void
HierarchyControlPanel::cm_cell_show_all ()
{
  mp_view->manager ()->transaction (tl::translate ("Show all cells")); 
  mp_view->show_all_cells ();
  mp_view->manager ()->commit ();
}

void 
HierarchyControlPanel::set_current_cell (int cv_index, const cell_path_type &path)
{
  if (cv_index < 0 || cv_index >= int (mp_cell_lists.size ())) {
    return;
  }

  if (path.size () > 0) {
    
    CellTreeModel *model = dynamic_cast <CellTreeModel *> (mp_cell_lists [cv_index]->model ());
    for (int c = 0; c < model->toplevel_items (); ++c) {
      CellTreeItem *item = model->toplevel_item (c);
      if (item->cell_index () == path.front ()) {
        item = find_child_item (path.begin () + 1, path.end (), item);
        if (item) {
          mp_cell_lists [cv_index]->scrollTo (model->model_index (item));
          mp_cell_lists [cv_index]->setCurrentIndex (model->model_index (item));
          break;
        }
      }
    } 

  }
}

void
HierarchyControlPanel::current_cell (int cv_index, HierarchyControlPanel::cell_path_type &path) const
{
  if (cv_index < 0 || cv_index >= int (mp_cell_lists.size ())) {
    return;
  }
  if (! mp_cell_lists [cv_index]->currentIndex ().isValid ()) {
    return;
  }
  
  CellTreeItem *item = (CellTreeItem *) mp_cell_lists [cv_index]->currentIndex ().internalPointer ();
  while (item) {
    path.push_back (item->cell_index ());
    item = item->parent ();
  }

  if (! path.empty ()) {
    std::reverse (path.begin (), path.end ());
  }
}

void
HierarchyControlPanel::set_background_color (QColor c)
{
  m_background_color = c;
  for (std::vector <QTreeView *>::const_iterator f = mp_cell_lists.begin (); f != mp_cell_lists.end (); ++f) {
    QPalette pl ((*f)->palette ());
    pl.setColor (QColorGroup::Base, c);
    (*f)->setPalette (pl);
  }
}

void
HierarchyControlPanel::set_text_color (QColor c)
{
  m_text_color = c;
  for (std::vector <QTreeView *>::const_iterator f = mp_cell_lists.begin (); f != mp_cell_lists.end (); ++f) {
    QPalette pl ((*f)->palette ());
    pl.setColor (QColorGroup::Text, c);
    (*f)->setPalette (pl);
  }
}

void
HierarchyControlPanel::signal_observed ()
{
  size_t i = 0; 
  for (std::vector <lay::CellView>::const_iterator cv = m_cellviews.begin (); cv != m_cellviews.end (); ++cv, ++i) {
    if (m_needs_update.size () > i) {
      m_needs_update [i] = true; 
    }
    if (m_force_close.size () > i) {
      m_force_close [i] = true; 
    }
  }
  m_do_update_content_dm ();
}

void 
HierarchyControlPanel::select_active (int cellview_index)
{
  if (cellview_index != m_active_index) {
    mp_selector->setCurrentItem (cellview_index);
    selection_changed (cellview_index);
  }
}

void
HierarchyControlPanel::selection_changed (int index)
{
  if (index != m_active_index) {
    m_active_index = index;
    int i;
    i = 0;
    for (std::vector <QTreeView *>::const_iterator f = mp_cell_lists.begin (); f != mp_cell_lists.end (); ++f, ++i) {
      if (i == index) {
        (*f)->show ();
      }
    }
    i = 0;
    for (std::vector <QTreeView *>::const_iterator f = mp_cell_lists.begin (); f != mp_cell_lists.end (); ++f, ++i) {
      if (i != index) {
        (*f)->hide ();
      }
    }
  }
}

CellTreeItem *
HierarchyControlPanel::find_child_item (cell_path_type::const_iterator start, cell_path_type::const_iterator end, CellTreeItem *p)
{
  if (start == end) {
    return p;
  } else {

    for (int n = 0; n < p->children (); ++n) {
      CellTreeItem *item = p->child (n);
      if (item && item->cell_index () == *start) {
        return find_child_item (start + 1, end, item);
      }
    }

    //  not found
    return 0;

  }
}

void
HierarchyControlPanel::do_update_content ()
{
  unsigned int i;

  for (i = 0; i < mp_view->cellviews (); ++i) {
    if (i >= m_force_close.size ()) {
      m_force_close.push_back (true);
    }
    if (i >= m_needs_update.size ()) {
      m_needs_update.push_back (true);
    }
    if (i >= m_cellviews.size ()) {
      m_force_close [i] = true;
      m_needs_update [i] = true;
    }
  }

  unsigned int n = std::min ((unsigned int) m_cellviews.size (), mp_view->cellviews ());
  for (i = 0; i < n; ++i) {

    if (&m_cellviews [i]->layout () != &mp_view->cellview (i)->layout ()) {
      m_needs_update [i] = true;
      m_force_close [i] = true;
    } else if (m_cellviews [i].combined_unspecific_path () != mp_view->cellview (i).combined_unspecific_path ()) {
      m_needs_update [i] = true;
    }

    if (m_needs_update [i] || m_visibility_needs_update) {
      mp_cell_lists [i]->doItemsLayout (); //  this schedules a redraw 
    }

    m_cellviews [i] = mp_view->cellview (i);

  }

  if (m_cellviews.size () < mp_view->cellviews ()) {
    for (; i < mp_view->cellviews (); ++i) {
      m_cellviews.push_back (mp_view->cellview (i));
    }
  } else if (m_cellviews.size () > mp_view->cellviews ()) {
    m_cellviews.erase (m_cellviews.begin () + mp_view->cellviews (), m_cellviews.end ()); 
  }

  while (mp_cell_lists.size () < m_cellviews.size ()) {

    int cv_index = int (mp_cell_lists.size ());

    HCPCellTreeWidget *cell_list = new HCPCellTreeWidget (this, "tree");
    cell_list->setModel (new CellTreeModel (cell_list, mp_view, cv_index));
    cell_list->setUniformRowHeights (true);

    QPalette pl;
    if (m_text_color.isValid ()) {
      pl.setColor (QColorGroup::Text, m_text_color);
    }
    if (m_background_color.isValid ()) {
      pl.setColor (QColorGroup::Base, m_background_color);
    }
    cell_list->setPalette (pl);

    cell_list->header ()->hide ();
    cell_list->setSelectionMode (QTreeView::SingleSelection);
    cell_list->setRootIsDecorated (true);
    cell_list->setIndentation (14);
    cell_list->setContextMenuPolicy (Qt::CustomContextMenu);

    connect (cell_list, SIGNAL(customContextMenuRequested (const QPoint &)), this, SLOT (context_menu (const QPoint &)));
    connect (cell_list, SIGNAL(cell_double_clicked (const QModelIndex &)), this, SLOT (double_clicked (const QModelIndex &)));
    connect (cell_list, SIGNAL(cell_middle_clicked (const QModelIndex &)), this, SLOT (middle_clicked (const QModelIndex &)));

    mp_cell_lists.push_back (cell_list);
    
  }

  while (mp_cell_lists.size () > m_cellviews.size ()) {
    delete mp_cell_lists.back ();
    mp_cell_lists.pop_back ();
  } 

  for (int cvi = 0; cvi < int (m_cellviews.size ()) && cvi < mp_selector->count (); ++cvi) {
    mp_selector->changeItem ((m_cellviews [cvi]->name () + " (@" + tl::to_string (cvi + 1) + ")").c_str (), cvi); 
  }
  while (mp_selector->count () < int (m_cellviews.size ())) {
    mp_selector->insertItem ((m_cellviews [mp_selector->count ()]->name () + " (@" + tl::to_string (mp_selector->count () + 1) + ")").c_str ());
  }
  while (mp_selector->count () > int (m_cellviews.size ())) {
    mp_selector->removeItem (mp_selector->count () - 1);
  } 

  if (m_active_index >= int (m_cellviews.size ())) {
    m_active_index = int (m_cellviews.size ()) - 1;
  }
  mp_selector->setCurrentItem (m_active_index);

  for (int i = 0; i < int (m_cellviews.size ()); ++i) {
    
    if (m_needs_update [i]) {

      //  draw the cells in the level of the current cell,
      //  add an "above" entry if there is a level above.
      //  highlight the current entry. If the index is 
      //  invalid, just clear the list.

      if (m_force_close [i]) {

        m_force_close [i] = false;

        mp_cell_lists [i]->collapse (mp_cell_lists [i]->rootIndex ());
        
        if (mp_cell_lists [i]->model ()) {
          delete mp_cell_lists [i]->model ();
        }
        mp_cell_lists [i]->setModel (new CellTreeModel (mp_cell_lists [i], mp_view, i));

      }

      //  locate the item and make selected
      mp_cell_lists [i]->clearSelection ();
      if (m_cellviews [i].unspecific_path ().size () > 0 || m_cellviews [i].specific_path ().size () > 0) {

        CellView::unspecific_cell_path_type path (m_cellviews [i].combined_unspecific_path ());
        
        CellTreeModel *model = dynamic_cast <CellTreeModel *> (mp_cell_lists [i]->model ());
        for (int c = 0; c < model->toplevel_items (); ++c) {
          CellTreeItem *item = model->toplevel_item (c);
          if (item->cell_index () == path.front ()) {
            item = find_child_item (path.begin () + 1, path.end (), item);
            if (item) {
              mp_cell_lists [i]->scrollTo (model->model_index (item));
              mp_cell_lists [i]->selectionModel ()->select (model->model_index (item), QItemSelectionModel::Select);
              break;
            }
          }
        } 

      }

      m_needs_update [i] = false;

    }

    if (i == m_active_index) {
      mp_cell_lists [i]->show ();
    } else {
      mp_cell_lists [i]->hide ();
    }

  }

  //  rearrange the geometry of the objects. This must be last since it requires
  //  a valid list view model
  rearrange (width (), height ());
}

CellTreeItem *
HierarchyControlPanel::current_item () const
{
  if (m_active_index < 0 || m_active_index >= int (mp_cell_lists.size ())) {
    return 0;
  }
  if (mp_cell_lists [m_active_index]->currentIndex ().isValid ()) {
    return (CellTreeItem *) mp_cell_lists [m_active_index]->currentIndex ().internalPointer ();
  } else {
    return 0;
  }
}

bool 
HierarchyControlPanel::has_focus () const
{
  return m_active_index >= 0 && m_active_index < int (mp_cell_lists.size ()) && mp_cell_lists [m_active_index]->hasFocus ();
}

void
HierarchyControlPanel::cut () 
{
  CellTreeItem *ci = current_item ();
  if (ci) {

    //  first copy
    copy ();

    //  then do a (shallow) delete
    db::Layout &layout = m_cellviews [m_active_index]->layout ();
    layout.delete_cell (ci->cell_index ());

  }
}

bool
HierarchyControlPanel::has_selection ()
{
  return (current_item () != 0);
}

void
HierarchyControlPanel::copy () 
{
  CellTreeItem *ci = current_item ();
  if (ci == 0 || m_active_index < 0 || m_active_index >= int (mp_cell_lists.size ())) {
    return;
  }

  db::Layout &layout = m_cellviews [m_active_index]->layout ();

  db::ClipboardValue<lay::CellClipboardData> *cd = new db::ClipboardValue<lay::CellClipboardData> ();
  cd->get ().add (layout, layout.cell (ci->cell_index ()), 1);

  db::Clipboard::instance ().clear ();
  db::Clipboard::instance () += cd;
}

void
HierarchyControlPanel::paste () 
{
  if (m_active_index < 0 || m_active_index >= int (mp_cell_lists.size ())) {
    return;
  }

  db::Layout &layout = m_cellviews [m_active_index]->layout ();

  std::vector<unsigned int> new_layers;

  //  paste the content into the active cellview.
  std::vector <db::Layout::cell_index_type> new_tops;
  for (db::Clipboard::iterator c = db::Clipboard::instance ().begin (); c != db::Clipboard::instance ().end (); ++c) {
    const db::ClipboardValue<lay::CellClipboardData> *value = dynamic_cast<const db::ClipboardValue<lay::CellClipboardData> *> (*c);
    if (value) {
      std::vector<unsigned int> nl = value->get ().insert (layout, 0, &new_tops);
      new_layers.insert (new_layers.end (), nl.begin (), nl.end ());
    }
  }

  //  Add new layers to the view if required.
  mp_view->add_new_layers (new_layers, m_active_index);
  mp_view->update_content ();

  //  select the first new top cell
  if (! new_tops.empty ()) {
    //  TODO: this does not work properly: since we are inside a transaction, bboxes are not updated
    //  correctly. Thus, the cell_fit does not work properly.
    mp_view->select_cell_fit (new_tops [0], m_active_index);
  }
}

} // namespace lay 
