
/*

  KLayout Layout Viewer
  Copyright (C) 2006-2007 Matthias Koefferlein

  This program is free software; you can redistribute it and/or modify
  it under the terms of the GNU General Public License as published by
  the Free Software Foundation; either version 2 of the License, or
  (at your option) any later version.

  This program is distributed in the hope that it will be useful,
  but WITHOUT ANY WARRANTY; without even the implied warranty of
  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
  GNU General Public License for more details.

  You should have received a copy of the GNU General Public License
  along with this program; if not, write to the Free Software
  Foundation, Inc., 51 Franklin St, Fifth Floor, Boston, MA  02110-1301  USA

*/


#ifndef HDR_layLayoutView
#define HDR_layLayoutView

#include <vector>
#include <map>
#include <set>
#include <string>

#include <QtGui/QFrame>

#include "layLayerProperties.h"
#include "layAnnotationShapes.h"
#include "layLayoutCanvas.h"
#include "layColorPalette.h"
#include "layCellView.h"
#include "layViewOp.h"
#include "layEditable.h"
#include "layPlugin.h"
#include "layDisplayState.h"
#include "layBookmarkList.h"
#include "gsiObject.h"
#include "tlException.h"
#include "dbInstElement.h"

class QSpinBox;

namespace db {
  class Layout;
  class Manager;
  class SaveLayoutOptions;
}

namespace lay {

class AbstractMenu;
class LayerControlPanel;
class HierarchyControlPanel;
class MouseTracker;
class ZoomService;
class SelectionService;
class MoveService;
class MainWindow;
class Browser;
class ColorButton;

/**
 *  @brief The coordinate tracker
 *
 *  This is the interface that the main window provides
 *  in order to display status information about current coordinate and 
 *  progress.
 */
class CoordTracker
{
public:
  virtual ~CoordTracker () { }

  virtual void current_pos (double x, double y, bool dbu_units) = 0;
  virtual void clear_current_pos () = 0;
};

/**
 *  @brief A layer display properties structure
 *
 *  The layer properties encapsulate the settings relevant for
 *  the display of a layer.
 *
 *  "brightness" is a index that indicates how much to make the
 *  color brighter to darker rendering the effective color 
 *  (eff_frame_color (), eff_fill_color ()). It's value is roughly between
 *  -255 and 255.
 */
struct LayerDisplayProperties
{
  LayerDisplayProperties ();

  bool operator== (const LayerDisplayProperties &d);
  bool operator!= (const LayerDisplayProperties &d);

  /**
   *  @brief render the effective frame color
   *  
   *  The effective frame color is computed from the frame color brightness and the
   *  frame color.
   */
  color_t eff_frame_color () const;

  /**
   *  @brief render the effective frame color
   *  
   *  The effective frame color is computed from the frame color brightness and the
   *  frame color.
   */
  color_t eff_fill_color () const;

  //  display styles
  color_t frame_color;
  color_t fill_color;
  int frame_brightness;
  int fill_brightness;
  unsigned int dither_pattern;
  bool visible;
  bool transparent;
  int width;
  bool marked;
  int animation;
  std::string name;
};

/**
 *  @brief The layout view object
 *
 *  The layout view is responsible for displaying one or a set of layouts.
 *  It is composed of a canvas and controls to control the appearance.
 *  It manages the layer display list, bookmark list etc.
 */
class LayoutView
  : public QFrame,
    public lay::Editables,
    public lay::Plugin,
    public gsi::Object<LayoutView>
{
Q_OBJECT

public:
  typedef lay::CellView::unspecific_cell_path_type cell_path_type;
  typedef lay::CellView::cell_index_type cell_index_type;
  typedef std::pair<std::string, DisplayState> bookmark_type;
  enum drop_small_cells_cond_type { DSC_Max = 0, DSC_Min = 1, DSC_Sum = 2 };

  /**
   *  @brief Constructor
   */
  LayoutView (db::Manager *mgr, lay::MainWindow *main_widget, QWidget *parent = 0, const char *name = "view");

  /**
   *  @brief Constructor (clone from another view)
   */
  LayoutView (lay::LayoutView *source, db::Manager *mgr, lay::MainWindow *main_widget, QWidget *parent = 0, const char *name = "main_window");

  /** 
   *  @brief Destructor
   */
  ~LayoutView ();

  /**
   *  @brief Determine if there is something to copy
   *
   *  This reimplementation of the lay::Editables interface additionally
   *  looks for content providers in the tree views for example.
   */
  bool has_selection ();

  /**
   *  @brief Paste from clipboard
   *
   *  This reimplementation of the lay::Editables interface additionally
   *  looks for paste receivers in the tree views for example.
   */
  void paste ();

  /**
   *  @brief Copy to clipboard
   *
   *  This reimplementation of the lay::Editables interface additionally
   *  looks for copy providers in the tree views for example.
   */
  void copy ();

  /**
   *  @brief Cut to clipboard
   *
   *  This reimplementation of the lay::Editables interface additionally
   *  looks for cut & copy providers in the tree views for example.
   */
  void cut ();

  /**
   *  @brief Get the window title of the view
   */
  std::string title () const;

  /**
   *  @brief Set the window title to an explicit string
   */
  void set_title (const std::string &t);

  /**
   *  @brief Reset the explicit title and enable the automatic naming
   */
  void reset_title ();

  /**
   *  @brief Fill the layer properties for a new layer
   *
   *  The layer's source must be set already to allow to compute the initial color.
   *  It is assumed that the layer is appended at the end of the layer list. This
   *  is important to set the dither pattern index accordingly.
   */
  void init_layer_properties (LayerProperties &props) const;

  /**
   *  @brief Create a set of new layers for the given layers of the given cellview
   */
  void add_new_layers (const std::vector <unsigned int> &layer_ids, int cv_index);

  /**
   *  @brief Set the layer properties of a layer with the given index
   */
  void set_properties (const LayerPropertiesConstIterator &iter, const LayerProperties &props);

  /**
   *  @brief Replace the layer node with a new one
   */
  void replace_layer_node (const LayerPropertiesConstIterator &iter, const LayerPropertiesNode &props);

  /**
   *  @brief Insert the given layer properties node into the list before the given index
   *
   *  This method returns a reference to the element created.
   */
  const LayerPropertiesNode &insert_layer (const LayerPropertiesConstIterator &iter, const LayerPropertiesNode &props);

  /**
   *  @brief Delete the layer properties node 
   *  
   *  This method deletes the object that the iterator points to and invalidates
   *  the iterator since the object that the iterator points to is no longer valid.
   */
  void delete_layer (LayerPropertiesConstIterator &iter);

  /**
   *  @brief Begin iterator for the layers
   *
   *  This iterator delivers recursively the layers of this view
   */
  LayerPropertiesConstIterator begin_layers () const
  {
    return m_layer_props.begin_const_recursive ();
  }

  /**
   *  @brief End iterator for the layers
   */
  LayerPropertiesConstIterator end_layers () const
  {
    return m_layer_props.end_const_recursive ();
  }

  /**
   *  @brief Replace the layer properties list
   */
  void set_properties (const LayerPropertiesList &list);

  /**
   *  @brief Clear the layer view list
   */
  void clear_layers ()
  {
    set_properties (LayerPropertiesList ());
  }

  /**
   *  @brief Access the layer properties list
   */
  const LayerPropertiesList &get_properties () const
  {
    return m_layer_props;
  }

  /**
   *  @brief Set the custom dither pattern 
   */
  void set_dither_pattern (const lay::DitherPattern &pattern);

  /**
   *  @brief Obtain the custom dither pattern 
   */
  const lay::DitherPattern &dither_pattern () const
  {
    return m_layer_props.dither_pattern ();
  }

  /**
   *  @brief Add an observer for a change in the annotations shape list
   *  
   *  If annotation shapes are added or removed, this observers are sent a signal.
   */
  void add_annotations_changed_observer (tl::Observer &observer)
  {
    m_annotations_changed_observers.add_observer (observer);
  }

  /**
   *  @brief Remove an observer for a change in the annotations shape list
   */
  void remove_annotations_changed_observer (tl::Observer &observer)
  {
    m_annotations_changed_observers.remove_observer (observer);
  }

  /**
   *  @brief Add an observer for a change in the hierarchy of the layouts
   *  
   *  If the hierarchy of a layout is changed, this observer is sent a signal.
   *  This may happen due to the removal or insertion of cells or instances.
   */
  void add_hier_changed_observer (tl::Observer &observer)
  {
    m_hier_changed_observers.add_observer (observer);
  }

  /**
   *  @brief Remove an observer for the hierarchy of the layouts
   */
  void remove_hier_changed_observer (tl::Observer &observer)
  {
    m_hier_changed_observers.remove_observer (observer);
  }

  /**
   *  @brief Add an observer for a change in the geometries of the layouts
   *  
   *  If something on the geometries of a cell in one the layouts changes, this
   *  event is triggered. This may happed due to the removal or insertion of shapes or cell instances.
   *  In general, this indicates the need for redrawing of the layout for example.
   */
  void add_geom_changed_observer (tl::Observer &observer)
  {
    m_geom_changed_observers.add_observer (observer);
  }

  /**
   *  @brief Remove an observer for a change in the geometries of the layouts
   */
  void remove_geom_changed_observer (tl::Observer &observer)
  {
    m_geom_changed_observers.remove_observer (observer);
  }

  /**
   *  @brief Add an observer to be notified before something on a cellview is changed
   *
   *  This notification is sent before something on the cellview (i.e. the current cell of 
   *  of set of the cellview) is changed. The integer event signals the index of the cellview to
   *  be changed.
   *  This event is followed by a corresponding cellview_changed event after the cellview
   *  has changed.
   */
  void add_cellview_about_to_change_observer (tl::Observer &observer)
  {
    m_cellview_about_to_change_observers.add_observer (observer);
  }

  /**
   *  @brief Remove an observer to be notified before something on a cellview is changed
   */
  void remove_cellview_about_to_change_observer (tl::Observer &observer)
  {
    m_cellview_about_to_change_observers.remove_observer (observer);
  }

  /**
   *  @brief Add an observer to be notified before something on the cellviews is changed
   *
   *  This notification is sent before something on the cellviews is changed, i.e. a cellview is deleted or inserted.
   *  Each of these events is followed by a cellviews_changed event after the change has been made.
   */
  void add_cellviews_about_to_change_observer (tl::Observer &observer)
  {
    m_cellviews_about_to_change_observers.add_observer (observer);
  }

  /**
   *  @brief Remove an observer to be notified before something on the cellviews is changed
   */
  void remove_cellviews_about_to_change_observer (tl::Observer &observer)
  {
    m_cellviews_about_to_change_observers.remove_observer (observer);
  }

  /**
   *  @brief Add a cellview list observer
   *  
   *  If a cellview is added or removed, this observer is sent a signal.
   */
  void add_cellview_list_observer (tl::Observer &observer)
  {
    m_cellviews_changed_observers.add_observer (observer);
  }

  /**
   *  @brief Remove a cellview list observer
   */
  void remove_cellview_list_observer (tl::Observer &observer)
  {
    m_cellviews_changed_observers.remove_observer (observer);
  }

  /**
   *  @brief Add a cellview observer
   *  
   *  If a cellview is changed (i.e. the cell is changed) this event is sent.
   *  The integer event will receive the index of the cellview that has changed.
   */
  void add_cellview_observer (tl::Observer &observer)
  {
    m_cellview_changed_observers.add_observer (observer);
  }

  /**
   *  @brief Remove a cellview observer
   */
  void remove_cellview_observer (tl::Observer &observer)
  {
    m_cellview_changed_observers.remove_observer (observer);
  }

  /**
   *  @brief Add a file open observer
   *  
   *  If a new file is loaded, this observer is triggered.
   */
  void add_file_open_observer (tl::Observer &observer)
  {
    m_file_open_observers.add_observer (observer);
  }

  /**
   *  @brief Remove a file open observer
   */
  void remove_file_open_observer (tl::Observer &observer)
  {
    m_file_open_observers.remove_observer (observer);
  }

  /**
   *  @brief Add a viewport changed observer
   *  
   *  If the viewport (the rectangle that is shown) changes, this observer
   *  is triggered.
   */
  void add_viewport_changed_observer (tl::Observer &observer)
  {
    m_viewport_changed_observers.add_observer (observer);
  }

  /**
   *  @brief Remove a viewport changed observer
   */
  void remove_viewport_changed_observer (tl::Observer &observer)
  {
    m_viewport_changed_observers.remove_observer (observer);
  }

  /**
   *  @brief Add a layer list observer
   *  
   *  If the layer list changes, the observers signal_observed(int) method is called. 
   *  The int's bit 0 is set, if the properties have changed. If the arguments bit 1 is
   *  set, the hierarchy has changed.
   */
  void add_layer_list_observer (tl::Observer &observer)
  {
    m_layer_list_observers.add_observer (observer);
  }

  /**
   *  @brief Remove a layer list observer
   */
  void remove_layer_list_observer (tl::Observer &observer)
  {
    m_layer_list_observers.remove_observer (observer);
  }

  /**
   *  @brief Add a cell visibility observer
   */
  void add_cell_visibility_observer (tl::Observer &observer)
  {
    m_cell_visibility_observers.add_observer (observer);
  }

  /**
   *  @brief Remove a cell visibility observer
   */
  void remove_cell_visibility_observer (tl::Observer &observer)
  {
    m_cell_visibility_observers.remove_observer (observer);
  }

  /**
   *  @brief Retrieve the index of the currently active layer
   *
   *  The active layer is the one that is active in the layer
   *  browser panel.
   *  This method returns a negative value, if no layer is active.
   */
  lay::LayerPropertiesConstIterator current_layer () const;

  /**
   *  @brief Return the layers that are selected in the layer browser
   *
   *  Returns an empty list if no layer is selected.
   */
  std::vector<lay::LayerPropertiesConstIterator> selected_layers () const;

  /**
   *  @brief Save the given cellview into the given file (with options)
   */
  void save_as (unsigned int index, const std::string &filename, bool gzip, const db::SaveLayoutOptions &options);

  /**
   *  @brief Implementation of the undo operations
   */
  virtual void undo (db::Op *op) throw ();

  /**
   *  @brief Implementation of the redo operations
   */
  virtual void redo (db::Op *op) throw ();

  /** 
   *  @brief Set the cellview at the given index
   *
   *  If cvindex is used as the cellview index to associate the 
   *  layout with. As a side effect, this method will emit a 
   *  title_changed signal, which means that the cellview passed
   *  should be correctly named before to reflect the correct
   *  title if no explicit title is set.
   */
  void set_layout (const lay::CellView &cv, unsigned int cvindex);

  /**
   *  @brief clear the cellviews
   */
  void clear_cellviews ();

  /**
   *  @brief erase one cellview with the given index
   */
  void erase_cellview (unsigned int index);

  /**
   *  @brief Save the layer properties 
   */
  void save_layer_props (const std::string &fn) throw (tl::Exception);

  /**
   *  @brief Load the layer properties 
   */
  void load_layer_props (const std::string &fn) throw (tl::Exception);

  /**
   *  @brief Save the screen content to a file
   */
  void save_screenshot (const std::string &fn) throw (tl::Exception);

  /**
   *  @brief Save an image file
   */
  void save_image (const std::string &fn, unsigned int width, unsigned int height) throw (tl::Exception);

  /**
   *  @brief Hierarchy level selection setter
   */
  void set_hier_levels (std::pair<int, int> l);

  /**
   *  @brief Hierarchy level selection setter
   */
  void set_min_hier_levels (int l)
  {
    set_hier_levels (std::pair<int, int> (l, get_max_hier_levels ()));
  }

  /**
   *  @brief Hierarchy level selection setter
   */
  void set_max_hier_levels (int l)
  {
    set_hier_levels (std::pair<int, int> (get_min_hier_levels (), l));
  }

  /**
   *  @brief Hierarchy level selection getter
   */
  std::pair<int, int> get_hier_levels () const;

  /**
   *  @brief Hierarchy level selection getter
   */
  int get_min_hier_levels () const
  {
    return get_hier_levels ().first;
  }

  /**
   *  @brief Hierarchy level selection getter
   */
  int get_max_hier_levels () const
  {
    return get_hier_levels ().second;
  }

  /**
   *  @brief Return true, if there is a "last" display state
   */
  bool has_last_display_state ();

  /**
   *  @brief Return true, if there is a "next" display state
   */
  bool has_next_display_state ();

  /**
   *  @brief Cell box/label color setter
   */
  void cell_box_color (QColor c);

  /**
   *  @brief Cell box/label getter
   */
  QColor cell_box_color () const
  {
    return m_box_color;
  }

  /**
   *  @brief Transform label flag setter
   */
  void cell_box_text_transform (bool xform);

  /**
   *  @brief Transform label flag getter
   */
  bool cell_box_text_transform () const
  {
    return m_box_text_transform;
  }

  /**
   *  @brief Label font setter
   */
  void cell_box_text_font (unsigned int f);

  /**
   *  @brief Label font setter
   */
  unsigned int cell_box_text_font () const
  {
    return m_box_font;
  }

  /** 
   *  @brief Visibility of cell boxes
   */
  void cell_box_visible (bool vis);

  /** 
   *  @brief Visibility of cell boxes
   */
  bool cell_box_visible () const
  {
    return m_cell_box_visible;
  }

  /**
   *  @brief Min instance label size setter
   */
  void min_inst_label_size (int px);

  /**
   *  @brief Min instance label size getter
   */
  int min_inst_label_size () const
  {
    return m_min_size_for_label;
  }

  /** 
   *  @brief Visibility of text objects
   */
  void text_visible (bool vis);

  /** 
   *  @brief Visibility of text objects
   */
  bool text_visible () const
  {
    return m_text_visible;
  }

  /** 
   *  @brief Show properties 
   */
  void show_properties_as_text (bool sp);

  /** 
   *  @brief Show properties 
   */
  bool show_properties_as_text () 
  {
    return m_show_properties;
  }

  /** 
   *  @brief Lazy rendering of text objects
   */
  void text_lazy_rendering (bool lzy);

  /** 
   *  @brief Lazy rendering of text objects
   */
  bool text_lazy_rendering () 
  {
    return m_text_lazy_rendering;
  }

  /**
   *  @brief Text object font setter
   */
  void text_font (unsigned int f);

  /**
   *  @brief Text object font getter
   */
  unsigned int text_font () const
  {
    return m_text_font;
  }

  /**
   *  @brief The default text size property
   */
  void default_text_size (double fs);

  /**
   *  @brief The default text size property
   */
  double default_text_size () const
  {
    return m_default_text_size;
  }

  /**
   *  @brief Apply text transformation property
   */
  void apply_text_trans (bool f);
  
  /**
   *  @brief Apply text transformation property
   */
  bool apply_text_trans () const
  {
    return m_apply_text_trans;
  }
  
  /**
   *  @brief Text object color
   */
  void text_color (QColor c);

  /**
   *  @brief Text object color
   */
  QColor text_color () const
  {
    return m_text_color;
  }

  /**
   *  @brief Fit new cell property
   *
   *  If this property is set to false, the coordinate window is not
   *  changed if the cell is changed.
   */
  void fit_new_cell (bool f);
  
  /**
   *  @brief Fit new cell
   *
   *  If this property is set to false, the coordinate window is not
   *  changed if the cell is changed.
   */
  bool fit_new_cell () const
  {
    return m_fit_new_cell;
  }
  
  /**
   *  @brief Set the color palette
   *
   *  The color palette is used for coloring new layers and is shown in the
   *  layer panel.
   */
  void set_palette (const lay::ColorPalette &);
  
  /**
   *  @brief Get the color palette
   */
  const lay::ColorPalette &get_palette () const
  {
    return m_palette;
  }
  
  /** 
   *  @brief Reload the given cellview into the current view
   *
   *  The cellview is given by index in the current view's cellview list.
   */
  void reload_layout (unsigned int cv_index) throw (tl::Exception);

  /** 
   *  @brief Load a (new) file into the layout
   *
   *  The add_cellview param controls whether to create a new cellview
   *  or clear all cellviews before.
   *
   *  @return The index of the cellview loaded.
   */
  unsigned int load_layout (const std::string &filename, bool add_cellview) throw (tl::Exception);

  /** 
   *  @brief Create a new, empty layout 
   *
   *  The add_cellview param controls whether to create a new cellview
   *  or clear all cellviews before.
   *
   *  @return The index of the cellview loaded.
   */
  unsigned int create_layout (bool add_cellview) throw (tl::Exception);

  /**
   *  @brief Deactivate the view
   *
   *  A view gets "deactivated" if it is pushed into the background.
   *  This is the opposite operation than activation. 
   */
  void deactivate ();

  /**
   *  @brief Activate the view
   *
   *  A view gets "activated" soon before it is shown. Activation 
   *  may involve allocation of resources only required in visible state.
   *  The active window is made owner of the status bar - i.e. only 
   *  this window may write to it.
   */
  void activate ();

  /**
   *  @brief Pass the current position from the mouse tracker to the status bar
   */ 
  void current_pos (double x, double y);

  /**
   *  @brief Get the index of the active cellview (shown in hierarchy browser)
   */
  int active_cellview_index () const;
  
  /**
   *  @brief Select a certain cellview for the active one
   */
  void set_active_cellview_index (int index);

  /**
   *  @brief Cell path of the current cell
   *
   *  The current cell is the one highlighted in the browser with the focus rectangle. The 
   *  current path is returned for the cellview given by cv_index.
   */
  void current_cell_path (int cv_index, cell_path_type &path) const;
  
  /**
   *  @brief Cell path of the current cell
   *
   *  This method version is provided for automation purposes mainly.
   */
  cell_path_type get_current_cell_path (int cv_index) const
  {
    cell_path_type r;
    current_cell_path (cv_index, r);
    return r;
  }
  
  /**
   *  @brief Cell path of the current cell in the active cellview 
   *
   *  This is a convenience function returning the path for the active cellview.
   */
  void current_cell_path (cell_path_type &path) const
  {
    current_cell_path (active_cellview_index (), path);
  }
  
  /**
   *  @brief Set the path to the current cell
   *
   *  The current cell is the one highlighted in the browser with the focus rectangle. The
   *  cell given by the path is hightlighted and scrolled into view.
   */
  void set_current_cell_path (int cv_index, const cell_path_type &path);
  
  /**
   *  @brief Set the path to the current cell is the current cellview
   *
   *  This is a convenience function setting the path for the active cellview.
   */
  void set_current_cell_path (const cell_path_type &path) 
  {
    set_current_cell_path (active_cellview_index (), path);
  }
  
  /**
   *  @brief Obtain the number of cellviews
   */
  unsigned int cellviews () const
  {
    return m_cellviews.size ();
  }
  
  /**
   *  @brief Obtain the cellviews as a vector
   */
  const std::vector<lay::CellView> &cellview_vector () const
  {
    return m_cellviews;
  }
  
  /**
   *  @brief Obtain the cell view reference for an index
   *
   *  If the index is invalid, an empty cell view reference is returned
   */
  const CellView &cellview (unsigned int index) const;

  /**
   *  @brief Obtain the list of annotation shapes
   */
  const lay::AnnotationShapes &annotation_shapes () const
  {
    return m_annotation_shapes;
  }

  /**
   *  @brief Obtain the list of annotation shapes (non-const version)
   */
  lay::AnnotationShapes &annotation_shapes () 
  {
    return m_annotation_shapes;
  }

  /** 
   *  @brief Select the list of cellviews for this window and fit cell
   *
   *  Warning: use with care!
   */
  void select_cellviews_fit (const std::vector <CellView> &cvs);

  /** 
   *  @brief Select the list of cellviews for this window
   *
   *  Warning: use with care!
   */
  void select_cellviews (const std::vector <CellView> &cvs);

  /** 
   *  @brief Shift and scale the window
   */
  void shift_window (double f, double dx, double dy);

  /** 
   *  @brief Goto window
   *
   *  Position the window to the new position x and y with a size of s (approximately).
   *  If s is negative or zero, the windows is just shifted.
   */
  void goto_window (const db::DPoint &p, double s = -1.0);

  /** 
   *  @brief Return the displayed window
   */
  db::DBox box () const;

  /**
   *  @brief Create a new cell with the given in the given cellview
   *
   *  @param cv_index The index of the cellview where to create the cell
   *  @param cell_name The name of the cell (an exception is thrown if a cell of that name already exists)
   *  @return The index of the new cell
   */
  db::Layout::cell_index_type new_cell (int cv_index, const std::string &cell_name);
  
  /**
   *  @brief Bookmark the current view under the given name
   */
  void bookmark_view (const std::string &name);

  /**
   *  @brief Show the bookmark management form
   */
  void manage_bookmarks ();

  /**
   *  @brief Obtain the bookmarks list
   */
  const BookmarkList &bookmarks () const
  {
    return m_bookmarks;
  }

  /**
   *  @brief Set the bookmarks list
   */
  void bookmarks (const BookmarkList &b);

  std::pair <bool, std::string> redo_available ();
  std::pair <bool, std::string> undo_available ();

  /**
   *  @brief Select a certain mode (by index)
   */
  void mode (unsigned int m);
  
  /**
   *  @brief Query the intrinsic mouse modes available
   */
  static unsigned int intrinsic_mouse_modes (std::vector<std::string> *descriptions);

  /**
   *  @brief Perform the layout view's initialisations on the main menu
   */
  static void init_menu (lay::AbstractMenu &menu);

  /**
   *  @brief Query the default mode
   */
  static unsigned int default_mode ();

  /**
   *  @brief Enable an editable service by index
   */
  void enable_editable (unsigned int m, bool en);

  /**
   *  @brief Get a list of cellview index and transform variants 
   */
  std::set< std::pair<db::CplxTrans, int> > cv_transform_variants () const;
  
  /**
   *  @brief Get the global transform variants for a given cellview index
   */
  std::vector<db::CplxTrans> cv_transform_variants (int cv_index) const;
  
  /**
   *  @brief Access to the hidden cell list
   */
  const std::vector <std::set <cell_index_type> > &hidden_cells () const
  { 
    return m_hidden_cells;
  }

  /**
   *  @brief Get the "dbu_coordinates" flag 
   *  
   *  If this flag is true, the property dialogs and other display functions should use 
   *  database units to display coordinates etc.
   */
  bool dbu_coordinates () const
  {
    return m_dbu_coordinates;
  }

  /**
   *  @brief Set the "dbu_coordinates" flag 
   *  
   *  If this flag is true, the property dialogs and other display functions should use 
   *  database units to display coordinates etc.
   */
  void dbu_coordinates (bool f); 

  /**
   *  @brief Get the "absolute_coordinates" flag 
   *  
   *  If this flag is true, the property dialogs and other display functions should use 
   *  absolute (on top level) coordinates for points etc.
   *  "absolute" may as well refer to orientation, not only to coordinates, 
   *  if transformations are considered.
   */
  bool absolute_coordinates () const
  {
    return m_absolute_coordinates;
  }

  /**
   *  @brief Set the "absolute_coordinates" flag 
   *  
   *  If this flag is true, the property dialogs and other display functions should use 
   *  absolute (on top level) coordinates for points etc.
   *  "absolute" may as well refer to orientation, not only to coordinates, 
   *  if transformations are considered.
   */
  void absolute_coordinates (bool f);

  /**
   *  @brief Obtain a new bitmap object
   */
  lay::Bitmap *bitmap (const lay::ViewOp &viewop);

  /**
   *  @brief Get the view object widget 
   *  
   *  This method intentionally delivers the ViewObjectWidget, not the 
   *  LayoutCanvas to emphasize that the LayoutCanvas object shall not
   *  be modified.
   */
  lay::ViewObjectWidget *view_object_widget () 
  {
    return mp_canvas;
  }

  /**
   *  @brief Get the current viewport 
   */
  const lay::Viewport &viewport () const
  {
    return mp_canvas->viewport ();
  }

  /**
   *  @brief Background color property
   */
  QColor background_color () const
  {
    return mp_canvas->background_color ();
  }

  /**
   *  @brief Foreground color property
   */
  QColor foreground_color () const
  {
    return mp_canvas->foreground_color ();
  }

  /**
   *  @brief Write accessor to the "drop small cells" flag
   */
  void drop_small_cells (bool m);

  /**
   *  @brief Read accessor to the "drop small cells" flag
   */
  bool drop_small_cells () const
  {
    return m_drop_small_cells;
  }

  /**
   *  @brief Write accessor to the "drop small cells" value
   */
  void drop_small_cells_value (unsigned int s);

  /**
   *  @brief Read accessor to the "drop small cells" value
   */
  unsigned int drop_small_cells_value () const
  {
    return m_drop_small_cells_value;
  }

  /**
   *  @brief Write accessor to the "drop small cells" condition 
   */
  void drop_small_cells_cond (drop_small_cells_cond_type t);

  /**
   *  @brief Read accessor to the "drop small cells" condition 
   */
  drop_small_cells_cond_type drop_small_cells_cond () const
  {
    return m_drop_small_cells_cond;
  }

  /**
   *  @brief Get the Drawings interface
   *  
   *  Although the Drawings interface is implemented by LayoutCanvas,
   *  it is a different interface from ViewObjectWidget. 
   */
  lay::Drawings *drawings () 
  {
    return mp_canvas;
  }

  /**
   *  @brief Attach the coordinate tracker widget to this view
   */
  void connect_coord_tracker (lay::CoordTracker *ct);

  /**
   *  @brief Select synchroneous mode or deselect it
   */
  void set_synchroneous (bool sync_mode);

  /**
   *  @brief Tell the state of synchroneous mode
   */
  bool synchroneous () const
  {
    return m_synchroneous;
  }

  /** 
   *  @brief Localize the plugin of the given Type
   *
   *  This method will return 0, if no such plugin is registered
   */
  template <class PI>
  PI *get_plugin () const
  {
    PI *pi = 0;
    for (std::vector<lay::Plugin *>::const_iterator p = mp_plugins.begin (); p != mp_plugins.end () && !pi; ++p) {
      pi = dynamic_cast<PI *> (*p);
    }
    return pi;
  }

  /** 
   *  @brief Localize the plugins of the given Type
   *
   *  This method will return 0, if no such plugin is registered
   */
  template <class PI>
  std::vector<PI *> get_plugins () const
  {
    std::vector<PI *> pi;
    for (std::vector<lay::Plugin *>::const_iterator p = mp_plugins.begin (); p != mp_plugins.end (); ++p) {
      if (dynamic_cast<PI *> (*p) != 0) {
        pi.push_back (dynamic_cast<PI *> (*p));
      }
    }
    return pi;
  }

  /**
   *  @brief Enable or disable the actions for edit functions 
   * 
   *  This method is used by non-modal dialogs that want to suppress any editing
   *  activities (like browsers) while they are open.
   */
  void enable_edits (bool enable);

  /**
   *  @brief Rename the given cellview to the given name
   *
   *  The name of the cell view is shown in the title
   */
  void rename_cellview (const std::string &name, int cellview_index);

  /** 
   *  @brief Descend into the hierarchy along the given specific path for the given cellview
   */
  void descend (const std::vector<db::InstElement> &path, int cellview_index);

  /** 
   *  @brief Ascend one level in the hierarchy for the given cellview
   *
   *  @return The instance element removed by ascending the path
   */
  db::InstElement ascend (int cellview_index);

  /** 
   *  @brief Select a cell by path for a certain cell view and fit cell
   */
  void select_cell_fit (const cell_path_type &path, int cellview_index);

  /** 
   *  @brief Select a cell by index for a certain cell view and fit cell
   */
  void select_cell_fit (cell_index_type index, int cellview_index);

  /** 
   *  @brief Select a cell by path for a certain cell view
   */
  void select_cell (const cell_path_type &path, int cellview_index);

  /** 
   *  @brief Select a cell by index for a certain cell view
   */
  void select_cell (cell_index_type index, int cellview_index);

  /**
   *  @brief Check if a cell is hidden
   */
  bool is_cell_hidden (cell_index_type ci, int cellview_index) const;

  /**
   *  @brief Add a cell to the hidden cell list
   */
  void hide_cell (cell_index_type ci, int cellview_index);

  /**
   *  @brief Remove a cell from the hidden cell list
   */
  void show_cell (cell_index_type ci, int cellview_index);

  /**
   *  @brief Clear the hidden cell list 
   */
  void show_all_cells ();

  /**
   *  @brief Update the layout view to the current state
   *
   *  This method must be called in order to trigger the update of
   *  the contents. It will be known, what needs to be updated.
   */
  void update_content ();

  /**
   *  @brief Force an unconditional update
   *
   *  This method is supposed to be used to ensure everything is updated.
   *  TODO: this method should become obsolete once there is a consistent
   *  signalling of states implemented.
   */
  void force_update_content ();

  /**
   *  @brief Internal method: check, if the layer tree is and an consistent state.
   *  
   *  This method is used by the layer tree model to check, if the tree has been brought into 
   *  a consistent state.
   *  HINT: for the layout this is solved more consistently with the "under construction" attribute
   *  of the layout. There is no equivalent object for the layer tree currently.
   */
  bool layer_model_updated ();

  /**
   *  @brief Get the default color for markers
   */
  QColor default_marker_color () const
  {
    return m_marker_color;
  }

  /**
   *  @brief Get the default line width for markers
   */
  int default_marker_line_width () const
  {
    return m_marker_line_width;
  }

  /**
   *  @brief Get the default vertex size for markers
   */
  int default_marker_vertex_size () const
  {
    return m_marker_vertex_size;
  }

  /**
   *  @brief Get the default halo flag for markers
   */
  int default_marker_halo () const
  {
    return m_marker_halo;
  }

  /**
   *  @brief Return true, if the cellview with the given index is editable
   */
  bool is_editable (unsigned int cv_index) const;

  /**
   *  @brief Return true, if any cellview is editable
   */
  bool is_editable () const;

public slots:
  /** 
   *  @brief Select last display state
   */
  void last_display_state ();

  /** 
   *  @brief Select next display state
   */
  void next_display_state ();

  void zoom_fit ();
  void zoom_box (const db::DBox &b);
  void zoom_trans (const db::DCplxTrans &b);
  void zoom_in ();
  void zoom_out ();
  void pan_left ();
  void pan_up ();
  void pan_right ();
  void pan_down ();
  void pan_center (const db::DPoint &p);

  /** 
   *  @brief Select a cell by index for a certain cell view
   *
   *  This will be forwarded to select_cell or select_cell_fit depending
   *  on the m_fit_new_cell flag.
   */
  void select_cell_dispatch (const cell_path_type &path, int cellview_index);

  /**
   *  @brief Goto a position/cell view that was saved with save_view 
   */
  void goto_view (const DisplayState &state);

  /**
   *  @brief Save a position/cell view that can be used for "goto_view"
   */
  void save_view (DisplayState &state);

  /**
   *  @brief Store the current state on the "last states" stack
   */
  void store_state ();

  /**
   *  @brief Cancel all edit operations and clear the selection
   */
  void cancel ();

  /**
   *  @brief Cancel all edit operations but leave selection
   */
  void cancel_edits ();

  /**
   *  @brief Select all levels of hierarchy available
   */
  void max_hier ();

  /**
   *  @brief Stop activities like shape browser and redrawing thread
   */
  void stop ();

  /**
   *  @brief Stop redrawing thread
   */
  void stop_redraw ();

  //  menu callbacks
  void cm_new_layer ();
  void cm_sel_flip_x ();
  void cm_sel_flip_y ();
  void cm_sel_rot_cw ();
  void cm_sel_rot_ccw ();
  void cm_sel_free_rot ();
  void cm_sel_move ();

  //  forwarded to the layer control panel
  void cm_select_all ();
  void cm_hide ();
  void cm_show ();
  void cm_rename ();
  void cm_delete ();
  void cm_insert ();
  void cm_group ();
  void cm_ungroup ();
  void cm_source ();
  void cm_sort_by_name ();
  void cm_sort_by_ild ();
  void cm_sort_by_idl ();
  void cm_sort_by_ldi ();
  void cm_sort_by_dli ();
  void cm_regroup_by_index ();
  void cm_regroup_by_datatype ();
  void cm_regroup_by_layer ();
  void cm_regroup_flatten ();
  void cm_expand_all ();
  void cm_add_missing ();
  void cm_remove_unused ();

  //  forwarded to the cell control panel
  void cm_cell_rename ();
  void cm_cell_delete ();
  void cm_cell_select ();
  void cm_open_current_cell ();
  void cm_cell_hide ();
  void cm_cell_show ();
  void cm_cell_show_all ();

  //  called by children and owner
  void redraw (bool clear = true);
  void redraw_layer (unsigned int index);
  void redraw_deco_layer ();
  void redraw_cell_boxes ();
  void layer_order_changed ();
  void prop_changed ();
  void visibility_changed ();
  void timer ();
  void menu_activated (const std::string &symbol);
  void deactivate_all_browsers ();
  void min_hier_changed (int i);
  void max_hier_changed (int i);

  //  event handlers used to connect to the layout object's events
  void signal_hier_changed ();
  void signal_bboxes_changed ();
  void signal_prop_ids_changed ();
  void signal_cell_name_changed ();
  void signal_annotations_changed ();

signals:
  void title_changed ();

protected:
  void set_view_ops ();

private:
  QTimer *mp_timer;
  lay::CoordTracker *mp_coord_tracker;
  lay::LayoutCanvas *mp_canvas;
  lay::LayerControlPanel *mp_control_panel;
  lay::HierarchyControlPanel *mp_hierarchy_panel;
  QSpinBox *mp_min_hier_spbx;
  QSpinBox *mp_max_hier_spbx;
  std::vector <CellView> m_cellviews;
  lay::AnnotationShapes m_annotation_shapes;
  std::vector <std::set <cell_index_type> > m_hidden_cells;
  std::string m_title;
  std::string m_def_lyr_file;
  bool m_synchroneous;

  int m_from_level, m_to_level;

  QColor m_ctx_color;
  int m_ctx_dimming;

  QColor m_box_color;
  bool m_box_text_transform;
  unsigned int m_box_font;
  int m_min_size_for_label;
  bool m_cell_box_visible;

  QColor m_marker_color;
  int m_marker_line_width;
  int m_marker_vertex_size;
  bool m_marker_halo;

  bool m_text_visible;
  bool m_text_lazy_rendering;
  bool m_show_properties;
  QColor m_text_color;
  bool m_apply_text_trans;
  double m_default_text_size;
  unsigned int m_text_font;

  bool m_drop_small_cells;
  unsigned int m_drop_small_cells_value;
  drop_small_cells_cond_type m_drop_small_cells_cond;

  bool m_fit_new_cell;
  bool m_dbu_coordinates;
  bool m_absolute_coordinates;

  bool m_activated;
  bool m_animated;
  unsigned int m_phase;

  lay::ColorPalette m_palette;

  std::vector <DisplayState> m_display_states;
  unsigned int m_display_state_ptr;

  lay::LayerPropertiesList m_layer_props;
  tl::Observed m_layer_list_observers;
  tl::Observed m_cell_visibility_observers;
  tl::Observed m_cellviews_about_to_change_observers;
  tl::Observed m_cellview_about_to_change_observers;
  tl::Observed m_cellviews_changed_observers;
  tl::Observed m_cellview_changed_observers;
  tl::Observed m_file_open_observers;
  tl::Observed m_viewport_changed_observers;
  tl::Observed m_hier_changed_observers;
  tl::Observed m_geom_changed_observers;
  tl::Observed m_annotations_changed_observers;
  tl::ObserverDelegate<LayoutView> m_hier_changed_observer;
  tl::ObserverDelegate<LayoutView> m_bboxes_changed_observer;
  tl::ObserverDelegate<LayoutView> m_prop_ids_changed_observer;
  tl::ObserverDelegate<LayoutView> m_cell_name_changed_observer;
  tl::ObserverDelegate<LayoutView> m_annotations_changed_observer;

  BookmarkList m_bookmarks;

  //  service and editable management
  unsigned int m_mode;
  
  //  services & editables
  lay::MouseTracker *mp_tracker;
  lay::ZoomService *mp_zoom_service;
  lay::SelectionService *mp_selection_service;
  lay::MoveService *mp_move_service;

  std::vector<lay::Plugin *> mp_plugins;

  db::LayerProperties m_new_layer_props;
  db::DPoint m_move_dist;
  int m_del_cell_mode;

  void init (db::Manager *mgr, lay::MainWindow *main_window);

  void do_redraw (int layer);
  void do_redraw ();
  void do_transform (const db::DCplxTrans &tr);

  void background_color (QColor c);
  void ctx_color (QColor c);
  void ctx_dimming (int percent);

  bool configure (const std::string &name, const std::string &value);

  void finish_cellviews_changed ();
  void create_initial_layer_props (int cv_index);
};

}

namespace gsi {

  template <> struct ClassTag<lay::LayoutView> {
    typedef CantCopySemanticsTag copy_semantics;
    typedef CantDefaultCreateSemanticsTag create_semantics;
  };

}

#endif

