
/*

  KLayout Layout Viewer
  Copyright (C) 2006-2007 Matthias Koefferlein

  This program is free software; you can redistribute it and/or modify
  it under the terms of the GNU General Public License as published by
  the Free Software Foundation; either version 2 of the License, or
  (at your option) any later version.

  This program is distributed in the hope that it will be useful,
  but WITHOUT ANY WARRANTY; without even the implied warranty of
  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
  GNU General Public License for more details.

  You should have received a copy of the GNU General Public License
  along with this program; if not, write to the Free Software
  Foundation, Inc., 51 Franklin St, Fifth Floor, Boston, MA  02110-1301  USA

*/


#include <iostream>
#include <fstream>

#include <QtGui/QWindowsStyle>
#include <QtGui/QCDEStyle>
#include <QtGui/QMenu>
#include <QtGui/QMenuBar>
#include <QtGui/QToolBar>
#include <QtGui/QStatusBar>
#include <QtGui/QInputDialog>
#include <QtGui/QMessageBox>
#include <QtGui/QPainter>

#include "tlInternational.h"
#include "tlTimer.h"
#include "tlLog.h"
#include "tlXMLParser.h"
#include "tlAssert.h"
#include "tlDeferredExecution.h"
#include "dbMemStatistics.h"
#include "dbManager.h"
#include "dbStream.h"
#include "dbSaveLayoutOptions.h"
#include "dbClipboard.h"
#include "layApplication.h"
#include "layVersion.h"
#include "layDialogs.h"
#include "layMainWindow.h"
#include "layProgress.h"
#include "layProgressWidget.h"
#include "layLayerControlPanel.h" // because of LabelWithBackground
#include "layFileDialog.h"
#include "layRedrawThread.h"
#include "layAbstractMenu.h"
#include "layExceptions.h"
#include "laySaveLayoutOptionsDialog.h"
#include "SettingsFormImpl.h"
#include "CellSelectionFormImpl.h"
#include "SelectCellViewFormImpl.h"
#include "LayoutPropertiesFormImpl.h"
#include "MainConfigPage.h"
#include "MainConfigPage2.h"
#include "HelpAboutDialog.h"
#include "gsi.h"
#include "gtf.h"
#include "rbaConsoleImpl.h"

namespace lay
{

// ------------------------------------------------------------
//  Declaration of the configuration options
//  The configuration pages are declared via two "dummy" plugins

const std::string cfg_grid ("grid-micron");
const std::string cfg_synchronized_views ("synchronized-views");
const std::string cfg_default_dbu ("dbu");


class MainPluginDeclaration
  : public lay::PluginDeclaration
{
public:
  virtual void get_options (std::vector < std::pair<std::string, std::string> > &options) const
  {
    options.push_back (std::pair<std::string, std::string> (cfg_grid, "0.001"));
  }

  virtual lay::ConfigPage *config_page (QWidget *parent, std::string &title) const 
  {
    title = tl::translate ("Grid");
    return new MainConfigPage (parent); 
  }
};

static tl::Registrar<lay::PluginDeclaration>::Class config_decl (new MainPluginDeclaration (), 1000, "MainPlugin(1)");

class MainPluginDeclaration2
  : public lay::PluginDeclaration
{
public:
  virtual void get_options (std::vector < std::pair<std::string, std::string> > &options) const
  {
    options.push_back (std::pair<std::string, std::string> (cfg_synchronized_views, "false"));
  }

  virtual lay::ConfigPage *config_page (QWidget *parent, std::string &title) const 
  {
    title = tl::translate ("Misc.");
    return new MainConfigPage2 (parent); 
  }
};

static tl::Registrar<lay::PluginDeclaration>::Class config_decl2 (new MainPluginDeclaration2 (), 1001, "MainPlugin(2)");

class MainPluginDeclaration3
  : public lay::PluginDeclaration
{
public:
  virtual void get_options (std::vector < std::pair<std::string, std::string> > &options) const
  {
    options.push_back (std::pair<std::string, std::string> (cfg_default_dbu, "0.001"));
  }
};

static tl::Registrar<lay::PluginDeclaration>::Class config_decl3 (new MainPluginDeclaration3 (), 1002, "MainPlugin(3)");

// -------------------------------------------------------------
//  The "grid" configuration page

MainConfigPage::MainConfigPage (QWidget *parent)
  : lay::ConfigPage (parent)
{
  mp_ui = new Ui::MainConfigPage ();
  mp_ui->setupUi (this);
}

MainConfigPage::~MainConfigPage ()
{
  delete mp_ui;
  mp_ui = 0;
}

void 
MainConfigPage::setup (lay::Plugin *root)
{
  double grid_micron = 0.0;
  root->config_get (cfg_grid, grid_micron);
  mp_ui->grid_edit->setText (tl::to_string (grid_micron).c_str ());
}

void 
MainConfigPage::commit (lay::Plugin *root)
{
  try {
    double g;
    tl::from_string (mp_ui->grid_edit->text ().ascii (), g);
    root->config_set (cfg_grid, g);
  } catch (...) { }
}

// ------------------------------------------------------------
//  The "misc" config page

MainConfigPage2::MainConfigPage2 (QWidget *parent)
  : lay::ConfigPage (parent)
{
  mp_ui = new Ui::MainConfigPage2 ();
  mp_ui->setupUi (this);
}

MainConfigPage2::~MainConfigPage2 ()
{
  delete mp_ui;
  mp_ui = 0;
}

void 
MainConfigPage2::setup (lay::Plugin *root)
{
  bool flag = false;
  root->config_get (cfg_synchronized_views, flag);
  mp_ui->sync_views_cbx->setChecked (flag);
}

void 
MainConfigPage2::commit (lay::Plugin *root)
{
  root->config_set (cfg_synchronized_views, mp_ui->sync_views_cbx->isChecked ());
}

// -------------------------------------------------------------

class ViewWidgetStack 
  : public QWidget
{
public:
  ViewWidgetStack (QWidget *parent = 0, const char *name = 0)
    : QWidget (parent, name)
  {
    mp_bglabel = new LabelWithBackground (this);
    mp_bglabel->setText (tl::translate ("Use File/Open to open a layout").c_str ());
    mp_bglabel->setAlignment (Qt::AlignVCenter | Qt::AlignHCenter);
    mp_bglabel->show ();
  }

  void addWidget (LayoutView *w)
  {
    m_widgets.push_back (w);
    updateGeometry ();
    raiseWidget (w);
  }

  void removeWidget (LayoutView *w)
  {
    for (std::vector <LayoutView *>::iterator child = m_widgets.begin (); child != m_widgets.end (); ++child) {
      if (*child == w) {
        m_widgets.erase (child, child + 1);
        if (m_widgets.size () == 0) {
          mp_bglabel->show ();
        }
        return;
      }
    }
  }

  void raiseWidget (LayoutView *w)
  {
    for (std::vector <LayoutView *>::iterator child = m_widgets.begin (); child != m_widgets.end (); ++child) {
      if (*child != w) {
        (*child)->deactivate ();
      }
    }
    mp_bglabel->hide ();
    w->activate ();
    w->show ();
    for (std::vector <LayoutView *>::iterator child = m_widgets.begin (); child != m_widgets.end (); ++child) {
      if (*child != w) {
        (*child)->hide ();
      }
    }
  }
  
  LayoutView *widget (size_t index)
  {
    if (index < m_widgets.size ()) {
      return m_widgets [index];
    } else {
      return 0;
    }
  }

  QWidget *background_widget ()
  {
    return mp_bglabel;
  }

protected:
  virtual void resizeEvent (QResizeEvent *)
  {
    updateGeometry ();
  }

  void updateGeometry ()
  {
    //  set the geometry of all children 
    for (std::vector <LayoutView *>::iterator child = m_widgets.begin (); child != m_widgets.end (); ++child) {
      (*child)->setGeometry (0, 0, width (), height ());
    }
    mp_bglabel->setGeometry (0, 0, width (), height ());
  }

  std::vector <LayoutView *> m_widgets;
  QLabel *mp_bglabel;
};

// -------------------------------------------------------------

static MainWindow *mw_instance = 0;

MainWindow &
MainWindow::instance ()
{
  tl_assert (mw_instance != 0);
  return *mw_instance; 
}

// -----------------------------------

MainWindow::MainWindow (QApplication *app, const char *name)
    : QMainWindow (0, name), 
      m_mode (std::numeric_limits<unsigned int>::max ()),
      mp_setup_form (new SettingsFormImpl (0, this, "setup_form")),
      m_current_view (0),
      m_edits_enabled (true),
      m_disable_tab_selected (false),
      m_exited (false),
      m_do_update_menu_dm (this, &MainWindow::do_update_menu),
      m_grid_micron (0.001), 
      m_new_cell_window_size (2.0),
      m_default_dbu (0.001),
      m_synchronized_views (false),
      m_synchroneous (false),
      m_disabled_edits (0),
      m_work_in_progress (false),
      mp_app (app)
{
 
  if (mw_instance != 0) {
    throw tl::Exception (tl::translate ("Only one instance of MainWindow may be created"));
  }
  mw_instance = this;

  init_menu ();

  mp_console = new rba::ConsoleImpl ();
  mp_console->setModal (false);

  mp_pr = new lay::ProgressReporter (this, app);

  QFrame *mf = new QFrame (this, "main_frame");
  setCentralWidget (mf);

  QVBoxLayout *vbl = new QVBoxLayout (mf);
  vbl->setMargin (0);
  vbl->setSpacing (0);

  mp_tab_bar = new QTabBar (mf);
  vbl->addWidget (mp_tab_bar);
  connect (mp_tab_bar, SIGNAL (currentChanged (int)), this, SLOT (view_selected (int)));

  mp_view_stack = new ViewWidgetStack (mf, "view_stack");
  vbl->addWidget (mp_view_stack);

  mp_tool_bar = new QToolBar (this);
  mp_tool_bar->setObjectName ("toolbar");
  mp_tool_bar->setMovable (false);
  mp_tool_bar->setToolButtonStyle (Qt::ToolButtonTextUnderIcon);
  addToolBar (Qt::TopToolBarArea, mp_tool_bar);

  QMenuBar *mbar = menuBar ();
  mbar->setObjectName ("menubar");

  mp_menu->build (mbar, mp_tool_bar);

  connect (mp_menu, SIGNAL (changed ()), this, SLOT (menu_changed ()));

  mp_status_bar = statusBar ();
  mp_status_bar->setObjectName ("status_bar");

  mp_msg_label = new QLabel (mp_status_bar, "msg_label");
  mp_status_bar->addWidget (mp_msg_label, 1);

  mp_right_stack = new QStackedWidget (mp_status_bar);

  mp_cp_frame = new QFrame (mp_right_stack);
  mp_right_stack->addWidget (mp_cp_frame);

  QHBoxLayout *cp_frame_ly = new QHBoxLayout (mp_cp_frame);
  cp_frame_ly->setMargin (0);
  cp_frame_ly->setSpacing (0);
  cp_frame_ly->insertSpacing (-1, 6);
  QLabel *xl = new QLabel ("x:", mp_cp_frame, "x");
  cp_frame_ly->addWidget (xl);
  cp_frame_ly->insertSpacing (-1, 6);
  mp_cpx_label = new QLabel (mp_cp_frame, "cpx_label");
  mp_cpx_label->setAlignment (Qt::AlignVCenter | Qt::AlignRight);
  mp_cpx_label->setMinimumSize (100, 0); // @@@
  cp_frame_ly->addWidget (mp_cpx_label);
  cp_frame_ly->insertSpacing (-1, 6);
  QLabel *yl = new QLabel ("y:", mp_cp_frame, "y");
  cp_frame_ly->addWidget (yl);
  cp_frame_ly->insertSpacing (-1, 6);
  mp_cpy_label = new QLabel (mp_cp_frame, "cpy_label");
  mp_cpy_label->setAlignment (Qt::AlignVCenter | Qt::AlignRight);
  mp_cpy_label->setMinimumSize (100, 0); // @@@
  cp_frame_ly->addWidget (mp_cpy_label);
  cp_frame_ly->insertSpacing (-1, 6);

  mp_progress_widget = new ProgressWidget (mp_pr, mp_right_stack);
  mp_right_stack->addWidget (mp_progress_widget);

  mp_right_stack->setCurrentIndex (0);
  mp_status_bar->addWidget (mp_right_stack);

  //  connect to the menus to provide the dynamic parts
  QMenu *edit_menu = mp_menu->menu ("edit_menu");
  tl_assert (edit_menu != 0);
  connect (edit_menu, SIGNAL (aboutToShow ()), this, SLOT (popup_menu_show ()));

  QMenu *zoom_menu = mp_menu->menu ("zoom_menu");
  tl_assert (zoom_menu != 0);
  connect (zoom_menu, SIGNAL (aboutToShow ()), this, SLOT (popup_menu_show ()));

  QMenu *bookmark_menu = mp_menu->menu ("bookmark_menu");
  tl_assert (bookmark_menu != 0);
  connect (bookmark_menu, SIGNAL (aboutToShow ()), this, SLOT (bookmark_menu_show ()));

  //  select the default mode
  select_mode (lay::LayoutView::default_mode ());

  //  create file dialogs:

  //  bookmarks file dialog
  mp_bookmarks_fdia = new lay::FileDialog (this, 
                            tl::translate ("Bookmarks File"), 
                            tl::translate ("All files (*);;Bookmark files (*.lyb)"), 
                            "lyb");
  //  layer properties
  mp_lprops_fdia = new lay::FileDialog (this, 
                            tl::translate ("Layer Properties File"), 
                            tl::translate ("All files (*);;Layer properties files (*.lyp)"), 
                            "lyp");
  //  screenshots
  mp_screenshot_fdia = new lay::FileDialog (this, 
                            tl::translate ("Screenshot"), 
                            tl::translate ("All files (*);;PNG files (*.png)"), 
                            "png");

  
  //  layout file dialog
  std::string fmts = tl::translate ("All files (*)");
  for (tl::Registrar<db::StreamFormatDeclaration>::iterator rdr = tl::Registrar<db::StreamFormatDeclaration>::begin (); rdr != tl::Registrar<db::StreamFormatDeclaration>::end (); ++rdr) {
    fmts += ";;" + rdr->file_format ();
  }
  mp_layout_fdia = new lay::FileDialog (this, tl::translate ("Layout File"), fmts);

  //  save layout options
  mp_layout_save_options = new lay::SaveLayoutOptionsDialog (this, tl::translate ("Save Layout Options"));

}

MainWindow::~MainWindow ()
{
  mw_instance = 0;

  //  explicitly delete the views here. Otherwise they
  //  are deleted by ~QWidget, which is too late since then
  //  the manager does not exist any longer.
  //  First empty the mp_views vector and then delete. This way,
  //  any callbacks issued during the deleting of the views do
  //  not find any invalid view pointers but rather nothing.

  std::vector <lay::LayoutView *> views;
  views.swap (mp_views);

  while (views.size () > 0) {
    delete views.back ();
    views.pop_back ();
  }

  //  delete the Menu after the views because they may want to access them in the destructor

  delete mp_menu;
  mp_menu = 0;

  delete mp_pr;
  mp_pr = 0;
}

void 
MainWindow::init_menu ()
{
  //  default menu layout
  
  MenuLayoutEntry empty_menu [] = {
    MenuLayoutEntry::last ()
  };

  MenuLayoutEntry file_menu [] = {
    MenuLayoutEntry ("new_layout:edit:edit_mode",  tl::translate ("New"),          SLOT (cm_new_layout ())),
    MenuLayoutEntry::separator ("post_new_group:edit_mode"),
    MenuLayoutEntry ("open:edit",         tl::translate ("Open"),                  SLOT (cm_open ())),
    MenuLayoutEntry ("open_same_panel:edit",tl::translate ("Open In Same Panel"),  SLOT (cm_open_too ()),             QKeySequence ("Shift+Ctrl+O")),
    MenuLayoutEntry ("open_new_panel:edit", tl::translate ("Open In New Panel"),   SLOT (cm_open_new_view ()),        QKeySequence ("Ctrl+O")),
    MenuLayoutEntry ("close:edit",        tl::translate ("Close"),                 SLOT (cm_close ()),                QKeySequence ("Ctrl+W")),
    MenuLayoutEntry ("clone",             tl::translate ("Clone Panel"),           SLOT (cm_clone ())),
    MenuLayoutEntry ("reload:edit",       tl::translate ("Reload"),                SLOT (cm_reload ()),               QKeySequence ("Ctrl+R")),
    MenuLayoutEntry::separator ("save_group"),
    MenuLayoutEntry ("save",              tl::translate ("Save"),                  SLOT (cm_save ())),
    MenuLayoutEntry ("save_as",           tl::translate ("Save As"),               SLOT (cm_save_as ())),
    MenuLayoutEntry::separator ("setup_group"),
    MenuLayoutEntry ("setup:edit",        tl::translate ("Setup"),                 SLOT (cm_setup ())),
    MenuLayoutEntry::separator ("misc_group"),
    MenuLayoutEntry ("screenshot:edit",   tl::translate ("Screenshot"),            SLOT (cm_screenshot ()),           QKeySequence ("Print")),
    MenuLayoutEntry ("layout_props",      tl::translate ("Layout Properties"),     SLOT (cm_layout_props ())),
    MenuLayoutEntry::separator ("layer_group"),
    MenuLayoutEntry ("load_layer_props:edit", tl::translate ("Load Layer Properties"), SLOT (cm_load_layer_props ())),
    MenuLayoutEntry ("save_layer_props:edit", tl::translate ("Save Layer Properties"), SLOT (cm_save_layer_props ())),
    MenuLayoutEntry::separator ("exit_group"),
    MenuLayoutEntry ("exit",              tl::translate ("Exit"),                  SLOT (cm_exit ()),                 QKeySequence ("Ctrl+Q")),
    MenuLayoutEntry::last ()
  };

  MenuLayoutEntry edit_layer_menu [] = {
    MenuLayoutEntry ("new_layer:edit:edit_mode",   tl::translate ("New Layer"),    SLOT (cm_new_layer ())),
    MenuLayoutEntry::last ()
  };

  MenuLayoutEntry edit_selection_menu [] = {
    MenuLayoutEntry ("sel_flip_x:edit_mode",       tl::translate ("Flip Horizontally"),  SLOT (cm_sel_flip_x ())),
    MenuLayoutEntry ("sel_flip_y:edit_mode",       tl::translate ("Flip Vertically"),    SLOT (cm_sel_flip_y ())),
    MenuLayoutEntry ("sel_rot_cw:edit_mode",       tl::translate ("Rotate Clockwise"),  SLOT (cm_sel_rot_cw ())),
    MenuLayoutEntry ("sel_rot_ccw:edit_mode",      tl::translate ("Rotate Counterclockwise"),    SLOT (cm_sel_rot_ccw ())),
    // does not work properly, since boxes cannot be freely rotated currently ..
    // MenuLayoutEntry ("sel_free_rot:edit_mode",     tl::translate ("Rotation By Angle .."),  SLOT (cm_sel_free_rot ())),
    MenuLayoutEntry ("sel_move:edit_mode",         tl::translate ("Move By .."),         SLOT (cm_sel_move ())),
    MenuLayoutEntry::last ()
  };

  MenuLayoutEntry edit_menu [] = {
    MenuLayoutEntry ("undo:edit",         tl::translate ("Undo"),                  SLOT (cm_undo ()),                 QKeySequence ("Ctrl+Z")),
    MenuLayoutEntry ("redo:edit",         tl::translate ("Redo"),                  SLOT (cm_redo ()),                 QKeySequence ("Ctrl+Y")),
    MenuLayoutEntry::separator ("new_group:edit_mode"),
    MenuLayoutEntry ("layer_menu:edit:edit_mode", tl::translate ("Layer"),         edit_layer_menu),
    MenuLayoutEntry ("selection_menu:edit:edit_mode", tl::translate ("Selection"), edit_selection_menu),
    MenuLayoutEntry::separator ("misc_group") ,
    MenuLayoutEntry ("delete:edit",       tl::translate ("Delete"),                SLOT (cm_delete ()),               QKeySequence ("Del")),
    MenuLayoutEntry ("show_properties:edit", tl::translate ("Properties"),         SLOT (cm_show_properties ())),
    MenuLayoutEntry::separator ("cpc_group"),
    MenuLayoutEntry ("copy:edit",         tl::translate ("Copy"),                  SLOT (cm_copy ()),                 QKeySequence ("Ctrl+C")),
    MenuLayoutEntry ("cut:edit",          tl::translate ("Cut"),                   SLOT (cm_cut ()),                  QKeySequence ("Ctrl+X")),
    MenuLayoutEntry ("paste:edit",        tl::translate ("Paste"),                 SLOT (cm_paste ()),                QKeySequence ("Ctrl+V")),
    MenuLayoutEntry::separator ("modes_group"),
    MenuLayoutEntry ("mode_menu",         tl::translate ("Mode"),                  empty_menu),
    MenuLayoutEntry ("select_menu",       tl::translate ("Select"),                empty_menu),
    MenuLayoutEntry::separator ("cancel_group"),
    MenuLayoutEntry ("cancel",            tl::translate ("Cancel"),                SLOT (cm_cancel ()),               QKeySequence ("Esc")),
    MenuLayoutEntry::last ()
  };

  MenuLayoutEntry bookmark_menu [] = {
    MenuLayoutEntry ("goto_bookmark_menu",tl::translate ("Goto Bookmark"),         empty_menu),
    MenuLayoutEntry ("bookmark_view",     tl::translate ("Bookmark This View"),    SLOT (cm_bookmark_view ())),
    MenuLayoutEntry::separator ("bookmark_mgm_group"),
    MenuLayoutEntry ("manage_bookmarks",  tl::translate ("Manage Bookmarks"),      SLOT (cm_manage_bookmarks ())),
    MenuLayoutEntry ("load_bookmarks",    tl::translate ("Load Bookmarks"),        SLOT (cm_load_bookmarks ())),
    MenuLayoutEntry ("save_bookmarks",    tl::translate ("Save Bookmarks"),        SLOT (cm_save_bookmarks ())),
    MenuLayoutEntry::last ()
  };

  MenuLayoutEntry zoom_menu [] = {
    MenuLayoutEntry ("max_hier",          tl::translate ("Full Hierarchy"),        SLOT (cm_max_hier ()),             QKeySequence ("*")),
    MenuLayoutEntry ("max_hier_0",        tl::translate ("Box Only"),              SLOT (cm_max_hier_0 ()),           QKeySequence ("0")),
    MenuLayoutEntry ("max_hier_1",        tl::translate ("Top Level Only"),        SLOT (cm_max_hier_1 ()),           QKeySequence ("1")),
    MenuLayoutEntry ("inc_max_hier",      tl::translate ("Increment Hierarchy"),   SLOT (cm_inc_max_hier ()),         QKeySequence ("+")),
    MenuLayoutEntry ("dec_max_hier",      tl::translate ("Decrement Hierarchy"),   SLOT (cm_dec_max_hier ()),         QKeySequence ("-")),
    MenuLayoutEntry::separator ("zoom_group"),
    MenuLayoutEntry ("zoom_fit",          tl::translate ("Zoom Fit"),              SLOT (cm_zoom_fit ()),             QKeySequence ("F2")),
    MenuLayoutEntry ("zoom_in",           tl::translate ("Zoom In"),               SLOT (cm_zoom_in ()),              QKeySequence ("Return")),
    MenuLayoutEntry ("zoom_out",          tl::translate ("Zoom Out"),              SLOT (cm_zoom_out ()),             QKeySequence ("Shift+Return")),
    MenuLayoutEntry::separator ("pan_group"),
    MenuLayoutEntry ("pan_up",            tl::translate ("Pan Up"),                SLOT (cm_pan_up ()),               QKeySequence ("Up")),
    MenuLayoutEntry ("pan_down",          tl::translate ("Pan Down"),              SLOT (cm_pan_down ()),             QKeySequence ("Down")),
    MenuLayoutEntry ("pan_left",          tl::translate ("Pan Left"),              SLOT (cm_pan_left ()),             QKeySequence ("Left")),
    MenuLayoutEntry ("pan_right",         tl::translate ("Pan Right"),             SLOT (cm_pan_right ()),            QKeySequence ("Right")),
    MenuLayoutEntry::separator ("redraw_group"),
    MenuLayoutEntry ("redraw",            tl::translate ("Redraw"),                SLOT (cm_redraw ())),
    MenuLayoutEntry::separator ("state_group"),
    MenuLayoutEntry ("last_display_state", tl::translate ("Last State"),           SLOT (cm_last_display_state ()), QKeySequence ("Shift+Tab")),
    MenuLayoutEntry ("next_display_state", tl::translate ("Next State"),           SLOT (cm_next_display_state ()), QKeySequence ("Tab")),
    MenuLayoutEntry::separator ("select_group"),
    MenuLayoutEntry ("select_cell:edit",  tl::translate ("Select Cell"),           SLOT (cm_select_cell ())),
    MenuLayoutEntry ("select_current_cell", tl::translate ("Select Current Cell"), SLOT (cm_select_current_cell ()), QKeySequence ("Ctrl+S")), 
    MenuLayoutEntry ("goto_position",     tl::translate ("Goto Position"),         SLOT (cm_goto_position ()),        QKeySequence ("Ctrl+G")),
    MenuLayoutEntry::last ()
  };

  MenuLayoutEntry help_menu [] = {
    MenuLayoutEntry ("about",             tl::translate ("About"),                 SLOT (cm_help_about ())),
    MenuLayoutEntry::last ()
  };

  MenuLayoutEntry tools_menu_with_ruby [] = {
    MenuLayoutEntry ("console",           tl::translate ("RBA Console"),           SLOT (cm_console ())),
    MenuLayoutEntry::separator ("tools_group"),
    MenuLayoutEntry::last ()
  };

  MenuLayoutEntry main_menu [] = {
    MenuLayoutEntry ("file_menu",         tl::translate ("&File"),                 file_menu),
    MenuLayoutEntry ("edit_menu",         tl::translate ("&Edit"),                 edit_menu),
    MenuLayoutEntry ("bookmark_menu",     tl::translate ("&Bookmarks"),            bookmark_menu),
    MenuLayoutEntry ("zoom_menu",         tl::translate ("&Display"),              zoom_menu),
    MenuLayoutEntry ("tools_menu",        tl::translate ("&Tools"),                lay::Application::instance ()->ruby_interpreter ().ruby_available () ? tools_menu_with_ruby : empty_menu),
    MenuLayoutEntry::separator ("help_group"),
    MenuLayoutEntry ("help_menu",         tl::translate ("&Help"),                 help_menu),
    MenuLayoutEntry ("@toolbar",          "",                                      empty_menu),
    MenuLayoutEntry::last ()
  };

  mp_menu = new AbstractMenu (this);
  mp_menu->init (main_menu);

  lay::LayoutView::init_menu (*mp_menu);

  //  make all plugins that return true on "implements_editable" into menu entries and
  //  set up the actions accordingly

  if (mp_menu->is_valid ("edit_menu.select_menu")) {

    unsigned int n_plugin = 0;
    int n = 1;
    for (tl::Registrar<lay::PluginDeclaration>::iterator cls = tl::Registrar<lay::PluginDeclaration>::begin (); cls != tl::Registrar<lay::PluginDeclaration>::end (); ++cls, ++n_plugin) {

      std::string title;
      if (cls->implements_editable (title)) {
        
        //  extract first part, which is the name, separated by a tab from the title.
        std::string name = tl::sprintf ("pi_enable_%d", n++);
        std::string t (title);
        const char *tab = strchr (t.c_str (), '\t');
        if (tab) {
          name = std::string (t.c_str (), 0, tab - t.c_str ());
          title = tab + 1;
        } 

        QAction *action = AbstractMenu::create_action (title);
        gtf::action_connect (action, SIGNAL (triggered ()), this, SLOT (enable ()));
        action->setData (QVariant (n_plugin));

        lay::Action handle (action);

        m_plugin_enabled.push_back (true);
        m_plugin_enable_actions.push_back (std::make_pair (n_plugin, handle));

        mp_menu->insert_item ("edit_menu.select_menu.end", name, handle);

      } else {
        m_plugin_enabled.push_back (false);
      }

    }

  }

  //  add all the custom menus from the plugins
  for (tl::Registrar<lay::PluginDeclaration>::iterator cls = tl::Registrar<lay::PluginDeclaration>::begin (); cls != tl::Registrar<lay::PluginDeclaration>::end (); ++cls) {

    std::vector<lay::MenuEntry> menu_entries;
    cls->get_menu_entries (menu_entries);

    for (std::vector<lay::MenuEntry>::const_iterator m = menu_entries.begin (); m != menu_entries.end (); ++m) {

      if (m->title.empty ()) {
        mp_menu->insert_separator (m->insert_pos, m->menu_name);
      } else {

        if (m->sub_menu) {
          mp_menu->insert_menu (m->insert_pos, m->menu_name, m->title);
        } else {

          QAction *action = AbstractMenu::create_action (m->title);
          action->setData (QVariant (QString (m->symbol.c_str ())));
          gtf::action_connect (action, SIGNAL (triggered ()), this, SLOT (generic_menu ()));
          mp_menu->insert_item (m->insert_pos, m->menu_name, action);

        }

      }

    }

  }

  //  Fill the mode menu file items from the intrinsic mouse modes followed by the 
  //  modes obtained from the plugins

  int n = 1;

  std::vector <std::string> mode_titles;
  lay::LayoutView::intrinsic_mouse_modes (&mode_titles);

  for (tl::Registrar<lay::PluginDeclaration>::iterator cls = tl::Registrar<lay::PluginDeclaration>::begin (); cls != tl::Registrar<lay::PluginDeclaration>::end (); ++cls) {
    std::string title;
    cls->implements_mouse_mode (title);
    mode_titles.push_back (title);
  }

  QActionGroup *mode_grp = new QActionGroup (this);
  mode_grp->setExclusive (true);

  unsigned int mode_index = 0;
  for (std::vector <std::string>::const_iterator t = mode_titles.begin (); t != mode_titles.end (); ++t, ++mode_index) {

    if (! t->empty ()) {

      //  extract first part, which is the name, separated by a tab from the title.
      std::string name = tl::sprintf ("mode_%d", n++);
      std::string title = *t;
      const char *tab = strchr (t->c_str (), '\t');
      if (tab) {
        name = std::string (t->c_str (), 0, tab - t->c_str ());
        title = tab + 1;
      } 

      QAction *action = AbstractMenu::create_action (title);
      mode_grp->addAction (action);

      action->setCheckable (true);
      action->setData (QVariant (mode_index));

      Action handle (action);
      m_mode_menu_actions.push_back (handle);

      mp_menu->insert_item ("edit_menu.mode_menu.end", name, handle);
      mp_menu->insert_item ("@toolbar.end", name, handle);

      gtf::action_connect (action, SIGNAL (triggered ()), this, SLOT (mode ()));

    }

  }

  //  if not in editable mode, hide all entries from "edit_mode" group
  if (! lay::Application::instance ()->is_editable ()) {
    std::vector<std::string> edit_mode_grp = mp_menu->group ("edit_mode");
    for (std::vector<std::string>::const_iterator g = edit_mode_grp.begin (); g != edit_mode_grp.end (); ++g) {
      mp_menu->action (*g).set_visible (false);
    }
  }
}

void
MainWindow::config_finalize ()
{
  //  make the changes visible in the setup form if the form is visible 
  mp_setup_form->setup ();
}

bool 
MainWindow::configure (const std::string &name, const std::string &value)
{
  if (name == cfg_grid) {

    double g;
    tl::from_string (value, g);
    m_grid_micron = g;
    return false; // not taken - let others set it too.

  } else if (name == cfg_synchronized_views) {

    bool flag;
    tl::from_string (value, flag);
    m_synchronized_views = flag;
    return true;

  } else if (name == cfg_default_dbu) {

    double d;
    tl::from_string (value, d);
    m_default_dbu = d;
    return false; // not taken - let others set it too.

  } else {
    return false;
  }

}

void 
MainWindow::enable_edits (bool enable)
{
  if (enable) {
    if (m_disabled_edits > 0) {
      --m_disabled_edits;
    }
    enable = (m_disabled_edits == 0);
  } else {
    ++m_disabled_edits;
  }

  std::vector<std::string> edit_grp = mp_menu->group ("edit");
  for (std::vector<std::string>::const_iterator g = edit_grp.begin (); g != edit_grp.end (); ++g) {
    mp_menu->action (*g).set_enabled (enable);
  }

  for (std::vector <lay::LayoutView *>::iterator vp = mp_views.begin (); vp != mp_views.end (); ++vp) {
    (*vp)->enable_edits (enable);
  }
  
  m_edits_enabled = enable;

  //  call this to establish a consistent set of options on the menu entries
  popup_menu_show ();
}

void
MainWindow::exit ()
{
  m_exited = true;
  QMainWindow::close ();
}

void
MainWindow::cm_exit ()
{
  BEGIN_PROTECTED

  int dirty_layouts = 0;
  std::string dirty_files;

  if (lay::Application::instance ()->is_editable ()) {
    std::vector <std::string> names;
    lay::LayoutHandle::get_names (names);
    for (std::vector <std::string>::const_iterator n = names.begin (); n != names.end (); ++n) {
      lay::LayoutHandle *handle = lay::LayoutHandle::find (*n);
      if (handle && handle->is_dirty ()) {
        ++dirty_layouts;
        if (! dirty_files.empty ()) {
          dirty_files += "\n";
        }
        dirty_files += handle->name ();
      }
    }
  }

  if (dirty_layouts == 0 || 
    QMessageBox::question (this, 
      tl::translate ("Save Needed").c_str (),
      (tl::translate ("The following layouts need saving.\nExit anyway and discard changes?\n\n") + dirty_files).c_str (),
      QMessageBox::Yes | QMessageBox::No,
      QMessageBox::No) == QMessageBox::Yes) {

    //  actually exit.
    exit ();

  }

  END_PROTECTED
}

lay::LayoutView *
MainWindow::view (int index)
{
  tl_assert (index >= 0 && index < int (mp_views.size ()));
  return mp_views [index];
}

const lay::LayoutView *
MainWindow::view (int index) const
{
  tl_assert (index >= 0 && index < int (mp_views.size ()));
  return mp_views [index];
}

int 
MainWindow::current_view_index () const
{
  if (m_current_view >= mp_views.size ()) {
    return -1;
  } else {
    return int (m_current_view);
  }
}

lay::LayoutView *
MainWindow::current_view ()
{
  if (m_current_view >= mp_views.size ()) {
    return 0;
  } else {
    return mp_views [m_current_view];
  }
}

const lay::LayoutView *
MainWindow::current_view () const
{
  if (m_current_view >= mp_views.size ()) {
    return 0;
  } else {
    return mp_views [m_current_view];
  }
}

void 
MainWindow::cm_show_properties ()
{
  if (current_view ()) {
    current_view ()->show_properties (this);
  }
}

void 
MainWindow::cm_delete ()
{
  BEGIN_PROTECTED

  if (current_view ()) {
    current_view ()->del ();
    //  because a "delete" might involve objects currently edited, we cancel the edit after we have deleted the object
    current_view ()->cancel ();
    current_view ()->clear_selection ();
  }

  END_PROTECTED
}

void 
MainWindow::cm_copy ()
{
  BEGIN_PROTECTED

  if (current_view () && current_view ()->has_selection ()) {
    current_view ()->copy ();
    current_view ()->clear_selection ();
  }

  END_PROTECTED
}

void 
MainWindow::cm_paste ()
{
  BEGIN_PROTECTED

  if (current_view () && ! db::Clipboard::instance ().empty ()) {
    current_view ()->cancel ();
    current_view ()->clear_selection ();
    current_view ()->paste ();
  }

  END_PROTECTED
}

void 
MainWindow::cm_cut ()
{
  BEGIN_PROTECTED

  if (current_view () && current_view ()->has_selection ()) {
    current_view ()->cut ();
    current_view ()->cancel (); //  see del() for reason why cancel is after cut
    current_view ()->clear_selection ();
  }

  END_PROTECTED
}

void 
MainWindow::mode ()
{
  QAction *action = dynamic_cast<QAction *> (sender ());
  if (action) {
    select_mode ((unsigned int) action->data ().toInt ());
  }
}

void
MainWindow::select_mode (unsigned int m)
{
  if (m_mode != m) {

    m_mode = m;
    for (std::vector <lay::LayoutView *>::iterator vp = mp_views.begin (); vp != mp_views.end (); ++vp) {
      (*vp)->mode (m);
    }

    //  update mode menu
    for (std::vector<lay::Action>::iterator i = m_mode_menu_actions.begin (); i != m_mode_menu_actions.end (); ++i) {
      i->set_checked (i->qaction ()->data ().toInt () == int (m_mode));
    }

  }
}

void 
MainWindow::enable ()
{
  BEGIN_PROTECTED

  QAction *action = dynamic_cast <QAction *> (sender ());
  tl_assert (action);

  int m = action->data ().toInt ();
  m_plugin_enabled [m] = ! m_plugin_enabled [m];
  for (std::vector <lay::LayoutView *>::iterator vp = mp_views.begin (); vp != mp_views.end (); ++vp) {
    (*vp)->enable_editable (m, m_plugin_enabled [m]);
  }

  END_PROTECTED
}

void 
MainWindow::cm_undo ()
{
  BEGIN_PROTECTED

  if (current_view () && m_manager.available_undo ().first) {
    for (std::vector <lay::LayoutView *>::iterator vp = mp_views.begin (); vp != mp_views.end (); ++vp) {
      (*vp)->clear_selection ();
      (*vp)->cancel ();
    }
    m_manager.undo ();
    current_view ()->update_content ();
  }

  END_PROTECTED
}

void
MainWindow::cm_redo ()
{
  BEGIN_PROTECTED

  if (current_view () && m_manager.available_redo ().first) {
    for (std::vector <lay::LayoutView *>::iterator vp = mp_views.begin (); vp != mp_views.end (); ++vp) {
      (*vp)->clear_selection ();
      (*vp)->cancel ();
    }
    m_manager.redo ();
    current_view ()->update_content ();
  }

  END_PROTECTED
}

void 
MainWindow::bookmark_menu_show ()
{
  if (mp_menu->is_valid ("bookmark_menu.goto_bookmark_menu")) {

    Action goto_bookmark_action = mp_menu->action ("bookmark_menu.goto_bookmark_menu");
    bool has_bookmarks = current_view () && current_view ()->bookmarks ().size () > 0;

    if (has_bookmarks && m_edits_enabled) {

      goto_bookmark_action.set_enabled (true);

      QMenu *goto_bookmark_menu = goto_bookmark_action.qaction ()->menu ();
      if (goto_bookmark_menu) {

        goto_bookmark_menu->clear ();

        if (current_view ()) {
          const lay::BookmarkList &bookmarks = current_view ()->bookmarks ();
          for (size_t i = 0; i < bookmarks.size (); ++i) {
            QAction *action = goto_bookmark_menu->addAction (bookmarks.name (i).c_str ());
            action->setObjectName (tl::sprintf ("bookmark_%d", i + 1).c_str ());
            gtf::action_connect (action, SIGNAL (triggered ()), this, SLOT (goto_bookmark ()));
            action->setData (QVariant (int (i))); 
          }
        }

      }

    } else {
      goto_bookmark_action.set_enabled (false);
    }

  }
}

void
MainWindow::goto_bookmark ()
{
  BEGIN_PROTECTED

  QAction *action = dynamic_cast <QAction *> (sender ());
  tl_assert (action);
  size_t id = size_t (action->data ().toInt ());
  if (current_view () && current_view ()->bookmarks ().size () > id) {
    current_view ()->goto_view (current_view ()->bookmarks ().state (id));
  }

  END_PROTECTED
}

void
MainWindow::cm_goto_position ()
{
  BEGIN_PROTECTED

  if (current_view ()) {

    while (true) {

      bool ok = false;

      db::DBox box (current_view ()->box ());
      std::string pos;
      pos += tl::micron_to_string (box.center ().x ()) + "," + tl::micron_to_string (box.center ().y ());
      pos += ",";
      pos += tl::micron_to_string (std::min (box.width (), box.height ()));

      QString text = QInputDialog::getText (tl::translate ("Enter Position").c_str (), tl::translate ("Enter position either as pair (x,y) or with window size (x,y,s)\n(x,y) will be the new window center position, s (if given) the new window size").c_str (),
                                            QLineEdit::Normal, QString (pos.c_str ()), &ok, this);

      if (! ok) {

        break;

      } else if (text.isEmpty ()) {

        throw tl::Exception (tl::translate ("Enter the position"));

      } else {

        double x = 0.0, y = 0.0, s = 0.0;
        tl::Extractor ex (text.ascii ());
        ex >> x >> "," >> y;
        if (! ex.at_end ()) {
          ex >> "," >> s >> tl::Extractor::end ();
          current_view ()->goto_window (db::DPoint (x, y), s);
        } else {
          current_view ()->goto_window (db::DPoint (x, y));
        }
        break;

      }

    }

  }

  END_PROTECTED
}

void
MainWindow::cm_manage_bookmarks ()
{
  BEGIN_PROTECTED

  if (current_view ()) {
    current_view ()->manage_bookmarks ();
  }

  END_PROTECTED
}

void
MainWindow::cm_bookmark_view ()
{
  BEGIN_PROTECTED

  if (current_view ()) {
    while (true) {
      bool ok = false;
      QString text = QInputDialog::getText (tl::translate ("Enter Bookmark Name").c_str (), tl::translate ("Bookmark name").c_str (),
                                            QLineEdit::Normal, QString::null, &ok, this);
      if (! ok) {
        break;
      } else if (text.isEmpty ()) {
        QMessageBox::critical (this, tl::translate ("Error").c_str (), tl::translate ("Enter a name for the bookmark").c_str ());
      } else {
        current_view ()->bookmark_view (text.ascii ());
        break;
      }
    }
  }

  END_PROTECTED
}

void
MainWindow::update_content ()
{
  BEGIN_PROTECTED

  mp_setup_form->setup ();
  if (current_view ()) {
    current_view ()->update_content ();
  }

  END_PROTECTED
}

void
MainWindow::cm_zoom_fit ()
{
  BEGIN_PROTECTED

  if (current_view ()) {
    current_view ()->zoom_fit ();
  }

  END_PROTECTED
}

void
MainWindow::cm_pan_left ()
{
  BEGIN_PROTECTED

  if (current_view ()) {
    current_view ()->pan_left ();
  }

  END_PROTECTED
}

void
MainWindow::cm_pan_right ()
{
  BEGIN_PROTECTED

  if (current_view ()) {
    current_view ()->pan_right ();
  }

  END_PROTECTED
}

void
MainWindow::cm_pan_up ()
{
  BEGIN_PROTECTED

  if (current_view ()) {
    current_view ()->pan_up ();
  }

  END_PROTECTED
}

void
MainWindow::cm_pan_down ()
{
  BEGIN_PROTECTED

  if (current_view ()) {
    current_view ()->pan_down ();
  }

  END_PROTECTED
}

void
MainWindow::cm_zoom_in ()
{
  BEGIN_PROTECTED

  if (current_view ()) {
    current_view ()->zoom_in ();
  }

  END_PROTECTED
}

void
MainWindow::cm_zoom_out ()
{
  BEGIN_PROTECTED

  if (current_view ()) {
    current_view ()->zoom_out ();
  }

  END_PROTECTED
}

void 
MainWindow::popup_menu_show ()
{
  BEGIN_PROTECTED

  if (mp_menu->is_valid ("edit_menu.undo")) {

    Action undo_action = mp_menu->action ("edit_menu.undo");

    std::string undo_txt (tl::translate ("&Undo"));
    bool undo_enable = false;
    if (current_view () && m_manager.available_undo ().first) {
      undo_txt += " - " + m_manager.available_undo ().second;
      undo_enable = true;
    }
    undo_action.set_title (undo_txt); 
    undo_action.set_enabled (undo_enable && m_edits_enabled);

  }

  if (mp_menu->is_valid ("edit_menu.redo")) {

    Action redo_action = mp_menu->action ("edit_menu.redo");

    std::string redo_txt (tl::translate ("&Redo"));
    bool redo_enable = false;
    if (current_view () && m_manager.available_redo ().first) {
      redo_txt += " - " + m_manager.available_redo ().second;
      redo_enable = true;
    }
    redo_action.set_title (redo_txt); 
    redo_action.set_enabled (redo_enable && m_edits_enabled);

  }

  if (mp_menu->is_valid ("edit_menu.copy")) {
    Action copy_action = mp_menu->action ("edit_menu.copy");
    copy_action.set_enabled (current_view () && current_view ()->has_selection () && m_edits_enabled);
  }

  if (mp_menu->is_valid ("edit_menu.cut")) {
    Action cut_action = mp_menu->action ("edit_menu.cut");
    cut_action.set_enabled (current_view () && current_view ()->has_selection () && m_edits_enabled);
  }

  if (mp_menu->is_valid ("edit_menu.paste")) {
    Action paste_action = mp_menu->action ("edit_menu.paste");
    paste_action.set_enabled (! db::Clipboard::instance ().empty () && m_edits_enabled);
  }

  if (mp_menu->is_valid ("zoom_menu.next_display_state")) {
    Action next_display_state_action = mp_menu->action ("zoom_menu.next_display_state");
    next_display_state_action.set_enabled (has_next_display_state ());
  } 

  if (mp_menu->is_valid ("zoom_menu.last_display_state")) {
    Action last_display_state_action = mp_menu->action ("zoom_menu.last_display_state");
    last_display_state_action.set_enabled (has_last_display_state ());
  }

  //  update mode menu
  for (std::vector< std::pair<unsigned int, lay::Action> >::iterator i = m_plugin_enable_actions.begin (); i != m_plugin_enable_actions.end (); ++i) {
    i->second.set_checkable (true);
    i->second.set_checked (m_plugin_enabled [i->first]);
  }

  END_PROTECTED
}

void
MainWindow::cm_redraw ()
{
  BEGIN_PROTECTED 
  redraw ();
  END_PROTECTED
}

void
MainWindow::redraw ()
{
  if (current_view ()) {
    current_view ()->redraw ();
  }
}

void
MainWindow::cm_cancel ()
{
  BEGIN_PROTECTED 
  cancel ();
  select_mode (lay::LayoutView::default_mode ());
  END_PROTECTED
}

void
MainWindow::cancel ()
{
  BEGIN_PROTECTED 

  for (std::vector <lay::LayoutView *>::iterator vp = mp_views.begin (); vp != mp_views.end (); ++vp) {
    (*vp)->cancel ();
  }

  END_PROTECTED
}

void
MainWindow::cm_save_layer_props ()
{
  BEGIN_PROTECTED

  if (current_view ()) {
    std::string fn;
    if (mp_lprops_fdia->get_save (fn)) {
      current_view ()->save_layer_props (fn);
    }
  } else {
    throw tl::Exception (tl::translate ("No view open to save the layer properties from"));
  }

  END_PROTECTED
}

void
MainWindow::load_layer_properties (const std::string &fn, bool all_views)
{
  if (all_views) {
    for (std::vector <lay::LayoutView *>::iterator vp = mp_views.begin (); vp != mp_views.end (); ++vp) {
      (*vp)->load_layer_props (fn);
    }
  } else if (current_view ()) {
    current_view ()->load_layer_props (fn);
  }
}

void
MainWindow::cm_load_layer_props ()
{
  BEGIN_PROTECTED

  if (current_view ()) {
    std::string fn;
    if (mp_lprops_fdia->get_open (fn)) {
      load_layer_properties (fn, false /*current view only*/);
    }
  } else {
    throw tl::Exception (tl::translate ("No view open to load the layer properties for"));
  }

  END_PROTECTED
}

void
MainWindow::cm_save_bookmarks ()
{
  BEGIN_PROTECTED

  if (current_view ()) {
    std::string fn;
    if (mp_bookmarks_fdia->get_save (fn)) {
      current_view ()->bookmarks ().save (fn);
    }
  } else {
    throw tl::Exception (tl::translate ("No view open to save the bookmarks from"));
  }

  END_PROTECTED
}

void
MainWindow::cm_load_bookmarks ()
{
  BEGIN_PROTECTED

  if (current_view ()) {
    std::string fn;
    if (mp_bookmarks_fdia->get_open (fn)) {
      BookmarkList bookmarks;
      bookmarks.load (fn);
      current_view ()->bookmarks (bookmarks);
    }
  } else {
    throw tl::Exception (tl::translate ("No view open to load the bookmarks for"));
  }

  END_PROTECTED
}

void
MainWindow::cm_select_current_cell ()
{
  BEGIN_PROTECTED

  lay::LayoutView *view = current_view ();
  if (view) {
    lay::LayoutView::cell_path_type path;
    int cvi = view->active_cellview_index ();
    view->current_cell_path (path);
    view->select_cell_fit (path, cvi);
  }

  END_PROTECTED
}

void
MainWindow::cm_open_current_cell ()
{
  BEGIN_PROTECTED

  if (current_view ()) {
    current_view ()->cm_open_current_cell ();
  }

  END_PROTECTED
}

void
MainWindow::cm_select_cell ()
{
  BEGIN_PROTECTED

  if (current_view ()) {

    CellSelectionFormImpl form (0, current_view (), "cell_selection_form");

    if (form.exec () == QDialog::Accepted && 
        form.selected_cellview_index () >= 0) {
      current_view ()->select_cell (form.selected_cellview ().combined_unspecific_path (), form.selected_cellview_index ()); 
      current_view ()->set_current_cell_path (form.selected_cellview_index (), form.selected_cellview ().combined_unspecific_path ());
      current_view ()->zoom_fit ();
    }

  } else {
    throw tl::Exception (tl::translate ("No view open to select a cell for"));
  }

  END_PROTECTED
}

void
MainWindow::cm_screenshot ()
{
  BEGIN_PROTECTED

  if (current_view ()) {
    std::string fn;
    if (mp_screenshot_fdia->get_save (fn)) {
      current_view ()->save_screenshot (fn);
    }
  } else {
    throw tl::Exception (tl::translate ("No view open to create a screenshot from"));
  }

  END_PROTECTED
}

void
MainWindow::cm_save_current_cell_as ()
{
  BEGIN_PROTECTED

  if (current_view ()) {

    int cv_index = current_view ()->active_cellview_index ();
    if (cv_index >= 0 && cv_index < int (current_view ()->cellviews ())) {

      LayoutView::cell_path_type path;
      current_view ()->current_cell_path (path);
      if (! path.empty ()) {

        std::string fn;
        if (mp_layout_fdia->get_save (fn)) {

          db::SaveLayoutOptions options (current_view ()->cellview (cv_index)->save_options ());
          options.set_dbu (current_view ()->cellview (cv_index)->layout ().dbu ());
          bool gzip = false;
          options.set_format_from_filename (fn, gzip);

          if (mp_layout_save_options->get_options (current_view (), cv_index, fn, gzip, options)) {
            options.add_cell (path.back ());
            current_view ()->save_as (cv_index, fn, gzip, options);
          }

        }

      }

    }

  }

  END_PROTECTED
}

void
MainWindow::cm_save ()
{
  do_save (false);
}

void
MainWindow::cm_save_as ()
{
  do_save (true);
}

void
MainWindow::do_save (bool as)
{
  BEGIN_PROTECTED

  if (current_view ()) {

    int cv_index = -1;
    if (current_view ()->cellviews () > 1) {
      SelectCellViewFormImpl form (0, current_view (), tl::translate ("Select Layout To Save"), true /*single selection*/);
      form.set_selection (current_view ()->active_cellview_index ());
      if (form.exec () == QDialog::Accepted) {
        cv_index = form.selected_cellview ();
      }
    } else if (current_view ()->cellviews () == 1) {
      cv_index = 0;
    }

    std::string fn;
    if (cv_index >= 0) {

      fn = current_view ()->cellview (cv_index)->filename ();
      if ((!as && !fn.empty ()) || mp_layout_fdia->get_save (fn)) {

        db::SaveLayoutOptions options (current_view ()->cellview (cv_index)->save_options ());
        options.set_dbu (current_view ()->cellview (cv_index)->layout ().dbu ());
        bool gzip = false;
        options.set_format_from_filename (fn, gzip);
        if (mp_layout_save_options->get_options (current_view (), cv_index, fn, gzip, options)) {
          current_view ()->save_as ((unsigned int) cv_index, fn, gzip, options);
        }

      }

    }

  } else {
    throw tl::Exception (tl::translate ("No view open to save"));
  }

  END_PROTECTED
}

void
MainWindow::cm_setup ()
{
  mp_setup_form->show ();
  mp_setup_form->setup ();
}

void
MainWindow::view_selected (int index)
{
  //  Hint: setting the focus to the tab bar avoids problem with dangling keyboard focus.
  //  Sometimes, the focus was set to the hierarchy level spin buttons which caught Copy&Paste
  //  events in effect.
  mp_tab_bar->setFocus ();

  if (! m_disable_tab_selected) {
    select_view (index);
  }
}

void
MainWindow::select_view (int index)
{
  bool dis = m_disable_tab_selected;
  m_disable_tab_selected = true; // prevent recursion

  tl_assert (index >= 0 && index < int (views ()));

  int view_index_org = current_view_index ();

  mp_tab_bar->setCurrentIndex (index);

  bool box_set = (m_synchronized_views && current_view () != 0);
  db::DBox box;
  if (box_set) {
    box = current_view ()->viewport ().box ();
  }

  m_current_view = index;

  if (current_view ()) {
    if (box_set) {
      current_view ()->zoom_box (box);
    }
    mp_view_stack->raiseWidget (current_view ());
    mp_setup_form->setup ();
  }

  m_current_view_observers.signal_observers ();
  m_current_view_observers.signal_observers (int (view_index_org));

  m_disable_tab_selected = dis;
}

void 
MainWindow::cm_open_too ()
{
  open (2);
}

void 
MainWindow::cm_open_new_view ()
{
  open (1);
}

void 
MainWindow::cm_open ()
{
  open (0);
}

void 
MainWindow::cm_new_layout ()
{
  BEGIN_PROTECTED

  lay::NewLayoutPropertiesDialog dialog (this);
  if (dialog.exec (m_new_cell_cell_name, m_default_dbu, m_new_cell_window_size)) {

    const lay::CellView &cellview = create_or_load_layout (0, 1 /*= new view*/);

    cellview->layout ().dbu (m_default_dbu);
    db::Layout::cell_index_type new_ci = cellview->layout ().add_cell (m_new_cell_cell_name.empty () ? 0 : m_new_cell_cell_name.c_str ());

    current_view ()->select_cell (new_ci, current_view ()->active_cellview_index ());
    current_view ()->zoom_box (db::DBox (-0.5 * m_new_cell_window_size, -0.5 * m_new_cell_window_size, 0.5 * m_new_cell_window_size, 0.5 * m_new_cell_window_size));
    current_view ()->set_hier_levels (std::make_pair (0, 1));

    //  make the current settings default for new calls of the dialog (and maybe for other
    //  components)
    config_set (cfg_default_dbu, m_default_dbu);
    config_finalize ();

  }

  END_PROTECTED
}

void 
MainWindow::cm_new_cell ()
{
  BEGIN_PROTECTED

  lay::LayoutView *curr = current_view ();
  if (! curr) {
    throw tl::Exception (tl::translate ("No view open to create a new cell inside"));
  }

  NewCellPropertiesDialog cell_prop_dia (this); 
  if (cell_prop_dia.exec (& curr->cellview (curr->active_cellview_index ())->layout (), m_new_cell_cell_name, m_new_cell_window_size)) {

    db::Layout::cell_index_type new_ci = curr->new_cell (curr->active_cellview_index (), m_new_cell_cell_name.c_str ());
    curr->select_cell (new_ci, curr->active_cellview_index ());
    curr->zoom_box (db::DBox (-0.5 * m_new_cell_window_size, -0.5 * m_new_cell_window_size, 0.5 * m_new_cell_window_size, 0.5 * m_new_cell_window_size));

    if (curr->get_max_hier_levels () < 1 || curr->get_min_hier_levels () > 0) {
      curr->set_hier_levels (std::make_pair (0, 1));
    }

  }

  END_PROTECTED
}

void 
MainWindow::cm_sel_move ()
{
  BEGIN_PROTECTED

  lay::LayoutView *curr = current_view ();
  if (! curr) {
    throw tl::Exception (tl::translate ("No view open for that operation"));
  }

  curr->cm_sel_move ();

  END_PROTECTED
}

void 
MainWindow::cm_sel_free_rot ()
{
  BEGIN_PROTECTED

  lay::LayoutView *curr = current_view ();
  if (! curr) {
    throw tl::Exception (tl::translate ("No view open for that operation"));
  }

  curr->cm_sel_free_rot ();

  END_PROTECTED
}

void 
MainWindow::cm_sel_rot_ccw ()
{
  BEGIN_PROTECTED

  lay::LayoutView *curr = current_view ();
  if (! curr) {
    throw tl::Exception (tl::translate ("No view open for that operation"));
  }

  curr->cm_sel_rot_ccw ();

  END_PROTECTED
}

void 
MainWindow::cm_sel_rot_cw ()
{
  BEGIN_PROTECTED

  lay::LayoutView *curr = current_view ();
  if (! curr) {
    throw tl::Exception (tl::translate ("No view open for that operation"));
  }

  curr->cm_sel_rot_cw ();

  END_PROTECTED
}

void 
MainWindow::cm_sel_flip_y ()
{
  BEGIN_PROTECTED

  lay::LayoutView *curr = current_view ();
  if (! curr) {
    throw tl::Exception (tl::translate ("No view open for that operation"));
  }

  curr->cm_sel_flip_y ();

  END_PROTECTED
}

void 
MainWindow::cm_sel_flip_x ()
{
  BEGIN_PROTECTED

  lay::LayoutView *curr = current_view ();
  if (! curr) {
    throw tl::Exception (tl::translate ("No view open for that operation"));
  }

  curr->cm_sel_flip_x ();

  END_PROTECTED
}

void 
MainWindow::cm_new_layer ()
{
  BEGIN_PROTECTED

  lay::LayoutView *curr = current_view ();
  if (! curr) {
    throw tl::Exception (tl::translate ("No view open to create a new layer inside"));
  }

  curr->cm_new_layer ();

  END_PROTECTED
}

void 
MainWindow::cm_layout_props ()
{
  BEGIN_PROTECTED

  lay::LayoutView *curr = current_view ();
  if (! curr) {
    throw tl::Exception (tl::translate ("No view open to show layout properties for"));
  }

  LayoutPropertiesFormImpl lp_form (this, curr, "layout_props_form");
  lp_form.exec ();

  END_PROTECTED
}

void 
MainWindow::cm_clone ()
{
  BEGIN_PROTECTED
  clone_current_view ();
  END_PROTECTED
}

void
MainWindow::clone_current_view ()
{
  lay::LayoutView *view = 0;
  lay::LayoutView *curr = current_view ();
  if (! curr) {
    throw tl::Exception (tl::translate ("No view open to clone"));
  }

  //  create a new view
  view = new lay::LayoutView (current_view (), &m_manager, this, mp_view_stack);
  connect (view, SIGNAL (title_changed ()), this, SLOT (view_title_changed ()));
  mp_views.push_back (view);

  //  we must resize the widget here to set the geometry properly. 
  //  This is required to make zoom_fit work.
  view->setGeometry (0, 0, mp_view_stack->width (), mp_view_stack->height ());
  view->show ();

  //  set initial attributes
  view->set_hier_levels (get_hier_levels ());

  //  select the current mode and select the enabled editables
  view->mode (m_mode);
  for (std::vector< std::pair<unsigned int, lay::Action> >::const_iterator i = m_plugin_enable_actions.begin (); i != m_plugin_enable_actions.end (); ++i) {
    view->enable_editable (i->first, m_plugin_enabled [i->first]);
  }

  lay::DisplayState state;
  current_view ()->save_view (state);
  view->goto_view (state);

  view->update_content ();

  int view_index_org = current_view_index ();

  m_current_view = mp_views.size () - 1;
  mp_view_stack->addWidget (view);
  int index = mp_tab_bar->insertTab (-1, view->title ().c_str ());
  mp_tab_bar->setCurrentIndex (index);

  m_new_view_observers.signal_observers ();
  m_current_view_observers.signal_observers ();
  m_current_view_observers.signal_observers (int (view_index_org));
}

void 
MainWindow::cm_close ()
{
  if (current_view ()) {

    std::vector <int> selected;

    if (current_view ()->cellviews () > 1) {

      SelectCellViewFormImpl form (0, current_view (), tl::translate ("Select Layouts To Close"));
      form.set_selection (current_view ()->active_cellview_index ());
    
      if (form.exec () == QDialog::Accepted) {
        selected = form.selected_cellviews ();
      }
        
    } else if (current_view ()->cellviews () > 0) {
      selected.push_back (0);
    }

    if (selected.size () > 0) {

      int dirty_layouts = 0;
      std::string dirty_files;

      if (lay::Application::instance ()->is_editable ()) {
        for (std::vector <int>::const_iterator i = selected.begin (); i != selected.end (); ++i) {
          if (current_view ()->cellview (*i)->is_dirty ()) {
            ++dirty_layouts;
            if (! dirty_files.empty ()) {
              dirty_files += "\n";
            }
            dirty_files += current_view ()->cellview (*i)->name ();
          }
        }
      }

      if (dirty_layouts == 0 || 
        QMessageBox::question (this, 
          tl::translate ("Save Needed").c_str (),
          (tl::translate ("The following layouts need saving.\nClose them anyway and discard changes?\n\n") + dirty_files).c_str (),
          QMessageBox::Yes | QMessageBox::No,
          QMessageBox::No) == QMessageBox::Yes) {

        BEGIN_PROTECTED

        //  Actually erase the selected cellviews
        if (current_view ()->cellviews () == selected.size ()) {
          //  all cellviews selected - simply close
          close_current_view ();
        } else {
          std::sort (selected.begin (), selected.end ());
          int offset = 0;
          for (std::vector <int>::const_iterator i = selected.begin (); i != selected.end (); ++i) {
            current_view ()->erase_cellview ((unsigned int)(*i - offset));
            ++offset;
          }
        }

        END_PROTECTED

      }

    }

  }

}

void 
MainWindow::close_current_view ()
{
  if (current_view ()) {

    BEGIN_PROTECTED

    //  this suppresses view_selected events that would otherwise be created
    m_disable_tab_selected = true;

    bool box_set = false;
    db::DBox box;
    if (m_synchronized_views) {
      box = current_view ()->viewport ().box ();
      box_set = true;
    }

    mp_tab_bar->removeTab (m_current_view);
    mp_view_stack->removeWidget (current_view ());
    delete current_view ();
    mp_views.erase (mp_views.begin () + m_current_view, mp_views.begin () + m_current_view + 1);

    int view_index_org = current_view_index ();

    if (m_current_view >= mp_views.size () && m_current_view > 0) {
      --m_current_view;
    }

    if (current_view ()) {
      if (box_set) {
        current_view ()->zoom_box (box);
      }
      mp_tab_bar->setCurrentTab (m_current_view);
      mp_view_stack->raiseWidget (current_view ());
    }

    m_disable_tab_selected = false;

    m_current_view_observers.signal_observers ();
    m_current_view_observers.signal_observers (int (view_index_org));

    END_PROTECTED

  }
}

void
MainWindow::cm_reload ()
{    
  BEGIN_PROTECTED 

  if (current_view ()) {

    std::vector <int> selected;

    if (current_view ()->cellviews () > 1) {

      SelectCellViewFormImpl form (0, current_view (), tl::translate ("Select Layouts To Reload"));
      form.select_all ();
    
      if (form.exec () == QDialog::Accepted) {
        selected = form.selected_cellviews ();
      }
        
    } else if (current_view ()->cellviews () > 0) {
      selected.push_back (0);
    }

    if (selected.size () > 0) {

      int dirty_layouts = 0;
      std::string dirty_files;

      if (lay::Application::instance ()->is_editable ()) {
        for (std::vector <int>::const_iterator i = selected.begin (); i != selected.end (); ++i) {
          if (current_view ()->cellview (*i)->is_dirty ()) {
            ++dirty_layouts;
            if (! dirty_files.empty ()) {
              dirty_files += "\n";
            }
            dirty_files += current_view ()->cellview (*i)->name ();
          }
        }
      }

      if (dirty_layouts == 0 || 
        QMessageBox::question (this, 
          tl::translate ("Save Needed").c_str (),
          (tl::translate ("The following layouts need saving.\nReload them anyway and discard changes?\n\n") + dirty_files).c_str (),
          QMessageBox::Yes | QMessageBox::No,
          QMessageBox::No) == QMessageBox::Yes) {

        //  Actually reload
        for (std::vector <int>::const_iterator i = selected.begin (); i != selected.end (); ++i) {
          reload_layout (*i);
        }

      }

    }

  }

  END_PROTECTED
}

void 
MainWindow::open (int mode)
{
  BEGIN_PROTECTED 

  static std::string fn;
  if (mp_layout_fdia->get_open (fn)) {
    load_layout (fn, mode);
  }

  END_PROTECTED
}

void
MainWindow::reload_layout (unsigned int cv_index)
{
  lay::LayoutView *view = current_view ();

  if (view && view->cellviews () > cv_index) {
    view->reload_layout (cv_index);
  }
}

const lay::CellView &
MainWindow::load_layout (const std::string &filename, int mode)
{
  return create_or_load_layout (&filename, mode);
}

const lay::CellView &
MainWindow::create_layout (int mode)
{
  return create_or_load_layout (0, mode);
}

const lay::CellView & 
MainWindow::create_or_load_layout (const std::string *filename, int mode)
{
  lay::LayoutView *view = 0;

  int view_index_org = current_view_index ();

  if (! current_view ()) {
    mode = 1;
  }

  if (mode == 1) {

    //  create a new view
    view = new lay::LayoutView (&m_manager, this, mp_view_stack);
    connect (view, SIGNAL (title_changed ()), this, SLOT (view_title_changed ()));

    mp_views.push_back (view);

    //  we must resize the widget here to set the geometry properly. 
    //  This is required to make zoom_fit work.
    view->setGeometry (0, 0, mp_view_stack->width (), mp_view_stack->height ());
    view->show ();

    //  set initial attributes
    view->set_synchroneous (synchroneous ());
    view->set_hier_levels (get_hier_levels ());

    //  select the current mode and select the enabled editables
    view->mode (m_mode);
    for (std::vector< std::pair<unsigned int, lay::Action> >::const_iterator i = m_plugin_enable_actions.begin (); i != m_plugin_enable_actions.end (); ++i) {
      view->enable_editable (i->first, m_plugin_enabled [i->first]);
    }

  } else {
    view = current_view ();
  }

  unsigned int cv_index = 0; 

  try {

    //  load or create the layout
    if (filename != 0) {
      cv_index = view->load_layout (*filename, mode == 2);
    } else {
      cv_index = view->create_layout (mode == 2);
    }

    //  make the new view the current one
    if (mode == 1) {
      m_current_view = mp_views.size () - 1;
      mp_view_stack->addWidget (mp_views.back ());
      int index = mp_tab_bar->insertTab (-1, current_view ()->title ().c_str ());
      mp_tab_bar->setCurrentIndex (index);
    } else if (mode == 0) {
      mp_tab_bar->setTabText (m_current_view, current_view ()->title ().c_str ());
    } else if (mode == 2) {
      mp_tab_bar->setTabText (m_current_view, current_view ()->title ().c_str ());
    }

    m_new_view_observers.signal_observers ();
    m_current_view_observers.signal_observers ();
    m_current_view_observers.signal_observers (int (view_index_org));

  } catch (...) {

    //  clean up in case of an error ..
    if (mode == 1) {
      delete mp_views.back ();
      mp_views.pop_back ();
    }

    throw;

  }

  return view->cellview (cv_index);
}

void 
MainWindow::view_title_changed ()
{
  //  transfer the current view's text into the tab bar
  unsigned int tabs = mp_tab_bar->count ();
  for (unsigned int i = 0; i < tabs && i < views (); ++i) {
    mp_tab_bar->setTabText (i, view (i)->title ().c_str ());
  }
}

double
MainWindow::grid_micron () const
{
  return m_grid_micron;
}

void 
MainWindow::cm_inc_max_hier ()
{
  BEGIN_PROTECTED 

  int new_to = get_max_hier_levels () + 1;
  set_hier_levels (std::make_pair (get_min_hier_levels (), new_to));

  END_PROTECTED
}

void 
MainWindow::cm_dec_max_hier ()
{
  BEGIN_PROTECTED 

  int new_to = get_max_hier_levels () > 0 ? get_max_hier_levels () - 1 : 0;
  set_hier_levels (std::make_pair (std::min (get_min_hier_levels (), new_to), new_to));

  END_PROTECTED
}

void 
MainWindow::cm_max_hier ()
{
  BEGIN_PROTECTED 

  if (current_view ()) {
    current_view ()->max_hier ();
  }

  END_PROTECTED
}

void 
MainWindow::cm_max_hier_0 ()
{
  BEGIN_PROTECTED 
  set_hier_levels (std::make_pair (std::min (get_min_hier_levels (), 0), 0));
  END_PROTECTED
}

void 
MainWindow::cm_max_hier_1 ()
{
  BEGIN_PROTECTED 
  set_hier_levels (std::make_pair (std::min (get_min_hier_levels (), 0), 1));
  END_PROTECTED
}

void 
MainWindow::set_hier_levels (std::pair<int, int> l)
{
  if (l != get_hier_levels ()) {
    if (current_view ()) {
      current_view ()->set_hier_levels (l);
    }
  } 
}

std::pair<int, int> 
MainWindow::get_hier_levels () const
{
  if (current_view ()) {
    return current_view ()->get_hier_levels ();
  } else {
    return std::make_pair (0, 0);
  }
}

void 
MainWindow::cm_last_display_state ()
{
  BEGIN_PROTECTED 

  if (has_last_display_state ()) {
    current_view ()->last_display_state ();
  }

  END_PROTECTED
}

bool 
MainWindow::has_last_display_state ()
{
  if (current_view ()) {
    return current_view ()->has_last_display_state ();
  } else {
    return false;
  }
}

void 
MainWindow::cm_next_display_state ()
{
  BEGIN_PROTECTED 

  if (has_next_display_state ()) {
    current_view ()->next_display_state ();
  }

  END_PROTECTED
}

bool 
MainWindow::has_next_display_state ()
{
  if (current_view ()) {
    return current_view ()->has_next_display_state ();
  } else {
    return false;
  }
}

void 
MainWindow::set_synchroneous (bool sync_mode)
{
  m_synchroneous = sync_mode;
  for (std::vector <lay::LayoutView *>::iterator vp = mp_views.begin (); vp != mp_views.end (); ++vp) {
    (*vp)->set_synchroneous (sync_mode);
  }
}

void
MainWindow::current_pos (double x, double y, bool dbu_units) 
{
  mp_cpx_label->setText (dbu_units ? tl::db_to_string (x).c_str () : tl::micron_to_string (x).c_str ());
  mp_cpy_label->setText (dbu_units ? tl::db_to_string (y).c_str () : tl::micron_to_string (y).c_str ());
}

void
MainWindow::clear_current_pos ()
{
  mp_cpx_label->setText ("");
  mp_cpy_label->setText ("");
}

void 
MainWindow::set_progress_text (const std::string &text)
{
  if (mp_progress_widget) {
    mp_progress_widget->set_text (text);
  }
}

void 
MainWindow::set_progress_value (double v, const std::string &value)
{
  if (mp_progress_widget) {
    mp_progress_widget->set_value (v, value);
  }
}

bool 
MainWindow::eventFilter (QObject *watched, QEvent *event)
{
  if (event->type () == QEvent::MouseButtonDblClick ||
      event->type () == QEvent::MouseButtonPress ||
      event->type () == QEvent::MouseButtonRelease ||
      event->type () == QEvent::MouseMove ||
      event->type () == QEvent::KeyPress ||
      event->type () == QEvent::KeyRelease) {

    QObject *o = watched;
    while (o) {
      //  if the watched object is a child of the progress widget, pass the event on to this ..
      if (o == mp_progress_widget) {
        return QObject::eventFilter(watched, event);
      }
      o = o->parent ();
    }
    return true;

  } else {
    // standard event processing
    return QObject::eventFilter(watched, event);
  }
}

void 
MainWindow::show_progress_bar (bool show)
{
  if (m_work_in_progress != show) {
    m_work_in_progress = show;
    if (show) {
      //  to avoid recursions of any kind, disallow any user interaction except 
      //  cancelling the operation
      mp_app->installEventFilter (this);
      QApplication::setOverrideCursor (QCursor (Qt::WaitCursor));
    } else {
      mp_app->removeEventFilter (this);
      QApplication::restoreOverrideCursor ();
    }
    //  HINT: enabling the scheduler is accumulative - make sure that this method is only called when necessary
    tl::DeferredMethodScheduler::instance ()->enable (! show);
  }
  mp_right_stack->setCurrentIndex (show ? 1 : 0);
}

//  Writing and Reading of configuration 

struct ConfigGetAdaptor
{
  ConfigGetAdaptor (const std::string &name)
    : mp_owner (0), m_name (name)
  {
    // .. nothing yet ..
  }
  
  std::string operator () () const
  {
    std::string s;
    mp_owner->config_get (m_name, s);
    return s;
  }

  bool at_end () const 
  {
    return m_done;
  }

  void start (const lay::MainWindow &owner) 
  {
    mp_owner = &owner;
    m_done = false;
  }

  void next () 
  {
    m_done = true;
  }

private:
  const lay::MainWindow *mp_owner;
  bool m_done;
  std::string m_name;
};
 
struct ConfigSetAdaptor
{
  ConfigSetAdaptor (const std::string &name)
    : m_name (name)
  {
    // .. nothing yet ..
  }
  
  void operator () (lay::MainWindow &w, const std::string &v) const
  {
    w.config_set (m_name, v);
  }

private:
  std::string m_name;
};
 
//  the configuration file's XML structure is built dynamically
static tl::XMLStruct<lay::MainWindow> 
config_structure ()
{
  tl::XMLElementList body;

  std::vector <std::string> names; 
  lay::Plugin::get_config_names (names);
  for (std::vector <std::string>::const_iterator n = names.begin (); n != names.end (); ++n) {
    body.append (tl::XMLMember<std::string, lay::MainWindow, ConfigGetAdaptor, ConfigSetAdaptor, tl::XMLStdConverter <std::string> > ( 
                       ConfigGetAdaptor (*n), ConfigSetAdaptor (*n), *n));
  }

  //  add a wildcard member to ignore all others
  body.append (tl::make_member<lay::MainWindow> ("*"));

  return tl::XMLStruct<lay::MainWindow> ("config", body);
}


bool 
MainWindow::write_config (const std::string &config_file)
{
  std::ofstream out (config_file.c_str ());
  if (! out.good ()) {
    return false;
  }
  config_structure ().write (out, *this); 
  return true;
}

bool
MainWindow::read_config (const std::string &config_file)
{
  bool ret = false;

  try {
    tl::XMLFileSource file (config_file);
    try {
      //  clear the custom dither pattern
      config_structure ().parse (file, *this); 
      ret = true;
    } catch (tl::XMLException &ex) {
      tl::error << tl::translate ("Problem reading config file:") << ex.msg ();
    } 
  } catch (tl::Exception &) {
    // .. ignore errors ..
  }

  config_end ();
  return ret;
}

void
MainWindow::cm_console ()
{
  mp_console->show ();
}

void
MainWindow::cm_help_about ()
{
  HelpAboutDialog help_about_dialog (this);
  help_about_dialog.exec ();
}

void 
MainWindow::generic_menu ()
{
  QAction *action = dynamic_cast <QAction *> (sender ());
  tl_assert (action);
  generic_menu (std::string (action->data ().toString ().toAscii ()));
}

void 
MainWindow::generic_menu (const std::string &symbol)
{
  BEGIN_PROTECTED 

  if (current_view ()) {
    current_view ()->menu_activated (symbol);
  } else {
    throw tl::Exception (tl::translate ("No view is active"));
  }

  END_PROTECTED
}

void
MainWindow::menu_changed ()
{
  //  delay actual rebuilding of the menu to collect multiple change events.
  m_do_update_menu_dm ();
}

void
MainWindow::do_update_menu ()
{
  mp_menu->build (menuBar (), mp_tool_bar);
}

void
MainWindow::cm_cell_delete ()
{
  BEGIN_PROTECTED 

  if (current_view ()) {
    current_view ()->cm_cell_delete ();
  }

  END_PROTECTED
}

void
MainWindow::cm_cell_rename ()
{
  BEGIN_PROTECTED 

  if (current_view ()) {
    current_view ()->cm_cell_rename ();
  }

  END_PROTECTED
}

void
MainWindow::cm_cell_select ()
{
  BEGIN_PROTECTED 

  if (current_view ()) {
    current_view ()->cm_cell_select ();
  }

  END_PROTECTED
}

void
MainWindow::cm_cell_hide ()
{
  BEGIN_PROTECTED 

  if (current_view ()) {
    current_view ()->cm_cell_hide ();
  }

  END_PROTECTED
}

void
MainWindow::cm_cell_show ()
{
  BEGIN_PROTECTED 

  if (current_view ()) {
    current_view ()->cm_cell_show ();
  }

  END_PROTECTED
}

void
MainWindow::cm_cell_show_all ()
{
  BEGIN_PROTECTED 

  if (current_view ()) {
    current_view ()->cm_cell_show_all ();
  }

  END_PROTECTED
}

void
MainWindow::cm_lv_select_all ()
{
  BEGIN_PROTECTED 

  if (current_view ()) {
    current_view ()->cm_select_all ();
  }

  END_PROTECTED
}

void
MainWindow::cm_lv_hide ()
{
  BEGIN_PROTECTED 

  if (current_view ()) {
    current_view ()->cm_hide ();
  }

  END_PROTECTED
}

void
MainWindow::cm_lv_show ()
{
  BEGIN_PROTECTED 

  if (current_view ()) {
    current_view ()->cm_show ();
  }

  END_PROTECTED
}

void
MainWindow::cm_lv_rename ()
{
  BEGIN_PROTECTED 

  if (current_view ()) {
    current_view ()->cm_rename ();
  }

  END_PROTECTED
}

void
MainWindow::cm_lv_delete ()
{
  BEGIN_PROTECTED 

  if (current_view ()) {
    current_view ()->cm_delete ();
  }

  END_PROTECTED
}

void
MainWindow::cm_lv_insert ()
{
  BEGIN_PROTECTED 

  if (current_view ()) {
    current_view ()->cm_insert ();
  }

  END_PROTECTED
}

void
MainWindow::cm_lv_group ()
{
  BEGIN_PROTECTED 

  if (current_view ()) {
    current_view ()->cm_group ();
  }

  END_PROTECTED
}

void
MainWindow::cm_lv_ungroup ()
{
  BEGIN_PROTECTED 

  if (current_view ()) {
    current_view ()->cm_ungroup ();
  }

  END_PROTECTED
}

void
MainWindow::cm_lv_source ()
{
  BEGIN_PROTECTED 

  if (current_view ()) {
    current_view ()->cm_source ();
  }

  END_PROTECTED
}

void
MainWindow::cm_lv_sort_by_name ()
{
  BEGIN_PROTECTED 

  if (current_view ()) {
    current_view ()->cm_sort_by_name ();
  }

  END_PROTECTED
}

void
MainWindow::cm_lv_sort_by_ild ()
{
  BEGIN_PROTECTED 

  if (current_view ()) {
    current_view ()->cm_sort_by_ild ();
  }

  END_PROTECTED
}

void
MainWindow::cm_lv_sort_by_idl ()
{
  BEGIN_PROTECTED 

  if (current_view ()) {
    current_view ()->cm_sort_by_idl ();
  }

  END_PROTECTED
}

void
MainWindow::cm_lv_sort_by_ldi ()
{
  BEGIN_PROTECTED 

  if (current_view ()) {
    current_view ()->cm_sort_by_ldi ();
  }

  END_PROTECTED
}

void
MainWindow::cm_lv_sort_by_dli ()
{
  BEGIN_PROTECTED 

  if (current_view ()) {
    current_view ()->cm_sort_by_dli ();
  }

  END_PROTECTED
}

void
MainWindow::cm_lv_regroup_by_index ()
{
  BEGIN_PROTECTED 

  if (current_view ()) {
    current_view ()->cm_regroup_by_index ();
  }

  END_PROTECTED
}

void
MainWindow::cm_lv_regroup_by_datatype ()
{
  BEGIN_PROTECTED 

  if (current_view ()) {
    current_view ()->cm_regroup_by_datatype ();
  }

  END_PROTECTED
}

void
MainWindow::cm_lv_regroup_by_layer ()
{
  BEGIN_PROTECTED 

  if (current_view ()) {
    current_view ()->cm_regroup_by_layer ();
  }

  END_PROTECTED
}

void
MainWindow::cm_lv_regroup_flatten ()
{
  BEGIN_PROTECTED 

  if (current_view ()) {
    current_view ()->cm_regroup_flatten ();
  }

  END_PROTECTED
}

void
MainWindow::cm_lv_expand_all ()
{
  BEGIN_PROTECTED 

  if (current_view ()) {
    current_view ()->cm_expand_all ();
  }

  END_PROTECTED
}

void
MainWindow::cm_lv_add_missing ()
{
  BEGIN_PROTECTED 

  if (current_view ()) {
    current_view ()->cm_add_missing ();
  }

  END_PROTECTED
}

void
MainWindow::cm_lv_remove_unused ()
{
  BEGIN_PROTECTED 

  if (current_view ()) {
    current_view ()->cm_remove_unused ();
  }

  END_PROTECTED
}

// ------------------------------------------------------------
//  Implementation of the "help about" dialog

HelpAboutDialog::HelpAboutDialog (QWidget *parent)
  : QDialog (parent)
{
  mp_ui = new Ui::HelpAboutDialog ();
  mp_ui->setupUi (this);

  std::string s = std::string (lay::Version::name ()) + " " + lay::Version::version () +
                  "\n\n" + lay::Version::about_text ();
  std::string t = tl::translate ("About ") + lay::Version::name ();

  setWindowTitle (t.c_str ());
  mp_ui->main->setText (s.c_str ());
}

HelpAboutDialog::~HelpAboutDialog ()
{
  delete mp_ui;
  mp_ui = 0;
}

} // namespace lay

