
/*

  KLayout Layout Viewer
  Copyright (C) 2006-2007 Matthias Koefferlein

  This program is free software; you can redistribute it and/or modify
  it under the terms of the GNU General Public License as published by
  the Free Software Foundation; either version 2 of the License, or
  (at your option) any later version.

  This program is distributed in the hope that it will be useful,
  but WITHOUT ANY WARRANTY; without even the implied warranty of
  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
  GNU General Public License for more details.

  You should have received a copy of the GNU General Public License
  along with this program; if not, write to the Free Software
  Foundation, Inc., 51 Franklin St, Fifth Floor, Boston, MA  02110-1301  USA

*/


#include "dbCellGraph.h"
#include "layMarker.h"
#include "laySnap.h"
#include "layBitmap.h"
#include "layViewOp.h"
#include "layRenderer.h"
#include "layLayoutView.h"
#include "tlAssert.h"

namespace lay
{

// ------------------------------------------------------------------------

MarkerBase::MarkerBase (lay::LayoutView *view)
  : lay::ViewObject (view->view_object_widget ()),
    m_line_width (-1), m_vertex_size (-1), m_halo (-1), m_dither_pattern (-1), m_frame_pattern (0), mp_view (view)
{ 
  // .. nothing yet ..
}

void 
MarkerBase::set_color (QColor color)
{
  if (color != m_color) {
    m_color = color;
    redraw ();
  }
}

void 
MarkerBase::set_line_width (int lw)
{
  if (m_line_width != lw) {
    m_line_width = lw;
    redraw ();
  }
}

void 
MarkerBase::set_vertex_size (int vs)
{
  if (m_vertex_size != vs) {
    m_vertex_size = vs;
    redraw ();
  }
}

void 
MarkerBase::set_halo (int halo)
{
  if (m_halo != halo) {
    m_halo = halo;
    redraw ();
  }
}

void 
MarkerBase::set_frame_pattern (int frame_pattern)
{
  if (m_frame_pattern != frame_pattern) {
    m_frame_pattern = frame_pattern;
    redraw ();
  }
}

void 
MarkerBase::set_dither_pattern (int dither_pattern)
{
  if (m_dither_pattern != dither_pattern) {
    m_dither_pattern = dither_pattern;
    redraw ();
  }
}

void 
MarkerBase::get_bitmaps (const Viewport &vp, ViewObjectCanvas &canvas, lay::Bitmap *&fill, lay::Bitmap *&contour, lay::Bitmap *&vertex, lay::Bitmap *&text)
{
  //  obtain bitmaps
  QColor color = m_color;
  if (! color.isValid ()) {
    color = mp_view->default_marker_color ();
  }
  if (! color.isValid ()) {
    color = canvas.foreground_color ();
  }

  int line_width = m_line_width < 0 ? mp_view->default_marker_line_width () : m_line_width;
  int vertex_size = m_vertex_size < 0 ? mp_view->default_marker_vertex_size () : m_vertex_size;
  bool halo = m_halo < 0 ? mp_view->default_marker_halo () : (m_halo != 0);
  
  if (halo) {

    std::vector <lay::ViewOp> ops;
    ops.resize (2);

    if (m_dither_pattern >= 0) {
      ops[0] = lay::ViewOp (canvas.background_color ().rgb (), lay::ViewOp::Copy, (unsigned int) m_dither_pattern, lay::ViewOp::Rect, 3, 0);
      ops[1] = lay::ViewOp (color.rgb (), lay::ViewOp::Copy, (unsigned int) m_dither_pattern, lay::ViewOp::Rect, line_width, 1);
      fill = canvas.bitmap (ops, vp.width (), vp.height ());
    } else {
      fill = 0;
    }

    ops[0] = lay::ViewOp (canvas.background_color ().rgb (), lay::ViewOp::Copy, (unsigned int) m_frame_pattern, lay::ViewOp::Rect, line_width > 0 ? line_width + 2 : 0, 0);
    ops[1] = lay::ViewOp (color.rgb (), lay::ViewOp::Copy, (unsigned int) m_frame_pattern, lay::ViewOp::Rect, line_width, 1);
    contour = canvas.bitmap (ops, vp.width (), vp.height ());

    ops[0] = lay::ViewOp (canvas.background_color ().rgb (), lay::ViewOp::Copy, 0, lay::ViewOp::Rect, vertex_size > 0 ? vertex_size + 2 : 0, 0);
    ops[1] = lay::ViewOp (color.rgb (), lay::ViewOp::Copy, 0, lay::ViewOp::Rect, vertex_size, 1);
    vertex = canvas.bitmap (ops, vp.width (), vp.height ());

    if (line_width == 1) {
      text = contour;
    } else {
      ops[0] = lay::ViewOp (canvas.background_color ().rgb (), lay::ViewOp::Copy, 0, lay::ViewOp::Rect, 3, 0);
      ops[1] = lay::ViewOp (color.rgb (), lay::ViewOp::Copy, 0, lay::ViewOp::Rect, 1, 1);
      text = canvas.bitmap (ops, vp.width (), vp.height ());
    }

  } else {

    if (m_dither_pattern >= 0) {
      fill = canvas.bitmap (lay::ViewOp (color.rgb (), lay::ViewOp::Copy, (unsigned int) m_dither_pattern, lay::ViewOp::Rect, line_width), vp.width (), vp.height ());
    } else {
      fill = 0;
    }

    contour = canvas.bitmap (lay::ViewOp (color.rgb (), lay::ViewOp::Copy, (unsigned int) m_frame_pattern, lay::ViewOp::Rect, line_width), vp.width (), vp.height ());
    vertex = canvas.bitmap (lay::ViewOp (color.rgb (), lay::ViewOp::Copy, 0, lay::ViewOp::Rect, vertex_size), vp.width (), vp.height ());
    if (line_width == 1) {
      text = contour;
    } else {
      text = canvas.bitmap (lay::ViewOp (color.rgb (), lay::ViewOp::Copy, 0, lay::ViewOp::Rect, 1), vp.width (), vp.height ());
    }

  }
}

// ------------------------------------------------------------------------

GenericMarkerBase::GenericMarkerBase (lay::LayoutView *view, unsigned int cv_index)
  : MarkerBase (view), mp_trans_vector (0), mp_view (view), m_cv_index (cv_index)
{ 
  // .. nothing yet ..
}

GenericMarkerBase::~GenericMarkerBase ()
{
  if (mp_trans_vector) {
    delete mp_trans_vector;
    mp_trans_vector = 0;
  }
}

void 
GenericMarkerBase::set (const db::CplxTrans &trans)
{
  if (mp_trans_vector) {
    delete mp_trans_vector;
    mp_trans_vector = 0;
  }
  m_trans = trans;
  redraw ();
}

void 
GenericMarkerBase::set (const db::CplxTrans &t1, const std::vector<db::CplxTrans> &trans)
{
  if (trans.size () == 1) {
    set (trans [0] * t1);
  } else {
    m_trans = t1;
    if (mp_trans_vector) {
      delete mp_trans_vector;
    }
    mp_trans_vector = new std::vector<db::CplxTrans> (trans);
    redraw ();
  }
}

// ------------------------------------------------------------------------

InstanceMarker::InstanceMarker (lay::LayoutView *view, unsigned int cv_index)
  : GenericMarkerBase (view, cv_index), m_inst ()
{ 
  // .. nothing yet ..
}

InstanceMarker::~InstanceMarker ()
{
  // .. nothing yet ..
}

void 
InstanceMarker::set (const db::Instance &instance, const db::CplxTrans &trans)
{
  m_inst = instance;
  GenericMarkerBase::set (trans);
}

void 
InstanceMarker::set (const db::Instance &instance, const db::CplxTrans &t1, const std::vector<db::CplxTrans> &trans)
{
  m_inst = instance;
  GenericMarkerBase::set (t1, trans);
}

void 
InstanceMarker::render (const Viewport &vp, ViewObjectCanvas &canvas)
{
  if (m_cv_index >= (unsigned int) (mp_view->cellviews ())) { 
    return;
  }
  const lay::CellView &cv = mp_view->cellview (m_cv_index);
  if (! cv.is_valid ()) {
    return;
  }
  
  lay::Bitmap *fill, *contour, *vertex, *text; 
  get_bitmaps (vp, canvas, fill, contour, vertex, text);
  if (contour == 0 && vertex == 0 && fill == 0 && text == 0) {
    return;
  }

  lay::Renderer r;
  r.reserve_edges (100);

  double dbu = cv->layout ().dbu ();

  r.set_font (db::Font (mp_view->text_font ()));
  r.apply_text_trans (mp_view->apply_text_trans ());
  r.default_text_size (db::Coord (mp_view->default_text_size () / dbu));
  r.set_precise (true);

  db::box_convert<db::CellInst> bc (cv->layout ());
  db::Box cell_box (bc (m_inst.cell_inst ().object ()));

  db::CplxTrans vpm (vp.trans () * dbu);
  for (db::CellInstArray::iterator arr = m_inst.cell_inst ().begin (); ! arr.at_end (); ++arr) {
    if (! mp_trans_vector) {
      r.draw (cell_box, (vpm * m_trans) * m_inst.cell_inst ().complex_trans (*arr), fill, contour, vertex, text);
    } else {
      for (std::vector<db::CplxTrans>::const_iterator tr = mp_trans_vector->begin (); tr != mp_trans_vector->end (); ++tr) {
        r.draw (cell_box, vpm * (*tr * m_trans) * m_inst.cell_inst ().complex_trans (*arr), fill, contour, vertex, text);
      }
    }
  }

}

// ------------------------------------------------------------------------

ShapeMarker::ShapeMarker (lay::LayoutView *view, unsigned int cv_index)
  : GenericMarkerBase (view, cv_index), m_shape ()
{ 
  // .. nothing yet ..
}

ShapeMarker::~ShapeMarker ()
{
  // .. nothing yet ..
}

void 
ShapeMarker::set (const db::Shape &shape, const db::CplxTrans &trans)
{
  m_shape = shape;
  GenericMarkerBase::set (trans);
}

void 
ShapeMarker::set (const db::Shape &shape, const db::CplxTrans &t1, const std::vector<db::CplxTrans> &trans)
{
  m_shape = shape;
  GenericMarkerBase::set (t1, trans);
}

void 
ShapeMarker::render (const Viewport &vp, ViewObjectCanvas &canvas)
{
  if (m_cv_index >= (unsigned int) (mp_view->cellviews ())) { 
    return;
  }
  const lay::CellView &cv = mp_view->cellview (m_cv_index);
  if (! cv.is_valid ()) {
    return;
  }
  
  lay::Bitmap *fill, *contour, *vertex, *text; 
  get_bitmaps (vp, canvas, fill, contour, vertex, text);
  if (contour == 0 && vertex == 0 && fill == 0 && text == 0) {
    return;
  }

  lay::Renderer r;
  r.reserve_edges (100);

  double dbu = cv->layout ().dbu ();

  r.set_font (db::Font (mp_view->text_font ()));
  r.apply_text_trans (mp_view->apply_text_trans ());
  r.default_text_size (db::Coord (mp_view->default_text_size () / dbu));
  r.set_precise (true);

  db::CplxTrans vpm = vp.trans () * dbu;
  if (mp_trans_vector) {
    for (std::vector<db::CplxTrans>::const_iterator tr = mp_trans_vector->begin (); tr != mp_trans_vector->end (); ++tr) {
      db::CplxTrans t = vpm * (*tr * m_trans);
      r.draw (m_shape, &cv->layout ().properties_repository (), t, fill, contour, vertex, text);
    }
  } else {
    db::CplxTrans t = vpm * m_trans;
    r.draw (m_shape, &cv->layout ().properties_repository (), t, fill, contour, vertex, text);
  }
}

// ------------------------------------------------------------------------

Marker::Marker (lay::LayoutView *view, unsigned int cv_index)
  : GenericMarkerBase (view, cv_index)
{ 
  m_type = None;
  m_object.any = 0;
}

Marker::~Marker ()
{
  remove_object ();
}

void
Marker::set ()
{
  remove_object ();
  redraw ();
}

void 
Marker::set (const db::Box &box, const db::CplxTrans &trans)
{
  remove_object ();

  m_type = Box;
  m_object.box = new db::Box (box);

  GenericMarkerBase::set (trans);
}

void 
Marker::set (const db::Box &box, const db::CplxTrans &t1, const std::vector<db::CplxTrans> &trans)
{
  remove_object ();

  m_type = Box;
  m_object.box = new db::Box (box);

  GenericMarkerBase::set (t1, trans);
}

void 
Marker::set (const db::Polygon &poly, const db::CplxTrans &trans)
{
  remove_object ();

  m_type = Polygon;
  m_object.polygon = new db::Polygon (poly);

  GenericMarkerBase::set (trans);
}

void 
Marker::set (const db::Polygon &poly, const db::CplxTrans &t1, const std::vector<db::CplxTrans> &trans)
{
  remove_object ();

  m_type = Polygon;
  m_object.polygon = new db::Polygon (poly);

  GenericMarkerBase::set (t1, trans);
}

void 
Marker::set (const db::Edge &edge, const db::CplxTrans &trans)
{
  remove_object ();

  m_type = Edge;
  m_object.edge = new db::Edge (edge);

  GenericMarkerBase::set (trans);
}

void 
Marker::set (const db::Edge &edge, const db::CplxTrans &t1, const std::vector<db::CplxTrans> &trans)
{
  remove_object ();

  m_type = Edge;
  m_object.edge = new db::Edge (edge);

  GenericMarkerBase::set (t1, trans);
}

void 
Marker::set (const db::Path &path, const db::CplxTrans &trans)
{
  remove_object ();

  m_type = Path;
  m_object.path = new db::Path (path);

  GenericMarkerBase::set (trans);
}

void 
Marker::set (const db::Path &path, const db::CplxTrans &t1, const std::vector<db::CplxTrans> &trans)
{
  remove_object ();

  m_type = Path;
  m_object.path = new db::Path (path);

  GenericMarkerBase::set (t1, trans);
}

void 
Marker::set (const db::Text &text, const db::CplxTrans &trans)
{
  remove_object ();

  m_type = Text;
  m_object.text = new db::Text (text);

  GenericMarkerBase::set (trans);
}

void 
Marker::set (const db::Text &text, const db::CplxTrans &t1, const std::vector<db::CplxTrans> &trans)
{
  remove_object ();

  m_type = Text;
  m_object.text = new db::Text (text);

  GenericMarkerBase::set (t1, trans);
}

void 
Marker::set (const db::CellInstArray &instance, const db::CplxTrans &trans)
{
  remove_object ();

  m_type = Instance;
  m_object.inst = new db::CellInstArray (instance);

  GenericMarkerBase::set (trans);
}

void 
Marker::set (const db::CellInstArray &instance, const db::CplxTrans &t1, const std::vector<db::CplxTrans> &trans)
{
  remove_object ();

  m_type = Instance;
  m_object.inst = new db::CellInstArray (instance);

  GenericMarkerBase::set (t1, trans);
}

void
Marker::remove_object ()
{
  if (m_type == Box) {
    delete m_object.box;
  } else if (m_type == Polygon) {
    delete m_object.polygon;
  } else if (m_type == Edge) {
    delete m_object.edge;
  } else if (m_type == Path) {
    delete m_object.path;
  } else if (m_type == Text) {
    delete m_object.text;
  } else if (m_type == Instance) {
    delete m_object.inst;
  } 

  m_type = None;
  m_object.any = 0;
}

void 
Marker::render (const Viewport &vp, ViewObjectCanvas &canvas)
{ 
  lay::Bitmap *fill, *contour, *vertex, *text; 
  get_bitmaps (vp, canvas, fill, contour, vertex, text);
  if (contour == 0 && vertex == 0 && fill == 0 && text == 0) {
    return;
  }

  lay::Renderer r;
  r.reserve_edges (100);

  const lay::CellView &cv = mp_view->cellview (m_cv_index);
  tl_assert (cv.is_valid ());
  double dbu = cv->layout ().dbu ();

  r.set_font (db::Font (mp_view->text_font ()));
  r.apply_text_trans (mp_view->apply_text_trans ());
  r.default_text_size (db::Coord (mp_view->default_text_size () / dbu));
  r.set_precise (true);

  db::CplxTrans vpm (vp.trans () * dbu);
  if (! mp_trans_vector) {
 
    db::CplxTrans t = vpm * m_trans;

    if (m_type == Box) {
      r.draw (*m_object.box, t, fill, contour, vertex, text);
    } else if (m_type == Polygon) {
      r.draw (*m_object.polygon, t, fill, contour, vertex, text);
    } else if (m_type == Path) {
      r.draw (*m_object.path, t, fill, contour, vertex, text);
    } else if (m_type == Text) {
      r.draw (*m_object.text, t, fill, contour, vertex, text);
    } else if (m_type == Edge) {
      r.draw (*m_object.edge, t, fill, contour, vertex, text);
    } else if (m_type == Instance) {
      db::box_convert<db::CellInst> bc (cv->layout ());
      db::Box cell_box (bc (m_object.inst->object ()));
      for (db::CellInstArray::iterator arr = m_object.inst->begin (); ! arr.at_end (); ++arr) {
        r.draw (cell_box, t * m_object.inst->complex_trans (*arr), fill, contour, vertex, text);
      }
    }

  } else {

    for (std::vector<db::CplxTrans>::const_iterator tr = mp_trans_vector->begin (); tr != mp_trans_vector->end (); ++tr) {

      db::CplxTrans t = vpm * (*tr * m_trans);

      if (m_type == Box) {
        r.draw (*m_object.box, t, fill, contour, vertex, text);
      } else if (m_type == Polygon) {
        r.draw (*m_object.polygon, t, fill, contour, vertex, text);
      } else if (m_type == Path) {
        r.draw (*m_object.path, t, fill, contour, vertex, text);
      } else if (m_type == Text) {
        r.draw (*m_object.text, t, fill, contour, vertex, text);
      } else if (m_type == Edge) {
        r.draw (*m_object.edge, t, fill, contour, vertex, text);
      } else if (m_type == Instance) {
        db::box_convert<db::CellInst> bc (cv->layout ());
        db::Box cell_box (bc (m_object.inst->object ()));
        for (db::CellInstArray::iterator arr = m_object.inst->begin (); ! arr.at_end (); ++arr) {
          r.draw (cell_box, t * m_object.inst->complex_trans (*arr), fill, contour, vertex, text);
        }
      }

    }

  }

}

// ------------------------------------------------------------------------

DMarker::DMarker (lay::LayoutView *view)
  : MarkerBase (view), mp_view (view)
{ 
  m_type = None;
  m_object.any = 0;
}

DMarker::~DMarker ()
{
  remove_object ();
}

void 
DMarker::set (const db::DBox &box)
{
  remove_object ();

  m_type = Box;
  m_object.box = new db::DBox (box);

  redraw ();
}

void 
DMarker::set (const db::DPolygon &poly)
{
  remove_object ();

  m_type = Polygon;
  m_object.polygon = new db::DPolygon (poly);

  redraw ();
}

void 
DMarker::set (const db::DEdge &edge)
{
  remove_object ();

  m_type = Edge;
  m_object.edge = new db::DEdge (edge);

  redraw ();
}

void 
DMarker::set (const db::DPath &path)
{
  remove_object ();

  m_type = Path;
  m_object.path = new db::DPath (path);

  redraw ();
}

void 
DMarker::set (const db::DText &text)
{
  remove_object ();

  m_type = Text;
  m_object.text = new db::DText (text);

  redraw ();
}

void
DMarker::remove_object ()
{
  if (m_type == Box) {
    delete m_object.box;
  } else if (m_type == Polygon) {
    delete m_object.polygon;
  } else if (m_type == Edge) {
    delete m_object.edge;
  } else if (m_type == Path) {
    delete m_object.path;
  } else if (m_type == Text) {
    delete m_object.text;
  } 

  m_type = None;
  m_object.any = 0;
}

void 
DMarker::render (const Viewport &vp, ViewObjectCanvas &canvas)
{ 
  lay::Bitmap *fill, *contour, *vertex, *text; 
  get_bitmaps (vp, canvas, fill, contour, vertex, text);
  if (contour == 0 && vertex == 0 && fill == 0 && text == 0) {
    return;
  }

  lay::Renderer r;
  r.reserve_edges (100);

  r.set_font (db::Font (mp_view->text_font ()));
  r.apply_text_trans (mp_view->apply_text_trans ());
  r.default_text_size (mp_view->default_text_size ());
  r.set_precise (true);

  db::DCplxTrans t = vp.trans ();

  if (m_type == Box) {
    r.draw (*m_object.box, t, fill, contour, vertex, text);
  } else if (m_type == Polygon) {
    r.draw (*m_object.polygon, t, fill, contour, vertex, text);
  } else if (m_type == Path) {
    r.draw (*m_object.path, t, fill, contour, vertex, text);
  } else if (m_type == Text) {
    r.draw (*m_object.text, t, fill, contour, vertex, text);
  } else if (m_type == Edge) {
    r.draw (*m_object.edge, t, fill, contour, vertex, text);
  }

}

}

