
/*

  KLayout Layout Viewer
  Copyright (C) 2006-2007 Matthias Koefferlein

  This program is free software; you can redistribute it and/or modify
  it under the terms of the GNU General Public License as published by
  the Free Software Foundation; either version 2 of the License, or
  (at your option) any later version.

  This program is distributed in the hope that it will be useful,
  but WITHOUT ANY WARRANTY; without even the implied warranty of
  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
  GNU General Public License for more details.

  You should have received a copy of the GNU General Public License
  along with this program; if not, write to the Free Software
  Foundation, Inc., 51 Franklin St, Fifth Floor, Boston, MA  02110-1301  USA

*/



#include <layRenderer.h>
#include <layBitmap.h>
#include <layFixedFont.h>
#include <laySnap.h>

#include <tlAlgorithm.h>

#include <algorithm>

namespace lay 
{

Renderer::Renderer ()
  : m_ortho (true),
    m_draw_texts (true),
    m_draw_properties (false),
    m_default_text_size (16),
    m_apply_text_trans (true),
    m_precise (false),
    m_font (db::DefaultFont)
{
  // .. nothing else ..
}

void 
Renderer::insert (const db::DBox &box, const std::string &text, db::Font font, db::HAlign halign, db::VAlign valign, db::fixpoint_trans trans)
{
  m_texts.push_back (lay::RenderText ());
  m_texts.back ().b = box;
  m_texts.back ().text = text;
  m_texts.back ().font = font;
  m_texts.back ().halign = halign;
  m_texts.back ().valign = valign;
  m_texts.back ().trans = trans;
}

void 
Renderer::insert (const db::DBox &b)
{
  insert (db::DEdge (b.p1 (), db::DPoint (b.p1 ().x (), b.p2 ().y ())));
  insert (db::DEdge (db::DPoint (b.p1 ().x (), b.p2 ().y ()), b.p2 ()));
  insert (db::DEdge (b.p2 (), db::DPoint (b.p2 ().x (), b.p1 ().y ())));
  insert (db::DEdge (db::DPoint (b.p2 ().x (), b.p1 ().y ()), b.p1 ()));
}

void 
Renderer::insert (const db::DEdge &e)
{
  if (m_edges.begin () == m_edges.end ()) {
    m_xmin = std::min (e.x1 (), e.x2 ());
    m_xmax = std::max (e.x1 (), e.x2 ());
    m_ymin = std::min (e.y1 (), e.y2 ());
    m_ymax = std::max (e.y1 (), e.y2 ());
  } else {
    m_xmin = std::min (m_xmin, std::min (e.x1 (), e.x2 ()));
    m_xmax = std::max (m_xmax, std::max (e.x1 (), e.x2 ()));
    m_ymin = std::min (m_ymin, std::min (e.y1 (), e.y2 ()));
    m_ymax = std::max (m_ymax, std::max (e.y1 (), e.y2 ()));
  }

  //  check, if the edge is neigther horizontal nor vertical - 
  //  reset the orthogonal flag in this case.
  if (m_ortho && fabs (e.x1 () - e.x2 ()) > render_epsilon 
              && fabs (e.y1 () - e.y2 ()) > render_epsilon) {
    m_ortho = false;
  }

  m_edges.push_back (e);
}

void 
Renderer::render_texts (lay::Bitmap &bitmap)
{
  for (std::vector<lay::RenderText>::const_iterator t = m_texts.begin (); t != m_texts.end (); ++t) {

    if (t->font == db::DefaultFont) {

      //  count the lines and max. characters per line

      unsigned int lines = 1;
      for (const char *cp = t->text.c_str (); *cp; ++cp) {
        if (*cp == '\012' || *cp == '\015') {
          if (*cp == '\015' && cp[1] == '\012') {
            ++cp;
          }
          ++lines;
        }
      }

      //  compute the actual top left position
      double y;
      if (t->valign == db::VAlignBottom) {
        y = t->b.bottom ();
        y += double (ff_line_height * (lines - 1) + ff_height);
      } else if (t->valign == db::VAlignCenter) {
        y = t->b.center ().y ();
        y += double ((ff_line_height * (lines - 1) + ff_height) / 2);
      } else {
        y = t->b.top ();
      }

      //  start generating the characters
       
      const char *cp1 = t->text.c_str ();
      while (*cp1) {

        unsigned int length = 0;
        const char *cp = cp1; 
        while (*cp && *cp != '\012' && *cp != '\015') {
          ++length;
          ++cp; 
        }

        double xx;
        if (t->halign == db::HAlignRight) {
          xx = t->b.right ();
          xx -= double (ff_width * length);
        } else if (t->halign == db::HAlignCenter) {
          xx = t->b.center ().x ();
          xx -= double (ff_width * length / 2);
        } else {
          xx = t->b.left ();
        }
        xx -= 0.5;

        if (y > -0.5 && y < double (bitmap.height () + ff_height - 1) - 0.5) {

          for ( ; cp1 != cp; ++cp1) {

            unsigned char c = (unsigned char)*cp1;
            size_t cc (c); // to suppress a compiler warning ..
            if (c >= ff_first_char && cc < sizeof (ff_data) / sizeof (ff_data [0]) + size_t (ff_first_char)
                && xx > -100.0 && xx < double (bitmap.width ())) {
              bitmap.fill_pattern (int (y + 0.5), int (floor (xx)), ff_data [c - ff_first_char], ff_height);
            }

            xx += double (ff_width);

          }

        } else {
          cp1 = cp;
        }

        //  next line
        if (*cp1 == '\012' || *cp1 == '\015') {
          if (*cp1 == '\015' && cp1[1] == '\012') {
            ++cp1;
          }
          ++cp1;
          y -= double (ff_line_height);
        }

      }

    } else {
     
      //  Create a sub-renderer so we do not need to clear *this
      lay::Renderer hr;

      db::DHershey ht (t->text, t->font);
      hr.reserve_edges (ht.count_edges ());
      ht.justify (t->b.transformed (t->trans.inverted ()), t->halign, t->valign);

      //  text becomes unreadable at a low scaling factor - don't draw then.
      if (ht.scale_factor () > 0.2) {

        db::DHershey::edge_iterator e = ht.begin_edges ();
        while (! e.at_end ()) {
          hr.insert ((*e).transformed (t->trans));
          ++e;
        }

      }

      hr.render_contour (bitmap);

    }

  }

}

void 
Renderer::render_vertices (lay::Bitmap &bitmap)
{
  //  a basic shortcut if there are no edges to render
  if (m_edges.end () == m_edges.begin ()) {
    return;
  }

  //  basic optimization: just a dot
  if (floor (m_xmax + 0.5) == floor (m_xmin + 0.5) &&
      floor (m_ymax + 0.5) == floor (m_ymin + 0.5)) {
    if (m_xmin > -0.5 && m_ymin > -0.5 && 
        m_xmin < double (bitmap.width ()) - 0.5 &&
        m_ymin < double (bitmap.height ()) - 0.5) {
      unsigned int yint = (unsigned int) (m_ymin + 0.5);
      unsigned int xint = (unsigned int) (m_xmin + 0.5);
      bitmap.fill (yint, xint, xint + 1);
    }
    return;
  }

  double xmax = bitmap.width ();
  double ymax = bitmap.height ();

  for (std::vector<lay::RenderEdge>::iterator e = m_edges.begin (); e != m_edges.end (); ++e) {

    double x, y;
    
    x = e->x1 () + 0.5;
    y = e->y1 () + 0.5;
    if (x >= 0.0 && x < xmax && y >= 0.0 && y < ymax) {
      unsigned int xint = (unsigned int)x;
      bitmap.fill ((unsigned int)y, xint, xint + 1);
    }
     
    x = e->x2 () + 0.5;
    y = e->y2 () + 0.5;
    if (x >= 0.0 && x < xmax && y >= 0.0 && y < ymax) {
      unsigned int xint = (unsigned int)x;
      bitmap.fill ((unsigned int)y, xint, xint + 1);
    }
     
  }
}

void 
Renderer::render_contour (lay::Bitmap &bitmap)
{
  //  a basic shortcut if there are no edges to render
  if (m_edges.end () == m_edges.begin ()) {
    return;
  }

  //  basic optimizations: outside window
  if (m_xmax < -0.5 || m_xmin > bitmap.width () - 0.5 ||
      m_ymax < -0.5 || m_ymin > bitmap.height () - 0.5) {
    return;
  }

  //  basic optimization: just a line or a dot
  if (floor (m_xmax + 0.5) == floor (m_xmin + 0.5)) {
    unsigned int y1int = (unsigned int) (std::max (0.0, std::min (m_ymin + 0.5, double (bitmap.height () - 1))));
    unsigned int y2int = (unsigned int) (std::max (0.0, std::min (m_ymax + 0.5, double (bitmap.height () - 1))));
    unsigned int xint = (unsigned int) (std::max (0.0, std::min (m_xmin + 0.5, double (bitmap.width () - 1))));
    for (unsigned int y = y1int; y <= y2int; ++y) {
      bitmap.fill (y, xint, xint + 1);
    }
    return;
  }
    
  if (floor (m_ymax + 0.5) == floor (m_ymin + 0.5)) {
    unsigned int x1int = (unsigned int) (std::max (0.0, std::min (m_xmin + 0.5, double (bitmap.width () - 1))));
    unsigned int x2int = (unsigned int) (std::max (0.0, std::min (m_xmax + 0.5, double (bitmap.width () - 1))));
    unsigned int yint = (unsigned int) (std::max (0.0, std::min (m_ymin + 0.5, double (bitmap.height () - 1))));
    bitmap.fill (yint, x1int, x2int + 1);
    return;
  }
  
  if (! m_ortho) {
    
    //  this is the generic case

    for (std::vector<lay::RenderEdge>::iterator e = m_edges.begin (); e != m_edges.end (); ++e) {

      //  This is the line render algorithm 
      //  The basic idea is to decompose the line into stripes
      //  associated with a integer y value. The stripe extends
      //  from x1 to x2 then. The algorithm tries to find a 
      //  set of pixels on the y line that covers the range x1
      //  to x2 as good as possible and advances to the next y 
      //  value then.

      //  TODO: the rendering would be somewhat more efficient if
      //  we would first clip the line and then render it. This
      //  way we could remove the tests in the rendering loop.

      if (e->y1 () < double (bitmap.height ()) - 0.5 && e->y2 () >= -0.5) {

        double y = std::max (floor (e->y1 () + 0.5), 0.0);
        double x = e->pos (y - 0.5);

        double dx = e->pos (y + 0.5) - x;
        double dx1 = (e->y2 () - e->y1 ()) < 1e-6 ? 0.0 : (e->x2 () - e->x1 ()) / (e->y2 () - e->y1 ());

        double y2m = e->y2 () - 0.5;

        unsigned int yeint = (unsigned int) std::min (double (bitmap.height () - 1), std::max (floor (e->y2 () + 0.5), 0.0));

        unsigned int xint = (unsigned int) (std::max (0.0, std::min (double (bitmap.width () - 1), x) + 0.5));
        unsigned int yint = (unsigned int) y;

        if (x < (double) bitmap.width () - 0.5 && x >= 0.0) {
          bitmap.fill (yint, xint, xint + 1);
        }

        if (e->x2 () > e->x1 ()) {

          while (yint <= yeint) {

            double xx;
            if (double (yint) > y2m) { 
              xx = e->x2 () + 0.5;
            } else {
              xx = x + dx;
              dx = dx1;
            }

            unsigned int xe;
            if (xx >= 0.0) {
              if (xx >= (double) bitmap.width ()) {
                if (x >= (double) bitmap.width () - 1) {
                  break; // done.
                }
                xe = bitmap.width () - 1;
              } else {
                xe = (unsigned int) (xx);
              }
              if (xe <= xint) {
                bitmap.fill (yint, xint, xint + 1);
              } else {
                bitmap.fill (yint, xint + 1, xe + 1);
                xint = xe;
              }
            } else {
              xint = 0;
            }

            x = xx;
            ++yint;

          }

        } else {

          while (yint <= yeint) {

            double xx;
            if (double (yint) > y2m) { 
              xx = e->x2 () - 0.5;
            } else {
              xx = x + dx;
              dx = dx1;
            }

            unsigned int xe;
            if (xx < double (bitmap.width () - 1)) {
              if (xx <= 0.0) {
                if (x <= 0.0) {
                  break; // done.
                }
                xe = 0;
              } else {
                xe = (unsigned int) (xx);
                if (double (xe) != xx) { 
                  ++xe;
                }
              }
              if (xe >= xint) {
                bitmap.fill (yint, xint, xint + 1);
              } else {
                bitmap.fill (yint, xe, xint);
                xint = xe;
              }
            } else {
              xint = bitmap.width ();
            }

            x = xx;
            ++yint;

          }

        }

      }

    }

  } else {
    
    //  this is the purely manhattan case

    for (std::vector<lay::RenderEdge>::iterator e = m_edges.begin (); e != m_edges.end (); ++e) {

      //  This is the line render algorithm 
      //  The basic idea is to decompose the line into stripes
      //  associated with a integer y value. The stripe extends
      //  from x1 to x2 then. The algorithm tries to find a 
      //  set of pixels on the y line that covers the range x1
      //  to x2 as good as possible and advances to the next y 
      //  value then.

      //  TODO: the rendering would be somewhat more efficient if
      //  we would first clip the line and then render it. This
      //  way we could remove the tests in the rendering loop.

      if (! e->is_horizontal ()) {

        //  vertical
        double x = e->x1 ();
        if (e->y1 () < double (bitmap.height ()) - 0.5  
            && e->y2 () >= -0.5
            && x < (double) bitmap.width () - 0.5 
            && x >= -0.5) {

          unsigned int xint  = (unsigned int) (std::max (0.0, std::min (double (bitmap.width () - 1), x) + 0.5));
          unsigned int yint  = (unsigned int) std::max (floor (e->y1 () + 0.5), 0.0);
          unsigned int yeint = (unsigned int) std::min (double (bitmap.height () - 1), std::max (floor (e->y2 () + 0.5), 0.0));

          while (yint <= yeint) {
            bitmap.fill (yint, xint, xint + 1);
            ++yint;
          }

        }

      } else {

        //  horizontal
        double x1 = e->x1 ();
        double x2 = e->x2 ();
        if (x1 > x2) {
          std::swap (x1, x2);
        }

        double y = e->y1 ();
        if (y < (double) bitmap.height () - 0.5 
            && y >= -0.5 
            && x1 < (double) bitmap.width () - 0.5 
            && x2 >= -0.5) {
        
          unsigned int x1int = (unsigned int) (std::max (0.0, std::min (double (bitmap.width () - 1), x1) + 0.5));
          unsigned int x2int = (unsigned int) (std::max (0.0, std::min (double (bitmap.width () - 1), x2) + 0.5));
          unsigned int yint  = (unsigned int) std::max (floor (y + 0.5), 0.0);
          bitmap.fill (yint, x1int, x2int + 1);

        }

      }

    }

  }

}

struct PosCompareF 
{
  bool operator() (const RenderEdge &a, const RenderEdge &b) const
  {
    return a.pos () < b.pos ();
  }
};

struct X1CompareF 
{
  bool operator() (const RenderEdge &a, const RenderEdge &b) const
  {
    return a.x1 () < b.x1 ();
  }
};

void 
Renderer::render_fill (lay::Bitmap &bitmap)
{
  //  a basic shortcut if there are no edges to render
  if (m_edges.end () == m_edges.begin ()) {
    return;
  }

  //  basic optimizations: outside window
  if (m_xmax < -0.5 || m_xmin > bitmap.width () - 0.5 ||
      m_ymax < -0.5 || m_ymin > bitmap.height () - 0.5) {
    return;
  }

  //  basic optimization: just a line or a dot
  if (floor (m_xmax + 0.5) == floor (m_xmin + 0.5)) {
    unsigned int y1int = (unsigned int) (std::max (0.0, std::min (m_ymin + 0.5, double (bitmap.height () - 1))));
    unsigned int y2int = (unsigned int) (std::max (0.0, std::min (m_ymax + 0.5, double (bitmap.height () - 1))));
    unsigned int xint = (unsigned int) (std::max (0.0, std::min (m_xmin + 0.5, double (bitmap.width () - 1))));
    for (unsigned int y = y1int; y <= y2int; ++y) {
      bitmap.fill (y, xint, xint + 1);
    }
    return;
  }
    
  if (floor (m_ymax + 0.5) == floor (m_ymin + 0.5)) {
    unsigned int x1int = (unsigned int) (std::max (0.0, std::min (m_xmin + 0.5, double (bitmap.width () - 1))));
    unsigned int x2int = (unsigned int) (std::max (0.0, std::min (m_xmax + 0.5, double (bitmap.width () - 1))));
    unsigned int yint = (unsigned int) (std::max (0.0, std::min (m_ymin + 0.5, double (bitmap.height () - 1))));
    bitmap.fill (yint, x1int, x2int + 1);
    return;
  }

  //  sort the edges so we can operate on the sorted list
  tl::sort (m_edges.begin (), m_edges.end ());

  double y = std::max (0.0, floor (m_edges.begin ()->y1 ()));
  std::vector<lay::RenderEdge>::iterator done = m_edges.begin ();

  if (! m_ortho) {
    
    //  this is generic case
    while (done != m_edges.end () && y < bitmap.height ()) {

      for ( ; done != m_edges.end (); ++done) {
        if (! done->done (y)) {
          break;
        }
      }

      std::vector<lay::RenderEdge>::iterator todo = done;

      for ( ; todo != m_edges.end (); ++todo) {
        if (todo->done (y)) {
          std::swap (*done, *todo);
          ++done;
        }
        if (todo->todo (y)) {
          break;
        } 
      }

      std::vector<lay::RenderEdge>::iterator e;
      for (e = done; e != todo; ++e) {
        e->set_pos (e->x1 () + e->slope () * (y - e->y1 ()));
      }

      PosCompareF f;
      tl::sort (done, todo, f);

      int c = 0;
      bool x1set = false;
      double x1 = 0;
      unsigned int yint = (unsigned int) (y + 0.5);

      for (e = done; e != todo; ++e) {
        if (! e->is_horizontal ()) {
          c += e->delta ();
          if (c == 0) {  //  this is implementing the != 0 rule
            if (e->pos () > 0) {
              unsigned int x1int = 0;
              if (x1 > 0.0) {
                x1int = (unsigned int) x1;
                if (double (x1int) != x1) {
                  ++x1int; 
                }
              }
              bitmap.fill (yint, x1int, (unsigned int) std::min (double (bitmap.width () - 1), e->pos ()) + 1);
            }
            x1set = false;
          } else if (!x1set) {
            x1 = e->pos ();
            x1set = true;
            if (x1 >= double (bitmap.width ())) {
              break;
            }
          }
        }
      }
      
      y += 1.0;

    }

  } else {

    //  this is the purely manhattan case
    //  @@@ the manhattan optimization is not really effective ..
    while (done != m_edges.end () && y < bitmap.height ()) {

      for ( ; done != m_edges.end (); ++done) {
        if (! done->done (y)) {
          break;
        }
      }

      std::vector<lay::RenderEdge>::iterator todo = done;
      for ( ; todo != m_edges.end (); ++todo) {
        if (todo->done (y)) {
          std::swap (*done, *todo);
          ++done;
        }
        if (todo->todo (y)) {
          break;
        } 
      }

      std::vector<lay::RenderEdge>::iterator e;

      X1CompareF f;
      tl::sort (done, todo, f);

      int c = 0;
      bool x1set = false;
      double x1 = 0;
      unsigned int yint = (unsigned int) (y + 0.5);

      for (e = done; e != todo; ++e) {
        if (! e->is_horizontal ()) {
          c += e->delta ();
          if (c == 0) {  //  this is implementing the != 0 rule
            if (e->x1 () > 0) {
              unsigned int x1int = 0;
              if (x1 > 0.0) {
                x1int = (unsigned int) x1;
                if (double (x1int) != x1) {
                  ++x1int; 
                }
              }
              bitmap.fill (yint, x1int, (unsigned int) std::min (double (bitmap.width () - 1), e->x1 ()) + 1);
            }
            x1set = false;
          } else if (!x1set) {
            x1 = e->x1 ();
            x1set = true;
            if (x1 >= double (bitmap.width ())) {
              break;
            }
          }
        }
      }
      
      y += 1.0;

    }

  }

}

template <class Box, class Trans>
bool 
Renderer::simplify_to_box (Box &b, const Trans &trans)
{
  const double small_size_threshold = 1.0;

  bool ortho = trans.is_ortho ();
  if (! m_precise && ((ortho && trans.ctrans (std::min (b.width (), b.height ())) < small_size_threshold) || 
                      (! ortho && trans.ctrans (std::max (b.width (), b.height ())) < small_size_threshold))) {

    if (trans.ctrans (b.width ()) < small_size_threshold) {
      typename Box::coord_type c = b.center ().x ();
      b.set_left (c);
      b.set_right (c);
    } 
    if (trans.ctrans (b.height ()) < small_size_threshold) {
      typename Box::coord_type c = b.center ().y ();
      b.set_top (c);
      b.set_bottom (c);
    } 

    return true;

  } else {
    return false;
  }
}

void 
Renderer::draw (const db::Shape &shape, const db::PropertiesRepository *prep, const db::CplxTrans &trans,
                lay::Bitmap *fill, lay::Bitmap *frame, lay::Bitmap *vertices, lay::Bitmap *text)
{
  if (shape.is_text ()) {
    
    db::Point p = shape.text_trans ().disp ();
    db::DPoint dp = trans * p;

    lay::Bitmap *bmp = frame ? vertices : frame;
    if ((vertices || frame) &&
        dp.x () < bmp->width () - 0.5 && dp.x () > -0.5 && 
        dp.y () < bmp->height () - 0.5 && dp.y () > -0.5) {

      clear ();

      db::Point pp (db::Point::from_double (dp));

      //  generate a dot (cross, if marked) at the text's position
      if (vertices) {
        vertices->fill (pp.y (), pp.x (), pp.x () + 1);
      }
      if (frame) {
        frame->fill (pp.y (), pp.x (), pp.x () + 1);
      }

    }

    if (m_draw_texts && text) {

      db::DPoint tp1 (dp + db::DPoint (2.0, 2.0));
      db::DCoord h = trans.ctrans (shape.text_size () > 0 ? shape.text_size () : m_default_text_size);
      db::DPoint tp2 (dp + db::DPoint (2.0, 2.0 + h));

      clear ();

      insert (db::DBox (tp1, tp2), shape.text_string (), shape.text_font () == db::NoFont ? m_font : shape.text_font (),
              db::HAlignLeft, db::VAlignBottom, 
              db::fixpoint_trans (m_apply_text_trans ? shape.text_trans ().rot () : db::fixpoint_trans::r0));
      
      render_texts (*text);

    }

    if (m_draw_properties && text && prep && shape.has_prop_id ()) {
      draw_propstring (shape.prop_id (), prep, dp, text, trans);
    }

  } else if (shape.is_polygon ()) {

    //  simplify to a rectangle if possible
    db::Box b (shape.bbox ());
    if (simplify_to_box (b, trans)) {
      draw (b, trans, fill, frame, vertices, text);
    } else {

      clear ();
      db::Shape::polygon_edge_iterator e = shape.begin_edge (); 
      for ( ; ! e.at_end (); ++e) {
        insert (trans * *e);
      }

      if (fill) {
        render_fill (*fill);
      }
      if (frame) {
        render_contour (*frame);
      }
      if (vertices) {
        render_vertices (*vertices);
      }

    }

    if (m_draw_properties && text && prep && shape.has_prop_id ()) {
      db::Shape::polygon_edge_iterator e = shape.begin_edge (); 
      db::DPoint pref = trans ((*e).p1 ());
      draw_propstring (shape.prop_id (), prep, pref, text, trans);
    }

  } else if (shape.is_edge ()) {

    draw (shape.edge (), trans, fill, frame, vertices, text);

    if (m_draw_properties && text && prep && shape.has_prop_id ()) {
      draw_propstring (shape.prop_id (), prep, trans (shape.edge ().p1 ()), text, trans);
    }

  } else if (shape.is_path ()) {

    //  simplify to a rectangle if possible
    db::Box b (shape.bbox ());
    if (simplify_to_box (b, trans)) {
      draw (b, trans, fill, frame, vertices, text);
    } else {

      //  instantiate the path and draw
      //  @@@ use the methods provided by the shape to produce the path ..
      db::Path path;
      shape.path (path);
      draw (path, trans, fill, frame, vertices, text);

    }

    if (m_draw_properties && text && prep && shape.has_prop_id ()) {
      db::DPoint pref = trans (*shape.begin_point ());
      draw_propstring (shape.prop_id (), prep, pref, text, trans);
    }

  } else if (shape.is_box ()) {

    draw (shape.box (), trans, fill, frame, vertices, text);

    if (m_draw_properties && text && prep && shape.has_prop_id ()) {
      draw_propstring (shape.prop_id (), prep, trans (shape.box ().p1 ()), text, trans);
    }

  }
}

void 
Renderer::draw (const db::Polygon &poly, const db::CplxTrans &trans,
                lay::Bitmap *fill, lay::Bitmap *frame, lay::Bitmap *vertices, lay::Bitmap *text)
{
  clear ();

  //  simplify to a rectangle if possible
  db::Box b (poly.box ());
  if (simplify_to_box (b, trans)) {
    insert (trans * b);
  } else {
    for (db::Polygon::polygon_edge_iterator e = poly.begin_edge (); ! e.at_end (); ++e) {
      insert (trans * *e);
    }
  }

  if (fill) {
    render_fill (*fill);
  }
  if (frame) {
    render_contour (*frame);
  }
  if (vertices) {
    render_vertices (*vertices);
  }

}

void 
Renderer::draw (const db::DPolygon &poly, const db::DCplxTrans &trans,
                lay::Bitmap *fill, lay::Bitmap *frame, lay::Bitmap *vertices, lay::Bitmap *text)
{
  clear ();

  //  simplify to a rectangle if possible
  db::DBox b (poly.box ());
  if (simplify_to_box (b, trans)) {
    insert (trans * b);
  } else {
    for (db::DPolygon::polygon_edge_iterator e = poly.begin_edge (); ! e.at_end (); ++e) {
      insert (trans * *e);
    }
  }

  if (fill) {
    render_fill (*fill);
  }
  if (frame) {
    render_contour (*frame);
  }
  if (vertices) {
    render_vertices (*vertices);
  }
}

void 
Renderer::draw (const db::ShortBox &box, const db::CplxTrans &trans,
                lay::Bitmap *fill, lay::Bitmap *frame, lay::Bitmap *vertices, lay::Bitmap *text)
{
  if (! box.empty ()) {

    if (trans.is_ortho ()) {
  
      clear ();
      insert (trans * db::Box (box));

      if (fill) {
        render_fill (*fill);
      }
      if (frame) {
        render_contour (*frame);
      }
      if (vertices) {
        render_vertices (*vertices);
      }

    } else {
      draw (db::Polygon (box), trans, fill, frame, vertices, text);
    }

  }
}

void 
Renderer::draw (const db::Box &box, const db::CplxTrans &trans,
                lay::Bitmap *fill, lay::Bitmap *frame, lay::Bitmap *vertices, lay::Bitmap *text)
{
  if (! box.empty ()) {
  
    if (trans.is_ortho ()) {
  
      clear ();
      insert (trans * box);

      if (fill) {
        render_fill (*fill);
      }
      if (frame) {
        render_contour (*frame);
      }
      if (vertices) {
        render_vertices (*vertices);
      }

    } else {
      draw (db::Polygon (box), trans, fill, frame, vertices, text);
    }

  }
}

void 
Renderer::draw (const db::DBox &box, const db::DCplxTrans &trans,
                lay::Bitmap *fill, lay::Bitmap *frame, lay::Bitmap *vertices, lay::Bitmap *text)
{
  if (! box.empty ()) {
  
    if (trans.is_ortho ()) {
  
      clear ();
      insert (trans * box);

      if (fill) {
        render_fill (*fill);
      }
      if (frame) {
        render_contour (*frame);
      }
      if (vertices) {
        render_vertices (*vertices);
      }

    } else {
      draw (db::DPolygon (box), trans, fill, frame, vertices, text);
    }

  }
}

void 
Renderer::draw (const db::Path &path, const db::CplxTrans &trans,
                lay::Bitmap *fill, lay::Bitmap *frame, lay::Bitmap *vertices, lay::Bitmap *text)
{
  clear ();

  //  simplify to a rectangle if possible
  db::Box b (path.box ());
  if (simplify_to_box (b, trans)) {
    draw (b, trans, fill, frame, vertices, text);
    return;
  } 

  //  generate the hull and produce the edges from this only if the path is considerably wide
  //  otherwise just render the spine
  
  double w = trans.ctrans (path.width ());

  bool thin = (w < 0.5);
  bool quite_thin = (w < 3.0);

  //  render the border
  if (! thin) {

    db::Path::pointlist_type pts;
    path.hull (pts);

    db::Path::pointlist_type::const_iterator p = pts.begin ();
    if (p != pts.end ()) {

      db::Path::pointlist_type::const_iterator pp = p;
      ++pp;

      while (pp != pts.end ()) {
        insert (trans * db::Edge (*p, *pp));
        p = pp;
        ++pp;
      }

      insert (trans * db::Edge (*p, pts [0]));
    }

    if (fill) {
      render_fill (*fill);
    }
    if (frame) {
      render_contour (*frame);
    }

  }

  //  render the spine edges
  if (! quite_thin || thin) {

    clear ();

    db::Path::iterator q = path.begin ();
    if (q != path.end ()) {

      db::Path::iterator qq = q;
      ++qq;

      bool first = true;
      while (qq != path.end ()) {

        db::Edge seg (*q, *qq);
        q = qq;
        ++qq;

        //  If the path is simplified, the spine is drawn instead of the contour.
        //  We must apply the begin/end extensions then.
        
        if (thin && qq == path.end ()) {
          if (path.extensions ().second != 0 && (seg.dx () != 0 || seg.dy () != 0)) {
            db::DPoint ed (seg.p2 () - seg.p1 ());
            ed *= 1.0 / ed.double_distance ();
            seg = db::Edge (seg.p1 (), seg.p2 () + db::Point::from_double (ed * double (path.extensions ().second)));
          }
        } 

        if (first) {
          first = false;
          if (thin && path.extensions ().first != 0 && (seg.dx () != 0 || seg.dy () != 0)) {
            db::DPoint ed (seg.p2 () - seg.p1 ());
            ed *= 1.0 / ed.double_distance ();
            seg = db::Edge (seg.p1 () - db::Point::from_double (ed * double (path.extensions ().first)), seg.p2 ());
          }
        }
          
        insert (trans * seg);

      }

    }

    if (frame) {
      render_contour (*frame);
    }
    if (vertices) {
      render_vertices (*vertices);
    }

  }
}

void 
Renderer::draw (const db::DPath &path, const db::DCplxTrans &trans,
                lay::Bitmap *fill, lay::Bitmap *frame, lay::Bitmap *vertices, lay::Bitmap *text)
{
  clear ();

  //  simplify to a rectangle if possible
  db::DBox b (path.box ());
  if (simplify_to_box (b, trans)) {
    draw (b, trans, fill, frame, vertices, text);
    return;
  } 

  //  generate the hull and produce the edges from this 

  db::DPath::pointlist_type pts;
  path.hull (pts);

  db::DPath::pointlist_type::const_iterator p = pts.begin ();
  if (p != pts.end ()) {

    db::DPath::pointlist_type::const_iterator pp = p;
    ++pp;

    while (pp != pts.end ()) {
      insert (trans * db::DEdge (*p, *pp));
      p = pp;
      ++pp;
    }

    insert (trans * db::DEdge (*p, pts [0]));
  }

  if (fill) {
    render_fill (*fill);
  }
  if (frame) {
    render_contour (*frame);
  }

  //  render the spine edges

  clear ();

  db::DPath::iterator q = path.begin ();
  if (q != path.end ()) {

    db::DPath::iterator qq = q;
    ++qq;

    while (qq != path.end ()) {
      insert (trans * db::DEdge (*q, *qq));
      q = qq;
      ++qq;
    }

  }

  if (frame) {
    render_contour (*frame);
  }
  if (vertices) {
    render_vertices (*vertices);
  }
}

void 
Renderer::draw (const db::Text &txt, const db::CplxTrans &trans,
                lay::Bitmap *fill, lay::Bitmap *frame, lay::Bitmap *vertices, lay::Bitmap *text)
{
  db::Point p = txt.trans ().disp ();
  db::DPoint dp = trans * p;

  lay::Bitmap *bmp = frame ? vertices : frame;
  if ((vertices || frame) &&
      dp.x () < bmp->width () - 0.5 && dp.x () > -0.5 && 
      dp.y () < bmp->height () - 0.5 && dp.y () > -0.5) {

    clear ();

    db::Point pp (db::Point::from_double (dp));

    //  generate a dot (cross, if marked) at the text's position
    if (vertices) {
      vertices->fill (pp.y (), pp.x (), pp.x () + 1);
    }
    if (frame) {
      frame->fill (pp.y (), pp.x (), pp.x () + 1);
    }

  }

  if (m_draw_texts && text) {

    db::fixpoint_trans fp (m_apply_text_trans ? txt.trans ().rot () : db::fixpoint_trans::r0);

    db::DPoint tp1 (db::DPoint (2.0, 2.0));
    db::DCoord h = trans.ctrans (txt.size () > 0 ? txt.size () : m_default_text_size);
    db::DPoint tp2 (tp1 + db::DPoint (0.0, h));

    clear ();

    insert (db::DBox (fp (tp1) + dp, fp (tp2) + dp), txt.string (), txt.font () == db::NoFont ? m_font : txt.font (),
            db::HAlignLeft, db::VAlignBottom, fp);
    
    render_texts (*text);

  }

}

void 
Renderer::draw (const db::DText &txt, const db::DCplxTrans &trans,
                lay::Bitmap *fill, lay::Bitmap *frame, lay::Bitmap *vertices, lay::Bitmap *text)
{
  db::DPoint p = txt.trans ().disp ();
  db::DPoint dp = trans * p;

  lay::Bitmap *bmp = frame ? vertices : frame;
  if ((vertices || frame) &&
      dp.x () < bmp->width () - 0.5 && dp.x () > -0.5 && 
      dp.y () < bmp->height () - 0.5 && dp.y () > -0.5) {

    clear ();

    db::Point pp (db::Point::from_double (dp));

    //  generate a dot (cross, if marked) at the text's position
    if (vertices) {
      vertices->fill (pp.y (), pp.x (), pp.x () + 1);
    }
    if (frame) {
      frame->fill (pp.y (), pp.x (), pp.x () + 1);
    }

  }

  if (m_draw_texts && text) {

    db::fixpoint_trans fp (m_apply_text_trans ? txt.trans ().rot () : db::fixpoint_trans::r0);

    db::DPoint tp1 (db::DPoint (2.0, 2.0));
    db::DCoord h = trans.ctrans (txt.size () > 0 ? txt.size () : m_default_text_size_dbl);
    db::DPoint tp2 (tp1 + db::DPoint (0.0, h));

    clear ();

    insert (db::DBox (fp (tp1) + dp, fp (tp2) + dp), txt.string (), txt.font () == db::NoFont ? m_font : txt.font (),
            db::HAlignLeft, db::VAlignBottom, fp);
    
    render_texts (*text);

  }

}

void 
Renderer::draw (const db::Edge &edge, const db::CplxTrans &trans,
                lay::Bitmap *fill, lay::Bitmap *frame, lay::Bitmap *vertices, lay::Bitmap *text)
{
  clear ();
  insert (trans * edge);

  if (frame) {
    render_contour (*frame);
  }
  if (vertices) {
    render_vertices (*vertices);
  }
}

void 
Renderer::draw (const db::DEdge &edge, const db::DCplxTrans &trans,
                lay::Bitmap *fill, lay::Bitmap *frame, lay::Bitmap *vertices, lay::Bitmap *text)
{
  clear ();
  insert (trans * edge);

  if (frame) {
    render_contour (*frame);
  }
  if (vertices) {
    render_vertices (*vertices);
  }
}

void 
Renderer::draw_propstring (db::PropertiesRepository::properties_id_type id, 
                           const db::PropertiesRepository *prep, const db::DPoint &pref, 
                           lay::Bitmap *text, const db::CplxTrans &trans)
{
  db::DPoint tp1 (pref + db::DPoint (2.0, -2.0));
  db::DPoint tp2 (pref + db::DPoint (2.0, -2.0 - trans.ctrans (m_default_text_size)));

  clear ();

  std::string ptext;

  const char *sep = "";
  const db::PropertiesRepository::properties_set &props = prep->properties (id);
  for (db::PropertiesRepository::properties_set::const_iterator p = props.begin (); p != props.end (); ++p) {
    ptext += sep;
    sep = "\n";
    ptext += prep->prop_name (p->first).to_string ();
    ptext += ": ";
    ptext += p->second.to_string ();
  }

  insert (db::DBox (tp1, tp2), ptext, m_font,
          db::HAlignLeft, db::VAlignTop, 
          db::fixpoint_trans (db::fixpoint_trans::r0));
  
  render_texts (*text);
}

}


