
/*

  KLayout Layout Viewer
  Copyright (C) 2006-2007 Matthias Koefferlein

  This program is free software; you can redistribute it and/or modify
  it under the terms of the GNU General Public License as published by
  the Free Software Foundation; either version 2 of the License, or
  (at your option) any later version.

  This program is distributed in the hope that it will be useful,
  but WITHOUT ANY WARRANTY; without even the implied warranty of
  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
  GNU General Public License for more details.

  You should have received a copy of the GNU General Public License
  along with this program; if not, write to the Free Software
  Foundation, Inc., 51 Franklin St, Fifth Floor, Boston, MA  02110-1301  USA

*/



#ifndef HDR_layRenderer
#define HDR_layRenderer

#include <dbEdge.h>
#include <dbPolygon.h>
#include <dbBox.h>
#include <dbPath.h>
#include <dbText.h>
#include <dbTrans.h>
#include <dbShape.h>
#include <dbHershey.h>

#include <vector>

namespace lay
{

class Bitmap;

const double render_epsilon = 1e-6;

/**
 *  @brief A renderer edge object
 */
class RenderEdge : public db::DEdge 
{
public:
  RenderEdge (const db::DEdge &d)
    : db::DEdge (d), m_d (true)
  {
    //  RenderEdges are normalized such that 
    //  p1 < p2. m_d is false if the points have been swapped.
    if (p2 () < p1 ()) {
      swap_points ();
      m_d = !m_d;
    }

    //  Horizontal is true, if the edge is horizontal within render_epsilon
    //  range.
    m_horizontal = (fabs (dy ()) < render_epsilon);

    //  Compute the slope to speed up position computation
    if (m_horizontal) {
      m_slope = 0.0;
    } else {
      m_slope = dx () / dy ();
    }
  }

  double pos () const
  {
    return m_pos;
  }

  void set_pos (double p) 
  { 
    m_pos = p;
  }

  double pos (double y) const
  {
    if (y > y2 ()) {
      return x2 ();
    } else if (y < y1 ()) {
      return x1 ();
    } else {
      return x1 () + m_slope * (y - y1 ());
    }
  }

  double slope () const
  {
    return m_slope;
  }

  bool is_horizontal () const 
  {
    return m_horizontal;
  }

  int delta () const
  {
    return m_d ? 1 : -1;
  }

  bool done (double y) const
  {
    return (y > y2 ());
  }

  bool todo (double y) const
  {
    return (y <= y1 ());
  }

  void update_pos (double y) 
  {
    m_pos = pos (y);
  }
  
private:
  bool m_d, m_horizontal;
  double m_pos;
  double m_slope;
};

/**
 *  @brief A rendered text object
 */
struct RenderText
{
  db::DBox b;
  std::string text;
  db::Font font;
  db::HAlign halign;
  db::VAlign valign;
  db::fixpoint_trans trans;
};

/**
 *  @brief A edge set and text rendering object
 *
 *  The renderer is supposed to render a set of edges to
 *  one or more bitmaps. One bitmap holds the contour
 *  information, one the fill. A third one holds the 
 *  vertex informations (dots).
 *  
 *  The intended use model is first to reserve a number of
 *  points if necessary, the fill the renderer with the
 *  edges and then render the content to the bitmaps.
 *
 *  The coordinate system of the bitmaps is 0,0..w-1,h-1
 *  with 0,0 being the lower left corner. 
 */

class Renderer 
{
public:

  /**
   *  @brief The default ctor
   */
  Renderer ();

  /**
   *  @brief Reserve space for n edges 
   *
   *  It is strongly recommended to use this method to
   *  reserve space in advance before any insert operation for edges or boxes.
   */
  void reserve_edges (size_t n);

  /**
   *  @brief Reserve space for n texts 
   *
   *  It is strongly recommended to use this method to
   *  reserve space in advance before any insert operation for texts.
   */
  void reserve_texts (size_t n);

  /** 
   *  @brief Clear the content
   */
  void clear ();

  /**
   *  @brief Insert edges using the provided sequence and a transform function
   *
   *  Inserts the edge sequence [from,to), transformed with f(E).
   *
   *  @param from The start of the sequence
   *  @param to The past-end iterator of the sequence
   *  @param f The function to apply before inserting the edge. It must
   *         provide a db::DEdge operator(Iter::value_type).
   */
  template <class Iter, class Func> 
  void insert_edges (Iter from, Iter to, Func f)
  {
    for (Iter e = from; e != to; ++e) {
      insert (f (*e));
    }
  }

  /** 
   *  @brief Insert a box 
   */
  void insert (const db::DBox &b);

  /** 
   *  @brief Insert an edge 
   *
   *  The edges must be a db::DEdge type.
   */
  void insert (const db::DEdge &e);

  /**
   *  @brief Insert a text
   *
   *  @box The box in which to inscribe the text (or the point at which to align if the width and height is zero)
   *  @text The text to render
   *  @font The font to use
   *  @halign The horizontal alignment
   *  @valign The vertical alignment
   *  @trans The orientation of the text
   */
  void insert (const db::DBox &box, const std::string &text, db::Font font, db::HAlign halign, db::VAlign valign, db::fixpoint_trans trans);

  /**
   *  @brief Render the interior of the object to the bitmap 
   */
  void render_fill (lay::Bitmap &bitmap);

  /**
   *  @brief Render the contour of the object to the bitmap 
   */
  void render_contour (lay::Bitmap &bitmap);

  /**
   *  @brief Render the vertices of the object to the bitmap 
   */
  void render_vertices (lay::Bitmap &bitmap);

  /**
   *  @brief Render the texts of the object to the bitmap
   */
  void render_texts (lay::Bitmap &bitmap);

  /**
   *  @brief Control whether draw(db::Shape..) draws precise (does not simplify to box)
   *
   *  Default is "not precise".
   */
  void set_precise (bool f)
  {
    m_precise = f;
  }

  /**
   *  @brief Control whether draw(db::Shape..) draws properties on the text layer
   */
  void draw_properties (bool f)
  {
    m_draw_properties = f;
  }

  /**
   *  @brief Control whether draw(db::Text..) draws texts but rather the pixel
   */
  void draw_texts (bool f)
  {
    m_draw_texts = f;
  }

  /**
   *  @brief The default text size to use for draw(db::Text..) 
   */
  void default_text_size (db::Coord sz)
  {
    m_default_text_size = sz;
  }

  /**
   *  @brief The default text size to use for draw(db::DText..) 
   */
  void default_text_size (double sz)
  {
    m_default_text_size_dbl = sz;
  }

  /**
   *  @brief The font to use for draw(db::Text..).
   */
  void set_font (db::Font f)
  {
    m_font = f;
  }

  /**
   *  @brief Apply text transformations to text or not for draw(db::Text..).
   */
  void apply_text_trans (bool f)
  {
    m_apply_text_trans = f;
  }

  /**
   *  @brief Render a generic shape into a set of bitmaps
   *
   *  The shape can either be a polygon, box, edge, path or text
   *
   *  @param shape The polygon to draw
   *  @param trans The transformation to apply
   *  @param fill The bitmap to which to render the interior. Can be 0 for not drawing this.
   *  @param frame The bitmap to which to render the contour. Can be 0 for not drawing this.
   *  @param vertices The bitmap to which to render the vertices. Can be 0 for not drawing this.
   *  @param texts Currenty not used. Used for drawing properties if drawing of these is required.
   */
  void draw (const db::Shape &shape, const db::PropertiesRepository *prep, const db::CplxTrans &trans,
             lay::Bitmap *fill, lay::Bitmap *frame, lay::Bitmap *vertices, lay::Bitmap *texts); 

  /**
   *  @brief Render a polygon into a set of bitmaps
   *
   *  This method clears the renderer object before drawing it
   *
   *  @param poly The polygon to draw
   *  @param trans The transformation to apply
   *  @param fill The bitmap to which to render the interior. Can be 0 for not drawing this.
   *  @param frame The bitmap to which to render the contour. Can be 0 for not drawing this.
   *  @param vertices The bitmap to which to render the vertices. Can be 0 for not drawing this.
   *  @param texts Currenty not used. No texts to draw.
   */
  void draw (const db::Polygon &poly, const db::CplxTrans &trans,
             lay::Bitmap *fill, lay::Bitmap *frame, lay::Bitmap *vertices, lay::Bitmap *texts); 

  /** 
   *  @brief Same for double coordinates
   */
  void draw (const db::DPolygon &poly, const db::DCplxTrans &trans,
             lay::Bitmap *fill, lay::Bitmap *frame, lay::Bitmap *vertices, lay::Bitmap *texts); 

  /**
   *  @brief Render a box into a set of bitmaps
   *
   *  This method clears the renderer object before drawing it
   *
   *  @param box The box to draw
   *  @param trans The transformation to apply
   *  @param fill The bitmap to which to render the interior. Can be 0 for not drawing this.
   *  @param frame The bitmap to which to render the contour. Can be 0 for not drawing this.
   *  @param vertices The bitmap to which to render the vertices. Can be 0 for not drawing this.
   *  @param texts Currenty not used. No texts to draw.
   */
  void draw (const db::Box &box, const db::CplxTrans &trans,
             lay::Bitmap *fill, lay::Bitmap *frame, lay::Bitmap *vertices, lay::Bitmap *texts); 

  /** 
   *  @brief Same for short representation
   */
  void draw (const db::ShortBox &box, const db::CplxTrans &trans,
             lay::Bitmap *fill, lay::Bitmap *frame, lay::Bitmap *vertices, lay::Bitmap *texts); 

  /** 
   *  @brief Same for double coordinates
   */
  void draw (const db::DBox &box, const db::DCplxTrans &trans,
             lay::Bitmap *fill, lay::Bitmap *frame, lay::Bitmap *vertices, lay::Bitmap *texts); 

  /**
   *  @brief Render a path into a set of bitmaps
   *
   *  This method clears the renderer object before drawing it
   *
   *  @param path The path to draw
   *  @param trans The transformation to apply
   *  @param fill The bitmap to which to render the interior. Can be 0 for not drawing this.
   *  @param frame The bitmap to which to render the contour. Can be 0 for not drawing this.
   *  @param vertices The bitmap to which to render the vertices. Can be 0 for not drawing this.
   *  @param texts Currenty not used. No texts to draw.
   */
  void draw (const db::Path &path, const db::CplxTrans &trans,
             lay::Bitmap *fill, lay::Bitmap *frame, lay::Bitmap *vertices, lay::Bitmap *texts);

  /** 
   *  @brief Same for double coordinates
   */
  void draw (const db::DPath &path, const db::DCplxTrans &trans,
             lay::Bitmap *fill, lay::Bitmap *frame, lay::Bitmap *vertices, lay::Bitmap *texts);

  /**
   *  @brief Render a text into a set of bitmaps
   *
   *  This method clears the renderer object before drawing it
   *
   *  @param text The text to draw
   *  @param trans The transformation to apply
   *  @param fill Not used currenty, since the text does not have an interior
   *  @param frame The bitmap to which to render the contour. Can be 0 for not drawing this.
   *  @param vertices The bitmap to which to render the vertices. Can be 0 for not drawing this.
   *  @param texts The bitmap to which to render the text.
   */
  void draw (const db::Text &text, const db::CplxTrans &trans,
             lay::Bitmap *fill, lay::Bitmap *frame, lay::Bitmap *vertices, lay::Bitmap *texts);

  /** 
   *  @brief Same for double coordinates
   */
  void draw (const db::DText &text, const db::DCplxTrans &trans,
             lay::Bitmap *fill, lay::Bitmap *frame, lay::Bitmap *vertices, lay::Bitmap *texts);

  /**
   *  @brief Render a edge into a set of bitmaps
   *
   *  This method clears the renderer object before drawing it
   *
   *  @param edge The box to draw
   *  @param trans The transformation to apply
   *  @param fill Not used currenty, since the text does not have an interior
   *  @param frame The bitmap to which to render the contour. Can be 0 for not drawing this.
   *  @param vertices The bitmap to which to render the vertices. Can be 0 for not drawing this.
   *  @param texts Currenty not used. No texts to draw.
   */
  void draw (const db::Edge &edge, const db::CplxTrans &trans,
             lay::Bitmap *fill, lay::Bitmap *frame, lay::Bitmap *vertices, lay::Bitmap *texts);

  /** 
   *  @brief Same for double coordinates
   */
  void draw (const db::DEdge &edge, const db::DCplxTrans &trans,
             lay::Bitmap *fill, lay::Bitmap *frame, lay::Bitmap *vertices, lay::Bitmap *texts);

  /**
   *  @brief Draw a properties string at the given position
   *
   *  The transformation is not used to transform the point (it must be given in micron units
   *  already) but to compute the position of the text, given the text height.
   */
  void draw_propstring (db::PropertiesRepository::properties_id_type id, 
                        const db::PropertiesRepository *prep, const db::DPoint &pref, 
                        lay::Bitmap *text, const db::CplxTrans &trans);
 
private:
  std::vector<lay::RenderEdge> m_edges;
  double m_xmin, m_xmax, m_ymin, m_ymax;
  bool m_ortho;
  std::vector<lay::RenderText> m_texts;
  bool m_draw_texts;
  bool m_draw_properties;
  db::Coord m_default_text_size;
  double m_default_text_size_dbl;
  bool m_apply_text_trans;
  bool m_precise;
  db::Font m_font;

  template <class Box, class Trans>
  bool simplify_to_box (Box &, const Trans &);
};

inline void 
Renderer::reserve_edges (size_t n)
{
  m_edges.reserve (n);
}

inline void 
Renderer::reserve_texts (size_t n)
{
  m_texts.reserve (n);
}

inline void 
Renderer::clear ()
{
  //  this implementation is efficent but does not free memory - 
  //  the idea is to let the Renderer object manage it's workspace.
  m_edges.erase (m_edges.begin (), m_edges.end ());
  //  might be manhattan
  m_ortho = true;
  //  see above
  m_texts.erase (m_texts.begin (), m_texts.end ());
}

} // namespace lay

#endif

