
/*

  KLayout Layout Viewer
  Copyright (C) 2006-2007 Matthias Koefferlein

  This program is free software; you can redistribute it and/or modify
  it under the terms of the GNU General Public License as published by
  the Free Software Foundation; either version 2 of the License, or
  (at your option) any later version.

  This program is distributed in the hope that it will be useful,
  but WITHOUT ANY WARRANTY; without even the implied warranty of
  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
  GNU General Public License for more details.

  You should have received a copy of the GNU General Public License
  along with this program; if not, write to the Free Software
  Foundation, Inc., 51 Franklin St, Fifth Floor, Boston, MA  02110-1301  USA

*/


#include <QtGui/QMouseEvent>
#include <QtGui/QWheelEvent>

#include "layViewObject.h"
#include "layBitmap.h"
#include "tlException.h"
#include "layExceptions.h"

#include <algorithm>

namespace lay
{

//  The distance by which the mouse must move in order to create a press/move/release
//  event rather than a single click event:
const int click_tolerance = 5;

// ---------------------------------------------------------------
//  ViewObject implementation

ViewObject::ViewObject (ViewObjectWidget *widget, bool _static)
  : mp_widget (widget), m_static (_static), m_visible (true)
{
  if (widget) {
    widget->m_objects.insert (this, false /*not owned by the widget*/);
    redraw ();
  }
}

ViewObject::ViewObject (const lay::ViewObject &d)
  : m_static (d.m_static)
{
  redraw ();
}

ViewObject &
ViewObject::operator= (const ViewObject &d)
{
  m_static = d.m_static;
  redraw ();
  return *this;
}

ViewObject::~ViewObject ()
{
  redraw ();
}

void 
ViewObject::visible (bool vis)
{
  if (vis != m_visible) {
    m_visible = vis;
    redraw ();
  }
}

void 
ViewObject::redraw ()
{
  if (widget ()) {
    if (m_static) {
      widget ()->touch ();
    } else {
      widget ()->update ();
    }
  }
}

void
ViewObject::thaw ()
{
  if (widget ()) {
    widget ()->thaw (this);
  }
}

void
ViewObject::freeze ()
{
  if (widget ()) {
    widget ()->freeze (this);
  }
}

// ---------------------------------------------------------------
//  ViewService implementation

ViewService::ViewService (ViewObjectWidget *widget)
  : mp_widget (widget), m_abs_grab (false), m_enabled (true)
{
  if (widget) {
    widget->m_services.insert (this);
  }
}

void
ViewService::mouse_enable (bool en)
{
  m_enabled = en;
  if (! en) {
    widget ()->ungrab_mouse (this);
  }
}

// ---------------------------------------------------------------
//  ViewObject implementation

ViewObjectWidget::ViewObjectWidget (QWidget *parent, const char *name)
  : QWidget (parent, name), 
    m_objects (this), m_services (this),
    m_needs_update_static (false),
    m_needs_update_bg (false),
    mp_active_service (0),
    m_mouse_pressed_state (false),
    m_mouse_buttons (0),
    m_in_mouse_move (false)
{
  setMouseTracking (true); 
}

ViewObjectWidget::~ViewObjectWidget ()
{
  //  release any mouse grabs now
  while (m_grabbed.begin () != m_grabbed.end ()) {
    ungrab_mouse (*m_grabbed.begin ());
  }
}

void
ViewObjectWidget::activate (lay::ViewService *service)
{
  if (mp_active_service != service) {
    if (mp_active_service) {
BEGIN_PROTECTED  
      mp_active_service->deactivated ();
END_PROTECTED
    }
    mp_active_service = service;
    if (mp_active_service) {
BEGIN_PROTECTED  
      mp_active_service->activated ();
END_PROTECTED
    }
  }
}

void 
ViewObjectWidget::mouseMoveEvent (QMouseEvent *e)
{
BEGIN_PROTECTED  
  m_mouse_pos = e->pos ();
  do_mouse_move (m_mouse_pos);
END_PROTECTED
}

void
ViewObjectWidget::do_mouse_move (const QPoint &qp)
{
  m_in_mouse_move = true;

  if (m_mouse_pressed_state && 
    (abs (qp.x () - m_mouse_pressed.x ()) > click_tolerance || abs (qp.y () - m_mouse_pressed.y ()) > click_tolerance)) {

    m_mouse_pressed_state = false;

    bool done = false; 
    
    db::DPoint p = m_trans.inverted () * db::DPoint (m_mouse_pressed.x (), height () - 1 - m_mouse_pressed.y ());

    for (std::list<ViewService *>::iterator g = m_grabbed.begin (); !done && g != m_grabbed.end (); ) {
      std::list<ViewService *>::iterator gg = g;
      ++gg;
      done = ((*g)->m_enabled && (*g)->mouse_press_event (p, m_mouse_buttons, true));
      g = gg;
    }

    if (! done && mp_active_service) {
      done = (mp_active_service->m_enabled && mp_active_service->mouse_press_event (p, m_mouse_buttons, true));
    }

    service_iterator svc = begin_services ();
    while (svc != end_services () && !done) {
      service_iterator next = svc;
      ++next;
      done = (svc->m_enabled && svc->mouse_press_event (p, m_mouse_buttons, false));
      svc = next;
    }

    if (! done) {
      mouse_press_event (p, m_mouse_buttons);
    }

  }

  if (! m_mouse_pressed_state) {

    bool done = false; 

    db::DPoint p = m_trans.inverted () * db::DPoint (qp.x (), height () - 1 - qp.y ());

    for (std::list<ViewService *>::iterator g = m_grabbed.begin (); !done && g != m_grabbed.end (); ) {
      std::list<ViewService *>::iterator gg = g;
      ++gg;
      done = ((*g)->m_enabled && (*g)->mouse_move_event (p, true));
      g = gg;
    }

    if (! done && mp_active_service) {
      done = (mp_active_service->m_enabled && mp_active_service->mouse_move_event (p, true));
    }

    service_iterator svc = begin_services ();
    while (svc != end_services () && !done) {
      service_iterator next = svc;
      ++next;
      done = (svc->m_enabled && svc->mouse_move_event (p, false));
      svc = next;
    }

    if (! done) {
      mouse_move_event (p);
    }

  }

  m_in_mouse_move = false;
}

static unsigned int 
qt_to_buttons (Qt::MouseButtons b, Qt::KeyboardModifiers m)
{
  return ((b & Qt::LeftButton) ? LeftButton : 0) |
         ((b & Qt::MidButton) ? MidButton : 0) |
         ((b & Qt::RightButton) ? RightButton : 0) |
         ((m & Qt::ShiftModifier) != 0 ? ShiftButton : 0) |
         ((m & Qt::ControlModifier) != 0 ? ControlButton : 0) |
         ((m & Qt::AltModifier) != 0 ? AltButton : 0);
}

void 
ViewObjectWidget::mouseDoubleClickEvent (QMouseEvent *e)
{
BEGIN_PROTECTED  
  setFocus ();

  bool done = false; 

  m_mouse_pos = e->pos ();
  m_mouse_pressed = e->pos ();
  m_mouse_pressed_state = false;

  unsigned int buttons = qt_to_buttons (e->buttons (), e->modifiers ());

  db::DPoint p = m_trans.inverted () * db::DPoint (e->pos ().x (), height () - 1 - e->pos ().y ());

  for (std::list<ViewService *>::iterator g = m_grabbed.begin (); !done && g != m_grabbed.end (); ) {
    std::list<ViewService *>::iterator gg = g;
    ++gg;
    done = ((*g)->m_enabled && (*g)->mouse_double_click_event (p, buttons, true));
    g = gg;
  }

  if (! done && mp_active_service) {
    done = (mp_active_service->m_enabled && mp_active_service->mouse_double_click_event (p, buttons, true));
  }

  service_iterator svc = begin_services ();
  while (svc != end_services () && !done) {
    service_iterator next = svc;
    ++next;
    done = (svc->m_enabled && svc->mouse_double_click_event (p, buttons, false));
    svc = next;
  }

  if (! done) {
    mouse_double_click_event (p, buttons);
  }
END_PROTECTED
}

void 
ViewObjectWidget::wheelEvent (QWheelEvent *e)
{
BEGIN_PROTECTED  
  e->ignore ();

  bool done = false; 

  unsigned int buttons = qt_to_buttons (e->buttons (), e->modifiers ());
  bool horizonal = (e->orientation () == Qt::Horizontal);

  db::DPoint p = m_trans.inverted () * db::DPoint (e->pos ().x (), height () - 1 - e->pos ().y ());

  for (std::list<ViewService *>::iterator g = m_grabbed.begin (); !done && g != m_grabbed.end (); ) {
    std::list<ViewService *>::iterator gg = g;
    ++gg;
    done = ((*g)->m_enabled && (*g)->wheel_event (e->delta (), horizonal, p, buttons, true));
    g = gg;
  }

  if (! done && mp_active_service) {
    done = (mp_active_service->m_enabled && mp_active_service->wheel_event (e->delta (), horizonal, p, buttons, true));
  }

  service_iterator svc = begin_services ();
  while (svc != end_services () && !done) {
    service_iterator next = svc;
    ++next;
    done = (svc->m_enabled && svc->wheel_event (e->delta (), horizonal, p, buttons, false));
    svc = next;
  }

  if (! done) {
    wheel_event (e->delta (), horizonal, p, buttons);
  }
END_PROTECTED
}

void 
ViewObjectWidget::mousePressEvent (QMouseEvent *e)
{
  setFocus ();

  m_mouse_pos = e->pos ();
  m_mouse_pressed = e->pos ();

  m_mouse_buttons = qt_to_buttons (e->buttons (), e->modifiers ());
  
  m_mouse_pressed_state = true;
}

void 
ViewObjectWidget::mouseReleaseEvent (QMouseEvent *e)
{
BEGIN_PROTECTED  
  bool done = false; 

  m_mouse_pos = e->pos ();
  db::DPoint p = m_trans.inverted () * db::DPoint (e->pos ().x (), height () - 1 - e->pos ().y ());

  for (std::list<ViewService *>::iterator g = m_grabbed.begin (); !done && g != m_grabbed.end (); ) {
    std::list<ViewService *>::iterator gg = g;
    ++gg;
    if ((*g)->m_enabled) {
      if (m_mouse_pressed_state) {
        done = (*g)->mouse_click_event (p, m_mouse_buttons, true);
      } else {
        done = (*g)->mouse_release_event (p, true);
      }
    }
    g = gg;
  }

  if (! done && mp_active_service && mp_active_service->m_enabled) {
    if (m_mouse_pressed_state) {
      done = mp_active_service->mouse_click_event (p, m_mouse_buttons, true);
    } else {
      done = mp_active_service->mouse_release_event (p, true);
    }
  }

  service_iterator svc = begin_services ();
  while (svc != end_services () && !done) {
    service_iterator next = svc;
    ++next;
    if (svc->m_enabled) {
      if (m_mouse_pressed_state) {
        done = svc->mouse_click_event (p, m_mouse_buttons, false);
      } else {
        done = svc->mouse_release_event (p, false);
      }
    }
    svc = next;
  }

  if (! done) {
    if (m_mouse_pressed_state) {
      mouse_click_event (p, m_mouse_buttons);
    } else {
      mouse_release_event (p);
    }
  }

END_PROTECTED

  m_mouse_pressed_state = false;
}

void
ViewObjectWidget::mouse_event_trans (const db::DCplxTrans &trans)
{
  if (trans != m_trans) {
    m_trans = trans;
    //  issue a move event in order to reposition the mouse in the new coordinate system
    //  since this may be called from within a mouse move event handler we need the recursion sentinel
    if (! m_in_mouse_move) {
      do_mouse_move (m_mouse_pos);
    }
  }
}

void 
ViewObjectWidget::drag_cancel ()
{
  for (service_iterator svc = begin_services (); svc != end_services (); ++svc) {
    svc->drag_cancel ();
  }
}

void 
ViewObjectWidget::do_render_bg (const lay::Viewport &vp, lay::ViewObjectCanvas &canvas)
{
  m_needs_update_bg = false;

  for (object_iterator obj = begin_objects (); obj != end_objects (); ++obj) {
    if (obj->is_visible ()) {
      obj->render_bg (vp, canvas);
    }
  }
}

void 
ViewObjectWidget::do_render (const lay::Viewport &vp, lay::ViewObjectCanvas &canvas, bool st)
{
  if (st) {
    m_needs_update_static = false;
  }

  for (object_iterator obj = begin_objects (); obj != end_objects (); ++obj) {
    if (obj->m_static == st && obj->is_visible ()) {
      obj->render (vp, canvas);
    }
  }

  canvas.sort_bitmaps ();
}

void 
ViewObjectWidget::grab_mouse (ViewService *obj, bool a)
{
  obj->m_abs_grab = a;
  m_grabbed.push_front (obj);
  if (a) {
    grabMouse ();
  }
}

void 
ViewObjectWidget::ungrab_mouse (ViewService *obj)
{
  bool any_abs_grabbed = false;

  for (std::list<ViewService *>::iterator g = m_grabbed.begin (); g != m_grabbed.end (); ) {
    std::list<ViewService *>::iterator gg = g;
    ++gg;
    if (*g == obj) {
      m_grabbed.erase (g);
    } else if ((*g)->m_abs_grab) {
      any_abs_grabbed = true;
    }
    g = gg;
  }

  if (! any_abs_grabbed && mouseGrabber () == this) {
    releaseMouse ();
  }
}

void 
ViewObjectWidget::freeze (ViewObject *obj)
{
  if (! obj->m_static) {
    obj->m_static = true;
    m_needs_update_static = true;
    //  no update needed since the display will not change through this.
  }
}

void 
ViewObjectWidget::thaw (ViewObject *obj)
{
  if (obj->m_static) {
    obj->m_static = false;
    m_needs_update_static = true;
    //  no update needed since the display will not change through this.
  }
}

void
ViewObjectWidget::touch ()
{
  if (! m_needs_update_static) {
    m_needs_update_static = true;
    update ();
  }
}

void
ViewObjectWidget::touch_bg ()
{
  if (! m_needs_update_bg) {
    m_needs_update_bg = true;
    update ();
  }
}

void 
ViewObjectWidget::added (tl::CollectionBase *coll, tl::Collectable *)
{
  if (coll == & m_objects) {
    touch ();
  }
}

void 
ViewObjectWidget::removed (tl::CollectionBase *coll, tl::Collectable *obj)
{
  if (coll == & m_objects) {
    touch ();
    update ();
  }
  if (coll == & m_services) {
    if (mp_active_service == obj) {
      mp_active_service = 0;
    }
  }
}

db::DBox 
ViewObjectWidget::mouse_event_viewport () const
{
  db::DPoint p1 = m_trans.inverted () * db::DPoint (0, 0);
  db::DPoint p2 = m_trans.inverted () * db::DPoint (width () - 1, height () - 1);
  return db::DBox (p1, p2);
}

// ---------------------------------------------------------------
//  ViewObjectCanvas implementation

ViewObjectCanvas::~ViewObjectCanvas ()
{
  clear_fg_bitmaps ();
}

lay::Bitmap *
ViewObjectCanvas::bitmap (const lay::ViewOp &style, unsigned int width, unsigned int height)
{
  std::map <lay::ViewOp, unsigned int>::iterator b = m_fg_bitmap_table.find (style);
  if (b == m_fg_bitmap_table.end ()) {

    //  we need to create a new bitmap
    m_fg_bitmap_table.insert (std::make_pair (style, mp_alloc_bitmaps.size ()));
    lay::Bitmap *bm = new lay::Bitmap (width, height);
    mp_fg_bitmaps.push_back (bm);
    mp_alloc_bitmaps.push_back (bm);
    m_fg_view_ops.push_back (style);
    return bm;
    
  } else {
    //  we can recycle a current one
    return mp_alloc_bitmaps [b->second];
  }
}

lay::Bitmap *
ViewObjectCanvas::bitmap (const std::vector<lay::ViewOp> &style, unsigned int width, unsigned int height)
{
  std::map <std::vector<lay::ViewOp>, unsigned int>::iterator b = m_fgv_bitmap_table.find (style);
  if (b == m_fgv_bitmap_table.end ()) {

    //  we need to create a new bitmap
    m_fgv_bitmap_table.insert (std::make_pair (style, mp_alloc_bitmaps.size ()));
    lay::Bitmap *bm = new lay::Bitmap (width, height);
    mp_alloc_bitmaps.push_back (bm);
    for (std::vector<lay::ViewOp>::const_iterator s = style.begin (); s != style.end (); ++s) {
      mp_fg_bitmaps.push_back (bm);
      m_fg_view_ops.push_back (*s);
    }
    return bm;
    
  } else {
    //  we can recycle a current one
    return mp_alloc_bitmaps [b->second];
  }
}

void 
ViewObjectCanvas::clear_fg_bitmaps ()
{
  for (std::vector <lay::Bitmap *>::iterator i = mp_alloc_bitmaps.begin (); i != mp_alloc_bitmaps.end (); ++i) {
    if (*i) {
      delete *i;
    }
  }
  mp_alloc_bitmaps.clear ();
  mp_fg_bitmaps.clear ();
  m_fg_view_ops.clear ();
  m_fg_bitmap_table.clear ();
  m_fgv_bitmap_table.clear ();
}

void
ViewObjectCanvas::sort_bitmaps ()
{
  //  sort the planes by view operator - this ensures a certain plane order as implied by
  //  the bitmap index of the plane operators.
  std::vector <std::pair <lay::ViewOp, lay::Bitmap *> > planes;
  planes.reserve (mp_fg_bitmaps.size ());
  for (unsigned int i = 0; i < mp_fg_bitmaps.size (); ++i) {
    planes.push_back (std::make_pair (m_fg_view_ops [i], mp_fg_bitmaps [i]));
  }
  std::sort (planes.begin (), planes.end ());
  for (unsigned int i = 0; i < mp_fg_bitmaps.size (); ++i) {
    m_fg_view_ops [i] = planes [i].first;
    mp_fg_bitmaps [i] = planes [i].second;
  }
}


}

