
/*

  KLayout Layout Viewer
  Copyright (C) 2006-2007 Matthias Koefferlein

  This program is free software; you can redistribute it and/or modify
  it under the terms of the GNU General Public License as published by
  the Free Software Foundation; either version 2 of the License, or
  (at your option) any later version.

  This program is distributed in the hope that it will be useful,
  but WITHOUT ANY WARRANTY; without even the implied warranty of
  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
  GNU General Public License for more details.

  You should have received a copy of the GNU General Public License
  along with this program; if not, write to the Free Software
  Foundation, Inc., 51 Franklin St, Fifth Floor, Boston, MA  02110-1301  USA

*/


#include "gsi.h"
#include "rba.h"
#include "tlString.h"
#include "tlInternational.h"
#include "tlException.h"
#include "tlAssert.h"

#include <map>
#include <set>
#include <memory>
#include <iostream>

// -------------------------------------------------------------------
//  RubyError implementation

std::string 
RubyError::msg () const
{
  std::string m (tl::translate ("Ruby error"));
  m += ": '";

  m += basic_msg ();

  m += tl::translate ("' (");
  m += cls ();
  m += ")";

  for (std::vector<std::string>::const_iterator bt = backtrace ().begin (); bt != backtrace ().end (); ++bt) {
    m += "\n  ";
    m += *bt;
  }

  return m;
}

// -------------------------------------------------------------------
//  This part is available only if Ruby is enabled

#ifdef HAVE_RUBY

#include "ruby.h"
#include "signal.h"

static std::map <VALUE, const gsi::ClassBase *> cls_map;
static std::map <const gsi::ClassBase *, VALUE> rev_cls_map;

// -------------------------------------------------------------------
//  Utilities to generate C++ exceptions from Ruby exceptions

static void
rba_get_backtrace (VALUE exception, std::vector<std::string> &bt)
{
  VALUE backtrace = rb_funcall (exception, rb_intern ("backtrace"), 0);

  if (TYPE (backtrace) == T_ARRAY) {
    unsigned int len = RARRAY (backtrace)->len;
    VALUE *el = RARRAY(backtrace)->ptr;
    bt.reserve (len);
    while (len-- > 0) {
      VALUE str = StringValue (*el++);
      bt.push_back (std::string (RSTRING (str)->ptr, RSTRING (str)->len));
    }
  }
}

static void 
rba_check_error_eval ()
{
  VALUE lasterr = rb_gv_get ("$!");
  if (lasterr != Qnil) {

    rb_gv_set ("$!", Qnil);

    VALUE klass = rb_class_path (CLASS_OF (lasterr));
    std::string eclass = std::string (RSTRING (klass)->ptr, 0, RSTRING (klass)->len);
    VALUE message = rb_obj_as_string(lasterr);
    std::string emsg = std::string (RSTRING (message)->ptr, 0, RSTRING (message)->len);

    std::vector <std::string> bt;
    rba_get_backtrace (lasterr, bt);

    throw RubyError (emsg.c_str (), eclass.c_str (), bt);

  }
}

static void
rba_check_error ()
{
  VALUE lasterr = rb_gv_get ("$!");
  if (lasterr != Qnil) {

    rb_gv_set ("$!", Qnil);

    VALUE klass = rb_class_path (CLASS_OF (lasterr));
    std::string eclass = std::string (RSTRING (klass)->ptr, 0, RSTRING (klass)->len);
    VALUE message = rb_obj_as_string(lasterr);
    std::string emsg = std::string (RSTRING (message)->ptr, 0, RSTRING (message)->len);

    ruby_set_current_source ();

    std::vector <std::string> bt;
    rba_get_backtrace (lasterr, bt);

    if (ruby_sourcefile) {
      throw RubyError (emsg.c_str (), ruby_sourcefile, ruby_sourceline, eclass.c_str (), bt);
    } else {
      throw RubyError (emsg.c_str (), eclass.c_str (), bt);
    }

  }
}

//  needed because StringValue is a macro:
static VALUE
rba_string_value_f (VALUE obj)
{
  return StringValue (obj);
}

//  string value retrieval with check 
static VALUE
rba_safe_string_value (VALUE obj)
{
  int error = 0;
  VALUE ret = rb_protect (rba_string_value_f, obj, &error);
  if (error) {
    rba_check_error ();
  }
  return ret;
}

// -------------------------------------------------------------------
//  The GSI object proxy stored within the Ruby object

class Proxy 
  : public gsi::Callee
{
public:
  struct VirtualFunction
  {
    VirtualFunction (ID id, VALUE o, const gsi::MethodBase *m)
      : method_id (id), rb_obj (o), method (m)
    {
      //  .. nothing yet ..
    }

    ID method_id;
    VALUE rb_obj;
    const gsi::MethodBase *method;
  };

  Proxy (const gsi::ClassBase *_cls_decl)
    : m_cls_decl (_cls_decl), m_obj (0), m_owned (false), m_const_ref (false), m_destroyed (false)
  { }

  ~Proxy ();

  void destroy ()
  {
    //  first create the object if it was not created yet and check if it has not been 
    //  destroyed already (the former is to ensure that the object is created at least)
    if (! m_obj) {
      if (m_destroyed) {
        throw tl::Exception (tl::translate ("Object has been destroyed already"));
      } else {
        m_obj = cls_decl ()->create ();
        m_owned = true;
      }
    }
    set (0, false, false);
    m_destroyed = true;
  }

  void set (void *obj, bool owned, bool const_ref) 
  {
    if (m_owned && m_obj && m_cls_decl) {
      m_cls_decl->destroy (m_obj);
      m_obj = 0;
    }
    m_owned = owned;
    m_const_ref = const_ref;
    m_obj = obj;
  }

  void detach ()
  {
    m_obj = 0;
    m_destroyed = true;
    m_const_ref = false;
    m_owned = false;
  }

  const gsi::ClassBase *cls_decl () const 
  {
    return m_cls_decl;
  }

  bool destroyed () const 
  {
    return m_destroyed;
  }

  bool const_ref () const 
  {
    return m_const_ref;
  }

  void *obj (VALUE self);

  bool owned () const 
  {
    return m_owned;
  }

  int add_callback (const VirtualFunction &vf)
  {
    m_vfuncs.push_back (vf);
    return int (m_vfuncs.size () - 1);
  }

  virtual void call (int id, gsi::SerialArgs &args, gsi::SerialArgs &ret);

private:
  const gsi::ClassBase *m_cls_decl;
  void *m_obj;
  bool m_owned;
  bool m_const_ref;
  bool m_destroyed;

  std::vector<VirtualFunction> m_vfuncs; 
};

// --------------------------------------------------------------------------
//  Ruby client data 

class RBAClientData
{
public:
  RBAClientData ()
    : self (Qnil)
  {
    //  .. nothing yet ..
  }

  ~RBAClientData () 
  {
    if (self != Qnil) {

      //  Compare if the classes are identical
      Proxy *p;
      Data_Get_Struct (self, Proxy, p);
      p->detach ();

      //  once the object is destroyed, we may release the Ruby reference
      rb_gc_unregister_address (&self);

    }
  }

  void reset_ref ()
  {
    self = Qnil;
  }

  void set_ref (VALUE s)
  {
    tl_assert (self == Qnil);
    tl_assert (s != Qnil);
    self = s;

    //  this avoids that our Ruby reference is deleted by the GC
    rb_gc_register_address (&self);
  }

  VALUE get_ref () const
  {
    return self;
  }

private:
  VALUE self;
};

// --------------------------------------------------------------------------
//  Ruby client data handler

class RBAClientHandler
  : public gsi::ClientHandler
{
public:
  virtual void attach (void *&data) 
  {
    data = new RBAClientData ();
  }

  virtual void detach (void *&data) 
  {
    delete (RBAClientData *) data;
  }
};

// -------------------------------------------------------------------
//  Type checks 
//
//  The type checks are somewhat more picky than the ruby2c functions.
//  The type checks are used to resolve overridden methods, so being picky
//  might be more appropriate.

template <class T>
struct test_type;

template <>
struct test_type<bool>
{
  static bool test (VALUE rval)
  {
    return true;  // everything can be converted to bool
  }
};

//  used for other integer types as well:
template <>
struct test_type<int>
{
  static bool test (VALUE rval)
  {
    unsigned int t = TYPE (rval);
    return (t == T_FIXNUM || t == T_BIGNUM);
  }
};

template <>
struct test_type<double>
{
  static bool test (VALUE rval)
  {
    unsigned int t = TYPE (rval);
    return (t == T_FLOAT);
  }
};

//  used for strings in general:
template <>
struct test_type<const char *>
{
  static bool test (VALUE rval)
  {
    return TYPE (rval) == T_STRING;
  }
};

template <>
struct test_type<tl::Variant>
{
  static bool test (VALUE rval)
  {
    unsigned int t = TYPE (rval);
    return (t == T_FALSE || t == T_TRUE || t == T_NIL || t == T_FIXNUM || t == T_BIGNUM || t == T_FLOAT || t == T_STRING || t == T_ARRAY /*be optimistic*/);
  }
};

// -------------------------------------------------------------------
//  Ruby to C conversion

template <class T>
struct ruby2c;

template <>
struct ruby2c<bool>
{
  static bool get (VALUE rval)
  {
    return bool (RTEST (rval));
  }
};

template <>
struct ruby2c<int>
{
  static int get (VALUE rval)
  {
    return NUM2INT (rval);
  }
};

template <>
struct ruby2c<unsigned int>
{
  static unsigned int get (VALUE rval)
  {
    return NUM2UINT (rval);
  }
};

template <>
struct ruby2c<long>
{
  static long get (VALUE rval)
  {
    return NUM2LONG (rval);
  }
};

template <>
struct ruby2c<unsigned long>
{
  static unsigned long get (VALUE rval)
  {
    return NUM2ULONG (rval);
  }
};

template <>
struct ruby2c<long long>
{
  static long long get (VALUE rval)
  {
    //  TODO: this is loosing precision!
    return (long long) NUM2DBL (rval);
  }
};

template <>
struct ruby2c<double>
{
  static double get (VALUE rval)
  {
    return NUM2DBL (rval);
  }
};

template <>
struct ruby2c<std::string>
{
  static std::string get (VALUE rval)
  {
    VALUE str = rba_safe_string_value (rval);
    return std::string (RSTRING (str)->ptr, RSTRING (str)->len);
  }
};

template <>
struct ruby2c<const char *>
{
  static const char *get (VALUE rval)
  {
    VALUE str = rba_safe_string_value (rval);
    return RSTRING (str)->ptr;
  }
};

template <>
struct ruby2c<tl::Variant>
{
  static tl::Variant get (VALUE rval)
  {
    if (FIXNUM_P (rval)) {
      return tl::Variant (ruby2c<long>::get (rval));
    } else if (rval == Qnil) {
      return tl::Variant ();
    } else if (rval == Qfalse) {
      return tl::Variant (0l);
    } else if (rval == Qtrue) {
      return tl::Variant (1l);
    } else if (TYPE (rval) == T_BIGNUM || TYPE (rval) == T_FLOAT) {
      return tl::Variant (ruby2c<double>::get (rval));
    } else if (TYPE (rval) == T_ARRAY) {

      unsigned int len = RARRAY (rval)->len;
      VALUE *el = RARRAY(rval)->ptr;

      static std::vector<tl::Variant> empty;
      tl::Variant r (empty.begin (), empty.end ());
      r.get_list ().reserve (len);
      while (len-- > 0) {
        r.get_list ().push_back (ruby2c<tl::Variant>::get (*el++));
      }
      return r;

    } else {
      return tl::Variant (ruby2c<const char *>::get (rval));
    }
  }
};

template <class R>
struct callarg
{
  static void set_value (gsi::SerialArgs &aa, VALUE arg, bool ref)
  {
    if (ref) {
      aa.set_value (typename gsi::type_traits<const R &>::tag (), ruby2c<R>::get (arg));
    } else {
      aa.set_value (typename gsi::type_traits<R>::tag (), ruby2c<R>::get (arg));
    }
  }

  static bool
  test_vector (VALUE arr)
  {
    if (TYPE (arr) != T_ARRAY) {
      return false;
    }

    unsigned int len = RARRAY (arr)->len;
    VALUE *el = RARRAY(arr)->ptr;
    while (len-- > 0) {
      if (! test_type<R>::test (*el++)) {
        return false;
      }
    }

    return true;
  }

  static void
  set_vector (gsi::SerialArgs &aa, VALUE arr, bool ref, std::list<gsi::TempObject> &tmp_stack)
  {
    Check_Type (arr, T_ARRAY);
    unsigned int len = RARRAY (arr)->len;
    VALUE *el = RARRAY(arr)->ptr;

    if (ref) {

      tmp_stack.push_back (gsi::TempObject ());

      std::vector<R> *v = new std::vector<R> ();
      v->reserve (len);
      while (len-- > 0) {
        v->push_back (ruby2c<R>::get (*el++));
      }
      tmp_stack.back ().set (v);

      aa.set_value (typename gsi::type_traits<const std::vector<R> &>::tag (), *v);

    } else {
      std::vector<R> *v = aa.set_value (typename gsi::type_traits< std::vector<R> >::tag (), std::vector<R> ());
      v->reserve (len);
      while (len-- > 0) {
        v->push_back (ruby2c<R>::get (*el++));
      }
    }
  }
};

inline void
callarg_string (gsi::SerialArgs &aa, VALUE arg, bool ref, std::list<gsi::TempObject> &tmp_stack)
{
  VALUE str = rba_safe_string_value (arg);

  if (ref) {

    tmp_stack.push_back (gsi::TempObject ());
    std::string *v = new std::string (RSTRING (str)->ptr, RSTRING (str)->len);
    tmp_stack.back ().set (v);

    aa.set_value (gsi::type_traits<const std::string &>::tag (), *v);

  } else {
    aa.set_value (gsi::type_traits<std::string>::tag (), std::string (RSTRING (str)->ptr, RSTRING (str)->len));
  }
}

inline void
callarg_var (gsi::SerialArgs &aa, VALUE arg, bool ref, std::list<gsi::TempObject> &tmp_stack)
{
  if (ref) {

    tmp_stack.push_back (gsi::TempObject ());
    tl::Variant *v = new tl::Variant (ruby2c<tl::Variant>::get (arg));
    tmp_stack.back ().set (v);

    aa.set_value (gsi::type_traits<const tl::Variant &>::tag (), *v);

  } else {
    aa.set_value (gsi::type_traits<tl::Variant>::tag (), ruby2c<tl::Variant>::get (arg));
  }
}

// -------------------------------------------------------------------
//  C to Ruby conversion

template <class R>
struct c2ruby;

template <>
struct c2ruby<bool>
{
  static VALUE get (bool c)
  {
    return c ? Qtrue : Qfalse;
  }
};

template <>
struct c2ruby<int>
{
  static VALUE get (int c)
  {
    return INT2NUM (c);
  }
};

template <>
struct c2ruby<unsigned int>
{
  static VALUE get (unsigned int c)
  {
    return UINT2NUM (c);
  }
};

template <>
struct c2ruby<long>
{
  static VALUE get (long c)
  {
    return LONG2NUM (c);
  }
};

template <>
struct c2ruby<unsigned long>
{
  static VALUE get (unsigned long c)
  {
    return ULONG2NUM (c);
  }
};

template <>
struct c2ruby<long long>
{
  static VALUE get (long long c)
  {
    //  TODO: this is loosing precision!
    return rb_float_new (double (c));
  }
};

template <>
struct c2ruby<double>
{
  static VALUE get (double c)
  {
    return rb_float_new (c);
  }
};

template <>
struct c2ruby<std::string>
{
  static VALUE get (const std::string &c)
  {
    return rb_str_new (c.c_str (), c.size ());
  }
};

template <>
struct c2ruby<tl::Variant>
{
  static VALUE get (const tl::Variant &c)
  {
    if (c.is_double ()) {
      return c2ruby<double>::get (c.to_double ());
    } else if (c.is_string ()) {
      return c2ruby<std::string>::get (c.to_string ());
    } else if (c.is_long ()) {
      return c2ruby<long>::get (c.to_long ());
    } else if (c.is_list ()) {
      VALUE ret = rb_ary_new ();
      for (std::vector<tl::Variant>::const_iterator i = c.get_list ().begin (); i != c.get_list ().end (); ++i) {
        rb_ary_push (ret, c2ruby<tl::Variant>::get (*i));
      }
      return ret;
    } else {
      return Qnil;
    }
  }
};

template <>
struct c2ruby<const char *>
{
  static VALUE get (const char *s)
  {
    return rb_str_new (s, strlen (s));
  }
};

template <class R> 
struct retargs
{
  static VALUE 
  get_value (gsi::SerialArgs &rr, bool ref)
  {
    if (ref) {
      return c2ruby<R>::get (rr.template get_value<const R &> (typename gsi::type_traits<const R &>::tag ()));
    } else {
      return c2ruby<R>::get (rr.template get_value<R> (typename gsi::type_traits<R>::tag ()));
    }
  }

  static VALUE 
  get_vector (gsi::SerialArgs &rr, bool ref)
  {
    if (ref) {

      const std::vector<R> &v = rr.template get_value<const std::vector<R> &> (typename gsi::type_traits<const std::vector<R> &>::tag ());

      VALUE ret = rb_ary_new ();
      for (typename std::vector<R>::const_iterator i = v.begin (); i != v.end (); ++i) {
        rb_ary_push (ret, c2ruby<R>::get (*i));
      }
      return ret;

    } else {

      std::vector<R> v = rr.template get_value< std::vector<R> > (typename gsi::type_traits< std::vector<R> >::tag ());

      VALUE ret = rb_ary_new ();
      for (typename std::vector<R>::const_iterator i = v.begin (); i != v.end (); ++i) {
        rb_ary_push (ret, c2ruby<R>::get (*i));
      }
      return ret;

    }
  }
};

inline VALUE 
ret_value_string (gsi::SerialArgs &rr, bool ref)
{
  if (ref) {
    const std::string &s = rr.get_value<const std::string &> (gsi::type_traits<const std::string &>::tag ());
    return c2ruby<std::string>::get (s);
  } else {
    return c2ruby<std::string>::get (rr.get_value<std::string> (gsi::type_traits<std::string>::tag ()));
  }
}

inline VALUE 
ret_value_var (gsi::SerialArgs &rr, bool ref)
{
  if (ref) {
    const tl::Variant &s = rr.get_value<const tl::Variant &> (gsi::type_traits<const tl::Variant &>::tag ());
    return c2ruby<tl::Variant>::get (s);
  } else {
    return c2ruby<tl::Variant>::get (rr.get_value<tl::Variant> (gsi::type_traits<tl::Variant>::tag ()));
  }
}

// -------------------------------------------------------------------
//  Test if an argument can be converted to the given type

bool
test_arg (const gsi::ArgType &atype, VALUE arg)
{
  if (atype.vector) {

    switch (atype.type) {
    case gsi::T_void:
      tl_assert (false);
      break;
    case gsi::T_bool:     
      return callarg<bool>::test_vector (arg);
    case gsi::T_int:      
    case gsi::T_uint:     
    case gsi::T_long:     
    case gsi::T_ulong:    
    case gsi::T_longlong:     
      return callarg<int>::test_vector (arg);
    case gsi::T_double:   
      return callarg<double>::test_vector (arg);
    case gsi::T_string:   
      return callarg<const char *>::test_vector (arg);
    case gsi::T_var:   
      return callarg<tl::Variant>::test_vector (arg);
    case gsi::T_object:
      {
        if (TYPE (arg) != T_ARRAY) {
          return false;
        }
        unsigned int len = RARRAY (arg)->len;
        VALUE *el = RARRAY(arg)->ptr;
        while (len-- > 0) {
          if (TYPE (*el) != T_DATA) {
            return false;
          }
          Proxy *p;
          Data_Get_Struct (*el, Proxy, p);
          if (!p->cls_decl ()->is_derived_from (atype.cls)) {
            return false;
          }
          ++el;
        }
      }
      break;
    case gsi::T_object_ref:
    case gsi::T_object_cref:
    case gsi::T_object_new:
    case gsi::T_string_ccptr:   
      tl_assert (false);  // @@@ not implemented yet.
      break;
    }

  } else {

    switch (atype.type) {
    case gsi::T_void:
      break;
    case gsi::T_bool:
      return test_type<bool>::test (arg);
    case gsi::T_int:
    case gsi::T_uint:
    case gsi::T_long:
    case gsi::T_ulong:
    case gsi::T_longlong:
      return test_type<int>::test (arg);
    case gsi::T_double:
      return test_type<double>::test (arg);
    case gsi::T_string_ccptr:
    case gsi::T_string:
      return test_type<const char *>::test (arg);
    case gsi::T_var:
      return test_type<tl::Variant>::test (arg);
    case gsi::T_object:
      {
        if (TYPE (arg) != T_DATA) {
          return false;
        }
        Proxy *p;
        Data_Get_Struct (arg, Proxy, p);
        if (!p->cls_decl ()->is_derived_from (atype.cls)) {
          return false;
        }
      }
      break;
    case gsi::T_object_ref:
    case gsi::T_object_cref:
      {
        if (arg != Qnil) {
          if (TYPE (arg) != T_DATA) {
            return false;
          }
          Proxy *p;
          Data_Get_Struct (arg, Proxy, p);
          if (!p->cls_decl ()->is_derived_from (atype.cls)) {
            return false;
          }
          if (p->const_ref () && atype.type == gsi::T_object_ref) {
            return false;
          }
        }
      }
      break;
    case gsi::T_object_new:
      tl_assert (false);  // An argument cannot be "new"
      break;
    }

  }

  return true;
}
      
// -------------------------------------------------------------------
//  Push an argument on the call or return stack

void
push_arg (const gsi::ArgType &atype, gsi::SerialArgs &aserial, VALUE arg, std::list<gsi::TempObject> &tmp_stack)
{
  if (atype.vector) {

    switch (atype.type) {
    case gsi::T_void:
      tl_assert (false);
      break;
    case gsi::T_bool:     
      callarg<bool>::set_vector (aserial, arg, atype.ref, tmp_stack); 
      break;
    case gsi::T_int:      
      callarg<int>::set_vector (aserial, arg, atype.ref, tmp_stack); 
      break;
    case gsi::T_uint:     
      callarg<unsigned int>::set_vector (aserial, arg, atype.ref, tmp_stack); 
      break;
    case gsi::T_long:     
      callarg<long>::set_vector (aserial, arg, atype.ref, tmp_stack); 
      break;
    case gsi::T_ulong:    
      callarg<unsigned long>::set_vector (aserial, arg, atype.ref, tmp_stack); 
      break;
    case gsi::T_longlong:     
      callarg<long long>::set_vector (aserial, arg, atype.ref, tmp_stack); 
      break;
    case gsi::T_double:   
      callarg<double>::set_vector (aserial, arg, atype.ref, tmp_stack); 
      break;
    case gsi::T_string:   
      callarg<std::string>::set_vector (aserial, arg, atype.ref, tmp_stack); 
      break;
    case gsi::T_var:   
      callarg<tl::Variant>::set_vector (aserial, arg, atype.ref, tmp_stack); 
      break;
    case gsi::T_object:
      {
        Check_Type (arg, T_ARRAY);
        unsigned int len = RARRAY (arg)->len;
        VALUE *el = RARRAY(arg)->ptr;

        std::vector<void *> vobj;
        vobj.reserve (len);
        while (len-- > 0) {
          Proxy *p;
          Data_Get_Struct (*el, Proxy, p);
          if (!p->cls_decl ()->is_derived_from (atype.cls)) {
            throw tl::Exception (tl::sprintf (tl::translate ("Unexpected object type (expected argument of class %s)").c_str (), atype.cls->name ().c_str ()));
          }
          vobj.push_back (p->obj (*el));
          ++el;
        }
        atype.cls->push_vector_of (aserial, atype.ref, tmp_stack, vobj);
      }
      break;
    case gsi::T_object_ref:
    case gsi::T_object_cref:
    case gsi::T_object_new:
    case gsi::T_string_ccptr:   
      tl_assert (false);  // @@@ not implemented yet.
      break;
    }

  } else {

    switch (atype.type) {
    case gsi::T_void:
      break;
    case gsi::T_bool:
      callarg<bool>::set_value (aserial, arg, atype.ref);
      break;
    case gsi::T_int:
      callarg<int>::set_value (aserial, arg, atype.ref);
      break;
    case gsi::T_uint:
      callarg<unsigned int>::set_value (aserial, arg, atype.ref);
      break;
    case gsi::T_long:
      callarg<long>::set_value (aserial, arg, atype.ref);
      break;
    case gsi::T_ulong:
      callarg<unsigned long>::set_value (aserial, arg, atype.ref);
      break;
    case gsi::T_longlong:
      callarg<long long>::set_value (aserial, arg, atype.ref);
      break;
    case gsi::T_double:
      callarg<double>::set_value (aserial, arg, atype.ref);
      break;
    case gsi::T_string_ccptr:
      callarg<const char *>::set_value (aserial, arg, false);
      break;
    case gsi::T_string:
      callarg_string (aserial, arg, atype.ref, tmp_stack);
      break;
    case gsi::T_var:
      callarg_var (aserial, arg, atype.ref, tmp_stack);
      break;
    case gsi::T_object:
      {
        Proxy *p;
        Data_Get_Struct (arg, Proxy, p);
        if (!p->cls_decl ()->is_derived_from (atype.cls)) {
          throw tl::Exception (tl::sprintf (tl::translate ("Unexpected object type (expected argument of class %s)").c_str (), atype.cls->name ().c_str ()));
        }
        atype.cls->copy_to (aserial, p->obj (arg));
      }
      break;
    case gsi::T_object_ref:
    case gsi::T_object_cref:
      {
        if (arg == Qnil) {
          aserial.set_value (gsi::type_traits<void *>::tag (), (void *) 0);
        } else {
          Proxy *p;
          Data_Get_Struct (arg, Proxy, p);
          if (!p->cls_decl ()->is_derived_from (atype.cls)) {
            throw tl::Exception (tl::sprintf (tl::translate ("Unexpected object type (expected argument of class %s)").c_str (), atype.cls->name ().c_str ()));
          }
          if (p->const_ref () && atype.type == gsi::T_object_ref) {
            throw tl::Exception (tl::sprintf (tl::translate ("Cannot pass a const reference of class %s to a non-const reference parameter").c_str (), atype.cls->name ().c_str ()));
          }
          aserial.set_value (gsi::type_traits<void *>::tag (), p->obj (arg));
        }
      }
      break;
    case gsi::T_object_new:
      tl_assert (false);  // An argument cannot be "new"
      break;
    }

  }
}
      
// -------------------------------------------------------------------
//  Some helper stuff for extracting arrays

struct ArrayData
{
  ArrayData (VALUE _ret, VALUE _klass)
    : ret (_ret), klass (_klass)
  { }

  VALUE ret;
  VALUE klass;
};

static void
make_array_f (void *data, void *cobj)
{
  ArrayData *ad = (ArrayData *)data;
  
  VALUE obj = rb_class_new_instance (0, 0, ad->klass);
  Proxy *p;
  Data_Get_Struct (obj, Proxy, p);
  p->set (cobj, true, false);
  rb_ary_push (ad->ret, obj);
}

// -------------------------------------------------------------------
//  Pop an argument from the call or return stack

VALUE
pop_arg (const gsi::ArgType &atype, gsi::SerialArgs &aserial)
{
  VALUE ret = Qnil;

  if (atype.iter) {
    tl_assert (false); // not implemented here 
  } else if (atype.vector) {

    switch (atype.type) {
    case gsi::T_void:
      break;
    case gsi::T_bool:
      ret = retargs<bool>::get_vector (aserial, atype.ref);
      break;
    case gsi::T_int:
      ret = retargs<int>::get_vector (aserial, atype.ref);
      break;
    case gsi::T_uint:
      ret = retargs<unsigned int>::get_vector (aserial, atype.ref);
      break;
    case gsi::T_long:
      ret = retargs<long>::get_vector (aserial, atype.ref);
      break;
    case gsi::T_ulong:
      ret = retargs<unsigned long>::get_vector (aserial, atype.ref);
      break;
    case gsi::T_longlong:
      ret = retargs<long long>::get_vector (aserial, atype.ref);
      break;
    case gsi::T_double:
      ret = retargs<double>::get_vector (aserial, atype.ref);
      break;
    case gsi::T_string:
      ret = retargs<std::string>::get_vector (aserial, atype.ref);
      break;
    case gsi::T_var:
      ret = retargs<tl::Variant>::get_vector (aserial, atype.ref);
      break;
    case gsi::T_object:
      {
        ret = rb_ary_new ();
        std::map <const gsi::ClassBase *, VALUE>::const_iterator c = rev_cls_map.find (atype.cls);
        tl_assert (c != rev_cls_map.end ());
        ArrayData ad (ret, c->second);
        c->first->get_vector_of (aserial, atype.ref, (void *) &ad, &make_array_f);
      }
      break;
    case gsi::T_object_ref:
    case gsi::T_object_cref:
    case gsi::T_object_new:
    case gsi::T_string_ccptr:
      tl_assert (false); // not implemented here 
      break;
    }

  } else {

    switch (atype.type) {
    case gsi::T_void:
      break;
    case gsi::T_bool:
      ret = retargs<bool>::get_value (aserial, atype.ref);
      break;
    case gsi::T_int:
      ret = retargs<int>::get_value (aserial, atype.ref);
      break;
    case gsi::T_uint:
      ret = retargs<unsigned int>::get_value (aserial, atype.ref);
      break;
    case gsi::T_long:
      ret = retargs<long>::get_value (aserial, atype.ref);
      break;
    case gsi::T_ulong:
      ret = retargs<unsigned long>::get_value (aserial, atype.ref);
      break;
    case gsi::T_longlong:
      ret = retargs<long long>::get_value (aserial, atype.ref);
      break;
    case gsi::T_double:
      ret = retargs<double>::get_value (aserial, atype.ref);
      break;
    case gsi::T_string_ccptr:
      ret = c2ruby<const char *>::get (aserial.get_value<const char *> (gsi::type_traits<const char *>::tag ()));
      break;
    case gsi::T_string:
      ret = ret_value_string (aserial, atype.ref);
      break;
    case gsi::T_var:
      ret = ret_value_var (aserial, atype.ref);
      break;
    case gsi::T_object:
      {
        std::map <const gsi::ClassBase *, VALUE>::const_iterator c = rev_cls_map.find (atype.cls);
        tl_assert (c != rev_cls_map.end ());
        //  create a instance without attaching it
        ret = rb_class_new_instance (0, 0, c->second);
        Proxy *p;
        Data_Get_Struct (ret, Proxy, p);
        p->set (atype.cls->take_and_create (aserial), true, false);
      }
      break;
    case gsi::T_object_ref:
    case gsi::T_object_cref:
      {
        void *obj = aserial.get_value<void *> (gsi::type_traits<void *>::tag ());
        if (obj == 0) {
          ret = Qnil;
        } else if (atype.cls->client_data (obj) != 0 && 
                   ((RBAClientData *) atype.cls->client_data (obj))->get_ref () != Qnil) {
          //  we have a that is located in C++ space but is supposed to get attached
          //  a Ruby object. If it already has, we simply return a reference to this
          ret = ((RBAClientData *) atype.cls->client_data (obj))->get_ref ();
        } else {
          std::map <const gsi::ClassBase *, VALUE>::const_iterator c = rev_cls_map.find (atype.cls);
          tl_assert (c != rev_cls_map.end ());
          //  create a instance without attaching it
          ret = rb_class_new_instance (0, 0, c->second);
          Proxy *p;
          Data_Get_Struct (ret, Proxy, p);
          p->set (obj, false, atype.type == gsi::T_object_cref);
          //  attach the ruby object to the C++ object if possible
          if (atype.cls->client_data (obj) != 0) {
            ((RBAClientData *) atype.cls->client_data (obj))->set_ref (ret);
          }
        }
      }
      break;
    case gsi::T_object_new:
      {
        void *obj = aserial.get_value<void *> (gsi::type_traits<void *>::tag ());
        if (obj == 0) {
          ret = Qnil;
        } else {
          std::map <const gsi::ClassBase *, VALUE>::const_iterator c = rev_cls_map.find (atype.cls);
          tl_assert (c != rev_cls_map.end ());
          //  create a instance without attaching it
          ret = rb_class_new_instance (0, 0, c->second);
          Proxy *p;
          Data_Get_Struct (ret, Proxy, p);
          p->set (obj, true, false);
          //  attach the ruby object to the C++ object if possible
          if (atype.cls->client_data (obj) != 0) {
            tl_assert (((RBAClientData *) atype.cls->client_data (obj))->get_ref () == Qnil);
            ((RBAClientData *) atype.cls->client_data (obj))->set_ref (ret);
          }
        }
      }
      break;
    }

  }

  return ret;
}

// -------------------------------------------------------------------
//  "each" implementation for iterators

template <class R>
struct yield_iter
{
  static void
  yield (gsi::SerialArgs &rr)
  {
    gsi::IterAdaptorAbstractBase *iter = (gsi::IterAdaptorAbstractBase *) rr.template get_value<gsi::IterAdaptorAbstractBase *> (typename gsi::type_traits<gsi::IterAdaptorAbstractBase *>::tag ());
    try {
      while (! iter->at_end ()) {
        rb_yield (c2ruby<R>::get (*((const R *) iter->v_get ())));
        iter->inc ();
      }
      delete iter;
    } catch (...) {
      delete iter;
      throw;
    }
  }
};

// -------------------------------------------------------------------
//  Proxy implementation

/**
 *  @brief Initialize the interpreter
 *
 *  It was learned (by bitter experience), that in particular rb_protect cannot simply be 
 *  called but requires some safety measures, at least in Ruby 1.8.5 as possibly 1.8.x. 
 *  What happens in rb_protect is that a RNode object is instantiated in some kind
 *  of compile steps that builds a temporary executable statement. Deep inside this
 *  functionality, the ruby_sourcefile variable is copied into the RNode object, as this
 *  variable is usually updated with the current sourcefile name. 
 *  In effect, the string referenced by this variable participates in the GC mark&sweep
 *  steps which leads to unpredicable results, if this variable is not set to a valid 
 *  string (ruby) buffer or 0.
 *
 *  As a consequence, this function must be called before rb_protect and likely other xx_protect
 *  derivatives.
 */
static void rb_protect_init ()
{
  ruby_sourcefile = 0;
  ruby_sourceline = 0;
}

//  A protect wrapper for rb_funcall2

struct rb_funcall2_params
{
  VALUE obj;
  ID id;
  int argc;
  VALUE *args;
};

static VALUE
rb_funcall2_wrap (VALUE args)
{
  //  HINT: this ugly cast is required since there is only one form of rb_protect
  const rb_funcall2_params *p = (const rb_funcall2_params *) args;
  return rb_funcall2 (p->obj, p->id, p->argc, p->args);
}

Proxy::~Proxy ()
{
  //  we know here that out Ruby counterpart hat been destroyed by the GC. 
  //  So we need to reset the reference
  if (m_owned && m_obj && cls_decl () && cls_decl ()->client_data (m_obj) != 0) {
    ((RBAClientData *) cls_decl ()->client_data (m_obj))->reset_ref ();
  }
  set (0, false, false);
  m_destroyed = true;
}

void 
Proxy::call (int id, gsi::SerialArgs &args, gsi::SerialArgs &ret) 
{
  const gsi::MethodBase *meth = m_vfuncs [id].method;

  std::vector<VALUE> argv;
  argv.reserve (std::distance (meth->begin_arguments (), meth->end_arguments ()));

  for (gsi::MethodBase::argument_iterator a = meth->begin_arguments (); a != meth->end_arguments (); ++a) {
    argv.push_back (pop_arg (*a, args));
  }

  tl_assert (id < int (m_vfuncs.size ()) && id >= 0);

  rb_funcall2_params p;
  p.obj  = m_vfuncs [id].rb_obj;
  p.id   = m_vfuncs [id].method_id;
  p.argc = argv.size ();
  p.args = argv.empty () ? 0 : &argv [0];

  int error = 0;
  //  HINT: the ugly (VALUE) cast is required since there is only one form of rb_protect
  rb_protect_init (); // see above
  VALUE rb_ret = rb_protect (&rb_funcall2_wrap, (VALUE) &p, &error);

  if (error) {
    rba_check_error ();
  }
  
  std::list<gsi::TempObject> tmp_stack;
  push_arg (meth->ret_type (), ret, rb_ret, tmp_stack);

  //  a Ruby callback must not leave temporary objects
  tl_assert (tmp_stack.empty ());
}

void *
Proxy::obj (VALUE self) 
{
  if (! m_obj) {

    if (m_destroyed) {
      throw tl::Exception (tl::translate ("Object has been destroyed already"));
    } else {

      //  delayed creation of a detached C++ object ..
      m_obj = cls_decl ()->create ();
      m_owned = true;

      //  manage attachement to the C++ object if required
      if (cls_decl ()->client_data (m_obj) != 0) {
        tl_assert (((RBAClientData *) cls_decl ()->client_data (m_obj))->get_ref () == Qnil);
        ((RBAClientData *) cls_decl ()->client_data (m_obj))->set_ref (self);
      }

      for (gsi::ClassBase::method_iterator m = cls_decl ()->begin_methods (); m != cls_decl ()->end_methods (); ++m) {

        if ((*m)->is_callback ()) {

          //  There is no place in the ruby API to determine whether a method is defined.
          //  Instead we explicitly call "method_defined?" to check, whether the given method 
          //  is defined.
          rb_funcall2_params p;
          p.obj  = rb_class_of (self);
          p.id   = rb_intern ("method_defined?");
          p.argc = 1;
          VALUE name = rb_str_new ((*m)->name ().c_str (), (*m)->name ().size ());
          p.args = &name;

          int error = 0;
          //  HINT: the ugly (VALUE) cast is required since there is only one form of rb_protect
          rb_protect_init (); // see above
          VALUE rb_ret = rb_protect (&rb_funcall2_wrap, (VALUE) &p, &error);
          if (! error && RTEST (rb_ret)) {

            //  Only if the class defines that method we can link the virtual method call to the 
            //  Ruby method
            ID method_id = rb_intern ((*m)->name ().c_str ());
            int id = add_callback (Proxy::VirtualFunction (method_id, self, *m));
            (*m)->set_callback (m_obj, gsi::Callback (id, this, (*m)->argsize (), (*m)->retsize ()));

          }

        }

      }

    }

  }
  return m_obj;
}

// -------------------------------------------------------------------
//  Ruby API 

VALUE gsi_init (int argc, VALUE *, VALUE self)
{
  try {

    Proxy *p;
    Data_Get_Struct (self, Proxy, p);
    const gsi::ClassBase *cls_decl_self = p->cls_decl ();

    /* 
     * This was supposed to be a check if the object can be created with new.
     * Unfortunately we can't do this, because a initialization happens even if an object is 
     * returned as reference: 
    if (! cls_decl_self->can_default_create ()) {
      throw tl::Exception (tl::translate ("No constructor provided for class '%s'").c_str (), cls_decl_self->name ().c_str ());
    }
    */

    if (argc > 0) {
      throw tl::Exception (tl::translate ("Constructor does not allow arguments").c_str (), cls_decl_self->name ().c_str ());
    }

  } catch (std::exception &ex) {
    rb_raise (rb_eRuntimeError, ex.what ());
  } catch (tl::Exception &ex) {
    rb_raise (rb_eRuntimeError, ex.msg ().c_str ());
  } catch (...) {
    rb_raise (rb_eRuntimeError, tl::translate ("Unspecific exception").c_str ());
  }

  return self;
}

VALUE gsi_destroy (VALUE self)
{
  try {

    //  Compare if the classes are identical
    Proxy *p;
    
    Data_Get_Struct (self, Proxy, p);
    p->destroy ();

  } catch (std::exception &ex) {
    rb_raise (rb_eRuntimeError, ex.what ());
  } catch (tl::Exception &ex) {
    rb_raise (rb_eRuntimeError, ex.msg ().c_str ());
  } catch (...) {
    rb_raise (rb_eRuntimeError, tl::translate ("Unspecific exception").c_str ());
  }

  return self;
}

VALUE gsi_destroyed (VALUE self)
{
  bool r = false; 

  try {

    //  return true if the object was destroyed already
    Proxy *p;
    Data_Get_Struct (self, Proxy, p);
    r = p->destroyed ();

  } catch (std::exception &ex) {
    rb_raise (rb_eRuntimeError, ex.what ());
  } catch (tl::Exception &ex) {
    rb_raise (rb_eRuntimeError, ex.msg ().c_str ());
  } catch (...) {
    rb_raise (rb_eRuntimeError, tl::translate ("Unspecific exception").c_str ());
  }

  return c2ruby<bool>::get (r);
}

VALUE gsi_init_copy (VALUE self, VALUE src)
{
  try {

    //  Compare if the classes are identical
    Proxy *p;
    
    Data_Get_Struct (src, Proxy, p);
    const gsi::ClassBase *cls_decl_src = p->cls_decl ();
    void *obj_src = p->obj (src);

    Data_Get_Struct (self, Proxy, p);
    const gsi::ClassBase *cls_decl_self = p->cls_decl ();
    void *obj_self = p->obj (self);

    if (cls_decl_src != cls_decl_self) {
      throw tl::Exception (tl::sprintf (tl::translate ("Type is not identical on copy").c_str ()));
    } 
    if (! cls_decl_self->can_copy ()) {
      throw tl::Exception (tl::translate ("No copy constructor provided for class '%s'").c_str (), cls_decl_self->name ().c_str ());
    }
    cls_decl_self->assign (obj_self, obj_src);

  } catch (std::exception &ex) {
    rb_raise (rb_eRuntimeError, ex.what ());
  } catch (tl::Exception &ex) {
    rb_raise (rb_eRuntimeError, ex.msg ().c_str ());
  } catch (...) {
    rb_raise (rb_eRuntimeError, tl::translate ("Unspecific exception").c_str ());
  }

  return self;
}

void free_proxy (void *p)
{
  delete ((Proxy *) p);
}


const gsi::ClassBase *
find_cclass (VALUE k)
{
  std::map <VALUE, const gsi::ClassBase *>::const_iterator cls;

  //  find the class that is bound to C++ (maybe a super class)
  while (k != rb_cObject) { 
    cls = cls_map.find (k);
    if (cls == cls_map.end ()) {
      //  if not found advance to super class
      k = RCLASS (k)->super;
    } else {
      break;
    }
  }

  tl_assert (cls != cls_map.end ());
  return cls->second;
}

VALUE alloc_proxy (VALUE klass)
{
  const gsi::ClassBase *cls = find_cclass (klass);
  Proxy *proxy = new Proxy (cls);
  return Data_Wrap_Struct (klass, 0, free_proxy, proxy);
}

VALUE
method_adaptor (int mid, int argc, VALUE *argv, VALUE self)
{
  VALUE ret = Qnil;
  std::list<gsi::TempObject> tmp_stack;

  try {

    const gsi::ClassBase *cls_decl;
    void *obj = 0;
    bool const_ref = false;

    if (TYPE (self) == T_CLASS) {
      //  we have a static method
      cls_decl = find_cclass (self);
    } else {
      //  we have an instance method
      Proxy *p;
      Data_Get_Struct (self, Proxy, p);
      cls_decl = p->cls_decl ();
      obj = p->obj (self);
      const_ref = p->const_ref ();
    }

    //  get number of candidates by argument count
    const gsi::MethodBase *meth = 0;
    unsigned int candidates = 0;
    for (gsi::ClassBase::overloaded_methods_iterator om = cls_decl->begin_overloaded_methods (mid); om != cls_decl->end_overloaded_methods (mid); ++om) {
      const gsi::MethodBase *m = cls_decl->begin_methods () [*om];
      if (std::distance (m->begin_arguments (), m->end_arguments ()) == argc) {
        ++candidates;
        meth = m;
      }
    }

    //  no candidate -> error
    if (! meth) {
      std::set<unsigned int> nargs;
      for (gsi::ClassBase::overloaded_methods_iterator om = cls_decl->begin_overloaded_methods (mid); om != cls_decl->end_overloaded_methods (mid); ++om) {
        const gsi::MethodBase *m = cls_decl->begin_methods () [*om];
        nargs.insert (std::distance (m->begin_arguments (), m->end_arguments ()));
      }
      std::string nargs_s;
      for (std::set<unsigned int>::const_iterator na = nargs.begin (); na != nargs.end (); ++na) {
        if (na != nargs.begin ()) {
          nargs_s += "/";
        }
        nargs_s += tl::to_string (*na);
      }
      throw tl::Exception (tl::sprintf (tl::translate ("Invalid number of arguments for method %s, class %s (got %d, expected %s)").c_str (), cls_decl->overloaded_name (mid).c_str (), cls_decl->name ().c_str (), argc, nargs_s.c_str ()));
    }

    //  more than one candidate -> refine by checking the arguments
    if (candidates > 1) {
      meth = 0;
      candidates = 0;
      for (gsi::ClassBase::overloaded_methods_iterator om = cls_decl->begin_overloaded_methods (mid); om != cls_decl->end_overloaded_methods (mid); ++om) {
        const gsi::MethodBase *m = cls_decl->begin_methods () [*om];
        bool is_valid = (std::distance (m->begin_arguments (), m->end_arguments ()) == argc);
        VALUE *av = argv;
        for (gsi::MethodBase::argument_iterator a = m->begin_arguments (); is_valid && a != m->end_arguments (); ++a, ++av) {
          if (! test_arg (*a, *av)) {
            is_valid = false;
          }
        }
        if (is_valid) {
          ++candidates;
          meth = m;
        }
      }
    }

    if (! meth) {
      throw tl::Exception (tl::sprintf (tl::translate ("No method with matching arguments for method %s, class %s").c_str (), cls_decl->overloaded_name (mid).c_str (), cls_decl->name ().c_str ()));
    }

    //  more than one candidate -> refine by checking the arguments
    if (candidates > 1) {
      throw tl::Exception (tl::sprintf (tl::translate ("Ambiguous overload variants for method %s, class %s - multiple method declarations match arguments").c_str (), cls_decl->overloaded_name (mid).c_str (), cls_decl->name ().c_str ()));
    }

    if (const_ref && ! meth->is_const ()) {
      throw tl::Exception (tl::sprintf (tl::translate ("Cannot call non-const method %s, class %s on a const reference").c_str (), cls_decl->overloaded_name (mid).c_str (), cls_decl->name ().c_str ()));
    }

    gsi::SerialArgs arglist (meth->argsize ());
    gsi::SerialArgs retlist (meth->retsize ());
    VALUE *av = argv;
    for (gsi::MethodBase::argument_iterator a = meth->begin_arguments (); a != meth->end_arguments (); ++a, ++av) {
      push_arg (*a, arglist, *av, tmp_stack);
    }

    meth->call (obj, arglist, retlist);

    if (meth->ret_type ().iter) {

      switch (meth->ret_type ().type) {
      case gsi::T_void:
        break;
      case gsi::T_bool:
        yield_iter<bool>::yield (retlist);
        break;
      case gsi::T_int:
        yield_iter<int>::yield (retlist);
        break;
      case gsi::T_uint:
        yield_iter<unsigned int>::yield (retlist);
        break;
      case gsi::T_long:
        yield_iter<long>::yield (retlist);
        break;
      case gsi::T_ulong:
        yield_iter<unsigned long>::yield (retlist);
        break;
      case gsi::T_longlong:
        yield_iter<long long>::yield (retlist);
        break;
      case gsi::T_double:
        yield_iter<double>::yield (retlist);
        break;
      case gsi::T_string_ccptr:
        yield_iter<const char *>::yield (retlist);
        break;
      case gsi::T_string:
        yield_iter<std::string>::yield (retlist);
        break;
      case gsi::T_var:
        yield_iter<tl::Variant>::yield (retlist);
        break;
      case gsi::T_object:
        {
          std::map <const gsi::ClassBase *, VALUE>::const_iterator c = rev_cls_map.find (meth->ret_type ().cls);
          tl_assert (c != rev_cls_map.end ());
          void *iter = c->first->create_obj_iterator (retlist);
          try {
            while (! c->first->obj_iterator_at_end (iter)) {
              VALUE obj = rb_class_new_instance (0, 0, c->second);
              Proxy *p;
              Data_Get_Struct (obj, Proxy, p);
              p->set (c->first->take_iterator_obj (iter), true, false);
              rb_yield (obj);
              c->first->inc_obj_iterator (iter);
            }
            c->first->delete_obj_iterator (iter);
          } catch (...) {
            c->first->delete_obj_iterator (iter);
            throw;
          }
        }
        break;
      case gsi::T_object_ref:
      case gsi::T_object_cref:
        {
          std::map <const gsi::ClassBase *, VALUE>::const_iterator c = rev_cls_map.find (meth->ret_type ().cls);
          tl_assert (c != rev_cls_map.end ());
          void *iter = c->first->create_obj_iterator (retlist);
          try {
            while (! c->first->obj_iterator_at_end (iter)) {
              void *cobj = const_cast<void *> (c->first->get_iterator_obj (iter));
              VALUE obj;
              if (cobj == 0) {
                obj = Qnil;
              } else if (meth->ret_type ().cls->client_data (cobj) != 0 && 
                         ((RBAClientData *) meth->ret_type ().cls->client_data (cobj))->get_ref () != Qnil) {
                //  we have a that is located in C++ space but is supposed to get attached
                //  a Ruby object. If it already has, we simply return a reference to this
                obj = ((RBAClientData *) meth->ret_type ().cls->client_data (cobj))->get_ref ();
              } else {
                obj = rb_class_new_instance (0, 0, c->second);
                Proxy *p;
                Data_Get_Struct (obj, Proxy, p);
                p->set (cobj, false, meth->ret_type ().type == gsi::T_object_cref);
                //  attach the ruby object to the C++ object if possible
                if (meth->ret_type ().cls->client_data (cobj) != 0) {
                  ((RBAClientData *) meth->ret_type ().cls->client_data (cobj))->set_ref (obj);
                }
              }
              rb_yield (obj);
              c->first->inc_obj_iterator (iter);
            }
            c->first->delete_obj_iterator (iter);
          } catch (...) {
            c->first->delete_obj_iterator (iter);
            throw;
          }
        }
        break;
      case gsi::T_object_new:
        tl_assert (false);
        break;
      }

    } else {
      ret = pop_arg (meth->ret_type (), retlist);
    }

  } catch (std::exception &ex) {
    rb_raise (rb_eRuntimeError, ex.what ());
  } catch (tl::Exception &ex) {
    rb_raise (rb_eRuntimeError, ex.msg ().c_str ());
  } catch (...) {
    rb_raise (rb_eRuntimeError, tl::translate ("Unspecific exception").c_str ());
  }

  return ret;
}

template <int N> 
VALUE method_adaptor (int argc, VALUE *argv, VALUE self)
{
  return method_adaptor (N, argc, argv, self);
}

VALUE (*(method_adaptors [])) (int, VALUE *, VALUE) =
{
  &method_adaptor<0x00>, &method_adaptor<0x01>, &method_adaptor<0x02>, &method_adaptor<0x03>, &method_adaptor<0x04>, &method_adaptor<0x05>, &method_adaptor<0x06>, &method_adaptor<0x07>,
  &method_adaptor<0x08>, &method_adaptor<0x09>, &method_adaptor<0x0a>, &method_adaptor<0x0b>, &method_adaptor<0x0c>, &method_adaptor<0x0d>, &method_adaptor<0x0e>, &method_adaptor<0x0f>,
  &method_adaptor<0x10>, &method_adaptor<0x11>, &method_adaptor<0x12>, &method_adaptor<0x13>, &method_adaptor<0x14>, &method_adaptor<0x15>, &method_adaptor<0x16>, &method_adaptor<0x17>,
  &method_adaptor<0x18>, &method_adaptor<0x19>, &method_adaptor<0x1a>, &method_adaptor<0x1b>, &method_adaptor<0x1c>, &method_adaptor<0x1d>, &method_adaptor<0x1e>, &method_adaptor<0x1f>,
  &method_adaptor<0x20>, &method_adaptor<0x21>, &method_adaptor<0x22>, &method_adaptor<0x23>, &method_adaptor<0x24>, &method_adaptor<0x25>, &method_adaptor<0x26>, &method_adaptor<0x27>,
  &method_adaptor<0x28>, &method_adaptor<0x29>, &method_adaptor<0x2a>, &method_adaptor<0x2b>, &method_adaptor<0x2c>, &method_adaptor<0x2d>, &method_adaptor<0x2e>, &method_adaptor<0x2f>,
  &method_adaptor<0x30>, &method_adaptor<0x31>, &method_adaptor<0x32>, &method_adaptor<0x33>, &method_adaptor<0x34>, &method_adaptor<0x35>, &method_adaptor<0x36>, &method_adaptor<0x37>,
  &method_adaptor<0x38>, &method_adaptor<0x39>, &method_adaptor<0x3a>, &method_adaptor<0x3b>, &method_adaptor<0x3c>, &method_adaptor<0x3d>, &method_adaptor<0x3e>, &method_adaptor<0x3f>,
  &method_adaptor<0x40>, &method_adaptor<0x41>, &method_adaptor<0x42>, &method_adaptor<0x43>, &method_adaptor<0x44>, &method_adaptor<0x45>, &method_adaptor<0x46>, &method_adaptor<0x47>,
  &method_adaptor<0x48>, &method_adaptor<0x49>, &method_adaptor<0x4a>, &method_adaptor<0x4b>, &method_adaptor<0x4c>, &method_adaptor<0x4d>, &method_adaptor<0x4e>, &method_adaptor<0x4f>,
  &method_adaptor<0x50>, &method_adaptor<0x51>, &method_adaptor<0x52>, &method_adaptor<0x53>, &method_adaptor<0x54>, &method_adaptor<0x55>, &method_adaptor<0x56>, &method_adaptor<0x57>,
  &method_adaptor<0x58>, &method_adaptor<0x59>, &method_adaptor<0x5a>, &method_adaptor<0x5b>, &method_adaptor<0x5c>, &method_adaptor<0x5d>, &method_adaptor<0x5e>, &method_adaptor<0x5f>,
  &method_adaptor<0x60>, &method_adaptor<0x61>, &method_adaptor<0x62>, &method_adaptor<0x63>, &method_adaptor<0x64>, &method_adaptor<0x65>, &method_adaptor<0x66>, &method_adaptor<0x67>,
  &method_adaptor<0x68>, &method_adaptor<0x69>, &method_adaptor<0x6a>, &method_adaptor<0x6b>, &method_adaptor<0x6c>, &method_adaptor<0x6d>, &method_adaptor<0x6e>, &method_adaptor<0x6f>,
  &method_adaptor<0x70>, &method_adaptor<0x71>, &method_adaptor<0x72>, &method_adaptor<0x73>, &method_adaptor<0x74>, &method_adaptor<0x75>, &method_adaptor<0x76>, &method_adaptor<0x77>,
  &method_adaptor<0x78>, &method_adaptor<0x79>, &method_adaptor<0x7a>, &method_adaptor<0x7b>, &method_adaptor<0x7c>, &method_adaptor<0x7d>, &method_adaptor<0x7e>, &method_adaptor<0x7f>,
  &method_adaptor<0x80>, &method_adaptor<0x81>, &method_adaptor<0x82>, &method_adaptor<0x83>, &method_adaptor<0x84>, &method_adaptor<0x85>, &method_adaptor<0x86>, &method_adaptor<0x87>,
  &method_adaptor<0x88>, &method_adaptor<0x89>, &method_adaptor<0x8a>, &method_adaptor<0x8b>, &method_adaptor<0x8c>, &method_adaptor<0x8d>, &method_adaptor<0x8e>, &method_adaptor<0x8f>,
  &method_adaptor<0x90>, &method_adaptor<0x91>, &method_adaptor<0x92>, &method_adaptor<0x93>, &method_adaptor<0x94>, &method_adaptor<0x95>, &method_adaptor<0x96>, &method_adaptor<0x97>,
  &method_adaptor<0x98>, &method_adaptor<0x99>, &method_adaptor<0x9a>, &method_adaptor<0x9b>, &method_adaptor<0x9c>, &method_adaptor<0x9d>, &method_adaptor<0x9e>, &method_adaptor<0x9f>,
  &method_adaptor<0xa0>, &method_adaptor<0xa1>, &method_adaptor<0xa2>, &method_adaptor<0xa3>, &method_adaptor<0xa4>, &method_adaptor<0xa5>, &method_adaptor<0xa6>, &method_adaptor<0xa7>,
  &method_adaptor<0xa8>, &method_adaptor<0xa9>, &method_adaptor<0xaa>, &method_adaptor<0xab>, &method_adaptor<0xac>, &method_adaptor<0xad>, &method_adaptor<0xae>, &method_adaptor<0xaf>,
  &method_adaptor<0xb0>, &method_adaptor<0xb1>, &method_adaptor<0xb2>, &method_adaptor<0xb3>, &method_adaptor<0xb4>, &method_adaptor<0xb5>, &method_adaptor<0xb6>, &method_adaptor<0xb7>,
  &method_adaptor<0xb8>, &method_adaptor<0xb9>, &method_adaptor<0xba>, &method_adaptor<0xbb>, &method_adaptor<0xbc>, &method_adaptor<0xbd>, &method_adaptor<0xbe>, &method_adaptor<0xbf>,
  &method_adaptor<0xc0>, &method_adaptor<0xc1>, &method_adaptor<0xc2>, &method_adaptor<0xc3>, &method_adaptor<0xc4>, &method_adaptor<0xc5>, &method_adaptor<0xc6>, &method_adaptor<0xc7>,
  &method_adaptor<0xc8>, &method_adaptor<0xc9>, &method_adaptor<0xca>, &method_adaptor<0xcb>, &method_adaptor<0xcc>, &method_adaptor<0xcd>, &method_adaptor<0xce>, &method_adaptor<0xcf>,
  &method_adaptor<0xd0>, &method_adaptor<0xd1>, &method_adaptor<0xd2>, &method_adaptor<0xd3>, &method_adaptor<0xd4>, &method_adaptor<0xd5>, &method_adaptor<0xd6>, &method_adaptor<0xd7>,
  &method_adaptor<0xd8>, &method_adaptor<0xd9>, &method_adaptor<0xda>, &method_adaptor<0xdb>, &method_adaptor<0xdc>, &method_adaptor<0xdd>, &method_adaptor<0xde>, &method_adaptor<0xdf>,
  &method_adaptor<0xe0>, &method_adaptor<0xe1>, &method_adaptor<0xe2>, &method_adaptor<0xe3>, &method_adaptor<0xe4>, &method_adaptor<0xe5>, &method_adaptor<0xe6>, &method_adaptor<0xe7>,
  &method_adaptor<0xe8>, &method_adaptor<0xe9>, &method_adaptor<0xea>, &method_adaptor<0xeb>, &method_adaptor<0xec>, &method_adaptor<0xed>, &method_adaptor<0xee>, &method_adaptor<0xef>,
  &method_adaptor<0xf0>, &method_adaptor<0xf1>, &method_adaptor<0xf2>, &method_adaptor<0xf3>, &method_adaptor<0xf4>, &method_adaptor<0xf5>, &method_adaptor<0xf6>, &method_adaptor<0xf7>,
  &method_adaptor<0xf8>, &method_adaptor<0xf9>, &method_adaptor<0xfa>, &method_adaptor<0xfb>, &method_adaptor<0xfc>, &method_adaptor<0xfd>, &method_adaptor<0xfe>, &method_adaptor<0xff>,
};

// --------------------------------------------------------------------------
//  RubyInterpreter implementation

static RubyInterpreter *rba_interpreter = 0;
static bool rba_ruby_destroyed = false;

static VALUE saved_stderr = Qnil;
static VALUE saved_stdout = Qnil;
static VALUE stdout_klass, stderr_klass;

static Console *installed_console = 0;

static VALUE 
stdout_write (VALUE self, VALUE a)
{
  if (installed_console) {
    installed_console->write_str (StringValuePtr (a), Console::OS_stdout);
  }
  return Qnil;
}

static VALUE 
stderr_write (VALUE self, VALUE a)
{
  if (installed_console) {
    installed_console->write_str (StringValuePtr (a), Console::OS_stderr);
  }
  return Qnil;
}

static void
rba_init ()
{
  //  Keep ruby_init from changing the SIGINT handler.
  //  The reason for ruby_init doing this is probably to enable the input methods
  //  to receive Ctrl+C. We however do not need this functionality and users complain 
  //  about now having Ctrl+C to kill the application.
  void (*org_sigint) (int) = signal (SIGINT, SIG_DFL);
  ruby_init ();
  signal (SIGINT, org_sigint);

  ruby_init_loadpath ();

  ruby_script ("embedded");

  typedef VALUE (*ruby_func)(ANYARGS);

  VALUE module = rb_define_module ("RBA");

  //  merge the extensions to the main declaration
  gsi::ClassBase::merge_declarations ();

  for (gsi::ClassBase::class_iterator c = gsi::ClassBase::begin_classes (); c != gsi::ClassBase::end_classes (); ++c) {

    //  there should be only main declarations since we merged
    tl_assert (c->declaration () == &*c);

    //  TODO: get rid of this const_cast hack
    (const_cast<gsi::ClassBase *> (&*c))->register_client_handler (new RBAClientHandler ());

    VALUE klass = rb_define_class_under (module, c->name ().c_str (), rb_cObject);
    cls_map.insert (std::make_pair (klass, &*c));
    rev_cls_map.insert (std::make_pair (&*c, klass));

    rb_define_alloc_func (klass, alloc_proxy);
    rb_define_method (klass, "initialize", (ruby_func) &gsi_init, -1);
    rb_define_method (klass, "initialize_copy", (ruby_func) &gsi_init_copy, 1);
    rb_define_method (klass, "destroy", (ruby_func) &gsi_destroy, 0);
    rb_define_method (klass, "destroyed", (ruby_func) &gsi_destroyed, 0);

    for (gsi::ClassBase::name_map_iterator nm = c->begin_name_map (); nm != c->end_name_map (); ++nm) {

      unsigned int mid = nm->second;

      //  the first overloaded method determines the type of method - this implies that all
      //  overloaded methods must be static or none.
      const gsi::MethodBase *m_first = c->begin_methods () [*c->begin_overloaded_methods (mid)];

      if (! m_first->is_callback ()) {
        if (! m_first->is_static ()) {
          tl_assert (mid < sizeof (method_adaptors) / sizeof (method_adaptors[0]));
          rb_define_method (klass, nm->first.c_str (), (ruby_func) method_adaptors[mid], -1);
        } else {
          tl_assert (mid < sizeof (method_adaptors) / sizeof (method_adaptors[0]));
          rb_define_module_function (klass, nm->first.c_str (), (ruby_func) method_adaptors[mid], -1);
        }
      }

    }

  }

  //  define new classes for handling stdout and stderr 
  stdout_klass = rb_define_class_under (module, "RBAstdout", rb_cObject);
  rb_define_method (stdout_klass, "write", (ruby_func) &stdout_write, 1);
  stderr_klass = rb_define_class_under (module, "RBAstderr", rb_cObject);
  rb_define_method (stderr_klass, "write", (ruby_func) &stderr_write, 1);

  //  register the saved stdout/stderr objects with the garbage collector
  rb_global_variable (& saved_stdout);
  rb_global_variable (& saved_stderr);

  //  create the handler objects
  VALUE empty_args [] = { Qnil };
  saved_stderr = rb_class_new_instance (0, empty_args, stderr_klass);
  saved_stdout = rb_class_new_instance (0, empty_args, stdout_klass);
}

RubyInterpreter::RubyInterpreter () 
{ 
  tl_assert (! rba_ruby_destroyed);
  tl_assert (! rba_interpreter);
  rba_interpreter = this;
  rba_init ();
}

RubyInterpreter::~RubyInterpreter () 
{
  ruby_finalize ();
  rba_ruby_destroyed = true;
  rba_interpreter = 0;
}

void
RubyInterpreter::require (const char *filename)
{
  int error = 0;
  rb_protect_init (); // see above
  rb_protect ((VALUE (*)(VALUE))rb_require, (VALUE)filename, &error);
  if (error) {
    rba_check_error ();
  }
}

void
RubyInterpreter::load_file (const char *filename)
{
  int error = 0;
  int wrap = 0; // @@@ make variable?
  rb_protect_init (); // see above
  rb_load_protect (rb_str_new (filename, strlen (filename)), wrap, &error);
  if (error) {
    rba_check_error ();
  }
}

void
RubyInterpreter::eval_string (const char *expr)
{
  int error = 0;
  rb_protect_init (); // see above
  rb_eval_string_protect (expr, &error);
  if (error) {
    rba_check_error_eval ();
  }
}

void
RubyInterpreter::eval_string_and_print (const char *expr)
{
  int error = 0;
  rb_protect_init (); // see above
  VALUE res = rb_eval_string_protect (expr, &error);
  if (error) {
    rba_check_error_eval ();
  } else if (installed_console && res != Qnil) {
    VALUE res_s = rb_obj_as_string (res);
    installed_console->write_str (StringValuePtr (res_s), Console::OS_stdout);
    installed_console->write_str ("\n", Console::OS_stdout);
  }
}

void
RubyInterpreter::define_variable (const std::string &name, const std::string &value)
{
  rb_gv_set (name.c_str (), rb_str_new (value.c_str (), value.size ()));
}

bool
RubyInterpreter::ruby_available () const
{
  return true;
}

void 
RubyInterpreter::install_console (Console *console)
{
  if ((console && ! installed_console) || (! console && installed_console)) {
    std::swap (saved_stderr, rb_stderr);
    std::swap (saved_stdout, rb_stdout);
  }
  installed_console = console;
}

#else

RubyInterpreter::RubyInterpreter () 
{ 
  // .. nothing yet ..
}

RubyInterpreter::~RubyInterpreter () 
{
  // .. nothing yet ..
}

void
RubyInterpreter::require (const char *)
{
  tl_assert (false);
}

void
RubyInterpreter::load_file (const char *)
{
  tl_assert (false);
}

void
RubyInterpreter::eval_string (const char *expr)
{
  tl_assert (false);
}

void
RubyInterpreter::eval_string_and_print (const char *expr)
{
  tl_assert (false);
}

void
RubyInterpreter::define_variable (const std::string &name, const std::string &value)
{
  tl_assert (false);
}

bool
RubyInterpreter::ruby_available () const
{
  return false;
}

void 
RubyInterpreter::install_console (Console *console)
{
  // .. nothing ..
}

#endif

