
/*

  KLayout Layout Viewer
  Copyright (C) 2006-2007 Matthias Koefferlein

  This program is free software; you can redistribute it and/or modify
  it under the terms of the GNU General Public License as published by
  the Free Software Foundation; either version 2 of the License, or
  (at your option) any later version.

  This program is distributed in the hope that it will be useful,
  but WITHOUT ANY WARRANTY; without even the implied warranty of
  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
  GNU General Public License for more details.

  You should have received a copy of the GNU General Public License
  along with this program; if not, write to the Free Software
  Foundation, Inc., 51 Franklin St, Fifth Floor, Boston, MA  02110-1301  USA

*/



#include <errno.h> 
#include <string.h> 

#include "tlStream.h"
#include "tlDeflate.h"
#include "tlAssert.h"

#include "tlException.h"
#include "tlString.h"

namespace tl
{

// ---------------------------------------------------------------
//  InputStream implementation

InputStream::InputStream (InputStreamBase &delegate)
  : m_pos (0), mp_bptr (0), mp_delegate (&delegate), mp_deflate (0)
{ 
  m_bcap = 4096; // initial buffer capacity
  m_blen = 0;
  mp_buffer = new char [m_bcap];
}

InputStream::~InputStream ()
{
  if (mp_deflate) {
    delete mp_deflate;
    mp_deflate = 0;
  } 
  if (mp_buffer) {
    delete[] mp_buffer;
    mp_buffer = 0;
  }
}

const char * 
InputStream::get (size_t n, bool bypass_deflate)
{
  //  if deflating, employ the deflate filter to get the data
  if (mp_deflate && ! bypass_deflate) {
    if (! mp_deflate->at_end ()) {

      const char *r = mp_deflate->get (n);
      tl_assert (r != 0);  //  since deflate did not report at_end()
      return r;

    } else {
      delete mp_deflate;
      mp_deflate = 0;
    }
  } 

  if (m_blen < n) {

    //  to keep move activity low, allocate twice as much as required
    if (m_bcap < n * 2) {

      while (m_bcap < n) {
        m_bcap *= 2;
      }

      char *buffer = new char [m_bcap];
      memcpy (buffer, mp_bptr, m_blen);
      delete [] mp_buffer;
      mp_buffer = buffer;

    } else {
      memmove (mp_buffer, mp_bptr, m_blen);
    }

    m_blen += mp_delegate->read (mp_buffer + m_blen, m_bcap - m_blen); 
    mp_bptr = mp_buffer;

  }

  if (m_blen >= n) {
    const char *r = mp_bptr;
    mp_bptr += n;
    m_blen -= n;
    m_pos += n;
    return r;
  } else {
    return 0;
  }

}

void
InputStream::unget (size_t n)
{
  if (mp_deflate) {
    mp_deflate->unget (n);
  } else {
    mp_bptr -= n;
    m_blen += n;
    m_pos -= n;
  }
}

void
InputStream::deflate ()
{
  tl_assert (mp_deflate == 0);
  mp_deflate = new tl::DeflateFilter (*this);
}

// ---------------------------------------------------------------
//  OutputStream implementation

OutputStream::OutputStream (OutputStreamBase &delegate)
  : m_pos (0), mp_delegate (&delegate)
{ 
  //  .. nothing yet ..
}

OutputStream::~OutputStream ()
{
  //  .. nothing yet ..
}

void
OutputStream::put (const char *b, size_t n)
{
  mp_delegate->write (b, n);
  m_pos += n;
}

// ---------------------------------------------------------------
//  Input file delegate implementation

InputFile::InputFile (const std::string &path) throw (tl::Exception)
  : m_file (NULL)
{
  m_source = path;
  m_file = fopen (path.c_str (), "rb");
  if (m_file == NULL) {
    throw FileOpenErrorException (m_source, errno);
  }
}

InputFile::~InputFile ()
{
  if (m_file != NULL) {
    fclose (m_file);
    m_file = NULL;
  }  
}

size_t 
InputFile::read (char *b, size_t n) throw (tl::Exception)
{
  tl_assert (m_file != NULL);
  size_t ret = fread (b, 1, n, m_file);
  if (ret < n) {
    if (ferror (m_file)) {
      throw FileReadErrorException (m_source, ferror (m_file));
    }
  }

  return ret;
}

// ---------------------------------------------------------------
//  Output file delegate implementation

OutputFile::OutputFile (const std::string &path) throw (tl::Exception)
  : m_file (NULL)
{
  m_source = path;
  m_file = fopen (path.c_str (), "wb");
  if (m_file == NULL) {
    throw FileOpenErrorException (m_source, errno);
  }
}

OutputFile::~OutputFile ()
{
  if (m_file != NULL) {
    fclose (m_file);
    m_file = NULL;
  }  
}

void 
OutputFile::write (const char *b, size_t n) throw (tl::Exception)
{
  tl_assert (m_file != NULL);
  size_t ret = fwrite (b, 1, n, m_file);
  if (ret < n) {
    if (ferror (m_file)) {
      throw FileWriteErrorException (m_source, ferror (m_file));
    }
  }
}

// ---------------------------------------------------------------
//  InputZLibFile implementation

InputZLibFile::InputZLibFile (const std::string &path) throw (tl::Exception)
  : m_zs (NULL)
{
  m_source = path;
  m_zs = gzopen (path.c_str (), "rb");
  if (m_zs == NULL) {
    throw FileOpenErrorException (m_source, errno);
  }
}

InputZLibFile::~InputZLibFile ()
{
  if (m_zs != NULL) {
    gzclose (m_zs);
    m_zs = NULL;
  }  
}

size_t 
InputZLibFile::read (char *b, size_t n) throw (tl::Exception)
{
  tl_assert (m_zs != NULL);
  int ret = gzread (m_zs, b, n);
  if (ret < 0) {
    int gz_err = 0;
    const char *em = gzerror (m_zs, &gz_err);
    if (gz_err == Z_ERRNO) {
      throw FileReadErrorException (m_source, errno);
    } else {
      throw ZLibReadErrorException (m_source, em);
    }
  }

  return ret;
}

// ---------------------------------------------------------------
//  OutputZLibFile implementation

OutputZLibFile::OutputZLibFile (const std::string &path) throw (tl::Exception)
  : m_zs (NULL)
{
  m_source = path;
  m_zs = gzopen (path.c_str (), "wb");
  if (m_zs == NULL) {
    throw FileOpenErrorException (m_source, errno);
  }
}

OutputZLibFile::~OutputZLibFile ()
{
  if (m_zs != NULL) {
    tl_assert (m_zs != NULL);
    gzclose (m_zs);
    m_zs = NULL;
  }  
}

void 
OutputZLibFile::write (const char *b, size_t n) throw (tl::Exception)
{
  tl_assert (m_zs != NULL);
  int ret = gzwrite (m_zs, (char *) b, n);
  if (ret < 0) {
    int gz_err = 0;
    const char *em = gzerror (m_zs, &gz_err);
    if (gz_err == Z_ERRNO) {
      throw FileWriteErrorException (m_source, errno);
    } else {
      throw ZLibWriteErrorException (m_source, em);
    }
  }
}

// ---------------------------------------------------------------
//  InputPipe delegate implementation

InputPipe::InputPipe (const std::string &path) throw (tl::Exception) 
  : m_file (NULL)
{
  m_source = path;
  m_file = popen (path.c_str (), "r");
  if (m_file == NULL) {
    throw FilePOpenErrorException (m_source, errno);
  }
}

InputPipe::~InputPipe ()
{
  if (m_file != NULL) {
    fclose (m_file);
    m_file = NULL;
  }  
}

size_t 
InputPipe::read (char *b, size_t n) throw (tl::Exception)
{
  tl_assert (m_file != NULL);
  size_t ret = fread (b, 1, n, m_file);
  if (ret < n) {
    if (ferror (m_file)) {
      throw FilePReadErrorException (m_source, ferror (m_file));
    }
  }

  return ret;
}

// ---------------------------------------------------------------
//  OutputPipe delegate implementation

OutputPipe::OutputPipe (const std::string &path) throw (tl::Exception) 
  : m_file (NULL)
{
  m_source = path;
  m_file = popen (path.c_str (), "w");
  if (m_file == NULL) {
    throw FilePOpenErrorException (m_source, errno);
  }
}

OutputPipe::~OutputPipe ()
{
  if (m_file != NULL) {
    fclose (m_file);
    m_file = NULL;
  }  
}

void 
OutputPipe::write (const char *b, size_t n) throw (tl::Exception)
{
  tl_assert (m_file != NULL);
  size_t ret = fwrite (b, 1, n, m_file);
  if (ret < n) {
    if (ferror (m_file)) {
      throw FilePWriteErrorException (m_source, ferror (m_file));
    }
  }
}

}

