
/*

  KLayout Layout Viewer
  Copyright (C) 2006-2007 Matthias Koefferlein

  This program is free software; you can redistribute it and/or modify
  it under the terms of the GNU General Public License as published by
  the Free Software Foundation; either version 2 of the License, or
  (at your option) any later version.

  This program is distributed in the hope that it will be useful,
  but WITHOUT ANY WARRANTY; without even the implied warranty of
  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
  GNU General Public License for more details.

  You should have received a copy of the GNU General Public License
  along with this program; if not, write to the Free Software
  Foundation, Inc., 51 Franklin St, Fifth Floor, Boston, MA  02110-1301  USA

*/


#include <stdlib.h>
#include <stdio.h>
#include <string.h>
#include <ctype.h>
#include <limits>

#include <tlString.h>
#include <tlInternational.h>

std::string 
tl::to_string (double d, const char *f)
{
  char b[32];
  ::sprintf (b, f, d); 
  return b;
}

std::string 
tl::to_string (float d, const char *f)
{
  char b[32];
  ::sprintf (b, f, d); 
  return b;
}

std::string 
tl::to_string (int d)
{
  char b[32];
  ::sprintf (b, "%d", d); 
  return b;
}

std::string 
tl::to_string (unsigned int d)
{
  char b[32];
  ::sprintf (b, "%u", d); 
  return b;
}

std::string 
tl::to_string (long d)
{
  char b[32];
  ::sprintf (b, "%ld", d); 
  return b;
}

std::string 
tl::to_string (unsigned long d)
{
  char b[32];
  ::sprintf (b, "%lu", d); 
  return b;
}

std::string
tl::to_string (const char *cp)
{
  return std::string (cp);
}

std::string
tl::to_string (const unsigned char *cp)
{
  return std::string ((const char *) cp);
}

std::string
tl::to_string (const char *cp, int length)
{
  return std::string (cp, 0, length);
}

std::string
tl::to_string (const unsigned char *cp, int length)
{
  return std::string ((const char *) cp, 0, length);
}

std::string
tl::to_string (bool b)
{
  return b ? "true" : "false";
}

std::string
tl::to_quoted_string (const std::string &s)
{
  std::string r;
  r.reserve (s.size () + 2);
  r += '\'';
  for (const char *c = s.c_str (); *c; ++c) {
    if (*c == '\'') {
      r += '\\';
    }
    r += *c;
  }
  r += '\'';
  return r;
}

std::string
tl::to_word_or_quoted_string (const std::string &s, const char *non_term)
{
  //  If the string does not contain non_term characters, we may simply keep it. 
  //  Otherwise we need to quote it.
  const char *cp;
  for (cp = s.c_str (); *cp && (isalnum (*cp) || strchr (non_term, *cp) != NULL); ++cp) {
    ;
  }
  if (*cp || s.empty ()) {
    return to_quoted_string (s);
  } else {
    return s;
  }
}

void
tl::from_string (const std::string &s, const char * &result)
{
  result = s.c_str ();
}

void
tl::from_string (const std::string &s, const unsigned char * &result)
{
  result = (unsigned char *) s.c_str ();
}

void 
tl::from_string (const std::string &s, double &v) throw (tl::Exception)
{
  const char *cp = s.c_str ();
  while (*cp && isspace (*cp)) {
    ++cp;
  }
  if (! *cp) {
    throw tl::Exception (tl::translate ("Got empty string where a real number was expected"));
  }
  char *cp_end = const_cast <char *> (cp);
  v = strtod (cp, &cp_end);
  while (*cp_end && isspace (*cp_end)) {
    ++cp_end;
  }
  if (*cp_end) {
    throw tl::Exception (tl::translate ("Not a valid real number: ") + s);
  }
}

template <class T>
void 
convert_string_to_int (const std::string &s, T &v) throw (tl::Exception)
{
  double x;
  tl::from_string (s, x);
  if (x < std::numeric_limits <T>::min ()) {
    throw tl::Exception (tl::translate ("Range underflow: ") + s);
  }
  if (x > std::numeric_limits <T>::max ()) {
    throw tl::Exception (tl::translate ("Range overflow: ") + s);
  }
  v = T (x);
  if (x != v) {
    throw tl::Exception (tl::translate ("Number cannot be represented precisely: ") + s);
  }
}

void 
tl::from_string (const std::string &s, int &v) throw (tl::Exception)
{
  convert_string_to_int (s, v);
}

void 
tl::from_string (const std::string &s, long &v) throw (tl::Exception)
{
  convert_string_to_int (s, v);
}

void 
tl::from_string (const std::string &s, unsigned int &v) throw (tl::Exception)
{
  convert_string_to_int (s, v);
}

void 
tl::from_string (const std::string &s, unsigned long &v) throw (tl::Exception)
{
  convert_string_to_int (s, v);
}

void
tl::from_string (const std::string &s, bool &b) throw (tl::Exception)
{
  std::string t (trim (s));
  if (t == "true") {
    b = true;
  } else if (t == "false") {
    b = false;
  } else if (t == "1") {
    b = true;
  } else if (t == "0") {
    b = false;
  } else {
    throw tl::Exception (tl::translate ("Invalid boolean value: ") + s);
  }
}


std::string 
tl::sprintf (const char *fmt, va_list args)
{
  char b[256];
  int n = vsnprintf (b, sizeof (b), fmt, args);
  if (n >= int (sizeof (b))) {

    //  the fixed buffer size above was not sufficient - 
    //  dynamically allocate enough memory
    char *bp = new char [n + 1];
    vsnprintf (bp, n + 1, fmt, args);
    std::string s (bp);
    delete [] bp;
    return s;

  } else {
    return std::string (b);
  }
}

std::string 
tl::sprintf (const char *fmt, ...)
{
  va_list args;
  va_start (args, fmt);

  char b[256];
  int n = vsnprintf (b, sizeof (b), fmt, args);
  if (n >= int (sizeof (b))) {

    //  the fixed buffer size above was not sufficient - 
    //  dynamically allocate enough memory
    char *bp = new char [n + 1];
    vsnprintf (bp, n + 1, fmt, args);
    std::string s (bp);
    delete [] bp;
    return s;

  } else {
    return std::string (b);
  }
}

std::string 
tl::trim (const std::string &s)
{
  const char *cp = s.c_str ();
  while (isspace (*cp) && *cp) {
    ++cp;
  }

  const char *cq = s.c_str () + s.size ();
  while (cq > cp && isspace (cq [-1])) {
    --cq;
  }

  return std::string (cp, 0, cq - cp);
}

// -------------------------------------------------------------------
//  tl::Extractor implementation

tl::Extractor::Extractor (const char *s)
  : m_cp (s)
{
  //  .. nothing yet ..
}

tl::Extractor & 
tl::Extractor::read (unsigned int &value)
{
  if (! try_read (value)) {
    error (tl::translate ("Expected an unsigned integer value"));
  }
  return *this;
}

tl::Extractor & 
tl::Extractor::read (unsigned long &value)
{
  if (! try_read (value)) {
    error (tl::translate ("Expected an unsigned long integer value"));
  }
  return *this;
}

tl::Extractor & 
tl::Extractor::read (double &value)
{
  if (! try_read (value)) {
    error (tl::translate ("Expected a real number"));
  }
  return *this;
}

tl::Extractor & 
tl::Extractor::read (int &value)
{
  if (! try_read (value)) {
    error (tl::translate ("Expected a integer value"));
  }
  return *this;
}

tl::Extractor & 
tl::Extractor::read (long &value)
{
  if (! try_read (value)) {
    error (tl::translate ("Expected a long integer value"));
  }
  return *this;
}

tl::Extractor & 
tl::Extractor::read (bool &value)
{
  if (! try_read (value)) {
    error (tl::translate ("Expected a boolean value ('true', 'false')"));
  }
  return *this;
}

tl::Extractor & 
tl::Extractor::read (std::string &value, const char *term)
{
  if (! try_read (value, term)) {
    error (tl::translate ("Expected a string"));
  }
  return *this;
}

tl::Extractor & 
tl::Extractor::read_word (std::string &value, const char *non_term)
{
  if (! try_read_word (value, non_term)) {
    error (tl::translate ("Expected a word string"));
  }
  return *this;
}

tl::Extractor & 
tl::Extractor::read_word_or_quoted (std::string &value, const char *non_term)
{
  if (! try_read_word (value, non_term) && ! try_read_quoted (value)) {
    error (tl::translate ("Expected a word or quoted string"));
  }
  return *this;
}

tl::Extractor & 
tl::Extractor::read_quoted (std::string &value)
{
  if (! try_read_quoted (value)) {
    error (tl::translate ("Expected a quoted string"));
  }
  return *this;
}

bool 
tl::Extractor::try_read (unsigned int &value)
{
  if (! *skip ()) {
    return false;
  } 

  if (! isdigit (*m_cp)) {
    return false;
  } 

  value = 0;
  while (isdigit (*m_cp)) {
    if ((value * 10) / 10 != value) {
      throw tl::Exception (tl::translate ("Range overflow on unsigned integer"));
    }
    value *= 10;
    value += (*m_cp - '0');
    ++m_cp;
  }

  return true;
}

bool 
tl::Extractor::try_read (int &value)
{
  if (! *skip ()) {
    return false;
  } 

  bool minus = false;
  if (*m_cp == '-') {
    minus = true;
    ++m_cp;
  } else if (*m_cp == '+') {
    ++m_cp;
  }

  if (! isdigit (*m_cp)) {
    return false;
  } 

  value = 0;
  while (isdigit (*m_cp)) {
    if ((value * 10) / 10 != value) {
      throw tl::Exception (tl::translate ("Range overflow on integer"));
    }
    value *= 10;
    value += (*m_cp - '0');
    ++m_cp;
  }

  if (minus) {
    value = -value;
  }

  return true;
}

bool 
tl::Extractor::try_read (unsigned long &value)
{
  if (! *skip ()) {
    return false;
  } 

  if (! isdigit (*m_cp)) {
    return false;
  } 

  value = 0;
  while (isdigit (*m_cp)) {
    if ((value * 10) / 10 != value) {
      throw tl::Exception (tl::translate ("Range overflow on unsigned long integer"));
    }
    value *= 10;
    value += (*m_cp - '0');
    ++m_cp;
  }

  return true;
}

bool 
tl::Extractor::try_read (long &value)
{
  if (! *skip ()) {
    return false;
  } 

  bool minus = false;
  if (*m_cp == '-') {
    minus = true;
    ++m_cp;
  } else if (*m_cp == '+') {
    ++m_cp;
  }

  if (! isdigit (*m_cp)) {
    return false;
  } 

  value = 0;
  while (isdigit (*m_cp)) {
    if ((value * 10) / 10 != value) {
      throw tl::Exception (tl::translate ("Range overflow on long integer"));
    }
    value *= 10;
    value += (*m_cp - '0');
    ++m_cp;
  }

  if (minus) {
    value = -value;
  }

  return true;
}

bool 
tl::Extractor::try_read (double &value)
{
  if (! *skip ()) {
    return false;
  } 

  char *cp_end = const_cast <char *> (m_cp);
  value = strtod (m_cp, &cp_end);

  if (cp_end == m_cp) {
    return false;
  } else {
    m_cp = cp_end;
    return true;
  }
}

bool 
tl::Extractor::try_read (bool &value)
{
  if (test ("0") || test ("false")) {
    value = false;
    return true;
  }
  if (test ("1") || test ("true")) {
    value = true;
    return true;
  }
  return false;
}

bool 
tl::Extractor::try_read_word (std::string &string, const char *non_term)
{
  if (! *skip ()) {
    return false;
  } 

  string.clear ();
  while (*m_cp && (isalnum (*m_cp) || strchr (non_term, *m_cp) != NULL)) {
    string += *m_cp;
    ++m_cp;
  }
  return ! string.empty ();
}

bool 
tl::Extractor::try_read_word_or_quoted (std::string &string, const char *non_term)
{
  return try_read_word (string, non_term) || try_read_quoted (string);
}

bool 
tl::Extractor::try_read_quoted (std::string &string)
{
  char q = *skip ();

  if (q != '\'' && q != '\"') {
    return false;
  }

  ++m_cp;
  string.clear ();
  while (*m_cp && *m_cp != q) {
    if (*m_cp == '\\' && m_cp[1]) {
      ++m_cp;
    }
    string += *m_cp;
    ++m_cp;
  }
  if (*m_cp == q) {
    ++m_cp;
  }
  return true;
}

bool 
tl::Extractor::try_read (std::string &string, const char *term)
{
  if (! *skip ()) {
    return false;
  } 

  string.clear ();
  while (*m_cp && ! isspace (*m_cp) && strchr (term, *m_cp) == NULL) {
    string += *m_cp;
    ++m_cp;
  }
  return true;
}

tl::Extractor & 
tl::Extractor::expect_end ()
{
  if (! at_end ()) {
    error (tl::translate ("Expected end of text"));
  }
  return *this;
}

tl::Extractor & 
tl::Extractor::expect_more ()
{
  if (at_end ()) {
    error (tl::translate ("Expected more text"));
  }
  return *this;
}

tl::Extractor & 
tl::Extractor::expect (const char *token)
{
  if (! test (token)) {
    error (tl::sprintf (tl::translate ("Expected '%s'").c_str (), token));
  }
  return *this;
}

bool 
tl::Extractor::test (const char *token)
{
  skip ();

  const char *cp = m_cp;
  while (*cp && *token) {
    if (*cp != *token) {
      return false;
    }
    ++cp;
    ++token;
  }
  
  if (! *token) {
    m_cp = cp;
    return true;
  } else {
    return false;
  }
}

const char *
tl::Extractor::skip ()
{
  while (isspace (*m_cp) && *m_cp) {
    ++m_cp;
  }
  return m_cp;
}

void
tl::Extractor::error (const std::string &msg)
{
  std::string m (msg);

  if (at_end ()) {
    m += tl::translate (", but text ended");
  } else {
    m += tl::translate (" here: ");
    for (unsigned int i = 0; i < 10 && *m_cp; ++i, ++m_cp) {
      m += *m_cp;
    }
    if (*m_cp) {
      m += " ..";
    }
  }

  throw tl::Exception (m);
}

// -------------------------------------------------------------------
//  tl::string implementation

tl::string::string (const char *c)
{
  if (c && *c) {
    m_capacity = m_size = strlen (c);
    allocator_t alloc;
    mp_rep = alloc.allocate (m_capacity + 1);
    strcpy (mp_rep, c);
  } else {
    mp_rep = 0;
    m_capacity = m_size = 0;
  }
}

tl::string::string (const char *c, size_t from, size_t to)
{
  m_capacity = m_size = to - from;
  if (m_size > 0) {
    allocator_t alloc;
    mp_rep = alloc.allocate (m_capacity + 1);
    strncpy (mp_rep, c + from, m_size);
    mp_rep [m_size] = 0;
  } else {
    mp_rep = 0;
  }
}

tl::string::string (const tl::string &s)
{
  m_capacity = m_size = s.size ();
  if (m_size > 0) {
    allocator_t alloc;
    mp_rep = alloc.allocate (m_capacity + 1);
    strncpy (mp_rep, s.c_str (), m_size);
    mp_rep [m_size] = 0;
  } else {
    mp_rep = 0;
  }
}

tl::string::string (const tl::string &s, size_t from, size_t to)
{
  m_capacity = m_size = to - from;
  if (m_size > 0) {
    allocator_t alloc;
    mp_rep = alloc.allocate (m_capacity + 1);
    strncpy (mp_rep, s.c_str () + from, m_size);
    mp_rep [m_size] = 0;
  } else {
    mp_rep = 0;
  }
}

tl::string::string (const std::string &s)
{
  m_capacity = m_size = s.size ();
  if (m_size > 0) {
    allocator_t alloc;
    mp_rep = alloc.allocate (m_capacity + 1);
    strncpy (mp_rep, s.c_str (), m_size);
    mp_rep [m_size] = 0;
  } else {
    mp_rep = 0;
  }
}

tl::string::string (const std::string &s, size_t from, size_t to)
{
  m_capacity = m_size = to - from;
  if (m_size > 0) {
    allocator_t alloc;
    mp_rep = alloc.allocate (m_capacity + 1);
    strncpy (mp_rep, s.c_str () + from, m_size);
    mp_rep [m_size] = 0;
  } else {
    mp_rep = 0;
  }
}

tl::string::~string ()
{
  if (mp_rep) {
    allocator_t alloc;
    alloc.deallocate (mp_rep, m_capacity + 1);
  }
  mp_rep = 0;
}

tl::string &
tl::string::operator= (const char *c)
{
  if (c && *c) {
    assign (c, 0, strlen (c));
  } else {
    m_size = 0;
    if (mp_rep) {
      mp_rep [0] = 0;
    }
  }
  return *this;
}

void 
tl::string::assign (const char *c, size_t from, size_t to)
{
  m_size = to - from;
  if (m_size > 0) {
    if (m_capacity < m_size) {
      allocator_t alloc;
      if (mp_rep) {
        alloc.deallocate (mp_rep, m_capacity + 1);
      }
      mp_rep = alloc.allocate (m_size + 1);
      m_capacity = m_size;
    }
    strncpy (mp_rep, c + from, m_size);
    mp_rep [m_size] = 0;
  } else {
    if (mp_rep) {
      mp_rep [0] = 0;
    }
  }
}

tl::string &
tl::string::operator= (const tl::string &s)
{
  if (&s != this) {
    m_size = s.size ();
    if (m_size > 0) {
      if (m_capacity < m_size) {
        allocator_t alloc;
        if (mp_rep) {
          alloc.deallocate (mp_rep, m_capacity + 1);
        }
        mp_rep = alloc.allocate (m_size + 1);
        m_capacity = m_size;
      }
      strncpy (mp_rep, s.mp_rep, m_size);
      mp_rep [m_size] = 0;
    } else {
      if (mp_rep) {
        mp_rep [0] = 0;
      }
    }
  }
  return *this;
}

void 
tl::string::assign (const tl::string &s, size_t from, size_t to)
{
  if (&s != this) {
    assign (s.c_str (), from, to);
  } else if (from != 0 || to != m_size) {
    tl::string substr (s, from, to);
    swap (substr);
  }
}

tl::string &
tl::string::operator= (const std::string &s)
{
  assign (s.c_str (), 0, s.size ());
  return *this;
}

void 
tl::string::assign (const std::string &s, size_t from, size_t to)
{
  assign (s.c_str (), from, to);
}

void
tl::string::clear ()
{
  if (mp_rep) {
    allocator_t alloc;
    alloc.deallocate (mp_rep, m_capacity + 1);
    mp_rep = 0;
  }
  m_size = 0;
  m_capacity = 0;
}

void 
tl::string::reserve (size_t n) 
{
  if (m_capacity < n) {
    allocator_t alloc;
    char *nrep = alloc.allocate (n + 1);
    strncpy (nrep, mp_rep, m_size);
    if (mp_rep) {
      alloc.deallocate (mp_rep, m_capacity + 1);
    }
    mp_rep = nrep;
    m_capacity = n; 
  }
}

bool 
tl::string::operator== (const char *c) const
{
  return (c[0] == c_str()[0] && strcmp (c, c_str()) == 0);
}

bool 
tl::string::operator== (const tl::string &s) const
{
  return (c_str()[0] == s.c_str()[0] && strcmp (c_str(), s.c_str()) == 0);
}

bool 
tl::string::operator!= (const char *c) const
{
  return (c[0] != c_str()[0] || strcmp (c, c_str()) != 0);
}

bool 
tl::string::operator!= (const tl::string &s) const
{
  return (c_str()[0] != s.c_str()[0] || strcmp (c_str(), s.c_str()) != 0);
}

bool 
tl::string::operator< (const char *c) const
{
  return strcmp (c_str(), c) < 0;
}
 
bool 
tl::string::operator< (const tl::string &s) const
{
  return strcmp (c_str(), s.c_str()) < 0;
}

bool 
tl::string::operator<= (const char *c) const
{
  return strcmp (c_str(), c) <= 0;
}
 
bool 
tl::string::operator<= (const tl::string &s) const
{
  return strcmp (c_str(), s.c_str()) <= 0;
}

bool 
tl::string::operator> (const char *c) const
{
  return strcmp (c_str(), c) > 0;
}
 
bool 
tl::string::operator> (const tl::string &s) const
{
  return strcmp (c_str(), s.c_str()) > 0;
}

bool 
tl::string::operator>= (const char *c) const
{
  return strcmp (c_str(), c) >= 0;
}
 
bool 
tl::string::operator>= (const tl::string &s) const
{
  return strcmp (c_str(), s.c_str()) >= 0;
}

