
/*

  KLayout Layout Viewer
  Copyright (C) 2006-2007 Matthias Koefferlein

  This program is free software; you can redistribute it and/or modify
  it under the terms of the GNU General Public License as published by
  the Free Software Foundation; either version 2 of the License, or
  (at your option) any later version.

  This program is distributed in the hope that it will be useful,
  but WITHOUT ANY WARRANTY; without even the implied warranty of
  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
  GNU General Public License for more details.

  You should have received a copy of the GNU General Public License
  along with this program; if not, write to the Free Software
  Foundation, Inc., 51 Franklin St, Fifth Floor, Boston, MA  02110-1301  USA

*/


#ifndef HDR_utHead
#define HDR_utHead

#include <vector>
#include <iostream>
#include <sstream>
#include <string.h>

#include <tlString.h>
#include <tlException.h>

namespace ut {

struct Exception 
  : public tl::Exception
{
  Exception (const std::string &msg)
    : tl::Exception (msg)
  { }
};

struct TestBase 
{
  TestBase (const std::string &test);
  virtual ~TestBase () { }

  bool do_test () const
  {
    std::cout << "Running " << m_test << std::endl;
    try {
      execute ();
      return true;
    } catch (tl::Exception &ex) {
      std::cerr << "Test " << m_test << " failed: " << std::endl
                << ex.msg () << std::endl;
      return false;
    }
  }

private:
  virtual void execute () const throw (tl::Exception) = 0;
  std::string m_test;
};

struct Registrar
{
  static void reg (const ut::TestBase *t) 
  {
    if (! ms_instance) {
      ms_instance = new Registrar ();
    }
    ms_instance->m_tests.push_back (t);
  }

  static int do_tests () 
  {
    int failed = 0;
    if (ms_instance) {
      for (std::vector <const ut::TestBase *>::const_iterator 
           t = ms_instance->m_tests.begin (); 
           t != ms_instance->m_tests.end (); ++t) {
        if (! (*t)->do_test ()) {
          ++failed;
        }
      }
    }
    return failed;
  }
  
private:
  static Registrar *ms_instance;

  Registrar () : m_tests () { }

  std::vector <const ut::TestBase *> m_tests;
};

ut::Registrar *ut::Registrar::ms_instance = 0;

TestBase::TestBase (const std::string &test)
  : m_test (test)
{
  ut::Registrar::reg (this);
}

static bool _verbose_flag = false;

bool verbose ()
{
  return _verbose_flag;
}

} // namespace ut


int 
main (int argc, const char *argv[]) 
{
  if (argc == 2 && ! strcmp (argv [1], "-v")) {
    ut::_verbose_flag = true;
  }
  int failed = ut::Registrar::do_tests ();
  if (failed > 0) {
    std::cerr << "Summary: " << failed << " test(s) failed" << std::endl;
  } else {
    std::cout << "All tests passed." << std::endl;
  }
  return failed;
}


#define TEST(NAME) \
struct TestImpl_##NAME \
  : public ut::TestBase \
{ \
  TestImpl_##NAME () : TestBase (__FILE__ ":" #NAME) { } \
  virtual void execute () const throw (tl::Exception); \
}; \
static TestImpl_##NAME TestImpl_Inst_##NAME; \
void TestImpl_##NAME::execute () const throw (tl::Exception)

#define EXPECT_EQ(WHAT,EQUALS) \
  if (!((WHAT) == (EQUALS))) { \
    std::stringstream sstr; \
    sstr << __LINE__ << ": " << #WHAT << " does not equal " << #EQUALS << " (actual value is " << (WHAT) << ")" << std::ends; \
    throw ut::Exception (sstr.str ()); \
  } 

#define EXPECT_NE(WHAT,EQUALS) \
  if (!((WHAT) != (EQUALS))) { \
    std::stringstream sstr; \
    sstr << __LINE__ << ": " << #WHAT << " does equal " << #EQUALS << " (actual value is " << (WHAT) << ")" << std::ends; \
    throw ut::Exception (sstr.str ()); \
  } 

#define EXPECT(WHAT) \
  if (!(WHAT)) { throw ut::Exception (__LINE__ << ": " << #WHAT " is not true"); } 

#define FAIL_ARG(MSG,WHAT) \
  { \
    std::stringstream sstr; \
    sstr << __LINE__ << ": " << MSG << ", value is " << (WHAT) << std::ends; \
    throw ut::Exception (sstr.str ()); \
  } 


#endif

